/*
 * bootinfo.c 
 *
 * Copyright (C) 2008 Borqs Ltd.
 * Author:  Ru Yi <yi.ru@borqs.com>
 *
 * This file is designed for Borqs linux phone platform collecting boot info 
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 *
 * Revision History:
 *                    Modification      Tracking
 * Author                 Date          Description of Changes
 * ----------------   ------------    -------------------------------
 * Leo He              Nov 9, 2009      Boot info refactoring
 *
 */

#include <asm/bootinfo.h>
#include <asm/setup.h>
#include <linux/module.h>
#include <linux/notifier.h>
#include <linux/reboot.h>

/*
 * All codes surrounded by "CONFIG_BOOT_INFO_SYS_INTERFACE" used for 
 * enable the sys interface for boot information.
 */

#ifdef  CONFIG_BOOT_INFO_SYS_INTERFACE
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/fs.h>
#include <linux/string.h>
#include <linux/sysdev.h>

static ssize_t powerup_reason_show(struct kobject *kobj,
			struct kobj_attribute *attr, char *buf)
{
	char *s = buf;
	const char* const __bootmode[BOOT_MODE_MAX] = BOOT_MODE_LIST;
	boot_mode_t bootmode = get_boot_mode();

	s += sprintf(s, "%s", __bootmode[bootmode]);

	return (s - buf);
}

static ssize_t powerup_reason_store(struct kobject *kobj,
			struct kobj_attribute *attr, const char *buf, size_t n)
{
	/* Do nothing here. */
	return n;
}

static ssize_t reboot_setting_show(struct kobject *kobj,
			struct kobj_attribute *attr, char *buf)
{
	char *s = buf;
	int i = 0;
	const char* const __bootreason[BOOT_REASON_MAX] = BOOT_REASON_LIST;
	boot_reason_t bootreason = get_boot_reason();

	for (; i < BOOT_REASON_MAX; ++i) {
		s += sprintf(s, "%s", __bootreason[i]);
		if (i == bootreason)
			s += sprintf(s, " [Active]");
		s += sprintf(s, "\n");
	}

	return (s - buf);
}

static ssize_t reboot_setting_store(struct kobject *kobj,
			struct kobj_attribute *attr, const char *buf, size_t n)
{
	char *p  = memchr(buf, '\n', n);
	int  len = (p == NULL) ? n : p - buf;  
	int  br  = BOOT_REASON_NORMAL;
	const char* const __bootreason[BOOT_REASON_MAX] = BOOT_REASON_LIST;

	for (; br < BOOT_REASON_MAX; ++br) {
		if (len == strlen(__bootreason[br])
				&& !strncmp(buf, __bootreason[br], len)) {
			set_boot_reason(br);
			return n;
		}
	}

	printk("%s(): boot reason invalid!", __func__);
	return -EINVAL;
}

static ssize_t boot_flex_show(struct kobject *kobj,
			struct kobj_attribute *attr, char *buf)
{
	char *s = buf;
	boot_flex_t bf = get_boot_flex();
	const char* const __bootflex[BOOT_FLEX_MAX] = BOOT_FLEX_LIST;

	s += sprintf(s, "boot flex: ");
	if (bf >= BOOT_FLEX_USBNET && bf < BOOT_FLEX_MAX)
		s += sprintf(s, "%s ", __bootflex[bf]);

	return (s - buf);
}

static ssize_t boot_flex_store(struct kobject *kobj,
			struct kobj_attribute *attr, const char *buf, size_t n)
{
	/* Do nothing here. */
	return n;
}

static ssize_t debug_info_show(struct kobject *kobj,
			struct kobj_attribute *attr, char *buf)
{
	char *s = buf;
	// TODO: show debug info
	return (s - buf);
}

static ssize_t debug_info_store(struct kobject *kobj,
			struct kobj_attribute *attr, const char * buf, size_t n)
{
	/* Do nothing here.*/
	return n;
}

#define create_kobj_attribute(__name)                          \
static struct kobj_attribute __name##_attr = {                 \
	.attr  = {.name = __stringify(__name), .mode = 0644,}, \
	.show  = __name##_show,                                \
	.store = __name##_store,                               \
}

create_kobj_attribute(powerup_reason);
create_kobj_attribute(reboot_setting);
create_kobj_attribute(boot_flex);
create_kobj_attribute(debug_info);

static struct attribute * __attrs[] = {
	&powerup_reason_attr.attr,
	&reboot_setting_attr.attr,
	&boot_flex_attr.attr,
	&debug_info_attr.attr,
	NULL,
};
static struct attribute_group attr_group    = { .attrs = __attrs, };
static struct kobject*        bootinfo_kobj = NULL;

static inline int __bootinfo_init(void)
{
	int ret = -ENOMEM;

	bootinfo_kobj = kobject_create_and_add("bootinfo", NULL);
	if (bootinfo_kobj == NULL) {
		printk("bootinfo_init: subsystem_register failed\n");
		return ret;
	}

	ret = sysfs_create_group(bootinfo_kobj, &attr_group);
	if (ret) {
		printk("bootinfo_init: subsystem_register failed\n");
		kobject_del(bootinfo_kobj);
		return ret;
	}

	return ret;
}

static inline void __bootinfo_exit(void)
{
	if (bootinfo_kobj != NULL) {
		sysfs_remove_group(bootinfo_kobj, &attr_group);
		kobject_del(bootinfo_kobj);
	}

	return;
}
#endif /* END OF CONFIG_BOOT_INFO_SYS_INTERFACE */

static boot_mode_t   bootmode   = BOOT_MODE_NORMAL;
static boot_reason_t bootreason = BOOT_REASON_NORMAL;
static boot_flex_t   bootflex   = BOOT_FLEX_MAX;

static int __init __setup_boot_mode(char *str)
{
	int i = 0, ret = 0;
	const char* const __bootmode[BOOT_MODE_MAX] = BOOT_MODE_LIST;

	for(i = 0; i < BOOT_MODE_MAX; i++) {
		if (strcmp(str, __bootmode[i]) == 0) {
			bootmode = (boot_mode_t)i;
			ret = 1;
			break;
		}
	}

	if (ret == 1) {
		printk("BOOT MODE FOUND: \"%s\"\n", __bootmode[bootmode]);
	} else {
		printk("BOOT MODE MISSED, default: \"%s\"\n", __bootmode[BOOT_MODE_NORMAL]);
	}

	return ret;
}
__setup(BOOT_MODE_TAG, __setup_boot_mode);

static int __init __setup_boot_flex(char *str)
{
	int i = 0;
	const char* const __bootflex[BOOT_FLEX_MAX] = BOOT_FLEX_LIST;

	printk("BOOT FLEX FOUND: \"%s\"\n", str);
	for (i = 0; i < BOOT_FLEX_MAX; ++i) {
		if (strcmp(str, __bootflex[i]) == 0)
			bootflex = (boot_flex_t) i;
	}

	return 0;
}
__setup(BOOT_FLEX_TAG, __setup_boot_flex);

boot_mode_t get_boot_mode(void)
{
	return bootmode;
}
EXPORT_SYMBOL(get_boot_mode);

boot_flex_t get_boot_flex(void)
{
	return bootflex;
}
EXPORT_SYMBOL(get_boot_flex);

int set_boot_reason(boot_reason_t br)
{
	int ret = 0;

	if (br >= BOOT_REASON_NORMAL && br < BOOT_REASON_MAX) {
		bootreason = br;
		ret = 1;
	} else {
		bootreason = BOOT_REASON_NORMAL;
		ret = 0;
		printk("Boot reason invalid(%d), use NORMAL as default!\n", br);
	}

	return ret;
}
EXPORT_SYMBOL(set_boot_reason);

boot_reason_t get_boot_reason(void)
{
	const char* const __bootreason[BOOT_REASON_MAX] = BOOT_REASON_LIST;
	printk("%s: apps boot reason[%s]\n", __func__, __bootreason[bootreason]);
	return bootreason;
}
EXPORT_SYMBOL(get_boot_reason);

static int __set_boot_reason(char* str)
{
	int i = 0;
	const char* const __bootreason[BOOT_REASON_MAX] = BOOT_REASON_LIST;

	if (str != NULL) {
		for(i = 0; i < BOOT_REASON_MAX; i++) {
			if (strcmp(str, __bootreason[i]) == 0) {
				set_boot_reason((boot_reason_t)i);
				break;
			}
		}
	}

	printk(" BOOT REASON: %s\n", __bootreason[bootreason]);
	return 1;
}

static int my_panic_event(struct notifier_block *this,
		unsigned long code, void *_cmd)
{
	set_boot_reason(BOOT_REASON_CRITICAL);
	return NOTIFY_DONE;
}

static int my_reboot_event (struct notifier_block *this,
		unsigned long code, void *_cmd) {
	if (code == SYS_RESTART) 
		__set_boot_reason(_cmd);
	return NOTIFY_DONE;
}

static struct notifier_block my_panic_block = {
	.notifier_call = my_panic_event,
};

static struct notifier_block my_reboot_block = {
	.notifier_call = my_reboot_event,
};

static int __init bootinfo_init(void)
{
	int ret = -ENOMEM;
	ret = register_reboot_notifier(&my_reboot_block);
	if (ret) {
		printk("%s: Fail to register application reboot notifier!\n", __func__);
	}

	ret = atomic_notifier_chain_register(&panic_notifier_list,
					&my_panic_block);
	if (ret) {
		printk("%s: Fail to register kernel panic notifier!\n", __func__);
	}
#ifdef CONFIG_BOOT_INFO_SYS_INTERFACE
	if (__bootinfo_init())
		printk("%s: __bootinfo_init() failed!\n", __func__);
#endif
	return ret;
}

static void __exit bootinfo_exit(void)
{
#ifdef CONFIG_BOOT_INFO_SYS_INTERFACE
	__bootinfo_exit();
#endif
	atomic_notifier_chain_unregister(&panic_notifier_list,
					&my_panic_block);
}

module_init(bootinfo_init);
module_exit(bootinfo_exit);

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Ru Yi <yi.ru@borqs.com>");
MODULE_DESCRIPTION("Boot information collector");

