/*
 *  arch/arm/mach-pxa/include/mach/memory.h
 *
 * Author:	Nicolas Pitre
 * Copyright:	(C) 2001 MontaVista Software Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#ifndef __ASM_ARCH_MEMORY_H
#define __ASM_ARCH_MEMORY_H

/* Override the ARM default */
#define CONSISTENT_DMA_SIZE	10 * 1024 * 1024

/*	The common memory map definitions for both FLATMEM and DISCONTIGMEM.
	RAM may have one or two banks:
	bank0: 0x80000000
	bank1: 0xc0000000
	The upper 16MB of RAM is assigned for CP use,
	and SHMEM driver accesses this space (ioremap's it)
	FLATMEM option does not allow use or "HW" area in the valid pfn range,
	so panic happens with two bank RAM: CP area appears INSIDE this range.
	To resolve this issue, bank0 physical address of 0xa0000000 is used,
	which is a HW alias to 0x80000000, assuming bank size is upto 512MB.
	The CP area for SHMEM uses physical address 0x9f000000,
	which is a HW alias to the top 16MB of bank0, and is located
	outside the valid pfn range [0xa0000000..0xc0000000+banksize].
	This works under the same assumption: bank size upto 512MB.
	Example of HW aliased physical addresses with RAM bank size up to 512MB:
	bank0 start: 		0x80000000, 0xa0000000
	bank0 last 16MB: 	0x9f000000, 0xbf000000
	bank1 start: 		0xc0000000, 0xe0000000
*/
/*
 * Physical DRAM offset.
 */
#define PHYS_OFFSET	UL(0xa0000000)

/*
 * The nodes are matched with the physical SDRAM banks as follows:
 *
 * 	node 0:  0xa0000000-0xbfffffff	-->  0xc0000000-0xdfffffff
 * 	node 1:  0xc0000000-0xcfffffff	-->  0xe0000000-0xefffffff
 *
 * This needs a node mem size of 29 bits, support 512M bank node.
 * Because the memory map limits, the Node 1 just support up to 
 * 256M bank node.
 */
#define __virt_to_phys(x)	((x) - PAGE_OFFSET + PHYS_OFFSET)
#define __phys_to_virt(x)	((x) - PHYS_OFFSET + PAGE_OFFSET)

#ifdef CONFIG_DISCONTIGMEM
/*
 * The below definitions are relevant for CONFIG_DISCONTIGMEM only.
 * Do not define NODE_MEM_SIZE_BITS to override defaults derived from it
 * in arch/arm/include/asm/memory.h.
 * Reason: PFN_TO_NID is different: 29 bits physical offset between nodes
 */
#define NODE_MEM_SIZE_MASK	((1 << 29) - 1)

/*
 * Given a kernel address, find the home node of the underlying memory.
 */
#define KVADDR_TO_NID(addr) \
	(((unsigned long)(addr) - PAGE_OFFSET) >> 29)

/*
 * Given a page frame number, convert it to a node id.
 */
#define PFN_TO_NID(pfn) \
	(((pfn) - PHYS_PFN_OFFSET) >> (29 - PAGE_SHIFT))

/*
 * Given a kaddr, LOCAL_MEM_MAP finds the owning node of the memory
 * and returns the index corresponding to the appropriate page in the
 * node's mem_map.
 */
#define LOCAL_MAP_NR(addr) \
	(((unsigned long)(addr) & NODE_MEM_SIZE_MASK) >> PAGE_SHIFT)

#endif

#if !defined(__ASSEMBLY__) && defined(CONFIG_PXA3xx_MVED)
void mved_adjust_zones(int node, unsigned long *zone_size,
			unsigned long *zhole_size);
#define arch_adjust_zones(node, size, holes) \
	mved_adjust_zones(node, size, holes)

#define ISA_DMA_THRESHOLD	(PHYS_OFFSET + SZ_256M - 1)
#endif

#if !defined(__ASSEMBLY__) && defined(CONFIG_MACH_ARMCORE) && defined(CONFIG_PCI)
void cmx2xx_pci_adjust_zones(int node, unsigned long *size,
			     unsigned long *holes);

#define arch_adjust_zones(node, size, holes) \
	cmx2xx_pci_adjust_zones(node, size, holes)

#define ISA_DMA_THRESHOLD	(PHYS_OFFSET + SZ_64M - 1)
#define MAX_DMA_ADDRESS		(PAGE_OFFSET + SZ_64M)
#endif

#endif
