/* 
 * arch/arm/mach-pxa/modem-beacon.c
 *
 * Support for the BEACON's modem driver.
 *
 * Copyright (C) 2009 Borqs, Inc.
 * Author: Emichael Li <emichael.li@borqs.com>
 *
 * This software is licensed under the terms of the GNU General Public
 * License version 2, as published by the Free Software Foundation, and
 * may be copied, distributed, and modified under those terms.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/platform_device.h>
#include <linux/delay.h>
#include <linux/modem.h>
#include <asm/arch/gpio.h>
#include <asm/arch/cherry.h>
#include "devices.h"

#undef MODEM_DEBUG

#ifdef MODEM_DEBUG
#define DENTER(fmt,args...)                                                   \
        do {                                                                  \
		printk(KERN_ERR "[BP] Enter %s: " fmt,                	      \
			__FUNCTION__ , ## args);                              \
        } while (0)
#define DLEAVE(fmt,args...)                                                   \
        do {                                                                  \
		printk(KERN_ERR "[BP] Leave %s: " fmt,                        \
			__FUNCTION__ , ## args);              		      \
        } while (0)
#else
#define DENTER(fmt,args...)                                                   \
        do {} while (0)
#define DLEAVE(fmt,args...)                                                   \
        do {} while (0)
#endif

#define AP_POWER_MODEM_GPIO		(mfp_to_gpio(MFP_CFG_PIN(AP_POWER_MODEM)))
#define AP_RTS_GPIO 			(mfp_to_gpio(MFP_CFG_PIN(AP_RTS)))
#define AP_RDY_GPIO 			(mfp_to_gpio(MFP_CFG_PIN(AP_RDY)))
#define AP_WAKEUP_GPIO			(mfp_to_gpio(MFP_CFG_PIN(AP_WAKEUP_MODEM)))
#define AP_SLEEP_GPIO			(mfp_to_gpio(MFP_CFG_PIN(AP_SLEEP_STS)))

#define BP_RTS_GPIO 			(mfp_to_gpio(MFP_CFG_PIN(MODEM_RTS)))
#define BP_RDY_GPIO 			(mfp_to_gpio(MFP_CFG_PIN(MODEM_RDY)))
#define BP_WAKEUP_GPIO			(mfp_to_gpio(MFP_CFG_PIN(MODEM_WAKEUP_AP)))
#define BP_SLEEP_GPIO 			(mfp_to_gpio(MFP_CFG_PIN(MODEM_SLEEP_STS)))
#define BP_LIFE_GPIO 			(mfp_to_gpio(MFP_CFG_PIN(MODEM_LIFE_STS)))

#define AP_BB_CON_DOWNLOAD02_GPIO	(mfp_to_gpio(MFP_CFG_PIN(AP_BB_CON_DOWNLOAD02)))

#define SPI_CLK_GPIO			(mfp_to_gpio(MFP_CFG_PIN(SSP3_MODEM_SPICLK)))
#define SPI_FRM_GPIO			(mfp_to_gpio(MFP_CFG_PIN(SSP3_SPI2_CSN1)))
#define SPI_TXD_GPIO			(mfp_to_gpio(MFP_CFG_PIN(SSP3_MODEM_SPIDI)))
#define SPI_RXD_GPIO			(mfp_to_gpio(MFP_CFG_PIN(SSP3_MODEM_SPIDO)))

extern void pxa_set_modem_info(struct modem_platform_data *mdata);

static void set_modem_off_state(void)
{
	gpio_direction_output(AP_POWER_MODEM_GPIO, 0);
	pxa3xx_mfp_set_lpm(gpio_to_mfp(AP_POWER_MODEM_GPIO), MFP_LPM_DRIVE_LOW);

	gpio_direction_input(AP_BB_CON_DOWNLOAD02_GPIO);

	pxa3xx_mfp_set_afds(SPI_CLK_GPIO, MFP_AF0, MFP_DS01X);
	pxa3xx_mfp_set_afds(SPI_FRM_GPIO, MFP_AF0, MFP_DS01X);
	pxa3xx_mfp_set_afds(SPI_TXD_GPIO, MFP_AF0, MFP_DS01X);
	pxa3xx_mfp_set_afds(SPI_RXD_GPIO, MFP_AF0, MFP_DS01X);

	gpio_direction_input(SPI_CLK_GPIO);
	gpio_direction_input(SPI_FRM_GPIO);
	gpio_direction_input(SPI_TXD_GPIO);
	gpio_direction_input(SPI_RXD_GPIO);

	gpio_direction_input(AP_RTS_GPIO);
	gpio_direction_input(AP_RDY_GPIO);
	gpio_direction_input(AP_WAKEUP_GPIO);
	gpio_direction_input(AP_SLEEP_GPIO);

	gpio_direction_input(BP_RTS_GPIO);
	gpio_direction_input(BP_RDY_GPIO);
	gpio_direction_input(BP_WAKEUP_GPIO);
	gpio_direction_input(BP_SLEEP_GPIO);
	gpio_direction_input(BP_LIFE_GPIO);
}

static void set_modem_booting_state(void)
{
	gpio_direction_output(AP_BB_CON_DOWNLOAD02_GPIO, 0);
	gpio_direction_output(AP_POWER_MODEM_GPIO, 1);
	pxa3xx_mfp_set_lpm(gpio_to_mfp(AP_POWER_MODEM_GPIO), MFP_LPM_DRIVE_HIGH);
}

static void set_modem_on_state(void)
{
	pxa3xx_mfp_set_afds(SPI_CLK_GPIO, MFP_AF1, MFP_DS08X);
	pxa3xx_mfp_set_afds(SPI_FRM_GPIO, MFP_AF1, MFP_DS08X);
	pxa3xx_mfp_set_afds(SPI_TXD_GPIO, MFP_AF1, MFP_DS08X);
	pxa3xx_mfp_set_afds(SPI_RXD_GPIO, MFP_AF1, MFP_DS08X);

	gpio_direction_output(AP_RTS_GPIO, 1);
	gpio_direction_output(AP_RDY_GPIO, 1);
	gpio_direction_output(AP_WAKEUP_GPIO, 1);
	gpio_direction_output(AP_SLEEP_GPIO, 0);

	gpio_direction_input(BP_RTS_GPIO);
	gpio_direction_input(BP_RDY_GPIO);
	gpio_direction_input(BP_WAKEUP_GPIO);
	gpio_direction_input(BP_SLEEP_GPIO);
	gpio_direction_input(BP_LIFE_GPIO);
}

int cherry_set_power_state(enum modem_power_state state)
{
	DENTER("on=%d\n", on);

	switch (state) {
	case MPS_OFF:
		set_modem_off_state();
		break;
	case MPS_BOOTING:
		set_modem_booting_state();
		break;
	case MPS_ON:
		set_modem_on_state();
		break;
	};

	DLEAVE("\n");
	return 0;
}


static int cherry_modem_ap_sleep(int sleep)
{

	if (1 == sleep) {
		//printk("====: set AP_SLEEP_GPIO to 0!\n");
		gpio_set_value(mfp_to_gpio(AP_SLEEP_GPIO), 0);
	}
	else {
		gpio_set_value(mfp_to_gpio(AP_SLEEP_GPIO), 1);
		//printk("====: set AP_SLEEP_GPIO to 1!\n");
	}

	return 0;
}


static int cherry_modem_ap_rts(int rts)
{
	DENTER("rts=%d\n", rts);

	/* according to the docs, if rts = 1, AP_RTS should be high */
	if (1 == rts)
		gpio_set_value(mfp_to_gpio(AP_RTS_GPIO), 1);
	else
		gpio_set_value(mfp_to_gpio(AP_RTS_GPIO), 0);

	DLEAVE("\n");
	return 0;
}

static int cherry_modem_ap_rdy(int rdy)
{
	DENTER("rdy=%d\n", rdy);

	if (1 == rdy)
		gpio_set_value(mfp_to_gpio(AP_RDY_GPIO), 0);
	else
		gpio_set_value(mfp_to_gpio(AP_RDY_GPIO), 1);

	DLEAVE("\n");
	return 0;
}

static int cherry_modem_ap_wakeup(int wakeup)
{
	int cnt;
	int bp_sleep;
	unsigned long oscr1, oscr2;
	int retval = 0;
	DENTER("wakeup=%d\n", wakeup);

	if (1 == wakeup) {
		oscr1 = OSCR;
		gpio_set_value(mfp_to_gpio(AP_WAKEUP_GPIO), 1);

		cnt = 100;
		while (cnt--) {
			bp_sleep = gpio_get_value(mfp_to_gpio(BP_SLEEP_GPIO));
			if (0 == bp_sleep)
				break;
			msleep(1);
		}

		if (1 == bp_sleep) {
			printk(KERN_ERR "Cannot wakeup Modem\n");
			WARN_ON(1);
			retval = 1;
		}
		oscr2 = OSCR;

		/*
		printk(KERN_DEBUG "wakeup time: %lu\n", (oscr2 - oscr1) / 3250);
		*/
	} else {
		gpio_set_value(mfp_to_gpio(AP_WAKEUP_GPIO), 0);
	}

	DLEAVE("\n");
	return retval;
}

static int cherry_modem_bp_rts(void)
{
	int ret;
	DENTER("\n");

	ret = gpio_get_value(mfp_to_gpio(BP_RTS_GPIO));

	if (ret == 0)
		ret = 1;
	else
		ret = 0;

	DLEAVE("ret=%d\n", ret);
	return ret;
}

static int cherry_modem_bp_rdy(void)
{
	int ret;
	DENTER("\n");

	ret = gpio_get_value(mfp_to_gpio(BP_RDY_GPIO));

	if (ret == 0)		/* by is rdy when gpio is low */
		ret = 1;
	else
		ret = 0;

	DLEAVE("ret=%d\n", ret);
	return ret;
}

static int cherry_modem_bp_wakeup(void)
{
	int ret;
	DENTER("\n");

	ret = gpio_get_value(mfp_to_gpio(BP_WAKEUP_GPIO));

	if (ret == 0)
		ret = 1;
	else
		ret = 0;

	DLEAVE("ret=%d\n", ret);
	return ret;
}

static int cherry_modem_bp_sleep(void)
{
	int ret;

	ret = gpio_get_value(mfp_to_gpio(BP_SLEEP_GPIO));
	if (ret == 0)
		ret = 1;
	else 
		ret = 0;
	return ret;
}

static int cherry_modem_bp_life(void)
{
	int ret;
	DENTER("\n");

	ret = gpio_get_value(mfp_to_gpio(BP_LIFE_GPIO));

	DLEAVE("ret=%d\n", ret);
	return ret;
}

static struct modem_platform_data cherry_modem_platform_data = {
	.set_power_state = cherry_set_power_state,
	.ap_rts = cherry_modem_ap_rts,
	.ap_rdy = cherry_modem_ap_rdy,
	.ap_wakeup = cherry_modem_ap_wakeup,
	.ap_sleep = cherry_modem_ap_sleep,

	.bp_rts = cherry_modem_bp_rts,
	.bp_rdy = cherry_modem_bp_rdy,
	.bp_wakeup = cherry_modem_bp_wakeup,
	.bp_sleep = cherry_modem_bp_sleep,
	.bp_life = cherry_modem_bp_life,

	.bp_rts_irq = IRQ_GPIO(BP_RTS_GPIO),
	.bp_rdy_irq = IRQ_GPIO(BP_RDY_GPIO),
	.bp_wakeup_irq = IRQ_GPIO(BP_WAKEUP_GPIO),
	.bp_sleep_irq = IRQ_GPIO(BP_SLEEP_GPIO),
	.bp_life_irq = IRQ_GPIO(BP_LIFE_GPIO),
};

static int __init cherry_modem_init(void)
{
	DENTER("\n");

	BUG_ON(gpio_request(AP_POWER_MODEM_GPIO, "power_bp"));
	BUG_ON(gpio_request(AP_RTS_GPIO, "ap_rts"));
	BUG_ON(gpio_request(AP_RDY_GPIO, "ap_rdy"));
	BUG_ON(gpio_request(AP_WAKEUP_GPIO, "ap_wk"));
	BUG_ON(gpio_request(AP_SLEEP_GPIO, "ap_slp"));
	BUG_ON(gpio_request(BP_RTS_GPIO, "bp_rts"));
	BUG_ON(gpio_request(BP_RDY_GPIO, "bp_rdy"));
	BUG_ON(gpio_request(BP_WAKEUP_GPIO, "bp_wk"));
	BUG_ON(gpio_request(BP_SLEEP_GPIO, "bp_slp"));
	BUG_ON(gpio_request(BP_LIFE_GPIO, "bp_life"));
	BUG_ON(gpio_request(AP_BB_CON_DOWNLOAD02_GPIO, "bb_dl"));
	BUG_ON(gpio_request(SPI_CLK_GPIO, "ipc_spi_clk"));
	BUG_ON(gpio_request(SPI_FRM_GPIO, "ipc_spi_frm"));
	BUG_ON(gpio_request(SPI_TXD_GPIO, "ipc_spi_txd"));
	BUG_ON(gpio_request(SPI_RXD_GPIO, "ipc_spi_rxd"));

	cherry_set_power_state(MPS_OFF);
	cherry_set_power_state(MPS_BOOTING);

	pxa_set_modem_info(&cherry_modem_platform_data);

	DLEAVE("\n");
	return 0;
}

module_init(cherry_modem_init);
