/*
 * arch/arm/mach-pxa/pxa3xx_pm_check.c 
 *
 * Copyright (C) 2008 Borqs Ltd.
 * Author:  Emichael Li <emichael.li@borqs.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/module.h>
#include <linux/pm.h>
#include <asm/gpio.h>
#include <mach/hardware.h>
#include <mach/pxa-regs.h>
#include <mach/pxa3xx-regs.h>
#include <mach/pxa3xx_pm_check.h>
#include <mach/micco.h>
#include <mach/mfp.h>

#ifdef CONFIG_PXA3xx_MICCO
/******************************************************************************/
/*                                                                            */
/*                        MICCO Power Check                                   */
/*                                                                            */
/******************************************************************************/
static struct micco_check (*pxa3xx_micco_setting)[] = NULL;

int register_pxa3xx_micco_setting(struct micco_check (*micco_setting)[])
{
	if (pxa3xx_micco_setting == NULL) {
		pxa3xx_micco_setting = micco_setting;
	} else {
		printk(KERN_ERR "Only one setting can be registered!\n");
		BUG();
	}
	return 0;
}

static int pxa3xx_micco_standby_check(struct pm_check *pmc)
{
	int i;
	int ret = 0;
	struct micco_check *pcheck;
	unsigned char reg, val, cur_val;
	static int count = 0;

	if (NULL == pxa3xx_micco_setting) {
		if (!count) {
			printk(KERN_WARNING "no mfp setting is registered\n");
			count = 1;
		}
		return 0;
	}

	for (i = 0; ; ++i) {
		pcheck = *pxa3xx_micco_setting + i;
		reg = pcheck->reg;
		val = pcheck->val;
		if (0xFF == reg)
			break;

		micco_read_cache(reg, &cur_val);
		if (cur_val != val) {
			printk(KERN_DEBUG "Micco 0x%x (0x%x: 0x%x) is not right\n", reg, cur_val, val);
			ret = 1;
		}
	}
	return ret;
}


struct pm_check pxa3xx_micco_check = {
	.name = "pxa3xx_pmic",
	.standby = pxa3xx_micco_standby_check,
};
#endif /* CONFIG_PXA3xx_MICCO */

/******************************************************************************/
/*                                                                            */
/*                         GPIO (MFP) Check                                   */
/*                                                                            */
/******************************************************************************/
static struct gpio_check (*pxa3xx_mfp_setting)[] = NULL;

int register_pxa3xx_mfp_setting(struct gpio_check (*mfp_setting)[])
{
	if (pxa3xx_mfp_setting == NULL) {
		pxa3xx_mfp_setting = mfp_setting;
	} else {
		printk(KERN_ERR "Only one setting can be registered!\n");
		BUG();
	}
	return 0;
}

#define MFP_PMC_MASK	(0xFF87)
static int pxa3xx_gpio_standby_check(struct pm_check *pmc)
{
	int i;
	int ret = 0;
	struct gpio_check *pcheck;
	int gpio, mfp, cur_mfp;
	static int count = 0;

	if (pxa3xx_mfp_setting == NULL) {
		if (!count) {
			printk(KERN_WARNING "no mfp setting is registered\n");
			count = 1;
		}
		return 0;
	}

	for (i = 0; ; ++i) {
		pcheck = *pxa3xx_mfp_setting + i;
		gpio = pcheck->gpio;
		mfp = pcheck->mfp;
		if (-1 == gpio)
			break;

		if (0 == mfp)
			continue;

		if( cpu_is_pxa935())
			mfp = (((mfp & MFPR_DRV_MASK) << 1) & MFPR_DRV_MASK) | (mfp & ~MFPR_DRV_MASK);

		cur_mfp = MFP_CFG_VAL(pxa3xx_mfp_read(gpio));
		if ((mfp & MFP_PMC_MASK) != (cur_mfp & MFP_PMC_MASK)) {
			printk(KERN_DEBUG "GPIO %d (0x%x: 0x%x) is not right\n", gpio, cur_mfp, mfp);
			ret = 1;
		}
	}
	return ret;
}

struct pm_check pxa3xx_gpio_check = {
	.name = "pxa3xx_gpio",
	.standby = pxa3xx_gpio_standby_check,
};


/******************************************************************************/
/*                                                                            */
/*                         pxa3xx clock Check                                 */
/*                                                                            */
/******************************************************************************/
static int pxa3xx_clock_standby_check(struct pm_check *pmc)
{
	int ret = 0;

	if (unlikely(OSCC & OSCC_PEN)) {
		ret = 1;
		printk(KERN_DEBUG "OSCC_PEN clock is not disabled by audio/bt/u2d driver\n");
	}

	if (unlikely(CKENA != CKENA_STANDBY)) {
		ret = 1;
		printk(KERN_DEBUG "CKENA(0x%x: 0x%x) is not right\n", CKENA, CKENA_STANDBY);
	}

	if (unlikely(CKENB != CKENB_STANDBY)) {
		ret = 1;
		printk(KERN_DEBUG "CKENB(0x%x: 0x%x) is not right\n", CKENB, CKENB_STANDBY);
	}

#ifdef CONFIG_CPU_PXA935
	if (unlikely(CKENC != CKENC_STANDBY)) {
		ret = 1;
		printk(KERN_DEBUG "CKENC(0x%x: 0x%x) is not right\n", CKENC, CKENC_STANDBY);
	}
#endif
	return ret;
}

struct pm_check pxa3xx_clock_check = {
	.name = "pxa3xx_clock",
	.standby = pxa3xx_clock_standby_check,
};


/******************************************************************************/
/*                                                                            */
/*                         module init/exit                                   */
/*                                                                            */
/******************************************************************************/
static int __init pxa3xx_pm_check_init(void)
{
	pm_check_register(&pxa3xx_clock_check);
	pm_check_register(&pxa3xx_gpio_check);
#ifdef CONFIG_PXA3xx_MICCO
	pm_check_register(&pxa3xx_micco_check);
#endif
	return 0;
}

static void __exit pxa3xx_pm_check_exit(void)
{
}

module_init(pxa3xx_pm_check_init);
module_exit(pxa3xx_pm_check_exit);
