/*
 *
 *  Bluetooth HCI UART driver
 *
 *
 *  Copyright (C) 2009, Marvell International Ltd.
 *
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include <linux/module.h>

#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/types.h>
#include <linux/fcntl.h>
#include <linux/interrupt.h>
#include <linux/ptrace.h>
#include <linux/poll.h>

#include <linux/slab.h>
#include <linux/tty.h>
#include <linux/errno.h>
#include <linux/string.h>
#include <linux/signal.h>
#include <linux/ioctl.h>
#include <linux/skbuff.h>
#include <linux/platform_device.h>
#include <linux/timer.h>

#include <net/bluetooth/bluetooth.h>
#include <net/bluetooth/hci_core.h>
#include <linux/proc_fs.h>
#include <linux/stat.h>

#include <asm/hci_ps.h>
#include "hci_uart.h"


u32 drvdbg = 0;

#ifndef CONFIG_BT_HCIUART_DEBUG
#undef  BT_DBG
#define	BT_DBG(fmt, arg...)  do {if (drvdbg)   printk(KERN_ALERT "%s: " fmt "\n" , __FUNCTION__ , ## arg);} while(0)
#endif

#ifdef CONFIG_PROC_FS
#define PROC_DIR	NULL
#else
#define PROC_DIR	proc_net
#endif

#define DEFAULT_BUF_SIZE 512

/** Default file permission */
#define DEFAULT_FILE_PERM  0644

/** Default time period in mili-second */
#define DEFAULT_TIME_PERIOD 2000


/** wakeup method DTR */
#define WAKEUP_METHOD_DTR       0
/** wakeup method break */
#define WAKEUP_METHOD_BREAK     1
/** wakeup method invalid */
#define  WAKEUP_METHOD_INVALID  0xff
/** ps mode disable */
#define PS_MODE_DISABLE         0
/** ps mode enable */
#define PS_MODE_ENABLE          1
/** ps cmd exit ps  */
#define PS_CMD_EXIT_PS          1
/** ps cmd enter ps */
#define PS_CMD_ENTER_PS         2
/** ps state awake */
#define PS_STATE_AWAKE                0
/** ps state SLEEP */
#define PS_STATE_SLEEP                1

/** OGF */
#define OGF				0x3F
/** Bluetooth command : Sleep mode */
#define BT_CMD_AUTO_SLEEP_MODE		0x23
/** Bluetooth Power State : Enable */
#define BT_PS_ENABLE			0x02
/** Bluetooth Power State : Disable */
#define BT_PS_DISABLE			0x03
/** Bluetooth command: Wakeup method */
#define BT_CMD_WAKEUP_METHOD    0x53

#define BT_HOST_WAKEUP_METHOD_NONE      0x00
#define BT_HOST_WAKEUP_METHOD_DTR       0x01
#define BT_HOST_WAKEUP_METHOD_BREAK     0x02
#define BT_HOST_WAKEUP_METHOD_GPIO      0x03
#define BT_HOST_WAKEUP_DEFAULT_GPIO     5

#define BT_CTRL_WAKEUP_METHOD_DSR       0x00
#define BT_CTRL_WAKEUP_METHOD_BREAK     0x01
#define BT_CTRL_WAKEUP_METHOD_GPIO      0x02
#define BT_CTRL_WAKEUP_DEFAULT_GPIO     4

#define  HCI_OP_AUTO_SLEEP_MODE 0xfc23
#define  HCI_OP_WAKEUP_METHOD   0xfc53
#define SEND_WAKEUP_METHOD_CMD          0x01
#define SEND_AUTO_SLEEP_MODE_CMD        0x02

typedef struct _BT_CMD {
    /** OCF OGF */
	u16	ocf_ogf;
    /** Length */
	u8	length;
    /** Data */
	u8	data[4];
} __attribute__ ((packed)) BT_CMD;


/** Proc directory entry */
static struct proc_dir_entry *proc_bt = NULL;

struct ps_data {
	u32    ps_mode;
	u32    cur_psmode;
	u32    ps_state;
	u32    ps_cmd;
	u32    interval;
	u32    wakeupmode;
	u32    cur_wakeupmode;
	u32    send_cmd;
	struct work_struct work;
	struct tty_struct *tty;
	struct timer_list ps_timer;
	u32    timer_on;

	chip_type_t	chip_type;
	void	(* chip_setpower)(int);
};

static struct ps_data  g_data;
int wakeupmode = WAKEUP_METHOD_DTR;
struct proc_data {
    /** Read length */
	int  rdlen;
    /** Read buffer */
	char *rdbuf;
    /** Write length */
	int  wrlen;
    /** Maximum write length */
	int  maxwrlen;
    /** Write buffer */
	char *wrbuf;
};


/** Debug dump buffer length */
#define DBG_DUMP_BUF_LEN 	64
/** Maximum number of dump per line */
#define MAX_DUMP_PER_LINE	16
/** Maximum data dump length */
#define MAX_DATA_DUMP_LEN	48
#define DBG_RAW_DATA        0x10
void hexdump(char *prompt, u8 * buf, int len)
{
    int i;
    char dbgdumpbuf[DBG_DUMP_BUF_LEN];
    char *ptr = dbgdumpbuf;
    
    if(g_data.chip_type != CHIP_MRVL8688)
        return;
    if(!drvdbg)
        return;
    printk(KERN_DEBUG "%s: len=%d\n", prompt,len);
    if(!(drvdbg & DBG_RAW_DATA))
        return;
    for (i = 1; i <= len; i++) {
        ptr += sprintf(ptr, "%02x ", *buf);
        buf++;
        if (i % MAX_DUMP_PER_LINE == 0) {
            *ptr = 0;
            printk(KERN_DEBUG "%s\n", dbgdumpbuf);
            ptr = dbgdumpbuf;
        }
    }
    if (len % MAX_DUMP_PER_LINE) {
        *ptr = 0;
        printk(KERN_DEBUG "%s\n", dbgdumpbuf);
    }
}


/** convert string to number */
int string_to_number(char *s)
{
	int r = 0;
	int base = 0;
	int pn = 1;

	if (strncmp(s, "-", 1) == 0) {
		pn = -1;
		s ++;
	}
	if ((strncmp(s, "0x", 2) == 0) || (strncmp(s, "0X", 2) == 0)) {
		base = 16;
		s += 2;
	}
	else
		base = 10;

	for (s = s; *s !=0; s++) {
		if ((*s >= '0') && (*s <= '9'))
			r = (r * base) + (*s - '0');
		else if ((*s >= 'A') && (*s <= 'F'))
			r = (r * base) + (*s - 'A' + 10);
		else if ((*s >= 'a') && (*s <= 'f'))
			r = (r * base) + (*s - 'a' + 10);
		else
			break;
	}

	return (r * pn);
}

/*
 * Builds and sends an PS command packet.
 */
static int send_ps_cmd(u8 cmd, struct hci_uart *hu)
{
    int err = 0;
    struct sk_buff *skb = NULL;
    BT_CMD *pCmd;

    BT_DBG("hu %p cmd 0x%x", hu, cmd);

    /* allocate packet */
    skb = bt_skb_alloc(sizeof(BT_CMD), GFP_ATOMIC);
    if (!skb) {
        BT_ERR("cannot allocate memory for HCILL packet");
        err = -ENOMEM;
        goto out;
    }
    
    pCmd = (BT_CMD *)skb->tail;		
    pCmd->ocf_ogf =  (OGF << 10)|BT_CMD_AUTO_SLEEP_MODE;
    pCmd->length = 1;
    if(cmd ==  PS_MODE_ENABLE)
        pCmd->data[0] = BT_PS_ENABLE;
    else
        pCmd->data[0] = BT_PS_DISABLE;

    bt_cb(skb)->pkt_type = HCI_COMMAND_PKT; 
    skb_put(skb,sizeof(BT_CMD)-4 + pCmd->length);
    skb->dev = (void *) hu->hdev;

    /* send packet */
    hu->proto->enqueue(hu, skb);
    hci_uart_tx_wakeup(hu);

out:
    return err;
}

/*
 * Builds and sends an wake up method command packet.
 */
static int send_wakeup_method_cmd(u8 cmd, struct hci_uart *hu)
{
    int err = 0;
    struct sk_buff *skb = NULL;
    BT_CMD *pCmd;

    BT_DBG("hu %p cmd 0x%x", hu, cmd);

    /* allocate packet */
    skb = bt_skb_alloc(sizeof(BT_CMD), GFP_ATOMIC);
    if (!skb) {
        BT_ERR("cannot allocate memory for HCILL packet");
        err = -ENOMEM;
        goto out;
    }
    
    pCmd = (BT_CMD *)skb->tail;		
    pCmd->ocf_ogf =  (OGF << 10)|BT_CMD_WAKEUP_METHOD;
    pCmd->length = 4;
    //pCmd->data[0] = BT_HOST_WAKEUP_METHOD_NONE;
    pCmd->data[0] = BT_HOST_WAKEUP_METHOD_DTR;
    pCmd->data[1] = BT_HOST_WAKEUP_DEFAULT_GPIO;
    if(cmd == WAKEUP_METHOD_DTR)
        pCmd->data[2] = BT_CTRL_WAKEUP_METHOD_DSR;
    else
        pCmd->data[2] = BT_CTRL_WAKEUP_METHOD_BREAK;

    pCmd->data[3] = BT_CTRL_WAKEUP_DEFAULT_GPIO;     

    bt_cb(skb)->pkt_type = HCI_COMMAND_PKT;    

    skb_put(skb,sizeof(BT_CMD)-4 + pCmd->length);
    skb->dev = (void *) hu->hdev;

    /* send packet */
    hu->proto->enqueue(hu, skb);
    hci_uart_tx_wakeup(hu);

out:
    return err;
}



/** This function handle the generic file close */
static void proc_on_close( struct inode *inode, struct file *file ) 
{
    struct proc_data *pdata = file->private_data;
    char *line;
    struct hci_uart *hu = NULL;
    u32 ps_mode = g_data.cur_psmode;
    u32 wakeupmode = g_data.cur_wakeupmode;
    if ( !pdata->wrlen ) 
        return;
    line = pdata->wrbuf;
    while( line[0] ) {
        if(!strncmp( line,"psmode",strlen("psmode"))){
            line += strlen("psmode") + 1;
            ps_mode = (u32)string_to_number(line);
            if(ps_mode > PS_MODE_ENABLE)
                ps_mode = g_data.cur_psmode;
        }
        if(!strncmp( line,"interval",strlen("interval"))){
            line += strlen("interval") + 1;
            g_data.interval = (u32)string_to_number(line);
        }
        if(!strncmp( line,"drvdbg",strlen("drvdbg"))){
            line += strlen("drvdbg") + 1;
            drvdbg = (u32)string_to_number(line);
        }
        if(!strncmp( line,"wakeupmode",strlen("wakeupmode"))){
            line += strlen("wakeupmode") + 1;
            wakeupmode = (u32)string_to_number(line);
            if(wakeupmode > WAKEUP_METHOD_BREAK)
                wakeupmode = g_data.cur_wakeupmode;
        }
        while( line[0] && line[0] != '\n' ) line++;
        if ( line[0] ) line++;
    }
    if((g_data.cur_psmode == PS_MODE_DISABLE) && 
        (ps_mode == PS_MODE_DISABLE) &&             
        (wakeupmode != g_data.cur_wakeupmode)) {
        g_data.wakeupmode = wakeupmode;
        if(g_data.tty) {
            hu =  (void *)g_data.tty->disc_data;
            if(hu->proto) {
                g_data.send_cmd |= SEND_WAKEUP_METHOD_CMD;
                send_wakeup_method_cmd(g_data.wakeupmode,hu);
            }
        }
    }
    if(ps_mode != g_data.cur_psmode) {
        g_data.ps_mode = ps_mode;
        if(g_data.tty) {
            hu =  (void *)g_data.tty->disc_data;
            if(hu->proto) {
                g_data.send_cmd |= SEND_AUTO_SLEEP_MODE_CMD;
                send_ps_cmd(g_data.ps_mode,hu);
            }
        }
    }
    return;
}

/** This function handle generic proc file close */
static int proc_close( struct inode *inode, struct file *file )
{
	struct proc_data *pdata = file->private_data;
	if(pdata){
		proc_on_close(inode, file);
		if(pdata->rdbuf)
			kfree(pdata->rdbuf);
		if(pdata->wrbuf)
			kfree(pdata->wrbuf);
		kfree(pdata);
	}
	return 0;
}

/** This function handle generic proc file read */
static ssize_t proc_read( struct file *file,
			  char __user *buffer,
			  size_t len,
			  loff_t *offset )
{
	loff_t pos = *offset;
	struct proc_data *pdata = (struct proc_data*)file->private_data;
	if((!pdata->rdbuf) || (pos < 0))
		return -EINVAL;
	if (pos >= pdata->rdlen)
		return 0;
	if (len > pdata->rdlen - pos)
		len = pdata->rdlen - pos;
	if (copy_to_user(buffer, pdata->rdbuf + pos, len))
		return -EFAULT;
	*offset = pos + len;
	return len;
}


/** This function handle generic proc file write */
static ssize_t proc_write( struct file *file,
			   const char __user *buffer,
			   size_t len,
			   loff_t *offset )
{
	loff_t pos = *offset;
	struct proc_data *pdata = (struct proc_data*)file->private_data;

	if (!pdata->wrbuf || (pos < 0))
		return -EINVAL;
	if (pos >= pdata->maxwrlen)
		return 0;
	if (len > pdata->maxwrlen - pos)
		len = pdata->maxwrlen - pos;
	if (copy_from_user(pdata->wrbuf + pos, buffer, len))
		return -EFAULT;
	if ( pos + len > pdata->wrlen )
		pdata->wrlen = len + file->f_pos;
	*offset = pos + len;
	return len;
}


/** This function handle the generic file open */
static int proc_open( struct inode *inode, struct file *file ) 
{
	struct proc_data *pdata;
	char *p;
	if ((file->private_data = kzalloc(sizeof(struct proc_data ), GFP_KERNEL)) == NULL){
		BT_ERR("Can not allocate memmory for proc_data\n");
		return -ENOMEM;
	}
	pdata = (struct proc_data *)file->private_data;
	if ((pdata->rdbuf = kmalloc(DEFAULT_BUF_SIZE, GFP_KERNEL )) == NULL) {
		BT_ERR("Can not allocate memory for rdbuf\n");
		kfree (file->private_data);
		return -ENOMEM;
	}
	if ((pdata->wrbuf = kzalloc(DEFAULT_BUF_SIZE, GFP_KERNEL )) == NULL) {
		BT_ERR("Can not allocate memory for wrbuf\n");
		kfree (pdata->rdbuf);
		kfree (file->private_data);
		return -ENOMEM;
	}
	pdata->maxwrlen = DEFAULT_BUF_SIZE;
	p = pdata->rdbuf;
	p += sprintf(p, "psmode=%d\n",g_data.ps_mode);
	p += sprintf(p, "psstate=%d\n",g_data.ps_state);
	p += sprintf(p, "interval=%d\n",g_data.interval);
	p += sprintf(p, "wakeupmode=%d\n",g_data.wakeupmode);
	p += sprintf(p, "current psmode=%d\n",g_data.cur_psmode);
   	p += sprintf(p, "current wakeupmode=%d\n",g_data.cur_wakeupmode);
   	p += sprintf(p, "sendcmd=%d\n",g_data.send_cmd);
	p += sprintf(p, "drvdbg=%d\n",drvdbg);
	pdata->rdlen = strlen(pdata->rdbuf);
	return 0;
}

static struct file_operations proc_rw_ops = {
	.read		= proc_read,
	.write		= proc_write,
	.open		= proc_open,
	.release	= proc_close
};



void ps_timeout_func(unsigned long context)
{
    struct ps_data *data = (struct ps_data *)context;
    struct tty_struct *tty = data->tty;
    struct hci_uart *hu = NULL; 
    data->timer_on = 0;
    if(!data->tty) 
        return;
    hu = (struct hci_uart *)tty->disc_data;        
    if (test_bit(HCI_UART_SENDING, &hu->tx_state)) {
        ps_start_timer();
    }
    else {
        data->ps_cmd = PS_CMD_ENTER_PS;
        schedule_work(&data->work);
    }
}


static void set_dtr(struct tty_struct *tty, int on_off)
{
    if (g_data.chip_setpower)
    g_data.chip_setpower(!on_off);
    msleep(100);

    return;
#if 0
    u32 old_state = 0;
    u32 new_state = 0;
    if (tty->driver->tiocmget) {
        old_state = tty->driver->tiocmget(tty, NULL);
        if(on_off)
            new_state = old_state | TIOCM_DTR;
        else
            new_state = old_state & ~TIOCM_DTR;
        if(new_state == old_state)
            return;
        if(tty->driver->tiocmset) {
            if(on_off) {
                BT_DBG("Set DTR ON");
                tty->driver->tiocmset(tty, NULL,TIOCM_DTR, 0);
            }
            else {
                BT_DBG("Clear DTR ");
                tty->driver->tiocmset(tty, NULL, 0, TIOCM_DTR);
            }
        }
    }
    return;
#endif
}

static void set_break(struct tty_struct *tty, int on_off)
{
    if (tty->driver->ops->break_ctl) {
        if(on_off) {
            BT_DBG("Turn on break");
            tty->driver->ops->break_ctl(tty, -1); /* turn on break */
        }
        else {
            BT_DBG("Turn off break");
            tty->driver->ops->break_ctl(tty, 0);  /*turn off break */
        }
    }
    return;
}


static void ps_control(struct ps_data *data, u8 ps_state)
{
    struct hci_uart *hu = NULL; 
    if(data->ps_state == ps_state)
        return;
    if(data->tty) {
        switch(data->cur_wakeupmode) {
            case WAKEUP_METHOD_DTR:
                if(ps_state == PS_STATE_AWAKE)
                    set_dtr(data->tty, 1);  //DTR ON
                else
                    set_dtr(data->tty, 0);  //DTR OFF
                data->ps_state = ps_state;
                break;
            case WAKEUP_METHOD_BREAK:
                if(ps_state == PS_STATE_AWAKE)
                    set_break(data->tty, 0); //break OFF
                else
                    set_break(data->tty, 1); //break ON
                data->ps_state = ps_state;                            
                break;          
            default:
                break;      
        }
        if(ps_state == PS_STATE_AWAKE) {
            hu = (struct hci_uart *)data->tty->disc_data;  
            /* actually send the packets */
            BT_DBG("Send tx data");
            hci_uart_tx_wakeup(hu);
        }
    }
}

static void ps_work_func(struct work_struct *work)
{
	struct ps_data *data = container_of(work, struct ps_data, work);
    if(data->tty){
        if((data->ps_cmd == PS_CMD_ENTER_PS) && (data->cur_psmode == PS_MODE_ENABLE)) {  
            ps_control(data,PS_STATE_SLEEP);
        }
        else  if(data->ps_cmd == PS_CMD_EXIT_PS) {
            ps_control(data,PS_STATE_AWAKE);
        }
    }
}
    
void ps_init_work(void) 
{
    memset(&g_data, 0, sizeof(g_data));
    g_data.interval = DEFAULT_TIME_PERIOD;
    g_data.timer_on    = 0;
    g_data.tty = NULL;
    g_data.ps_state = PS_STATE_AWAKE;
    g_data.ps_mode = PS_MODE_ENABLE;
    g_data.ps_cmd = 0;
    g_data.send_cmd = 0;
    if(wakeupmode !=  WAKEUP_METHOD_BREAK)
        g_data.wakeupmode =  WAKEUP_METHOD_DTR;
    else
        g_data.wakeupmode =  WAKEUP_METHOD_BREAK;
    g_data.cur_psmode = PS_MODE_DISABLE;
    g_data.cur_wakeupmode = WAKEUP_METHOD_INVALID;
    INIT_WORK(&g_data.work,ps_work_func);
}   


/** This function init proc entry  */
int proc_init(void)
{
	u8 ret = 0;
	struct proc_dir_entry *entry;
	if(!proc_bt){
		proc_bt = proc_mkdir("mbt_uart", PROC_DIR);
		if(!proc_bt){
			BT_ERR("Could not mkdir mbt_uart!\n");
		    ret = -1;
			goto done;
		}			
		entry = create_proc_entry("config", S_IFREG |DEFAULT_FILE_PERM, proc_bt);
		if(entry){
			entry->owner = THIS_MODULE;
			entry->proc_fops =&proc_rw_ops;
		}
	}
    ps_init_work();
done:	
	return ret;
}	

/** remove proc file */
void proc_remove(void)
{
	if(proc_bt){
		remove_proc_entry("config",proc_bt);
		remove_proc_entry("mbt_uart", PROC_DIR);
		proc_bt = NULL;
	}
	return;
}



void ps_init_timer(struct tty_struct *tty)
{
   if(g_data.chip_type != CHIP_MRVL8688)
       return;
   init_timer(&g_data.ps_timer);
   g_data.timer_on = 0;
   g_data.tty = tty;
   g_data.ps_timer.function = ps_timeout_func;
   g_data.ps_timer.data = (u32)&g_data;
   return;
}

void ps_start_timer(void)
{
    if(g_data.chip_type != CHIP_MRVL8688)
        return;
    if(g_data.cur_psmode ==  PS_MODE_ENABLE) {
        g_data.timer_on = 1;
        mod_timer(&g_data.ps_timer, jiffies + (g_data.interval * HZ)/1000);
    }
}

void ps_cancel_timer(void)
{
    if(g_data.chip_type != CHIP_MRVL8688)
        return;
    if(g_data.timer_on)
        del_timer(&g_data.ps_timer);
    //g_data.tty = NULL;
    return;
}

int ps_wakeup(void)
{
    if(g_data.chip_type != CHIP_MRVL8688)
        return 0;
    if(g_data.ps_state == PS_STATE_AWAKE) 
        return 0;
    g_data.ps_cmd = PS_CMD_EXIT_PS;
   	schedule_work(&g_data.work);
    return 1;
}


void ps_init(void)
{
    struct hci_uart *hu;
    if(g_data.chip_type != CHIP_MRVL8688)
        return;

    hu = (void *)g_data.tty->disc_data;
    g_data.send_cmd = 0;
    if(hu->proto){
        //if(g_data.cur_wakeupmode != g_data.wakeupmode) {
            g_data.send_cmd |= SEND_WAKEUP_METHOD_CMD;
            send_wakeup_method_cmd(g_data.wakeupmode,hu);
        //}
        //if(g_data.cur_psmode != g_data.ps_mode) {
            g_data.send_cmd |= SEND_AUTO_SLEEP_MODE_CMD;
            send_ps_cmd(g_data.ps_mode,hu);
        //}
    }
}


void ps_check_event_packet(struct sk_buff *skb)
{
    struct hci_event_hdr *hdr = (void *) skb->data;
    struct hci_ev_cmd_complete *ev = NULL;
    u8 event = hdr->evt;
    u16 opcode;
    u8 status = 0;
    if(g_data.chip_type != CHIP_MRVL8688)
        return;
    if(!g_data.send_cmd)
        return;
    if(event == HCI_EV_CMD_COMPLETE) {
        ev = (void *)(skb->data + sizeof(struct hci_event_hdr));
        opcode = __le16_to_cpu(ev->opcode);
        switch(opcode) {
            case HCI_OP_AUTO_SLEEP_MODE:
                status = *((u8 *)ev + sizeof(struct hci_ev_cmd_complete));
                if(!status)
                    g_data.cur_psmode = g_data.ps_mode;
                else
                    g_data.ps_mode = g_data.cur_psmode;
                g_data.send_cmd &= ~SEND_AUTO_SLEEP_MODE_CMD;
                if(g_data.cur_psmode == PS_MODE_ENABLE) 
                    ps_start_timer();
                else
                    ps_wakeup();
                BT_DBG("status=%d,ps_mode=%d", status,g_data.cur_psmode);
                break;
            case HCI_OP_WAKEUP_METHOD:
                status = *((u8 *)ev + sizeof(struct hci_ev_cmd_complete));
                g_data.send_cmd &= ~SEND_WAKEUP_METHOD_CMD;
                if(!status)
                    g_data.cur_wakeupmode = g_data.wakeupmode;
                else
                    g_data.wakeupmode = g_data.cur_wakeupmode;
                BT_DBG("status=%d,wakeupmode=%d", status,g_data.cur_wakeupmode);
                break;
            default:
                break;                    
        }
    }
    return;
}

#if defined(CONFIG_MRVL8688_CHIP_LPM_SWITCH)
static ssize_t store_lpm_ctrl(struct class *dev, const char *buf, size_t count)
{
	char *ptr;
	int val;
	val = simple_strtoul(buf, &ptr, 10);
	if (ptr == buf)
 		return -EINVAL;
	g_data.chip_setpower(val);

	return count;
}
static CLASS_ATTR(lpm_ctrl, S_IWUSR, NULL, store_lpm_ctrl); 
#endif

static int __devinit hci_ps_probe(struct platform_device *pdev)
{
	struct hci_ps_platform_data *pdata;

	pdata = pdev->dev.platform_data;
        if (!pdata) {
                dev_err(&pdev->dev,"platform data missing\n");
                return -ENODEV;
        }

	g_data.chip_type = pdata->chip_type;
	g_data.chip_setpower = pdata->chip_setpower;

#if defined(CONFIG_MRVL8688_CHIP_LPM_SWITCH)
	if(class_create_file(bt_class, &class_attr_lpm_ctrl)) 
		printk("Bluetooth lpm_ctrl interface create failed\n"); 
#endif

	return 0;
}

static int __devexit hci_ps_remove(struct platform_device *pdev)
{
	return 0;
}

static struct platform_driver hci_ps_device_driver = {
	.probe = hci_ps_probe,
	.remove = hci_ps_remove,
	.driver = {
		.name = "hci_ps",
	},
};

static int __init hci_ps_init(void)
{
	return platform_driver_register(&hci_ps_device_driver);
}

static void  __exit hci_ps_exit(void)
{
	platform_driver_unregister(&hci_ps_device_driver);
}

module_init(hci_ps_init);
module_exit(hci_ps_exit);
module_param(wakeupmode, int, WAKEUP_METHOD_BREAK);
