/*
 * micco_headset.c 
 *
 * Copyright (C) 2008 Borqs Ltd.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 */
#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/platform_device.h>

#include <linux/errno.h>
#include <linux/types.h>
#include <linux/kdev_t.h>
#include <linux/interrupt.h>
#include <linux/spinlock.h>
#include <linux/delay.h>
#include <linux/cdev.h>
#include <linux/device.h>
#include <linux/proc_fs.h>
#include <linux/miscdevice.h>
#include <linux/accessory.h>
#include <linux/earlysuspend.h>
#include <linux/latesuspend.h>
#include <linux/wakelock.h>
#include <mach/micco.h>
#include <mach/pxa3xx_pmic.h>
#include <mach/micco_headset.h>
#include <mach/gpio.h>
#include <mach/mfp.h>

#define HEADSET_WITH_MIC_ON	(1 << 0)
#define HEADSET_NO_MIC_ON	(1 << 1)
#define TTY_DEV_ON		(1 << 2)

#define HEADSET_ON		(1 << 0)
#define MICPHONE_ON		(1 << 1)
#define DETECTING		(1 << 31)

/* FIXME:
 * we have to assume the headset with micphone no initilization,
 * after detecting, the micphone bit would be masked if the there is 
 * true a headset without micphone. 
 */

struct micco_headset_platform_data *headset_pdata;
static atomic_t g_headset_state;
static struct delayed_work headset_dwork;
static struct timer_list hookswitch_down_timer;
static struct timer_list hookswitch_up_timer;
extern void hookswitch_keyevent(int);

#ifdef CONFIG_HAS_WAKELOCK
static struct wake_lock micco_headset_lock;
#endif

static int micco_headset_get_property(struct accessory *acy,
				  enum accessory_property pacy,
				  union accessory_propval *val)
{
	val->intval = atomic_read(&g_headset_state);

	switch (pacy) {
	case ACCESSORY_PROP_ONLINE:
		if ( val->intval & DETECTING ) 
			val->intval = 0;
		break;
	default:
		return -EINVAL;
	}

	return 0;
}

static enum accessory_property micco_headset_props[] = {
	ACCESSORY_PROP_ONLINE,
};
static struct accessory micco_headset_accessory = {
	.name = "headset",
	.type = ACCESSORY_TYPE_HEADSET,
	.properties = micco_headset_props,
	.num_properties = ARRAY_SIZE(micco_headset_props),
	.get_property = micco_headset_get_property,
	.put_property = NULL,

};

static int headset_state_convert(int old_state)
{
	int new_state = 0;
	if ((old_state & HEADSET_ON) && (old_state & MICPHONE_ON)) {
		new_state = HEADSET_WITH_MIC_ON;
		printk("--->>>>>with mic headset insert \n");
	} else if (old_state & HEADSET_ON) {
		new_state = HEADSET_NO_MIC_ON;
		printk("--->>>>>without mic headset insert \n");
	} else {
		new_state = old_state;
		printk("--->>>>>headset remove\n");
	}

	return new_state;
}

static void mask_hook_irq(void)
{
	int headset_state = atomic_read(&g_headset_state);
	headset_state |= DETECTING;
	atomic_set(&g_headset_state, headset_state);
	del_timer(&hookswitch_down_timer);
	del_timer(&hookswitch_up_timer);
	
}

static void headset_detect_workqueue(struct work_struct *work)
{
	u8 val;	
	int headset_state = 0;
	/***** uart detect *****/	
	if (headset_pdata && headset_pdata->detecting_headset) {
		if (headset_pdata->detecting_headset()) {
			printk(KERN_DEBUG "%s: This is a uart\n", __func__);
			return;
		} 
	}
	micco_read(MICCO_STATUS_B, &val);
	headset_state = atomic_read(&g_headset_state);
	if (headset_state & DETECTING) {
		/****** headset detect ******/
		if (val & MICCO_STATUS_B_HEADSET) {
			printk(KERN_DEBUG "%s: headset on\n", __func__);
			headset_state |= HEADSET_ON;
			if (headset_pdata && headset_pdata->insert)
				headset_pdata->insert();
		} else {
			printk(KERN_DEBUG "%s: headset off\n", __func__);
			headset_state &= ~(HEADSET_ON | MICPHONE_ON);
			if (headset_pdata && headset_pdata->remove)
				headset_pdata->remove();

		}

		/****** micphone detect ******/
		if (val & MICCO_STATUS_B_HOOKSWITCH) {
			printk(KERN_DEBUG "%s: micphone off\n", __func__);
			headset_state &= ~MICPHONE_ON;
		} else if (val & MICCO_STATUS_B_HEADSET) {
			printk(KERN_DEBUG "%s: micphone on\n", __func__);
			headset_state |= MICPHONE_ON;
			headset_state &= ~DETECTING;
		}
	headset_state = headset_state_convert(headset_state);
	atomic_set(&g_headset_state, headset_state);
	accessory_changed(&micco_headset_accessory);
	}
}

#ifdef CONFIG_GPIO_DETECT_HEADSET
static struct work_struct gpio_fast_work;

static void gpio_work(struct work_struct *work)
{
	printk(KERN_DEBUG "headset gpio interrupt\n");
	if (headset_pdata && headset_pdata->remove)
		headset_pdata->remove();
	if( headset_pdata )
		mutex_lock(&headset_pdata->worker_mutex);
	cancel_delayed_work_sync(&headset_dwork);
	schedule_delayed_work(&headset_dwork, msecs_to_jiffies(1000));
	if( headset_pdata )
		mutex_unlock(&headset_pdata->worker_mutex);
}

extern int set_irq_type(unsigned int irq, unsigned int type);
static irqreturn_t gpio_interrupt(int irq, void *dev_id)
{
#ifdef CONFIG_HAS_WAKELOCK
	wake_lock_timeout(&micco_headset_lock, 5 * HZ);
#endif
	mask_hook_irq();
	schedule_work(&gpio_fast_work);
	return IRQ_HANDLED;
}
#endif /* GPIO_DETECT_HEADSET */

static void headset_worker(unsigned long event)
{
#ifdef CONFIG_HAS_WAKELOCK
	wake_lock_timeout(&micco_headset_lock, 5 * HZ);
#endif
	mask_hook_irq();
	printk(KERN_DEBUG "headset micco interrupt\n");
	if (headset_pdata && headset_pdata->remove)
		headset_pdata->remove();
	if( headset_pdata )
		mutex_lock(&headset_pdata->worker_mutex);
	cancel_delayed_work_sync(&headset_dwork);
	schedule_delayed_work(&headset_dwork, msecs_to_jiffies(1000));
	if( headset_pdata )
		mutex_unlock(&headset_pdata->worker_mutex);
	return ;
}

static void hookswitch_worker(unsigned long event)
{
	int headset_state;
	u8 val;
#ifdef CONFIG_HAS_WAKELOCK
	wake_lock_timeout(&micco_headset_lock, 5 * HZ);
#endif
	headset_state = atomic_read(&g_headset_state);
	if (!(headset_state & DETECTING)) {
		printk(KERN_DEBUG "hookswitch micco interrupt\n");
		micco_read(MICCO_STATUS_B, &val);
		if (val & MICCO_STATUS_B_HOOKSWITCH) {
			mod_timer(&hookswitch_down_timer, (jiffies + HZ / 3));
		} else {
			mod_timer(&hookswitch_up_timer, (jiffies + HZ / 3));
		}
	}

	return;

}

static void hookswitch_down_report(unsigned long data)
{
	printk(KERN_DEBUG "%s: hookswitch press\n", __func__);
	hookswitch_keyevent(1);
}

static void hookswitch_up_report(unsigned long data)
{
	printk(KERN_DEBUG "%s: hookswitch release!\n", __func__);
	hookswitch_keyevent(0);	
}

static int headset_on_line(void)
{
	u8 val;
	micco_read(MICCO_STATUS_B, &val);
	return ((val & MICCO_STATUS_B_HEADSET) ? 1 : 0);
}

#ifdef  CONFIG_GPIO_DETECT_HEADSET
static int init_gpio_detect(void)
{
	int ret;
	
	ret = gpio_request(headset_pdata->gpio_detect, "headset_gpio_det");
	if (ret < 0) {
		printk(KERN_ERR "request headset detect gpio failed\n");
		goto failed_request_headset_detect_gpio;
	}
	ret = gpio_direction_input(headset_pdata->gpio_detect);
	if (ret < 0) {
		printk(KERN_ERR "set headset gpio input failed\n");
		goto failed_set_detect_gpio_input;
	}
	headset_pdata->gpio_irq_num =	\
		gpio_to_irq(headset_pdata->gpio_detect);
	if (headset_pdata->gpio_irq_num < 0) {
		ret = headset_pdata->gpio_irq_num;
		goto failed_get_detect_gpio_irq_num;
	}
	ret = request_irq(headset_pdata->gpio_irq_num, gpio_interrupt,
			 IRQF_TRIGGER_RISING | IRQF_TRIGGER_FALLING, "headset__det", NULL);
	if (ret < 0) {
		printk(KERN_ERR "headset_pdata->gpio_detect failed\n");
		goto failed_request_detect_irq;
	}

	return ret;

failed_request_detect_irq:
failed_get_detect_gpio_irq_num:
failed_set_detect_gpio_input:
	gpio_free(headset_pdata->gpio_detect);
failed_request_headset_detect_gpio:
	return ret;

}
#endif

static int __init headset_detect_probe(struct platform_device *pdev)
{
	u8 val;
        int ret = 0;

	headset_pdata = pdev->dev.platform_data;
	
	micco_read(MICCO_IRQ_MASK_D, &val);
	val &= ~(IRQ_MASK_D_HEADSET | IRQ_MASK_D_HOOKSWITCH);
	micco_write(MICCO_IRQ_MASK_D, val);

	micco_read(MICCO_MISC, &val);
	val &= ~MICCO_MISC_REMCON_AUTO;
	val |= MICCO_MISC_REMCON_ENABLE | MICCO_MISC_REMCON_FILTER;
	micco_write(MICCO_MISC, val);
	atomic_set(&g_headset_state, DETECTING);

	INIT_DELAYED_WORK(&headset_dwork, headset_detect_workqueue);
	ret = pmic_callback_register(PMIC_EVENT_HEADSET, headset_worker);
	if (ret < 0) {
                printk(KERN_ERR "Cann't callback register headset.\n");
		return ret;
	}
	ret = pmic_callback_register(PMIC_EVENT_HOOKSWITCH, hookswitch_worker);
	if (ret < 0) {
		printk(KERN_ERR "cann't callback register hookswitch.\n");
		return ret;
	}
	init_timer(&hookswitch_down_timer);
	hookswitch_down_timer.function = hookswitch_down_report;
	init_timer(&hookswitch_up_timer);
	hookswitch_up_timer.function = hookswitch_up_report;
	if( headset_pdata )
		mutex_init(&headset_pdata->worker_mutex);
#ifdef  CONFIG_GPIO_DETECT_HEADSET
	INIT_WORK(&gpio_fast_work,gpio_work);
	ret = init_gpio_detect();
	if (ret) {
		printk(KERN_ERR "request headset detect gpio failed\n");
		return ret;
	}
#endif
	/* register accessory detection code */
	ret = accessory_register(&pdev->dev, &micco_headset_accessory);
	if (ret < 0) {
		printk(KERN_ERR "micco headset detect register accessory failed.\n");
		return ret;
	}
        if (headset_on_line()) {
                schedule_delayed_work(&headset_dwork, 0);
        }
	return ret;
}

static int __devexit headset_detect_remove(struct platform_device *dev)
{
	pmic_callback_unregister(PMIC_EVENT_HEADSET, headset_worker);
	pmic_callback_unregister(PMIC_EVENT_HOOKSWITCH, hookswitch_worker);
#ifdef  CONFIG_GPIO_DETECT_HEADSET
	gpio_free(headset_pdata->gpio_detect);
	free_irq(headset_pdata->gpio_irq_num, 0);
#endif
	return 0;
}

#ifdef CONFIG_PM

#ifdef CONFIG_MICCO_HEADSET_DETECT_DENOISE
void micco_headset_android_suspend_handler(struct early_suspend *h)
{
	return;
}

void micco_headset_android_resume_handler(struct early_suspend *h)
{
	return;
}

static struct early_suspend micco_headset_android_suspend = {
        .level = 5,
	.suspend = micco_headset_android_suspend_handler,
	.resume = micco_headset_android_resume_handler,
};
#endif

void micco_headset_android_late_suspend_handler(struct late_suspend *h)
{
	u8 val;
	u8 i;

	for (i = 0; i < 5; i++) {
		if (micco_read(MICCO_AUDIO_REG_BASE + i, &val) < 0)
			continue;
		/* If one of audio multiplexers is eabled, the headset detector will not be suspended  */
		if (val)
			return;
	}

	printk("%s reached\n", __FUNCTION__);
	micco_read(MICCO_MISC, &val);
	val &= ~MICCO_MISC_REMCON_ENABLE;
	val |= MICCO_MISC_REMCON_AUTO | MICCO_MISC_REMCON_FILTER;
	micco_write(MICCO_MISC, val);
	printk("%s returned\n", __FUNCTION__);
}

void micco_headset_android_late_resume_handler(struct late_suspend *h)
{
	u8 val;

	printk(KERN_DEBUG "%s reached\n", __func__);
	
	micco_read(MICCO_MISC, &val);
	val &= ~MICCO_MISC_REMCON_AUTO;
	val |= MICCO_MISC_REMCON_ENABLE | MICCO_MISC_REMCON_FILTER;
	micco_write(MICCO_MISC, val);
	
	printk(KERN_DEBUG "%s returned\n", __func__);
}

static struct late_suspend micco_headset_android_late_suspend = {
        .level = 5,
	.suspend = micco_headset_android_late_suspend_handler,
	.resume = micco_headset_android_late_resume_handler,
};

static int micco_headset_suspend(struct platform_device *pdev,
		  	pm_message_t state)
{
	return 0;
}

static int micco_headset_resume(struct platform_device *pdev)
{
	return 0;
}

#else

#define micco_headset_suspend NULL
#define micco_headset_resume NULL

#endif /* CONFIG_PM */


static struct platform_driver headset_detect_driver = {
	.probe = headset_detect_probe,
	.remove = __devexit_p(headset_detect_remove),
	.suspend = micco_headset_suspend,
	.resume	= micco_headset_resume,
	.driver = {
 		.name	= "micco-headset",
 	},
};

static int __init headset_detect_init(void)
{
	wake_lock_init(&micco_headset_lock, WAKE_LOCK_SUSPEND, "micco_headset");
#ifdef CONFIG_MICCO_HEADSET_DETECT_DENOISE
	register_early_suspend(&micco_headset_android_suspend);
#endif
	register_late_suspend(&micco_headset_android_late_suspend);
 	return platform_driver_register(&headset_detect_driver);
}

static void __exit headset_detect_exit(void)
{
	wake_lock_destroy(&micco_headset_lock);
#ifdef CONFIG_MICCO_HEADSET_DETECT_DENOISE
	unregister_early_suspend(&micco_headset_android_suspend);
#endif
	unregister_late_suspend(&micco_headset_android_late_suspend);
 	platform_driver_unregister(&headset_detect_driver);
}

module_init(headset_detect_init);
module_exit(headset_detect_exit);
