/*
 *  lis311dl.c - ST LIS311DL accelerometer driver
 *
 *  Copyright (C) 2007-2008 Yan Burman
 *  Copyright (C) 2008 Eric Piel
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/dmi.h>
#include <linux/module.h>
#include <linux/types.h>
#include <linux/platform_device.h>
#include <linux/interrupt.h>
#include <linux/input.h>
#include <linux/kthread.h>
#include <linux/semaphore.h>
#include <linux/delay.h>
#include <linux/wait.h>
#include <linux/poll.h>
#include <linux/freezer.h>
#include <linux/uaccess.h>
#include <linux/proc_fs.h>
#include <linux/i2c.h>
#include <linux/sensor-input.h>
#include <asm/atomic.h>
#include "lis311dl.h"

static struct i2c_client *g_client;

static int lis311dl_read(u8 reg, u8 *pval)
{
	int ret;
	int status;

	if (g_client == NULL)	/* No global client pointer? */
		return -1;
	ret = i2c_smbus_read_byte_data(g_client, reg);
	if (ret >= 0) {
		*pval = ret;
		status = 0;
	} else {
		status = -EIO;
	}

	return status;
}

static int lis311dl_write(u8 reg, u8 val)
{
	int ret;
	int status;

	if (g_client == NULL)	/* No global client pointer? */
		return -1;
	ret = i2c_smbus_write_byte_data(g_client, reg, val);
	if (ret == 0)
		status = 0;
	else
		status = -EIO;

	return status;
}

static inline void lis311dl_poweroff(void)
{
	/* disable X,Y,Z axis and power down */
	lis311dl_write(CTRL_REG1, 0x00);
}

static void lis311dl_poweron(void)
{
	u8 val;

	lis311dl_write(CTRL_REG1,
		CTRL1_PD1|CTRL1_PD0|CTRL1_Xen|CTRL1_Yen|CTRL1_Zen);

	lis311dl_write(FF_WU_CFG, 0);
	/*
	 * BDU: LSB and MSB values are not updated until both have been read.
	 *      So the value read will always be correct.
	 * IEN: Interrupt for free-fall and DD, not for data-ready.
	 */
	lis311dl_read(CTRL_REG2, &val);
	val |= CTRL2_BDU | CTRL2_IEN;
	lis311dl_write(CTRL_REG2, val);
}

void lis311dl_get_xyz(u8 *x, u8 *y, u8 *z)
{
	lis311dl_read(OUTX, x);
	lis311dl_read(OUTY, y);
	lis311dl_read(OUTZ, z);
}

static void lis311dl_report(struct input_dev *idev)
{
	u8 x, y, z;
	s8 status;
	lis311dl_read(STATUS_REG, &status);
	if((status & 0x7) == 0x7) {
		lis311dl_read(OUTX, &x);
		lis311dl_read(OUTY, &y);
		lis311dl_read(OUTZ, &z);

		printk(KERN_DEBUG "lis311dl: x = %d, y = %d, z = %d\n", x, y, z);
		input_report_abs(idev, ABS_X, x);
		input_report_abs(idev, ABS_Y, y);
		input_report_abs(idev, ABS_Z, z);
		input_sync(idev);
	}
}

#ifdef	CONFIG_PROC_FS
#define	LIS331DL_PROC_FILE	"driver/lis311dl"
static struct proc_dir_entry *lis311dl_proc_file;
static int index;

static ssize_t lis311dl_proc_read(struct file *filp,
		char *buffer, size_t length, loff_t *offset)
{
	u8 reg_val;

	if ((index < 0) || (index > 0x3f))
		return 0;

	lis311dl_read(index, &reg_val);
	printk(KERN_INFO "lis311dl: register 0x%x: 0x%x\n", index, reg_val);
	return 0;
}

static ssize_t lis311dl_proc_write(struct file *filp,
		const char *buff, size_t len, loff_t *off)
{
	u8 reg_val;
	char messages[256], vol[256];

	if (len > 256)
		len = 256;

	if (copy_from_user(messages, buff, len))
		return -EFAULT;

	if ('-' == messages[0]) {
		/* set the register index */
		memcpy(vol, messages+1, len-1);
		index = (int) simple_strtoul(vol, NULL, 16);
	} else {
	    /* set the register value */
		reg_val = (int)simple_strtoul(messages, NULL, 16);
		lis311dl_write(index, reg_val & 0xFF);
	}

	return len;
}

static struct file_operations lis311dl_proc_ops = {
	.read = lis311dl_proc_read,
	.write = lis311dl_proc_write,
};

static void create_lis311dl_proc_file(void)
{
	lis311dl_proc_file = create_proc_entry(LIS331DL_PROC_FILE, 0644, NULL);
	if (lis311dl_proc_file) {
		lis311dl_proc_file->owner = THIS_MODULE;
		lis311dl_proc_file->proc_fops = &lis311dl_proc_ops;
	} else
		printk(KERN_INFO "lis311dl: proc file create failed!\n");
}

static void remove_lis311dl_proc_file(void)
{
	extern struct proc_dir_entry proc_root;
	remove_proc_entry(LIS331DL_PROC_FILE, &proc_root);
}

#endif
static int __devinit lis311dl_probe(struct i2c_client *client,
			const struct i2c_device_id *id)
{
	int chip_id;

	chip_id = i2c_smbus_read_byte_data(client, 0x0F);
	if (chip_id < 0) {
		printk(KERN_ERR "lis311dl: unavailable!\n");
		g_client = NULL;
		return -ENXIO;
	} else {
		printk(KERN_ERR "lis311dl: chip id 0x%02x detected.\n", chip_id);
	}

	g_client = client;

	sensor_input_add(INPUT_G_SENSOR, "lis311dl", lis311dl_report, lis311dl_poweron,
			lis311dl_poweroff);

#ifdef	CONFIG_PROC_FS
	create_lis311dl_proc_file();
#endif
	return 0;
}

static int lis311dl_remove(struct i2c_client *client)
{
	sensor_input_del("lis311dl");

#ifdef	CONFIG_PROC_FS
	remove_lis311dl_proc_file();
#endif
	return 0;
}

static const struct i2c_device_id lis311dl_id[] = {
	{ "lis311dl", 0 },
	{ }
};

static struct i2c_driver lis311dl_driver = {
	.driver = {
		.name	= "lis311dl",
	},
	.id_table 	= lis311dl_id,
	.probe		= lis311dl_probe,
	.remove		= lis311dl_remove,
	.suspend	= NULL,
	.resume		= NULL,
};

static int __init lis311dl_i2c_init(void)
{
	return i2c_add_driver(&lis311dl_driver);
}

static void __exit lis311dl_i2c_exit(void)
{
	i2c_del_driver(&lis311dl_driver);
}

MODULE_DESCRIPTION("ST LIS311DL three-axis digital accelerometer (I2C) driver");
MODULE_AUTHOR("Bin Yang <bin.yang@marvell.com>");
MODULE_LICENSE("GPL");

module_init(lis311dl_i2c_init);
module_exit(lis311dl_i2c_exit);
