/*
 *  i2c_adap_pxa.c
 *
 *  I2C adapter for the PXA I2C bus access.
 *
 *  Copyright (C) 2002 Intrinsyc Software Inc.
 *  Copyright (C) 2004-2005 Deep Blue Solutions Ltd.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2 as
 *  published by the Free Software Foundation.
 *
 *
 *  History:
 *	Jul 2009 - [APSE]     Initial version for Linux-2.6.28 platform EVB3 PV2
 *                        with introduction to multiple instances supporting
 *	Jul 2009 - [PTK]      Use i2c_pxa_reset() in case of an error on I2C bus
 *	Aug 2009 - [PTK YANM] Reconsidering with sync to latest SAAR-platform code
 *		- Strict and multiple bus busy checking by (IBMR_IDLE ISR_EBB | ISR_IBB | ISR_UB)
 *		- i2c_pxa_abort() and i2c_pxa_unit_restart() to recover from error
 *		- udelay(I2C_ACCESS2_DELAY_US/2) after CLOCK setting !
 *		- do_xfer_retry with xfer_retries=1
 *		- "ADDR Arbitration Loss with MasterSTOP" handling
 *		- Delete cpu_is_pxa3xx related code and cpu_is_pxa9 differentiation in most cases except
 *		    No dvfm "CG" (Clock Gating mode) for cpu_is_pxa930 only
 *          Add differentiation for the IWCR
 *		- Adjust delays/retries for block/paging (more than one byte) read/write
 *		- Use ILCR/IWCR to adjust SCLK close to maxmium frequency
 *		- Parsing data-msg for debug and customization
 *		- CUSTOMIZATION_PER_SLAVE_DEVICE: SLOW, i2c_pxa_custom_scream_silent(), CUSTOM_BYTE2BYTE_DELAY
 *		- CONFIG_PROC_FS: to control slow/fast, customization and DEBUG_ERR_SIMULATE
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/i2c.h>
#include <linux/i2c-id.h>
#include <linux/init.h>
#include <linux/time.h>
#include <linux/sched.h>
#include <linux/delay.h>
#include <linux/errno.h>
#include <linux/interrupt.h>
#include <linux/i2c-pxa.h>
#include <linux/platform_device.h>
#include <linux/err.h>
#include <linux/clk.h>

#include <linux/proc_fs.h>
#include <asm/uaccess.h>

#include <mach/hardware.h>
#include <asm/irq.h>
#include <asm/io.h>
#include <mach/i2c.h>
#include <mach/pxa-regs.h>
//#include <mach/pxa2xx-gpio.h>
#ifdef CONFIG_PXA3xx
#include <mach/pxa3xx-regs.h>
#endif
#if defined(CONFIG_PXA3xx_DVFM)
#include <mach/dvfm.h>
#include <mach/pxa3xx_dvfm.h>
#include <asm/atomic.h>
#endif
#include <linux/wakelock.h>

/* Incorrect "Successful" request may cause BUS-ERROR failure for next request
* So save the request message in debug buffer; it will be printed as "previous"
* with failed "current" request.
*/
#define SAVE_MSG_REQUEST_DEBUG_BUF

#define ICR_START	(1 << 0)	   /* start bit */
#define ICR_STOP	(1 << 1)	   /* stop bit */
#define ICR_ACKNAK	(1 << 2)	   /* send ACK(0) or NAK(1) */
#define ICR_TB		(1 << 3)	   /* transfer byte bit */

#define ICR_MA		(1 << 4)	   /* master abort */
#define ICR_SCLE	(1 << 5)	   /* master clock enable */
#define ICR_IUE		(1 << 6)	   /* unit enable */
#define ICR_GCD		(1 << 7)	   /* general call disable */

#define ICR_ITEIE	(1 << 8)	   /* enable tx interrupts */
#define ICR_IRFIE	(1 << 9)	   /* enable rx interrupts */
#define ICR_BEIE	(1 << 10)	   /* enable bus error ints */
#define ICR_SSDIE	(1 << 11)	   /* slave STOP detected int enable */

#define ICR_ALDIE	(1 << 12)	   /* enable arbitration interrupt */
#define ICR_SADIE	(1 << 13)	   /* slave address detected int enable */
#define ICR_UR		(1 << 14)	   /* unit reset */
#define ICR_FM		(1 << 15)	   /* fast mode */
                        /*16*/
#define ICR_MSDIE	(1 << 17)	   /* master stop detected int enable */
#define ICR_MSDE	(1 << 18)	   /* master stop detected enable */

#define ISR_RWM		(1 << 0)	   /* read/write mode */
#define ISR_ACKNAK	(1 << 1)	   /* ack/nak status */
#define ISR_UB		(1 << 2)	   /* unit busy */
#define ISR_IBB		(1 << 3)	   /* bus busy */
#define ISR_SSD		(1 << 4)	   /* slave stop detected */
#define ISR_ALD		(1 << 5)	   /* arbitration loss detected */
#define ISR_ITE		(1 << 6)	   /* tx buffer empty */
#define ISR_IRF		(1 << 7)	   /* rx buffer full */
#define ISR_GCAD	(1 << 8)	   /* general call address detected */
#define ISR_SAD		(1 << 9)	   /* slave address detected */
#define ISR_BED		(1 << 10)	   /* bus error no ACK/NAK */
#define ISR_EBB		(1 << 11)	   /* early bus busy */
#define ISR_MSD		(1 << 12)	   /* master stop detected */

#define ICR_LOG      (i2c->icr_log)
#define ISR_LOG      (i2c->isr_log)
#define IBMR_LOG     (i2c->ibmr_log)
#define ENTRY_CNTR   (i2c->entry_cntr)
#define ERROR_CNTR   (i2c->error_cntr)


struct i2c_algo_custom {
	u8	slave_addr;
	u8	clock_adj;
	u8	byte2byte_delay;
	u8	scream_silent;
	u8	spare[5];
};

struct i2c_msg_debug {
	u16 addr;
	u16 flags;
	u16 len;
	u16 gap;
	u8  buf[4];
};

struct pxa_i2c {
	wait_queue_head_t	wait;
	struct i2c_msg		*msg;
	unsigned int		msg_num;
	unsigned int		msg_idx;
	unsigned int		msg_ptr;
	unsigned int		slave_addr;  /* This is "my Master's" but not remote address*/
	unsigned int		icr_backup;

	struct i2c_adapter	adap;
	struct clk		*clk;
#ifdef CONFIG_I2C_PXA_SLAVE
	struct i2c_slave_client *slave;
#endif

	void __iomem		*reg_base;

	unsigned long		iobase;
	unsigned long		iosize;

	int			irq;
	int			use_pio;
	int			flags;	/* specify i2c speed mode */

	int		access2_delay;
	int		busbusy_delay;

	u32				icr_log;
	u32				isr_log;
	u32				ibmr_log;
	u32				entry_cntr;
	u32				error_cntr;
    int				debug_track;
#ifdef SAVE_MSG_REQUEST_DEBUG_BUF
    int						debug_num_msgs;
	struct i2c_msg_debug	debug_msgs[2];
#endif
	/* Parse data-msg and Customization per remote slave request */
	u8		req_slave_addr;
	char	req_read;			/* 'R' read or 'W' write */
	u8		req_num_data_bytes;
	u8		align;
	struct i2c_algo_custom *p_custom;
};

#define _IBMR(i2c)	((i2c)->reg_base + 0)
#define _IDBR(i2c)	((i2c)->reg_base + 8)
#define _ICR(i2c)	((i2c)->reg_base + 0x10)
#define _ISR(i2c)	((i2c)->reg_base + 0x18)
#define _ISAR(i2c)	((i2c)->reg_base + 0x20)
#define _ILCR(i2c)	((i2c)->reg_base + 0x28)
#define _IWCR(i2c)	((i2c)->reg_base + 0x30)

/*
 * I2C Slave mode address for this controller but not remote slave
 */
#define I2C_PXA_SLAVE_ADDR      0x1

/*************************** DEBUG ***********************************/
/** DBG_GPIO
*
*  The GPIOS are correct for SAAR platform only and may differe on otheer platforms
*
* GPIO:52 is used for the debugging on SAAR board (J9 B27)
* For GPIOs <63:32>
*    MFPR=0x40e102C4: 0xA840
*    Registers are GPDRx where x=1
*    Bit_in_reg=52-32=20   = (1<<20)
*    GSDR1=0x40E00404  SET  Direction 1=Output
*    GPDR1=0x40E00010       Direction to Read DIR status
*    GPSR1=0x40E0001C  SET  Output to 1   (D10 led ON)
*    GPCR1=0x40E00028  CLR  Output to 0   (D10 led OFF)
*---------------
* GPIO:51: bit 19 = (1<<19)  ; MFPR=0x40e102C0 (J9 B23)
**/
//#define DBG_GPIO
#ifdef  DBG_GPIO

void DBG_GPIO_CFG(void) {
 if(__raw_readl((void *)&(__REG(0x40e102C4))) != 0x0000A840)
    __raw_writel(0x0000A840, (void *)&(__REG(0x40e102C4)));  //MFPR
 if(! (__raw_readl((void *)&(__REG(0x40E00010))) & (1<<20)))
    __raw_writel((1<<20), (void *)&(__REG(0x40E00404)));  //Direction SET

 if(     (__raw_readl((void *)&(__REG(0x40e102C0))) != 0x0000A840)
   || (! (__raw_readl((void *)&(__REG(0x40E00010))) & (1<<19))) )
 {
    __raw_writel(0x0000A840, (void *)&(__REG(0x40e102C0)));  //MFPR
    __raw_writel((1<<19), (void *)&(__REG(0x40E00404)));  //Direction SET
    __raw_writel((1<<19), (void *)&(__REG(0x40E00028)));  //init-state off
 }
}

#define DBG_GPIO_DRV_IN()    //__raw_writel((1<<20), (void *)&(__REG(0x40E00028))) //off
#define DBG_GPIO_DRV_OUT()   //__raw_writel((1<<20), (void *)&(__REG(0x40E0001C))) //on
//#define DBG_GPIO_IRQH_IN()   __raw_writel((1<<20), (void *)&(__REG(0x40E0001C))) //on
//#define DBG_GPIO_IRQH_OUT()  __raw_writel((1<<20), (void *)&(__REG(0x40E00028))) //off

#define DBG_GPIO_IRQH_IN()   __raw_writel((1<<20), (void *)&(__REG(0x40E00028))) //off
#define DBG_GPIO_IRQH_OUT()  __raw_writel((1<<20), (void *)&(__REG(0x40E0001C))) //on

#define DBG_GPIO_ERR_ON()    __raw_writel((1<<19), (void *)&(__REG(0x40E0001C))) //on
#define DBG_GPIO_ERR_OFF()   __raw_writel((1<<19), (void *)&(__REG(0x40E00028))) //off
#define DBG_GPIO_WAIT_ON()  // __raw_writel((1<<19), (void *)&(__REG(0x40E0001C))) //on
#define DBG_GPIO_WAIT_OFF() // __raw_writel((1<<19), (void *)&(__REG(0x40E00028))) //off

#else//DBG_GPIO

#define DBG_GPIO_CFG()
#define DBG_GPIO_DRV_IN()
#define DBG_GPIO_DRV_OUT()
#define DBG_GPIO_IRQH_IN()
#define DBG_GPIO_IRQH_OUT()
#define DBG_GPIO_ERR_ON()
#define DBG_GPIO_ERR_OFF()
#define DBG_GPIO_WAIT_ON()
#define DBG_GPIO_WAIT_OFF()

#endif//DBG_GPIO

//#define DBG_MIPSRAM    /* Platform INDEPENDENT */
#ifdef DBG_MIPSRAM
#include <linux/mipsram.h>
//To STOP logging use stop OLD/NEW and then place END marker as below
//    MIPSRAM_get_descriptor()->buffer = NULL;        for OLD versions to stop logging
//    MIPSRAM_get_descriptor()->logging_stopped = 1;  for NEW versions to stop logging
//  MIPS_RAM_MARK_END_OF_LOG(MIPSRAM_get_descriptor(), 0xCAFE);
// The user events are also possible as: void MIPSRAM_Trace(unsigned long id);
#endif//DBG_MIPSRAM



#ifdef DEBUG

struct bits {
	u32	mask;
	const char *set;
	const char *unset;
};
#define PXA_BIT(m, s, u)	{ .mask = m, .set = s, .unset = u }

static inline void
decode_bits(const char *prefix, const struct bits *bits, int num, u32 val)
{
	printk("%s %08x: ", prefix, val);
	while (num--) {
		const char *str = val & bits->mask ? bits->set : bits->unset;
		if (str)
			printk("%s ", str);
		bits++;
	}
}

static const struct bits isr_bits[] = {
	PXA_BIT(ISR_RWM,	"RX",		"TX"),
	PXA_BIT(ISR_ACKNAK,	"NAK",		"ACK"),
	PXA_BIT(ISR_UB,		"Bsy",		"Rdy"),
	PXA_BIT(ISR_IBB,	"BusBsy",	"BusRdy"),
	PXA_BIT(ISR_EBB,	"EarBusBsy",	"BusRdy"),
	PXA_BIT(ISR_SSD,	"SlaveStop",	NULL),
	PXA_BIT(ISR_ALD,	"ALD",		NULL),
	PXA_BIT(ISR_ITE,	"TxEmpty",	NULL),
	PXA_BIT(ISR_IRF,	"RxFull",	NULL),
	PXA_BIT(ISR_GCAD,	"GenCall",	NULL),
	PXA_BIT(ISR_SAD,	"SlaveAddr",	NULL),
	PXA_BIT(ISR_BED,	"BusErr",	NULL),
};

static void decode_ISR(unsigned int val)
{
	decode_bits(KERN_DEBUG "ISR", isr_bits, ARRAY_SIZE(isr_bits), val);
	printk("\n");
}

static const struct bits icr_bits[] = {
	PXA_BIT(ICR_START,  "START",	NULL),
	PXA_BIT(ICR_STOP,   "STOP",	NULL),
	PXA_BIT(ICR_ACKNAK, "ACKNAK",	NULL),
	PXA_BIT(ICR_TB,     "TB",	NULL),
	PXA_BIT(ICR_MA,     "MA",	NULL),
	PXA_BIT(ICR_SCLE,   "SCLE",	"scle"),
	PXA_BIT(ICR_IUE,    "IUE",	"iue"),
	PXA_BIT(ICR_GCD,    "GCD",	NULL),
	PXA_BIT(ICR_ITEIE,  "ITEIE",	NULL),
	PXA_BIT(ICR_IRFIE,  "IRFIE",	NULL),
	PXA_BIT(ICR_BEIE,   "BEIE",	NULL),
	PXA_BIT(ICR_SSDIE,  "SSDIE",	NULL),
	PXA_BIT(ICR_ALDIE,  "ALDIE",	NULL),
	PXA_BIT(ICR_SADIE,  "SADIE",	NULL),
	PXA_BIT(ICR_UR,     "UR",		"ur"),
};

#ifdef CONFIG_I2C_PXA_SLAVE
static void decode_ICR(unsigned int val)
{
	decode_bits(KERN_DEBUG "ICR", icr_bits, ARRAY_SIZE(icr_bits), val);
	printk("\n");
}
#endif

static unsigned int i2c_debug = DEBUG;

static void i2c_pxa_show_state(struct pxa_i2c *i2c, int lno, const char *fname)
{
	dev_dbg(&i2c->adap.dev, "state:%s:%d: ISR=%08x, ICR=%08x, IBMR=%02x\n", fname, lno,
		readl(_ISR(i2c)), readl(_ICR(i2c)), readl(_IBMR(i2c)));
}

#define show_state(i2c) i2c_pxa_show_state(i2c, __LINE__, __FUNCTION__)

#else /* DEBUG ***********************************/

#define i2c_debug	0

#define show_state(i2c) do { } while (0)
#define decode_ISR(val) do { } while (0)
#define decode_ICR(val) do { } while (0)

#endif /* DEBUG ***********************************/

/*****************************************************************
*         Configuration MODEs and Time/Retries
******************************************************************
*/
#define I2C_BUS_ERROR             (-6000)
#define I2C_NULL_BUFFER           (-4000)
#define I2C_TIMEOUT_DONT_RETRY    (-3000)
#define I2C_MASTER_ABORT          (-3010)

//writing 1 to ISR status bits to clear them
#define I2C_ISR_CLEAR  (ISR_SSD | ISR_ALD | ISR_ITE | ISR_IRF | ISR_GCAD | ISR_SAD | ISR_BED | ISR_MSD)
#define I2C_IBMR_IDLE   3 /* IBMR[SCL:1] and IBMR[SDA:0] are High */

//===== The modes below are configured from upper layers
// If the value is different from "0"-default we Force to specified mode
#define HANDLER_MODE__DFLT_PIO_IRQ     0 /* 0: Default ; 1: Polling only; 2: IRQ only */
#define HANDLER_MODE__DFLT_SLOW_FAST   0 /* 0: Default ; 1: Slow only ;   2: Fast only */


/* Delays are accounted for FAST 400kHz mode, for SLOW 100kHz the FAST_SLOW_RATE is used
 * More than one byte may be read/write by second Master in block/paging mode => MAX_TRANSACT_TIME_US
 */
#define MAX_XFER_TIMEOUT_US       80000   // 80mSec guard after successfull start
#define MAX_TRANSACT_TIME_US       1200   // 3+(4*2)+1=12 *10_SlowCycles = 120*10uS
#define FAST_SLOW_RATE            (400/100)
#define I2C_ACCESS2_DELAY_US      (5+1)  // 2 cycles * 1/400kHz  =  5usec
#define I2C_BUSBUSY_DELAY_US        80   //RETRY: Shortest transaction 26bits. The time is 1/400kHz*26=65usec
#define NUM_RETRIES_ON_BUS_BUSY    125   // (3ctrl+4data bytes)*9bits / 2bitsAccessDelay * FAST_SLOW_RATE
#define NUM_RETRIES_EXHAUSTED        3


#define OSCR_MASK		(~1UL)

#define CUSTOMIZATION_PER_SLAVE_DEVICE

#if !defined (CUSTOMIZATION_PER_SLAVE_DEVICE)
#define i2c_pxa_msg_parse(A,B,C)            /* */
#define i2c_pxa_custom_scream_silent(I2C)   (0)
#define i2c_pxa_custom_clock_adj(I2C)      I2C->icr_backup
#define CUSTOM_BYTE2BYTE_DELAY(I2C)         /* */
#else
/**********************************************************************
*        CUSTOMIZATION PER SLAVE DEVICE
***********************************************************************/
#define MAX_CUSTOMIZED_SLAVE_DEVICES   4
/*------------ {{{ CUSTOMER DATA --------------*/
struct i2c_algo_custom  i2c_pxa_custom[MAX_CUSTOMIZED_SLAVE_DEVICES+1] = 
{
/*	slave_addr	clock_adj	byte2byte_delay	scream_silent	spare[5]	*/
	{ 0x34,		0,			0,				0xFA,			{0}	}, /* PMIC */
	{0} /*ZERO line as END must be present, index=MAX_CUSTOMIZED_SLAVE_DEVICES */
};

int i2c_custom_ForceClck(u8 slave_addr, u8 set_none0_slow1_fast2)
{
	int i = 0;
	if(i2c_pxa_custom[MAX_CUSTOMIZED_SLAVE_DEVICES-1].slave_addr != 0) {
		printk("No free place for cutomization found\n");
		return -1;
	}
	while (i2c_pxa_custom[i].slave_addr != 0) {
		if(i2c_pxa_custom[i].slave_addr == slave_addr) {
			i2c_pxa_custom[i].clock_adj = set_none0_slow1_fast2;
			return 0;
		}
		i++;
	}
	i2c_pxa_custom[i].slave_addr = slave_addr;
	i2c_pxa_custom[i].clock_adj  = set_none0_slow1_fast2;
	return 0;
}
/*-------------- }}} END CUSTOMER DATA ---------*/

static void i2c_pxa_msg_parse(struct pxa_i2c *i2c, struct i2c_msg *msg, int num)
{
	int i =  (num > 1)  ?  1 : 0;

	i2c->req_slave_addr     = msg[i].addr & 0x7f;
	i2c->req_read           =(msg[i].flags & I2C_M_RD) ? 'R' : 'W';
	i2c->req_num_data_bytes = msg[i].len;
}

#if !defined(SAVE_MSG_REQUEST_DEBUG_BUF)
#define i2c_pxa_msg_save(i2c, msg, num)
#else
static void i2c_pxa_msg_save(struct pxa_i2c *i2c, struct i2c_msg *msg, int num)
{
	struct i2c_msg_debug *p;

	i2c->debug_num_msgs = num;
	/* struct i2c_msg: (u16_addr, u16_flags), (u16_len, u16), u32_bufP */
	p = i2c->debug_msgs;
	p[0].addr  = msg[0].addr;
	p[0].flags = msg[0].flags;
	p[0].len   = msg[0].len;
	memcpy(&(p[0].buf[0]), (u8 *)msg[0].buf, 4);

	if (num > 1) {
		p[1].addr  = msg[1].addr;
		p[1].flags = msg[1].flags;
		p[1].len   = msg[1].len;
		memcpy(&(p[1].buf[0]), (u8 *)msg[1].buf, 4);
	}
}
#endif/*SAVE_MSG_REQUEST_DEBUG_BUF*/

static void i2c_pxa_msg_print(struct pxa_i2c *i2c, struct i2c_msg *msg, int num)
{
	u8  *bufP;
	char text[80];
	int  offs;
	int  i, k, len;

	printk(KERN_ERR "i2c: Current: num_messages=%d \n", num);
	num = (num > 1) ? 2 : 1;

	for (k = 0; k < num; k++) {
		bufP = msg[k].buf;
		offs = sprintf(text, "i2c:   slave_%02x, flags_%04x, len_%d:",
								msg[k].addr, msg[k].flags, msg[k].len);
		len = (msg[k].len <= 4) ? msg[k].len : 4;
		for (i = 0; i < len; i++)
			offs += sprintf(text+offs, " %02x", bufP[i]);
		printk(KERN_ERR "%s \n", text);
	}

#ifdef SAVE_MSG_REQUEST_DEBUG_BUF
  {
	struct i2c_msg_debug *msgD = i2c->debug_msgs;

	printk(KERN_ERR "i2c: Previous: num_messages=%d \n", i2c->debug_num_msgs);
	num = (i2c->debug_num_msgs > 1) ? 2 : 1;

	for (k = 0; k < num; k++) {
		bufP = &(i2c->debug_msgs[k].buf[0]);
		offs = sprintf(text, "i2c:   slave_%02x, flags_%04x, len_%d:",
								msgD[k].addr, msgD[k].flags, msgD[k].len);
		len = (msgD[k].len <= 4) ? msgD[k].len : 4;
		for (i = 0; i < len; i++)
			offs += sprintf(text+offs, " %02x", bufP[i]);
		printk(KERN_ERR "%s \n", text);
	}
  }
#endif/*SAVE_MSG_REQUEST_DEBUG_BUF*/
	printk(KERN_ERR "\n");
}


static struct i2c_algo_custom* i2c_pxa_custom_hit(struct pxa_i2c *i2c)
{
	struct i2c_algo_custom *p = i2c->p_custom;

	if(p != NULL) {
		while (p->slave_addr != 0) {
			if(p->slave_addr == i2c->req_slave_addr)
				break;
			p++;
		}
	}
	return p;
}

/* There are some specific addresses which do NOT give an ACK:
* SlaveAddr=0x34, the operation is write (SlaveAddr<<1), and the write applied into 0xFA..FF.
* Check them and do NOT print error report
*/
static int i2c_pxa_custom_scream_silent(struct pxa_i2c *i2c)
{
	struct i2c_algo_custom *p = i2c_pxa_custom_hit(i2c);

	if(p == NULL) return 0;

	if(p->scream_silent) {
		if(i2c->msg->buf[0/*i2c->msg_ptr*/] >= p->scream_silent)
			return 1;	/* Action */
	}
	return 0;
}

static int i2c_pxa_custom_clock_adj(struct pxa_i2c *i2c)
{
	struct i2c_algo_custom *p = i2c_pxa_custom_hit(i2c);

	if(p == NULL) return i2c->icr_backup;

	if(p->clock_adj == 1) {
		return (i2c->icr_backup & ~((1 << 16) | (1 << 15)) ); /* set SLOW=100k */
	} else if(p->clock_adj == 2) {
		return (i2c->icr_backup | (1 << 15) );                /* set FAST=400k */
	}
	return i2c->icr_backup; /* not changed */
}

#ifndef CUSTOM_BYTE2BYTE_DELAY_IN_USE
#define CUSTOM_BYTE2BYTE_DELAY(I2C)    /* */
#else
#define CUSTOM_BYTE2BYTE_DELAY(I2C)    i2c_pxa_custom_byte2byte_delay(I2C)
static void i2c_pxa_custom_byte2byte_delay(struct pxa_i2c *i2c)
{
	struct i2c_algo_custom *p = i2c_pxa_custom_hit(i2c);

	if(p == NULL) return;

	if(p->byte2byte_delay) {
		udelay(100);
	}
	return;
}
#endif//CUSTOM_BYTE2BYTE_DELAY

/*** END CUSOTMIZATION DATA AND PROCEDURES *****************************/
#endif/*CUSTOMIZATION_PER_SLAVE_DEVICE*/


#ifdef	CONFIG_PROC_FS
#define	I2C_PXA_PROC_FILE          "driver/i2c-pxa"
#define	I2C_PXA_PROC_NUM_DEVS      3
static struct proc_dir_entry *i2c_pxa_proc_file;
static struct pxa_i2c *proc_i2c[I2C_PXA_PROC_NUM_DEVS];
static int    err_simulate;

/*#define DEBUG_HW_STOP_ON_ERR*/
#ifdef DEBUG_HW_STOP_ON_ERR
static int    stop_on_error_req;
static int    stop_on_error;
#endif

#define DEBUG_ERR_SIMULATE       0xF9 /*Non-exist slave device address */
#if defined(DEBUG_ERR_SIMULATE)
static void i2c_pxa_buserr_simulate(struct i2c_msg *msg, int num)
{
  if(err_simulate) {
	int i =  (num > 1)  ?  1 : 0;
	msg[i].addr = DEBUG_ERR_SIMULATE;
	err_simulate--;
  }
}
#endif

static ssize_t i2c_pxa_proc_read(struct file *filp,
		char *buffer, size_t count, loff_t *offset)
{
	return 0;
}

static ssize_t i2c_pxa_proc_write(struct file *filp,
		const char *buff, size_t len, loff_t *off)
{
	char messages[256];

	if (len > 256)
		len = 256;

	if (copy_from_user(messages, buff, len))
		return -EFAULT;

	if ('?' == messages[0]) {
		/* echo ? > /proc/driver/i2c-pxa */
		printk("\n possible commands:\n"
		         "    i2c0slow .. i2c2slow  force I2C clock to 100kHz\n"
		         "    i2c0fast .. i2c2fast  force I2C clock to 400kHz\n"
#if defined (CUSTOMIZATION_PER_SLAVE_DEVICE)
				 "   Force clock only for specified NN hexAddress Slave-Device\n"
				 "    slave.fast_NN   force  to 400kHz\n"
				 "    slave.slow_NN   force  to 100kHz\n"
				 "    slave.nocl_NN   NO force\n"
#endif
		         "    err1       simulate one Bus-Error=accessToNonExist\n"
				 "               use also  echo 1 > /dev/i2c-0 or 1\n"
#ifdef DEBUG_HW_STOP_ON_ERR
				 "    stop       DEBUG: drop all I2C requests and reset CP upon ERROR\n"
#endif
				 );
	} else if (strncmp(messages, "i2c0slow", 5) == 0) {
		proc_i2c[0]->icr_backup &= ~(1 << 15);
	} else if (strncmp(messages, "i2c0fast", 5) == 0) {
		proc_i2c[0]->icr_backup |= (1 << 15);
	} else if (strncmp(messages, "i2c1slow", 5) == 0) {
		proc_i2c[1]->icr_backup &= ~(1 << 15);
	} else if (strncmp(messages, "i2c1fast", 5) == 0) {
		proc_i2c[1]->icr_backup |= (1 << 15);
	} else if (strncmp(messages, "i2c2slow", 5) == 0) {
		proc_i2c[2]->icr_backup &= ~(1 << 15);
	} else if (strncmp(messages, "i2c2fast", 5) == 0) {
		proc_i2c[2]->icr_backup |= (1 << 15);
#if defined (CUSTOMIZATION_PER_SLAVE_DEVICE)
	} else if (strncmp(messages, "slave.",6) == 0) {
		u8   clck_op=0xFF; /*invalid*/
		if (strncmp(messages, "slave.fast_",11) == 0)
			clck_op = 2;
		else
		if (strncmp(messages, "slave.slow_",11) == 0)
			clck_op = 1;
		else
		if (strncmp(messages, "slave.nocl_",11) == 0)
			clck_op = 0;

		if (clck_op != 0xFF) {
			u8   slave_addr;
			char vol[256];
			memcpy(vol, messages+11, len-11);
			slave_addr = (u8)simple_strtoul(vol, NULL, 16);
			if(slave_addr == 0) {
				clck_op = 0xFF;
			} else {
				printk("\n Force clock for slave device address 0x%X\n", slave_addr);
				i2c_custom_ForceClck(slave_addr, clck_op);
			}
		}
		if (clck_op == 0xFF)
			printk("\n Incorrect format. Use ? for help\n");
#endif/*CUSTOMIZATION_PER_SLAVE_DEVICE*/
	} else if (strncmp(messages, "err1", 4) == 0) {
		err_simulate = 1;
#ifdef DEBUG_HW_STOP_ON_ERR
	} else if (strncmp(messages, "stop", 4) == 0) {
		stop_on_error_req = 1;
#endif
	} else {
		printk("\n Unknown command\n");
	}
	return len;
}

static struct file_operations i2c_pxa_proc_ops = {
	.read  = i2c_pxa_proc_read,
	.write = i2c_pxa_proc_write,
};

static void create_i2c_pxa_proc_file(struct pxa_i2c *i2c, int id)
{
	if(id >= I2C_PXA_PROC_NUM_DEVS)
		return;

	proc_i2c[id] = i2c;

	if(i2c_pxa_proc_file != NULL) /* already done */
		return;

	i2c_pxa_proc_file = create_proc_entry(I2C_PXA_PROC_FILE, 0644, NULL);
	if (i2c_pxa_proc_file) {
		i2c_pxa_proc_file->owner = THIS_MODULE;
		i2c_pxa_proc_file->proc_fops = &i2c_pxa_proc_ops;
	} else
		printk(KERN_INFO "proc file create failed!\n");
}

static void remove_i2c_pxa_proc_file(int id)
{
	if((id != 0) || (i2c_pxa_proc_file == NULL))
		return;

	remove_proc_entry(I2C_PXA_PROC_FILE, NULL);
	i2c_pxa_proc_file = NULL;
}

#endif /*CONFIG_PROC_FS*/



static unsigned int read_time(void)
{
#ifdef CONFIG_PXA_32KTIMER
	return OSCR4;
#else
	return OSCR;
#endif
}


#ifdef CONFIG_PXA3xx_DVFM
static int dvfm_dev_idx;

static void set_dvfm_constraint(void)
{
	/* Disable Low power mode */
	dvfm_disable_op_name("D1", dvfm_dev_idx);
	dvfm_disable_op_name("D2", dvfm_dev_idx);
	if (!cpu_is_pxa930()) /* except the Tavor-P */
		dvfm_disable_op_name("CG", dvfm_dev_idx);
	/* make sure enough delay after exiting from D2 */
	udelay(10);
}

static void unset_dvfm_constraint(void)
{
	/* Enable Low power mode */
	dvfm_enable_op_name("D1", dvfm_dev_idx);
	dvfm_enable_op_name("D2", dvfm_dev_idx);
	if (!cpu_is_pxa930()) /* except the Tavor-P */
		dvfm_enable_op_name("CG", dvfm_dev_idx);
}

#else
static void set_dvfm_constraint() {}
static void unset_dvfm_constraint() {}
#endif

static struct wake_lock pxa_i2c_suspend_lock;

#define eedbg(lvl, x...) do { if ((lvl) < 1) { printk(KERN_DEBUG "" x); } } while(0)

static void i2c_pxa_master_complete(struct pxa_i2c *i2c, int ret);
static irqreturn_t i2c_pxa_handler(int this_irq, void *dev_id);


static void i2c_pxa_scream_blue_murder(struct pxa_i2c *i2c, const char *whyTxt)
{
    ERROR_CNTR++;

	printk(KERN_ERR "i2c: Time[%u] %s \n", read_time(), whyTxt);
#ifdef CONFIG_I2C_DEBUG_DEVICE_BUSY
	printk(KERN_DEBUG "Current:%s[%d]\n", current->comm, current->pid);
	printk("i2c: last_device = 0x%x,current_device = 0x%x\n",
	       i2c->adap.last_device, i2c->adap.current_device);
#endif	
printk(KERN_ERR "i2c: %s:%x, slave_0x%X, %c-operation, %d data-bytes\n",
	         i2c->adap.name,(unsigned int)i2c, i2c->req_slave_addr, i2c->req_read, i2c->req_num_data_bytes);
	printk(KERN_ERR "i2c: msg_num=%d, msg_idx=%d, msg_ptr=%d;  track=%d, Entry/Err=%d/%d\n",
		            i2c->msg_num, i2c->msg_idx, i2c->msg_ptr,  i2c->debug_track, ENTRY_CNTR, ERROR_CNTR);
	printk(KERN_ERR "i2c: logged  icr=%08x isr=%08x ibmr=%x\n",
                                  ICR_LOG, ISR_LOG, IBMR_LOG);
    printk(KERN_ERR "i2c: current ICR=%08x ISR=%08x IBMR=%x\n",
	                      readl(_ICR(i2c)), readl(_ISR(i2c)), readl(_IBMR(i2c)) );
	printk(KERN_ERR "\n");
    i2c->debug_track=0;
}

static inline int i2c_pxa_is_slavemode(struct pxa_i2c *i2c)
{
	return !(readl(_ICR(i2c)) & ICR_SCLE);
}

static void i2c_pxa_abort(struct pxa_i2c *i2c)
{
	unsigned long icr;

	if (i2c_pxa_is_slavemode(i2c)) {
		dev_dbg(&i2c->adap.dev, "%s: called in slave mode\n", __func__);
		return;
	}

	if (readl(_IBMR(i2c)) == I2C_IBMR_IDLE)
		return;

	if( readl(_ISR(i2c)) & ISR_MSD )
	{
		writel(ISR_MSD, _ISR(i2c));
		return;
	}

	if(readl(_ICR(i2c)) & ICR_START)
	{  //Under START state the MA-bit cannot be used
		unsigned long timeout = jiffies + HZ/4;
		while (time_before(jiffies, timeout) && (readl(_IBMR(i2c)) & 0x1) == 0)
		{
			icr = readl(_ICR(i2c));

			icr &= ~ICR_START;
			icr |= ICR_ACKNAK | ICR_STOP | ICR_TB;

			writel(icr, _ICR(i2c));
			udelay(i2c->access2_delay);
		}
	}
	else
	{ //MA Master Abort bit used
		icr = readl(_ICR(i2c));
		icr &= ~(ICR_START | ICR_ACKNAK | ICR_STOP | ICR_TB);
		icr |= ICR_MA;
		writel(I2C_ISR_CLEAR,  _ISR(i2c));
		writel(icr, _ICR(i2c));
		udelay(i2c->access2_delay);
	}
	writel(readl(_ICR(i2c)) & ~(ICR_MA | ICR_START | ICR_STOP), _ICR(i2c));
	if( in_interrupt() )
		mdelay(1);
	else
		msleep(1); /*udelay(MAX_TRANSACT_TIME_US);*/
}


//=== Check Bus-Busy - 2 procedures:
//  is_bus_busy()
//  is_busy_now() =  is_bus_busy() + read_ISR - ISR_UB
//
#ifdef CONFIG_I2C_PXA_SLAVE
static int is_bus_busy(u32 isr)
{
	int ret = 0;

	if (isr & (ISR_IBB | ISR_UB | ISR_EBB))
		ret = 1;
	return ret;
}

static int is_busy_now(struct pxa_i2c *i2c)
{
	u32 isr;
	int ret = 0;

	isr = readl(_ISR(i2c));
	if (isr & (ISR_IBB | ISR_EBB))
		ret = 1;
	return ret;
}


static int i2c_pxa_set_master(struct pxa_i2c *i2c)
{
	u32 isr;

	if (i2c_debug)
		dev_dbg(&i2c->adap.dev, "setting to bus master\n");

	isr = readl(_ISR(i2c));
	if (is_bus_busy(isr)) {
		dev_dbg(&i2c->adap.dev, "%s: unit is busy\n", __func__);
		if (i2c_debug > 1)
			dev_dbg(&i2c->adap.dev, "%s: %ld: ISR=%08x, ICR=%08x, IBMR=%02x\n",
				__func__, (long)jiffies, readl(_ISR(i2c)),
				readl(_ICR(i2c)), readl(_IBMR(i2c)));

		if (isr & ISR_SAD) {
			if (i2c_debug > 0)
				dev_dbg(&i2c->adap.dev, "%s: Slave detected\n",
					__func__);
		}
		return I2C_RETRY;
	} else if (readl(_IBMR(i2c)) == I2C_IBMR_IDLE) {
		/* wait for unit and bus being not busy, and we also do a
		 * quick check of the i2c lines themselves to ensure they've
		 * gone high...
		 */
		if (i2c_debug > 0)
			dev_dbg(&i2c->adap.dev, "%s: done\n", __func__);
		/* set as master */
		writel(readl(_ICR(i2c)) | ICR_SCLE, _ICR(i2c));
		return 0;
	}
	if (i2c_debug > 0)
		dev_dbg(&i2c->adap.dev, "%s: did not free\n", __func__);

	return I2C_RETRY;
}

static int i2c_pxa_wait_slave(struct pxa_i2c *i2c)
{
	unsigned long timeout = jiffies + HZ*1;
	u32 isr;

	/* wait for stop */

	show_state(i2c);

	while (time_before(jiffies, timeout)) {
		if (i2c_debug > 1)
			dev_dbg(&i2c->adap.dev, "%s: %ld: ISR=%08x, ICR=%08x, IBMR=%02x\n",
				__func__, (long)jiffies, readl(_ISR(i2c)), readl(_ICR(i2c)), readl(_IBMR(i2c)));

		isr = readl(_ISR(i2c));
		if (!is_bus_busy(isr) || (isr & ISR_SAD)
			|| ((readl(_ICR(i2c)) & ICR_SCLE) == 0)) {
			if (i2c_debug > 1)
				dev_dbg(&i2c->adap.dev, "%s: done\n", __func__);
			return 1;
		}

		msleep(1);
	}

	if (i2c_debug > 0)
		dev_dbg(&i2c->adap.dev, "%s: did not free\n", __func__);
	return 0;
}

/*
 * clear the hold on the bus, and take of anything else
 * that has been configured
 */
static void i2c_pxa_set_slave(struct pxa_i2c *i2c, int errcode)
{
	show_state(i2c);

	if (errcode < 0) {
		udelay(i2c->access2_delay);   /* simple delay */
	} else {
		/* we need to wait for the stop condition to end */

		/* if we where in stop, then clear... */
		if (readl(_ICR(i2c)) & ICR_STOP) {
			udelay(i2c->access2_delay);
			writel(readl(_ICR(i2c)) & ~ICR_STOP, _ICR(i2c));
		}

		if (!i2c_pxa_wait_slave(i2c)) {
			dev_err(&i2c->adap.dev, "%s: wait timedout\n",
				__func__);
			return;
		}
	}

	writel(readl(_ICR(i2c)) & ~(ICR_STOP|ICR_ACKNAK|ICR_MA), _ICR(i2c));
	writel(readl(_ICR(i2c)) & ~ICR_SCLE, _ICR(i2c));

	if (i2c_debug) {
		dev_dbg(&i2c->adap.dev, "ICR now %08x, ISR %08x\n", readl(_ICR(i2c)), readl(_ISR(i2c)));
		decode_ICR(readl(_ICR(i2c)));
	}
}
#else
#define i2c_pxa_set_slave(i2c, err)	do { } while (0)
#endif//CONFIG_I2C_PXA_SLAVE


static void i2c_pxa_unit_restart(struct pxa_i2c *i2c, int full)
{
	if(full)
	{
		// If bus is not IDLE, abort any transfer currently under way (should send STOP)
		i2c_pxa_abort(i2c);
	}
	//Exec reset according to 9.8, but KEEP in reset
	writel(ICR_UR, _ICR(i2c));
	writel(I2C_ISR_CLEAR, _ISR(i2c));
	writel( 0 , _ICR(i2c));
	writel(i2c->slave_addr, _ISAR(i2c));
	if(full)
		if( in_interrupt())
			mdelay(1);
		else
			msleep(1);
}


static void i2c_pxa_reset(struct pxa_i2c *i2c)
{
	u32 icr;

	pr_info("Resetting I2C Controller Unit.  ");

	i2c_pxa_unit_restart(i2c,1);

	/* set control register values */
	writel((I2C_ICR_INIT & ~ICR_SCLE), _ICR(i2c));

	/* Adjust the SCLK close to maxmium frequency */
//	if(cpu_is_pxa950) {
//		writel(0x082C36A3, _ILCR(i2c));
//		writel(0x00001434, _IWCR(i2c));
//	} else {
		writel(0x082C26A3, _ILCR(i2c));
//	}
#ifdef CONFIG_I2C_PXA_SLAVE
	dev_info(&i2c->adap.dev, "Enabling slave mode\n");
	writel(readl(_ICR(i2c)) | ICR_SADIE | ICR_ALDIE | ICR_SSDIE, _ICR(i2c));
#endif

	i2c_pxa_set_slave(i2c, 0);

	icr = readl(_ICR(i2c));
	if (i2c->flags & PXA_I2C_FAST_MODE) {
		icr |= (1 << 15);
		icr &= ~(1 << 16);
		pr_info("Specify FAST I2C speed.\n");
	} else //if ((i2c->flags & PXA_I2C_STANDARD_MODE) == 0)
		pr_info("Specify SLOW I2C speed.\n");

    i2c->icr_backup = icr | ICR_IUE | ICR_SCLE;

	/* enable unit */
	writel(icr | ICR_IUE | ICR_SCLE, _ICR(i2c));
	udelay(100);
}


#ifdef CONFIG_I2C_PXA_SLAVE
/*
 * PXA I2C Slave mode
 */

static void i2c_pxa_slave_txempty(struct pxa_i2c *i2c, u32 isr)
{
	if (isr & ISR_BED) {
		/* what should we do here? */
		i2c_pxa_scream_blue_murder(i2c, "Bus ERROR i2c_pxa_slave_txempty");
	} else {
		int ret = 0;

		if (i2c->slave != NULL)
			ret = i2c->slave->read(i2c->slave->data);

		writel(ret, _IDBR(i2c));
		writel(readl(_ICR(i2c)) | ICR_TB, _ICR(i2c));   /* allow next byte */
	}
}

static void i2c_pxa_slave_rxfull(struct pxa_i2c *i2c, u32 isr)
{
	unsigned int byte = readl(_IDBR(i2c));

	if (i2c->slave != NULL)
		i2c->slave->write(i2c->slave->data, byte);

	writel(readl(_ICR(i2c)) | ICR_TB, _ICR(i2c));
}

static void i2c_pxa_slave_start(struct pxa_i2c *i2c, u32 isr)
{
	int timeout;

	if (i2c_debug > 0)
		dev_dbg(&i2c->adap.dev, "SAD, mode is slave-%cx\n",
		       (isr & ISR_RWM) ? 'r' : 't');

	if (i2c->slave != NULL)
		i2c->slave->event(i2c->slave->data,
				 (isr & ISR_RWM) ? I2C_SLAVE_EVENT_START_READ : I2C_SLAVE_EVENT_START_WRITE);

	/*
	 * slave could interrupt in the middle of us generating a
	 * start condition... if this happens, we'd better back off
	 * and stop holding the poor thing up
	 */
	writel(readl(_ICR(i2c)) & ~(ICR_START|ICR_STOP), _ICR(i2c));
	writel(readl(_ICR(i2c)) | ICR_TB, _ICR(i2c));

	timeout = 0x10000;

	while (1) {
		if ((readl(_IBMR(i2c)) & 2) == 2)
			break;

		timeout--;

		if (timeout <= 0) {
			dev_err(&i2c->adap.dev, "timeout waiting for SCL high\n");
			break;
		}
	}

	writel(readl(_ICR(i2c)) & ~ICR_SCLE, _ICR(i2c));
}

static void i2c_pxa_slave_stop(struct pxa_i2c *i2c)
{
	if (i2c_debug > 2)
		dev_dbg(&i2c->adap.dev, "ISR: SSD (Slave Stop)\n");

	if (i2c->slave != NULL)
		i2c->slave->event(i2c->slave->data, I2C_SLAVE_EVENT_STOP);

	if (i2c_debug > 2)
		dev_dbg(&i2c->adap.dev, "ISR: SSD (Slave Stop) acked\n");

	/*
	 * If we have a master-mode message waiting,
	 * kick it off now that the slave has completed.
	 */
	if (i2c->msg)
		i2c_pxa_master_complete(i2c, I2C_RETRY);
}
#else //CONFIG_I2C_PXA_SLAVE
static void i2c_pxa_slave_txempty(struct pxa_i2c *i2c, u32 isr)
{
	if (isr & ISR_BED) {
		/* what should we do here? */
		i2c_pxa_scream_blue_murder(i2c, "Bus ERROR i2c_pxa_slave_txempty");
	} else {
		writel(0, _IDBR(i2c));
		writel(readl(_ICR(i2c)) | ICR_TB, _ICR(i2c));
	}
}

static void i2c_pxa_slave_rxfull(struct pxa_i2c *i2c, u32 isr)
{
	writel(readl(_ICR(i2c)) | ICR_TB | ICR_ACKNAK, _ICR(i2c));
}

static void i2c_pxa_slave_start(struct pxa_i2c *i2c, u32 isr)
{
	int timeout;

	/*
	 * slave could interrupt in the middle of us generating a
	 * start condition... if this happens, we'd better back off
	 * and stop holding the poor thing up
	 */
	writel(readl(_ICR(i2c)) & ~(ICR_START|ICR_STOP), _ICR(i2c));
	writel(readl(_ICR(i2c)) | ICR_TB | ICR_ACKNAK, _ICR(i2c));

	timeout = 0x10000;

	while (1) {
		if ((readl(_IBMR(i2c)) & 2) == 2)
			break;

		timeout--;

		if (timeout <= 0) {
			dev_err(&i2c->adap.dev, "timeout waiting for SCL high\n");
			break;
		}
	}

	writel(readl(_ICR(i2c)) & ~ICR_SCLE, _ICR(i2c));
}

static void i2c_pxa_slave_stop(struct pxa_i2c *i2c)
{
	if (i2c->msg)
		i2c_pxa_master_complete(i2c, I2C_RETRY);
}
#endif//CONFIG_I2C_PXA_SLAVE

/*
 * PXA I2C Master mode
 */

static inline unsigned int i2c_pxa_addr_byte(struct i2c_msg *msg)
{
	unsigned int addr = (msg->addr & 0x7f) << 1;

	if (msg->flags & I2C_M_RD)
		addr |= 1;

	return addr;
}

static inline void i2c_pxa_start_message(struct pxa_i2c *i2c)
{
	u32 icr;

	/*
	 * Step 1: target slave address into IDBR
	 */
	writel(i2c_pxa_addr_byte(i2c->msg), _IDBR(i2c));

	/*
	 * Step 2: initiate the write.
	 *   Do NOT enable ICR_ALDIE so HW will resolve Arbitration automatically
	 */
	icr = readl(_ICR(i2c)) & ~(ICR_STOP | ICR_ALDIE);
	ICR_LOG = icr;
	icr |= ICR_START | ICR_TB | ICR_MSDE;

	writel(icr, _ICR(i2c));
}

static inline void i2c_pxa_stop_message(struct pxa_i2c *i2c)
{
	u32 icr, isr , i=5;

	icr = readl(_ICR(i2c));
	/* Workaround: Wait for a while before clearing ICR_STOP */
	//if (cpu_is_pxa935()) {
		if (icr & ICR_MSDE) {
			do {
				isr = readl(_ISR(i2c));
			} while ( ~(isr & ISR_MSD) & i--);
			writel(isr ,_ISR(i2c));
		}
	/*
	 * Clear the STOP and ACK flags
	 */
	icr &= ~(ICR_STOP | ICR_ACKNAK);
	writel(icr, _ICR(i2c));
}

/*
 * We are protected by the adapter bus mutex.
 */
static int i2c_pxa_do_xfer(struct pxa_i2c *i2c, struct i2c_msg *msg, int num)
{
    unsigned long cpsr;
	int ret, retry;
	u32 icr, isr, ibmr;
    int  polling_delay;
	long timeout;

	i2c->msg     = msg;
	i2c->msg_num = num;
	i2c->msg_idx = 0;
	i2c->msg_ptr = 0;
	ICR_LOG = 0xCAFE;

    /* Wait for bus free: check twice for the Bus status bits and IBMR to be sure.
     * This still can't guaranty "no race condition with other master" but makes best efforts.
     * If decided the bus is IDLE
     *   do NOT enable ICR_ALDIE so possible arbitration will be fixed by HW
     **/
    writel(i2c_pxa_addr_byte(i2c->msg), _IDBR(i2c)); //target slave address into IDBR
    retry = 2;
	local_irq_save(cpsr);
    do
    {
        retry--;
        ibmr = readl(_IBMR(i2c));
        isr  = readl(_ISR(i2c));
        IBMR_LOG = ibmr;
        ISR_LOG  = isr;
        if( (ibmr != I2C_IBMR_IDLE) || (isr & (ISR_EBB | ISR_IBB | ISR_UB)) ) {
            local_irq_restore(cpsr); udelay(i2c->access2_delay);
            return(I2C_RETRY);
        }
        if(retry == 0)
            break;

        //"Delay" and Re-Check Bus IDLE
        icr = readl(_ICR(i2c));
        ICR_LOG = icr;
        isr |= I2C_ISR_CLEAR;
        writel(isr, _ISR(i2c));

    }while(retry>0);

    // Bus is FREE, go ahead and start transaction
	i2c->debug_track=1;
    icr |= ICR_START | ICR_TB | ICR_MSDE;
    writel(icr, _ICR(i2c));
    //local_irq_restore(cpsr)   - see below

    if (i2c->use_pio)
    {
        local_irq_restore(cpsr);
        polling_delay = 14 * i2c->access2_delay/2; //MIN: 3*8+2 cycles = 26 => use about half for polling
        timeout = (MAX_XFER_TIMEOUT_US/polling_delay)+1; //80mSec
        while(timeout-- && (i2c->msg_num > 0) )
        {
            i2c_pxa_handler(0, i2c);
            udelay(polling_delay);
        }
    }
    else
    {
        //local_irq_restore(cpsr) do NOT restore,
        // the wait_event_timeout() should go into SUSPEND before IRQ-end-event wakes ip this task
        // The wait_event_timeout() enables the irq itself
        //WORKAROUND:
        //  Sometimes wake_up() does not wake-up the task with wait_event_timeout()
        //  In that case the task runs again upon TimeOut but condition is already TRUE.
        //  Let's use wait_event_timeout() with short (7.8mS is minimum) and go to long TO only if condition still FALSE
        timeout = wait_event_timeout(i2c->wait, i2c->msg_num == 0, 1); //7.8 mSec
        if(i2c->msg_num != 0)
            wait_event_timeout(i2c->wait, i2c->msg_num == 0, 5*HZ); //5s is long enough for camera
    }

	i2c_pxa_stop_message(i2c);

	ret = i2c->msg_idx;  //return code in i2c->msg_idx

	if(i2c->msg_num != 0) {
		i2c_pxa_scream_blue_murder(i2c, "ERROR: XFER timeout - started but never finished");
	    ret = I2C_TIMEOUT_DONT_RETRY;
	}

    if (ret == I2C_RETRY)
        udelay(i2c->busbusy_delay);

	return ret;
}

/*
 * i2c_pxa_master_complete - complete the message and wake up.
 */
static void i2c_pxa_master_complete(struct pxa_i2c *i2c, int ret)
{
	i2c->msg_ptr = 0;
	i2c->msg = NULL;
	i2c->msg_idx ++;
	i2c->msg_num = 0;
	if (ret)
		i2c->msg_idx = ret;
	if (!i2c->use_pio)
		wake_up(&i2c->wait);
}


static void i2c_pxa_irq_txempty(struct pxa_i2c *i2c, u32 isr)
{
	u32 icr = readl(_ICR(i2c));
	ICR_LOG = icr;

    if ((isr & (ISR_BED | ISR_ALD)) == (ISR_BED | ISR_ALD)) {
        i2c_pxa_scream_blue_murder(i2c, "wrong ISR_BED & ISR_ALD");
        i2c_pxa_unit_restart(i2c, 1);
        return;
    }
    if (isr & ISR_ALD) {
      writel(ISR_ALD, _ISR(i2c));
      if ((i2c->msg_ptr != 0) || (i2c->msg_idx != 0))
      {
        DBG_GPIO_ERR_ON();
        i2c_pxa_scream_blue_murder(i2c, "INFO but not error\n"
		    "i2c:        Arbitration Loss Detected, go with RETRY");
        DBG_GPIO_ERR_OFF();
		i2c_pxa_master_complete(i2c, I2C_RETRY);
      }
	  else
	  {
		//Arbitration on ADDRESS stage should be resolved by HW (just clear ALD-bit)
		//But check the master is NOT stopped; if stopped go to retry
		if(isr & ISR_MSD) {
			i2c_pxa_scream_blue_murder(i2c, "ERROR: ADDR Arbitration Loss with MasterSTOP");
			i2c_pxa_master_complete(i2c, I2C_MASTER_ABORT);
		} else {
			i2c->debug_track=7;
		}
	  }
      return;
    }
    if (isr & ISR_BED) {
		/*
		 * I2C bus error - either the device NAK'd us, or
		 * something more serious happened.  If we were NAK'd
		 * on the initial ADDRESS phase, we can retry.
		 */
	    writel(ISR_BED, _ISR(i2c));
        if (i2c->msg_ptr == 0 && i2c->msg_idx == 0)
        {
            DBG_GPIO_ERR_ON();
			i2c_pxa_scream_blue_murder(i2c, "Bus ERROR or no answer from slave-device");
            DBG_GPIO_ERR_OFF();
        }
		else {
			if(i2c_pxa_custom_scream_silent(i2c)) {
		        i2c_pxa_master_complete(i2c, 0);
				return;
			}
			i2c_pxa_scream_blue_murder(i2c, "Bus ERROR on Data stage");
		}
        i2c_pxa_master_complete(i2c,I2C_BUS_ERROR);
        return;
	}

	icr &= ~(ICR_START|ICR_STOP|ICR_ACKNAK|ICR_TB);

 again:
	if (isr & ISR_RWM) {
		/*
		 * Read mode.  We have just sent the address byte, and
		 * now we must initiate the transfer.
		 */
		if (i2c->msg_ptr == i2c->msg->len - 1 &&
		    i2c->msg_idx == i2c->msg_num - 1)
			icr |= ICR_STOP | ICR_ACKNAK;

		icr |= ICR_ALDIE | ICR_TB;
        i2c->debug_track=2;

	} else if (i2c->msg_ptr < i2c->msg->len) {
		/*
		 * Write mode.  Write the next data byte.
		 */
        if(i2c->msg->buf == NULL) {
				i2c_pxa_scream_blue_murder(i2c, "NULL buffer for TX");
				i2c_pxa_unit_restart(i2c, 1);
				i2c_pxa_master_complete(i2c, I2C_NULL_BUFFER);
                return;
		}
		writel(i2c->msg->buf[i2c->msg_ptr++], _IDBR(i2c));

		icr |= ICR_ALDIE | ICR_TB;

		/*
		 * If this is the last byte of the last message, send
		 * a STOP.
		 */
		if (i2c->msg_ptr == i2c->msg->len &&
		    i2c->msg_idx == i2c->msg_num - 1)
			icr |= ICR_STOP;
        i2c->debug_track=3;
	} else if (i2c->msg_idx < i2c->msg_num - 1) {
		/*
		 * Next segment of the message.
		 */
		i2c->msg_ptr = 0;
		i2c->msg_idx ++;
		i2c->msg++;

		/*
		 * If we aren't doing a repeated start and address,
		 * go back and try to send the next byte.  Note that
		 * we do not support switching the R/W direction here.
		 */
		if (i2c->msg->flags & I2C_M_NOSTART)
			goto again;

		/*
		 * Write the next address.
		 */
		writel(i2c_pxa_addr_byte(i2c->msg), _IDBR(i2c));

		/*
		 * And trigger a repeated start, and send the byte.
		 */
		icr &= ~ICR_ALDIE;
		icr |= ICR_START | ICR_TB | ICR_MSDE;
		/* Master stop detected enabled */
		//if (cpu_is_pxa935())
			icr |= ICR_MSDE;
        i2c->debug_track=4;
	} else {
		if (i2c->msg->len == 0) {
			/*
			 * Device probes have a message length of zero
			 * and need the bus to be reset before it can
			 * be used again.
			 */
			i2c_pxa_unit_restart(i2c, 1);
            i2c->debug_track=51;
		}
        else
            i2c->debug_track=52;
		i2c_pxa_master_complete(i2c, 0);
	}
	writel(icr, _ICR(i2c));
	show_state(i2c);
}

static void i2c_pxa_irq_rxfull(struct pxa_i2c *i2c, u32 isr)
{
	u32 icr = readl(_ICR(i2c)) & ~(ICR_START|ICR_STOP|ICR_ACKNAK|ICR_TB);
	ICR_LOG = icr;

    i2c->debug_track=6;
	/*
	 * Read the byte.
	 */
    if(i2c->msg->buf == NULL) {
			i2c_pxa_scream_blue_murder(i2c, "NULL buffer for RX");
			i2c_pxa_unit_restart(i2c, 1);
			i2c_pxa_master_complete(i2c, I2C_NULL_BUFFER);
            return;
	}
	i2c->msg->buf[i2c->msg_ptr++] = readl(_IDBR(i2c));

	if (i2c->msg_ptr < i2c->msg->len) {
		/*
		 * If this is the last byte of the last
		 * message, send a STOP.
		 */
		if (i2c->msg_ptr == i2c->msg->len - 1)
			icr |= ICR_STOP | ICR_ACKNAK;

		icr |= ICR_ALDIE | ICR_TB;
        i2c->debug_track=61;
	} else {
        i2c->debug_track=62;
		i2c_pxa_master_complete(i2c, 0);
	}
	writel(icr, _ICR(i2c));
}

static irqreturn_t i2c_pxa_handler(int this_irq, void *dev_id)
{
	struct pxa_i2c *i2c = dev_id;
    u32 isr;

	CUSTOM_BYTE2BYTE_DELAY(i2c);

DBG_GPIO_IRQH_IN();
    isr = readl(_ISR(i2c));
	ISR_LOG = isr;

/****
	if (i2c_debug > 2 && 0) {
		dev_dbg(&i2c->adap.dev, "%s: ISR=%08x, ICR=%08x, IBMR=%02x\n",
			__func__, isr, readl(_ICR(i2c)), readl(_IBMR(i2c)));
		decode_ISR(isr);
	}
	show_state(i2c);
*****/

	if (isr & ISR_SAD) {
		writel(ISR_SAD, _ISR(i2c));
        i2c_pxa_scream_blue_murder(i2c, "spurious SLAVE irq (ISR_SAD=1)");
		i2c_pxa_slave_start(i2c, isr);
	}
	if (isr & ISR_SSD) {
		writel(ISR_SSD, _ISR(i2c));
        i2c_pxa_scream_blue_murder(i2c, "spurious SLAVE irq (ISR_SSD=1)");
		i2c_pxa_slave_stop(i2c);
	}

	if (i2c_pxa_is_slavemode(i2c)) {
		i2c_pxa_scream_blue_murder(i2c, "spurious SLAVE irq");
		if (isr & ISR_ITE) {
			writel(ISR_ITE, _ISR(i2c));
			i2c_pxa_slave_txempty(i2c, isr);
		}
		if (isr & ISR_IRF) {
			writel(ISR_IRF, _ISR(i2c));
			i2c_pxa_slave_rxfull(i2c, isr);
		}
	} else if (i2c->msg) {
		if (isr & (ISR_ITE | ISR_ALD | ISR_BED)) {
			writel(ISR_ITE, _ISR(i2c));
			i2c_pxa_irq_txempty(i2c, isr);
		}
		if (isr & ISR_IRF) {
			writel(ISR_IRF, _ISR(i2c));
			i2c_pxa_irq_rxfull(i2c, isr);
		}
	} else {
		i2c_pxa_scream_blue_murder(i2c, "spurious irq");
		i2c_pxa_unit_restart(i2c, 1);
	}

DBG_GPIO_IRQH_OUT();
	return IRQ_HANDLED;
}

static int i2c_pxa_xfer(struct i2c_adapter *adap, struct i2c_msg msgs[], int num)
{
	#define NUM_XFER_RETRIES  1
	struct pxa_i2c *i2c = adap->algo_data;
	int ret, i;
	int exh_retries = 0, xfer_retries = NUM_XFER_RETRIES;

	if (num <= 0)
		BUG();

#ifdef DEBUG_HW_STOP_ON_ERR
	if (stop_on_error)
		return num;
#endif

#ifdef CONFIG_I2C_DEBUG_DEVICE_BUSY
	adap->current_device = msgs->addr;
	memset(adap->last_task_comm, 0, sizeof(adap->last_task_comm));
	memcpy(adap->last_task_comm, current->comm, sizeof(current->comm));
	adap->last_pid = current->pid;
#endif
    if(in_interrupt()) {
		printk(KERN_ERR "i2c_pxa_xfer called in interrupt context. Does nothing but Exit with ok\n");
		return(num);
	}

#if defined(DEBUG_ERR_SIMULATE)
	i2c_pxa_buserr_simulate(msgs, num);
#endif
	i2c_pxa_msg_parse(i2c, msgs, num);

	set_dvfm_constraint();
	wake_lock(&pxa_i2c_suspend_lock);

	clk_enable(i2c->clk);   /* Possible but not must */

    ENTRY_CNTR++;
    DBG_GPIO_CFG();
    DBG_GPIO_DRV_IN();

    i2c_pxa_unit_restart(i2c, 0); //Reset since we could be after D2

do_xfer_retry:	
	i2c->debug_track=0;
	// Unit on, Clock Master Mode, Fast/Slow mode...
	//writel(i2c->icr_backup, _ICR(i2c));
	// i2c_pxa_custom_clock_adj(i2c) may modify the clock
	writel(i2c_pxa_custom_clock_adj(i2c), _ICR(i2c));
	udelay(I2C_ACCESS2_DELAY_US/2);  //WORKAROUND: wait a bit the clock is configured and stable

	while(1)
	{
		for (i = adap->retries; i >= 0; i--) {  // NUM_RETRIES_ON_BUS_BUSY
			ret = i2c_pxa_do_xfer(i2c, msgs, num);
			if (ret == I2C_RETRY)
				continue;	//udelay() done in the i2c_pxa_do_xfer()
			goto out;
		}
		if(exh_retries++<NUM_RETRIES_EXHAUSTED) {
			msleep(1); /* about 7.8 mS but non-blocking */
		} else{
			printk(KERN_ERR "i2c: ERROR: Bus Busy exhausted retries on Entry=%d  (ISR_%04x IBMR_%02x)\n",
			                            ENTRY_CNTR, readl(_ISR(i2c)), readl(_IBMR(i2c)) );
			ret = -EREMOTEIO;
			break;
		}
	}
out:
    if(ret<0)
    {
		if (xfer_retries == NUM_XFER_RETRIES)
			i2c_pxa_msg_print(i2c, msgs, num); /* print once only */

        i2c_pxa_unit_restart(i2c, 1);
		if (xfer_retries > 0)
		{
			xfer_retries--;
			goto do_xfer_retry;
		} else {
			ret = -EREMOTEIO;
			printk(KERN_ERR "\n");
		}
#ifdef DEBUG_HW_STOP_ON_ERR
		if (stop_on_error_req) {
			printk(KERN_ERR "i2c: DROP ALL i2c requests since now (reset CP also)!!!\n=======================\n");
			DBG_GPIO_DRV_OUT();
			stop_on_error = 1;
			/* Set CP into RESET */
			__raw_writel(0, (void *)&(__REG(0x40f5001c)));
		}
#endif
    }
    else
       i2c_pxa_unit_restart(i2c, 0);

	i2c_pxa_msg_save(i2c, msgs, num);

	clk_disable(i2c->clk);   /* Possible but not must */
	wake_unlock(&pxa_i2c_suspend_lock);
	unset_dvfm_constraint();
	DBG_GPIO_DRV_OUT();
#ifdef CONFIG_I2C_DEBUG_DEVICE_BUSY
	adap->last_device = msgs->addr;
#endif
	return ret;
}

static u32 i2c_pxa_functionality(struct i2c_adapter *adap)
{
	return I2C_FUNC_I2C | I2C_FUNC_SMBUS_EMUL;
}

static const struct i2c_algorithm i2c_pxa_algorithm = {
	.master_xfer	= i2c_pxa_xfer,
	.functionality	= i2c_pxa_functionality,
};

static const struct i2c_algorithm i2c_pxa_pio_algorithm = {
	.master_xfer	= i2c_pxa_xfer,
	.functionality	= i2c_pxa_functionality,
};

extern int pxa_gpio_mode(int gpio_mode);
static void i2c_pxa_enable(struct platform_device *dev)
{
	if (cpu_is_pxa27x()) {
		switch (dev->id) {
		case 0:
			pxa_gpio_mode(GPIO117_I2CSCL_MD);
			pxa_gpio_mode(GPIO118_I2CSDA_MD);
			break;
		case 1:
			local_irq_disable();
			PCFR |= PCFR_PI2CEN;
			local_irq_enable();
			break;
		}
	}
}

static void i2c_pxa_disable(struct platform_device *dev)
{
	if (cpu_is_pxa27x() && dev->id == 1) {
		local_irq_disable();
		PCFR &= ~PCFR_PI2CEN;
		local_irq_enable();
	}
}

#define res_len(r)		((r)->end - (r)->start + 1)
static int i2c_pxa_probe(struct platform_device *dev)
{
	struct pxa_i2c *i2c;
	struct resource *res;
	struct i2c_pxa_platform_data *plat = dev->dev.platform_data;
	int ret;
	int irq;

	res = platform_get_resource(dev, IORESOURCE_MEM, 0);
	irq = platform_get_irq(dev, 0);
	if (res == NULL || irq < 0)
		return -ENODEV;

	if (!request_mem_region(res->start, res_len(res), res->name))
		return -ENOMEM;

	i2c = kzalloc(sizeof(struct pxa_i2c), GFP_KERNEL);
	if (!i2c) {
		ret = -ENOMEM;
		goto emalloc;
	}

	i2c->adap.owner   = THIS_MODULE;

	init_waitqueue_head(&i2c->wait);

	/*
	 * If "dev->id" is negative we consider it as zero.
	 * The reason to do so is to avoid sysfs names that only make
	 * sense when there are multiple adapters.
	 */
	i2c->adap.nr = dev->id != -1 ? dev->id : 0;
	snprintf(i2c->adap.name, sizeof(i2c->adap.name), "pxa_i2c-i2c.%u",
		 i2c->adap.nr);

	i2c->clk = clk_get(&dev->dev, "I2CCLK");
	if (IS_ERR(i2c->clk)) {
		ret = PTR_ERR(i2c->clk);
		goto eclk;
	}

	i2c->reg_base = ioremap(res->start, res_len(res));
	if (!i2c->reg_base) {
		ret = -EIO;
		goto eremap;
	}

	i2c->iobase = res->start;
	i2c->iosize = res_len(res);

	i2c->irq = irq;

	i2c->slave_addr = I2C_PXA_SLAVE_ADDR;

#ifdef CONFIG_I2C_DEBUG_DEVICE_BUSY
	i2c->adap.last_device = 0;
	i2c->adap.current_device = 0;
#endif
#ifdef CONFIG_I2C_PXA_SLAVE
	if (plat) {
		i2c->slave_addr = plat->slave_addr;
		i2c->slave = plat->slave;
	}
#endif

	clk_enable(i2c->clk);
	i2c_pxa_enable(dev);

	if (plat) {
		i2c->adap.class = plat->class;
		i2c->use_pio = plat->use_pio;
		i2c->flags = plat->flags;
	}

/*   HANDLER_MODE__DFLT_SLOW_FAST == 0 - use default from upper configurator */
#if (HANDLER_MODE__DFLT_SLOW_FAST == 1)
	i2c->flags &= ~PXA_I2C_FAST_MODE;
#endif
#if (HANDLER_MODE__DFLT_SLOW_FAST == 2)
	i2c->flags |= PXA_I2C_FAST_MODE;
#endif

#if (HANDLER_MODE__DFLT_PIO_IRQ == 1)
	i2c->use_pio = 1;
#elif (HANDLER_MODE__DFLT_PIO_IRQ == 2)
	i2c->use_pio = 0;
#endif

	if (i2c->use_pio) {
		i2c->adap.algo = &i2c_pxa_pio_algorithm;
		pr_info("%s: PIO polling algo-mode configured\n", i2c->adap.name);
	} else {
		i2c->adap.algo = &i2c_pxa_algorithm;
		ret = request_irq(irq, i2c_pxa_handler, IRQF_DISABLED,
				  i2c->adap.name, i2c);
		if (ret) {
			pr_info("%s: IRQ algo-mode configuration failed\n", i2c->adap.name);
			goto ereqirq;
		}
		pr_info("%s: IRQ mode configured\n", i2c->adap.name);
	}

	i2c->adap.retries = NUM_RETRIES_ON_BUS_BUSY;
	i2c->busbusy_delay = I2C_BUSBUSY_DELAY_US;
	i2c->access2_delay = I2C_ACCESS2_DELAY_US;

	if(!(i2c->flags & PXA_I2C_FAST_MODE))  {
		i2c->access2_delay *= FAST_SLOW_RATE;
	}

#if defined (CUSTOMIZATION_PER_SLAVE_DEVICE)
	if((i2c->adap.nr == 0) || (i2c->adap.nr == 2))
		i2c->p_custom = &i2c_pxa_custom[0];
	else
		i2c->p_custom = NULL;
#endif

	i2c_pxa_reset(i2c);

	i2c->adap.algo_data = i2c;
	i2c->adap.dev.parent = &dev->dev;

	ret = i2c_add_numbered_adapter(&i2c->adap);
	if (ret < 0) {
		printk(KERN_INFO "I2C: Failed to add bus\n");
		goto eadapt;
	}

	platform_set_drvdata(dev, i2c);

#ifdef CONFIG_I2C_PXA_SLAVE
	printk(KERN_INFO "I2C: %s: PXA I2C adapter, slave address %d\n",
	       i2c->adap.dev.bus_id, i2c->slave_addr);
#else
	//printk(KERN_INFO "I2C: %s: PXA I2C adapter\n",
	//       i2c->adap.dev.bus_id);
#endif

#ifdef	CONFIG_PROC_FS
	create_i2c_pxa_proc_file(i2c, i2c->adap.nr);
#endif

	//DBG_GPIO_CFG();
	clk_disable(i2c->clk);
	return 0;

eadapt:
	if (!i2c->use_pio)
		free_irq(irq, i2c);
ereqirq:
	clk_disable(i2c->clk);
	i2c_pxa_disable(dev);
	iounmap(i2c->reg_base);
eremap:
	clk_put(i2c->clk);
eclk:
	kfree(i2c);
emalloc:
	release_mem_region(res->start, res_len(res));
	return ret;
}

static int __exit i2c_pxa_remove(struct platform_device *dev)
{
	struct pxa_i2c *i2c = platform_get_drvdata(dev);

	clk_enable(i2c->clk);

	platform_set_drvdata(dev, NULL);

#ifdef	CONFIG_PROC_FS
	remove_i2c_pxa_proc_file(i2c->adap.nr);
#endif

	i2c_del_adapter(&i2c->adap);
	if (!i2c->use_pio)
		free_irq(i2c->irq, i2c);

	clk_disable(i2c->clk);
	clk_put(i2c->clk);
	i2c_pxa_disable(dev);

	iounmap(i2c->reg_base);
	release_mem_region(i2c->iobase, i2c->iosize);
	kfree(i2c);

	return 0;
}

static struct platform_driver i2c_pxa_driver = {
	.probe		= i2c_pxa_probe,
	.remove		= __exit_p(i2c_pxa_remove),
	.driver		= {
		.name	= "pxa2xx-i2c",
		.owner	= THIS_MODULE,
	},
};

static int __init i2c_adap_pxa_init(void)
{
#ifdef CONFIG_PXA3xx_DVFM
	dvfm_register("I2C", &dvfm_dev_idx);
#endif
	wake_lock_init(&pxa_i2c_suspend_lock, WAKE_LOCK_SUSPEND, "pxa_i2c");
	return platform_driver_register(&i2c_pxa_driver);
}

static void __exit i2c_adap_pxa_exit(void)
{
	platform_driver_unregister(&i2c_pxa_driver);
#ifdef CONFIG_PXA3xx_DVFM
	dvfm_unregister("I2C", &dvfm_dev_idx);
#endif
	wake_lock_destroy(&pxa_i2c_suspend_lock);
}

MODULE_LICENSE("GPL");

subsys_initcall(i2c_adap_pxa_init);
module_exit(i2c_adap_pxa_exit);
