/* drivers/i2c/chips/akm8973.c - akm8973 compass driver
 *
 * Copyright (C) 2007-2008 HTC Corporation.
 * Author: Hou-Kun Chen <houkun.chen@gmail.com>
 *
 * This software is licensed under the terms of the GNU General Public
 * License version 2, as published by the Free Software Foundation, and
 * may be copied, distributed, and modified under those terms.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 */

/*
 * Revised by AKM 2009/04/02
 * 
 */
/*
 *Design notes
 1. read report magenetic x/y/z values
 2. write will receive from library the yaw/pitch/roll data
 3. report these datas to app by input/sensor-core
 4. add test mode
 *
 */

#include <linux/interrupt.h>
#include <linux/i2c.h>
#include <linux/slab.h>
#include <linux/irq.h>
#include <linux/miscdevice.h>
#include <asm/gpio.h>
#include <asm/uaccess.h>
#include <linux/delay.h>
#include <linux/input.h>
#include <linux/poll.h>
#include <linux/workqueue.h>
#include <linux/freezer.h>
#include <mach/akm8973.h>
#include <linux/sensors.h>
#include <linux/earlysuspend.h>

#define DEBUG 0 
#define FAC_TEST_RD 0
#define FAC_TEST_WR 1
#define FAC_TEST_DISABLE_INT 2
#define FAC_TEST_ENABLE_INT 3
static  u8 fac_test_addr=0xE0;
static  u8 fac_test_len=7;
static int compass_blkread(u8 reg, u8 *pval, u8 len);
static int compass_active(void);
static int compass_active_e2p(void);
static unsigned short normal_i2c[] = { I2C_CLIENT_END };
I2C_CLIENT_INSMOD;
static struct i2c_client *this_client;
static struct akm8973_platform_data *pdata;
static int reserve_akm_switch_on;
static int reserver_dummy_switch_on;

static compass_device_t compass_device = {
	.name 			= "compass_device",
	.rf_count		= ATOMIC_INIT(0),
};

static int AKI2C_RxData(char *rxData, int length)
{
	struct i2c_msg msgs[] = {
		{
		 .addr = this_client->addr,
		 .flags = 0,
		 .len = 1,
		 .buf = rxData,
		 },
		{
		 .addr = this_client->addr,
		 .flags = I2C_M_RD,
		 .len = length,
		 .buf = rxData,
		 },
	};

	if (i2c_transfer(this_client->adapter, msgs, 2) < 0) {
		printk(KERN_ERR "AKI2C_RxData: transfer error\n");
		return -EIO;
	} else
		return 0;
}

static int AKI2C_TxData(char *txData, int length)
{
	struct i2c_msg msg[] = {
		{
		 .addr = this_client->addr,
		 .flags = 0,
		 .len = length,
		 .buf = txData,
		 },
	};

#if DEBUG
	printk(KERN_INFO "%s\n", __FUNCTION__);
#endif
	if (i2c_transfer(this_client->adapter, msg, 1) < 0) {
		printk(KERN_ERR "AKI2C_TxData: transfer error\n");
		return -EIO;
	} else
		return 0;
}

static int AKECS_StartMeasure(void)
{
	char buffer[2];

	/* Set measure mode */
	buffer[0] = AKECS_REG_MS1;
	buffer[1] = AKECS_MODE_MEASURE;

	/* Set data */
	return AKI2C_TxData(buffer, 2);
}

static int AKECS_PowerDown(void)
{
	char buffer[2];
	int ret;

	/* Set powerdown mode */
	buffer[0] = AKECS_REG_MS1;
	buffer[1] = AKECS_MODE_POWERDOWN;
	/* Set data */
	ret = AKI2C_TxData(buffer, 2);
	if (ret < 0)
		return ret;

	/* Dummy read for clearing INT pin */
	buffer[0] = AKECS_REG_TMPS;
	/* Read data */
	ret = AKI2C_RxData(buffer, 1);
	if (ret < 0)
		return ret;

	return ret;
}

static int AKECS_StartE2PRead(void)
{
	char buffer[2];

	/* Set E2P mode */
	buffer[0] = AKECS_REG_MS1;
	buffer[1] = AKECS_MODE_E2P_READ;
	/* Set data */
	return AKI2C_TxData(buffer, 2);
}

static int AKECS_SetMode(char mode)
{
	int ret;

	switch (mode) {
	case AKECS_MODE_MEASURE:
		ret = AKECS_StartMeasure();
		break;
	case AKECS_MODE_E2P_READ:
		ret = AKECS_StartE2PRead();
		break;
	case AKECS_MODE_POWERDOWN:
		ret = AKECS_PowerDown();
		break;
	default:
		return -EINVAL;
	}

	/* wait at least 300us after changing mode */
	msleep(1);
	return ret;
}

/*
 * Borqs 
 * 
 */

//wake up poll fd to read data
static void compass_wake_up_int(unsigned long data)
{
	compass_status_t *status_t;

	spin_lock(&compass_device.list_lock);
	list_for_each_entry(status_t, &compass_device.open_list, list) {
		atomic_set(&status_t->g_rdy, 1);
	}
	spin_unlock(&compass_device.list_lock);

	//wake_up_interruptible(&compass_device.compass_wait);
}

//the scheduled work,read i2c data
static char akm_buf[3] = {0,0,0};
static void compass_int_routine(struct work_struct *work)
{
#if DEBUG
	printk(KERN_DEBUG "compass int trigger.\n");	
#endif
	char x,y,z;

	compass_blkread(COMPASS_X, akm_buf, 3);
	x = akm_buf[0];
	y = akm_buf[1];
	z = akm_buf[2];
#if DEBUG
	printk(KERN_DEBUG "compass read: x=%d,y=%d,z=%d\n",
				(signed char)x, (signed char)y, (signed char)z);
#endif
	compass_wake_up_int(0);
	wake_up_interruptible(&compass_device.compass_wait);
}

//interrupt handler,start the work and set next timer(2s)
static irqreturn_t compass_irq(int irq, void *dev_id)
{
	schedule_delayed_work(&compass_device.compass_work, 0);
	//mod_timer(&compass_device.timer, jiffies + msecs_to_jiffies(2000));
	return IRQ_HANDLED;
}

static int compass_dev_open(struct inode *inode, struct file *file)
{
	int ret;
	compass_status_t *status_p;

	printk(KERN_DEBUG "E-compass open\n");
	
	if (down_interruptible(&compass_device.mutex))
		return -ERESTARTSYS;

	if ((file->f_flags & O_ACCMODE) == O_WRONLY) {
		printk(KERN_INFO "COMPASS's device node is writeonly\n");
		ret = -EPERM;
	} else {
		printk(KERN_DEBUG "COMPASS's device :You are opening !\n");
		ret = 0;
	}
	
	status_p = kzalloc(sizeof(compass_status_t), GFP_KERNEL);
	if (status_p == NULL) { 
		printk(KERN_INFO "No mem!\n");
		up(&compass_device.mutex);
		return -ENOMEM;
	}
	file->private_data = status_p;
	atomic_set(&status_p->g_rdy, 1);
	spin_lock(&compass_device.list_lock);
	list_add(&status_p->list, &compass_device.open_list);
	spin_unlock(&compass_device.list_lock);

	atomic_inc(&compass_device.rf_count);
	printk(KERN_INFO "rf_count = %d\n", atomic_read(&compass_device.rf_count));
	up(&compass_device.mutex);
	return ret;
}

/* Without block, return x/y/z values from pool */
static ssize_t compass_dev_read(struct file *file, char __user * buffer,
				  size_t size, loff_t * f_pos)
{
	compass_status_t *status_p;
	struct input_event event[3];
	char x, y, z;
	int ret;

	x = akm_buf[0];
	y = akm_buf[1];
	z = akm_buf[2];
#if DEBUG
	printk(KERN_DEBUG "compass read: x=%d,y=%d,z=%d\n",
				(signed char)x, (signed char)y, (signed char)z);
#endif
	get_monotonic_time(&event[0].time);
	event[1].time = event[0].time;
	event[2].time = event[0].time;

	//magnetic datas
	event[0].type = EV_ABS;
	event[1].type = EV_ABS;
	event[2].type = EV_ABS;

	event[0].value = x;
	event[0].code = ABS_HAT0X;

	event[1].value = y;
	event[1].code = ABS_HAT0Y;

	event[2].value = z;
	event[2].code = ABS_BRAKE;

	//add values range detection
	if (copy_to_user(buffer, event, sizeof(event))) {
		printk(KERN_INFO "Copy Akm_sensor data Fail!\n");
		ret = -EIO;	
	} else
	ret = sizeof(event);
	status_p = file->private_data;
	atomic_set(&status_p->g_rdy, 0);
	return ret;
}

static ssize_t compass_dev_write(struct file *file, const char __user * buffer,
				  size_t size, loff_t * f_pos)
{
	/*Follow library write style*/
	/*
	 * buf[0]: bytes length to write
	 * buf[1]: register values
	 */
	void __user *argp = (void __user *)buffer;
	int ret = 0;

	char rwbuf[16] = {0};
	if (copy_from_user(&rwbuf, argp, sizeof(rwbuf)))
			return -EFAULT;

	if (rwbuf[0] < 2)
			return -EINVAL;
	ret = AKI2C_TxData(&rwbuf[1], rwbuf[0]);
	if (ret < 0) {
		printk(KERN_INFO "AKi2c write faild \n");
		return ret;
	}
	return 0;
}

static unsigned int compass_dev_poll(struct file *filep, poll_table * pt)
{
	unsigned int mask = 0;
	compass_status_t *status_p;
	
	if (down_interruptible(&compass_device.mutex))
		return -ERESTARTSYS;

	status_p = filep->private_data;
	if (status_p == NULL) {
		up(&compass_device.mutex);
		return -EFAULT;
	}

	poll_wait(filep, &compass_device.compass_wait, pt);

	if (atomic_read(&status_p->g_rdy)) 
		mask |= POLLIN | POLLRDNORM;

	up(&compass_device.mutex);
	return mask;
}

static int compass_dev_release(struct inode *inode, struct file *filp)
{
	compass_status_t *status_p;

	if (down_interruptible(&compass_device.mutex))
		return -ERESTARTSYS;

	status_p = filp->private_data;
	if (status_p == NULL) {
		up(&compass_device.mutex);
		return -EFAULT;
	}
	filp->private_data = NULL;
	
	spin_lock(&compass_device.list_lock);
	list_del(&status_p->list);
	spin_unlock(&compass_device.list_lock);

	kfree(status_p);

	atomic_dec(&compass_device.rf_count);

	up(&compass_device.mutex);
	
	return 0;
}


struct input_event orientation[3];
int dr = 0;
static int compass_dummy_open(struct inode *inode, struct file *file)
{
	/*For dummy node, always be okay*/
	return 0;
}

static ssize_t compass_dummy_read(struct file *file, char __user * buffer,
				  size_t size, loff_t * f_pos)
{
	int ret;

	/* If data not ready, we ignore this operation*/
	if(!dr)
		return 0;
	dr = 0;
#if DEBUG
	printk(KERN_INFO "compassE read(xyz)\n");
#endif
	//check values range here
	if (copy_to_user(buffer, orientation, sizeof(orientation))) {
		printk(KERN_INFO "Copy orientation data Fail!\n");
		ret = -EIO;	
	} else
		ret = sizeof(orientation);
	return ret;
}

static ssize_t compass_dummy_write(struct file *file, const char __user * buffer,
				  size_t size, loff_t * f_pos)
{
	/* Make sure data types not change during translation*/
	int x, y, z;
	struct input_event event[3];
	s16 buf[3];

#if DEBUG
	printk(KERN_DEBUG "Dummy write orientation (xyz)\n");
#endif
	if (copy_from_user(&buf, buffer, sizeof(buf)))
			return -EFAULT;
	/* Get data from buffer,update by interrupt work */
	x = buf[0];
	y = buf[1];
	z = buf[2];
#if DEBUG
	printk(KERN_DEBUG "compass read: x=%d,y=%d,z=%d\n",
				(signed char)x, (signed char)y, (signed char)z);
#endif

	get_monotonic_time(&event[0].time);
	event[1].time = event[0].time;
	event[2].time = event[0].time;

	event[0].type = EV_ABS;
	event[1].type = EV_ABS;
	event[2].type = EV_ABS;

	//orientation type datas
	event[0].value = x;
	event[0].code = ABS_RX;

	event[1].value = y;
	event[1].code = ABS_RY;

	event[2].value = z;
	event[2].code = ABS_RZ;

	//copy event values to orientation array
	memcpy(orientation, event, sizeof(event));
	dr = 1; /*data ready*/

	return 0;
}

static unsigned int compass_dummy_poll(struct file *filep, poll_table * pt)
{
	unsigned int mask = 0;

	if(dr)
		mask |= POLLIN | POLLRDNORM;
	return mask;
}

static int compass_dummy_release(struct inode *inode, struct file *filp)
{
	/* For dummy node, always close okay */
	return 0;
}

static void akm8973_early_suspend(struct early_suspend *handler)
{
	printk(KERN_DEBUG "AKM8973 suspend!!!!!!!!!!\n");
	reserve_akm_switch_on = compass_device.akm_switch_on;
	reserver_dummy_switch_on = compass_device.dummy_switch_on;
	AKECS_PowerDown();
	compass_device.akm_switch_on = SWITCH_OFF;
	compass_device.dummy_switch_on = SWITCH_OFF;
}

static void akm8973_early_resume(struct early_suspend *handler)
{
	printk(KERN_DEBUG "AKM8973 resume!!!!!!!!!!\n");
	compass_device.akm_switch_on = reserve_akm_switch_on;
	compass_device.dummy_switch_on = reserver_dummy_switch_on;
}

static struct file_operations akm8973_dev_fops = {
	.owner = THIS_MODULE,
	.open = compass_dev_open,
	.read = compass_dev_read,
	.write = compass_dev_write,
	.poll = compass_dev_poll,
	.release = compass_dev_release,
};

static struct file_operations akm_dummy_fops = {
	.owner = THIS_MODULE,
	.open = compass_dummy_open,
	.read = compass_dummy_read,
	.write = compass_dummy_write,
	.poll = compass_dummy_poll,
	.release = compass_dummy_release,
};

// STEP1: initialize of register and check initial state
int step_result = 0;
static void check_initial(void)
{
	char ehga[3];
	char buf[5];
	char msg[RBUFF_SIZE + 1] = {0};
	int j;
	//poweron and reset ---> operate gpio pins
	if (pdata != NULL) {
		pdata->fac_reset();
	}

#if 1
	//set eeprom access mode
	AKECS_StartE2PRead();
	udelay(300);

	//read magnetic sensor gain x/y/z
	compass_blkread(AKECS_REG_EHXGA, ehga, 3);
	AKECS_PowerDown();

	//write ehga values to h*ga registers
	buf[0] = 4; //length
	buf[1] = AKECS_REG_HXGA; //address
	buf[2] = ehga[0];//data buffer 
	buf[3] = ehga[1]; 
	buf[4] = ehga[2]; 
	AKI2C_TxData(&buf[1], buf[0]);
#endif

	//check status and temperature
	AKECS_StartMeasure();
	mdelay(13);
	//AKECS_TransRBuff(msg, RBUFF_SIZE+1);
	compass_blkread(AKECS_REG_ST, msg, 5);
	for (j = 0; j < 5; j++) {
		printk(" %02x", msg[j]);
	}
	step_result = 1;
}

//STEP2: check akm function
static void check_function(void)
{
	char buffer[6] = {0};
	//set measure mode
	AKECS_StartMeasure();
	mdelay(13);

	//read ST,judge int pin 
	compass_blkread(AKECS_REG_ST, buffer, 1);
	//expired after interrupt
	if ((buffer[0] & 0x1) != 1) {
		printk("No interrupt!!!\n");
	} else {
		printk("Interrupt Pin High!! Value:1 Interrupt!!!\n");
	}
	//read TEMP data
	compass_blkread(AKECS_REG_TMPS, buffer, 1);
	if (!buffer[0]) {
		printk("Temperature error!!!\n");
		step_result = -1;
		return;
	} else {
		printk("Temperature value: %02xH\n", buffer[0]);
	}
	step_result = 1;
}

static enum sensors_property akm_properties[] = {
	/* int */
	SENSORS_PROP_INTERVAL,
	SENSORS_PROP_THRESHOLD,
	SENSORS_PROP_MAXRANGE,		/* read only */
	SENSORS_PROP_RESOLUTION,	/* read only */
	SENSORS_PROP_VERSION,		/* read only */
	SENSORS_PROP_CURRENT,		/* read only */
	/* char */
	SENSORS_PROP_SWITCH,
	SENSORS_PROP_VENDOR,		/* read only */

	/*add for akm8973 ioctl operations*/
	SENSORS_PROP_MODE,
	SENSORS_PROP_TEST_STEP,
	SENSORS_PROP_REGISTER,		/* string values */
	SENSORS_PROP_FACTEST,
};

static int akm_put_property(struct sensors_dev *sdev,
				 enum sensors_property property,
				 union sensors_propval *val)
{
	int mode, ret;

	switch (property) {
	/*mode can be measure/e2p/powerdown*/
	case SENSORS_PROP_MODE:
		mode = val->intval;
#if DEBUG
		printk("sensors property: %d, mode=%x\n", property, mode);
#endif
		ret = AKECS_SetMode((char)mode);

		if (ret < 0)
			return ret;
		break;
	case SENSORS_PROP_SWITCH:
        compass_device.akm_switch_on = val->intval;
		break;
	case SENSORS_PROP_INTERVAL:
		if (down_interruptible(&compass_device.mutex))
			return -ERESTARTSYS;	
		if (val->intval > COMPASS_MAX_INTERVAL) {
			compass_device.akm_interval = COMPASS_MAX_INTERVAL;
		}else if (val->intval < COMPASS_MIN_INTERVAL ) {
			compass_device.akm_interval = COMPASS_MIN_INTERVAL;
		}else{
			compass_device.akm_interval = val->intval;
		}
		up(&compass_device.mutex);
		break;

	case SENSORS_PROP_TEST_STEP:
		if(val->intval == 1)
			check_initial();
		if(val->intval == 2)
			check_function();
		break;
	case SENSORS_PROP_FACTEST:
		if (val->strval[0] == FAC_TEST_RD){
			fac_test_addr = val->strval[2];
			if (val->strval[1] * 4 < PRO_STR_SIZE) //return 4 charactors for every reg 
				fac_test_len =  val->strval[1];
			else{
				printk(KERN_ERR "The length [%d] of Fac test reading is too long!\n",val->strval[1]);
				printk(KERN_ERR "The max length of Fac test reading is %d\n",PRO_STR_SIZE/4);
				return -EINVAL;
			} 
			return 3;
		}else if(val->strval[0] == FAC_TEST_WR) {
			ret = AKI2C_TxData(&val->strval[2], val->strval[1]+1);
			if (ret < 0) {
				printk(KERN_ERR "FAC test AKi2c write faild \n");
				return ret;
			}
			return  val->strval[1]+3;
		}else if (val->strval[0] == FAC_TEST_DISABLE_INT){
			disable_irq(this_client->irq);	
			return 1;
		}else if(val->strval[0] == FAC_TEST_ENABLE_INT){
			enable_irq(this_client->irq);	
			return 1;
		}else{
			printk(KERN_ERR "FAC test flag error!!\n");
		}
		break;
	default:
		return -EINVAL;
	}

	return 0; 
}


static int akm_get_property(struct sensors_dev *sdev,
				 enum sensors_property property,
				 union sensors_propval *val)
{
	//int interval = 0;
	char ehga[3];
	char tmp[PRO_STR_SIZE];
	//add some support future 
	int i;	
	switch (property) {
	case SENSORS_PROP_THRESHOLD:
		val->intval = 0;
		break;

	case SENSORS_PROP_INTERVAL:
		val->intval = compass_device.akm_interval;
		break;

	case SENSORS_PROP_MAXRANGE:
		val->intval = 128;	
		break;

	case SENSORS_PROP_RESOLUTION:
		val->intval = 1;
		break;

	case SENSORS_PROP_VERSION:
		val->intval = 1;	
		break;

	case SENSORS_PROP_CURRENT:
		val->intval = 30;	
		break;

	case SENSORS_PROP_MODE:
		break;
	case SENSORS_PROP_TEST_STEP:
		val->intval = step_result;	
		break;
	case SENSORS_PROP_VENDOR:
		sprintf(val->strval, "AKM");
		break;
	case SENSORS_PROP_SWITCH:
		val->intval = compass_device.akm_switch_on;	
		break;
	case SENSORS_PROP_REGISTER:		/* read only */
		printk(KERN_DEBUG "AKM get Register Values\n");
		compass_active_e2p();
		compass_blkread(AKECS_REG_EHXGA, ehga, 3);
		return	sprintf(val->strval, "%d %d %d", ehga[0], ehga[1], ehga[2]); //need future parse function
	case SENSORS_PROP_FACTEST:
		tmp[0] = fac_test_addr;
		AKI2C_RxData(tmp,fac_test_len);
		for (i=0; i<fac_test_len;i++){
			sprintf(&val->strval[i*4] , "%-3d ",tmp[i]);
		}
		val->strval[i*4] = 0;
		return i*4;
	default:
		return -EINVAL;
	}
	return 0;
}

static int akm_put_property_dummy(struct sensors_dev *sdev,
				 enum sensors_property property,
				 union sensors_propval *val)
{
	int mode, ret;

	switch (property) {
	case SENSORS_PROP_THRESHOLD:
		break;
	case SENSORS_PROP_INTERVAL:
		break;
	case SENSORS_PROP_MODE:
		mode = val->intval;
		ret = AKECS_SetMode((char)mode);
		if (ret < 0)
			return ret;
		break;
	case SENSORS_PROP_SWITCH:
		compass_device.dummy_switch_on = val->intval;
		break;
	case SENSORS_PROP_TEST_STEP:
		if(val->intval == 1)
			check_initial();
		if(val->intval == 2)
			check_function();
		break;
	default:
		return -EINVAL;
	}

	return 0; 
}


static int akm_get_property_dummy(struct sensors_dev *sdev,
				 enum sensors_property property,
				 union sensors_propval *val)
{
	int interval = 0;
	char ehga[3];
	//add some support future 
		
	switch (property) {
	case SENSORS_PROP_THRESHOLD:
		val->intval = 0;
		break;

	case SENSORS_PROP_INTERVAL:
		val->intval = interval;
		break;

	case SENSORS_PROP_MAXRANGE:
		val->intval = 128;	
		break;

	case SENSORS_PROP_RESOLUTION:
		val->intval = 1;
		break;

	case SENSORS_PROP_VERSION:
		val->intval = 1;	
		break;

	case SENSORS_PROP_CURRENT:
		val->intval = 30;	
		break;

	case SENSORS_PROP_MODE:
		break;
	case SENSORS_PROP_TEST_STEP:
		val->intval = step_result;	
		break;
	case SENSORS_PROP_VENDOR:
		sprintf(val->strval, "AKM");
		break;
	case SENSORS_PROP_SWITCH:
		val->intval = compass_device.dummy_switch_on;	
		break;
	case SENSORS_PROP_REGISTER:		/* read only */
		compass_active_e2p();
		compass_blkread(AKECS_REG_EHXGA, ehga, 3);
		return	sprintf(val->strval, "%d %d %d", ehga[0], ehga[1], ehga[2]);
	default:
		return -EINVAL;
	}

	return 0;
}

struct sensors_dev akm8973_magnetic_dev = {
	.id = SENSORS_MAGNETIC_FIELD,
	.name = "magnetic",
	.properties = akm_properties,
	.num_properties = ARRAY_SIZE(akm_properties),
	.get_property = akm_get_property,
	.put_property = akm_put_property,
	.fops = &akm8973_dev_fops,
};

struct sensors_dev akm8973_orientation_dev = {
	.id = SENSORS_ORIENTATION,
	.name = "orientation",
	.properties = akm_properties,
	.num_properties = ARRAY_SIZE(akm_properties),
	.get_property = akm_get_property_dummy,
	.put_property = akm_put_property_dummy,
	.fops = &akm_dummy_fops,
};

static int compass_blkread(u8 reg, u8 *pval, u8 len)
{
	int ret;
	struct i2c_client *compass_client = this_client;
	if (!compass_client)
		return -EINVAL;
	ret = i2c_smbus_read_i2c_block_data(compass_client, reg, len, pval);

	return ret;
}

static int compass_active(void)
{
	return AKECS_StartMeasure();
}

static int compass_active_e2p(void)
{
	return AKECS_StartE2PRead();
}

static struct early_suspend akm8973_early_suspend_desc = {
	.level = 1,
	.suspend = akm8973_early_suspend,
	.resume = akm8973_early_resume,
};

static int  __devinit akm8973_probe(struct i2c_client *client, const struct i2c_device_id *id)
{
	int err;

	if (!i2c_check_functionality(client->adapter, I2C_FUNC_I2C)) {
		return -ENODEV;
	}

	compass_device.compass_client = client;	
	this_client = client;

	INIT_DELAYED_WORK(&compass_device.compass_work, compass_int_routine);
	init_waitqueue_head(&compass_device.compass_wait);
	INIT_LIST_HEAD(&compass_device.open_list);
	init_MUTEX(&compass_device.mutex);
	spin_lock_init(&compass_device.list_lock);
	compass_device.akm_switch_on = SWITCH_OFF;
	compass_device.dummy_switch_on = SWITCH_OFF;
	compass_device.akm_interval = COMPASS_DEFAULT_INTERVAL; 
	//setup_timer(&compass_device.timer, compass_wake_up_int, 0);

	pdata = client->dev.platform_data;
	pdata->init_irq();
	err = request_irq(client->irq, compass_irq, IRQF_TRIGGER_RISING, "AK-COMPASS", client);	
	if (err) {
		printk(KERN_ERR "akm8973 request iqr failed!\n");
		goto failed_request_irq;
	}

	if (pdata->reset_pin) {
		err = pdata->reset_pin();
		if (err) {
			printk(KERN_ERR "akm8973 reset failed!\n");
			goto failed_reset_pin;
		}
	}
	err = compass_active();
	if (err < 0) {
		printk(KERN_ERR "Akm8973 device probe failed!\n");
		goto failed_check_chip;
	} 

	err = sensors_dev_register(&akm8973_magnetic_dev);
	if (err) {
		printk(KERN_INFO "akm8973 compass register fail!\n");
		goto failed_register_dev_magnetic;
	} else {
		printk(KERN_INFO "akm8973 compass register OK!\n");
	}
	
	err = sensors_dev_register(&akm8973_orientation_dev);
	if (err) {
		printk(KERN_INFO "akm8973 compass register fail!\n");
		goto failed_register_dev_orientation;
	} else {
		printk(KERN_INFO "akm8973 orientation dummy sensor device register OK!\n");
	}


#ifdef CONFIG_HAS_EARLYSUSPEND
	register_early_suspend(&akm8973_early_suspend_desc);
#endif
	return 0;

failed_register_dev_orientation:		
	sensors_dev_unregister(&akm8973_magnetic_dev);
failed_register_dev_magnetic:
failed_check_chip:
failed_reset_pin:
	free_irq(client->irq, NULL);
failed_request_irq:
	return err;
}

static int __devexit akm8973_remove(struct i2c_client *client)
{
#if DEBUG
	printk(KERN_INFO "AK8973 compass driver: remove\n");
#endif
	free_irq(client->irq, NULL);
#ifdef CONFIG_HAS_EARLYSUSPEND
	unregister_early_suspend(&akm8973_early_suspend_desc);
#endif
	return 0;
}

static const struct i2c_device_id akm8973_id[] = {
	{ "akm8973", 0 },
	{ }
};

static struct i2c_driver akm8973_driver = {
	.class = I2C_CLASS_HWMON,
	.probe = akm8973_probe,
	.remove = akm8973_remove,
	.id_table = akm8973_id,
	.driver = {
		   .owner = THIS_MODULE,
		   .name = "akm8973",
		   },
};

static int __init akm8973_init(void)
{
	int ret;
	ret = i2c_add_driver(&akm8973_driver);
	
	if (ret)
		printk(KERN_WARNING "AK8973 compass: driver registration failed, module not inserted.\n");

	return ret;
}

static void __exit akm8973_exit(void)
{
#if DEBUG
	printk(KERN_INFO "AK8973 compass driver: exit\n");
#endif
	sensors_dev_unregister(&akm8973_magnetic_dev);
	sensors_dev_unregister(&akm8973_orientation_dev);

	i2c_del_driver(&akm8973_driver);
}

module_init(akm8973_init);
module_exit(akm8973_exit);

MODULE_AUTHOR("BORQS");
MODULE_DESCRIPTION("AK8973 compass driver");
MODULE_LICENSE("GPL");
