/*
 *  drivers/i2c/chips/compass_sensor.c
 *
 * Copyright (C) 2008 Borqs, Inc.
 * 
 * This software is licensed under the terms of the GNU General Public
 * License version 2, as published by the Free Software Foundation, and
 * may be copied, distributed, and modified under those terms.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
*/

#include <linux/init.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/sysctl.h>
#include <linux/platform_device.h>
#include <linux/i2c.h>
#include <linux/cdev.h>
#include <linux/major.h>
#include <linux/interrupt.h>
#include <linux/poll.h>
#include <asm/ioctl.h>
#include <asm/uaccess.h>
#include <mach/pxa-regs.h>
#include <mach/pxa3xx-regs.h>
#include <mach/compass_sensor.h>
#include <mach/mfp-pxa9xx.h>
#include <mach/gpio.h>
#include <linux/input.h>
#include <linux/sensors.h>
#include <linux/delay.h>

static int interval = 100;
static int xyz_direction, x_dir, y_dir, z_dir;
static compass_device_t compass_device = {
	.name 			= "compass_device",
	.rf_count		= ATOMIC_INIT(0),
};
static void compass_power_on(void);
static void compass_power_off(void);

static int compass_read(int reg, int *pval)
{
	int ret;
	struct i2c_client *compass_client = compass_device.compass_client;
	if (!compass_client)
		return -EINVAL;
	ret = i2c_smbus_read_byte_data(compass_client,reg);
	if (ret >= 0) {
		*pval = ret;
		ret = 0;
	} else
		ret = -EIO;

	return ret;
}

static int compass_write(u8 reg, u8 val)
{
	int ret;

	if (compass_device.compass_client == NULL)
		return -1;

	ret = i2c_smbus_write_byte_data(compass_device.compass_client, reg, val);
	if (ret == 0)
		ret = 0;
	else
		ret = -EIO;

	return ret;
}

static void compass_active(void)
{
	compass_write(COMPASS_MODE, SENSOR_MEASURE);
}
static void compass_initial(struct compass_platform_data *pdata)
{
	int ret;
	printk("compass is initializing...\n");
#if 0
	ret = gpio_request(pdata->power_pin, "compass power");
	if (ret){
		printk("gpio_requset failed, return :%d\n",ret);
	}
#endif
	ret = gpio_request(pdata->reset_pin, "compass reset");
	if (ret){
		printk("gpio_requset failed, return :%d\n",ret);
	}
	//power on high
	//gpio_direction_output(pdata->power_pin, 1);
	//mdelay(100);

	gpio_direction_output(pdata->reset_pin, 0);
	mdelay(2000);

	gpio_direction_output(pdata->reset_pin, 1);
	mdelay(20);

	compass_active();
#if 1 
	int x,y,z,i,st;
	 compass_read(COMPASS_X, &x);
	 compass_read(COMPASS_STATUS, &st);
	 compass_read(COMPASS_Y, &y);
	 compass_read(COMPASS_Z, &z);
	printk(KERN_INFO "compass read: x=%d,y=%d,z=%d,status = 0x%x\n",
			(signed char)x, (signed char)y, (signed char)z, st);
	//compass_active();
	mdelay(100);
#endif
	return 0;
}
static void compass_wake_up_int(unsigned long data)
{
#if 0
	compass_status_t *status_t;

	spin_lock(&compass_device.list_lock);
	list_for_each_entry(status_t, &compass_device.open_list, list) {
		atomic_set(&status_t->g_rdy, 1);
	}
	spin_unlock(&compass_device.list_lock);

	wake_up_interruptible(&compass_device.compass_wait);
#endif
	schedule_work(&compass_device.compass_work);
	mod_timer(&compass_device.timer, jiffies + msecs_to_jiffies(interval));

}
static void compass_int_routine(struct work_struct *work)
{
	printk(KERN_INFO "compass int trigger.\n");	
	int x,y,z,i,st;
	 compass_read(COMPASS_X, &x);
	 compass_read(COMPASS_STATUS, &st);
	 compass_read(COMPASS_Y, &y);
	 compass_read(COMPASS_Z, &z);
	printk(KERN_INFO "compass read: x=%d,y=%d,z=%d,status = 0x%x\n",
			(signed char)x, (signed char)y, (signed char)z, st);

}

static irqreturn_t compass_irq(int irq, void *dev_id)
{
	schedule_work(&compass_device.compass_work);
	mod_timer(&compass_device.timer, jiffies + msecs_to_jiffies(2000));
	return IRQ_HANDLED;
}


static int compass_dev_open(struct inode *inode, struct file *file)
{
	int ret;
	compass_status_t *status_p;

	printk(KERN_INFO "compassDLE open\n");
	
	if (down_interruptible(&compass_device.mutex))
		return -ERESTARTSYS;

	if ((file->f_flags & O_ACCMODE) == O_WRONLY) {
		printk(KERN_INFO "COMPASS's device node is readonly\n");
		ret = -EPERM;
	} else {
		printk(KERN_INFO "COMPASS's device :You are reading !\n");
		ret = 0;
	}
	
	status_p = kzalloc(sizeof(compass_status_t), GFP_KERNEL);
	if (status_p == NULL) { 
		printk(KERN_INFO "No mem!\n");
		up(&compass_device.mutex);
		return -ENOMEM;
	}
	file->private_data = status_p;
	atomic_set(&status_p->g_rdy, 1);
	spin_lock(&compass_device.list_lock);
	list_add(&status_p->list, &compass_device.open_list);
	spin_unlock(&compass_device.list_lock);

	atomic_inc(&compass_device.rf_count);
	printk(KERN_INFO "rf_count = %d\n", atomic_read(&compass_device.rf_count));
	up(&compass_device.mutex);

	return ret;
}
static ssize_t compass_dev_read(struct file *file, char __user * buffer,
				  size_t size, loff_t * f_pos)
{
	struct input_event event[3];
	char x, y, z;
	int retx, rety, retz;
	int ret;
	compass_status_t *status_p;
	printk(KERN_INFO "compassE read(xyz)\n");

	if (down_interruptible(&compass_device.mutex))
		return -ERESTARTSYS;

	status_p = file->private_data;
	if( status_p == NULL) {
		up(&compass_device.mutex);
		return -EFAULT;
	}
	
	retx = compass_read(COMPASS_X, &x);
	rety = compass_read(COMPASS_Y, &y);
	retz = compass_read(COMPASS_Z, &z);
	compass_active();
	if (retx | rety | retz) {
		printk(KERN_INFO "compass read Error!\n");
		ret = -EIO;
		goto read_out;
	}

	printk(KERN_INFO "compass read: x=%d,y=%d,z=%d\n",
			(signed char)x, (signed char)y, (signed char)z);

	do_gettimeofday(&event[0].time);
	event[1].time = event[0].time;
	event[2].time = event[0].time;

	event[0].type = EV_ABS;
	event[1].type = EV_ABS;
	event[2].type = EV_ABS;

	event[0].value = (int)((signed char)y * x_dir * MG_PER_SAMPLE);
	event[0].code = ABS_X;

	event[1].value = (int)((signed char)x * y_dir * MG_PER_SAMPLE);
	//event[1].value = (int)((signed char)x * (-1) *  MG_PER_SAMPLE);
	event[1].code = ABS_Y;

	event[2].value = (int)((signed char)z * z_dir * MG_PER_SAMPLE);
	event[2].code = ABS_Z;

	if (copy_to_user(buffer, event, sizeof(event))) {
		printk(KERN_INFO "Copy Gsensor data Fail!\n");
		ret = -EIO;	
	} else
		ret = sizeof(event);
read_out:
	atomic_set(&status_p->g_rdy, 0);
	up(&compass_device.mutex);
	return ret;
}
static unsigned int compass_dev_poll(struct file *filep, poll_table * pt)
{

	unsigned int mask = 0;
	compass_status_t *status_p;
	
	if (down_interruptible(&compass_device.mutex))
		return -ERESTARTSYS;

	status_p = filep->private_data;
	if (status_p == NULL) {
		up(&compass_device.mutex);
		return -EFAULT;
	}

	poll_wait(filep, &compass_device.compass_wait, pt);

	if (atomic_read(&status_p->g_rdy)) 
		mask |= POLLIN | POLLRDNORM;

	up(&compass_device.mutex);
	return mask;
}

static int compass_dev_release(struct inode *inode, struct file *filp)
{
	compass_status_t *status_p;

	if (down_interruptible(&compass_device.mutex))
		return -ERESTARTSYS;

	status_p = filp->private_data;
	if (status_p == NULL) {
		up(&compass_device.mutex);
		return -EFAULT;
	}
	filp->private_data = NULL;
	
	spin_lock(&compass_device.list_lock);
	list_del(&status_p->list);
	spin_unlock(&compass_device.list_lock);

	kfree(status_p);

	atomic_dec(&compass_device.rf_count);

	up(&compass_device.mutex);
	
	return 0;
}

static struct file_operations compass_dev_fops = {
	.owner = THIS_MODULE,
	.open = compass_dev_open,
	.read = compass_dev_read,
	.poll = compass_dev_poll,
	.release = compass_dev_release,
};

struct sensors_dev compass_sensor_dev = {
	.id = SENSORS_COMPASS,
	.name = "compass",
	.fops = &compass_dev_fops,
};

#if 0
static irqreturn_t compass_irq(int irq, void *dev_id)
{
	printk(KERN_INFO "compass come to irq routine \n");
	return IRQ_HANDLED;
}

static void timer_func(unsigned long data)
{
	u8 value_x,value_y,value_z;
	value_x = compass_read(compass_client,0xc2);
	value_y = compass_read(compass_client,0xc3);
	value_z = compass_read(compass_client,0xc4);
	printk(KERN_INFO "x = %d,y = %d, z = %d\n",value_x,value_y,value_z);
	//mod_timer(&compass_timer, jiffies + msecs_to_jiffies(1000));

}
#endif
static ssize_t compass_show(struct class_device *cdev, char *buf)
{
	int val;
	char *s = buf;

	compass_read(COMPASS_STATUS, &val);
	s += sprintf(s, "status register c0: %d\n", (signed char)val);
	compass_read(COMPASS_TEMP, &val);
	s += sprintf(s, "tem register c0: %d\n", (signed char)val);
	compass_read(COMPASS_X, &val);
	s += sprintf(s, "x : %d\n", (signed char)val);
	compass_read(COMPASS_Y, &val);
	s += sprintf(s, "y: %d\n", (signed char)val);
	compass_read(COMPASS_Z, &val);
	s += sprintf(s, "z: %d\n", (signed char)val);
	compass_active();
	return (s - buf);
}

static ssize_t compass_store(struct class_device *cdev, const char *buf,
			    size_t count)
{
	int compass_answer;
	sscanf(buf, "compass store =%d", &compass_answer);
	return count;
}
static ssize_t switch_compass_show(struct class_device *cdev, char *buf)
{
	char *s = buf;
	if (compass_device.switch_on)
		s += sprintf(s, "switch is enable  \n");
	else
		s += sprintf(s, "switch is disable  \n");
	return (s - buf);
}

static ssize_t switch_compass_store(struct class_device *cdev, const char *buf, size_t count)
{
	char temp_switch = -1;
	if (!strncmp(buf, "on", 2))
		temp_switch = 1;
	else if (!strncmp(buf, "off", 3))
		temp_switch = 0;
	else
		printk(KERN_INFO "Input on or off  \n");

	if ((temp_switch == 0) && 
		compass_device.switch_on) {
			compass_device.switch_on = SWITCH_OFF;
			compass_power_off();
			printk(KERN_INFO "compass power off! \n");
	} else if ((temp_switch == 1) && 
			!compass_device.switch_on) {
			compass_device.switch_on = SWITCH_ON;
			compass_power_on();
			printk(KERN_INFO "compass power on !\n");
	}
	
	return count; 
}

static ssize_t interval_compass_show(struct class_device *cdev, char *buf)
{
	char *s = buf;
	s += sprintf(s, "interval_compass_show \n interval = %d\n",interval);
	return (s - buf);
}

static ssize_t interval_compass_store(struct class_device *cdev, const char *buf, size_t count)
{
	sscanf(buf, "%d\n", &interval);

	if (interval > 2000) {
		interval = COMPASS_MAX_INTERVAL;
	}

	if (interval < 0) {
		interval = COMPASS_MIN_INTERVAL;
	}
	
	printk(KERN_INFO "interval = %d\n", interval);

	return count; 
}
static CLASS_DEVICE_ATTR(compass, S_IRUGO | S_IWUGO, 
			compass_show, 
			compass_store);
static CLASS_DEVICE_ATTR(switch, S_IRUGO | S_IWUGO, 
			switch_compass_show,
			switch_compass_store);

static CLASS_DEVICE_ATTR(interval, S_IRUGO | S_IWUGO, 
			interval_compass_show,
			interval_compass_store);
static struct attribute* compass_attr[] = {
	&class_device_attr_compass.attr,
	&class_device_attr_switch.attr,
	&class_device_attr_interval.attr,
	NULL,
};

static struct attribute_group compass_group = {
	.attrs = compass_attr,
};	
static void compass_early_suspend(struct early_suspend *h)
{
	compass_power_off();
	compass_device.switch_on = SWITCH_OFF;
}

static void compass_early_resume(struct early_suspend *h)
{
	compass_power_on();
	compass_device.switch_on = SWITCH_ON;
}

static struct early_suspend compass_early_suspend_desc = {
	.level  = 1,
	.suspend = compass_early_suspend,
	.resume = compass_early_resume,
};

#if 0
static int compass_register_create_sysfile(void)
{
	int ret;
	
	struct class_device *cdev_pointer;
	
	ret = sensors_dev_register(&compass_sensor_dev);
	if (ret) {
		printk(KERN_INFO "compass register fail!\n");
		sensors_dev_unregister(&compass_sensor_dev);
		return -EIO;
	}
	printk(KERN_INFO "compass register OK!\n");
	
	cdev_pointer = compass_sensor_dev.cdev;
	ret = sysfs_create_group(&cdev_pointer->kobj, &compass_group);
	return ret;
}
#endif
static int compass_probe(struct i2c_client *client)
{
	struct compass_platform_data *pdata;
	int ret;

	printk(KERN_INFO "compass probe\n");
	if (client == NULL) {
		printk(KERN_ERR "Client is NUll!\n");
		return -EFAULT;
	}
	register_early_suspend(&compass_early_suspend_desc);

	compass_device.compass_client = client;
	
	INIT_WORK(&compass_device.compass_work, compass_int_routine);
	init_waitqueue_head(&compass_device.compass_wait);
	INIT_LIST_HEAD(&compass_device.open_list);
	init_MUTEX(&compass_device.mutex);
	spin_lock_init(&compass_device.list_lock);
	compass_device.switch_on = SWITCH_ON;
	setup_timer(&compass_device.timer, compass_wake_up_int, 0);

	compass_device.device = &compass_sensor_dev;
	
#if 0
	ret = compass_register_create_sysfile();
	if (ret) {
		printk(KERN_ERR "compass create sysfile fail!\n");
	}
#endif
	pdata = client->dev.platform_data;
	pdata->init_irq();
	ret = request_irq(client->irq, compass_irq, IRQF_TRIGGER_FALLING | IRQF_TRIGGER_RISING, "COMPASS", client);	
	printk(KERN_INFO "compass: request_irq = %d\n", ret);
	compass_initial(pdata);
}
static int compass_remove(struct i2c_client *client)
{	
	free_irq(client->irq, NULL);
	unregister_early_suspend(&compass_early_suspend_desc);
	return 0;
}

static void compass_power_on(void)
{
	
	return ;
}

static void compass_power_off(void)
{
	
	return ;
}

#ifdef CONFIG_PM

static int compass_suspend(struct i2c_client *client, pm_message_t mesg)
{
	/*Make sure the device is power off when suspend*/
	compass_power_off();
	return 0 ;
}

static int compass_resume(struct i2c_client *client)
{
	/*Wake up the device*/
	compass_power_on();
	return 0 ;
}

#else

#define compass_suspend 	NULL
#define compass_resume 	NULL

#endif

static struct i2c_driver compass_driver = {
	.driver = {
		.name = "compass_sensor",
	},
	.probe = compass_probe,
	.remove = compass_remove,
	.suspend = compass_suspend,
	.resume = compass_resume,
};

static int __init compass_init(void)
{
	int ret;

	ret = i2c_add_driver(&compass_driver);
	if (ret) 
		printk(KERN_WARNING "COMPASS: Driver registration failed, module not inserted.\n");
	else
		printk(KERN_WARNING "COMPASS: i2c_add_driver() OK!\n");
	compass_device.driver = &compass_driver;
	
	return ret;
}

static void __exit compass_exit(void)
{
	i2c_del_driver(&compass_driver);
	sensors_dev_unregister(&compass_sensor_dev);
}

module_init(compass_init);
module_exit(compass_exit);

MODULE_DESCRIPTION("Sensors Compass");
MODULE_LICENSE("GPL");
