/*
 *  IMAGIS GPU Driver
 *
 * This software program is licensed subject to the GNU General Public License
 * (GPL).Version 2,June 1991, available at http://www.fsf.org/copyleft/gpl.html
 */

#include <linux/init.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/i2c.h>
#include <linux/proc_fs.h>
#include <asm/uaccess.h>
#include <mach/ism2300s.h>

#define IMAGIS_REG_NUM		(0x100)

static struct imagis_regs imagis_regs[IMAGIS_REG_NUM];

/* Unique ID allocation */
struct imagis_data {
	struct i2c_client 	*g_client;
};

static struct imagis_data *imagis_d;

int imagis_read(u8 reg, u8 *pval)
{
	int ret;
	int status;

	if (imagis_d->g_client == NULL) {	/* No global client pointer? */
		return -1;
	}

	ret = i2c_smbus_read_byte_data(imagis_d->g_client, reg);
	if (ret >= 0) {
		*pval = ret;
		imagis_regs[reg].inited = 1;
		imagis_regs[reg].data = ret;
		status = 0;
	} else
	{
		printk("%s: error read reg %d\n",__func__, reg);
		status = -EIO;
	}

	return status;
}

int imagis_read_block(u8 reg, u8 *pval, u8 len)
{
	int ret;

	if (imagis_d->g_client == NULL) {	/* No global client pointer? */
		return -1;
	}

	ret = i2c_smbus_read_i2c_block_data(imagis_d->g_client, reg, len, pval);
	if (ret != len)
	{
		printk("%s: error read reg %d-%d\n",__func__, reg,reg+len-1);
		ret = -EIO;
	}

	return ret;

}

int imagis_write(u8 reg, u8 val)
{
	int ret;
	int status;

	//printk("%s: reg 0x%x,val 0x%x\n",__func__,reg,val);
	if (imagis_d->g_client == NULL) {	/* No global client pointer? */
		return -1;
	}

	ret = i2c_smbus_write_byte_data(imagis_d->g_client, reg, val);
	if (ret == 0) {
		status = 0;
	} else{
		printk("%s: error write reg %d\n",__func__, reg);
		status = -EIO;
	}

	return status;
}

#ifdef CONFIG_PM

#define	imagis_suspend		NULL
#define	imagis_resume		NULL
#endif				/* CONFIG_PM */

#ifdef	CONFIG_PROC_FS
#define	IMAGIS_PROC_FILE	"driver/imagis"
static struct proc_dir_entry *imagis_proc_file;
static int index;

static ssize_t imagis_proc_read(struct file *filp, char __user *buffer, size_t length, loff_t *off)
{
	u8 reg_val;
        const char fmt[]="register 0x%02x: 0x%02x\n";
        char buf[sizeof(fmt)];/* %2x is 4 chars while it yields 2 chars in output */
        size_t len;

	if ((index < 0) || (index >= IMAGIS_REG_NUM))
		return 0;

        /* It would be correct to first produce the full output again on the second call with *off>0 
           However, the total size required is so small that re-reading imagis reg again is redundant */
        if(*off) return 0;

	imagis_read(index, &reg_val);
        
	len=sprintf(buf, fmt, (unsigned)index, (unsigned)reg_val);
        len-=*off;
        if(length<len) return -EFAULT;/*unreasonably short buffer supplied*/
        if(len>length) len=length;
        if (copy_to_user(buffer, buf+(*off), len))
		return -EFAULT;
        *off+=len;
	return len;
}

static ssize_t imagis_proc_write(struct file *filp,
		const char *buff, size_t len, loff_t *off)
{
	u8 reg_val;
	char messages[256], vol[256];

	if (len > 256)
		len = 256;

	if (copy_from_user(messages, buff, len))
		return -EFAULT;

	if (strncmp(messages, "dump", 4) == 0){
		int i;
		for (i=1; i <0xE; i++){
			imagis_read(i, &reg_val);
			printk(KERN_INFO " reg %x = 0x%x\n",i , reg_val);
		}	

	}
	else if ('-' == messages[0]) {
		/* set the register index */
		memcpy(vol, messages+1, len-1);
		index = (int) simple_strtoul(vol, NULL, 16);
	} else {
		/* set the register value */
		reg_val = (int)simple_strtoul(messages, NULL, 16);
		imagis_write(index, reg_val & 0xFF);
	}

	return len;
}

static struct file_operations imagis_proc_ops = {
	.read = imagis_proc_read,
	.write = imagis_proc_write,
};

#define	IMAGIS_ALL_PROC_FILE	"driver/imagis_all"
static ssize_t imagis_all_proc_read(struct file *filp,
                char *buffer, size_t length, loff_t *offset)
{
        u8 reg_val;
	int imagis_reg;

	printk(KERN_INFO "-------Group 00-------\n");
	imagis_write(0x00,0x00);
	for (imagis_reg = 0; imagis_reg < IMAGIS_REG_NUM; ++imagis_reg) {
        	imagis_read(imagis_reg, &reg_val);
        	printk(KERN_INFO "register 0x%x: 0x%x\n", imagis_reg, reg_val);
	}

	printk(KERN_INFO "-------Group 01-------\n");
	imagis_write(0x00,0x01);
	for (imagis_reg = 0; imagis_reg < IMAGIS_REG_NUM; ++imagis_reg) {
        	imagis_read(imagis_reg, &reg_val);
        	printk(KERN_INFO "register 0x%x: 0x%x\n", imagis_reg, reg_val);
	}

	printk(KERN_INFO "-------Group 02-------\n");
	imagis_write(0x00,0x02);
	for (imagis_reg = 0; imagis_reg < IMAGIS_REG_NUM; ++imagis_reg) {
        	imagis_read(imagis_reg, &reg_val);
        	printk(KERN_INFO "register 0x%x: 0x%x\n", imagis_reg, reg_val);
	}	
        return 0;
}

static struct file_operations imagis_all_proc_ops = {
        .read = imagis_all_proc_read,
};

static void create_imagis_proc_file(void)
{
	imagis_proc_file = create_proc_entry(IMAGIS_PROC_FILE, 0644, NULL);
	if (imagis_proc_file) {
		imagis_proc_file->owner = THIS_MODULE;
		imagis_proc_file->proc_fops = &imagis_proc_ops;
	} else
		printk(KERN_INFO "proc file create failed!\n");
        imagis_proc_file = create_proc_entry(IMAGIS_ALL_PROC_FILE, 0444, NULL);
        if (imagis_proc_file) {
                imagis_proc_file->owner = THIS_MODULE;
                imagis_proc_file->proc_fops = &imagis_all_proc_ops;
        } else
                printk(KERN_INFO "proc file create failed!\n");
}

static void remove_imagis_proc_file(void)
{
	remove_proc_entry(IMAGIS_ALL_PROC_FILE, NULL);
	remove_proc_entry(IMAGIS_PROC_FILE, NULL);
}

#endif

static int __init imagis_probe(struct i2c_client *client,
			const struct i2c_device_id *id)
{
	struct imagis_platform_data *pdata;
	int chip_id;
	
	pdata = client->dev.platform_data;
	if (pdata->platform_init)
	      pdata->platform_init();

	chip_id = i2c_smbus_read_byte_data(client, 0x01);
	if (chip_id < 0) {
		printk(KERN_WARNING "imagis unavailable!\n");
		imagis_d->g_client = NULL;
		return -ENXIO;
	} else {
		printk(KERN_INFO "imagis(chip id:0x%x) detected.\n", chip_id);
	}

	imagis_d->g_client = client;


#ifdef	CONFIG_PROC_FS
	create_imagis_proc_file();
#endif

	return 0;
}

static int imagis_remove(struct i2c_client *client)
{
#ifdef	CONFIG_PROC_FS
	remove_imagis_proc_file();
#endif

	return 0;
}

static const struct i2c_device_id imagis_id[] = {
	{ "ism2300s", 0 },
	{ }
};
static struct i2c_driver imagis_driver = {
	.driver = {
		.name	= "ism2300s",
	},
	.probe		= imagis_probe,
	.remove		= imagis_remove,
	.id_table	= imagis_id,
	.suspend	= imagis_suspend,
	.resume		= imagis_resume,
};

static int __init imagis_init(void)
{
	imagis_d = kzalloc(sizeof(struct imagis_data), GFP_KERNEL);
        if (!imagis_d) {
                return -ENOMEM;
        }

	return i2c_add_driver(&imagis_driver);
}

static void __exit imagis_exit(void)
{
	i2c_del_driver(&imagis_driver);
	kfree(imagis_d);
}

subsys_initcall(imagis_init);
module_exit(imagis_exit);

MODULE_DESCRIPTION("Imagis Driver");
MODULE_LICENSE("GPL");
