/*
 * Tavor Levante PMIC Management Routines
 *
 *  linux/i2c/drivers/chips/levante.c
 *
 *  driver for pxa935 - Levante
 *
 *  Copyright (C) 2006, Marvell Corporation
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2 as
 *  published by the Free Software Foundation.
 */

#include <linux/init.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/i2c.h>
#include <linux/proc_fs.h>
#include <linux/delay.h>         /* for msleep definition */


#include <asm/uaccess.h>
#include <mach/levante_hw.h>
#include <mach/levante.h>
#include <mach/micco.h>


#include <mach/hardware.h>
#include <mach/pxa-regs.h>
#include <mach/tavor_common.h>

/* Internal Definitions */
#define TSI_SAMPLES_NUM			3
#define TSI_MAX_SAMPLE_DIF_X    100//12
#define TSI_MAX_SAMPLE_DIF_Y    100//9

#define TRUE        1
#define FALSE       0


static u8 pmicId;
/* Unique ID allocation */
static struct i2c_client *g_client;

unsigned int levante_pxa_ticks_to_msec(unsigned int ticks)
{
#ifdef CONFIG_PXA_32KTIMER
	return (ticks * 5*5*5) >> 12;
#else
	return ticks / 3250;
#endif
}

int levante_enable_pen_down_irq(int enable)
{
	int ret;
	u8 val;

	printk(LVT_DEBUG "----->levante_enable_pen_down_irq begin\n");

	if (enable) {
		printk(LVT_DEBUG_MUST \
		"----->levante_enable_pen_down_irq:enable\n");

		{   /* before enabling the pen IRQ need to make sure pen measurment is enable and x,y measurment is disable */
			int status;
    		u8 val;

			status = sanremo_read(SANREMO_MEAS_ENABLE3, &val);
			val &= ~(SANREMO_MEAS_EN3_TSIX | SANREMO_MEAS_EN3_TSIY);
    		val |= (SANREMO_MEAS_EN3_PENDET);
    		status = sanremo_write(SANREMO_MEAS_ENABLE3, val);
		}

		/* enable pen down IRQ */
		/*for A1 only - before enable pen-det clear pen event in status register - in B0 need to enable the bit that doit auto*/
		sanremo_read(SANREMO_INT_STATUS_3, &val);
		if ((pmicId >= LEVATNE_A0_ID) && (pmicId <=  LEVATNE_A1_ID)) {
			val &= ~(SANREMO_INT_STS3_PEN);
			sanremo_write(SANREMO_INT_STATUS_3, val);
		}
		ret = sanremo_read(SANREMO_INT_EN_3, &val);
		printk(LVT_DEBUG_MUST "----->levante_enable_pen_down_irq: SANREMO_INT_EN_3 = [0x%x]  \n",val);
        if (!(val & SANREMO_INT_EN_PEN)) {
			val |= SANREMO_INT_EN_PEN;
			ret = sanremo_write(SANREMO_INT_EN_3, val);
			if(ret != 0)
				printk(LVT_DEBUG_MUST "----->levante_enable_pen_down_irq - fail \n");
			   else
				printk(LVT_DEBUG_MUST "----->levante_enable_pen_down_irq - success \n");
		}

	} else {

		/* disable pen down IRQ */
		printk(LVT_DEBUG_MUST \
		"----->levante_enable_pen_down_irq:disable\n");
		ret = sanremo_read(SANREMO_INT_EN_3, &val);
		printk(LVT_DEBUG_MUST "----->levante_enable_pen_down_irq: SANREMO_INT_EN_3 = [0x%x]  \n",val);
		if (val & SANREMO_INT_EN_PEN) {
			val &= ~SANREMO_INT_EN_PEN;
			ret = sanremo_write(SANREMO_INT_EN_3, val);
			if(ret != 0)
				printk(LVT_DEBUG_MUST "----->levante_disable_pen_down_irq - fail \n");
			   else
				printk(LVT_DEBUG_MUST "----->levante_disable_pen_down_irq - success \n");
		}
	}
	printk(LVT_DEBUG "------->levante_enable_pen_down_irq end\n");


	return ret;
}

static int calc_avg(int* buff, int size, int max_diff)
{
    int result = 0, i;
    int max_sample = buff[0];
    int min_sample = buff[0];
    int used_samples = size;

    for (i=0; i< size; ++i)
    {
        if(buff[i] == -1)
        {
            used_samples--;
            continue;
        }

        result += buff[i];
        max_sample =  max_sample < buff[i] ? buff[i] : min_sample;
        min_sample =  min_sample > buff[i] ? buff[i] : min_sample;
    }
    if( (max_sample - min_sample) > max_diff ) // this is not a valid sample there are big diffs
    {
        printk(LVT_DEBUG_MUST "levante xy measurment averaging failed, sample diff is too big (%d)!\n", (max_sample - min_sample));
        return -EIO;
    }
    result /= used_samples;
    return result;
}

int levante_tsi_measurements_config(void)
{
    int status;
    u8 val = 0;

	printk(LVT_DEBUG_MUST "levante_tsi_measurements_config begin \n" );

    // set measurements timing
 	if((pmicId >= LEVATNE_A0_ID) && (pmicId <=  LEVATNE_A1_ID) )
		val = (SANREMO_MEASOFFTIME1MEAS_OFFTIME1_A1(0x0A) | SANREMO_MEASOFFTIME1_BAT_DET_EN_A1);

	if  (pmic_id() >= LEVATNE_B0_ID){
		val = (SANREMO_MEASOFFTIME1MEAS_OFFTIME1_B(0x00) | SANREMO_MEASOFFTIME1MEAS_EN_SLP_B);
	}
	status = sanremo_write(SANREMO_MEAS_OFF_TIME1, val);
    if (status)
        return -EIO;

	status = sanremo_read(SANREMO_DEBOUNCE_REG, &val);
    val |= SANREMO_DEBOUNCE_PEN_DET(0x03); /* need 8ms(0x03) for BSP only*/
    status |= sanremo_write(SANREMO_DEBOUNCE_REG, val);
    if (status)
        return -EIO;

	if  (pmicId >= LEVATNE_B0_ID){
		val = 0x20; //for B0 SANREMO_MEAS_OFF_TIME2 in the next write.
	}
	status = sanremo_write(SANREMO_MEAS_OFF_TIME2, val);
    if (status)
        return -EIO;

    val = 0x50;// set TSI pre-bias (migth be different between EVB and SAAR depends on Touch panel HW )
    status = sanremo_write(SANREMO_PD_PREBIAS_TIME, val);
    if (status)
        return -EIO;

   val = 0x07;//migth be different between EVB and SAAR depends on Touch panel HW
   status = sanremo_write(SANREMO_TSI_PREBIAS_TIME, val);
    if (status)
        return -EIO;

	printk(LVT_DEBUG_MUST "levante_tsi_measurements_config end \n" );

    return 0;
}

int levante_tsi_disable_all_measurements(void)
{
    int status;
    u8 val;

    // disable TSI measurements
    status = sanremo_read(SANREMO_MEAS_ENABLE3, &val);
    if (status)
        return -EIO;

    val &= ~(SANREMO_MEAS_EN3_PENDET | SANREMO_MEAS_EN3_TSIX | SANREMO_MEAS_EN3_TSIY);
    status = sanremo_write(SANREMO_MEAS_ENABLE3, val);
    if (status)
        return -EIO;

    return 0;
}
int levante_tsi_enable_xy_measurements(void)
{
    int status;
    u8 val;

    /*disable TSI measurements*/
    status = sanremo_read(SANREMO_MEAS_ENABLE3, &val);
    if (status)
	return -EIO;

    val |= (SANREMO_MEAS_EN3_TSIX | SANREMO_MEAS_EN3_TSIY);
    status = sanremo_write(SANREMO_MEAS_ENABLE3, val);
    if (status)
	return -EIO;

    return 0;
}

int levante_tsi_disable_xy_measurements(void)
{
    int status;
    u8 val;

    // disable TSI measurements
    status = sanremo_read(SANREMO_MEAS_ENABLE3, &val);
    if (status)
        return -EIO;

    val &= ~(SANREMO_MEAS_EN3_TSIX | SANREMO_MEAS_EN3_TSIY);
    status = sanremo_write(SANREMO_MEAS_ENABLE3, val);
    if (status)
        return -EIO;

    return 0;
}

int levante_tsi_readxy(u16 *x, u16 *y, u8 *pen_state)
{
    struct levante_platform_data *pdata = g_client->dev.platform_data;
    int i, x_avg, y_avg, status;
    int x_arr[TSI_SAMPLES_NUM], y_arr[TSI_SAMPLES_NUM];
    u8 val[4];

	*x = *y = 0; /*initilize to zero*/

    for (i = 0; i < TSI_SAMPLES_NUM; ++i) {

		status = sanremo_read_block(SANREMO_TSIX_MEAS1,val, sizeof(val));

		if (status) {
			x_arr[0] = x_arr[1] = x_arr[2] = -1;
			y_arr[0] = y_arr[1] = y_arr[2] = -1;
			printk(LVT_DEBUG_MUST \
			"---->levante_tsi_readxy - fail sanremo_read_block \n");
			return -EIO;
		}

		*pen_state = (val[1] & (1<<6));

		x_arr[i] = ((val[0]<<4) | (val[1]&0x0f));
		y_arr[i] = ((val[2]<<4) | (val[3]&0x0f));
	}

    /*calculate the average*/
    x_avg = calc_avg(x_arr, TSI_SAMPLES_NUM, TSI_MAX_SAMPLE_DIF_X);
    y_avg = calc_avg(y_arr, TSI_SAMPLES_NUM, TSI_MAX_SAMPLE_DIF_Y);

    /*if the average value is not valid*/
	if ((x_avg <= 0) || (y_avg <= 0)) {
		printk(LVT_DEBUG_MUST \
		"------->levante_tsi_readxy - average value is not valid \n");
		return -EAGAIN;
		}

    *x = x_avg;
    *y = y_avg;

	printk(LVT_DEBUG "------->levante_tsi_readxy pen_state[%x]cmp[%x] \n",*pen_state,val[1] );
	printk(LVT_DEBUG "------->levante_tsi_readxy X[%d] Y[%d] \n", *x, *y );

    return 0;
}

/* levante Audio functions */
int levante_codec_read(u8 reg, u8 *val)
{
	if (/*(reg >= SANREMO_CODEC_REG_BASE1) &&*/\
		 (reg <= SANREMO_CODEC_REG_END1))
		reg = reg - SANREMO_CODEC_REG_BASE1 + SANREMO_AUDIO_REG_BASE1;
	else if ((reg >= SANREMO_CODEC_REG_BASE2) \
			 && (reg <= SANREMO_CODEC_REG_END2))
		reg = reg - SANREMO_CODEC_REG_BASE2 + SANREMO_AUDIO_REG_BASE2;
	else
		return -EINVAL;
    return sanremo_read(reg, val);
}

int levante_codec_write(u8 reg, u8 val)
{
	if (/*(reg >= SANREMO_CODEC_REG_BASE1) &&*/\
		(reg <= SANREMO_CODEC_REG_END1))
		reg = reg - SANREMO_CODEC_REG_BASE1 + SANREMO_AUDIO_REG_BASE1;
	else
	if ((reg >= SANREMO_CODEC_REG_BASE2) && (reg <= SANREMO_CODEC_REG_END2))
		reg = reg - SANREMO_CODEC_REG_BASE2 + SANREMO_AUDIO_REG_BASE2;
	else {
		printk(KERN_ERR
		"levante_codec_write; Out of range ERROR; reg= 0x%x\n", reg);
		return -EINVAL;
	}
    return sanremo_write(reg, val);
}


#define VIBRA_OFF_VALUE	0
#define VIBRA_ON_VALUE	1

int levante_control_vibrator(unsigned char value)
{
	if (value == VIBRA_OFF_VALUE) {
		portofino_write(PORTOFINO_PF_VIBRATOR1, 0x00);
		portofino_write(PORTOFINO_PF_VIBRATOR2, 0xFF);

	} else if (value == VIBRA_ON_VALUE) {
		portofino_write(PORTOFINO_PF_VIBRATOR1, 0x01);
		portofino_write(PORTOFINO_PF_VIBRATOR2, 0xFF);
	}
	return 0;
}
EXPORT_SYMBOL(levante_control_vibrator);

int levante_set_vibrator(unsigned char value)
{
	u8 val;

	printk(LVT_DEBUG_MUST  \
	"\n LEVANTE: inside levante_set_vibrator value:%d precentage\n", value);
	if ((value > 0) && (value < 100))	{
		/*the steps are of 0.4%,so to covert from
		percent each percent = 2.5,
		to avoid working with floting point we
		double by 5 and devide by 2*/
		val = (value*5)>>1;
		portofino_write(PORTOFINO_PF_VIBRATOR2, val);
		printk(LVT_DEBUG_MUST \
		"\n\rLEVANTE: inside levante_set_vibrator value = %x \n", val);
	} else if (value == 100) {
		printk(LVT_DEBUG_MUST \
		"\n\rLEVANTE: inside levante_set_vibrator 100 value \n") ;
		portofino_read(PORTOFINO_PF_VIBRATOR1, &val);
		val |= 0x80;
		portofino_write(PORTOFINO_PF_VIBRATOR1, val);
	}

	return 0;
}
EXPORT_SYMBOL(levante_set_vibrator);

/*
  This function configures the ADC as requires for CP implementation.
  CP does not "own" the ADC configuration registers and relies on AP:
  Reason: enable automatic ADC measurements needed for CP to get VBAT and RF temperature readings.
 */
static void levante_enable_adc_for_cp(void)
{
	u8 val;

	sanremo_read(SANREMO_MEAS_ENABLE1, &val);
	val |= (SANREMO_MEAS_EN1_TINT | SANREMO_MEAS_EN1_RFTMP); /* Enable auto ADC (RF temprature, and VBAT) */
	sanremo_write(SANREMO_MEAS_ENABLE1, val);

	val = 0 ;

	val |= (SANREMO_GPADC_MISC1_GPFSM_EN);

	if((pmicId >= LEVATNE_A0_ID) && (pmicId <=  LEVATNE_A1_ID) )
		val |= (SANREMO_GPPADC_GP_PREBIAS_TIME_A1(0x03)| SANREMO_GPPADC_SLOW_MODE(0x01));

	if  (pmicId >= LEVATNE_B0_ID)
		val |= (SANREMO_GPPADC_GP_PREBIAS_TIME_B0(0x01) | SANREMO_GPPADC_SLOW_MODE(0x01));

	sanremo_write(SANREMO_GPADC_MISC1, val);

	/* set for Temprature mesurmants*/
	if  (pmicId >= LEVATNE_B0_ID) {
		val = 0;
		val = (SANREMO_GPADC0_GP_BIAS_A0 | \
				SANREMO_GPADC1_GP_BIAS_A1 | \
				SANREMO_GPADC2_GP_BIAS_A2 | \
				SANREMO_GPADC3_GP_BIAS_A3);
		sanremo_write(SANREMO_GPADC_MISC2, val);
	}
}
/*USB ID-GPADC2 measured enable & ALS measured enabled*/
static void levante_enable_usb_als_measure(void)
{
	u8 val;

	sanremo_read(SANREMO_MEAS_ENABLE1, &val);
	val |= (SANREMO_MEAS_EN1_GPADC2 | SANREMO_MEAS_EN1_GPADC3); /* Enable auto ADC (RF temprature, and VBAT) */
	sanremo_write(SANREMO_MEAS_ENABLE1, val);
}

int levante_check_trim(void)
{
	static u8 trimVal, val;

	if (!trimVal) {
		/* ---> enter test page*/
		sanremo_write(0xFA, 0x00);
		sanremo_write(0xFB, 0x00);
		sanremo_write(0xFF, 0x00);
		sanremo_read(SANREMO_TEST_PAGE1_D6, &val);
		trimVal = ((val & 0xF0) >> 4);/*LSB*/
		sanremo_read(SANREMO_TEST_PAGE1_D7, &val);
		trimVal |= ((val & 0x03) << 4);/*MSB*/
		/*<--- exit test page*/
		sanremo_write(0xFE, 0x00);
		sanremo_write(0xFC, 0x00);

		printk(LVT_DEBUG_MUST \
		"levante_probe:triming value for B0[%d] \n", trimVal);
		printk(LVT_DEBUG_MUST \
		"levante_probe:triming {57 is bad value, 31-36 is good} \n");
	}
	if (trimVal == 57)
		printk(LVT_DEBUG_MUST \
		"levante_probe:!!!*** BAD TRIMING[%d] ***!!!! \n" \
		, trimVal);
	return trimVal;
}


#ifdef	CONFIG_PM
static int levante_suspend(struct i2c_client *client, pm_message_t state)
{
    return 0;
}

static int levante_resume(struct i2c_client *client)
{
	return 0;
}
#else
#define	levante_suspend		NULL
#define	levante_resume		NULL
#endif


#ifdef	CONFIG_PROC_FS
#define	LEVANTE_PROC_FILE	"driver/levante"
static struct proc_dir_entry *levante_proc_file;

static ssize_t levante_proc_read(struct file *filp,
        char __user *buffer, size_t length, loff_t *offset)
{
	printk(LVT_DEBUG_MUST "levante_proc_read - no action done\n");

	return 0;
}

static ssize_t levante_proc_write(struct file *filp,
		const char *buff, size_t len, loff_t *off)
{
	char messages[256];
	if	(copy_from_user(messages, buff, len))
		return -EFAULT;

	if	('?' == messages[0])	{
		printk(LVT_DEBUG_MUST	\
		"\n 1. echo viboff > /proc/driver/levante\n");
	  printk(LVT_DEBUG_MUST	\
		"\n 2. echo vibon > /proc/driver/levante\n");
	  printk(LVT_DEBUG_MUST	\
		"\n 3. echo vib [val] > /proc/driver/levante\n");
	  printk(LVT_DEBUG_MUST	\
		"\n 4. echo blon > /proc/driver/levante\n");
	  printk(LVT_DEBUG_MUST	\
		"\n 5. echo bloff > /proc/driver/levante\n");
	  printk(LVT_DEBUG_MUST \
		"\n 6. echo chargeron > /proc/driver/levante\n");
	  printk(LVT_DEBUG_MUST \
		"\n 7. echo tsiopen > /proc/driver/levante\n\n");
	} else if (strncmp(messages, "chargeron", 9) == 0) {
#if defined(CONFIG_CHARGE_MICCO) || defined(CONFIG_CHARGE_LEVANTE)
		charger_on();
		printk(LVT_DEBUG_MUST "Levante charger on\n");
#else
		printk(LVT_DEBUG_MUST "Levante charger is disabled\n");
#endif
	}
#if defined(CONFIG_TOUCHSCREEN_LEVANTE)
	 else if (strncmp(messages, "tsiopen", 7) == 0) {
		levante_open_tsi();
		printk(LVT_DEBUG_MUST "\nLevante tsiopen\n");
	}
#endif
	 else if (strncmp(messages, "viboff", 6) == 0) {
		levante_control_vibrator(VIBRA_OFF_VALUE);
		printk(LVT_DEBUG_MUST "Levante vibrator off\n");
	} else if (strncmp(messages, "vibon", 5) == 0) {
		levante_control_vibrator(VIBRA_ON_VALUE);
		printk(LVT_DEBUG_MUST "Levante vibrator on\n");
	} else if (strncmp(messages, "blon", 4) == 0) {
#ifdef CONFIG_BACKLIGHT_LCD_SUPPORT
		portofino_write(PORTOFINO_WLED1A, 0xFF);
		portofino_write(PORTOFINO_WLED2A, 0xFF);
		lcd_backlight_power_on(0xFF);/*100% intensity*/
		kp_bl_power_on(0xFF);
		printk(LVT_DEBUG_MUST "Levante LCD backlight on\n");
		printk(LVT_DEBUG_MUST "Levante KP LCD backlight on\n");
#else
		printk(LVT_DEBUG_MUST \
			"Levante KP LCD backlight not supported\n");
#endif
	} else if (strncmp(messages, "bloff", 5) == 0) {
#ifdef CONFIG_BACKLIGHT_LCD_SUPPORT
		portofino_write(PORTOFINO_WLED1A, 0x00);
		portofino_write(PORTOFINO_WLED2A, 0x00);
		lcd_backlight_power_off();
		kp_bl_power_off();
		printk(LVT_DEBUG_MUST "Levante LCD backlight off\n");
		printk(LVT_DEBUG_MUST "Levante KP LCD backlight off\n");
#else
		printk(LVT_DEBUG_MUST \
			"Levante KP LCD backlight not supported\n");
#endif
	} else if (strncmp(messages, "bl", 2) == 0) {
#ifdef CONFIG_BACKLIGHT_LCD_SUPPORT
		unsigned val, orig_val = 0;
		sscanf(messages, "bl%d", &val);
		orig_val = val;
		if (val >= 100) {
			val = 0xFF;
			orig_val = 100;
		} else
			/* convert precentage to WLED_DUTY*/
			val = (val*256/100);

		portofino_write(PORTOFINO_WLED1A, val);/*bl*/
		portofino_write(PORTOFINO_WLED2A, val);/*kpbl*/
		lcd_backlight_power_on(val);/*100% intensity*/
		kp_bl_power_on(val);
		printk(LVT_DEBUG_MUST \
		"Levante LCD backlight intensity:%d precentage\n", orig_val);
		printk(LVT_DEBUG_MUST \
		"Levante KP LCD backlight intensity:%d precentage\n", orig_val);
#else
		printk(LVT_DEBUG_MUST \
			"Levante KP LCD backlight not supported\n");
#endif
	} else if (strncmp(messages, "vib", 3) == 0) {
		unsigned val = 0;
		sscanf(messages, "vib%x", &val);
		levante_set_vibrator(val);
	} else {
		printk(LVT_DEBUG_MUST \
			"Levante - no argument in: echo [xxx] > ... \n");
	}

	printk(LVT_DEBUG_MUST "levante_proc_write - done\n\n");

	return len;
}

static struct file_operations levante_proc_ops = {
	.read = levante_proc_read,
	.write = levante_proc_write,
};

static void create_levante_proc_file(void)
{
	levante_proc_file = create_proc_entry(LEVANTE_PROC_FILE, 0644, NULL);
	if (levante_proc_file) {
		levante_proc_file->owner = THIS_MODULE;
		levante_proc_file->proc_fops = &levante_proc_ops;
	} else
		printk(LVT_DEBUG_MUST "proc file create failed!\n");
}

static void remove_levante_proc_file(void)
{
	remove_proc_entry(LEVANTE_PROC_FILE, NULL);
}

#endif
static int __devinit levante_probe(struct i2c_client *client,
			 const struct i2c_device_id *id)
{
    struct levante_platform_data *pdata;

    g_client = client;

	printk(LVT_DEBUG_MUST "levante_probe - start\n");

	/* workaround for pen detect on EVB3 & RF temprature test page */
 	if((pmicId >= LEVATNE_A0_ID) && (pmicId <=  LEVATNE_A1_ID) ){

   	/* ---> enter test page*/
   		sanremo_write(0xFA, 0x00);
   		sanremo_write(0xFB, 0x00);
   		sanremo_write(0xFF ,0x00 );
   		/*---------no Bias current and Yn at the mux during pen detection--------------*/
   		sanremo_write(SANREMO_TEST_PAGE1_C8 ,0x0F );
   		sanremo_write(SANREMO_TEST_PAGE1_C9 ,0x08 );//(selects YN)
   		sanremo_write(SANREMO_TEST_PAGE1_CA ,0xFF );
   		sanremo_write(SANREMO_TEST_PAGE1_CB ,0x0F );
   		sanremo_write(SANREMO_TEST_PAGE1_CC ,0xF3 ); //(mask enabled for debug 10/11)
   		sanremo_write(SANREMO_TEST_PAGE1_CD ,0xFF );
   		sanremo_write(SANREMO_TEST_PAGE1_CF ,0x04 );
		/*-----------------------*/
   		/*<--- exit test page*/
   		sanremo_write(0xFE, 0x00);
   		sanremo_write(0xFC, 0x00);
		printk(LVT_DEBUG_MUST "levante_probe - workaround pen detect & RF temprature test page \n");
	}else
   	    printk(LVT_DEBUG_MUST "levante_probe - *not active* workaround pen detect & RF temprature test page \n");

	#if defined(LEVANTE_USB_ID_RESISTOR_TBD) //this is un-defined will be->TBD.
	if  (pmicId >= LEVATNE_B0_ID){
			u8 val;
			sanremo_read(SANREMO_GP_BIAS_1, &val);
			val |= (SANREMO_GP_BIAS_1_GP3(0x05));
			sanremo_write(SANREMO_GP_BIAS_1, val);
		}
	#endif
		/* Configure the automatic ADC required for CP operation */
    	levante_enable_adc_for_cp();
		if  (pmicId >= LEVATNE_B0_ID){
			levante_enable_usb_als_measure(); /* enabling this harm in A1 the charging measurmanets so only for B0*/
		}
    	pdata = g_client->dev.platform_data;
    	pdata->first_xy_sample = TRUE;

    	pdata->platform_init();

#ifdef	CONFIG_PROC_FS
		create_levante_proc_file();
#endif
		/*  Buck1 Settings - Initialization Sequences 1*/
		sanremo_write(SANREMO_VBUCK_SET_SLEEP1, 0x24);
		sanremo_write(SANREMO_GENERAL_USE, 0x00);
		/*for power only 1 modif*/
		sanremo_write(SANREMO_SLEEP_MODE1_REG, 0xAA);

	if  ((pmicId >= LEVATNE_A0_ID) && (pmicId <=  LEVATNE_A1_ID) ){
		/* IRQ is masked during the power-up sequence and will not be released
		 * until they have been read for the first time - in A1 we need to write to it */
		sanremo_write(SANREMO_INT_STATUS_1, 0x1F);
		sanremo_write(SANREMO_INT_STATUS_2, 0xFF);
		sanremo_write(SANREMO_INT_STATUS_3, 0xFF);
	}

	if  (pmicId >= LEVATNE_B0_ID){
		/* IRQ is masked during the power-up sequence and will not be released
		 * until they have been read for the first time */
		u8 status1, status2, status3;
		sanremo_read(SANREMO_INT_STATUS_1, &status1);
		sanremo_read(SANREMO_INT_STATUS_2, &status2);
		sanremo_read(SANREMO_INT_STATUS_3, &status3);

		levante_check_trim();

	}

	printk(LVT_DEBUG_MUST "levante_probe - done\n");

    return 0;
}

static int levante_remove(struct i2c_client *client)
{
#ifdef	CONFIG_PROC_FS
	remove_levante_proc_file();
#endif

	return 0;
}

static const struct i2c_device_id levante_id[] = {
	{ "levante", 0 },
	{ }
};

static struct i2c_driver levante_driver = {
	.driver = {
		.name	= "levante",
	},
	.probe		= levante_probe,
	.remove		= levante_remove,
	.id_table	= levante_id,
	.suspend	= levante_suspend,
	.resume		= levante_resume,
};

static int __init levante_init(void)
{
   int retVal = -EINVAL;
   pmicId = pmic_id();
   if ((pmicId >= LEVATNE_A0_ID) && (pmicId <=  LEVATNE_END_ID)){
	   retVal=i2c_add_driver(&levante_driver);
   }
   return (retVal);
}

static void __exit levante_exit(void)
{

	i2c_del_driver(&levante_driver);
}


subsys_initcall(levante_init);
module_exit(levante_exit);

MODULE_AUTHOR("Yael,Shemla <yshemla@marvell.com > Yossef,Hanin <yhanin@marvell.com>");
MODULE_DESCRIPTION("Levante Driver");
MODULE_LICENSE("GPL");
