/*
 *  drivers/i2c/chips/lis331dl_sensor.c
 *
 * Copyright (C) 2008 Borqs, Inc.
 * 
 * This software is licensed under the terms of the GNU General Public
 * License version 2, as published by the Free Software Foundation, and
 * may be copied, distributed, and modified under those terms.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
*/

#include <linux/init.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/sysctl.h>
#include <linux/platform_device.h>
#include <linux/i2c.h>
#include <linux/cdev.h>
#include <linux/major.h>
#include <linux/interrupt.h>
#include <linux/poll.h>
#include <asm/ioctl.h>
#include <asm/uaccess.h>
#include <mach/pxa-regs.h>
#include <mach/pxa3xx-regs.h>
#include <mach/lis331.h>
#include <mach/mfp.h>
#include <mach/gpio.h>
#include <linux/input.h>
#include <linux/sensors.h>
#include <linux/earlysuspend.h>

static int interval = 100;
static int int_mod = 0; //0:real-time 1:interupt
static int xyz_direction, x_dir, y_dir, z_dir, switch_xy;
static lis331dl_device_t lis331dl_device = {
	.name 			= "lis331dl_device",
	.rf_count		= ATOMIC_INIT(0),
};

static void lis331dl_power_on(void);
static void lis331dl_power_off(void);

static int lis331_check_id(void)
{
	u8 chip_id;
	chip_id = 0;

	printk(KERN_INFO "lis331dl check chip id\n");

	/* WHO_AM_I */
	chip_id = i2c_smbus_read_byte_data(lis331dl_device.lis331_client, WHO_AM_I);
	
	if (chip_id != LIS331_LIS302DL) {
		printk(KERN_ERR "LIS331DLE unavailable!\n");
		return -ENXIO;
	} else {
		printk(KERN_INFO "LIS331DLE(chip id:0x%02x) detected.\n",
		       chip_id);
		return 0;
	}
}

static int lis331_read(u8 reg, u8 *pval)
{
	int ret;
	
	if (lis331dl_device.lis331_client == NULL)
		return -EINVAL;

	ret = i2c_smbus_read_byte_data(lis331dl_device.lis331_client, reg);

	if (ret >= 0) {
		*pval = ret;
		ret = 0;
	} else
		ret = -EIO;

	return ret;
}

static int lis331_read_block(u8 reg, u8 *pval, u8 len)
{
	int ret;
	
	if (lis331dl_device.lis331_client == NULL)
		return -EINVAL;

	ret = i2c_smbus_read_i2c_block_data(lis331dl_device.lis331_client, reg | MSB_ENABLE, len, pval);

	if (ret != len)
		ret = -EIO;

	return ret;
}

static int lis331_write(u8 reg, u8 val)
{
	int ret;

	if (lis331dl_device.lis331_client == NULL)
		return -1;

	ret = i2c_smbus_write_byte_data(lis331dl_device.lis331_client, reg, val);
	if (ret == 0)
		ret = 0;
	else
		ret = -EIO;

	return ret;
}

static void lis331_init(void)
{
	u8 ctrl_reg3, ff_wu_cfg, ff_wu_ths, ff_wu_duration;
	/*Add code for single/double click */
	u8 click_cfg,click_ths_yx,click_ths_z;
	//u8 click_timelimit,click_latency,click_window;
#ifdef CONFIG_LIS331_400HZ 
	u8 ctrl_reg1;
	//ctrl_reg1:1100 0111 The selected date rate will be 400hz
	lis331_read(Ctrl_Reg1,&ctrl_reg1);
	ctrl_reg1 |=DATE_RATE_400HZ;
	lis331_write(Ctrl_Reg1,ctrl_reg1);
#endif	
	// ctrl_reg3
	lis331_read(Ctrl_Reg3, &ctrl_reg3);
	ctrl_reg3 = (LIS331DL_FF_WU_1);	// FF_WU 
	//ctrl_reg3 |= ( 1 << 7 ); // FF_WU
	lis331_write(Ctrl_Reg3, ctrl_reg3);

	// FF_WU_CFG
	//lis331_read( FF_WU_CFG, & ff_wu_cfg );
	ff_wu_cfg = (u8) XYZ_LIE;
	// OR combination, LIR
	lis331_write(FF_WU_CFG, ff_wu_cfg);

	ff_wu_ths = (u8) (THRESH_OLD | THRESH_DCRM);
	lis331_write(FF_WU_THS, ff_wu_ths);

	ff_wu_duration = (u8) DURATION;
	lis331_write(FF_WU_DURATION, ff_wu_duration);
	
	click_cfg = (u8) LIS331DL_SINGLE_CLICK;
	lis331_write(CLICK_CFG, click_cfg);

	click_ths_yx = (u8) LIS331DL_CLICK_THS_YX;
	lis331_write(CLICK_THS_Y_X,click_ths_yx);

	click_ths_z = (u8) LIS331DL_CLICK_THS_Z;
	lis331_write(CLICK_THS_Z, click_ths_z);
		
}


static void lis331dl_wake_up_int(unsigned long data)
{
	lis331_status_t *status_t;

	spin_lock(&lis331dl_device.list_lock);
	list_for_each_entry(status_t, &lis331dl_device.open_list, list) {
		atomic_set(&status_t->g_rdy, 1);
	}
	spin_unlock(&lis331dl_device.list_lock);

	wake_up_interruptible(&lis331dl_device.lis331_wait);
	if(!int_mod && (lis331dl_device.switch_on == SWITCH_ON))
		mod_timer(&lis331dl_device.timer, jiffies + msecs_to_jiffies(interval));
}

static void lis331_int_routine(struct work_struct *work)
{
	u8 interrupt_type;
	struct i2c_client *p_client = lis331dl_device.lis331_client;
	
	struct lis331_platform_data *p_data = p_client->dev.platform_data;
	
	if (!p_data) {
		printk(KERN_INFO "Fail to get the platform data\n.");
	} 
	lis331_read(Ctrl_Reg3, &interrupt_type);
	interrupt_type &= INTERRUPT_TYPE_MASK;
	
	if (interrupt_type == INTERRUPT_FF_WU) {
		printk(KERN_DEBUG "GSENSOR WAKEUP_FREEFALL INTERRUPT GENERATED!\n");
	} else if (interrupt_type == INTERRUPT_DATA_READY) {
		printk(KERN_DEBUG "GSENSOR DATA_READY  INTERRUPT GENERATED!\n");
	} else if (interrupt_type == INTERRUPT_CLICK) { 
		printk(KERN_DEBUG "GSENSOR CLICK INTERRUPT GENERATED!\n");
	} else {
		printk(KERN_DEBUG "Unknown interrupt generated!\n");
	}
}

static irqreturn_t lis331_isr(int irq, void *dev_id)
{
	if (interval == LIS331_DEBUG_FLOOD) 
		schedule_work(&lis331dl_device.lis331_work);
	if(int_mod)
		mod_timer(&lis331dl_device.timer, jiffies + msecs_to_jiffies(LIS331DL_MIN_INTERVAL));
	return IRQ_HANDLED;
}

static int lis331dle_dev_open(struct inode *inode, struct file *file)
{
	int ret;
	lis331_status_t *status_p;

	printk(KERN_INFO "LIS331DLE open\n");
	
	if (down_interruptible(&lis331dl_device.mutex))
		return -ERESTARTSYS;

	if ((file->f_flags & O_ACCMODE) == O_WRONLY) {
		printk(KERN_INFO "LIS331DLE's device node is writeonly\n");
		ret = -EPERM;
	} else {
		printk(KERN_INFO "LIS331DLE's device :You are opening !\n");
		ret = 0;
	}
	
	status_p = kzalloc(sizeof(lis331_status_t), GFP_KERNEL);
	if (status_p == NULL) { 
		printk(KERN_INFO "No mem!\n");
		up(&lis331dl_device.mutex);
		return -ENOMEM;
	}
	file->private_data = status_p;
	atomic_set(&status_p->g_rdy, 1);
	spin_lock(&lis331dl_device.list_lock);
	list_add(&status_p->list, &lis331dl_device.open_list);
	spin_unlock(&lis331dl_device.list_lock);

	atomic_inc(&lis331dl_device.rf_count);
	printk(KERN_INFO "rf_count = %d\n", atomic_read(&lis331dl_device.rf_count));
	up(&lis331dl_device.mutex);
	if(!int_mod)
		mod_timer(&lis331dl_device.timer, jiffies + msecs_to_jiffies(10*interval));

	return ret;
}

static int ox, oy, oz;
static ssize_t lis331dle_dev_read(struct file *file, char __user * buffer,
				  size_t size, loff_t * f_pos)
{
	struct input_event event[3];
	int coorduser_x, coorduser_y;
	int x, y, z;
	int ret;
	lis331_status_t *status_p;
	signed char xyz[5];

	if (down_interruptible(&lis331dl_device.mutex))
		return -ERESTARTSYS;

	status_p = file->private_data;
	if( status_p == NULL) {
		up(&lis331dl_device.mutex);
		return -EFAULT;
	}
	
	ret = lis331_read_block(OutX, xyz, 5);

	if (ret < 0) {
		printk(KERN_INFO "LIS331DLE read Error!\n");
		goto read_out;
	}

	x = xyz[0]; 
	y = xyz[2];
	z = xyz[4];

	/*----------------------LOW PASS FILTER*/
	x = (xyz[0] * 205 + ox * 51) / 256;
	y = (xyz[2] * 205 + oy * 51) / 256;
	z = (xyz[4] * 205 + oz * 51) / 256;
	ox = x;
	oy = y;
	oz = z;

	if (switch_xy) {
		coorduser_x = x;
		coorduser_y = y;
	} else {
		coorduser_x = y;
		coorduser_y = x;
	}

	get_monotonic_time(&event[0].time);
	event[1].time = event[0].time;
	event[2].time = event[0].time;

	event[0].type = EV_ABS;
	event[1].type = EV_ABS;
	event[2].type = EV_ABS;

	event[0].value = (int)((signed char)coorduser_x * x_dir * MG_PER_SAMPLE);
	event[0].code = ABS_X;

	event[1].value = (int)((signed char)coorduser_y * y_dir * MG_PER_SAMPLE);
	event[1].code = ABS_Y;

	event[2].value = (int)((signed char)z * z_dir * MG_PER_SAMPLE);
	event[2].code = ABS_Z;

	if (copy_to_user(buffer, event, sizeof(event))) {
		printk(KERN_INFO "Copy Gsensor data Fail!\n");
		ret = -EIO;	
	} else
		ret = sizeof(event);
read_out:
	atomic_set(&status_p->g_rdy, 0);
	up(&lis331dl_device.mutex);
	return ret;
}

static unsigned int lis331dle_dev_poll(struct file *filep, poll_table * pt)
{

	unsigned int mask = 0;
	lis331_status_t *status_p;
	
	if (down_interruptible(&lis331dl_device.mutex))
		return -ERESTARTSYS;

	status_p = filep->private_data;
	if (status_p == NULL) {
		up(&lis331dl_device.mutex);
		return -EFAULT;
	}

	poll_wait(filep, &lis331dl_device.lis331_wait, pt);

	if (atomic_read(&status_p->g_rdy)) 
		mask |= POLLIN | POLLRDNORM;

	up(&lis331dl_device.mutex);
	return mask;
}

static int lis331dle_dev_release(struct inode *inode, struct file *filp)
{
	lis331_status_t *status_p;

	if (down_interruptible(&lis331dl_device.mutex))
		return -ERESTARTSYS;

	status_p = filp->private_data;
	if (status_p == NULL) {
		up(&lis331dl_device.mutex);
		return -EFAULT;
	}
	filp->private_data = NULL;
	
	spin_lock(&lis331dl_device.list_lock);
	list_del(&status_p->list);
	spin_unlock(&lis331dl_device.list_lock);

	kfree(status_p);

	atomic_dec(&lis331dl_device.rf_count);

	up(&lis331dl_device.mutex);
	
	return 0;
}

static struct file_operations lis331dle_dev_fops = {
	.owner = THIS_MODULE,
	.open = lis331dle_dev_open,
	.read = lis331dle_dev_read,
	.poll = lis331dle_dev_poll,
	.release = lis331dle_dev_release,
};

static int lis331dl_get_property(struct sensors_dev *sdev,
				 enum sensors_property property,
				 union sensors_propval *val)
{
	switch (property) {
	case SENSORS_PROP_THRESHOLD:
		val->intval = 0;
		break;

	case SENSORS_PROP_INTERVAL:
		val->intval = interval;
		break;

	case SENSORS_PROP_MAXRANGE:
		val->intval = 128;	
		break;

	case SENSORS_PROP_RESOLUTION:
		val->intval = 1;
		break;

	case SENSORS_PROP_VERSION:
		val->intval = 1;	
		break;

	case SENSORS_PROP_CURRENT:
		val->intval = 30;	
		break;

	case SENSORS_PROP_SWITCH:
		val->intval = lis331dl_device.switch_on ? 1 : 0;
		break;

	case SENSORS_PROP_VENDOR:
		sprintf(val->strval, "ST");
		break;

	default:
		return -EINVAL;
	}

	return 0;
}


static int lis331dl_put_property(struct sensors_dev *sdev,
				 enum sensors_property property,
				 union sensors_propval *val)
{
	switch (property) {
	case SENSORS_PROP_SWITCH:
		if (val->intval && lis331dl_device.switch_on) {
			lis331dl_device.switch_on = SWITCH_ON;
			lis331dl_power_on();
		} else if (!val->intval && !lis331dl_device.switch_on) {
			lis331dl_device.switch_on = SWITCH_OFF;
			lis331dl_power_off();
		}
		break;

	case SENSORS_PROP_THRESHOLD:
		break;

	case SENSORS_PROP_INTERVAL:
		interval = val->intval;

		if (interval > 2000) {
			interval = LIS331DL_MAX_INTERVAL;
			int_mod = 1;
			printk(KERN_DEBUG "INT MODE NOW!\n");
		}

		if (interval < LIS331DL_MIN_INTERVAL) {
			interval = LIS331DL_MIN_INTERVAL;
			int_mod = 0;
			printk(KERN_DEBUG "REAL-TIME MODE NOW!\n");
		}
	
		break;
	default:
		return -EINVAL;
	}

	return 0; 
}




static enum sensors_property lis331dl_properties[] = {
	/* int */
	SENSORS_PROP_INTERVAL,
	SENSORS_PROP_THRESHOLD,
	SENSORS_PROP_MAXRANGE,		/* read only */
	SENSORS_PROP_RESOLUTION,	/* read only */
	SENSORS_PROP_VERSION,		/* read only */
	SENSORS_PROP_CURRENT,		/* read only */
	/* char */
	SENSORS_PROP_SWITCH,
	SENSORS_PROP_VENDOR,		/* read only */
};

struct sensors_dev lis331dl = {
	.id = SENSORS_ACCELERATION,
	.name = "gsensor",
	.num_properties = ARRAY_SIZE(lis331dl_properties),
	.properties = lis331dl_properties,
	.get_property = lis331dl_get_property,
	.put_property = lis331dl_put_property,
	.fops = &lis331dle_dev_fops
};


#ifdef CONFIG_HAS_EARLYSUSPEND
static void lis331_early_suspend(struct early_suspend *h)
{
	printk(KERN_DEBUG "Lis331 Suspend!\n");
	lis331dl_power_off();
	lis331dl_device.switch_on = SWITCH_OFF;
}

static void lis331_early_resume(struct early_suspend *h)
{
	printk(KERN_DEBUG "Lis331 Resume!\n");
	lis331dl_power_on();
	lis331dl_device.switch_on = SWITCH_ON;
	if(atomic_read(&lis331dl_device.rf_count) && !int_mod)
		mod_timer(&lis331dl_device.timer, jiffies + msecs_to_jiffies(interval));
}

static struct early_suspend lis331_early_suspend_desc = {
	.level  = 1,
	.suspend = lis331_early_suspend,
	.resume = lis331_early_resume,
};

#endif

static int lis331dl_probe(struct i2c_client *client, const struct i2c_device_id *id)
{
	struct lis331_platform_data *pdata;
	int ret;
	u8 ctrl_reg1;

	printk(KERN_INFO "LIS331DLE probe\n");
	if (client == NULL) {
		printk(KERN_ERR "Client is NUll!\n");
		return -EFAULT;
	}
	lis331dl_device.lis331_client = client;
	ret = lis331_check_id();
	if (ret != 0) {
		printk(KERN_ERR "Fail to check lis331.\n");
		return ret;
	}
	
	INIT_WORK(&lis331dl_device.lis331_work, lis331_int_routine);
	init_waitqueue_head(&lis331dl_device.lis331_wait);
	INIT_LIST_HEAD(&lis331dl_device.open_list);
	init_MUTEX(&lis331dl_device.mutex);
	spin_lock_init(&lis331dl_device.list_lock);
	lis331dl_device.switch_on = SWITCH_ON;
	setup_timer(&lis331dl_device.timer, lis331dl_wake_up_int, 0);

	lis331dl_device.device = &lis331dl;
	
	lis331_read(Ctrl_Reg1, &ctrl_reg1);
	ctrl_reg1 |= LIS331_POWER_UP;
	lis331_write(Ctrl_Reg1, ctrl_reg1);

	ret = sensors_dev_register(&lis331dl);
	if (ret) {
		printk(KERN_ERR "lis331dl create sysfile fail!\n");
	}

	pdata = client->dev.platform_data;
	xyz_direction = pdata->xyz_dir & XYZ_DIR_MASK;
	x_dir = ((xyz_direction & X_POS)? 1:-1); 
	y_dir = ((xyz_direction & Y_POS)? 1:-1); 
	z_dir = ((xyz_direction & Z_POS)? 1:-1); 
	switch_xy = ((xyz_direction & SWITCH_XY)? 1:0);

	pdata->init_irq();

	ret = request_irq(client->irq, lis331_isr, IRQF_TRIGGER_RISING, "LIS331DLE", client);	
	printk(KERN_INFO "LIS331DLE: request_irq = %d\n", ret);

	lis331_init();

#ifdef CONFIG_HAS_EARLYSUSPEND
	register_early_suspend(&lis331_early_suspend_desc);
#endif

	return 0;
}

static int lis331dl_remove(struct i2c_client *client)
{
	free_irq(client->irq, NULL);
#ifdef CONFIG_HAS_EARLYSUSPEND
	unregister_early_suspend(&lis331_early_suspend_desc);
#endif
	return 0;
}

static void lis331dl_power_on(void)
{
	u8 ctrl_reg1;
	lis331_read(Ctrl_Reg1, &ctrl_reg1);
	ctrl_reg1 |= LIS331_POWER_UP;
	lis331_write(Ctrl_Reg1, ctrl_reg1);
	return ;
}

static void lis331dl_power_off(void)
{
	u8 ctrl_reg1;
	lis331_read(Ctrl_Reg1, &ctrl_reg1);
	ctrl_reg1 &= LIS331_POWER_OFF;
	lis331_write(Ctrl_Reg1, ctrl_reg1);
	return ;
}

#ifdef CONFIG_PM

static int lis331dl_suspend(struct i2c_client *client, pm_message_t mesg)
{
	/*Make sure the device is power off when suspend*/
	lis331dl_power_off();
	return 0 ;
}

static int lis331dl_resume(struct i2c_client *client)
{
	/*Wake up the device*/
	lis331dl_power_on();
	return 0 ;
}

#else

#define lis331dl_suspend 	NULL
#define lis331dl_resume 	NULL

#endif

static const struct i2c_device_id lis331_id[] = {
	{ "lis331", 0 },
	{ }
};

static struct i2c_driver lis331_driver = {
	.driver = {
		.name = "lis331",
	 },
	.probe 	= lis331dl_probe,
	.remove = lis331dl_remove,
	.suspend = lis331dl_suspend,
	.resume = lis331dl_resume,
	.id_table = lis331_id,
};

static int __init lis331dl_init(void)
{
	int ret;

	ret = i2c_add_driver(&lis331_driver);
	if (ret) 
		printk(KERN_WARNING "LIS331DLE: Driver registration failed, module not inserted.\n");

	lis331dl_device.driver = &lis331_driver;
	
	return ret;
}

static void __exit lis331dl_exit(void)
{
	i2c_del_driver(&lis331_driver);
	sensors_dev_unregister(&lis331dl);
}

module_init(lis331dl_init);
module_exit(lis331dl_exit);

MODULE_DESCRIPTION("Sensors LIS331DL");
MODULE_LICENSE("GPL");
