/*
 * Tavor Portofino PMIC Management Routines
 *
 *  linux/i2c/drivers/chips/portofino.c
 *
 *  driver for pxa935/950 - portofino.c
 *
 *  Copyright (C) 2006, Marvell Corporation
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2 as
 *  published by the Free Software Foundation.
 */


#include <linux/init.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/interrupt.h>
#include <linux/i2c.h>
#include <linux/proc_fs.h>

#include <asm/uaccess.h>
#include <mach/pxa3xx_pmic.h>
#include <mach/levante_hw.h>
#include <mach/levante.h>
#include <mach/tavor_common.h>


#define PORTOFINO_REG_NUM		0x1C

static int porto_print = 0;
#define PORTO_PRINT_WRITE(x,y)     if(porto_print) printk("======>PORTO_PRINT_WRITE: 0x%x<-0x%x \n",y,x);
#define PORTO_PRINT_READ(x,y)      if(porto_print) printk("======> PORTO_PRINT_READ: 0x%x->0x%x \n",x,y);

static struct pxa3xx_pmic_regs portofino_regs[PORTOFINO_REG_NUM];

/* Unique ID allocation */
static struct i2c_client *g_client;

int portofino_read(u8 reg, u8 *pval)
{
	int ret;
	int status;

	if (g_client == NULL)	/* No global client pointer? */
		return -1;
	ret = i2c_smbus_read_byte_data(g_client, reg);
	if (ret >= 0) {
		PORTO_PRINT_READ(reg,ret);
		*pval = ret;
        portofino_regs[reg].hit = ~portofino_regs[reg].mask;
        portofino_regs[reg].data = ret;
		status = 0;
	} else {
		status = -EIO;
		printk(LVT_DEBUG_MUST "\nportofino_read fail\n");
	}


	return status;
}

int portofino_write(u8 reg, u8 val)
{
	int ret;
	int status;

	if (g_client == NULL)	/* No global client pointer? */
		return -1;

	ret = i2c_smbus_write_byte_data(g_client, reg, val);

	PORTO_PRINT_WRITE(val,reg);

	if (ret == 0) {
        portofino_regs[reg].hit = ~portofino_regs[reg].mask;
        portofino_regs[reg].data = val;
		status = 0;
    }
	else{
		status = -EIO;
		printk(LVT_DEBUG_MUST "\nportofino_write fail\n");
	}

	return status;
}

static int portofino_initchip(void)
{
	int i;

    memset(&portofino_regs, 0,
        (sizeof(struct pxa3xx_pmic_regs) * PORTOFINO_REG_NUM));
    for (i = 0; i < PORTOFINO_REG_NUM; i++)
        portofino_regs[i].mask = 1;
	return 0;
}

#ifdef	CONFIG_PM
static int portofino_suspend(struct i2c_client *client, pm_message_t state)
{

    disable_irq(client->irq);
 /*   if ((get_pm_state() == PM_SUSPEND_MEM)) {
        portofino_read(MICCO_OVER1, &val);

    }  */
	return 0;
}

static int portofino_resume(struct i2c_client *client)
{
    int i;

    for (i = 0; i < PORTOFINO_REG_NUM; i++)
        portofino_regs[i].hit = 0;

    enable_irq(client->irq);
/*    if ((get_pm_state() == PM_SUSPEND_MEM)) {
        portofino_write(MICCO_OVER1, portofino_regs[MICCO_OVER1].data);
    }  */
	return 0;
}
#else
#define	portofino_suspend		NULL
#define	portofino_resume		NULL
#endif

#ifdef	CONFIG_PROC_FS
#define	PORTOFINO_PROC_FILE	"driver/portofino"
static struct proc_dir_entry *portofino_proc_file;
static int index = 0x00;

static ssize_t portofino_proc_read(struct file *filp,
        char __user *buffer, size_t length, loff_t *offset)
{
	u8 reg_val;

	if ((index < 0) || (index > PORTOFINO_REG_NUM))
		return 0;

	portofino_read(index, &reg_val);
	printk(KERN_ALERT "register 0x%x: 0x%x\n", index, reg_val);
	return 0;
}

static ssize_t portofino_proc_write(struct file *filp,
		const char *buff, size_t len, loff_t *off)
{
	u8 reg_val;
	char messages[256], vol[256];

	if (len > 256)
		len = 256;

	if (copy_from_user(messages, buff, len))
		return -EFAULT;

	if ('-' == messages[0]) {
		/* set the register index */
		memcpy(vol, messages+1, len-1);
		index = (int) simple_strtoul(vol, NULL, 16);
    }else if ('?' == messages[0]) {
		printk(LVT_DEBUG_MUST "\n write: echo -0x10 > /proc/driver/portofino \n");
		printk(LVT_DEBUG_MUST " read(reg from write): cat /proc/driver/portofino \n");
		printk(LVT_DEBUG_MUST " read/write trace on: echo traceon > /proc/driver/portofino \n");
	}else if (strncmp(messages, "traceon", 7) == 0){
		porto_print = 1;
		printk(LVT_DEBUG_MUST "\nportofino_proc_write:enable portofino read/write traces \n");
	} else {
		/* set the register value */
		reg_val = (int)simple_strtoul(messages, NULL, 16);
		portofino_write(index, reg_val & 0xFF);
	}

	return len;
}

static struct file_operations portofino_proc_ops = {
	.read = portofino_proc_read,
	.write = portofino_proc_write,
};

static void create_portofino_proc_file(void)
{
	portofino_proc_file = create_proc_entry(PORTOFINO_PROC_FILE, 0644, NULL);
	if (portofino_proc_file) {
		portofino_proc_file->owner = THIS_MODULE;
		portofino_proc_file->proc_fops = &portofino_proc_ops;
	} else
		printk(KERN_ALERT "proc file create failed!\n");
}

static void remove_portofino_proc_file(void)
{
	remove_proc_entry(PORTOFINO_PROC_FILE, NULL);
}

#endif
static int __devinit portofino_probe(struct i2c_client *client,
			 const struct i2c_device_id *id)
{
	struct portofino_platform_data *pdata;
	int ret;
	int chip_id;

	printk(LVT_DEBUG_MUST "--->>>portofino_probe!!!\n");

	chip_id = i2c_smbus_read_byte_data(client, PORTOFINO_DIE_REVISION_ID);
	if(chip_id <0 ){
		printk(LVT_DEBUG_MUST "I2c fail portofino unavailable!\n");
		g_client = NULL;
		return -ENXIO;
    } else if (chip_id != PORTOFINO_A0_ID &&
				chip_id != PORTOFINO_A1_ID &&
				chip_id != PORTOFINO_END_ID) {
		printk(LVT_DEBUG_MUST "Id[0x%02x] is not portofino; --->>> portofino unavailable!!!\n",chip_id);
		g_client = NULL;
		return -ENXIO;
	} else {
		printk(KERN_ALERT "portofino(die & revision id:0x%02x) detected.\n", chip_id);
	}

	g_client = client;

	ret = portofino_initchip();
	if (ret != 0)
		printk(KERN_ALERT "Initialize Portofino failed with ret 0x%x\n",ret);

	pdata = client->dev.platform_data;
	/* init spinlock */
	spin_lock_init(&pdata->lock);

    //pdata->platform_init();

#ifdef	CONFIG_PROC_FS
	create_portofino_proc_file();
#endif
	return 0;
}

static int portofino_remove(struct i2c_client *client)
{
#ifdef	CONFIG_PROC_FS
	remove_portofino_proc_file();
#endif
	return 0;
}

static const struct i2c_device_id portofino_id[] = {
	{ "portofino", 0 },
	{ }
};

static struct i2c_driver portofino_driver = {
	.driver = {
		.name	= "portofino",
	},
	.probe		= portofino_probe,
	.remove		= portofino_remove,
	.id_table	= portofino_id,
	.suspend	= portofino_suspend,
	.resume		= portofino_resume,
};

static int __init portofino_init(void)
{
   int retVal = -EINVAL;
   if  ((pmic_id() >= LEVATNE_A0_ID) && (pmic_id() <=  LEVATNE_END_ID) ){
	   retVal=i2c_add_driver(&portofino_driver);
   }
   return (retVal);
}

static void __exit portofino_exit(void)
{
	i2c_del_driver(&portofino_driver);
}

subsys_initcall(portofino_init);
module_exit(portofino_exit);

MODULE_DESCRIPTION("Portofino Driver");
MODULE_LICENSE("GPL");
