/*
 * Tavor SanRemo PMIC Management Routines
 *
 *  linux/i2c/drivers/chips/sanremo.c
 *
 *  driver for pxa935 - sanremo.c
 *
 *  Copyright (C) 2006, Marvell Corporation
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2 as
 *  published by the Free Software Foundation.
 */

#include <linux/init.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/interrupt.h>
#include <linux/workqueue.h>
#include <linux/i2c.h>
#include <linux/proc_fs.h>
#include <linux/mutex.h>

#include <asm/irq.h>
#include <mach/gpio.h>
#include <asm/uaccess.h>
#include <mach/pxa3xx_pmic.h>
#include <mach/levante_hw.h>
#include <mach/levante.h>

#include <mach/hardware.h>
#include <mach/pxa-regs.h>
#include <mach/tavor_common.h>


#define SANREMO_REG_NUM		256 /*0..0xff*/

static int sanremo_print = 0;
#define SANREMO_PRINT_WRITE(x,y)     if(sanremo_print) printk("======>SANREMO_PRINT_WRITE: 0x0%x<-0x0%x \n",y,x)
#define SANREMO_PRINT_READ(x,y)      if(sanremo_print) printk("======> SANREMO_PRINT_READ: 0x0%x->0x0%x \n",x,y)

static int pmicId = 0;
struct sanremo_read_cache {
	u8  data;
	u8  hit;
	u8  cachable;
	u8  spare;
};

static struct sanremo_read_cache   sanremo_regs[SANREMO_REG_NUM];
static struct power_supply_module *sanremo_power_module;

/* Unique ID allocation */
static struct i2c_client *g_client;


/************** READ & Read BLOCK *****************************************/
int sanremo_read(u8 reg, u8 *pval)
{
	int ret;
	int status;

	if (g_client == NULL)	/* No global client pointer? */
		return -1;

	/* Cache read of the Sanremo register */
	if (sanremo_regs[reg].hit) {
		*pval = sanremo_regs[reg].data;
		return 0;
	}

	ret = i2c_smbus_read_byte_data(g_client, reg);
	if (ret >= 0) {
	 	SANREMO_PRINT_READ(reg,ret);
		*pval = ret;
		sanremo_regs[reg].data = ret;
		sanremo_regs[reg].hit = sanremo_regs[reg].cachable;
		status = 0;
	} else{
		status = -EIO;
	}
	return status;
}
EXPORT_SYMBOL(sanremo_read);

int sanremo_read_block(u8 reg, u8 *pval, u8 length)
{
    int ret;
	u8  i;

    if (g_client == NULL)   /* No global client pointer? */
        return -1;

	/* The block-read is NON cachable */
    ret = i2c_smbus_read_i2c_block_data(g_client, reg, length, pval);

	for ( i=0; i<length; i++)
	 SANREMO_PRINT_READ(reg+i,pval[i]);

	ret = (ret < 0) ? -EIO : 0;
	return ret;
}


/************** WRITE & Write BLOCK *****************************************/
int sanremo_write(u8 reg, u8 val)
{
	int ret;
	int status;

	if (g_client == NULL)	/* No global client pointer? */
		return -1;


	sanremo_regs[reg].hit = 0;

	ret = i2c_smbus_write_byte_data(g_client, reg, val);

	SANREMO_PRINT_WRITE(val,reg);

	if (ret == 0) {
		sanremo_regs[reg].data = val;
		sanremo_regs[reg].hit = sanremo_regs[reg].cachable;
		status = 0;
	} else {
		if( reg == 0xFF || reg == 0xFB || reg == 0xFA ){/*write to those register will not return ACK*/
	  		status = 0;
		} else {
			status = -EIO;
		}
	}
	return status;
}

int sanremo_write_block(u8 reg,u8 length, const u8 *pval)
{
	int ret;
	int status = 0;
	u8  i;

	if (g_client == NULL)	/* No global client pointer? */
		return -ENODEV;

	for ( i=0; i<length; i++)
		sanremo_regs[reg+i].hit = 0;
	ret = i2c_smbus_write_block_data(g_client, reg,length, pval);
	if (ret == 0) {
		for ( i=0; i<length; i++)
		{
			SANREMO_PRINT_WRITE(pval[i],reg+i);
			sanremo_regs[reg+i].data = pval[i];
			sanremo_regs[reg+i].hit  = sanremo_regs[reg+i].cachable;
		}
		status = 0;
	} else {
		if( reg == 0xFF || reg == 0xFB || reg == 0xFA ){/*write to those register will not return ACK*/
		 	status = 0;
	    }else{
			status = -EIO;
		}
	}
	return status;
}

/************** Lock/Unlock for READ/Modify/WRITE ***********/
void sanremo_rw_lock(void)
{
	struct sanremo_platform_data *pdata = g_client->dev.platform_data;
	if (g_client == NULL)	return;
	mutex_lock(&pdata->m_lock);
}

void sanremo_rw_unlock(void)
{
	struct sanremo_platform_data *pdata = g_client->dev.platform_data;
	mutex_unlock(&pdata->m_lock);
}


static int sanremo_initchip(void)
{
	memset(&sanremo_regs, 0,
		(sizeof(struct sanremo_read_cache) * SANREMO_REG_NUM));

	/* Init Read-cache	 */
	/* The behavior is not always trivial */
	/* Let's add registers "one-by-one"   */
	sanremo_regs[ SANREMO_CHIP_ID  ].cachable = 1;
	sanremo_regs[ SANREMO_INT_EN_1 ].cachable = 1;
	sanremo_regs[ SANREMO_INT_EN_2 ].cachable = 1;
	sanremo_regs[ SANREMO_INT_EN_3 ].cachable = 1;

	sanremo_regs[ SANREMO_CHG_CTRL1 ].cachable = 1;
	sanremo_regs[ SANREMO_CHG_CTRL2 ].cachable = 1;
	sanremo_regs[ SANREMO_CHG_CTRL3 ].cachable = 1;
	sanremo_regs[ SANREMO_CHG_CTRL4 ].cachable = 1;
	sanremo_regs[ SANREMO_CHG_CTRL5 ].cachable = 1;
	sanremo_regs[ SANREMO_CHG_CTRL6 ].cachable = 1;
	sanremo_regs[ SANREMO_CHG_CTRL7 ].cachable = 1;

	sanremo_regs[ /*MEAS GP Bias*/ 0x4F ].cachable = 1;
	sanremo_regs[ SANREMO_MEAS_ENABLE1  ].cachable = 1;
	sanremo_regs[ SANREMO_MEAS_ENABLE2  ].cachable = 1;
	sanremo_regs[ SANREMO_MEAS_ENABLE3  ].cachable = 1;
	sanremo_regs[ SANREMO_MEAS_OFF_TIME1].cachable = 1;
	sanremo_regs[ SANREMO_MEAS_OFF_TIME2].cachable = 1;

	sanremo_regs[ SANREMO_TSI_PREBIAS_TIME     ].cachable = 1;
	sanremo_regs[ SANREMO_PD_PREBIAS_TIME      ].cachable = 1;
	sanremo_regs[ SANREMO_GPADC_MISC1 ].cachable = 1;
	sanremo_regs[ SANREMO_GPADC_MISC2 ].cachable = 1;

	printk(LVT_DEBUG_MUST"------->LEVANTE: sanremo_initchip - disable WDT is done in the OBM!!!\n");
	return 0;
}


static void sanremo_event_clear(u8* status /*,u8 num_regs*/)
{
   	/* clear interrupt but only for active(s) */
	if (status[0] && status[1] && status[2])
		sanremo_write_block(SANREMO_INT_STATUS_1, 3, &status[0]);
	else
	if (status[0] && status[1])
		sanremo_write_block(SANREMO_INT_STATUS_1, 2, &status[0]);
	else
	if (status[1] && status[2])
		sanremo_write_block(SANREMO_INT_STATUS_2, 2, &status[1]);
	else
	if (status[0])
		sanremo_write(SANREMO_INT_STATUS_1, status[0]);
	else
	if (status[1])
		sanremo_write(SANREMO_INT_STATUS_2, status[1]);
	else
	if (status[2])
		sanremo_write(SANREMO_INT_STATUS_3, status[2]);
}


static unsigned int sanremo_event_change(u8* status)
{
	unsigned int ret = 0;

	sanremo_read_block(SANREMO_INT_STATUS_1, status, 3);

	if (status[0] & SANREMO_INT_STS1_ONKEY)
	{
		printk(LVT_DEBUG_MUST "------->sanremo_event_change::SANREMO_INT_STS1_ONKEY\n");
		ret |= PMIC_EVENT_ONKEY;
	}

    if (status[0] & SANREMO_INT_STS1_EXTON)
	{
		printk(LVT_DEBUG_MUST "------->sanremo_event_change::SANREMO_INT_STS1_EXTON \n");
		ret |= PMIC_EVENT_EXTON;
	}

	if (status[0] & SANREMO_INT_STS1_CHG)
	{
		printk(LVT_DEBUG_MUST "------->sanremo_event_change::SANREMO_INT_STS1_CHG (Pmic:CHDET+VBUS)\n");
		ret |= (PMIC_EVENT_CHDET | PMIC_EVENT_VBUS);
	}

	if (status[0] & SANREMO_INT_STS1_BAT)
	{
		printk(LVT_DEBUG_MUST "------->sanremo_event_change::SANREMO_INT_STS1_BAT \n");
		ret |= PMIC_EVENT_BAT_DET;
	}

/*	if (status[0] & SANREMO_INT_STS1_RTC)
 		ret |= ;  */

	if (status[1] & SANREMO_INT_STS2_VBAT)
	{
		printk(LVT_DEBUG_MUST "------->sanremo_event_change::SANREMO_INT_STS2_VBAT \n");
		ret |= PMIC_EVENT_VBAT_TS;
	}

	if (status[1] & SANREMO_INT_STS2_VCHG)
	{
		printk(LVT_DEBUG_MUST "------->sanremo_event_change::SANREMO_INT_STS2_VCHG \n");
		ret |= PMIC_EVENT_VCHG_TS;
	}

	if (status[1] & SANREMO_INT_STS2_VSYS)
   {
		printk(LVT_DEBUG_MUST "------->sanremo_event_change::SANREMO_INT_STS2_VSYS \n");
		ret |= PMIC_EVENT_VSYS;
   }

	if (status[1] & SANREMO_INT_STS2_TINT)
	{
		printk(LVT_DEBUG_MUST "------->sanremo_event_change::SANREMO_INT_STS2_TINT \n");
		ret |= PMIC_EVENT_TINT;
	}

	if (status[1] & SANREMO_INT_STS2_GPADC0)
	{
		printk(LVT_DEBUG_MUST "------->sanremo_event_change::SANREMO_INT_STS2_GPADC0 \n");
	 	ret |= PMIC_EVENT_RF_TEMP;
	}

	if (status[1] & SANREMO_INT_STS2_GPADC1)
	{
		printk(LVT_DEBUG_MUST "------->sanremo_event_change::SANREMO_INT_STS2_GPADC1 \n");
		ret |= PMIC_EVENT_TBAT;
	}

/*	if (status[1] & SANREMO_INT_STS2_GPADC2)
		ret |= ;

	if (status[1] & SANREMO_INT_STS2_GPADC3)
		ret |= ;       */


    if (status[2] & SANREMO_INT_STS3_PEN)
	{
		printk(LVT_DEBUG_MUST "------->sanremo_event_change::SANREMO_INT_STS3_PEN \n");
		ret |= PMIC_EVENT_TOUCH;
	}

	if (status[2] & SANREMO_INT_STS3_HEADSET)
   {
		printk(LVT_DEBUG_MUST "------->sanremo_event_change::SANREMO_INT_STS3_HEADSET \n");
		ret |= PMIC_EVENT_HSDETECT;
   }

	if (status[2] & SANREMO_INT_STS3_HOOK)
   {
		printk(LVT_DEBUG_MUST "------->sanremo_event_change::SANREMO_INT_STS3_HOOK \n");
		ret |= PMIC_EVENT_HOOKSWITCH;
   }

	if (status[2] & SANREMO_INT_STS3_MICIN)
	{
		printk(LVT_DEBUG_MUST "------->sanremo_event_change::SANREMO_INT_STS3_MICIN \n");
		ret |= PMIC_EVENT_MICIN;
   		if  ((pmicId >= LEVATNE_A0_ID) && (pmicId <=  LEVATNE_A1_ID) )
	    	ret |= PMIC_EVENT_HSDETECT; /*this is workaround for levante A0 and A1 in B0 it should be off*/
	}

	if (status[2] & SANREMO_INT_STS3_CHG_FAIL)
   {
		printk(LVT_DEBUG_MUST "------->sanremo_event_change::SANREMO_INT_STS3_CHG_FAIL \n");
		ret |= PMIC_EVENT_CAHRG_FAIL;
   }

	if (status[2] & SANREMO_INT_STS3_CHG_DONE)
   {
		printk(LVT_DEBUG_MUST "------->sanremo_event_change:: SANREMO_INT_STS3_CHG_DONE \n");
		ret |= PMIC_EVENT_CHARG_COMP;
   }

	if (status[2] & SANREMO_INT_STS3_CHG_IOVER)
   {
		printk(LVT_DEBUG_MUST "------->sanremo_event_change::SANREMO_INT_STS3_CHG_IOVER \n");
		ret |= PMIC_EVENT_IOVER;
   }

	return ret;
}


/*============== PMIC interrupt handling workaround ========
* All PMIC events are muxed into one physical INT signal.
* The High level is inactive, low is active.
* While any event is active the INT level is low.
* The level goes up when all events are cleared.
* There are 3 EVENT registers (1/2/3) on SANREMO.
* The clear is done when write 1 into appropriated EVENT-bit.
*
* The TAVOR IRQ works on FALLING-EDGE
*
* TS Pen-Detect event occured it is present in the 3 register.
* USB event is in the reg 2.
*
* SCENARIO:
* The Pen-Detect event causes the INT High-to-Low and IRQ occured ans handler called.
* The handler reads reg 2, then 3
*              At this point just before reading the reg-3
*             new USB-event occured and written into reg-2
* Since the SW has already read the 2 !!_before_event_!! the 2 stays with active event inside.
* The INT-level stays active low, no falling edge, no new IRQ generated and never would be generated.
*
* The solution is:
*   Read the IRQ/GPIO level after "worker".
*   Normally it should be High-Inactive,
*   but if there is no new IRQ found and level is still Low-Active
*   repeat the "worker".
*/
#define MAX_WORKER_REPEAT      4
static unsigned int sanremo_int_pin;
static int sanremo_irq_in_workqueue;

static int sanremo_worker_repeat(u32 repeat)
{
	if  ((pmicId >= LEVATNE_A0_ID) && (pmicId <=  LEVATNE_A1_ID) ){

		if(repeat >= MAX_WORKER_REPEAT)
		{
			printk("\nPanic: PMIC interrupt stays LOW-active and never released\n\n");
			return 0;
		}
		if((sanremo_irq_in_workqueue == 0) && (gpio_get_value(sanremo_int_pin) == 0))
		{
			printk("WARNING: PMIC interrupt stays LOW-active\n");
			return 1;
		}
	}
	return 0;
}

static void sanremo_worker_init(unsigned int irq)
{
	int err;
	sanremo_int_pin = IRQ_TO_GPIO(irq);
	err = gpio_request(sanremo_int_pin, "SANREMO_IRQ_GPIO");
	if (err) {
		gpio_free(sanremo_int_pin);
		printk(KERN_ERR "Request GPIO_%d failed with %d\n", sanremo_int_pin, err);
		return;
	}
	gpio_direction_input(sanremo_int_pin);
	sanremo_irq_in_workqueue = 0;
}
/*============== End PMIC interrupt handling workaround ========*/


static void sanremo_worker(struct work_struct *work)
{
	unsigned int event;
	u8 event_regs[3] = {0,0,0};  /* STATUS 1 2 3 */
	u32 repeat = 0;

	sanremo_irq_in_workqueue = 0; /* no mutual protection required */

	do {
		event = sanremo_event_change(event_regs);

		if  ((pmicId >= LEVATNE_A0_ID) && (pmicId <=  LEVATNE_A1_ID) )
			sanremo_event_clear(event_regs);

		pmic_event_handle(event);

	}while (sanremo_worker_repeat(repeat++));
}

/*
 * Levante interrupt service routine.
 * In the ISR we need to check the Status bits in Levante and according
 * to those bits to check which kind of IRQ had happened.
 */
static irqreturn_t sanremo_irq_handler(int irq, void *dev_id)
{
	struct i2c_client *client = dev_id;
	struct sanremo_platform_data *pdata = client->dev.platform_data;

	/* clear the irq */
	pdata->ack_irq();
	sanremo_irq_in_workqueue = 1;
	schedule_work(&pdata->work);

	return IRQ_HANDLED;
}

#ifdef	CONFIG_PM
static int sanremo_suspend(struct i2c_client *client, pm_message_t state)
{
/*    unsigned char val; */

    disable_irq(client->irq);
 /*   if ((get_pm_state() == PM_SUSPEND_MEM)) {
        sanremo_read(MICCO_OVER1, &val);
        sanremo_regs[MICCO_OVER1].data = val;
        sanremo_read(MICCO_APP_OVER2, &val);
        sanremo_regs[MICCO_APP_OVER2].data = val;
        sanremo_read(MICCO_APP_OVER3, &val);
        sanremo_regs[MICCO_APP_OVER3].data = val;
        sanremo_read(MICCO_BUCK_SLEEP, &val);
        sanremo_regs[MICCO_BUCK_SLEEP].data = val;
        sanremo_read(MICCO_SYSCTRL_A, &val);
        sanremo_regs[MICCO_SYSCTRL_A].data = val;

        sanremo_write(MICCO_BUCK_SLEEP, 0x6D);
        sanremo_write(MICCO_SYSCTRL_A, 0xff);

      	sanremo_write(MICCO_OVER1, 0x04);
        sanremo_write(MICCO_APP_OVER2, 0x30);
        sanremo_write(MICCO_APP_OVER3, 0x7c);

    }  */
    return 0;
}

static int sanremo_resume(struct i2c_client *client)
{
    int i;

    printk(LVT_DEBUG "LEVANTE: inside sanremo_resume\n");

    for (i = 0; i < SANREMO_REG_NUM; i++)
        sanremo_regs[i].hit = 0;

    enable_irq(client->irq);
    sanremo_irq_handler(client->irq, client);
/*    if ((get_pm_state() == PM_SUSPEND_MEM)) {
        sanremo_write(MICCO_OVER1, sanremo_regs[MICCO_OVER1].data);
        sanremo_write(MICCO_APP_OVER2, sanremo_regs[MICCO_APP_OVER2].data);
        sanremo_write(MICCO_APP_OVER3, sanremo_regs[MICCO_APP_OVER3].data);
        sanremo_write(MICCO_BUCK_SLEEP, sanremo_regs[MICCO_BUCK_SLEEP].data);
        sanremo_write(MICCO_SYSCTRL_A, sanremo_regs[MICCO_SYSCTRL_A].data);
    }  */
	return 0;
}
#else
#define	sanremo_suspend		NULL
#define	sanremo_resume		NULL
#endif


static int is_sanremo_vbus_assert(void)
{
	int ret;
	u8 val;

	ret = sanremo_read(SANREMO_STATUS, &val);
	if (ret) {
		printk(KERN_ERR "I2C operation error 0x%x\n", ret);
		return ret;
	}

	if (val & SANREMO_STS_CHG)
		return 1;
	else
		return 0;
}

static int is_sanremo_avbusvld(void)
{
	printk(LVT_DEBUG "LEVANTE: is_sanremo_avbusvld - stub TBD \n");
	return 0;
}

static int is_sanremo_asessvld(void)
{
	printk(LVT_DEBUG "LEVANTE: is_sanremo_asessvld - stub TBD \n");
	return 0;
}

static int is_sanremo_bsessvld(void)
{
	printk(LVT_DEBUG "LEVANTE: is_sanremo_bsessvld - stub TBD \n");
	return 0;
}

static int is_sanremo_srp_ready(void)
{
	printk(LVT_DEBUG "LEVANTE: is_sanremo_srp_ready - stub TBD \n");
	return 0;
}

static int is_sanremo_onkey_assert(void)
{
	u8 val;

	sanremo_read(SANREMO_STATUS, &val);

	return (!!(val & SANREMO_STS_ONKEY));
}

static int sanremo_set_pump(int enable)
{
	printk(LVT_DEBUG "LEVANTE: sanremo_set_pump - stub TBD \n");
	return 0;
}
static int sanremo_set_vbus_supply(int enable, int srp)
{
	printk(LVT_DEBUG "LEVANTE: sanremo_set_vbus_supply - stub TBD \n");
	return 0;
}
static int sanremo_set_usbotg_a_mask(void)
{
	printk(LVT_DEBUG "LEVANTE: sanremo_set_usbotg_a_mask - stub TBD \n");
	return 0;
}
static int sanremo_set_usbotg_b_mask(void)
{
	printk(LVT_DEBUG "LEVANTE: sanremo_set_usbotg_b_mask - stub TBD \n");
	return 0;
}

static struct pmic_ops sanremo_pmic_ops = {
/*  .get_voltage        = get_sanremo_voltage,
    .set_voltage        = set_sanremo_voltage,
*/
    .is_vbus_assert     = is_sanremo_vbus_assert,

    .is_avbusvld        = is_sanremo_avbusvld,
    .is_asessvld        = is_sanremo_asessvld,
    .is_bsessvld        = is_sanremo_bsessvld,
    .is_srp_ready       = is_sanremo_srp_ready,

    .is_onkey_assert	= is_sanremo_onkey_assert,
    
    .set_pump           = sanremo_set_pump,
    .set_vbus_supply    = sanremo_set_vbus_supply,
    .set_usbotg_a_mask  = sanremo_set_usbotg_a_mask,
    .set_usbotg_b_mask  = sanremo_set_usbotg_b_mask,

};

#ifdef	CONFIG_PROC_FS
#define	SANREMO_PROC_FILE	"driver/sanremo"
static struct proc_dir_entry *sanremo_proc_file;
static int index;

/*
 * (1) cat /proc/driver/sanremo
 * will print the value of register located in address of index.
 */
static ssize_t sanremo_proc_read(struct file *filp,
	char __user *buffer, size_t count, loff_t *offset)
{
	u8 reg_val;
	ssize_t len;
	const char fmt[] =    "register 0x%02x: 0x%02x\n";
	const char fmtErr[] = "register 0x%02x: failed\n";

	char buf[sizeof(fmt)];

	if ((index < 0) || (index > SANREMO_REG_NUM))
		return 0;

	sanremo_regs[index].hit = 0; /* Force to read but not take from cache */

	if (!sanremo_read(index, &reg_val))
		len = sprintf(buf, fmt, index, reg_val);
	else
		len = sprintf(buf, fmtErr, index);

	return simple_read_from_buffer(buffer, count, offset, buf, len);
}

/*
 * (1) echo -0x0b > /proc/driver/sanremo
 * will set index to 0x0b such that next proc_read will be done from this new index value.
 *
 * (2) echo 0x0b > /proc/driver/sanremo
 * will set value of 0x0b for the register in address of index
 *
* */
static ssize_t sanremo_proc_write(struct file *filp,
		const char *buff, size_t len, loff_t *off)
{
	u8 reg_val;
	char messages[256], vol[256];

	if (len > 256)
		len = 256;

	if (copy_from_user(messages, buff, len))
		return -EFAULT;

	if ('-' == messages[0]) {
		/* set the register index */
		memcpy(vol, messages+1, len-1);
		index = (int) simple_strtoul(vol, NULL, 16);
		printk(LVT_DEBUG_MUST "\nsanremo_proc_write:register # was set = 0x%x\n", index);
    }else if ('?' == messages[0]) {
		printk(LVT_DEBUG_MUST "\n write: echo -0x56 > /proc/driver/sanremo \n");
		printk(LVT_DEBUG_MUST " read(reg from write): cat /proc/driver/sanremo \n");
		printk(LVT_DEBUG_MUST " read/write trace on: echo traceon > /proc/driver/sanremo \n");
		printk(LVT_DEBUG_MUST " charger test page trace on: echo trcht > /proc/driver/sanremo \n");
		printk(LVT_DEBUG_MUST " charger other register trace on: echo trchr > /proc/driver/sanremo \n");
	}else if (strncmp(messages, "traceon", 7) == 0){
		sanremo_print = 1;
		printk(LVT_DEBUG_MUST "\nsanremo_proc_write:enable sanremo read/write traces \n");
	}else if (strncmp(messages, "trcht", 5) == 0){

		sanremo_print = 1;

		sanremo_write(0xFA, 0x00);
   		sanremo_write(0xFB, 0x00);
		sanremo_write(0xFF ,0x00);
		sanremo_read(0xD3 ,&reg_val );
		sanremo_read(0xD4 ,&reg_val );
		sanremo_read(0xD5 ,&reg_val );
		sanremo_read(0xD6 ,&reg_val );
		sanremo_read(0xD7 ,&reg_val );
		/*-----------------------*/
   		/*<--- exit test page*/
   		sanremo_write(0xFE, 0x00);
   		sanremo_write(0xFC, 0x00);

		sanremo_print = 0;

	}else if (strncmp(messages, "trchr", 5) == 0){

		printk("======> read charger register \n");
		sanremo_read(0x48 ,&reg_val );
		printk("======> SANREMO_PRINT_READ: 0x0%x->0x%x \n",0x48,reg_val);
		sanremo_read(0x49 ,&reg_val );
		printk("======> SANREMO_PRINT_READ: 0x0%x->0x%x \n",0x49,reg_val);
		sanremo_read(0x4A ,&reg_val );
		printk("======> SANREMO_PRINT_READ: 0x0%x->0x%x \n",0x4A,reg_val);
		sanremo_read(0x4B ,&reg_val );
		printk("======> SANREMO_PRINT_READ: 0x0%x->0x%x \n",0x4B,reg_val);
		sanremo_read(0x4C ,&reg_val );
		printk("======> SANREMO_PRINT_READ: 0x0%x->0x%x \n",0x4C,reg_val);
		sanremo_read(0x4D ,&reg_val );
		printk("======> SANREMO_PRINT_READ: 0x0%x->0x%x \n",0x4D,reg_val);
		sanremo_read(0x4E ,&reg_val );
		printk("======> SANREMO_PRINT_READ: 0x0%x->0x%x \n",0x4E,reg_val);
		sanremo_read(0x52 ,&reg_val );
		printk("======> SANREMO_PRINT_READ: 0x0%x->0x%x \n",0x52,reg_val);
		sanremo_read(0x5b ,&reg_val );
		printk("======> SANREMO_PRINT_READ: 0x0%x->0x%x \n",0x5b,reg_val);
		sanremo_read(0x63 ,&reg_val );
		printk("======> SANREMO_PRINT_READ: 0x0%x->0x%x \n",0x63,reg_val);
		sanremo_read(0x6d ,&reg_val );
		printk("======> SANREMO_PRINT_READ: 0x0%x->0x%x \n",0x6d,reg_val);

	}else{
		/* set the register value */
		reg_val = (int)simple_strtoul(messages, NULL, 16);
		sanremo_write(index, reg_val & 0xFF);
		printk(LVT_DEBUG_MUST "sanremo_proc_write:register write 0x%x <- %x \n", index, (reg_val & 0xFF));
	}
	return len;
}

static struct file_operations sanremo_proc_ops = {
	.read = sanremo_proc_read,
	.write = sanremo_proc_write,
};

static void create_sanremo_proc_file(void)
{
	sanremo_proc_file = create_proc_entry(SANREMO_PROC_FILE, 0644, NULL);
	if (sanremo_proc_file) {
		sanremo_proc_file->owner = THIS_MODULE;
		sanremo_proc_file->proc_fops = &sanremo_proc_ops;
	} else
		printk(LVT_DEBUG_MUST "proc file create failed!\n");
}

static void remove_sanremo_proc_file(void)
{
	remove_proc_entry(SANREMO_PROC_FILE, NULL);
}

#endif

static int sanremo_remove(struct i2c_client *client)
{
	pmic_set_ops(NULL);

#ifdef	CONFIG_PROC_FS
	remove_sanremo_proc_file();
#endif

	free_irq(client->irq, NULL);

	return 0;
}

static int __devinit sanremo_probe(struct i2c_client *client,
			 const struct i2c_device_id *id)
{
	struct sanremo_platform_data *pdata;
	int ret;
	int chip_id;
	int i;

	printk(LVT_DEBUG_MUST "--->>>sanremo_probe!!!\n");

	chip_id = i2c_smbus_read_byte_data(client, SANREMO_CHIP_ID);
	if(chip_id < 0){
		printk(LVT_DEBUG_MUST "i2c fail -> sanremo unavailable!!!\n");
		g_client = NULL;
		return -ENXIO;
	}else if ( (chip_id < LEVATNE_A0_ID) || (chip_id >  LEVATNE_END_ID) ){
		printk(LVT_DEBUG_MUST "Id[0x%02x] is not sanremo -->>>>sanremo unavailable!!!\n", chip_id);
		g_client = NULL;
		return -ENXIO;
	} else {
		printk(LVT_DEBUG_MUST "sanremo(chip id:0x%02x) detected.\n", chip_id);
	}

	g_client = client;

	ret = sanremo_initchip();
	if (ret != 0)
		printk(LVT_DEBUG_MUST "Initialize Sanremo failed with ret 0x%x\n",ret);

	pdata = client->dev.platform_data;

	mutex_init(&pdata->m_lock);

	/* init workqueue */
	INIT_WORK(&pdata->work, sanremo_worker);
	/* init irq */
	pdata->init_irq();
	ret = request_irq(client->irq, sanremo_irq_handler, IRQF_TRIGGER_FALLING,
			"Sanremo", client);
	if (ret) {
		printk(LVT_DEBUG_MUST "Request IRQ for Sanremo failed, return:%d\n",
				ret);
		return ret;
	}

	sanremo_worker_init(client->irq);

	for (i = 0; pdata->power_chips[i].power_supply_modules != NULL; i++) {
		if (pdata->power_chips[i].chip_id == chip_id) {
			sanremo_power_module =
			  (struct power_supply_module*) pdata->power_chips[i].power_supply_modules;
			break;
		}
	}
	if (pdata->power_chips[i].power_supply_modules == NULL)
		panic("This Sanremo chip is not supported");

#ifdef	CONFIG_PROC_FS
	create_sanremo_proc_file();
#endif

	pmic_set_ops(&sanremo_pmic_ops);

	return 0;
}

static const struct i2c_device_id sanremo_id[] = {
	{ "sanremo", 0 },
	{ }
};
static struct i2c_driver sanremo_driver = {
	.driver = {
		.name	= "sanremo",
	},
	.probe		= sanremo_probe,
	.remove		= sanremo_remove,
	.id_table	= sanremo_id,
	.suspend	= sanremo_suspend,
	.resume		= sanremo_resume,
};

static int __init sanremo_init(void)
{
   int retVal = -EINVAL;
   pmicId = pmic_id();
   if  ((pmicId >= LEVATNE_A0_ID) && (pmicId <=  LEVATNE_END_ID) ){
	   retVal=i2c_add_driver(&sanremo_driver);
   }
   return (retVal);
}

static void __exit sanremo_exit(void)
{
	flush_scheduled_work();
	i2c_del_driver(&sanremo_driver);
}

subsys_initcall(sanremo_init);
module_exit(sanremo_exit);

MODULE_AUTHOR("Yael,Shemla <yshemla@marvell.com > Yossef,Hanin <yhanin@marvell.com>");
MODULE_DESCRIPTION("Sanremo Driver");
MODULE_LICENSE("GPL");
