/*
 *  linux/drivers/input/touchscreen/levante_touch.c
 *
 *  touch screen driver for pxa935 - Levante
 *
 *  Copyright (C) 2006, Marvell Corporation (fengwei.yin@Marvell.com)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2 as
 *  published by the Free Software Foundation.
 */
#include <linux/module.h>
#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/input.h>
#include <linux/interrupt.h>
#include <linux/workqueue.h>
#include <linux/wait.h>
#include <linux/delay.h>
#include <linux/platform_device.h>
#include <linux/freezer.h>
#include <linux/proc_fs.h>

#include <asm/irq.h>
#include <mach/hardware.h>

#include <mach/pxa-regs.h>
#include <mach/levante.h>
#include <mach/levante_hw.h>
#include <mach/pxa3xx_pmic.h>
#include <asm/uaccess.h>
#include <mach/tavor_common.h>

DEFINE_MUTEX(tsi_lock);

static int TS_print;
#define DIS_BET_2_POINTS 20 /* Pen Accuracy distance^2 = x^2 + y^2 */
#define TOUCHSCREEN_CONSOLE_PRINT_PENDETECT  if(TS_print) printk("levante_ts_interrupt  PEN-DETECT\n");
#define TOUCHSCREEN_CONSOLE_PRINT_XY         if(TS_print) printk("======> Touch-Screen:  X=%d    Y=%d\n", tem_x, tem_y);
#define TOUCHSCREEN_CONSOLE_PRINT_PENUP      if(TS_print) printk("======> Touch-Screen: PEN-UP\n");

/*#define DEBUG_PRINT_ON*/
#ifdef DEBUG_PRINT_ON
#define PRINTK(TXT)            printk(TXT)
#define PRINTK1(TXT,P1)        printk(TXT, P1)
#define PRINTK2(TXT,P1,P2)     printk(TXT, P1, P2)
#else
#define PRINTK(TXT)
#define PRINTK1(TXT,P1)
#define PRINTK2(TXT,P1,P2)
#endif

#ifdef TSI_MIPS_RAM
	#ifdef CONFIG_PXA_MIPSRAM
	#include <linux/mipsram.h>
	#endif /* CONFIG_PXA_MIPSRAM */
#endif


/* levante_ts_opt:
 * 		=0, 1, 2, 3:
 * 		sampling rate is higher while the opt value is higher.
 *		Touch timer block time is higher while the opt value is higher.
 */

static u8 pmicId = 0;
static u16 	tem_x_pre = 0xffff;
static u16 	tem_y_pre = 0xffff;

static int levante_ts_opt = 0;

enum {
	TSI_EVENT_NONE = 0,
	TSI_EVENT_TOUCH = 1,
	TSI_EVENT_LIFT = 2
};

enum {
	TSI_PEN_UNKNOW = 0,
	TSI_PEN_DOWN = 1,
	TSI_PEN_UP = 2
};

struct levante_ts_data {
	spinlock_t			ts_lock;
	int						suspended;
	int						pen_state;
	int						use_count;
	struct timer_list 		*ts_timer;
	struct	work_struct		work;
};

static struct levante_ts_data *levante_td;
static struct input_dev *levante_ts_input_dev;

static int levante_ts_proc_write(struct file *file, const char __user *buffer,
			   unsigned long count, void *data)
{
	static char kbuf[1024];
	int arg;

	if (count >= 1024)
		return -EINVAL;
	if (copy_from_user(kbuf, buffer, count))
		return -EFAULT;

	if ('?' == kbuf[0]) {
		printk("HowTo: {tson \\ tsoff} echo tson > /proc/driver/levante_ts\n");
	} else if (strncmp(kbuf, "tson", 4) == 0) {
		TS_print = 1;
		printk(LVT_DEBUG_MUST "\nTSI prints on\n");
	} else if (strncmp(kbuf, "tsoff", 4) == 0) {
		TS_print = 0;
		printk(LVT_DEBUG_MUST "\nTSI prints off\n");
	} else {
		sscanf(kbuf, "%d", &arg);
		if(arg <= 3)
			levante_ts_opt = arg;
		printk(LVT_DEBUG_MUST "levante ts opt = %d\n", levante_ts_opt);
	}
	return count;
}

int levante_ts_read(u16 *p_x,  u16 *p_y, int *p_pen_state)
{
	int ret;

	ret = levante_tsi_readxy(p_x, p_y, (u8*)p_pen_state);
	return ret;
}

/*(xd*xd + yd*yd) < (Diameter*Diameter)*/
static int levante_evaluate_point(int iX_Pre, int iY_Pre, int iX, int iY)
{
    int tempx,tempy,diff;
    int fRet=0;

    if ((iX_Pre < 0) || (iX < 0)){  /* not valid point*/
        return -EINVAL;
    }

	if( tem_x_pre == 0xffff && tem_y_pre ==0xffff ){/*the first time no need to evaluate*/
		return 0;
	}

 	tempx = abs( iX - iX_Pre);
	tempy = abs( iY - iY_Pre);

	diff = ((tempx ^ 2) + (tempy ^ 2));

	 /* if the dis between 2 points are less then DIS_BET_2_POINTS but
	  greater then zero ( zero means its the same point which was reported)
	  then it's valid */
	fRet =  ((diff < (DIS_BET_2_POINTS * DIS_BET_2_POINTS)) ? 0 : -EINVAL);

	PRINTK1(LVT_DEBUG "-------->levante_evaluate_point2 ok - (diff) = %d \n",diff);

    if (fRet < 0) {
		printk(LVT_DEBUG_MUST "-------->!!!levante_evaluate_point skip!!! - diff [%d] pre[x=%d y=%d] current[x=%d y=%d]\n",diff,iX_Pre, iY_Pre, iX, iY);
	}
	return fRet;
}

/* setting the pen debounce 0ms=0x00/2ms=0x01/4ms=0x02/8ms=0x03*/
static int levante_tsi_set_pen_debounce(u8 debounce)
{  /* change debounce to 0ms*/
	u8 val;
	sanremo_read(SANREMO_DEBOUNCE_REG, &val);

	if (debounce == 0)
		val &= 0xFC;
	else
		val |= SANREMO_DEBOUNCE_PEN_DET(debounce);

	sanremo_write(SANREMO_DEBOUNCE_REG, val);
}

/* ===================== RUN-TIME HANDLING ==================================== */

/*  The context is SW-IRQ => I2C operations cannot be used here,
 *    activate the thread instead.
*/
static void levante_ts_timer_handler(unsigned long d)
{
	if (levante_td->use_count > 0) {
		schedule_work(&levante_td->work);
	}
}

/*
 * The callback function for levante touch.
 *
 *  The context is NOT interrupt but worker => I2C operations may be used here
 *
 */
void levante_ts_interrupt(unsigned long event)
{
 u8 status;
#ifdef DEBUG_PRINT_ON
	static u16 i=0;
 	PRINTK1(LVT_DEBUG "------>levante_ts_interrupt begin = %d\n", i++);
#endif
   	if  ((pmicId >= LEVATNE_A0_ID) && (pmicId <=  LEVATNE_A1_ID) )
			sanremo_read(SANREMO_STATUS, &status);
	else if(pmicId >= LEVATNE_B0_ID)
			sanremo_read(SANREMO_STATUS_B0, &status);

	/* handle only pen-down events , pen-up is detected while measure x,y*/
	if (status & SANREMO_STS_PEN_DOWN) {
		if (levante_td->use_count > 0) {
			if (TSI_PEN_UP ==  levante_td->pen_state) {

				TOUCHSCREEN_CONSOLE_PRINT_PENDETECT;
				levante_td->pen_state = TSI_PEN_DOWN;
				/*Disable pen IRQ*/
				levante_enable_pen_down_irq(0);

				if  ((pmicId >= LEVATNE_A0_ID) && (pmicId <=  LEVATNE_A1_ID) ){
					/*pen detect event is detected, before entering in always on mode*/
					sanremo_write(0xFA, 0x00);
   					sanremo_write(0xFB, 0x00);
   					sanremo_write(0xFF ,0x00 );
   					/*-----------------------*/
   					sanremo_write(SANREMO_TEST_PAGE1_CC ,0xFF );
		  			/*-----------------------*/
   					/*<--- exit test page*/
   					sanremo_write(0xFE, 0x00);
   					sanremo_write(0xFC, 0x00);
				}

				/* change to 0ms*/
				levante_tsi_set_pen_debounce(0x0);
				levante_tsi_enable_xy_measurements();

				/* activate timer for x,y readings*/
				/* for first mesurmant wait 5ms -
				*in Tavor it will be 7.8msec since
				* 1 jiffies = 7.8ms == 256 * 1/32K*/
				levante_td->ts_timer->expires =
				jiffies + msecs_to_jiffies(5);
				/* check for pending before mod */
				if (timer_pending(levante_td->ts_timer) == 0)
					mod_timer(levante_td->ts_timer,
					levante_td->ts_timer->expires);
				/* initilize for this pen-down event*/
				tem_x_pre = tem_y_pre = 0xffff;
			}
		}
		PRINTK1(LVT_DEBUG "-------->levante_ts_interrupt end = %d\n", i);
		printk(LVT_DEBUG_MUST "-------->levante_ts_interrupt - measure x,y start... for seeing x,y value:echo tson > /proc/driver/levante_ts \n");
	}else
		printk(LVT_DEBUG_MUST \
		"------->levante_ts_interrupt end - ignor pen-up \n");

	return;
}
EXPORT_SYMBOL(levante_ts_interrupt);

extern int is_android(void);

static void levante_ts_worker(struct work_struct *work)
{
	int valid_point = -EINVAL;
	u8 pen_state;
	static u16 tem_x = 0xffff;
	static u16 tem_y = 0xffff;
	int retVal = 0;
	unsigned int time_stamp_before,time_stamp_after,diffmesc,next_delay;

  	PRINTK(LVT_DEBUG "-------->levante_ts_worker - start\n");

#ifdef TSI_MIPS_RAM
	#ifdef CONFIG_PXA_MIPSRAM
		MIPS_RAM_ADD_TRACE((unsigned int)0x70000);/*0x70000 is the trace ID need to make sure it is updated in MipsRam_linux.xls in event sheet*/
	#endif/* CONFIG_PXA_MIPSRAM */
#endif

	/* use 32Khz  (OSCR use 13Mhz if there is a
	need of usec resulotion)
	need to include #include <asm/hardware.h>
	#include <asm/arch/pxa-regs.h> */
	time_stamp_before = OSCR4;

	mutex_lock(&tsi_lock);

	retVal = levante_tsi_readxy(&tem_x, &tem_y, &pen_state);

	printk(LVT_DEBUG \
	"---->levante_ts_worker - pen state [%d][1:up 0:down]\n", pen_state);

	if (pen_state) {

		//check current location vs. previous - if too far then measure again don't send to user.
		if (!retVal)
			valid_point = levante_evaluate_point(tem_x_pre,tem_y_pre,tem_x,tem_y);

		/* report the point only if xy meas is ok and eval is valid */
		if (!retVal && !valid_point) {
 			// send measure to user here.
			input_report_abs(levante_ts_input_dev,	\
				ABS_X, (tem_x & 0xfff)-250/*12b vaild*/);
			input_report_abs(levante_ts_input_dev,	\
				ABS_Y, 4096-(tem_y & 0xfff)-250/*12b vaild*/);
			if (!is_android())
				input_report_abs(levante_ts_input_dev,
						ABS_PRESSURE, 255);
			if (is_android()) {
				input_report_key(levante_ts_input_dev,
						BTN_TOUCH, 1);
				input_sync(levante_ts_input_dev);
			}

			tem_x_pre = tem_x;/* update for next points evaluate*/
			tem_y_pre = tem_y;/* update for next points evaluate*/

			printk(LVT_DEBUG \
			"--->levante_ts_worker-tem_x_pre[%d],tem_y_pre[%d] \n"
			, tem_x_pre, tem_y_pre);
			printk(LVT_DEBUG \
			"--->levante_ts_worker-report pen down x[%d],y[%d]\n"
			, tem_x, tem_y);
			TOUCHSCREEN_CONSOLE_PRINT_XY;
		} else
			printk(LVT_DEBUG_MUST \
			"levante_ts_worker - NOT VALID point don't report\n");

		time_stamp_after = OSCR4; /* we now know when to trigger the next measuremt in order to have 1 in 10msec*/
		diffmesc = levante_pxa_ticks_to_msec(time_stamp_after - time_stamp_before);
		PRINTK1(LVT_DEBUG "-------->levante_ts_worker - diffmesc[%d msec] \n",diffmesc);

		if( diffmesc < LEVNATE_XY_MES_CYC_MS)
			next_delay =  LEVNATE_XY_MES_CYC_MS - diffmesc;/*we want a mesurment every LEVNATE_XY_MES_CYC_MS (10ms)
															* cycle,less might overloading the system
															* WM standart normal 100 in a sec*/
		 else
			next_delay = 0;/*if diffmesc > LEVNATE_XY_MES_CYC_MS - it seems we spent lots of time calc -> don't delay next mesure*/

		PRINTK1(LVT_DEBUG_MUST "-------->levante_ts_worker - next delay[%d] \n",next_delay);
		levante_td->ts_timer->expires = jiffies + msecs_to_jiffies(next_delay);/* since OS tick is 7.8msec - next_delay usualy == 1 tick */
		/* check for pending beofre mod timer*/
		if (timer_pending(levante_td->ts_timer) == 0)
			mod_timer(levante_td->ts_timer,
			levante_td->ts_timer->expires);
	} else {
		/* Report a pen up event */

		printk(LVT_DEBUG_MUST "-------->levante_ts_worker - pen-up event \n");

		if( levante_td->pen_state == TSI_PEN_DOWN){/*Only do if the previous state was pen-down,if it is pen-up then nothing to do here*/

			levante_td->pen_state = TSI_PEN_UP;

			if (!is_android()) {
				input_report_abs(levante_ts_input_dev,
						ABS_PRESSURE, 0);
				input_report_abs(levante_ts_input_dev,
						ABS_TOOL_WIDTH, 1);
			}
			if (is_android()) {
				input_report_key(levante_ts_input_dev,
						BTN_TOUCH, 0);
				input_sync(levante_ts_input_dev);
			}

			printk(LVT_DEBUG_MUST "-------->levante_ts_worker-reported to user:pen-up event\n");

			if  ((pmicId >= LEVATNE_A0_ID) && (pmicId <=  LEVATNE_A1_ID) ){

				levante_tsi_disable_all_measurements();

				/* each time duty cyclemode is entered*/
				sanremo_write(0xFA, 0x00);
	   			sanremo_write(0xFB, 0x00);
	   			sanremo_write(0xFF ,0x00 );
	   			/*-----------------------*/
	   			sanremo_write(SANREMO_TEST_PAGE1_CC ,0xF3 );
		  		/*-----------------------*/
	   			/*<--- exit test page*/
	   			sanremo_write(0xFE, 0x00);
	   			sanremo_write(0xFC, 0x00);

				printk(LVT_DEBUG_MUST \
				"levante_ts_worker-reported -WA for A1 dis\n");
			}

			/* change back to 8ms*/
			levante_tsi_set_pen_debounce(0x03);
			/* enable pen IRQ*/
			levante_enable_pen_down_irq(1);
			/* check for pending beofre delete*/
			if (timer_pending(levante_td->ts_timer) == 0)
				del_timer(levante_td->ts_timer);
		} else
		 printk(LVT_DEBUG_MUST \
		 "levante_ts_worker - pen is already up - doing nothing \n");
	}

#ifdef TSI_MIPS_RAM
	#ifdef CONFIG_PXA_MIPSRAM
		MIPS_RAM_ADD_TRACE((unsigned int)0x70001);/*0x70001 is the trace ID need to make sure it is updated in MipsRam_linux.xls in event sheet*/
	#endif/* CONFIG_PXA_MIPSRAM */
#endif
	mutex_unlock(&tsi_lock);

	return;
}


/* ===================== INIT and CONTROL ================================ */

static int levante_ts_open(struct input_dev *idev)
{
	unsigned long flags;

	pr_debug("%s: enter", __func__);

	if (levante_td->suspended) {
		printk(LVT_DEBUG_MUST "touch has been suspended!\n");
		return -1;
	}

	spin_lock_irqsave(&levante_td->ts_lock, flags);
	if (levante_td->use_count++ == 0) {
		spin_unlock_irqrestore(&levante_td->ts_lock, flags);

		levante_td->ts_timer =
			(struct timer_list *)kmalloc(sizeof(struct timer_list), GFP_KERNEL);
		init_timer(levante_td->ts_timer);
		levante_td->ts_timer->function = levante_ts_timer_handler;
		levante_td->ts_timer->data = (int)levante_td;
	} else {
		spin_unlock_irqrestore(&levante_td->ts_lock, flags);
	}

	return 0;
}

void levante_open_tsi(void)
{
	u8 val;
	levante_ts_open(levante_ts_input_dev);
	/* workaround since if
	 * TSI is not working need to disable it */
	sanremo_read(SANREMO_MEAS_ENABLE3, &val);
	val &= ~(SANREMO_MEAS_EN3_COULOMB_COUNTER);
	sanremo_write(SANREMO_MEAS_ENABLE3, val);
	printk(LVT_DEBUG_MUST \
		"levante_open_tsi - WA COULOMB_COUNTER disable");
}

/* Kill the touchscreen thread and stop the touch digitiser. */
static void levante_ts_close(struct input_dev *idev)
{
	unsigned long flags;

	printk(LVT_DEBUG_MUST "levante_ts_close!\n");
	pr_debug("%s: enter with use count = %d", __func__,
		levante_td->use_count);

	spin_lock_irqsave(&levante_td->ts_lock, flags);
	if (--levante_td->use_count == 0) {
		spin_unlock_irqrestore(&levante_td->ts_lock, flags);
		kfree(levante_td->ts_timer);
	} else {
		spin_unlock_irqrestore(&levante_td->ts_lock, flags);
	}
}

static int levante_ts_probe(struct platform_device *pdev)
{
	int ret;
	struct proc_dir_entry *levante_ts_proc_entry;


	printk(LVT_DEBUG_MUST "------->levante_ts_probe - start\n");

	levante_td = kzalloc(sizeof(struct levante_ts_data), GFP_KERNEL);
	if (!levante_td) {
		ret = -ENOMEM;
		printk(LVT_DEBUG_MUST "------->levante_ts_probe - kzalloc fail \n");
		goto levante_ts_out;
	}
	spin_lock_init(&levante_td->ts_lock);
	INIT_WORK(&levante_td->work, levante_ts_worker);
	levante_td->pen_state = TSI_PEN_UP;
	levante_td->suspended = 0;
	levante_td->use_count = 0;

	/* register input device */
	levante_ts_input_dev = input_allocate_device();
	if (levante_ts_input_dev == NULL) {
		printk(LVT_DEBUG_MUST "------->levante_ts_probe - fail to allocate device\n");
		return -ENOMEM;
	}

	levante_ts_input_dev->name = "levante-ts";

	levante_ts_input_dev->phys = "levante-ts/input0";
	levante_ts_input_dev->dev.parent = &pdev->dev;

	levante_ts_input_dev->open = levante_ts_open;
	levante_ts_input_dev->close = levante_ts_close;

	__set_bit(EV_ABS, levante_ts_input_dev->evbit);
	__set_bit(ABS_X, levante_ts_input_dev->absbit);
	__set_bit(ABS_Y, levante_ts_input_dev->absbit);
	__set_bit(ABS_PRESSURE, levante_ts_input_dev->absbit);

	__set_bit(EV_SYN, levante_ts_input_dev->evbit);
	__set_bit(EV_KEY, levante_ts_input_dev->evbit);
	__set_bit(BTN_TOUCH, levante_ts_input_dev->keybit);

	input_set_abs_params(levante_ts_input_dev, ABS_X, LEVANTE_X_AXIS_MIN, LEVANTE_X_AXIS_MAX, 0, 0);
	input_set_abs_params(levante_ts_input_dev, ABS_Y, LEVANTE_Y_AXIS_MIN, LEVANTE_Y_AXIS_MAX, 0, 0);
	input_set_abs_params(levante_ts_input_dev, ABS_PRESSURE, LEVANTE_PRESSURE_MIN, LEVANTE_PRESSURE_MAX, 0, 0);
	input_set_abs_params(levante_ts_input_dev, ABS_TOOL_WIDTH, LEVANTE_TOOL_WIDTH_MIN, LEVANTE_TOOL_WIDTH_MAX, 0, 0);

	ret = input_register_device(levante_ts_input_dev);
	if (ret) {
		printk(LVT_DEBUG_MUST "------->levante_ts_probe - fail to register device\n");
		return ret;
	}

	/* IRQ come from levante. So we just implemate
	 * a callback IRQ function for levante.
	 */
	ret = pmic_callback_register(PMIC_EVENT_TOUCH, levante_ts_interrupt);
	if (ret < 0)
	{
		printk(LVT_DEBUG_MUST "------->levante_ts_probe - register on PMIC_EVENT_TOUCH fail \n");
		goto pmic_cb_out;
	}

	pmicId = pmic_id();

	levante_enable_pen_down_irq(1);
	levante_tsi_measurements_config(); /*for detecting pen-down*/
	levante_ts_proc_entry = create_proc_entry("driver/levante_ts", 0, NULL);
	if (levante_ts_proc_entry) {
		levante_ts_proc_entry->write_proc = levante_ts_proc_write;
	}

	printk(LVT_DEBUG_MUST "------->levante_ts_probe - end \n");

	return 0;

pmic_cb_out:
	input_unregister_device(levante_ts_input_dev);
levante_ts_out:
	kfree(levante_td);

 	printk(LVT_DEBUG_MUST "------->levante_ts_probe fail - end \n");
 	return ret;
}

static int levante_ts_remove(struct platform_device *pdev)
{
	input_unregister_device(levante_ts_input_dev);

	return 0;
}

#if CONFIG_PM
static int levante_ts_resume(struct platform_device *pdev)
{
	levante_enable_pen_down_irq(1);
	levante_tsi_measurements_config();
	levante_td->suspended = 0;

	return 0;
}

static int levante_ts_suspend(struct platform_device *pdev, pm_message_t state)
{
	levante_td->suspended = 1;
	levante_tsi_disable_all_measurements();
	levante_enable_pen_down_irq(0);

	return 0;
}
#else
#define levante_ts_resume NULL
#define levante_ts_suspend NULL
#endif

static struct platform_driver levante_ts_driver = {
	.driver = {
		.name 	= "levante-ts",
	},
	.probe		= levante_ts_probe,
	.remove		= levante_ts_remove,
	.resume 	= levante_ts_resume,
	.suspend 	= levante_ts_suspend,
};

static int __init levante_ts_init(void)
{
	return platform_driver_register(&levante_ts_driver);
}

static void __exit levante_ts_exit(void)
{
	/* We move these codes here because we want to detect the
	 * pen down event even when touch driver is not opened.
	 */
	levante_tsi_disable_all_measurements();
	levante_enable_pen_down_irq(0);
	pmic_callback_unregister(PMIC_EVENT_TOUCH, levante_ts_interrupt);

	platform_driver_unregister(&levante_ts_driver);
}

module_init(levante_ts_init);
module_exit(levante_ts_exit);

MODULE_AUTHOR("Yin, Fengwei <fengwei.yin@marvell.com> Yael,Shemla <yshemla@marvell.com > Yossef,Hanin <yhanin@marvell.com>");
MODULE_DESCRIPTION("Levante touch screen driver");
MODULE_LICENSE("GPL");
