/*
 *  linux/drivers/input/touchscreen/micco_touch.c
 *
 *  touch screen driver for Littleton Platform
 *
 *  Copyright (C) 2006, Marvell Corporation (fengwei.yin@Marvell.com)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2 as
 *  published by the Free Software Foundation.
 */
#include <linux/module.h>
#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/input.h>
#include <linux/interrupt.h>
#include <linux/wait.h>
#include <linux/delay.h>
#include <linux/platform_device.h>
#include <linux/freezer.h>
#include <linux/proc_fs.h>
#include <linux/earlysuspend.h>
#include <asm/irq.h>
#include <mach/hardware.h>

#include <mach/pxa-regs.h>
#include <mach/micco.h>
#include <mach/pxa3xx_pmic.h>
#include <asm/uaccess.h>

#ifdef CONFIG_PXA3xx_DVFM
#include <mach/dvfm.h>
#include <mach/pxa3xx_dvfm.h>
#endif

/* CONFIG_TOUCHSCREEN_MICCO_TIMER is defined in the .config
* When this flag is defined the I2C driver is called directly from the Timer-IRQ.
* The I2C driver does NOT supports IRQ-context when working
* in the interrupt but not polling (pio) mode.
* From the other hand the NON TIMER TouchScreen is working well.
* This flag is used only here; tet's undefine it.
*/
#undef CONFIG_TOUCHSCREEN_MICCO_TIMER
static int TS_print;
#define TOUCHSCREEN_CONSOLE_PRINT_XY     if (TS_print) printk(KERN_INFO "======> Touch-Screen:  X=0x%x    Y=0x%x\n", tem_x, tem_y)
#define TOUCHSCREEN_CONSOLE_PRINT_PENUP  if (TS_print) printk(KERN_INFO "======> Touch-Screen: PEN-UP\n");

#define INVALID_XY		0xFFF
#define VALID_MIN_XY	0x30
#define VALID_MAX_XY    0x3D0
#define VALID_MAX_DIF   0xE0

/* micco_ts_opt: 
 * 		=0, 1, 2, 3: 
 * 		sampling rate is higher while the opt value is higher.
 *		Touch timer block time is higher while the opt value is higher.
 */
static int micco_ts_opt = 0;
enum {
	TSI_EVENT_NONE = 0,
	TSI_EVENT_TOUCH = 1,
	TSI_EVENT_LIFT = 2
};

enum {
	TSI_PEN_UNKNOW = 0,
	TSI_PEN_DOWN = 1,
	TSI_PEN_UP = 2
};

struct micco_ts_eval {
	u16		X_pre;
	u16		Y_pre;
	u32		error_on_measure;
};

struct micco_ts_data {
	spinlock_t			ts_lock;
	struct task_struct		*thread;
	int				suspended;
	int				pen_state;
	int				use_count;
#ifdef CONFIG_TOUCHSCREEN_MICCO_TIMER
	struct timer_list 		*ts_timer;
	int 				step;
#else
	//struct task_struct		*thread;
 	wait_queue_head_t		ts_wait_queue;
 	struct completion		thread_init;
 	struct completion		thread_exit;
#endif

#ifdef CONFIG_PXA3xx_DVFM
	int 				dvfm_dev_idx;
	struct dvfm_lock 		dvfm_lock;
	struct timer_list 		ui_timer;
	unsigned int 			ui_timeout_val;
#endif
};

static struct micco_ts_data *micco_td;
static struct input_dev *micco_ts_input_dev;

#ifdef CONFIG_PXA3xx_DVFM
static void set_dvfm_constraint(struct micco_ts_data *p)
{
	spin_lock_irqsave(&p->dvfm_lock.lock, p->dvfm_lock.flags);

	dvfm_disable_op_name("416M", p->dvfm_dev_idx);
	dvfm_disable_op_name("208M", p->dvfm_dev_idx);
	dvfm_disable_op_name("156M", p->dvfm_dev_idx);
	dvfm_disable_op_name("D0CS", p->dvfm_dev_idx);

	spin_unlock_irqrestore(&p->dvfm_lock.lock, p->dvfm_lock.flags);
}

static void unset_dvfm_constraint(struct micco_ts_data *p)
{
	spin_lock_irqsave(&p->dvfm_lock.lock, p->dvfm_lock.flags);

	dvfm_enable_op_name("D0CS",p->dvfm_dev_idx);
	dvfm_enable_op_name("156M",p->dvfm_dev_idx);
	dvfm_enable_op_name("208M",p->dvfm_dev_idx);
	dvfm_enable_op_name("416M",p->dvfm_dev_idx);
	spin_unlock_irqrestore(&p->dvfm_lock.lock, p->dvfm_lock.flags);
}

static void ui_timer_handler(unsigned long data)
{
	struct micco_ts_data *p = (struct micco_ts_data *)data;
	unset_dvfm_constraint(p);
}
#endif

static int micco_ts_proc_write(struct file *file, const char __user *buffer,
			   unsigned long count, void *data)
{
	static char kbuf[1024];
	int arg;

	if (count >= 1024)
		return -EINVAL;
	if (copy_from_user(kbuf, buffer, count))
		return -EFAULT;

	if ('?' == kbuf[0]) {
		printk(KERN_INFO "HowTo: {tson \\ tsoff} echo tson > /proc/driver/micco_ts\n");
	} else if (strncmp(kbuf, "tson", 4) == 0) {
		TS_print = 1;
		printk(KERN_INFO "TSI prints on\n");
	} else if (strncmp(kbuf, "tsoff", 4) == 0) {
		TS_print = 0;
		printk(KERN_INFO "TSI prints off\n");
	} else {
		sscanf(kbuf, "%d", &arg);
		if (arg <= 3)
			micco_ts_opt = arg;
		printk(KERN_INFO "micco ts opt = %d\n", micco_ts_opt);
	}
	return count;
}

int micco_ts_read(u16 *x,  u16 *y, int *pen_state)
{
	int ret;

	ret = micco_tsi_readxy(x, y, *pen_state);
	return ret;
}


/*************************************************************************
* return ==-2  - pen up decided
* return ==-1  - bad measurement; restored from previous; not for report
* return == 0  - bad measurement; restored from previous
* return == 1  - good measure for reporting
* return == 7  - bad measure not restored but approximated
*/
static int micco_evaluate_point(struct micco_ts_eval* buf, u16* pX, u16* pY, u32 cntr)
{
    int fRet = 1;

  if(buf != NULL) /* NULL: No evaluation required The XY data not modified (transparent) */
  {
    u16 dif_X, dif_Y;
	int xy_range_valid;
	u16 tmpX, tmpY;

	dif_X = dif_Y = 0;

	if(cntr == 0) {
		/* Init on the first measurement evaluation */
		buf->X_pre = INVALID_XY;
		buf->Y_pre = INVALID_XY;
		buf->error_on_measure = 0;
	}

	xy_range_valid = (*pX > VALID_MIN_XY) && (*pX < VALID_MAX_XY) &&
	                 (*pY > VALID_MIN_XY) && (*pY < VALID_MAX_XY);

	if(!xy_range_valid) {
        fRet = -1;  /* the measuremet has failed, retun -1 or 0 */
    }else {
		 /* Still no "previous" data; cannot evaluate, just update "previous" */
		if((buf->X_pre == INVALID_XY) || (buf->Y_pre == INVALID_XY))
			cntr = 0;

		if(cntr != 0){
 			dif_X = abs( *pX - buf->X_pre );
			dif_Y = abs( *pY - buf->Y_pre );
			fRet = (dif_X < VALID_MAX_DIF) && (dif_Y < VALID_MAX_DIF);
			if(fRet==0) {
				if (cntr == 1) { /* prev_0 was not good, drop it, report current_1 */
					buf->X_pre = *pX;
					buf->Y_pre = *pY;
					return fRet;
				}else if((buf->error_on_measure+1) == cntr) {
					/* 2 consequtive errors. Obviously real "jump" => re-sync */
					buf->X_pre = *pX;
					buf->Y_pre = *pY;
					buf->error_on_measure = 0;
					fRet = 1;
				}
			}
		}
	}
	/*------------------------------------------------------------------
	* If pen goes up at time of the measurement the last value is incorrect.
	* To exclude it lets report about PREVIOUS but not current.
	* The last will be dropped out "automatically".
	* If cntr==0 we still haven't the previous so return the current.
	*/
    if (fRet > 0) {
		if(cntr == 0){
			buf->X_pre = *pX;
			buf->Y_pre = *pY;
			fRet = 0;
		}else{
			tmpX = buf->X_pre;
			tmpY = buf->Y_pre;
			buf->X_pre = *pX;
			buf->Y_pre = *pY;
			*pX = tmpX;
			*pY = tmpY;
		}
	}else {
		if(cntr != 0){
			printk(KERN_DEBUG "--->TSI skip BAD point{pre_d_new}  X{%x_%x_%x}  Y{%x_%x_%x} cntr_%u\n",
											 buf->X_pre, dif_X, *pX,  buf->Y_pre, dif_Y, *pY, cntr);
			*pX = buf->X_pre;
			*pY = buf->Y_pre;
			buf->error_on_measure = cntr;
		}
	}

  }/*if buf==NULL: No evaluation required */
  
	return fRet;
}

static int micco_ts_access_check_recovery(int ret_code, int line_no)
{
	if(ret_code != 0)
	{
		printk(KERN_INFO "TSI recovery from MICCO read/write error (%d)\n", line_no);
		micco_td->pen_state = TSI_PEN_UP;
		micco_tsi_enable_pen(1);
		micco_enable_pen_down_irq(1);
	}
	return ret_code;
}

/*
 * The callback function for micco touch.
 * Just turn on the LCD backligtht when PEN DOWN on touch.
 * Leave all the touch X,Y read to a kernel thread.
 *
 * TODO:
 *	1. Move the backlight out of the touch driver.
 *	2. Make backlight align with kernel LCD backlight driver.
 *	3. Make user application handle the touch activity
 *	   turn on the LCD backlight.
 */
void micco_ts_interrupt(unsigned long event)
{
	if (micco_td->use_count > 0) {
		if (TSI_PEN_UP ==  micco_td->pen_state) {
#ifdef CONFIG_TOUCHSCREEN_MICCO_TIMER
			micco_td->pen_state = TSI_PEN_DOWN;
			micco_td->ts_timer->expires = jiffies;
			micco_td->step = 0;
			add_timer(micco_td->ts_timer);
#else
	/* No mutual for the "pen_state"
	 * In occasional worst case the thread executes one additional loop */
			u8 pen_down;
			int ret = micco_read(MICCO_STATUS_A, &pen_down);
			if(	micco_ts_access_check_recovery(ret, __LINE__) ) {
	 			wake_up_interruptible(&micco_td->ts_wait_queue);
				return;
			}
			if (pen_down & 0x40) { /* do for PEN-DOWN only, ignore PEN-UP */
				micco_enable_pen_down_irq(0);
	 			wake_up_interruptible(&micco_td->ts_wait_queue);
			}
#endif
		}/*pen_state*/
	}/*use_count*/
}
EXPORT_SYMBOL(micco_ts_interrupt);


/************************************************************************* 
* A delay needed between micco_tsi_enable_tsi() and micco_ts_read()
* The different HW may requires different configuration in
*    MICCO_TSI_CONTROL_1: TSI_DELAY  b_2..0, TSI_SKIP b_5..3
* This is done in the micco_tsi_poweron(),
* but may be reconfigured here with RECONFIGURE_TSI_DELAY 
* The delay should be set accordingly.
*   TSI_DELAY=3slots TSI_SKIP=3  -> 0x1B
*   TSI_DELAY=1slots TSI_SKIP=3  -> 0x19
*   TSI_DELAY=0slots TSI_SKIP=3  -> 0x18
*/
#define RECONFIGURE_TSI_DELAY   0x19
static unsigned int micco_ts_data_ready_delay(void)
{
	int ret;
	u8 tsi_delay;
	unsigned int delay_us = 999;
#ifdef RECONFIGURE_TSI_DELAY
	tsi_delay = RECONFIGURE_TSI_DELAY;
	ret = micco_write(MICCO_TSI_CONTROL_1, tsi_delay);
#else
	ret = micco_read(MICCO_TSI_CONTROL_1, &tsi_delay);
	tsi_delay &= 0x07;
#endif
	if(ret==0) {
	  switch(tsi_delay)	{
		case 0:		delay_us =  10;		break;
		case 1:		delay_us = 200;		break;
		case 2:		delay_us = 500;		break;
		case 3:		delay_us = 700;		break;
		default:  /*delay_us = 999;*/	break;
	  }
	}
	return(delay_us);
}


#ifdef CONFIG_TOUCHSCREEN_MICCO_TIMER
#error The Touch-Screen handling by Timer IRQ is obsolete. Only thread mode is supported
static void micco_ts_timer_handler(unsigned long d)
{
	struct micco_ts_data *ts_data = (struct micco_ts_data *)d;
	static u16 tem_x = 0xffff;
	static u16 tem_y = 0xffff;
	int ret = 0, state;
	u8 val;
	unsigned long next = 0;
	int cur_step;

	cur_step = ts_data->step; 
_start:
	switch (ts_data->step) {
		case 0:
			tem_x = 0xffff;
			tem_y = 0xffff;
			if (0 == micco_read(MICCO_STATUS_A, &val)) {
				/* pen down */
				if (val & 0x40) {
					if (TSI_PEN_UP ==  ts_data->pen_state) {
						ts_data->pen_state = TSI_PEN_DOWN;
						pr_debug("%s: touch pen down!", __func__);
					}

					/* Enable the auto measure of the TSI.
					 * This will disable pen down interupt automatically.
					 */
					micco_tsi_enable_tsi(1);
					ts_data->step = 1;
					if (micco_ts_opt >= 3) {
						mdelay(1);
					}else {
						next = 1;
						break;
					}
				} else {
					ts_data->step = 3;
				}
			}else {
				next = 1;	
				break;
			}
		case 1:
			micco_read(MICCO_EVENT_C, &val);
			if(val & 0x20) {
				ts_data->step = 2;
			}else {
				pr_debug("%s: micco c polling val = 0x%x",
						__func__, val);
			}
			if (micco_ts_opt >= 1) {
				if(ts_data->step == 1) {
					mdelay(1);
					goto _start;
				}
			}else{
				next = 1;
				break;
			}
		case 2:
			pr_debug("%s: micco c val = 0x%x when TSI ready",
					__func__, val);

			micco_ts_read(&tem_x, &tem_y, &state);
			pr_debug("%s: tem_x:0x%x; tem_y:0x%x; pen_state:0x%x",
					__func__, tem_x, tem_y, state);
		case 3:
			micco_tsi_enable_tsi(0);
			micco_td->step = 4;
			if (micco_ts_opt >= 2) {
				mdelay(1);
			}else {
				next = 1;
				break;
			}
		case 4:
			cur_step = 4;
			ret = micco_read(MICCO_STATUS_A, &val);
			pr_debug("%s: after tsi disable, before pen_down" \
					"enabled status a = 0x%x", __func__, val);
			if (val & 0x40) {
				/* Pen still down. The X, Y data is valid.
				 * Report it.
				 */
				input_report_abs(micco_ts_input_dev,
						ABS_X, tem_x & 0xfff);
				input_report_abs(micco_ts_input_dev,
						ABS_Y, tem_y & 0xfff);
			/*	input_report_abs(micco_ts_input_dev,
						ABS_PRESSURE, 255); */
				input_report_key(micco_ts_input_dev,
						BTN_TOUCH, 1);
				input_sync(micco_ts_input_dev);
				next = 1;
			} else {	/* Pen is up now */
				/* Report a pen up event */
			/*	input_report_abs(micco_ts_input_dev,
			 *			ABS_PRESSURE, 0);
			 *	input_report_abs(micco_ts_input_dev,
			 *			ABS_TOOL_WIDTH, 1); */
				input_report_key(micco_ts_input_dev,
						BTN_TOUCH, 0);
				input_sync(micco_ts_input_dev);
				pr_debug("%s: report pen up", __func__);
				micco_tsi_enable_pen(1);
				ts_data->pen_state = TSI_PEN_UP;
				next = 0xffffffff;
			}
			micco_td->step = 0;
			break;
		default:
			break;		
	}

	if(next != 0xffffffff) {
		ts_data->ts_timer->expires = jiffies+msecs_to_jiffies(next);
		ts_data->ts_timer->data = (unsigned long)ts_data;
		add_timer(ts_data->ts_timer);
	}

	return;
}
#else /*CONFIG_TOUCHSCREEN_MICCO_TIMER*/

/***********************************************************************
 * The touchscreen sample reader thread
 * The thread is activated by Pen-Down interrupt event.
 * Since that it is working with POLLING and the Pen-Down is masked.
 * The micco could give only one - or PEN-DOWN status or TSI measurement.
 * So while polling we need to
 *   micco_tsi_enable_tsi(1); GET-TSI; micco_tsi_enable_tsi(0); check Pen-Up
 */
#define MICCO_ACCESS_CHECK_RECOVERY		if(	micco_ts_access_check_recovery(ret, __LINE__) ) continue;

static int micco_ts_thread(void *d)
{
	struct micco_ts_data *ts_data = d;
	struct micco_ts_eval     eval_buf = {0};
 	struct micco_ts_eval  *p_eval_buf = &eval_buf; /* if NULL no evaluation applied */
	u16 tem_x = 0xffff;
	u16 tem_y = 0xffff;
	int ret = 0, state;
	u32 cntr = 0;
	u8  pen_down;
	unsigned int delay_us = micco_ts_data_ready_delay();
	int pen_up_already_reported = 0;
        struct task_struct *tsk = current;
        struct sched_param param = { .sched_priority = 2 };

	DEFINE_WAIT(ts_wait);
	/* set up thread context */

	ts_data->thread = current;
	daemonize("micco_ts_thread");

        /* improve micco_ts_thread priority */
        sched_setscheduler(tsk, SCHED_FIFO, &param);

	/* init is complete */
	complete(&ts_data->thread_init);

	/* touch reader loop */
	while (1) {
		if (!ts_data->thread)
			break;
		/* if the pen state is up, sleep */
		if (TSI_PEN_UP == ts_data->pen_state) {
			prepare_to_wait(&ts_data->ts_wait_queue, &ts_wait, TASK_INTERRUPTIBLE);

			if (TSI_PEN_UP == ts_data->pen_state)
				schedule();

			finish_wait(&ts_data->ts_wait_queue, &ts_wait);
		}
		try_to_freeze();

		/* Check pen state down or up */
		ret = micco_read(MICCO_STATUS_A, &pen_down);
		MICCO_ACCESS_CHECK_RECOVERY;

		if (pen_down & 0x40) {
			ts_data->pen_state = TSI_PEN_DOWN;
#ifdef CONFIG_PXA3xx_DVFM
			if (!mod_timer(&micco_td->ui_timer, jiffies + micco_td->ui_timeout_val))
				set_dvfm_constraint(micco_td);
#endif
			/* Enable the auto measure of the TSI and wait a while for result ready.
			 * This will disable pen down interupt automatically.
			 */
			ret = micco_tsi_enable_tsi(1);
			MICCO_ACCESS_CHECK_RECOVERY;
			udelay(delay_us);
			//msleep(7);  /* Time to measure and also Polling non blocking delay = 1..2 jiffes (7.8mS+) */

			/* Read X/Y and report */
			micco_ts_read(&tem_x, &tem_y, &state);
			if(0 < micco_evaluate_point(p_eval_buf, &tem_x, &tem_y, cntr)) {
				TOUCHSCREEN_CONSOLE_PRINT_XY; //Debug
				input_report_abs(micco_ts_input_dev, ABS_X, tem_x & 0xfff);
				input_report_abs(micco_ts_input_dev, ABS_Y, tem_y & 0xfff);
			/*	input_report_abs(micco_ts_input_dev, ABS_PRESSURE, 255); */
				input_report_key(micco_ts_input_dev, BTN_TOUCH, 1);
				input_sync(micco_ts_input_dev);
				pen_up_already_reported = 0;
			}
			/* Check PEN-status => need to disable the auto tsi and
			 *  wait a while to read the correct pen_down state
			 * Don't enable the pen down interrupt here.
			 */
			micco_tsi_enable_tsi(0);
			//udelay(delay_us);
			msleep(7);  /* Time to measure and also Polling non blocking delay = 1..2 jiffes (7.8mS+) */
			cntr++;
		}
		else
		{ /* Pen is UP now */
			if((ts_data->pen_state != TSI_PEN_UP) && !pen_up_already_reported) {
			/*	input_report_abs(micco_ts_input_dev, ABS_PRESSURE, 0);
			 *	input_report_abs(micco_ts_input_dev, ABS_TOOL_WIDTH, 1); */
				input_report_key(micco_ts_input_dev,BTN_TOUCH, 0);
				input_sync(micco_ts_input_dev);
				TOUCHSCREEN_CONSOLE_PRINT_PENUP;
				pen_up_already_reported = 1;
			}
			cntr = 0;

			/* pen up/down debounce */
			msleep(7);
			ts_data->pen_state = TSI_PEN_UP;
			ret = micco_enable_pen_down_irq(1);
			MICCO_ACCESS_CHECK_RECOVERY;
		}

	}/*while(1)*/

	/* Always enable the pen down detection before exit from thread */
	ret = micco_tsi_enable_pen(1);
	micco_ts_access_check_recovery(ret, __LINE__);
	complete_and_exit(&ts_data->thread_exit, 0);
	return 0;
}
#endif /*CONFIG_TOUCHSCREEN_MICCO_TIMER*/

#ifdef CONFIG_PXA3xx_DVFM
static ssize_t cleartouch_uitimeout_show(struct device *dev, struct device_attribute *attr, char *buf)
{
	return sprintf(buf, "%d\n", micco_td->ui_timeout_val);
}

static ssize_t cleartouch_uitimeout_store(struct device *dev, struct device_attribute *attr, const char *buf, size_t size)
{
	unsigned int ui_timeout = simple_strtoul(buf, NULL, 10);
	if (ui_timeout > 0 && ui_timeout <= HZ) {
		micco_td->ui_timeout_val = ui_timeout;
		return size;
	} else
		printk(KERN_ERR "Error: range must be at (0,HZ)\n");

	return -EINVAL;
}

static DEVICE_ATTR(uitimeout, S_IRUGO | S_IWUSR, cleartouch_uitimeout_show, cleartouch_uitimeout_store);
#endif

static int micco_ts_open(struct input_dev *idev)
{
	unsigned long flags;
	int ret;

	pr_debug("%s: enter", __func__);

	if (micco_td->suspended) {
		printk(KERN_INFO "touch has been suspended!\n");
		return -1;
	}

	spin_lock_irqsave(&micco_td->ts_lock, flags);
	if (micco_td->use_count++ == 0) {
		spin_unlock_irqrestore(&micco_td->ts_lock, flags);

#ifdef CONFIG_TOUCHSCREEN_MICCO_TIMER
		micco_td->ts_timer = 
			(struct timer_list *)kmalloc(sizeof(struct timer_list), GFP_KERNEL);
		init_timer(micco_td->ts_timer);
		micco_td->ts_timer->function = micco_ts_timer_handler;
		micco_td->ts_timer->data = (unsigned long)micco_td;
#else
 		init_completion(&micco_td->thread_init);
 		ret = kernel_thread(micco_ts_thread, micco_td, 0);
 		if (ret < 0)
 			return ret;
 
 		wait_for_completion(&micco_td->thread_init);
#endif
	} else {
		spin_unlock_irqrestore(&micco_td->ts_lock, flags);
	}

	return 0;
}

/* Kill the touchscreen thread and stop the touch digitiser. */
static void micco_ts_close(struct input_dev *idev)
{
	unsigned long flags;

	pr_debug("%s: enter with use count = %d", __func__,
		micco_td->use_count);

	spin_lock_irqsave(&micco_td->ts_lock, flags);
	if (--micco_td->use_count == 0) {
		spin_unlock_irqrestore(&micco_td->ts_lock, flags);

#ifdef CONFIG_TOUCHSCREEN_MICCO_TIMER
		kfree(micco_td->ts_timer);
#else
		pr_debug("%s: kill thread", __func__);
		/* kill thread */
		if (micco_td->thread) {
			init_completion(&micco_td->thread_exit);
			micco_td->thread = NULL;
			wake_up_interruptible(&micco_td->ts_wait_queue);
			wait_for_completion(&micco_td->thread_exit);
		}
#endif
	} else {
		spin_unlock_irqrestore(&micco_td->ts_lock, flags);
	}
}

static int __init micco_ts_probe(struct platform_device *pdev)
{
	int ret;
	struct proc_dir_entry *micco_ts_proc_entry;	

	micco_td = kzalloc(sizeof(struct micco_ts_data), GFP_KERNEL);
	if (!micco_td) {
		ret = -ENOMEM;
		goto micco_ts_out;
	}
	spin_lock_init(&micco_td->ts_lock);
#ifndef CONFIG_TOUCHSCREEN_MICCO_TIMER
	init_waitqueue_head(&micco_td->ts_wait_queue);
#endif
	micco_td->pen_state = TSI_PEN_UP;
	micco_td->suspended = 0;
	micco_td->use_count = 0;

	/* register input device */
	micco_ts_input_dev = input_allocate_device();
	if (micco_ts_input_dev == NULL) {
		printk(KERN_ERR "%s: failed to allocate input dev\n",__FUNCTION__);
		return -ENOMEM;
	}

	micco_ts_input_dev->name = "micco-ts";

	micco_ts_input_dev->phys = "micco-ts/input0";
	micco_ts_input_dev->dev.parent = &pdev->dev;

	micco_ts_input_dev->open = micco_ts_open;
	micco_ts_input_dev->close = micco_ts_close;

	__set_bit(EV_ABS, micco_ts_input_dev->evbit);
	__set_bit(ABS_X, micco_ts_input_dev->absbit);
	__set_bit(ABS_Y, micco_ts_input_dev->absbit);
/*	__set_bit(ABS_PRESSURE, micco_ts_input_dev->absbit);*/

	__set_bit(EV_SYN, micco_ts_input_dev->evbit);
	__set_bit(EV_KEY, micco_ts_input_dev->evbit);
	__set_bit(BTN_TOUCH, micco_ts_input_dev->keybit);

	input_set_abs_params(micco_ts_input_dev, ABS_X, 0, 480, 0, 0);
	input_set_abs_params(micco_ts_input_dev, ABS_Y, 0, 800, 0, 0);
/*	input_set_abs_params(micco_ts_input_dev, ABS_PRESSURE, 0, 255, 0, 0);
 *	input_set_abs_params(micco_ts_input_dev, ABS_TOOL_WIDTH, 0, 15, 0, 0);*/

	ret = input_register_device(micco_ts_input_dev);
	if (ret) {
		printk(KERN_ERR
			"%s: unabled to register input device, ret = %d\n",
			__FUNCTION__, ret);
		return ret;
	}

#ifdef CONFIG_PXA3xx_DVFM
	micco_td->dvfm_dev_idx = -1;
        micco_td->dvfm_lock.lock = SPIN_LOCK_UNLOCKED;

	if (!dvfm_register("micco-ts", &(micco_td->dvfm_dev_idx))) {
		printk("registered dvfm_dev_idx=%d\n",micco_td->dvfm_dev_idx);
		init_timer(&micco_td->ui_timer);
		micco_td->ui_timer.function = ui_timer_handler;
		micco_td->ui_timer.data = (unsigned long)micco_td;
		micco_td->ui_timeout_val = HZ*5;
		ret = device_create_file( &micco_ts_input_dev->dev, &dev_attr_uitimeout);
		if (ret) {
			printk(KERN_ERR "register ui_timeout interface error.\n");
		}
	}
#endif

	/* The Littleton IRQ come from Micco. So we just implemate
	 * a callback IRQ function for Micco.
	 */
	ret = pmic_callback_register(PMIC_EVENT_TOUCH, micco_ts_interrupt);
	if (ret < 0)
		goto pmic_cb_out;

	/* Enable the touch IRQ here for lcd backlight */
	micco_enable_pen_down_irq(1);
	micco_tsi_poweron();

	micco_ts_proc_entry = create_proc_entry("driver/micco_ts", 0, NULL);
	if (micco_ts_proc_entry) { 
		micco_ts_proc_entry->write_proc = micco_ts_proc_write;
	} 
	return 0;

pmic_cb_out:
	input_unregister_device(micco_ts_input_dev);
micco_ts_out:
	kfree(micco_td);

	return ret;
}

void micco_touch_suspend_handler(struct early_suspend *h)
{
	micco_td->suspended = 1;
	micco_tsi_poweroff();
	micco_enable_pen_down_irq(0);

	if ( del_timer(&micco_td->ui_timer))
	{
		if( likely(micco_td) )
			unset_dvfm_constraint(micco_td);
	}
	return;
}

void micco_touch_resume_handler(struct early_suspend *h)
{
	micco_enable_pen_down_irq(1);
	micco_tsi_poweron();
	micco_td->suspended = 0;

	return;
}

static struct early_suspend micco_touch_suspend = {
	.level = 5,
	.suspend = micco_touch_suspend_handler,
	.resume = micco_touch_resume_handler,
};

static int micco_ts_remove(struct platform_device *pdev)
{
#ifdef CONFIG_PXA3xx_DVFM
	if (micco_td->dvfm_dev_idx >=0) {
		del_timer(&micco_td->ui_timer);
		dvfm_unregister("micco-ts", &micco_td->dvfm_dev_idx);
		micco_td->dvfm_dev_idx = -1;
	}
#endif
	input_unregister_device(micco_ts_input_dev);

	return 0;
}

#if CONFIG_PM
static int micco_ts_resume(struct platform_device *pdev)
{
	micco_enable_pen_down_irq(1);
	micco_tsi_poweron();
	micco_td->suspended = 0;

	return 0;
}

static int micco_ts_suspend(struct platform_device *pdev, pm_message_t state)
{
	micco_td->suspended = 1;
	micco_tsi_poweroff();
	micco_enable_pen_down_irq(0);

	return 0;
}
#else
#define micco_ts_resume NULL
#define micco_ts_suspend NULL
#endif

static struct platform_driver micco_ts_driver = {
	.driver = {
		.name 	= "micco-ts",
	},
	.probe		= micco_ts_probe,
	.remove		= micco_ts_remove,
	.resume 	= micco_ts_resume,
	.suspend 	= micco_ts_suspend,
};

static int __init micco_ts_init(void)
{
	register_early_suspend(&micco_touch_suspend);
	return platform_driver_register(&micco_ts_driver);
}

static void __exit micco_ts_exit(void)
{
	/* We move these codes here because we want to detect the
	 * pen down event even when touch driver is not opened.
	 */
	micco_tsi_poweroff();
	micco_enable_pen_down_irq(0);
	pmic_callback_unregister(PMIC_EVENT_TOUCH, micco_ts_interrupt);

	unregister_early_suspend(&micco_touch_suspend);
	platform_driver_unregister(&micco_ts_driver);
}

module_init(micco_ts_init);
module_exit(micco_ts_exit);

MODULE_AUTHOR("Yin, Fengwei <fengwei.yin@marvell.com>");
MODULE_DESCRIPTION("Littleton Platfrom touch screen driver");
MODULE_LICENSE("GPL");
