/* drivers/input/keyboard/synaptics_i2c_rmi.c
 *
 * Copyright (C) 2007 Google, Inc.
 *
 * This software is licensed under the terms of the GNU General Public
 * License version 2, as published by the Free Software Foundation, and
 * may be copied, distributed, and modified under those terms.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 */

#include <linux/module.h>
#include <linux/delay.h>
#include <linux/earlysuspend.h>
#include <linux/hrtimer.h>
#include <linux/i2c.h>
#include <linux/input.h>
#include <linux/interrupt.h>
#include <linux/io.h>
#include <linux/platform_device.h>
#include <linux/synaptics_i2c_rmi_tm1539.h>
#ifdef CONFIG_PXA3xx_DVFM
#include <mach/dvfm.h>
#include <mach/pxa3xx_dvfm.h>
#endif

static struct workqueue_struct *synaptics_wq;

struct coordinate_map {
	int x_data;
	int y_data;
	int z_data;
	int w_data;
	int down;
	int need_up;
};

struct rmi_function_descriptor {
	uint8_t queryBaseAddr;
	uint8_t commandBaseAddr;
	uint8_t controlBaseAddr;
	uint8_t dataBaseAddr;
	uint8_t interruptSrcCnt;
	uint8_t functionNum;
};

struct rmi_module_info {
	unsigned char rmi_maj_ver;
	unsigned char rmi_min_ver;
	unsigned char mfgid;
	unsigned char properties;
	unsigned char prod_info[2];
	unsigned char data_code[3];
	unsigned short tester_id;
	unsigned short serial_num;
	char prod_id[10];
};

struct synaptics_ts_data {
	struct i2c_client *client;
	struct input_dev *input_dev;
	struct hrtimer timer;
	struct work_struct  work;
	struct synaptics_i2c_rmi_platform_data *pdata;
	struct rmi_function_descriptor fn01descriptor;
	struct rmi_function_descriptor fn11descriptor;
	struct rmi_module_info rmi_info;
	struct coordinate_map finger_data[2];
	struct early_suspend early_suspend;
	int use_irq;
	unsigned short fn11FingersSupported;
	uint32_t last_keystate;
	int (*power)(int on);

#ifdef CONFIG_PXA3xx_DVFM
	int dvfm_dev_idx;
	struct dvfm_lock dvfm_lock;
	struct timer_list ui_timer;
	unsigned int ui_timeout_val;
#endif
};

#ifdef CONFIG_HAS_EARLYSUSPEND
static void synaptics_ts_early_suspend(struct early_suspend *h);
static void synaptics_ts_late_resume(struct early_suspend *h);
#endif

#ifdef CONFIG_PXA3xx_DVFM
static void set_dvfm_constraint(struct synaptics_ts_data *p)
{
	spin_lock_irqsave(&p->dvfm_lock.lock, p->dvfm_lock.flags);

	dvfm_disable_op_name("416M", p->dvfm_dev_idx);
	dvfm_disable_op_name("208M", p->dvfm_dev_idx);
	dvfm_disable_op_name("156M", p->dvfm_dev_idx);

	spin_unlock_irqrestore(&p->dvfm_lock.lock, p->dvfm_lock.flags);
}

static void unset_dvfm_constraint(struct synaptics_ts_data *p)
{
	spin_lock_irqsave(&p->dvfm_lock.lock, p->dvfm_lock.flags);

	dvfm_enable_op_name("156M",p->dvfm_dev_idx);
	dvfm_enable_op_name("208M",p->dvfm_dev_idx);
	dvfm_enable_op_name("416M",p->dvfm_dev_idx);
	spin_unlock_irqrestore(&p->dvfm_lock.lock, p->dvfm_lock.flags);
}

static void ui_timer_handler(unsigned long data)
{
        struct synaptics_ts_data *p = (struct synaptics_ts_data *)data;
       	unset_dvfm_constraint(p);
}
#else
static void set_dvfm_constraint(struct synaptics_ts_data *p) {}
static void unset_dvfm_constraint(struct synaptics_ts_data *p) {}
static void ui_timer_handler(unsigned long data) {}
#endif

#ifdef CONFIG_PXA3xx_DVFM
static ssize_t synaptics_ts_uitimeout_show(struct device *dev, struct device_attribute *attr, char *buf)
{
	struct synaptics_ts_data *ts = dev_get_drvdata(dev);
	return sprintf(buf, "%d\n", ts->ui_timeout_val);
}

static ssize_t synaptics_ts_uitimeout_store(struct device *dev, struct device_attribute *attr, const char *buf, size_t size)
{
	unsigned int ui_timeout = simple_strtoul(buf, NULL, 10);
	struct synaptics_ts_data *ts = dev_get_drvdata(dev);
	if( ui_timeout > 0 && ui_timeout <= HZ )
	{
		ts->ui_timeout_val = ui_timeout;
		return size;
	}else
		printk(KERN_ERR "Error: range must be at (0,HZ)\n");
	
	return -EINVAL;
}

static DEVICE_ATTR(uitimeout, S_IRUGO | S_IWUSR, synaptics_ts_uitimeout_show, synaptics_ts_uitimeout_store);
#endif

static int synaptics_write(struct synaptics_ts_data *ts, void *buf, int buf_sz)
{
	int retries = 10;
	int ret;

	do {
		ret = i2c_master_send(ts->client, (char *)buf, buf_sz);
	} while ((ret < buf_sz) && (--retries > 0));

	if (ret < 0)
		pr_info("%s: Error while trying to write %d bytes\n", __func__,
			buf_sz);
	else if (ret != buf_sz) {
		pr_info("%s: Write %d bytes, expected %d\n", __func__,
			ret, buf_sz);
		ret = -EIO;
	}
	return ret;
}

static int synaptics_set_addr(struct synaptics_ts_data *ts, uint16_t addr)
{
	int ret;
	char txbuf;

	/* Note: addr on the wire is LSB first */
	txbuf = addr & 0xff;
	ret = synaptics_write(ts, &txbuf, sizeof(char));
	if (ret < 0)
		pr_info("%s: Can't send obp addr 0x%4x\n", __func__, addr);

	return ret >= 0 ? 0 : ret;
}

static int synaptics_read(struct synaptics_ts_data *ts, void *buf, int buf_sz)
{
	int retries = 10;
	int ret;

	do {
		memset(buf, 0, buf_sz);
		ret = i2c_master_recv(ts->client, (char *)buf, buf_sz);
	} while ((ret < 0) && (--retries > 0));

	if (ret < 0)
		pr_info("%s: Error while trying to read %d bytes\n", __func__,
			buf_sz);
	else if (ret != buf_sz) {
		pr_info("%s: Read %d bytes, expected %d\n", __func__,
			ret, buf_sz);
		ret = -EIO;
	}

	return ret >= 0 ? 0 : ret;
}

static int synaptics_read_addr(struct synaptics_ts_data *ts, uint16_t addr,
			    void *buf, int buf_sz)
{
	int ret;

	ret = synaptics_set_addr(ts, addr);
	if (ret != 0)
		return ret;

	return synaptics_read(ts, buf, buf_sz);
}

static int synaptics_write_addr(struct synaptics_ts_data *ts, uint16_t addr,
			     void *buf, int buf_sz)
{
	int ret;
	uint8_t write_buf[128];
	char txbuf;

	if (buf_sz + sizeof(uint16_t) > sizeof(write_buf)) {
		pr_err("%s: Buffer too large (%d)\n", __func__, buf_sz);
		return -EINVAL;
	}

	txbuf = addr & 0xff;
	memcpy(write_buf, (void *)&txbuf, sizeof(char));
	memcpy((void *)write_buf + sizeof(char), buf, buf_sz);

	ret = synaptics_write(ts, write_buf, buf_sz + sizeof(char));

	if (ret < 0) {
		pr_err("%s: Could not write %d bytes.\n", __func__, buf_sz);
		return ret;
	}

	return 0;
}

static void synaptics_ts_work_func(struct work_struct *work)
{
	int ret;
	unsigned char interruptstate = 0;
	unsigned char values[2] = {0, 0};
	unsigned char data[12] = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
	int fn11FingerRegisters;
	int finger;
	int touch_x = 0;
	int touch_y = 0;
	int touch_wx = 0;
	int touch_wy = 0;
	uint32_t keystate = 0;
	int i;

	struct synaptics_ts_data *ts = container_of(work, struct synaptics_ts_data, work);

	//printk("synaptics_ts_work_func\n");

	ret = synaptics_read_addr(ts, ts->fn01descriptor.dataBaseAddr + 1, &interruptstate, 1);
	
	//calculate touch&key changes and save the data temporarily.
	fn11FingerRegisters = (ts->fn11FingersSupported + 3) / 4;
	ret = synaptics_read_addr(ts, ts->fn11descriptor.dataBaseAddr, &values, fn11FingerRegisters);
	for (finger = 0; finger < ts->fn11FingersSupported; finger++)
	{
		int reg = finger / 4;
		int fingerShift = (finger % 4) * 2;
		int fingerStatus = (values[reg] >> fingerShift) & 0x3;
		if ((1 == fingerStatus) || (2 == fingerStatus))
		{
			ret = synaptics_read_addr(ts, ts->fn11descriptor.dataBaseAddr + finger * 5 + fn11FingerRegisters, &data, 5);
			touch_x = data[0] << 4;
			touch_x |= (data[2] & 0xf);
			touch_y = data[1] << 4;
			touch_y |= ((data[2] >> 4) & 0xf);
			touch_wx = data[3] & 0xf;
			touch_wy = (data[3] >> 4) & 0xf;
			if (touch_y < ts->pdata->abs_max_y + 1)
			{
				//touch
				ts->finger_data[finger].down = 1;
				ts->finger_data[finger].x_data = touch_x;
				ts->finger_data[finger].y_data = touch_y;
				ts->finger_data[finger].w_data = (touch_wx + touch_wy) / 2;
				ts->finger_data[finger].z_data = data[4];
#if 0//cyj
				if (0 == finger)
					printk("x, y, f: %d  %d  %d\n", touch_x, touch_y, fingerStatus);
				if (1 == finger)
					printk("                                     x, y, f: %d  %d  %d\n", touch_x, touch_y, fingerStatus);
				if ((finger != 0) && (finger != 1))
				{
					while (1)
					{
						printk("1 while(1)\n");
						msleep(500);
					}
				}
#endif
			}
			else
			{
				if (1 == ts->finger_data[finger].down)
				{
					ts->finger_data[finger].need_up = 1;
				}
				else
				{
					//keys
					for (i = 0; i < ts->pdata->key_array.num_keys; ++i)
					{
						struct synaptics_key *key = &ts->pdata->key_array.keys[i];
						if ((key->start_position <= touch_x)
						 && (touch_x <= key->end_position))
						{
							keystate |= (1 << (key->channel));
						}
					}
				}
			}
		}
		else
		{
			if (1 == ts->finger_data[finger].down)
			{
				ts->finger_data[finger].need_up = 1;
			}
		}
	}

	//update key states if needed.
	if (ts->last_keystate != keystate)
	{
		for (i = 0; i < ts->pdata->key_array.num_keys; ++i)
		{
			struct synaptics_key *key = &ts->pdata->key_array.keys[i];
			uint32_t bit = 1 << (key->channel);
			if ((ts->last_keystate & bit) != (keystate & bit))
			{
#if 0//cyj
				printk("x, y: %d  %d\n", touch_x, touch_y);
				printk("key->code = 0x%x, press = 0x%x\n", key->code, keystate & bit);
#endif
				input_report_key(ts->input_dev, key->code, keystate & bit);
			}
		}
		input_sync(ts->input_dev);
		ts->last_keystate = keystate;
	}

	//update touch state if needed.
	for (finger = 0; finger < ts->fn11FingersSupported; finger++)
	{
		if (1 == ts->finger_data[finger].need_up)
		{
			ts->finger_data[finger].z_data = 0;
		}
		if (1 == ts->finger_data[finger].down)
		{
			input_report_abs(ts->input_dev, ABS_MT_TOUCH_MAJOR, ts->finger_data[finger].z_data);
			input_report_abs(ts->input_dev, ABS_MT_WIDTH_MAJOR, ts->finger_data[finger].w_data);
			input_report_abs(ts->input_dev, ABS_MT_POSITION_X, ts->finger_data[finger].x_data);
			input_report_abs(ts->input_dev, ABS_MT_POSITION_Y, ts->finger_data[finger].y_data);
			input_mt_sync(ts->input_dev);
#if 0//cyj
				if (1 == finger)
					printk("                                     ");
				printk("z: %d, w: %d\n", ts->finger_data[finger].z_data, ts->finger_data[finger].w_data);
#endif
		}
		if (1 == ts->finger_data[finger].need_up)
		{
			memset(&ts->finger_data[finger], 0, sizeof(struct coordinate_map));
		}
	}
	input_sync(ts->input_dev);

	if (ts->use_irq)
		enable_irq(ts->client->irq);
}

static enum hrtimer_restart synaptics_ts_timer_func(struct hrtimer *timer)
{
	struct synaptics_ts_data *ts = container_of(timer, struct synaptics_ts_data, timer);
	/* printk("synaptics_ts_timer_func\n"); */

	queue_work(synaptics_wq, &ts->work);

	hrtimer_start(&ts->timer, ktime_set(0, 12500000), HRTIMER_MODE_REL);
	return HRTIMER_NORESTART;
}

static irqreturn_t synaptics_ts_irq_handler(int irq, void *dev_id)
{
	struct synaptics_ts_data *ts = dev_id;

	//printk("synaptics_ts_irq_handler\n");

#ifdef CONFIG_PXA3xx_DVFM
	if (mod_timer(&ts->ui_timer, jiffies + ts->ui_timeout_val))
        { /* timer is already active */}
        else
		set_dvfm_constraint(ts);
#endif

	disable_irq_nosync(ts->client->irq);
	queue_work(synaptics_wq, &ts->work);
	return IRQ_HANDLED;
}

static int synaptics_get_module_info(struct synaptics_ts_data *ts)
{
	int ret = 0;
	int PDT_index;//Page Descriptor Table index
	unsigned char fn11Queries[9];
	unsigned char std_queries[21];
	unsigned char fn11_sensitivity = 0;
	struct rmi_function_descriptor rmi_fd;

	for (PDT_index = 0x00E9; PDT_index > 0x000a; PDT_index -= 6)
	{
		if (!(ret = synaptics_read_addr(ts, PDT_index, &rmi_fd, sizeof(rmi_fd))))
		{
			if (rmi_fd.functionNum)
			{
				switch(rmi_fd.functionNum & 0xff)
				{
					case 0x01:
						printk("  RMI Device Control\n");
						ts->fn01descriptor.queryBaseAddr = rmi_fd.queryBaseAddr;
						ts->fn01descriptor.commandBaseAddr = rmi_fd.commandBaseAddr;
						ts->fn01descriptor.controlBaseAddr = rmi_fd.controlBaseAddr;
						ts->fn01descriptor.dataBaseAddr = rmi_fd.dataBaseAddr;
						ts->fn01descriptor.interruptSrcCnt = rmi_fd.interruptSrcCnt;
						ts->fn01descriptor.functionNum = rmi_fd.functionNum;
						break;
					case 0x11:
						printk("  2-D Sensor\n");
						ret = synaptics_read_addr(ts, rmi_fd.queryBaseAddr, &fn11Queries, sizeof(fn11Queries));
						ts->fn11FingersSupported = (fn11Queries[1] & 0x7) + 1;
						printk("  FingersSupported = 0x%x\n", ts->fn11FingersSupported);
						if (ts->fn11FingersSupported > 2)
						{
							ts->fn11FingersSupported = 2;
						}

						ts->fn11descriptor.queryBaseAddr = rmi_fd.queryBaseAddr;
						ts->fn11descriptor.commandBaseAddr = rmi_fd.commandBaseAddr;
						ts->fn11descriptor.controlBaseAddr = rmi_fd.controlBaseAddr;
						ts->fn11descriptor.dataBaseAddr = rmi_fd.dataBaseAddr;
						ts->fn11descriptor.interruptSrcCnt = rmi_fd.interruptSrcCnt;
						ts->fn11descriptor.functionNum = rmi_fd.functionNum;
						ret = synaptics_read_addr(ts, ts->fn11descriptor.controlBaseAddr, std_queries, 14);
						printk("Sensor Max X: %d\n", ((std_queries[6]) | ((std_queries[7] & 0xf) << 8)));
						printk("Sensor Max Y: %d\n", ((std_queries[8]) | ((std_queries[9] & 0xf) << 8)));
						printk("reprot mode: 0x%x\n", (std_queries[0]));
						printk("palm reject: 0x%x\n", (std_queries[1]));
						printk("delta x: 0x%x\n", (std_queries[2]));
						printk("delta y: 0x%x\n", (std_queries[3]));
						printk("velocity: 0x%x\n", (std_queries[4]));
						printk("acceleration: 0x%x\n", (std_queries[5]));
						break;
					case 0x08:
						printk("  BIST\n");
						break;
					case 0x19:
						printk("  Buttons\n");
						break;
					case 0x30:
						printk("  LED/GPIOs\n");
						break;
					case 0x32:
						printk("  Timer\n");
						break;
					case 0x34:
						printk("  Flash Update\n");
						break;
					default:
						printk("  Unknown Function %02x\n", rmi_fd.functionNum & 0xff);
						break;
				}
			}
			else
			{
				printk("A zero in the function number\n");
				/*A zero in the function number signals the end of the PDT*/
				break;
			}
		}
		else
		{
			printk("failed to read next PDT entry\n");
			/*failed to read next PDT entry - end PDT scan - this may result in an imcomplete
			set of recognized functions - should probably return an error but the driver
			may still be viable for diagnostics and debugging so let's let it continue.*/
			break;
		}
	}

	ret = synaptics_read_addr(ts, ts->fn01descriptor.queryBaseAddr, std_queries, sizeof(std_queries));

	ts->rmi_info.rmi_maj_ver = 4;
	ts->rmi_info.rmi_min_ver = 0;
	ts->rmi_info.mfgid = std_queries[0];
	ts->rmi_info.properties = std_queries[1];
	ts->rmi_info.prod_info[0] = std_queries[2];
	ts->rmi_info.prod_info[1] = std_queries[3];
	ts->rmi_info.data_code[0] = std_queries[4] & 0x1f;
	ts->rmi_info.data_code[1] = std_queries[5] & 0x0f;
	ts->rmi_info.data_code[2] = std_queries[6] & 0x1f;
	ts->rmi_info.tester_id = ((std_queries[7] & 0x7f) << 8) | (std_queries[8] & 0x7f);
	ts->rmi_info.serial_num = ((std_queries[9] & 0x7f) << 8) | (std_queries[10] & 0x7f);

	printk("Properties: %x\n", ts->rmi_info.properties);
	printk("Product Info: %x %x\n", ts->rmi_info.prod_info[0], ts->rmi_info.prod_info[1]);
	printk("Date Code, Year: %d Month: %d Day: %d\n", ts->rmi_info.data_code[0], ts->rmi_info.data_code[1], ts->rmi_info.data_code[2]);
	printk("Tester ID: %d\n", ts->rmi_info.tester_id);
	printk("Serial Number: %x\n", ts->rmi_info.serial_num);

	if (1 == ts->rmi_info.mfgid)
		printk("(synaptics)\n");
	else
		ret = -ENODEV;

	if ((1 << 6) & fn11Queries[1])
	{
		ret = synaptics_read_addr(ts, 0x005E, &fn11_sensitivity, 1);
		fn11_sensitivity = ts->pdata->touch_sensitivity;
		ret = synaptics_write_addr(ts, 0x005E, (void *)&fn11_sensitivity, 1);
		msleep(20);
		ret = synaptics_read_addr(ts, 0x005E, &fn11_sensitivity, 1);
		printk("after set fn11_sensitivity = 0x%x\n", fn11_sensitivity);
	}

	return ret;
}

static int synaptics_ts_probe(
	struct i2c_client *client, const struct i2c_device_id *id)
{
	struct synaptics_ts_data *ts;
	struct synaptics_i2c_rmi_platform_data *pdata = client->dev.platform_data;
	int ret = 0;
	unsigned char interruptstate = 0;

	int i;

	if (!i2c_check_functionality(client->adapter, I2C_FUNC_I2C)) {
		printk(KERN_ERR "synaptics_ts_probe: need I2C_FUNC_I2C\n");
		ret = -ENODEV;
		goto err_check_functionality_failed;
	}

	ts = kzalloc(sizeof(*ts), GFP_KERNEL);
	if (ts == NULL) {
		ret = -ENOMEM;
		goto err_alloc_data_failed;
	}
	INIT_WORK(&ts->work, synaptics_ts_work_func);

	ts->pdata = pdata;
	ts->client = client;
	i2c_set_clientdata(client, ts);
	if (pdata)
		ts->power = pdata->power;
	if (ts->power) {
		ret = ts->power(1);
		if (ret < 0) {
			printk(KERN_ERR "synaptics_ts_probe power on failed\n");
			goto err_power_failed;
		}
	}

	ret = synaptics_get_module_info(ts);
	if (ret < 0) {
		printk(KERN_ERR "synaptics_ts_probe get module info failed\n");
		goto err_power_failed;
	}

	ts->input_dev = input_allocate_device();
	if (ts->input_dev == NULL) {
		ret = -ENOMEM;
		printk(KERN_ERR "synaptics_ts_probe: Failed to allocate input device\n");
		goto err_input_dev_alloc_failed;
	}
	ts->input_dev->name = "synaptics-rmi-touchscreen";
	input_set_drvdata(ts->input_dev, ts);

	set_bit(EV_SYN, ts->input_dev->evbit);
	for (i = 0; i < pdata->key_array.num_keys; ++i)
		input_set_capability(ts->input_dev, EV_KEY, pdata->key_array.keys[i].code);

	set_bit(EV_ABS, ts->input_dev->evbit);
	input_set_capability(ts->input_dev, EV_KEY, BTN_TOUCH);
	input_set_capability(ts->input_dev, EV_KEY, BTN_2);

	input_set_abs_params(ts->input_dev, ABS_MT_POSITION_X, pdata->abs_min_x, pdata->abs_max_x, pdata->fuzz_x, 0);
	input_set_abs_params(ts->input_dev, ABS_MT_POSITION_Y, pdata->abs_min_y, pdata->abs_max_y, pdata->fuzz_y, 0);
	input_set_abs_params(ts->input_dev, ABS_MT_TOUCH_MAJOR, pdata->abs_min_p, pdata->abs_max_p, pdata->fuzz_p, 0);
	input_set_abs_params(ts->input_dev, ABS_MT_WIDTH_MAJOR, pdata->abs_min_w, pdata->abs_max_w, pdata->fuzz_w, 0);

	memset(&ts->finger_data[0], 0, (sizeof(struct coordinate_map) * (ts->fn11FingersSupported)));

	ret = input_register_device(ts->input_dev);
	if (ret) {
		printk(KERN_ERR "synaptics_ts_probe: Unable to register %s input device\n", ts->input_dev->name);
		goto err_input_register_device_failed;
	}

	if (client->irq) {
		ret = request_irq(client->irq, synaptics_ts_irq_handler, IRQF_DISABLED | pdata->irqflags, "synaptics_ts_int", ts);
		if (ret != 0) {
			printk(KERN_ERR "synaptics_ts_probe: request_irq (%d) failed\n", client->irq);
			ts->use_irq = 0;
		}
		else {
			ts->use_irq = 1;
		}
	}
	if (!ts->use_irq) {
		hrtimer_init(&ts->timer, CLOCK_MONOTONIC, HRTIMER_MODE_REL);
		ts->timer.function = synaptics_ts_timer_func;
		hrtimer_start(&ts->timer, ktime_set(1, 0), HRTIMER_MODE_REL);
	}

#if defined(CONFIG_PXA3xx_DVFM)
	ts->dvfm_dev_idx = -1;
        ts->dvfm_lock.lock = SPIN_LOCK_UNLOCKED;

	if(!dvfm_register((char *)ts->input_dev->name, &ts->dvfm_dev_idx))
	{
		printk("registered dvfm_dev_idx=%d\n",ts->dvfm_dev_idx);
		init_timer(&ts->ui_timer);
		ts->ui_timer.function = ui_timer_handler;
		ts->ui_timer.data = (unsigned long)ts;
		ts->ui_timeout_val = 5*HZ;
		ret = device_create_file( &ts->input_dev->dev, &dev_attr_uitimeout);
		if (ret) {
			printk(KERN_ERR "register ui_timeout interface error.\n");
		}
	}
#endif

#ifdef CONFIG_HAS_EARLYSUSPEND
	ts->early_suspend.level = EARLY_SUSPEND_LEVEL_BLANK_SCREEN + 1;
	ts->early_suspend.suspend = synaptics_ts_early_suspend;
	ts->early_suspend.resume = synaptics_ts_late_resume;
	register_early_suspend(&ts->early_suspend);
#endif

	//read interrupt state to enable next interrupt
	ret = synaptics_read_addr(ts, ts->fn01descriptor.dataBaseAddr + 1, &interruptstate, 1);
	printk("interruptstate = 0x%x\n", interruptstate);

	printk(KERN_INFO "synaptics_ts_probe: Start touchscreen %s in %s mode\n", ts->input_dev->name, ts->use_irq ? "interrupt" : "polling");
	return 0;

err_input_register_device_failed:
	input_free_device(ts->input_dev);

err_input_dev_alloc_failed:
err_power_failed:
	kfree(ts);
err_alloc_data_failed:
err_check_functionality_failed:
	return ret;
}

static int synaptics_ts_remove(struct i2c_client *client)
{
	struct synaptics_ts_data *ts = i2c_get_clientdata(client);

#ifdef CONFIG_PXA3xx_DVFM
	if( ts->dvfm_dev_idx >=0 )
	{
		del_timer(&ts->ui_timer);
		dvfm_unregister((char *)ts->input_dev->name, &ts->dvfm_dev_idx);
		ts->dvfm_dev_idx = -1;
	}
#endif

	unregister_early_suspend(&ts->early_suspend);
	if (ts->use_irq)
		free_irq(client->irq, ts);
	else
		hrtimer_cancel(&ts->timer);
	input_unregister_device(ts->input_dev);

	kfree(ts);
	return 0;
}

static int synaptics_ts_suspend(struct i2c_client *client, pm_message_t mesg)
{
	int ret;
	unsigned char command;
	struct synaptics_ts_data *ts = i2c_get_clientdata(client);

	if (ts->use_irq)
		disable_irq(client->irq);
	else
		hrtimer_cancel(&ts->timer);
	ret = cancel_work_sync(&ts->work);
	if (ret && ts->use_irq) /*if work was pending disable-count is now 2 */
		enable_irq(client->irq);
	
	command = 0x01;
	ret = synaptics_write_addr(ts, ts->fn01descriptor.controlBaseAddr, (void *)&command, 1);
	if (ret < 0)
	{
		printk(KERN_ERR "synaptics_ts_suspend: synaptics_write_addr failed\n");
	}

	if (ts->power) {
		ret = ts->power(0);
		if (ret < 0)
			printk(KERN_ERR "synaptics_ts_suspend power off failed\n");
	}
	return 0;
}

static int synaptics_ts_resume(struct i2c_client *client)
{
	int ret;
	unsigned char command;
	struct synaptics_ts_data *ts = i2c_get_clientdata(client);

	if (ts->power) {
		ret = ts->power(1);
		if (ret < 0)
			printk(KERN_ERR "synaptics_ts_resume power on failed\n");
	}

	if (ts->use_irq)
		enable_irq(client->irq);
	
	command = 0x00;
	ret = synaptics_write_addr(ts, ts->fn01descriptor.controlBaseAddr, (void *)&command, 1);
	if (ret < 0)
	{
		printk(KERN_ERR "synaptics_ts_resume: synaptics_write_addr failed\n");
	}

	if (!ts->use_irq)
		hrtimer_start(&ts->timer, ktime_set(1, 0), HRTIMER_MODE_REL);

	return 0;
}

#ifdef CONFIG_HAS_EARLYSUSPEND
static void synaptics_ts_early_suspend(struct early_suspend *h)
{
	struct synaptics_ts_data *ts;
	ts = container_of(h, struct synaptics_ts_data, early_suspend);
	synaptics_ts_suspend(ts->client, PMSG_SUSPEND);
}

static void synaptics_ts_late_resume(struct early_suspend *h)
{
	struct synaptics_ts_data *ts;
	ts = container_of(h, struct synaptics_ts_data, early_suspend);
	synaptics_ts_resume(ts->client);
}
#endif

static const struct i2c_device_id synaptics_ts_id[] = {
	{ SYNAPTICS_I2C_RMI_NAME, 0 },
	{ }
};

static struct i2c_driver synaptics_ts_driver = {
	.probe		= synaptics_ts_probe,
	.remove		= synaptics_ts_remove,
#ifndef CONFIG_HAS_EARLYSUSPEND
	.suspend	= synaptics_ts_suspend,
	.resume		= synaptics_ts_resume,
#endif
	.id_table	= synaptics_ts_id,
	.driver = {
		.name	= SYNAPTICS_I2C_RMI_NAME,
	},
};

static int __devinit synaptics_ts_init(void)
{
	synaptics_wq = create_singlethread_workqueue("synaptics_wq");
	if (!synaptics_wq)
		return -ENOMEM;
	return i2c_add_driver(&synaptics_ts_driver);
}

static void __exit synaptics_ts_exit(void)
{
	i2c_del_driver(&synaptics_ts_driver);
	if (synaptics_wq)
		destroy_workqueue(synaptics_wq);
}

module_init(synaptics_ts_init);
module_exit(synaptics_ts_exit);

MODULE_DESCRIPTION("Synaptics Touchscreen Driver TM1539");
MODULE_LICENSE("GPL");
