/*
 *
 *
 */
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/module.h>
#include <linux/bitops.h>
#include <linux/types.h>
#include <linux/errno.h>
#include <linux/sched.h>
#include <linux/interrupt.h>
#include <linux/slab.h>
#include <linux/uaccess.h>
#include <linux/delay.h>
#include "ts0710.h"
#include "mux_buff.h"

#define SET_PF(ctr) 		((ctr) | (1 << 4))
#define CLR_PF(ctr) 		((ctr) & 0xef)
#define GET_PF(ctr) 		(((ctr) >> 4) & 0x1)

#define SHORT_PAYLOAD_SIZE 	127

#define EA 			1
#define FCS_SIZE 		1
#define FLAG_SIZE 		2

#define TS0710_MAX_HDR_SIZE 	5
#define DEF_TS0710_MTU 		(1024 * 2)

#define TS0710_BASIC_FLAG 	0xF9

/* the control field */
#define SABM 			0x2f
#define SABM_SIZE 		4
#define UA 			0x63
#define UA_SIZE 		4
#define DM 			0x0f
#define DISC 			0x43
#define UIH 			0xef

/* the type field in a multiplexer command packet */
#define TEST 			0x8
#define FCON 			0x28
#define FCOFF 			0x18
#define MSC 			0x38
#define RPN 			0x24
#define RLS 			0x14
#define PN 			0x20
#define NSC 			0x4

/* V.24 modem control signals */
#define FC 			0x2
#define RTC 			0x4
#define RTR 			0x8
#define IC 			0x40
#define DV 			0x80

#define CTRL_CHAN 		0	/* The control channel */
#define MCC_CR 			0x2
#define MCC_CMD 		1	/* Multiplexer command cr */
#define MCC_RSP 		0	/* Multiplexer response cr */

#ifdef __LITTLE_ENDIAN_BITFIELD
struct address_field {
	u8 ea:1;
	u8 cr:1;
	u8 d:1;
	u8 server_chn:5;
} __attribute__ ((packed));

struct short_length {
	u8 ea:1;
	u8 len:7;
} __attribute__ ((packed));

struct long_length {
	u8 ea:1;
	u8 l_len:7;
	u8 h_len;
} __attribute__ ((packed));

struct short_frame_head {
	struct address_field addr;
	u8 control;
	struct short_length length;
} __attribute__ ((packed));

struct short_frame {
	struct short_frame_head h;
	u8 data[0];
} __attribute__ ((packed));

struct long_frame_head {
	struct address_field addr;
	u8 control;
	struct long_length length;
	u8 data[0];
} __attribute__ ((packed));

struct long_frame {
	struct long_frame_head h;
	u8 data[0];
} __attribute__ ((packed));

/* Typedefinitions for structures used for the multiplexer commands */
struct mcc_type {
	u8 ea:1;
	u8 cr:1;
	u8 type:6;
} __attribute__ ((packed));

struct mcc_short_frame_head {
	struct mcc_type type;
	struct short_length length;
	u8 value[0];
} __attribute__ ((packed));

struct mcc_short_frame {
	struct mcc_short_frame_head h;
	u8 value[0];
} __attribute__ ((packed));

struct mcc_long_frame_head {
	struct mcc_type type;
	struct long_length length;
	u8 value[0];
} __attribute__ ((packed));

struct mcc_long_frame {
	struct mcc_long_frame_head h;
	u8 value[0];
} __attribute__ ((packed));

/* MSC-command */
struct v24_sigs {
	u8 ea:1;
	u8 fc:1;
	u8 rtc:1;
	u8 rtr:1;
	u8 reserved:2;
	u8 ic:1;
	u8 dv:1;
} __attribute__ ((packed));

struct brk_sigs {
	u8 ea:1;
	u8 b1:1;
	u8 b2:1;
	u8 b3:1;
	u8 len:4;
} __attribute__ ((packed));

struct msc_msg_data {
	struct address_field dlci;
	u8 v24_sigs;
} __attribute__ ((packed));

struct pn_msg_data {
	u8 dlci:6;
	u8 res1:2;

	u8 frame_type:4;
	u8 credit_flow:4;

	u8 prior:6;
	u8 res2:2;

	u8 ack_timer;
	u8 frame_sizel;
	u8 frame_sizeh;
	u8 max_nbrof_retrans;
	u8 credits;
} __attribute__ ((packed));
#else
#error Only littel-endianess supported now!
#endif

#define TS0710_FRAME_SIZE(len)						\
	((len) > SHORT_PAYLOAD_SIZE ?					\
	 (len) + FLAG_SIZE + sizeof(struct long_frame) + FCS_SIZE :	\
	 (len) + FLAG_SIZE + sizeof(struct short_frame) + FCS_SIZE)

#define TS0710_MCC_FRAME_SIZE(len) \
	TS0710_FRAME_SIZE((len) + sizeof(struct mcc_short_frame))

#define MAX_FRAME_SIZE TS0710_MCC_FRAME_SIZE(sizeof(struct pn_msg_data))

enum {
	REJECTED = 0,
	DISCONNECTED,
	CONNECTING,
	NEGOTIATING,
	CONNECTED,
	DISCONNECTING,
	FLOW_STOPPED,
};

struct dlci_struct {
	u8 state;
	u16 mtu;
	int clients;
	int dlci;
	int type;
#define CMD_CHN		1
#define DATA_CHN	2
	struct mutex lock;
	wait_queue_head_t open_wait;
	wait_queue_head_t ioctl_wait;
	wait_queue_head_t close_wait;
	wait_queue_head_t write_wait;
	spinlock_t 	io_lock;
	struct list_head rx_pool;
	struct delayed_work	rx_work;
	unsigned long flag;
#define RX_RUNNING	1
#define RX_PENDING	2
};

static struct dlci_struct dlcis[TS0710_MAX_CHN];

#define EREJECTED 901		/* link connection request is rejected */

#define TS0710MUX_SEND_BUF_SIZE (TS0710_FRAME_SIZE(DEF_TS0710_MTU))

#define TS0710MUX_SERIAL_BUF_SIZE 2048

enum recv_state {
	RECV_STATE_IDLE = 0,
	RECV_STATE_ADDR,
	RECV_STATE_CONTROL,
	RECV_STATE_LEN,
	RECV_STATE_LEN2,
	RECV_STATE_DATA,
	RECV_STATE_FCS,
	RECV_STATE_END,
};

static u8 crctable[256];

static struct workqueue_struct *wqueue;

static LIST_HEAD(send_list);
static DEFINE_SPINLOCK(send_lock);
static void process_send_buf(unsigned long);
static DECLARE_TASKLET(send_tsk, process_send_buf, 0);

#define TS0710_DEBUG

#ifdef TS0710_DEBUG
#define DEBUG_SEND	(30)
#define DEBUG_RECV	(29)
#define DEBUG_RESEND	(28)
static unsigned long debug_mask = 0;
module_param_named(debug_mask, debug_mask, ulong, S_IRUGO | S_IWUSR);

static void mux_dump_data(const unsigned char *buf, int count)
{
	int i;

	if (count <= 0)
		return;

	pr_info("	(%d): ", count);
	for (i = 0; i < count; i++) {
		printk(" %02X", buf[i]);
	}
	printk("\n");
}
#endif

static DEFINE_SPINLOCK(free_list_lock);
static struct list_head mbuf_free_list;
#define TS0710_MAX_MBUF_LIST	512
static atomic_t mbuf_count;

static struct mux_buff *mbuf_get_free_list(void)
{
	struct mux_buff *mbuf;
	unsigned long flags;

	spin_lock_irqsave(&free_list_lock, flags);
	if (list_empty(&mbuf_free_list)) {
		spin_unlock_irqrestore(&free_list_lock, flags);
		if (atomic_dec_return(&mbuf_count) < 0) {
			return 0;
		}
		if (!(mbuf = mux_buff_alloc(GFP_KERNEL))) {
			pr_info("%s: alloc mbuf failed\n", __func__);
			return 0;
		}
		if (!(mbuf->buf = kzalloc(DEF_TS0710_MTU, GFP_ATOMIC))) {
			mux_buff_free(mbuf);
			pr_info("%s: alloc mbuf head failed\n", __func__);
			return 0;
		}
		pr_info("%s: count %d\n", __func__, atomic_read(&mbuf_count));
		mbuf->d_p = mbuf->buf;
		mbuf->len = DEF_TS0710_MTU;
		mbuf->d_len = 0;
	} else {
		mbuf = list_first_entry(&mbuf_free_list, struct mux_buff, list);
		list_del_init(&mbuf->list);
	}
	spin_unlock_irqrestore(&free_list_lock, flags);
	return mbuf;
}

static void mbuf_put_free_list(struct mux_buff *mbuf)
{
	unsigned long flags;

	if (!mbuf)
		return;

	mbuf->d_p = mbuf->buf;
	mbuf->d_len = 0;
	mbuf->dlci = 0;
	mbuf->type = CMD_CHN;
	memset(mbuf->buf, 0, mbuf->len);

	spin_lock_irqsave(&free_list_lock, flags);
	list_add(&mbuf->list, &mbuf_free_list);
	spin_unlock_irqrestore(&free_list_lock, flags);
}

static inline int mcc_is_cmd(u8 type)
{
	return type & MCC_CR;
}

static inline int mcc_is_rsp(u8 type)
{
	return !(type & MCC_CR);
}


static inline int addr2dlci(u8 addr)
{
	return (addr >> 2) & 0x3f;
}

static int valid_dlci(u8 dlci)
{
	if (((dlci < TS0710_MAX_CHN) && (dlci > 0)))
		return 1;
	else
		return 0;
}

static void crc_create_table(u8 table[])
{
	int i, j;

	u8 data;
	u8 code_word = 0xe0;
	u8 sr = 0;

	for (j = 0; j < 256; j++) {
		data = (u8) j;

		for (i = 0; i < 8; i++) {
			if ((data & 0x1) ^ (sr & 0x1)) {
				sr >>= 1;
				sr ^= code_word;
			} else {
				sr >>= 1;
			}

			data >>= 1;
			sr &= 0xff;
		}

		table[j] = sr;
		sr = 0;
	}
}

static u8 crc_start(void)
{
	return 0xff;
}

static u8 crc_calc(u8 fcs, u8 c)
{
	return crctable[fcs ^ c];
}

static u8 crc_end(u8 fcs)
{
	return 0xff - fcs;
}

#define CRC_VALID 0xcf
static int crc_check(u8 fcs)
{
#if 0
	return fcs == CRC_VALID;
#endif
}

static u8 crc_data(u8 *data, int length)
{
	u8 fcs = crc_start();

	while (length--)
		fcs = crc_calc(fcs, *data++);

	return crc_end(fcs);
}

static void mbuf_put_tail(spinlock_t *lock, struct list_head *head,
		struct mux_buff *mbuf)
{
	unsigned long flags;

	spin_lock_irqsave(lock, flags);
	list_add_tail(&mbuf->list, head);
	spin_unlock_irqrestore(lock, flags);
}

/* add a request to the tail of a list */
static void mbuf_put(spinlock_t *lock, struct list_head *head,
		struct mux_buff *mbuf)
{
	unsigned long flags;

	spin_lock_irqsave(lock, flags);
	list_add(&mbuf->list, head);
	spin_unlock_irqrestore(lock, flags);
}

/* remove a request from the head of a list */
static struct mux_buff *mbuf_get(spinlock_t *lock, struct list_head *head)
{
	unsigned long flags;
	struct mux_buff *mbuf;

	spin_lock_irqsave(lock, flags);
	if (list_empty(head)) {
		mbuf = 0;
	} else {
		mbuf = list_first_entry(head, struct mux_buff, list);
		list_del_init(&mbuf->list);
	}
	spin_unlock_irqrestore(lock, flags);
	return mbuf;
}

static void set_pkt_header(u8 *data, int len, int addr_ea,
				int addr_cr, int addr_dlci,
				int control)
{
	struct short_frame *pkt = (struct short_frame *)(data + 1);

	pkt->h.addr.ea = addr_ea;
	pkt->h.addr.cr = addr_cr;
	pkt->h.addr.d = addr_dlci & 0x1;
	pkt->h.addr.server_chn = addr_dlci >> 1;
	pkt->h.control = control;

	if ((len) > SHORT_PAYLOAD_SIZE) {
		struct long_frame *long_pkt = (struct long_frame *)(data + 1);
		long_pkt->h.length.ea = 0;
		long_pkt->h.length.l_len = len & 0x7F;
		long_pkt->h.length.h_len = (len >> 7) & 0xFF;
	} else {
		pkt->h.length.ea = 1;
		pkt->h.length.len = len;
	}
}

static void *pkt_data(u8 *data)
{
	struct short_frame *pkt = (struct short_frame *)(data + 1);
	if (pkt->h.length.ea == 1)
		return pkt->data;
	else
		return pkt->data + 1;
}

int (*ts0710_tx)(u8 *buf, int len);

static int send_pkt(struct mux_buff *mbuf)
{
	u8 *d, *data = mbuf->buf;
	int len, header_len;
	struct short_frame *pkt = (struct short_frame *)(data + 1);
	

	if (pkt->h.length.ea == 1) {
		len = pkt->h.length.len;
		d = pkt->data;
		header_len = sizeof(*pkt);
	} else {
		struct long_frame *long_pkt = (struct long_frame *)(data + 1);
		len = (long_pkt->h.length.h_len << 7) |
			long_pkt->h.length.l_len;
		d = pkt->data+1;
		header_len = sizeof(*long_pkt);
	}

	data[0] = TS0710_BASIC_FLAG;
	d[len] = crc_data(data + 1, header_len);
	d[len+1] = TS0710_BASIC_FLAG;

	mbuf->d_len = TS0710_FRAME_SIZE(len);
	mbuf->d_p = mbuf->buf;

	mbuf_put_tail(&send_lock, &send_list, mbuf);
	tasklet_schedule(&send_tsk);

	return 0;
}

static int send_cmd(u8 dlci, u8 cmd)
{
	struct mux_buff *mbuf;

	if (!(mbuf = mbuf_get_free_list())) {
		pr_err("%s: mbuf is NULL\n", __func__);
		return 0;
	}
	mbuf->dlci = dlci;
	set_pkt_header(mbuf->buf, 0, 1, MCC_RSP, dlci, SET_PF(cmd));

	return send_pkt(mbuf);
}

static int send_ua(u8 dlci)
{
	return send_cmd(dlci, UA);
}

static int send_dm(u8 dlci)
{
	return send_cmd(dlci, DM);
}

static int send_sabm(u8 dlci)
{
	return send_cmd(dlci, SABM);
}

static int send_disc(u8 dlci)
{
	return send_cmd(dlci, DISC);
}

static int send_uih(u8 dlci, const u8 *data, int len)
{
	struct mux_buff *mbuf;
	struct dlci_struct *d = &dlcis[dlci];

	if (!(mbuf = mbuf_get_free_list())) {
		pr_err("%s: mbuf is NULL\n", __func__);
		return -EAGAIN;
	}
	mbuf->dlci = dlci;
	mbuf->type = d->type;
	set_pkt_header(mbuf->buf, len, 1, MCC_CMD, dlci, CLR_PF(UIH));
	memcpy(pkt_data(mbuf->buf), data, len);
	
	return send_pkt(mbuf);
}

static void set_mcc_header(u8 *frame, int len, int cr, int cmd)
{
	struct mcc_short_frame *mcc_pkt;
	set_pkt_header(frame, sizeof(struct mcc_short_frame) + len,
			      1, MCC_CMD, CTRL_CHAN, CLR_PF(UIH));

	mcc_pkt = pkt_data(frame);
	mcc_pkt->h.type.ea = EA;
	mcc_pkt->h.type.cr = cr;
	mcc_pkt->h.type.type = cmd;
	mcc_pkt->h.length.ea = EA;
	mcc_pkt->h.length.len = len;
}

static void *mcc_data(u8 *frame)
{
	return ((struct mcc_short_frame *)pkt_data(frame))->value;
}

#if 0
static int send_fcon(int dlci, int cr)
{
	struct mux_buff *mbuf;

	if (!(mbuf = mbuf_get_free_list())) {
		pr_err("%s: mbuf is NULL\n", __func__);
		return -EAGAIN;
	}
	mbuf->dlci = dlci;

	set_mcc_header(mbuf->buf, 0, cr, FCON);

	return send_pkt(mbuf);
}

static int send_fcoff(int dlci, int cr)
{
	struct mux_buff *mbuf;

	if (!(mbuf = mbuf_get_free_list())) {
		pr_err("%s: mbuf is NULL\n", __func__);
		return -EAGAIN;
	}
	mbuf->dlci = dlci;

	set_mcc_header(frame, 0, cr, FCOFF);
	return send_pkt(mbuf);
}
#endif

static int send_pn(u8 prior, int frame_size,
		u8 credit_flow, u8 credits, u8 dlci, u8 cr)
{
	struct pn_msg_data *pn;
	struct mux_buff *mbuf;

	if (!(mbuf = mbuf_get_free_list())) {
		pr_err("%s: mbuf is NULL\n", __func__);
		return 0;
	}
	set_mcc_header(mbuf->buf, sizeof(struct pn_msg_data), cr, PN);

	pn = mcc_data(mbuf->buf);
	pn->res1 = 0;
	pn->res2 = 0;
	pn->dlci = dlci;
	pn->frame_type = 0;
	pn->credit_flow = credit_flow;
	pn->prior = prior;
	pn->ack_timer = 0;
	pn->frame_sizel = frame_size & 0xff;
	pn->frame_sizeh = frame_size >> 8;
	pn->credits = credits;
	pn->max_nbrof_retrans = 0;

	return send_pkt(mbuf);
}

static int send_nsc(u8 type, int cr)
{
	struct mcc_type *t;
	struct mux_buff *mbuf;

	if (!(mbuf = mbuf_get_free_list())) {
		pr_err("%s: mbuf is NULL\n", __func__);
		return 0;
	}

	set_mcc_header(mbuf->buf, sizeof(struct mcc_type), cr, NSC);
	t = mcc_data(mbuf->buf);
	t->ea = 1;
	t->cr = mcc_is_cmd(type);
	t->type = type >> 2;

	return send_pkt(mbuf);
}

static int send_msc(u8 dlci, int cr, u8 value)
{
	struct msc_msg_data *msc;
	struct mux_buff *mbuf;

	if (!(mbuf = mbuf_get_free_list())) {
		pr_err("%s: mbuf is NULL\n", __func__);
		return 0;
	}
	set_mcc_header(mbuf->buf, sizeof(struct msc_msg_data), cr, MSC);
	msc = mcc_data(mbuf->buf);
	msc->dlci.ea = 1;
	msc->dlci.cr = 1;
	msc->dlci.d = dlci & 1;
	msc->dlci.server_chn = (dlci >> 1) & 0x1f;
	msc->v24_sigs = value;
	pr_info("%s: send %s to dlci %d %02x\n", __func__, cr ? "MSC":"MSR", dlci, value);

	return send_pkt(mbuf);
}

static void handle_msc(u8 type, u8 *buf, int len)
{
	u8 dlci = *buf >> 2;
	u8 v24_sigs = *(buf + 1);
	struct dlci_struct *d = &dlcis[dlci];
	
	if ((d->state != CONNECTED) && (d->state != FLOW_STOPPED)) {
		send_dm(dlci);
		return;
	}
	pr_info("%s:  dlci %d len %d\n", __func__, dlci, len);

	if (mcc_is_cmd(type)) {
		pr_info("%s: recveive MCC_MSC %02x\n", __func__, v24_sigs);
		if (v24_sigs & RTC) {
			d->type = DATA_CHN;
			pr_info("%s: AT to DATA\n", __func__);
		} else {
			d->type = CMD_CHN;
			pr_info("%s: DATA to AT\n", __func__);
		}
		if (v24_sigs & FC) {
			pr_info("%s: FLOW STOP\n", __func__);
#if 0
			if (d->state == CONNECTED) {
				d->state = FLOW_STOPPED;
			}
#endif
		} else {
			pr_info("%s: FLOW START\n", __func__);
#if 0
			if (d->state == FLOW_STOPPED) {
				d->state = CONNECTED;
			}
			wake_up_interruptible(&d->write_wait);
			queue_delayed_work(wqueue, &d->rx_work, 0);
			tasklet_schedule(&send_tsk);
#endif
		}
		send_msc(dlci, MCC_RSP, v24_sigs);
	} else {
		pr_info("%s: recveive MSR %02x\n", __func__, v24_sigs);
		if (v24_sigs & RTC) {
			d->type = DATA_CHN;
			pr_info("%s: AT to DATA\n", __func__);
		} else {
			d->type = CMD_CHN;
			pr_info("%s: DATA to AT\n", __func__);
		}
		if (v24_sigs & FC) {
			if (d->state == CONNECTED) {
				d->state = FLOW_STOPPED;
			}
			pr_info("%s: FLOW STOP accepted\n", __func__);
		} else {
			if (d->state == FLOW_STOPPED) {
				d->state = CONNECTED;
				/* TODO: flow control not supported */
			}
			wake_up_interruptible(&d->write_wait);
			queue_delayed_work(wqueue, &d->rx_work, 0);
			tasklet_schedule(&send_tsk);
			pr_info("%s: FLOW START accepted\n", __func__);
		}
	}
	wake_up_interruptible(&d->ioctl_wait);
}

static void handle_pn(u8 type, char *buf, int len)
{
	u8 dlci;
	u16 frame_size;
	struct pn_msg_data *pn = (struct pn_msg_data *)buf;
	struct dlci_struct *d;

	pr_info("%s:  len %d\n", __func__, len);
	if (len != 8) {
		pr_err("%s: reveived pn on length:%d != 8\n", __func__, len);
		return;
	}
	dlci = pn->dlci;
	d = &dlcis[dlci];
	frame_size = pn->frame_sizel | (pn->frame_sizeh << 8);

	if (mcc_is_cmd(type)) {
		/* TODO: this looks like it will only ever shrink mtu */
		frame_size = min(frame_size, d->mtu);
		send_pn(pn->prior, frame_size, 0, 0, dlci, MCC_RSP);
		d->mtu = frame_size;
	} else {
		frame_size = min(frame_size, d->mtu);
		d->mtu = frame_size;

		if (d->state == NEGOTIATING) {
			d->state = CONNECTING;
			wake_up_interruptible(&d->open_wait);
		}
	}
}

static void handle_mcc(u8 *buf, u8 control, int len)
{
	u8 type = *buf;
	u8 mcc_len = *(buf + 1);
	buf += 2; len -= 2;

	pr_info("%s: mcc_len:%d len:%d\n", __func__, mcc_len, len);

	switch (type >> 2) {
	case TEST:
		pr_warning("%s: test command unimplemented\n", __func__);
		break;

	case FCON:
		pr_warning("%s: flow control unimplemented\n", __func__);
		break;

	case FCOFF:
		pr_warning("%s: flow control unimplemented\n", __func__);
		break;

	case MSC:
		handle_msc(type, buf, len);
		break;

	case PN:
		handle_pn(type, buf, len);
		break;

	case NSC:
		pr_warning("%s: received non supported cmd response\n", __func__);
		break;

	default:
		pr_warning("%s: received a non supported command\n", __func__);
		send_nsc(type, MCC_RSP);
		break;
	}
}

int __attribute__((weak)) ts0710_recv_push_to_chan(int dlci, unsigned char *buf, int len)
{
	pr_info("%s: need %s()\n", __func__, __func__);
	return len;
}

static void reset_dlci(u8 j)
{
	struct mux_buff *mbuf;
	struct dlci_struct *d;
	if (j >= TS0710_MAX_CHN)
		return;

	d = &dlcis[j];
	d->state = DISCONNECTED;
	d->mtu = DEF_TS0710_MTU;
	wake_up_interruptible(&d->open_wait);
	wake_up_interruptible(&d->ioctl_wait);
	wake_up_interruptible(&d->close_wait);
	wake_up_interruptible(&d->write_wait);

	while ((mbuf = mbuf_get(&d->io_lock, &d->rx_pool))) {
		mbuf_put_free_list(mbuf);
	}

}

/* TODO: look at this */
static void reset_con(void)
{
	int j;
	struct mux_buff *mbuf;
	for (j = 0; j < TS0710_MAX_CHN; j++)
		reset_dlci(j);

	while((mbuf = mbuf_get(&send_lock, &send_list))) {
		mbuf_put_free_list(mbuf);
	}
}

static void upon_disconnect(void)
{
	reset_con();
}

static void rx_worker(struct work_struct *work)
{
	int ret, dlci;
	struct mux_buff *mbuf;
	struct dlci_struct *d = container_of(work, struct dlci_struct, rx_work.work);

	dlci = d->dlci;
	if (test_bit(RX_RUNNING, &d->flag))
		return;
	else
		set_bit(RX_RUNNING, &d->flag);

	while ((mbuf = mbuf_get(&d->io_lock, &d->rx_pool))) {
#ifdef TS0710_DEBUG
		if (test_bit(mbuf->dlci, &debug_mask) && test_bit(DEBUG_RECV, &debug_mask)) {
			pr_info("%s: dlci(%d):\n", __func__,  d->dlci);
			mux_dump_data(mbuf->buf, mbuf->d_len);
		}
#endif
		if (d->state == FLOW_STOPPED) {
			pr_info("%s: dlci(%d): flow stopped\n", __func__,  d->dlci);
			mbuf_put(&d->io_lock, &d->rx_pool, mbuf);
			break;
		}
#ifdef VT_DATA_CHANNEL
		if (unlikely((mbuf->type == DATA_CHN) && (mbuf->type != d->type))) {
#ifdef TS0710_DEBUG
			pr_info("%s: dlci(%d) data and type mimatch\n", __func__, mbuf->dlci);
			mux_dump_data(mbuf->buf, mbuf->d_len);
#endif
			mbuf_put_free_list(mbuf);
			continue;
		}
		/* workaround for VT data */
		if (dlci == VT_CMD_CHANNEL && mbuf->type == DATA_CHN) {
			dlci = VT_DATA_CHANNEL;
		}
#ifdef TS0710_DEBUG
		if (dlci == VT_CMD_CHANNEL) {
			pr_info("%s: dlci(%d): type %d\n", __func__,  dlci, mbuf->type);
			mux_dump_data(mbuf->buf, mbuf->d_len);
		}
#endif
#endif
		ret = ts0710_recv_push_to_chan(dlci, mbuf->d_p, mbuf->d_len);
		if (unlikely(ret < mbuf->d_len)) {
			if (ret > 0) {
				mbuf->d_p += ret;
				mbuf->d_len -= ret;
			}
			mbuf_put(&d->io_lock, &d->rx_pool, mbuf);
#ifdef VT_DATA_CHANNEL
			/* workaround for VT data */
			if (ret < 0 && dlci == VT_DATA_CHANNEL) {
				pr_info("%s: VT data not open\n", __func__);
				ts0710_flow_control(VT_CMD_CHANNEL, 0);
			}
#endif
#ifdef TS0710_DEBUG
			pr_info("%s: dlci(%d): ret %d len %d\n", __func__, dlci, ret, mbuf->d_len);
#endif
			queue_delayed_work(wqueue, &d->rx_work, 1);
			break;
		}
		mbuf_put_free_list(mbuf);
	}
	clear_bit(RX_RUNNING, &d->flag);
}

static void process_send_buf(unsigned long data)
{
	int ret;
	struct mux_buff *mbuf;

	while((mbuf = mbuf_get(&send_lock, &send_list))) {
		struct dlci_struct *d = &dlcis[mbuf->dlci];
#ifdef TS0710_DEBUG
		if (test_bit(mbuf->dlci, &debug_mask) && test_bit(DEBUG_SEND, &debug_mask)) {
			pr_info("%s: dlci(%d):\n", __func__, mbuf->dlci);
			mux_dump_data(mbuf->d_p, mbuf->d_len);
		}
		if (test_bit(mbuf->dlci, &debug_mask) && test_bit(DEBUG_RESEND, &debug_mask)) {
			pr_info("%s: dlci(%d): resend (%d)\n", __func__, mbuf->dlci, mbuf->d_len);
			if (mbuf->type == CMD_CHN)
				mux_dump_data(mbuf->d_p, mbuf->d_len);
			clear_bit(DEBUG_RESEND, &debug_mask);
		}
#endif
		if (unlikely((mbuf->type == DATA_CHN) && (mbuf->type != d->type))) {
#ifdef TS0710_DEBUG
			pr_info("%s: dlci(%d) data and type mimatch\n", __func__, mbuf->dlci);
			mux_dump_data(mbuf->buf, mbuf->d_len);
#endif
			mbuf_put_free_list(mbuf);
			continue;
		}
		if (unlikely((ret = ts0710_tx(mbuf->d_p, mbuf->d_len)) < mbuf->d_len)) {
			if (ret > 0) {
				mbuf->d_p += ret;
				mbuf->d_len -= ret;
			}
			mbuf_put(&send_lock, &send_list, mbuf);
			if (ret == -EAGAIN) {
#ifdef TS0710_DEBUG
				if (test_bit(mbuf->dlci, &debug_mask))
					pr_info("%s: dlci(%d) BP sleep (%d)\n", __func__, 
						mbuf->dlci, mbuf->d_len);
#endif
				break;
			}
			if (ret == -EFAULT) {
#ifdef TS0710_DEBUG
				if (test_bit(mbuf->dlci, &debug_mask))
					pr_info("%s: dlci(%d) reschedule (%d)\n", __func__,
						mbuf->dlci, mbuf->d_len);
#endif
			}
			tasklet_schedule(&send_tsk);
			break;
		} else
			mbuf_put_free_list(mbuf);
	}
}

void ts0710_schedule_send_task(void)
{
	unsigned long flags;
	spin_lock_irqsave(&send_lock, flags);
	if (unlikely(list_empty(&send_list))) {
		spin_unlock_irqrestore(&send_lock, flags);
		return;
	}
	spin_unlock_irqrestore(&send_lock, flags);
#ifdef TS0710_DEBUG
	set_bit(DEBUG_RESEND, &debug_mask);
#endif
	tasklet_schedule(&send_tsk);

}

static void process_dlci0_buf(struct mux_buff *mbuf, u8 control)
{
	int oldstate;
	struct dlci_struct *d = &dlcis[0];

	switch (CLR_PF(control)) {
	case SABM:
		d->state = CONNECTED;
		send_ua(0);
		wake_up_interruptible(&d->open_wait);
		pr_info("%s: SABM dlci 0 connected\n", __func__);
		break;

	case UA:
		if (d->state == CONNECTING) {
			d->state = CONNECTED;
			wake_up_interruptible(&d->open_wait);
			pr_info("%s: UA dlci 0 connected\n", __func__);
		} else if (d->state == DISCONNECTING) {
			upon_disconnect();
		} else {
			pr_warning("%s: dlci 0 invalid UA packet\n", __func__);
		}
		break;

	case DM:
		oldstate = d->state;
		upon_disconnect();
		if (oldstate == CONNECTING)
			d->state = REJECTED;
		break;

	case DISC:
		send_ua(0);
		upon_disconnect();
		break;

	case UIH:
		if (GET_PF(control)) {
			pr_warning("%s: packet with P/F set.\n", __func__);
			break;
		}

		if ((d->state != CONNECTED) && (d->state != FLOW_STOPPED)) {
			pr_warning("%s: dlci 0 not connected.\n", __func__);
			send_dm(0);
			break;
		}
#ifdef TS0710_DEBUG
		pr_info("%s: mcc on channel 0\n", __func__);
		mux_dump_data(mbuf->buf, mbuf->d_len);
#endif
		handle_mcc(mbuf->buf, control, mbuf->d_len);
		break;
	default:
		pr_err("%s: illegal packet: dlci 0 CLR_PF(control) %02x\n",
				__func__, CLR_PF(control));
		break;
	}
	mbuf_put_free_list(mbuf);
}

/*
 *
 *
 */
static void process_mux_buf(struct mux_buff *mbuf, u8 addr, u8 control)
{
	int len = mbuf->d_len;
	struct dlci_struct *d;
	int dlci;

	dlci = addr2dlci(addr);

	d = &dlcis[dlci];

	switch (CLR_PF(control)) {
	case SABM:
		if (!valid_dlci(dlci)) {
			pr_warning("%s: SABM invalid dlci(%d).\n", __func__, dlci);
			send_dm(dlci);
			break;
		}
		d->state = CONNECTED;
		send_ua(dlci);
		wake_up_interruptible(&d->open_wait);
		pr_info("%s: SABM dlci(%d) connected\n", __func__, dlci);
		break;

	case UA:
		if (!valid_dlci(dlci)) {
			pr_warning("%s: UA invalid dlci(%d)\n", __func__, dlci);
			break;
		}
		if (d->state == CONNECTING) {
			d->state = CONNECTED;
			wake_up_interruptible(&d->open_wait);
			pr_info("%s: UA dlci(%d) connected\n", __func__, dlci);
		} else if (d->state == DISCONNECTING) {
			d->state = DISCONNECTED;
			wake_up_interruptible(&d->open_wait);
			wake_up_interruptible(&d->close_wait);
			reset_dlci(dlci);
		} else {
			pr_warning("%s: dlci(%d) invalid UA packet\n", __func__, dlci);
		}
		break;

	case DM:
		if (!valid_dlci(dlci)) {
			pr_warning("%s: DM invalid dlci(%d)\n", __func__, dlci);
			break;
		}
		if (d->state == CONNECTING)
			d->state = REJECTED;
		else
			d->state = DISCONNECTED;
		pr_info("%s: DM dlci(%d) state %d\n", __func__, dlci, d->state);
		wake_up_interruptible(&d->open_wait);
		wake_up_interruptible(&d->close_wait);
		reset_dlci(dlci);
		break;

	case DISC:
		if (!valid_dlci(dlci)) {
			pr_warning("%s: DISC invalid dlci(%d)\n", __func__, dlci);
			break;
		}
		if (!dlci) {
			send_ua(dlci);
			upon_disconnect();
			break;
		}
		d->state = DISCONNECTED;
		send_ua(dlci);
		wake_up_interruptible(&d->open_wait);
		wake_up_interruptible(&d->close_wait);
		reset_dlci(dlci);
		pr_info("%s: DM dlci(%d) disconnected\n", __func__, dlci);
		break;

	case UIH:
		if (unlikely(!valid_dlci(dlci))) {
			pr_warning("%s: UIH invalid dlci(%d)\n", __func__, dlci);
			send_dm(dlci);
			break;
		}

		if (unlikely(GET_PF(control))) {
			pr_warning("%s: packet with P/F set.\n", __func__);
			break;
		}

		if (unlikely((d->state != CONNECTED) && (d->state != FLOW_STOPPED))) {
			pr_warning("%s: dlci(%d) not connected.\n", __func__, dlci);
			send_dm(dlci);
			break;
		}

		if (unlikely(len > d->mtu || len == 0)) {
			pr_warning("%s: dlci(%d): uih_len:%d is bigger than mtu:%d.\n",
				    __func__, dlci, len, d->mtu);
			break;
		}
		mbuf->d_p = mbuf->buf;
		mbuf->type = d->type;
		mbuf->dlci = d->dlci;
		mbuf_put_tail(&d->io_lock, &d->rx_pool, mbuf);
		queue_delayed_work(wqueue, &d->rx_work, 0);
		return;
	default:
#ifdef TS0710_DEBUG
		pr_err("%s: illegal packet: dlci(%d) CLR_PF(control) %02x\n",
				__func__, dlci, CLR_PF(control));
		mux_dump_data(mbuf->buf, mbuf->d_len);
#endif
		break;
	}

	mbuf_put_free_list(mbuf);
}

int ts0710_rx_insert_char(u8 c)
{
//	static u8 fcs = 0;
	static u8 control = 0, addr = 0;
	static int state = RECV_STATE_IDLE;
	static int len = 0;

	static struct mux_buff *mbuf = 0;

	/*
	 *+----------------------------------------------------------+
	 *|Flag|Address|Control|Length Indicator|Information|FCS|Flag|
	 *|----+-------+-------+----------------+-----------+---+----|
	 *| 1  |   1   |   1   |   1 or 2       |length data| 1 | 1  |
	 *+----------------------------------------------------------+
	 */

	if (!mbuf && !(mbuf = mbuf_get_free_list())) {
		pr_err("%s: mbuf is NULL\n", __func__);
		return 0;
	}

	switch (state) {
	case RECV_STATE_IDLE:
		if (c == TS0710_BASIC_FLAG) {
//			fcs = crc_start();
			state = RECV_STATE_ADDR;
		}
		break;
	case RECV_STATE_ADDR:
		if (c != TS0710_BASIC_FLAG) {
//			fcs = crc_calc(fcs, c);
			addr = c;
			state = RECV_STATE_CONTROL;
		} else {
			pr_err("%s: RX wrong data %02x\n", __func__, c);
		}
		break;

	case RECV_STATE_CONTROL:
//		fcs = crc_calc(fcs, c);
		control = c;
		state = RECV_STATE_LEN;
		break;

	case RECV_STATE_LEN:
//		fcs = crc_calc(fcs, c);
		len = c >> 1;
		if (c & 0x1) {
			if (len)
				state = RECV_STATE_DATA;
			else 
				state = RECV_STATE_FCS;
		} else
			state = RECV_STATE_LEN2;
		break;

	case RECV_STATE_LEN2:
//		fcs = crc_calc(fcs, c);
		len |= c << 7;
		state = RECV_STATE_DATA;
		if (len > DEF_TS0710_MTU) {
			pr_info("%s:len (%d) > DLCI_MAX_BUFF_SIZE\n", __func__, len);
		}
		break;

	case RECV_STATE_DATA:
		*mbuf->d_p = c;
		mbuf->d_p++;
		mbuf->d_len++;
		if (mbuf->d_len == len) {
			/* FCS byte */
//			fcs = crc_calc(fcs, c);
			state = RECV_STATE_FCS;
		}
		break;

	case RECV_STATE_FCS:
		state = RECV_STATE_END;
		break;
	case RECV_STATE_END:
		if (unlikely(c != TS0710_BASIC_FLAG)) {
			pr_info("%s: wrong end flag %02x\n", __func__, c);
			break;
		}
		if (unlikely(!addr2dlci(addr))) {
			process_dlci0_buf(mbuf, control);
		} else
			process_mux_buf(mbuf, addr, control);
		mbuf = 0;
		state = RECV_STATE_IDLE;
		break;
	}
	return 1;
}
EXPORT_SYMBOL(ts0710_rx_insert_char);

#define TS0710MUX_TIME_OUT 250

/* call with dlci locked held */
static int wait_for_open(int dlci)
{
	int try = 8;
	int ret;
	struct dlci_struct *d = &dlcis[dlci];

	while (try--) {
		mutex_unlock(&d->lock);
		ret = wait_event_interruptible_timeout(d->open_wait,
						       d->state != CONNECTING,
						       TS0710MUX_TIME_OUT);
		/*
		 * It is possible that d->state could have changed back to
		 * to connecting between being woken up and aquiring the lock.
		 * The side effect is that this open() will return turn -ENODEV.
		 */

		mutex_lock(&d->lock);
		if (ret == 0)
			continue;

		if (ret == -ERESTARTSYS)
			return -EAGAIN;

		if (d->state == REJECTED)
			return -EREJECTED;

		if (d->state != CONNECTED)
			return -ENODEV;

		return 0;
	}

	return -ENODEV;
}

static int open_dlci(u8 dlci)
{
	struct dlci_struct *d = &dlcis[dlci];
	int try;
	int retval;

#ifdef VT_DATA_CHANNEL
	/* workaround for VT data */
	if (dlci == VT_DATA_CHANNEL) {
		pr_info("%s: open VT data channel\n", __func__);
		ts0710_flow_control(VT_CMD_CHANNEL, 1);
		return 0;
	}
#endif
	pr_info("%s: dlci(%d)\n", __func__, dlci);
	mutex_lock(&d->lock);
	if (d->clients > 0) {
		if (d->state == CONNECTED || d->state == FLOW_STOPPED) {
			d->clients++;
			mutex_unlock(&d->lock);
			return 0;
		}

		if (d->state != CONNECTING) {
			mutex_unlock(&d->lock);
			return -EREJECTED;
		}
		retval = wait_for_open(dlci);
		d->clients++;
		mutex_unlock(&d->lock);
		return retval;
	}

	if (d->state != DISCONNECTED && d->state != REJECTED) {
		pr_err("%s: DLCI(%d): state invalid on open\n", __func__, dlci);
		mutex_unlock(&d->lock);
		return -ENODEV;
	}

	/* we are the first to try to open the dlci */
	d->state = CONNECTING;
	try = (dlci == 0 ? 10 : 3);
	while (try--) {
		send_sabm(dlci);

		mutex_unlock(&d->lock);
		retval = wait_event_interruptible_timeout(d->open_wait,
							  d->state !=
							  CONNECTING,
							  TS0710MUX_TIME_OUT);
		mutex_lock(&d->lock);

		if (retval == 0)
			continue;

		if (retval == -ERESTARTSYS) {
			retval = -EAGAIN;
			break;
		}

		if (d->state == REJECTED) {
			retval = -EREJECTED;
			break;
		}

		if (d->state != CONNECTED) {
			retval = -ENODEV;
			break;
		}

		d->clients++;

		retval = 0;
		break;
	}

	if (try < 0)
		retval = -ENODEV;

	if (d->state == CONNECTING)
		d->state = DISCONNECTED;

	/* other ttys might be waiting on this dlci */
	wake_up_interruptible(&d->open_wait);

	mutex_unlock(&d->lock);
	return retval;
}

int ts0710_connect(int dlci)
{
	int ret;
	if ((ret = open_dlci(0)) < 0) {
		pr_err("%s: open dhci 0 failed\n", __func__);
		return ret;
	}

	return open_dlci(dlci);
}
EXPORT_SYMBOL(ts0710_connect);

int ts0710_room_size(int dlci) 
{
	unsigned long flags;
	struct dlci_struct *d = &dlcis[dlci];

#ifdef VT_DATA_CHANNEL
	/* workaround for VT data */
	if (unlikely(dlci == VT_DATA_CHANNEL)) {
		d = &dlcis[VT_CMD_CHANNEL];
	}
#endif
	spin_lock_irqsave(&free_list_lock, flags);
	if (unlikely(list_empty(&mbuf_free_list) && (atomic_read(&mbuf_count) < 0))) {
		spin_unlock_irqrestore(&free_list_lock, flags);
		return 0;
	}
	spin_unlock_irqrestore(&free_list_lock, flags);

	return (d->mtu - TS0710_MAX_HDR_SIZE - FCS_SIZE - FLAG_SIZE);
}
EXPORT_SYMBOL(ts0710_room_size);

int ts0710_send(int dlci, const u8 *buf, int count)
{
	int err, ret, c;
	struct dlci_struct *d0 = &dlcis[0], *d = &dlcis[dlci];

	if (unlikely(count <= 0))
		return count;

	if (unlikely(d0->state == FLOW_STOPPED)) {
		/* TODO: this should block */
		pr_info("%s: Flow stopped on all, channel %d\n", __func__, dlci);
		return 0;
	}

#ifdef VT_DATA_CHANNEL
#ifdef TS0710_DEBUG
	 if (dlci == VT_CMD_CHANNEL) {
		pr_info("%s: dlci(%d): type %d\n", __func__,  d->dlci, d->type);
		mux_dump_data(buf, count);
	}
#endif
	/* workaround for VT data */
	if (unlikely(dlci == VT_DATA_CHANNEL)) {
		d = &dlcis[VT_CMD_CHANNEL];
	       	if (d->type == CMD_CHN)
			return count;
		else
			dlci = VT_CMD_CHANNEL;
	}
#endif

	if (unlikely(d->state == FLOW_STOPPED)) {
		/* TODO: this should block */
		pr_info("%s: Flow stopped, channel %d\n", __func__, dlci);
		if (wait_event_interruptible(d->write_wait, d->state == CONNECTED))
			return -ERESTARTSYS;
	}

	if (likely(d->state == CONNECTED)) {
		c = min(count, (d->mtu - TS0710_MAX_HDR_SIZE - FCS_SIZE - FLAG_SIZE));
		if (c <= 0) {
			pr_info("%s: c = %d\n", __func__, c);
			err = 0;
			goto err;
		}
		ret = send_uih(dlci, buf, c);
		/*
		 * TODO: should check write notify flag and call back
		 * into the tty layer
		 */
		if (unlikely(ret)) {
			pr_err("%s: ret %d\n", __func__, ret);
			return ret;
		} else
			return c;
	}
	pr_warning("%s: %d while not connected\n", __func__, dlci);
	return -EFAULT;

err:
	return err;
}
EXPORT_SYMBOL(ts0710_send);

int ts0710_switch(int dlci, int cmd)
{
	int i, v24_sig;
	int ret, type;
	struct dlci_struct *d = &dlcis[dlci];
	struct mux_buff *mbuf;

	if (cmd == TS0710MUX_IO_MSC_AT) {
		v24_sig = (EA | DV);
		type = CMD_CHN;
	} else {
		v24_sig = (EA | RTC | DV);
		type = DATA_CHN;
	}
	pr_info("%s: dlci(%d) type = %s\n", __func__, dlci, type == CMD_CHN ? "DATA to AT":"AT to DATA");

#ifdef VT_DATA_CHANNEL
	/* workaround for VT data */
	if (dlci == VT_DATA_CHANNEL) {
		d = &dlcis[VT_CMD_CHANNEL];
		dlci = VT_CMD_CHANNEL;
	}
#endif
	for (i = 0; i < 3; i++) {
		if (send_msc(dlci, MCC_CMD, v24_sig) < 0) {
			continue;
		 } else {
			break;
		 }
	}

	pr_info("%s: wait event %d\n", __func__, type);
	ret = wait_event_interruptible(d->ioctl_wait, d->type == type);
	pr_info("%s: wait event return %d\n", __func__, d->type);
	while ((mbuf = mbuf_get(&d->io_lock, &d->rx_pool))) {
		if (mbuf->type == d->type || mbuf->type == CMD_CHN) {
			mbuf_put(&d->io_lock, &d->rx_pool, mbuf);
			break;
		}
#ifdef TS0710_DEBUG
		pr_info("%s: discarding useless data, len %d\n", __func__, mbuf->d_len);
		mux_dump_data(mbuf->buf, mbuf->d_len);
#endif
		mbuf_put_free_list(mbuf);
	}
	pr_info("%s: return %d\n", __func__, d->type);
	return 0;
}
EXPORT_SYMBOL(ts0710_switch);

void ts0710_flow_control(int dlci, int on)
{
	int i, v24_sig;
	struct dlci_struct *d = &dlcis[dlci];

#ifdef VT_DATA_CHANNEL
	/* workaround for VT data */
	if (dlci == VT_DATA_CHANNEL) {
		d = &dlcis[VT_CMD_CHANNEL];
		dlci = VT_CMD_CHANNEL;
	}
#endif
	if (on) {
		v24_sig = (EA | DV);
	} else {
		v24_sig = (EA | FC | DV);
		d->state = FLOW_STOPPED;
	}

	if (d->type == DATA_CHN)
		v24_sig |= RTC;
	for (i = 0; i < 3; i++) {
		if (send_msc(dlci, MCC_CMD, v24_sig) < 0) {
			continue;
		 } else {
			break;
		 }
	}

}
EXPORT_SYMBOL(ts0710_flow_control);

static int close_dlci(int dlci)
{
	struct dlci_struct *d = &dlcis[dlci];
	int try;
	int retval;

#ifdef VT_DATA_CHANNEL
	/* workaround for VT data */
	if (dlci == VT_DATA_CHANNEL) {
		ts0710_switch(dlci, TS0710MUX_IO_MSC_AT);
		ts0710_flow_control(dlci, 1);
		pr_info("%s: close VT data channel\n", __func__);
		return 0;
	}
#endif

	pr_info("%s: dlci(%d)\n", __func__, dlci);
	mutex_lock(&d->lock);

	if (d->clients > 1) {
		d->clients--;
		mutex_unlock(&d->lock);
		return 0;
	}

	if (d->state == DISCONNECTED ||
	    d->state == REJECTED ||
	    d->state == DISCONNECTING) {
		d->clients--;
		mutex_unlock(&d->lock);
		return 0;
	}

	d->state = DISCONNECTING;
	try = 3;
	while (try--) {
		send_disc(dlci);
		mutex_unlock(&d->lock);
		retval = wait_event_interruptible_timeout(d->close_wait,
							  d->state !=
							  DISCONNECTING,
							  TS0710MUX_TIME_OUT);
		mutex_lock(&d->lock);

		if (retval == 0)
			continue;

		if (retval == -ERESTARTSYS) {
			retval = -EAGAIN;
			break;
		}

		if (d->state != DISCONNECTED) {
			retval = -EIO;
			break;
		}

		retval = 0;
		break;
	}

	if (try < 0)
		retval = -EIO;

	if (d->state != DISCONNECTED) {
		if (dlci == 0) {
			upon_disconnect();
		} else {
			d->state = DISCONNECTED;
			wake_up_interruptible(&d->close_wait);
			reset_dlci(dlci);
		}
	}

	d->clients--;

	mutex_unlock(&d->lock);
	return retval;
}

int ts0710_disconnect(int dlci)
{
	int ret;
	if ((ret = close_dlci(dlci)) < 0) {
		return ret;
	}
	return close_dlci(0);

}
EXPORT_SYMBOL(ts0710_disconnect);

static struct delayed_work bp_assert_work;

static void ts0710_bp_assert_worker(struct work_struct *work)
{
	reset_con();
}

void ts0710_bp_assert(void)
{
	schedule_delayed_work(&bp_assert_work, 10);
}
EXPORT_SYMBOL(ts0710_bp_assert);

static int __init ts0710_init(void)
{
	int err;
	int i;
	struct dlci_struct *d;

	crc_create_table(crctable);
	INIT_LIST_HEAD(&mbuf_free_list);
	for (i = 0; i < TS0710_MAX_CHN; i++) {
		d = &dlcis[i];
		mutex_init(&d->lock);
		d->mtu = DEF_TS0710_MTU;
		d->dlci = i;
		d->type = CMD_CHN;
		spin_lock_init(&d->io_lock);
		INIT_LIST_HEAD(&d->rx_pool);
		INIT_DELAYED_WORK(&d->rx_work, rx_worker);
		init_waitqueue_head(&d->open_wait);
		init_waitqueue_head(&d->ioctl_wait);
		init_waitqueue_head(&d->close_wait);
		init_waitqueue_head(&d->write_wait);
	}
	INIT_DELAYED_WORK(&bp_assert_work, ts0710_bp_assert_worker);
	reset_con();
	atomic_set(&mbuf_count, TS0710_MAX_MBUF_LIST);
	wqueue = create_singlethread_workqueue("ipc mux");
	if (!wqueue) {
                pr_err("%s:failed to create work queue.\n", __func__);
		err = -ESRCH;
		goto err0;
	}

	pr_info("%s: TS 07.10 Multiplexer registered\n", __func__);

	return 0;

err0:
	for (i = 0; i < TS0710_MAX_CHN; i++) {
		d = &dlcis[i];
		cancel_delayed_work_sync(&d->rx_work);
	}

	return err;
}

void __exit ts0710_exit(void)
{
	int i;
	struct dlci_struct *d;

	for (i = 0; i < TS0710_MAX_CHN; i++) {
		d = &dlcis[i];
		cancel_delayed_work_sync(&d->rx_work);
	}
	if (wqueue) {
		destroy_workqueue(wqueue);
		wqueue = NULL;
	}
	tasklet_kill(&send_tsk);
}

module_init(ts0710_init);
module_exit(ts0710_exit);

MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("TS 07.10 Multiplexer");
