#include <linux/module.h>
#include <linux/types.h>
#include <linux/tty.h>
#include <linux/tty_flip.h>
#include <linux/kernel.h>
#include <linux/string.h>

#include "ts0710.h"

#define TS0710MUX_IO_MSC_AT  0x54F2
#define TS0710MUX_IO_MSC_DATA  0x54F3


struct ts0710_tty_channel_data {
	atomic_t ref_count;
	struct tty_struct *tty;
	struct work_struct work; /* wakeup tty */
};

struct ts0710_tty_data {
	struct ts0710_tty_channel_data		chan[NR_MUXS];
};

static struct tty_driver *driver;

#define TS0710MUX_MAJOR 245
#define TS0710MUX_MINOR_START 0

static const u8 tty2dlci[] = {1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17};
static const u8 dlci2tty[] = {0, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16};

int ts0710_recv_push_to_chan(int dlci, u8 *data, int len)
{
	int ret;
	int line = dlci2tty[dlci];
	struct ts0710_tty_data *td = driver->driver_state;
	struct tty_struct *tty;
     	
	if (!(tty = td->chan[line].tty)) {
#ifdef VT_DATA_CHANNEL
		/* workaround for VT data */
		if (dlci == VT_DATA_CHANNEL) {
			pr_info("%s: VT data channel\n", __func__);
			return -ENODEV;
		}
#endif
		pr_info("%s: mux %d no open.\n", __func__, line);
		return len;
	}

	ret = tty_insert_flip_string(tty, data, len);
	tty_flip_buffer_push(tty);
	return ret;
}

static int ts0710_tty_open(struct tty_struct *tty, struct file *filp)
{
	struct ts0710_tty_data *td = tty->driver->driver_state;
	int err;
	int line;

	line = tty->index;
	if ((line < 0) || (line >= NR_MUXS)) {
		pr_err("ts0710: tty index out of range.\n");
		err = -ENODEV;
		goto err;
	}

	atomic_inc(&td->chan[line].ref_count);

	td->chan[line].tty = tty;

	err = ts0710_connect(tty2dlci[line]);
	if (err < 0)
		goto err;

	return 0;

err:
	return err;
}


static void ts0710_tty_close(struct tty_struct *tty, struct file *filp)
{
	struct ts0710_tty_data *td = tty->driver->driver_state;

	if (atomic_dec_and_test(&td->chan[tty->index].ref_count)) {
		ts0710_disconnect(tty2dlci[tty->index]);

		td->chan[tty->index].tty = NULL;

		/*
		 * the old code did:
		 *   wake_up_interruptible(&tty->read_wait);
		 *   wake_up_interruptible(&tty->write_wait);
		 *
		 * I belive this is unecessary
		 */
	}
}

static void tty_wakeup_worker(struct work_struct *work)
{
	struct ts0710_tty_channel_data *cd = 
		container_of(work, struct ts0710_tty_channel_data, work);
	if (cd->tty)
		tty_wakeup(cd->tty);
}

static int ts0710_tty_write(struct tty_struct *tty,
			     const unsigned char *buf, int count)
{
	int dlci, line;
	int ret;
	struct ts0710_tty_data *td = tty->driver->driver_state;

	line = tty->index;
	if ((line < 0) || (line >= NR_MUXS))
		return -ENODEV;

	dlci = tty2dlci[line];

	ret = ts0710_send(dlci, buf, count);
	if (ret > 0) {
		schedule_work(&td->chan[line].work);
	}
	return ret;
}


static int ts0710_tty_write_room(struct tty_struct *tty)
{
	int dlci;
	int line;

	line = tty->index;
	if ((line < 0) || (line >= NR_MUXS))
		return -ENODEV;

	dlci = tty2dlci[line];

	return ts0710_room_size(dlci);
}

static void ts0710_tty_flush_buffer(struct tty_struct *tty)
{
	pr_info("%s: on line %d\n", __func__, tty->index);
	ts0710_schedule_send_task();
}

static int ts0710_tty_chars_in_buffer(struct tty_struct *tty)
{
	return 0;
}

static void ts0710_tty_throttle(struct tty_struct *tty)
{
	int dlci;
	int line;

	pr_info("%s:on line %d\n", __func__, tty->index);
	line = tty->index;
	if ((line < 0) || (line >= NR_MUXS))
		return;

	dlci = tty2dlci[line];

	ts0710_flow_control(dlci, 0);
}

static void ts0710_tty_unthrottle(struct tty_struct *tty)
{	int dlci;
	int line;

	pr_info("%s:on line %d\n", __func__, tty->index);
	line = tty->index;
	if ((line < 0) || (line >= NR_MUXS))
		return;

	dlci = tty2dlci[line];

	ts0710_flow_control(dlci, 1);
}

static int ts0710_tty_ioctl(struct tty_struct *tty, struct file *file,
		     unsigned int cmd, unsigned long arg)
{
	int line;

	line = tty->index;
	if ((line < 0) || (line >= NR_MUXS))
		return -ENODEV;

	switch (cmd) {
	case TS0710MUX_IO_MSC_DATA:
		ts0710_switch(tty2dlci[line], TS0710MUX_IO_MSC_DATA);
		pr_info("%s: %d change as data channel\n", __func__, line);
		return 0;
	case TS0710MUX_IO_MSC_AT:
		tty_ldisc_flush(tty);
		ts0710_switch(tty2dlci[line], TS0710MUX_IO_MSC_AT);
		pr_info("%s: %d change as cmd channel\n", __func__, line);
		return 0;

	default:
		break;
	}

	return -ENOIOCTLCMD;
}

static const struct tty_operations ts0710_tty_ops = {
	.open = ts0710_tty_open,
	.close = ts0710_tty_close,
	.write = ts0710_tty_write,
	.write_room = ts0710_tty_write_room,
	.flush_buffer = ts0710_tty_flush_buffer,
	.chars_in_buffer = ts0710_tty_chars_in_buffer,
	.throttle = ts0710_tty_throttle,
	.unthrottle = ts0710_tty_unthrottle,
	.ioctl = ts0710_tty_ioctl,
};

static int ts0710_tty_init(void)
{
	struct ts0710_tty_data *td;
	int err;
	int i;

	driver = alloc_tty_driver(NR_MUXS);
	if (driver == NULL) {
		err = -ENOMEM;
		goto err0;
	}

	td = kzalloc(sizeof(*td), GFP_KERNEL);
	if (td == NULL) {
		err = -ENOMEM;
		goto err1;
	}

	for (i = 0; i < NR_MUXS; i++) {
		atomic_set(&td->chan[i].ref_count, 0);
		INIT_WORK(&td->chan[i].work, tty_wakeup_worker);
	}

	driver->driver_state = td;

	driver->driver_name = "ts0710mux";
	driver->name = "ttyMUX";
	driver->major = TS0710MUX_MAJOR;
	driver->major = 234;
	driver->minor_start = TS0710MUX_MINOR_START;
	driver->type = TTY_DRIVER_TYPE_SERIAL;
	driver->subtype = SERIAL_TYPE_NORMAL;
	driver->init_termios = tty_std_termios;
	driver->init_termios.c_iflag = 0;
	driver->init_termios.c_oflag = 0;
	driver->init_termios.c_cflag = B38400 | CS8 | CREAD;
	driver->init_termios.c_lflag = 0;
	driver->flags = TTY_DRIVER_RESET_TERMIOS | TTY_DRIVER_REAL_RAW;

	driver->other = NULL;
	driver->owner = THIS_MODULE;

	tty_set_operations(driver, &ts0710_tty_ops);

	if (tty_register_driver(driver)) {
		pr_err("ts0710: can't register tty driver\n");
		err = -EINVAL;
		goto err2;
	}

	return 0;

err2:
	kfree(td);
err1:
	put_tty_driver(driver);
err0:
	return err;
}

void ts0710_tty_remove(void)
{
	struct ts0710_tty_data *td = driver->driver_state;
	tty_unregister_driver(driver);
	kfree(td);
	put_tty_driver(driver);
}

module_init(ts0710_tty_init);
