/*
 * gpio sleep debug  
 *
 * Copyright (C) 2008 Borqs Co.
 * Author: Ru Yi <yi.ru@borqs.com>
 *
 * This software is licensed under the terms of the GNU General Public
 * License version 2, as published by the Free Software Foundation, and
 * may be copied, distributed, and modified under those terms.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 */

#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/platform_device.h>
#include <linux/leds.h>
#include <linux/err.h>
#include <asm/arch/leds-debug.h>
#if defined(CONFIG_PXA3xx_DVFM)
#include <asm/arch/dvfm.h>
#include <asm/arch/pxa3xx_dvfm.h>
#endif
#include "leds.h"

#define DRVNAME "gpio_sleep_debug"
static struct sleep_led_platform_data *sleep_led_pdata = NULL;
static int dvfm_dev_idx;

static void leds_brightness_set(struct led_classdev *led_cdev, 
					enum led_brightness value)
{
	if(sleep_led_pdata)
		sleep_led_pdata->set_brightness(value ? LED_FULL : LED_OFF);
#ifdef CONFIG_PXA3xx_DVFM		
	if(value)
		dvfm_disable_op_name("D2", dvfm_dev_idx);
	else
		dvfm_enable_op_name("D2", dvfm_dev_idx);
#endif

	return;
}

static struct led_classdev sleep_led = {
	.name = "gpio_sleep_debug",
	.flags = 0,
	.brightness = 0,
	.brightness_set = leds_brightness_set,
	.default_trigger = "sleep",
};

int enable_sleep_led(void)
{
	led_trigger_set_default(&sleep_led);
	return 0;
}

int disable_sleep_led(void)
{
	led_trigger_set(&sleep_led, NULL);
	if (sleep_led.brightness)
		led_set_brightness(&sleep_led, LED_OFF);
	return 0;
}

#ifdef CONFIG_PM
static int gpio_sleep_debug_suspend(struct platform_device *dev,
		pm_message_t state)
{
	led_classdev_suspend(&sleep_led);
	return 0;
}

static int gpio_sleep_debug_resume(struct platform_device *dev)
{
	led_classdev_resume(&sleep_led);
	return 0;
}
#else
#define gpio_sleep_debug_suspend NULL
#define gpio_sleep_debug_resume NULL
#endif

static ssize_t sleep_led_enable_show(struct device *dev, struct device_attribute *attr, char *buf)
{
	int value = sleep_led.trigger ? 1 : 0;

	return sprintf(buf, "%d", value);
}

static ssize_t sleep_led_enable_store(
		struct device *dev, struct device_attribute *attr,
		const char *buf, size_t size)
{
	int value;

	sscanf(buf, "%d", &value);
	if (value)
		enable_sleep_led();
	else
		disable_sleep_led();

	return size;
}

static DEVICE_ATTR(enable, S_IRUGO | S_IWUGO, sleep_led_enable_show, sleep_led_enable_store);

static int __init gpio_sleep_debug_probe(struct platform_device *pdev)
{
	int ret = -EINVAL;
	sleep_led_pdata = pdev->dev.platform_data;

	ret = led_classdev_register(&pdev->dev, &sleep_led);
	if (ret < 0)
		return ret;

	ret = device_create_file(&pdev->dev, &dev_attr_enable);
	if (ret < 0)
		led_classdev_unregister(&sleep_led);

#if defined(CONFIG_PXA3xx_DVFM) 
	dvfm_register("wakelock", &dvfm_dev_idx);
#endif
	return ret;
}

static int gpio_sleep_debug_remove(struct platform_device *pdev)
{
#if defined(CONFIG_PXA3xx_DVFM) 
	dvfm_unregister("wakelock", &dvfm_dev_idx);
#endif
	led_classdev_unregister(&sleep_led);
	return 0;
}

static struct platform_driver gpio_sleep_debug_driver = {
	.probe		= gpio_sleep_debug_probe,
	.remove		= gpio_sleep_debug_remove,
	.suspend	= gpio_sleep_debug_suspend,
	.resume		= gpio_sleep_debug_resume,
	.driver		= {
		.name		= DRVNAME,
		.owner		= THIS_MODULE,
	},
};

static int __init gpio_sleep_debug_init(void)
{
	int ret;

	ret = platform_driver_register(&gpio_sleep_debug_driver);
	if (ret < 0)
	      return ret;
	return 0;
}

static void __exit gpio_sleep_debug_exit(void)
{
	platform_driver_unregister(&gpio_sleep_debug_driver);
}

module_init(gpio_sleep_debug_init);
module_exit(gpio_sleep_debug_exit);

MODULE_AUTHOR("Ru Yi <yi.ru@borqs.com>");
MODULE_DESCRIPTION("Borqs gpio led sleep debug");
MODULE_LICENSE("GPL");
