/*
 * LEDs driver for PXA 
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/platform_device.h>
#include <linux/leds.h>
#include <linux/err.h>
#include <asm/gpio.h>
#include <mach/micco.h>
#include <mach/leds-pxa.h>

#define DRVNAME "pxa-leds"

static void (*keyboard_backlight_set)(int);
static void (*button_backlight_set)(int);
static void (*battery_charging_full_backlight_set)(int);
static void (*battery_charging_backlight_set)(int);
static void (*battery_full_backlight_set)(int);

static void keyboard_leds_brightness_set(struct led_classdev *led_cdev, 
					enum led_brightness value)
{
	pr_debug("keyboard leds: backlight o%s\n", value ? "n" : "ff");

 	if (keyboard_backlight_set)
	      keyboard_backlight_set(value);
}

static struct led_classdev keyboard_led = {
	.name = "keyboard-backlight",
	.brightness = 0,
	.brightness_set = keyboard_leds_brightness_set,
};

static void button_leds_brightness_set(struct led_classdev *led_cdev, 
					enum led_brightness value)
{
	pr_debug("button leds: backlight o%s\n", value ? "n" : "ff");

 	if (button_backlight_set)
	      button_backlight_set(value);
}

static struct led_classdev button_led = {
	.name = "button-backlight",
	.brightness = 0,
	.brightness_set = button_leds_brightness_set,
};

static void battery_charging_full_leds_brightness_set(struct led_classdev *led_cdev, 
					enum led_brightness value)
{
        pr_debug("battery_charging_full leds: backlight o%s\n", value ? "n" : "ff");

        if (battery_charging_full_backlight_set)
              battery_charging_full_backlight_set(value);
}

static struct led_classdev battery_charging_full_led = {
	.name 			= "battery_charging_full",
	.brightness 		= 0,
	.brightness_set 	= battery_charging_full_leds_brightness_set,	
	.default_trigger	= "battery-charging-or-full",
};

static void battery_charging_leds_brightness_set(struct led_classdev *led_cdev, 
					enum led_brightness value)
{
        pr_debug("battery_charging_full leds: backlight o%s\n", value ? "n" : "ff");

        if (battery_charging_backlight_set)
              battery_charging_backlight_set(value);
}

static struct led_classdev battery_charging_led = {
	.name 			= "battery_charging",
	.brightness 		= 0,
	.brightness_set 	= battery_charging_leds_brightness_set,
	.default_trigger	= "battery-charging",
};

static void battery_full_leds_brightness_set(struct led_classdev *led_cdev,
                                        enum led_brightness value)
{
        pr_debug("battery_full leds: backlight o%s\n", value ? "n" : "ff");

        if (battery_full_backlight_set)
              battery_full_backlight_set(value);
}

static struct led_classdev battery_full_led = {
        .name                   = "battery_full",
        .brightness             = 0,
        .brightness_set         = battery_full_leds_brightness_set,
        .default_trigger        = "battery-full",
};


#ifdef CONFIG_PM
static int pxa_led_suspend(struct platform_device *dev,
		pm_message_t state)
{
	led_classdev_suspend(&keyboard_led);
	led_classdev_suspend(&button_led);
	return 0;
}

static int pxa_led_resume(struct platform_device *dev)
{
	led_classdev_resume(&button_led);
	led_classdev_resume(&keyboard_led);
	return 0;
}
#else
#define pxa_led_suspend NULL
#define pxa_led_resume NULL
#endif

static int __init pxa_led_probe(struct platform_device *pdev)
{
	struct leds_platform_data *pdata;
	int ret = -EINVAL;

	pdata = pdev->dev.platform_data;
	if (pdata == NULL) {
		dev_err(&pdev->dev, "no platform data defined\n");
		ret = -EINVAL;
		return ret;	
	}

	keyboard_backlight_set = pdata->set_keyboard_brightness;
	ret = led_classdev_register(&pdev->dev, &keyboard_led);
	if (ret < 0)
		return ret;

	button_backlight_set = pdata->set_button_brightness;
	ret = led_classdev_register(&pdev->dev, &button_led);
	if (ret < 0)
		return ret;


	battery_charging_full_backlight_set = pdata->set_battery_charging_full_brightness;
	ret = led_classdev_register(&pdev->dev, &battery_charging_full_led);
	if (ret < 0)
		return ret;

	battery_charging_backlight_set = pdata->set_battery_charging_brightness;
	ret = led_classdev_register(&pdev->dev, &battery_charging_led);
	if (ret < 0)
		return ret;

	battery_full_backlight_set = pdata->set_battery_full_brightness;
	ret = led_classdev_register(&pdev->dev, &battery_full_led);
	if (ret < 0)
		return ret;

	return 0;
}

static int pxa_led_remove(struct platform_device *pdev)
{
	led_classdev_unregister(&battery_full_led);
	led_classdev_unregister(&battery_charging_led);
	led_classdev_unregister(&battery_charging_full_led);
	led_classdev_unregister(&button_led);
	led_classdev_unregister(&keyboard_led);
	return 0;
}

static struct platform_driver pxa_led_driver = {
	.probe		= pxa_led_probe,
	.remove		= pxa_led_remove,
	.suspend	= pxa_led_suspend,
	.resume		= pxa_led_resume,
	.driver		= {
		.name		= DRVNAME,
		.owner		= THIS_MODULE,
	},
};

static int __init pxa_led_init(void)
{
	int ret;

	ret = platform_driver_register(&pxa_led_driver);
	if (ret < 0)
	      return ret;
	return 0;
}

static void __exit pxa_led_exit(void)
{
	platform_driver_unregister(&pxa_led_driver);
}

module_init(pxa_led_init);
module_exit(pxa_led_exit);
