#ifndef _IF101_H_
#define _IF101_H_
#include <linux/wait.h>		// wait_queue_head_t
#include <linux/semaphore.h>
#include <linux/completion.h>
#include <linux/types.h>
#include <linux/list.h>

#define CONFIG_INNODEV_V2_DEBUG

#define inno_msg(fmt, arg...)	printk(KERN_INFO "[inno]"fmt"\n", ##arg)
#define inno_dbg(fmt, arg...)	//printk(KERN_INFO "[inno]"fmt"\n", ##arg)
#define inno_err(fmt, arg...)	printk(KERN_ERR "[inno:ERROR!!!]"fmt"\n", ##arg) 

// character device major and name
#define INNODEV_MAJOR		99
#define INNODEV_CHRDEV_NAME		"innodev"

// if101 i2c slave address
#define INNO_I2C_SLAVE_ADDR		0x10

#define INNO_BUFFER_SIZE	(0x10E00)


typedef int (*i2c_init_func_t)(void);
typedef int (*i2c_read_func_t)(u16 reg, u8* data);
typedef int (*i2c_write_func_t)(u16 reg, u8 data);
typedef int (*i2c_exit_func_t)(void);
typedef struct inno_i2c_driver{
	i2c_init_func_t		init;
	i2c_read_func_t		read;
	i2c_write_func_t	write;
	i2c_exit_func_t		exit;
}inno_i2c_driver_t;

#define INNO_SPI_MASTER		0
#define INNO_SPI_SLAVE		1
typedef int (*spi_init_func_t)(u8 type);
typedef int (*spi_read_func_t)(u8* pdata, u32 len);
typedef int (*spi_write_func_t)(u8* pdata, u32 len);
typedef int (*spi_uninit_func_t)(void);
typedef int (*spi_sync_cs_func_t)(void);	//slave mode to download fireware
typedef int (*spi_drive_cs_func_t)(int drive);
typedef struct inno_spi_driver{
	spi_init_func_t		init;
	spi_read_func_t		read;
	spi_write_func_t	write;
	spi_uninit_func_t	uninit;
	spi_sync_cs_func_t	sync_cs;
	spi_drive_cs_func_t	drive_cs;
	
	spi_read_func_t		read2;
	spi_write_func_t	write2;
}inno_spi_driver_t;

typedef struct inno_cmdset {
    u32 (*get_lgx_length)(u8 channelID);
    u8  (*get_intr_status)(void);
    void (*fetch_lgx_data)(u8 chanelID, u8* buf, u32 len);
    void (*fetch_lgx_data2)(u8 chanelID, u8* buf, u32 len);    
}inno_cmdset_t;

typedef enum TRANSFER_DATA_SIZE {
	TRANS_1BYTE = 0,			/* transfer 8bits data every time */
	TRANS_2BYTES,				/* transfer 16bits data every time */
	TRANS_4BYTES,				/* transfer 32bits data every time */
	TRANS_BURST				/* continous transfer until nCS high */
}trans_size_t;

typedef enum INTR_TYPE {
	FALLING_EDGE = 0,
	RISING_EDGE
}intr_type_t;

typedef enum SPI_CLK_MODE {
	SPI_CLK_MODE_0 = 0,		//idle low, transmit falling edge, receive rising edge.
	SPI_CLK_MODE_1,			//idle high, transmit rising edge, receive falling edge.
	SPI_CLK_MODE_2,			//idle low, transmit rising edge, receive falling edge.
	SPI_CLK_MODE_3			//idle high, transmit falling edge, receive rising edge.
}spi_clk_mode_t;

typedef enum SPI_INTERFACE {
	NORMAL_SPI = 0,		// CLK, CS, TXD, RXD (4pins)
	ONEWIRE_SPI,		// CLK, CS, TRXD (3pins)
	FOURWIRES_SPI,		// CLK, CS, TRXD0, TRXD1, TRXD2, TRXD3 (6pins)
}spi_interface_t;

typedef struct inno_configuration{
	spi_interface_t	spi_interface;
	spi_clk_mode_t	spi_clk_mode;	
	intr_type_t		intr_type;
	trans_size_t	trans_size;
	int				i2c_cmd_mode;	/* if set 1, use i2c send command, otherwise use spi send command */
}inno_cfg_t;

typedef enum Logic_Channel_ID {
	LG0 = 0,
	LG1,
	UAM,
	LG_END
}lgx_t;

struct inno_lgx_ids{
	char* 	name;
	int 	id;
};

typedef struct inno_buffer{
	void* start;
	int 	own_id;
	int		valid;
	u32		valid_size;
	struct semaphore sem;
	void		*vaddr;		/* vmap() return virtual address */
	struct page	**pages;	/* physical pages */
	int		page_num;	/* physical pages count */
	int		bufsize;	/* buffer size */
}inno_buffer_t;

#define MAX_INNOLG_NUM	LG_END	// maxium logic channel number
typedef struct inno_lgx{
	struct inno_lgx_ids*	ids;
	int use_count;
	struct semaphore sem;
	
	wait_queue_head_t	read_wait;
	wait_queue_head_t	uam_wait;
	u32		bytes_available;	// byte size available by if101
	struct inno_buffer	inno_buffer;
}inno_lgx_t;

typedef struct inno_device{
	struct inno_configuration cfg;
	struct inno_lgx*	lgxs[MAX_INNOLG_NUM];

	inno_i2c_driver_t*		i2c_driver;
	inno_spi_driver_t*		spi_driver;
	inno_cmdset_t*		cmdset;	

	int 	use_count;

	struct semaphore	sem;	// inno device semaphore
}inno_device_t;

typedef enum INNO_RETURN_VALUE {
	INNO_SUCCESS = 0,
	INNO_ERROR,
	INNO_TIMEOUT,	// timeout error
	INNO_PARAM_ERR,	//parameter error
}inno_return_value_t;

//=========== UAM ==============//
#define UAM_RESET_CMD 0xc1
#define UAM_PPS_CMD 0xc4
#define SET_UAM_CLOCK_CMD 0xc5
#define SEND_UAM_CMD 0xc2
#define GET_UAM_CMD 0xc3

#define SCARD_PROTOCOL_UNDEFINED    0x00000000  // There is no active protocol.
#define SCARD_PROTOCOL_T0           0x00000001  // T=0 is the active protocol.
#define SCARD_PROTOCOL_T1           0x00000002  // T=1 is the active protocol.

// Block waiting integer default value as definded by ISO
#define T1_BWI_DEFAULT		4
// Character waiting integer default value as definded by ISO
#define T1_CWI_DEFAULT		13

typedef enum _OP_MODE
{
	OP_SPECIFIC,
	OP_NEGOTIABLE,
}OP_MODE_T;

typedef struct _CLOCK_RATE_CONVERSION {

	const unsigned long F;
	const unsigned long fs; 

} CLOCK_RATE_CONVERSION, *PCLOCK_RATE_CONVERSION;

typedef struct _BIT_RATE_ADJUSTMENT {

	const unsigned long DNumerator;
	const unsigned long DDivisor;

} BIT_RATE_ADJUSTMENT, *PBIT_RATE_ADJUSTMENT;

typedef struct _PTS_DATA {
#define PTS_TYPE_DEFAULT 0x00
#define PTS_TYPE_OPTIMAL 0x01
#define PTS_TYPE_USER    0x02

    unsigned char Type;  

    unsigned char Fl;    // Fl value for PTS

    unsigned char Dl;     	    // Dl value for PTS

    unsigned long  CLKFrequency;    // New clock frequency

    unsigned long DataRate;    // New baud rate to be used after pts

    unsigned char StopBits;    // new stop bits to be used after pts
} PTS_DATA, *PPTS_DATA;

typedef struct _SCARD_CARD_CAPABILITIES
{
	// Flag that indicates that the current card uses invers convention
	int InversConvention;

	// Calculated etu 
	unsigned long	etu;
      
    // Answer To Reset string returned by card.
	struct {
		unsigned char Buffer[64];
		unsigned char Length;
	} ATR;

	struct {
		unsigned char Buffer[16];
		unsigned char Length;
	} HistoricalChars;

	PCLOCK_RATE_CONVERSION 	ClockRateConversion;
	PBIT_RATE_ADJUSTMENT 	BitRateAdjustment;

	unsigned char FI;	// Clock rate conversion 
	unsigned char DI;	// Bit rate adjustment
	unsigned char II;	// Maximum programming current
	unsigned char P;	// Programming voltage in .1 Volts
	unsigned char N;	// Extra guard time in etu 
	unsigned long GT;// Calculated guard time in micro seconds

	struct {
		unsigned long Supported;		// This is a bit mask of the supported protocols
		unsigned long Selected;		// The currently selected protocol
	} Protocol;

	// T=0 specific data
	struct {
		unsigned char WI;		// Waiting integer
		unsigned long WT;		// Waiting time in micro seconds
	} T0;

	// T=1 specific data
	struct {
		unsigned char IFSC;		// Information field size of card

		// Character waiting integer and block waiting integer
		unsigned char CWI;
		unsigned char BWI;

		unsigned char EDC;		// Error detection code

		// Character and block waiting time in micro seconds
		unsigned long CWT;
		unsigned long BWT;

		unsigned long BGT;		// Block guarding time in micro seconds
	} T1;

    PTS_DATA PtsData;

    unsigned char Reserved[100 - sizeof(PTS_DATA)];

} SCARD_CARD_CAPABILITIES, *PSCARD_CARD_CAPABILITIES;


//=======IOCTL COMMANDS==============

#define INNO_IOC_MAGIC		'i'
#define INNO_GET_INTR_TYPE			_IOR(INNO_IOC_MAGIC, 1, int)
#define INNO_MMIS_READ				_IOWR(INNO_IOC_MAGIC, 2, int)
#define INNO_MMIS_WRITE				_IOWR(INNO_IOC_MAGIC, 3, int)
#define INNO_MMIS_CMD				_IOWR(INNO_IOC_MAGIC, 4, int)
#define INNO_READ_REG				_IOWR(INNO_IOC_MAGIC, 5, int)
#define INNO_WRITE_REG				_IOWR(INNO_IOC_MAGIC, 6, int)
#define INNO_SCAN_FREQUENCY			_IOWR(INNO_IOC_MAGIC, 7, int)
#define INNO_SCAN_FREQUENCY_DOT		_IOWR(INNO_IOC_MAGIC, 8, int)
#define INNO_SET_FREQUENCY			_IOWR(INNO_IOC_MAGIC, 9, int)
#define INNO_SET_FREQUENCY_DOT		_IOWR(INNO_IOC_MAGIC, 10, int)
#define INNO_SET_CHANNEL_CONFIG		_IOWR(INNO_IOC_MAGIC, 11, int)
#define INNO_GET_CHANNEL_CONFIG		_IOWR(INNO_IOC_MAGIC, 12, int)
#define INNO_GET_SYS_STATE			_IOWR(INNO_IOC_MAGIC, 13, int)
#define INNO_SET_PM					_IOWR(INNO_IOC_MAGIC, 14, int)
#define INNO_GET_FW_VER				_IOWR(INNO_IOC_MAGIC, 15, int)
#define INNO_FW_DOWNLOAD			_IOWR(INNO_IOC_MAGIC, 16, int)
#define INNO_SET_CP_TYPE			_IOWR(INNO_IOC_MAGIC, 17, int)
#define INNO_UPDATE_FW				_IOWR(INNO_IOC_MAGIC, 22, int)
#define INNO_GET_BUFFERLEN			_IOR(INNO_IOC_MAGIC, 23, int)
#define INNO_ENABLE_IRQ				_IOWR(INNO_IOC_MAGIC, 24, int)
#define INNO_DISABLE_IRQ			_IOWR(INNO_IOC_MAGIC, 25, int)
#define INNO_SHUTDOWN_IRQ			_IOWR(INNO_IOC_MAGIC, 26, int)

#define INNO_UAM_SET_CMD			_IOW(INNO_IOC_MAGIC, 31, int)
#define INNO_UAM_SET_CMDL			_IOW(INNO_IOC_MAGIC, 32, int)
#define INNO_UAM_READ_RES			_IOR(INNO_IOC_MAGIC, 33, int)
#define INNO_UAM_READ_STATUS			_IOR(INNO_IOC_MAGIC, 34, int)
#define INNO_UAM_READ_STATUS_LEN                _IOR(INNO_IOC_MAGIC, 35, int)

/*
 * IOCTL COMMAND PARAMETER
 */
//INNO_READ_REG & INNO_WRITE_REG command  parameter
typedef struct inno_reg_data{	
	u16 reg;
	u16 data;
}inno_reg_t;

// INNO_SCAN_FREQUENCY & INNO_SCAN_FREQUENCY_DOT command parameter
struct inno_freq_scan_area {
	u16		start;
	u16		end;
};

// INNO_SET_CP_TYPE	command parameter
#define CP_TYPE_MASK 	0x1
enum inno_cp_type {
	SHORT_CP = 0,
	LONG_CP,
	CP_40TS = SHORT_CP,
	CP_36TS = LONG_CP
};

// INNO_SET_PM command parameter
struct inno_pm_control {
	u8	sys_pause:1;	// System
	u8	stun_off:1;		// S-Tuner
	u8	utun_off:1;		// U-Tuner
	u8	tspwr_off:1;	// TS Power
};

// INNO_SET_CHANNEL_CONFIG	command parameter
#define CFG_LGCH0				0x00
#define CFG_LGCH1				0x01
#define	CFG_CLOSE_LGCH			0x01
#define CFG_TS_INVALID			0x3F
#define MODULATE_BPSK			0x00
#define MODULATE_QPSK			0x01
#define MODULATE_16QAM			0x02
#define MODULATE_RESERVED		0x03
#define LDPC_1DB2				0x00		// 1/2 LDPC
#define LDPC_3DB4				0x01		// 3/4 LDPC
#define ITLV_RESERVED			0x00
#define ITLV_MODE1				0x01
#define ITLV_MODE2				0x02
#define ITLV_MODE3				0x03
#define RS_240_240				0x00	//RS(240, 240)
#define RS_240_224				0x01	//RS(240, 224)
#define RS_240_192				0x02	//RS(240, 192)
#define RS_240_176				0x03	//RS(240, 176)
struct inno_channel_config{
	u8	ch_id:1;		//0--logic channel 0, 1--logic channel 1
	u8	ch_close:1;		//1--close this logic channel
	u8	ts_start:6;
	u8	ts_count;	
	u8	ldpc:2;
	u8 	itlv:2;
	u8	rs:2;
	u8	modulate:2;
};


// INNO_GET_SYS_STATE command parameter
struct inno_sys_state{
	u8	stattype;
	union{
		u8	sync_state;
		u32	signal_strength;
		u32	ldpc_total_count;
		u32 ldpc_err_count;
		u32	rs_total_count;
		u32	rs_err_count;
	}statdata;
};

// INNO_GET_FW_VER command parameter
struct inno_fw_ver {
	u8	major;
	u8	minor;
};

// INNO_FW_DOWNLOAD command parameter
struct inno_fw_bin {
	u8	data[1024*24];
};

// INNO_UAM_TRANSFER command parameter
struct inno_uam_parameter {
	unsigned char pBufIn[260];
	unsigned int bufInLen;
	unsigned char pBufOut[256];
	unsigned int pBufOutLen;
	unsigned short sw;
};

#endif
