/************************************************
 * FileName:	inno_core.c
 *
 * Description:
 *		core driver for innofidei if101 v2 chip.
 * Author:		
 *		Jimmy.li <lizhengming@innofidei.com>
 * Created:
 * 		Dec.14, 2007
 * History:
 * 		Dec.25, 2007	-- 		add AP Interface Functions
 *		Jan.3, 2008		--		add inno_download_fireware function.
 *		Jan.7, 2008		--		add inno_set_cp_type function.
 *		Jan.11, 2008	--		implement inno_download_fireware function.
 * Copyright (c) 2007, Innofidei Inc
 *
 ***********************************************/
 
#include <linux/init.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/fs.h>
#include <linux/sched.h>
#include <linux/mm.h>
#include <asm/page.h>
#include <linux/kdev_t.h>
#include <linux/semaphore.h>
#include <linux/slab.h>
#include <linux/delay.h>
#include <asm/uaccess.h>
#include <asm/irq.h>
#include <linux/interrupt.h>
#include <linux/workqueue.h>
#include <linux/vmalloc.h>
#include <linux/device.h>


#include "if208.h"
#include "if101_control_proto.h"
#include "mmis_cmdlist.h"

#define MAKEWORD(low,high)	((unsigned short)((unsigned char)(low)) | (((unsigned short)(unsigned char)(high))<<8))

DECLARE_MUTEX(inno_spi_mutex);
DECLARE_MUTEX(inno_iic_mutex);

EXPORT_SYMBOL(inno_spi_mutex);
EXPORT_SYMBOL(inno_iic_mutex);

extern inno_i2c_driver_t 	inno_i2c_driver;
extern inno_spi_driver_t 	inno_spi_driver;
extern inno_cmdset_t        inno_cmdset_ops;

/* global variable */
struct inno_device  g_inno_dev = {
	.cfg		= {
					.spi_interface 	= NORMAL_SPI,
					.spi_clk_mode 	= SPI_CLK_MODE_0,
					.intr_type		= FALLING_EDGE,
					.trans_size		= TRANS_BURST,
					.i2c_cmd_mode	= 0,
				  },
	.i2c_driver		=	&inno_i2c_driver,
	.spi_driver		=	&inno_spi_driver,
	.cmdset         =   &inno_cmdset_ops,
};

static struct inno_fw_ver g_fw_ver = {0, 0};

//#define	IRQ_BUG_DEBUG
#ifdef	IRQ_BUG_DEBUG
static int irq_bug_debug = 0;
#endif
static struct work_struct	irq_work;
static struct workqueue_struct	*irq_wq;

/*==================== UAM =============*/
OP_MODE_T	OperationMode_UAM;
SCARD_CARD_CAPABILITIES  cardCapabilities_UAM;

/*========================
 * Interrupt relate
 *=======================*/
extern int inno_irq_setup(irqreturn_t (*handler)(int, void*), int irq_type);
extern void inno_irq_release(void);

static void inno_delay_ms(int ms);
static int INNO_Send_UAM_Cmd(unsigned char *cmd, int cmdLen);
static int INNO_Read_UAM_Rsp(unsigned char *rsp, unsigned int *rsplen);
static int inno_init_uam(void);
int inno_spi_r(unsigned char cmd, unsigned char *rsp);

static irqreturn_t inno_intr_handler(int irq, void* device)
{
#ifdef	IRQ_BUG_DEBUG
	if(irq_wq)
	{
		irq_bug_debug++;
		if((irq_bug_debug>=120) && ((irq_bug_debug%60)==0))
		{
			inno_err("inno_intr_handler: irq worker no response, count=%d!", irq_bug_debug);
			show_state();
		}
		//BUG_ON(irq_bug_debug>100);
	}
#endif

	inno_dbg("--- IF208 hardware interrupt trigger ---");	
	if(irq_wq)
		queue_work(irq_wq, &irq_work);

	return IRQ_HANDLED;
}


static int inno_do_request(int lgx_id)
{
	int ret = 0;
	struct inno_lgx* lgx = g_inno_dev.lgxs[lgx_id];
	struct inno_buffer* buffer = &lgx->inno_buffer;
	u32 size;
	inno_cmdset_t* mmis_cmd = g_inno_dev.cmdset;

	down(&buffer->sem);
	buffer->valid = 0;
	up(&buffer->sem);

	// get logic channel available data length
	size = mmis_cmd->get_lgx_length(lgx_id);
	inno_dbg("%s size = 0x%x, bufsize: 0x%x", lgx->ids->name, size, buffer->bufsize);
	if(size > (buffer->bufsize)){
		inno_err("error: %s data size 0x%x, buffer is not enough.", lgx->ids->name, size);
		return -EINVAL;
	}
	
	down(&lgx->sem);
	lgx->bytes_available = size;
	up(&lgx->sem);

	down(&buffer->sem);
	mmis_cmd->fetch_lgx_data(lgx_id, (u8*)buffer->start, size);	
	buffer->valid = 1;
	buffer->valid_size = size;
	buffer->own_id = lgx_id;
	up(&buffer->sem);


	wake_up_interruptible(&lgx->read_wait);
	
	return ret;
}

static int inno_uam_request(void)
{
	int ret = 0;
	struct inno_lgx* lgx = g_inno_dev.lgxs[UAM];
	struct inno_buffer* buffer = &lgx->inno_buffer;
	u32 size;
	inno_cmdset_t* mmis_cmd = g_inno_dev.cmdset;

	down(&buffer->sem);
	buffer->valid = 0;
	up(&buffer->sem);

	// get logic channel available data length
	size = mmis_cmd->get_lgx_length(UAM);
	inno_dbg("UAM size = 0x%x, bufsize: 0x%x", size, buffer->bufsize);
	if(size > (buffer->bufsize)){
		inno_err("error: %s data size 0x%x, buffer is not enough.", lgx->ids->name, size);
		return -EINVAL;
	}
	
	down(&lgx->sem);
	lgx->bytes_available = size;
	up(&lgx->sem);

	down(&buffer->sem);
	mmis_cmd->fetch_lgx_data(UAM, (u8*)buffer->start, size);
	buffer->valid = 1;
	buffer->valid_size = size;
	buffer->own_id = UAM;
	up(&buffer->sem);
	
	wake_up_interruptible(&lgx->uam_wait);
	
	return ret;
}


/*==========================
 *  Kernel Thread
 *=========================*/

/*
 * irq thread ( CALL THREAD-1)
 *
 */
static void irq_work_handler(struct work_struct *_work)
{
	inno_device_t* inno_dev = &g_inno_dev;
	inno_cmdset_t* mmis_cmd = inno_dev->cmdset;
	unsigned char status = 0;
	unsigned char channel = 0;

#ifdef	IRQ_BUG_DEBUG
	irq_bug_debug=0;
#endif

	//inno_msg("irq_work_handler: enter");
	inno_spi_r(0x20, &status);
	inno_spi_r(0x04, &channel);
	if (20 == channel) {
		inno_msg("RS error[530MHz]: %d%%", status);
	} else if (32 == channel) {
		inno_msg("RS error[666MHz]: %d%%", status);
	} else if (43 == channel) {
		inno_msg("RS error[754MHz]: %d%%", status);
	} else {
		inno_dbg("No validate value");
	}
	
//	down(&inno_dev->sem);
	{
		u8 intr_status;
		
		intr_status = mmis_cmd->get_intr_status();
		inno_msg("======== intr_status [%x] ============", intr_status);	
		
		// we only interest in opened logic channel
		if(g_inno_dev.lgxs[UAM] && (intr_status & UAM_DATA_RDY)){
			//inno_msg("UAM data ready");
			inno_uam_request();
		}

		if(g_inno_dev.lgxs[LG0] && (intr_status & LG0_DATA_RDY)){
			//inno_msg("LG0 data ready");
			inno_do_request(LG0);
		}

		if(g_inno_dev.lgxs[LG1] && (intr_status & LG1_DATA_RDY)){
			//inno_msg("LG1 data ready");
			inno_do_request(LG1);
		}
	}
//	up(&inno_dev->sem);

	//inno_msg("irq_work_handler: leave");
}

extern void inno_chip_power_on(void);
extern void inno_chip_power_off(void);
extern int inno_check_firmware(void);

/*
 * return 0 -- successful
 */
int innodev_startup(inno_device_t* inno_dev)
{
	int ret = 0;
	int ret_uam = 0;
	struct inno_channel_config cfg;
//	int i = 0;
	
	down(&inno_dev->sem);

	inno_msg("====== innodev_startup ===== use count = %d", inno_dev->use_count);
	if(inno_dev->use_count++ != 0)
		goto out;

	inno_msg("!! startup innodev !!");

	/* power on if101 chip */
	inno_chip_power_on();

#ifndef CONFIG_INNODEV_SPIONLY
	//initialize i2c driver & spi driver
	if(inno_dev->i2c_driver && inno_dev->i2c_driver->init)
	{
		ret = inno_dev->i2c_driver->init();
		if(ret)
			goto out;
	}
#endif
	if(inno_dev->spi_driver && inno_dev->spi_driver->init)
		inno_dev->spi_driver->init(INNO_SPI_MASTER);

#if 0 //+Mason Test 20091223
		inno_msg("========= Begin Mason Test FW =========="); 

		while(1){
			inno_chip_power_on();
			mdelay(500);
			inno_check_firmware();
			if( inno_get_fireware_version(&g_fw_ver)==INNO_SUCCESS )
				inno_msg("firmware version %d.%d", g_fw_ver.major, g_fw_ver.minor);
			else inno_msg("get fw version fail");
			inno_chip_power_off();
			inno_msg("i=%d",++i);
			mdelay(500);
		}
		
		inno_msg("========= End Mason Test FW =========="); 
#endif //-Mason Test 20091223

	ret = inno_check_firmware();
	if( ret<0 )
		goto err2;

	if((g_fw_ver.major==0) && (g_fw_ver.minor==0)){
		if( inno_get_fireware_version(&g_fw_ver)==INNO_SUCCESS )
			inno_msg("firmware version %d.%d", g_fw_ver.major, g_fw_ver.minor);
	}

#ifdef	IRQ_BUG_DEBUG
	irq_bug_debug=0;
#endif
	INIT_WORK(&(irq_work), irq_work_handler);
	// start irq thread
	irq_wq = create_workqueue("inno_irqtask");
	if(!irq_wq){
		ret = -ENOMEM;
		inno_err("inno_startup: create_workqueue failed");
		goto err2;
	}

	ret = inno_irq_setup(inno_intr_handler, inno_dev->cfg.intr_type);
	if(ret<0){
		inno_err("inno_irq_setup return failed");
		goto err1;
	}

    // AAAAAAAA 2009/11/02
#if 1
    inno_delay_ms(100);
    cfg.ch_id = 0;
    cfg.ch_close = 1;
    cfg.ts_start = 0;
    cfg.ts_count = 0;
    cfg.ldpc = 0;
    cfg.itlv = 0;
    cfg.rs = 0;
    cfg.modulate = 0;
    inno_set_channel_config(&cfg);
    inno_delay_ms(1000);

	if(inno_dev->lgxs[UAM]!= NULL){
	    ret_uam = inno_init_uam();
	    if(ret_uam != INNO_SUCCESS)
	        inno_err("Init UAM fail\n");
	    else
	        inno_msg("Init UAM OK");
	}

    inno_delay_ms(100);
#endif

	goto out;
	
err1:
	if(irq_wq){
		destroy_workqueue(irq_wq);
		irq_wq = NULL;
	}

err2:
#ifndef CONFIG_INNODEV_SPIONLY
	inno_dev->i2c_driver->exit();
#endif
	inno_dev->spi_driver->uninit();
	inno_chip_power_off();

out:
	if(ret)
		inno_dev->use_count--;
	
	up(&inno_dev->sem);

	return ret;
}

int innodev_shutdown(inno_device_t* inno_dev)
{
	down(&inno_dev->sem);

	inno_msg("======innodev_shutdown ===== use count = %d", inno_dev->use_count);
	if(!(--inno_dev->use_count)){
		inno_msg("!! shutdown innodev !!");

		inno_irq_release();
		inno_msg("!! inno_irq_release !!");

		/* Clean up the kernel thread */
		if(irq_wq){
			destroy_workqueue(irq_wq);
			irq_wq = NULL;
		}
		inno_msg("!! destroy_workqueue !!");

#ifndef CONFIG_INNODEV_SPIONLY
		// uninitialize i2c driver & spi driver
		if(inno_dev->i2c_driver && inno_dev->i2c_driver->exit)
			inno_dev->i2c_driver->exit();
#endif
		if(inno_dev->spi_driver && inno_dev->spi_driver->uninit)
			inno_dev->spi_driver->uninit();
		inno_msg("!! spi_driver deinit !!");

		// cut down if101 chip power supply
		inno_chip_power_off();
	}

	up(&inno_dev->sem);
	return 0;
}
EXPORT_SYMBOL(innodev_startup);
EXPORT_SYMBOL(innodev_shutdown);


/*============================================
 * inno device initialization/uninitialization 
 *==========================================*/

int innodev_init(void)
{
	int ret = 0;
	inno_msg("======innodev_init ===== use count = %d", g_inno_dev.use_count);
	init_MUTEX(&g_inno_dev.sem);

	down(&g_inno_dev.sem);
	g_inno_dev.use_count = 0;
	up(&g_inno_dev.sem);

	return ret;
}

void innodev_uninit(void)
{
	inno_msg("======innodev_uninit ===== use count = %d", g_inno_dev.use_count);
	if(!g_inno_dev.use_count)
		innodev_shutdown(&g_inno_dev);
}

EXPORT_SYMBOL(innodev_init);
EXPORT_SYMBOL(innodev_uninit);


/**************************************
 *
 * IF101 & AP Communication API
 * 
 * ***********************************/

static void inno_delay_ms(int ms)
{
	if( ms>=10 ){
		msleep(ms);
	}else
		mdelay(ms);
}

#ifdef CONFIG_INNODEV_SPIONLY
static int inno_spi_w(unsigned char cmd, unsigned char data)
{
	g_inno_dev.spi_driver->write(&cmd, 1);
	g_inno_dev.spi_driver->write(&data, 1);
	return 0;
}

int inno_spi_r(unsigned char cmd, unsigned char *rsp)
{
	g_inno_dev.spi_driver->write(&cmd, 1);
	g_inno_dev.spi_driver->read(rsp, 1);
	return 0;
}

static int inno_spi_cmd_read(unsigned char cmd, unsigned char *rsp)
{
	int ret;
	down(&inno_spi_mutex);
	ret = inno_spi_r(cmd, rsp);
	up(&inno_spi_mutex);
	return ret;
}
#endif

/*
 * Description:
 * 		Send one byte of command frame to AIF(AP interface)
 * Parameters:
 * 		@cmd_filed	-	command frame field value
 * 		@type		-	the type of @cmd_field, should be CMD_CODE, CMD_DATA, CMD_CHECKSUM
 */
static int inno_i2c_send_cmd_byte(u8 cmd_field, u8 type)
{
#ifdef CONFIG_INNODEV_SPIONLY
	unsigned char status = 0;

	int j = 0;
	for(j = 0; j<5000; j++){
		inno_spi_r(SPI_R_CMD_STATUS_ADDR, &status);
		if((status & CMD_BUSY) == 0){
			inno_spi_w(SPI_W_CMD_DATA_ADDR, cmd_field);
			inno_spi_w(SPI_W_CMD_STATUS_ADDR, type);
			return INNO_SUCCESS;
		}
	}
#else
	inno_i2c_driver_t*	i2c_driver = g_inno_dev.i2c_driver;
	u8 cmd_status = 0;

	int i = 0;
	for(i=0; i<100; i++){
		i2c_driver->read(I2C_CMD_STATUS_REG, &cmd_status);
		if((cmd_status & CMD_STATUS_HDEN)==0){	//HDEN==0
			i2c_driver->write(I2C_CMD_DATA_REG, cmd_field);
			i2c_driver->write(I2C_CMD_STATUS_REG, type);	// set HDEN
			return INNO_SUCCESS;
		}
		inno_delay_ms(1);
	}
#endif

	inno_err("send cmd byte timeout error");
	return -INNO_TIMEOUT;
}

/*
 * Description:
 *		Receive one byte of response frame from AIF (AP Interface)
 * Parameters:
 * 		@response	-	store one byte response value.
 */
static int inno_i2c_read_response_byte(u8* response)
{
#ifdef CONFIG_INNODEV_SPIONLY
	unsigned char Local_Sen_Status = 0;
	int j = 0;

	for(j=0; j<10000; j++){
		inno_spi_r(SPI_R_RSP_STATUS_ADDR, &Local_Sen_Status);
		if((Local_Sen_Status & RSP_DATA_VALID) == RSP_DATA_VALID){
			inno_spi_r(SPI_R_RSP_DATA_ADDR, response);
			inno_spi_w(SPI_W_RSP_STATUS_ADDR, 0x0);
			return INNO_SUCCESS;
		}		
	}
#else
	inno_i2c_driver_t* i2c_driver = g_inno_dev.i2c_driver;
	u8 response_status = 0;
	
	int i = 0;
	for(i=0; i<100; i++){
		i2c_driver->read(I2C_RESPONSE_STATUS_REG, &response_status);
		if(response_status & RESPONSE_STATUS_CDEN){	//CDEN==1
			i2c_driver->read(I2C_RESPONSE_DATA_REG, response);
			i2c_driver->write(I2C_RESPONSE_STATUS_REG, 0);	// clear CDEN
			return INNO_SUCCESS;
		}
		inno_delay_ms(1);
	}
#endif
	
	inno_err("read respone byte timeout error");
	return -INNO_TIMEOUT;
}

/*
 * cmd_frame contain 8bytes, which including 1byte CMD_CODE, 6bytes CMD_DATA, and 1byte CMD_CHECKSUM
 */
static int inno_i2c_send_cmd_frame(struct inno_cmd_frame* cmd_frame)
{
	u8 ret = INNO_SUCCESS;
	int i;
	unsigned char end;
/*
#ifdef CONFIG_INNODEV_SPIONLY
	down(&inno_spi_mutex);
#else
	down(&inno_iic_mutex);
#endif
*/
	ret = inno_i2c_send_cmd_byte(cmd_frame->code, CMD_FIELD_CODE);
	if(INNO_SUCCESS != ret)	
		goto out;

	for(i=0; i<6; i++){
		ret = inno_i2c_send_cmd_byte(cmd_frame->data[i], CMD_FIELD_DATA);
		if(INNO_SUCCESS != ret)	
			goto out;
	}
	cmd_frame->checksum = cmd_frame_checksum(cmd_frame);
	ret = inno_i2c_send_cmd_byte(cmd_frame->checksum, CMD_FIELD_CHECKSUM);
	if(INNO_SUCCESS != ret)	
		goto out;

	//inno_delay_ms(200);
	ret = inno_i2c_read_response_byte(&end);
	if(INNO_SUCCESS != ret)	
		goto out;
	if(end == 0xF9)
		ret = INNO_SUCCESS;
	else{
		inno_err("FW response 0x%x error", end);
		ret = -INNO_ERROR;
	}

out:
/*
#ifdef CONFIG_INNODEV_SPIONLY
	up(&inno_spi_mutex);
#else
	up(&inno_iic_mutex);
#endif
*/
	return ret;
}

/*
 * response_frame contain 8bytes, which including 1byte SENSE_CODE, 6bytes SENSE_DATA, and 1byte CRC_CHECKSUM.
 */
static int inno_i2c_read_response_frame(struct inno_response_frame* resp_frame)
{
	u8 ret = INNO_SUCCESS;
	int i = 0;
/*
#ifdef CONFIG_INNODEV_SPIONLY
	down(&inno_spi_mutex);
#else
	down(&inno_iic_mutex);
#endif
*/
	ret = inno_i2c_read_response_byte(&resp_frame->code);
	if(INNO_SUCCESS != ret)
		goto out;

	for(i=0; i<6; i++){
		ret = inno_i2c_read_response_byte(&resp_frame->data[i]);
		if(INNO_SUCCESS != ret)
			goto out;
	}
	ret = inno_i2c_read_response_byte(&resp_frame->checksum);
	if(INNO_SUCCESS != ret)
		goto out;
	if(response_frame_checksum(resp_frame) != resp_frame->checksum)
		ret = -INNO_ERROR;

out:
	/*
#ifdef CONFIG_INNODEV_SPIONLY
	up(&inno_spi_mutex);
#else
	up(&inno_iic_mutex);
#endif
*/
	return ret;

}

#ifndef CONFIG_INNODEV_SPIONLY
static int inno_i2c_reg_read(u16 reg, u8 *data)
{
	int ret;
	down(&inno_iic_mutex);
	ret = g_inno_dev.i2c_driver->read(reg, data);
	up(&inno_iic_mutex);
	return ret;
}
#endif

static void dump_cmd_frame(struct inno_cmd_frame* cmd_frame)
{
	inno_msg("cmd code=0x%x, data=0x%x 0x%x 0x%x 0x%x 0x%x 0x%x, checksum=%x", \
				cmd_frame->code, \
				cmd_frame->data[0], \
				cmd_frame->data[1], \
				cmd_frame->data[2], \
				cmd_frame->data[3], \
				cmd_frame->data[4], \
				cmd_frame->data[5], \
				cmd_frame->checksum);
}

static void dump_response_frame(struct inno_response_frame* resp_frame)
{
	inno_msg("resp code=0x%x, data=0x%x 0x%x 0x%x 0x%x 0x%x 0x%x, checksum=%x", \
				resp_frame->code, \
				resp_frame->data[0], \
				resp_frame->data[1], \
				resp_frame->data[2], \
				resp_frame->data[3], \
				resp_frame->data[4], \
				resp_frame->data[5], \
				resp_frame->checksum);
}

/*=====================================
 *
 * API (AP Interface) Functions
 *
 * ===================================*/

/*
 */
int inno_get_channel_config(struct inno_channel_config* cfg)
{
	struct inno_cmd_frame 		cmd_frame = {0};
	struct inno_response_frame	resp_frame = {0};
	int ret;

	down(&inno_spi_mutex);
	cmd_frame.code  = CMD_GET_CONFIG;

	ret = inno_i2c_send_cmd_frame(&cmd_frame);
	if(INNO_SUCCESS != ret)
		goto out;
	ret = inno_i2c_read_response_frame(&resp_frame);
	if(INNO_SUCCESS != ret)
		goto out;
/*
 |---------------------------------------|
 | B7 | B6 | B5 | B4 | B3 | B2 | B1 | B0 |
 |---------------------------------------|
 |MODULATE |   RS    |  ITLV   |  LPDC   |     
 |---------------------------------------|
*/
#define LDPC_MASK		(0x03<<0)
#define ITLV_MASK		(0x03<<2)
#define RS_MASK			(0x03<<4)
#define MODULATE_MASK	(0x03<<6)
	switch(cfg->ch_id){
		case CFG_LGCH0:
		{
			cfg->ts_start = resp_frame.data[0];
			cfg->ts_count = resp_frame.data[1];
			cfg->ch_close = (cfg->ts_count==0)?1:0;
			cfg->ldpc	  = (resp_frame.data[2] & LDPC_MASK);
			cfg->itlv	  = (resp_frame.data[2] & ITLV_MASK)>>2;
			cfg->rs	  	  = (resp_frame.data[2] & RS_MASK)>>4;
			cfg->modulate = (resp_frame.data[2] & MODULATE_MASK)>>6;
			
			break;
		}
		case CFG_LGCH1:
		{
			cfg->ts_start = resp_frame.data[3];
			cfg->ts_count = resp_frame.data[4];
			cfg->ch_close = (cfg->ts_count==0)?1:0;
			cfg->ldpc	  = (resp_frame.data[5] & LDPC_MASK);
			cfg->itlv	  = (resp_frame.data[5] & ITLV_MASK)>>2;
			cfg->rs	  	  = (resp_frame.data[5] & RS_MASK)>>4;
			cfg->modulate = (resp_frame.data[5] & MODULATE_MASK)>>6;
			break;
		}
	}
out:
	up(&inno_spi_mutex);
	return ret;

}


/*
 * Set logic channel configure
 */
int inno_set_channel_config(struct inno_channel_config* cfg)
{
	struct inno_cmd_frame cmd_frame = {0};	
	int ret;
	int retry = 0;
	
	if(!cfg){
		ret = INNO_PARAM_ERR;
		goto out;
	}
	cmd_frame.code = CMD_SET_LG_CONFIG;

	if(cfg->ch_close){	//shut down logic channel
		switch(cfg->ch_id){
			case CFG_LGCH0:
				cmd_frame.data[1] = 0;	//close lg0
				cmd_frame.data[3] = CFG_TS_INVALID; 	//ignore lg1
				break;
			case CFG_LGCH1:
				cmd_frame.data[4] = 0;	//close lg1
				cmd_frame.data[0] = CFG_TS_INVALID;	//ignore lg0
				break;
		}
	}else{
		switch(cfg->ch_id){ 
			case CFG_LGCH0:
				cmd_frame.data[0] = cfg->ts_start;
				cmd_frame.data[1] = cfg->ts_count;
				cmd_frame.data[2] = (cfg->modulate<<6) | (cfg->rs <<4) | (cfg->itlv<<2) | cfg->ldpc;
				cmd_frame.data[3] = CFG_TS_INVALID; 	//ignore lg1
				break;
			case CFG_LGCH1:
				cmd_frame.data[0] = CFG_TS_INVALID;	//ignore lg0
				cmd_frame.data[3] = cfg->ts_start;
				cmd_frame.data[4] = cfg->ts_count;
				cmd_frame.data[5] = (cfg->modulate<<6) | (cfg->rs <<4) | (cfg->itlv<<2) | cfg->ldpc;
				break;
		}
	}
	down(&inno_spi_mutex);
	for(retry = 0; retry < 5; retry ++){
		ret = inno_i2c_send_cmd_frame(&cmd_frame);
		if(ret == INNO_SUCCESS)
			break;
		inno_msg("inno_set_channel_config retry=%d\n", retry); 
	}
	up(&inno_spi_mutex);
	
	//dump_cmd_frame(&cmd_frame);

	//inno_delay_ms(200);

out:
	return ret;
}

/*
 * Get system information
 */
int inno_get_system_state(struct inno_sys_state* state)
{
#ifdef CONFIG_INNODEV_SPIONLY
	unsigned char data;

	unsigned char reg_0 = 0;	// 0-7bit
	unsigned char reg_1 = 0;	// 8-15bit
	unsigned char reg_2 = 0;	// 16-23bit
	unsigned char reg_3 = 0;	// 24-31bit
//AAAAA
	switch(state->stattype){
		case STATTYPE_SYNC_STATE:
			inno_spi_cmd_read(SPI_R_STATUS, &data);
			state->statdata.sync_state = (data & 0x80) >> 7;
			inno_msg("sync_state = %x", state->statdata.sync_state);
			break;
		case STATTYPE_SIGNAL_STRENGTH:
			inno_spi_cmd_read(SPI_R_STATUS, &data);
			state->statdata.signal_strength = data & 0x7F;
			inno_msg("signal_strength = %d", state->statdata.signal_strength);
			break;
		case STATTYPE_LDPC_TOTAL_COUNT:
			inno_spi_cmd_read(SPI_R_LDPC_Total_0, &reg_0);
			inno_spi_cmd_read(SPI_R_LDPC_Total_1, &reg_1);
			inno_spi_cmd_read(SPI_R_LDPC_Total_2, &reg_2);
			inno_spi_cmd_read(SPI_R_LDPC_Total_3, &reg_3);
			state->statdata.ldpc_total_count = (reg_3 << 24) | (reg_2 << 16) | (reg_1 << 8) | reg_0;
			inno_msg("ldpc_total_count = %x", state->statdata.ldpc_total_count);
			break;
		case STATTYPE_LDPC_ERR_COUNT:
			inno_spi_cmd_read(SPI_R_LDPC_Error_0, &reg_0);
			inno_spi_cmd_read(SPI_R_LDPC_Error_1, &reg_1);
			inno_spi_cmd_read(SPI_R_LDPC_Error_2, &reg_2);
			inno_spi_cmd_read(SPI_R_LDPC_Error_3, &reg_3);
			state->statdata.ldpc_err_count = (reg_3 << 24) | (reg_2 << 16) | (reg_1 << 8) | reg_0;
			inno_msg("ldpc_err_count = %x", state->statdata.ldpc_err_count);
			break;
		case STATTYPE_RS_TOTAL_COUNT:
			inno_spi_cmd_read(SPI_R_RS_Total_0, &reg_0);
			inno_spi_cmd_read(SPI_R_RS_Total_1, &reg_1);
			inno_spi_cmd_read(SPI_R_RS_Total_2, &reg_2);
			inno_spi_cmd_read(SPI_R_RS_Total_3, &reg_3);
			state->statdata.rs_total_count = (reg_3 << 24) | (reg_2 << 16) | (reg_1 << 8) | reg_0;
			inno_msg("rs_total_count = %x", state->statdata.rs_total_count);
			break;
		case STATTYPE_RS_ERR_COUNT:
			inno_spi_cmd_read(SPI_R_RS_Error_0, &reg_0);
			inno_spi_cmd_read(SPI_R_RS_Error_1, &reg_1);
			inno_spi_cmd_read(SPI_R_RS_Error_2, &reg_2);
			inno_spi_cmd_read(SPI_R_RS_Error_3, &reg_3);
			state->statdata.rs_err_count = (reg_3 << 24) | (reg_2 << 16) | (reg_1 << 8) | reg_0;
			inno_msg("rs_err_count = %x", state->statdata.rs_err_count);
			break;
		default:
			return -EINVAL;
	}
	return 0;
#else
	struct inno_cmd_frame cmd_frame = {0};
	struct inno_response_frame resp_frame = {0};
	int ret;
	unsigned char data;

	if(state->stattype==STATTYPE_SYNC_STATE){
		inno_i2c_reg_read(0xCC, &data);
		state->statdata.sync_state = (data == 0x8);
		inno_msg("sync_state = %d", state->statdata.sync_state);
		return 0;
	}
	else if(state->stattype==STATTYPE_SIGNAL_STRENGTH){
		inno_i2c_reg_read(0xBE, &data);
		state->statdata.signal_strength = data;
		inno_msg("signal_strength = %d", state->statdata.signal_strength);
		return 0;
	}

	cmd_frame.code = CMD_GET_ERROR_INFO;
	cmd_frame.data[0] = state->stattype;

	ret = inno_i2c_send_cmd_frame(&cmd_frame);
	if(INNO_SUCCESS != ret)
		goto out;
	
	dump_cmd_frame(&cmd_frame);
	ret = inno_i2c_read_response_frame(&resp_frame);
	dump_response_frame(&resp_frame);
	if(INNO_SUCCESS != ret){
		inno_err("%s : read response error", __FUNCTION__);
		goto out;
	}
	switch(resp_frame.data[0]){
		case STATTYPE_SYNC_STATE:
			state->statdata.sync_state = resp_frame.data[1];
			inno_msg("sync_state = %x", state->statdata.sync_state);
			break;
		case STATTYPE_SIGNAL_STRENGTH:
			state->statdata.signal_strength = resp_frame.data[3]<<16 | resp_frame.data[2]<<8 | resp_frame.data[1];
			inno_msg("signal_strength = %x", state->statdata.signal_strength);
			break;
		case STATTYPE_LDPC_TOTAL_COUNT:
		case STATTYPE_LDPC_ERR_COUNT:
		case STATTYPE_RS_TOTAL_COUNT:
		case STATTYPE_RS_ERR_COUNT:
			state->statdata.ldpc_total_count = resp_frame.data[4]<<24 | resp_frame.data[3]<<16 | \
											   resp_frame.data[2]<<8 | resp_frame.data[1];
			inno_msg("ldpc_total_count = %x", state->statdata.ldpc_total_count);
			break;
	}

out:
	return ret;
#endif

}

/*
 * if101 power management
 */
int inno_set_pm(struct inno_pm_control* pmctl)
{
	struct inno_cmd_frame cmd_frame = {0};
	int ret;
	
	cmd_frame.code  = CMD_SET_PM;
	cmd_frame.data[0] = ((pmctl->sys_pause?PM_SYS_SUSPEND_BIT:PM_SYS_RESUME_BIT) | \
						(pmctl->stun_off?PM_STUN_OFF_BIT:PM_STUN_ON_BIT)	|	\
						(pmctl->utun_off?PM_UTUN_OFF_BIT:PM_STUN_ON_BIT)	|	\
						(pmctl->tspwr_off?PM_TSPWR_OFF_BIT:PM_TSPWR_ON_BIT));
	down(&inno_spi_mutex);
	ret = inno_i2c_send_cmd_frame(&cmd_frame);
	up(&inno_spi_mutex);

	return ret;
}

/*
 * TS0 auto scan
 */
int inno_scan_frequency_dot(u8 start_fdot, u8 end_fdot)
{
	struct inno_cmd_frame cmd_frame ={0};
	int ret = -INNO_ERROR;

	inno_msg("%s start_fdot=%d, end_fdot=%d.", __FUNCTION__, start_fdot, end_fdot);
	
	if(start_fdot > end_fdot){
		ret = -INNO_PARAM_ERR;
		goto out;
	}

	cmd_frame.code = CMD_SCAN_FREQUENCY;
	cmd_frame.data[0] = start_fdot;
	cmd_frame.data[1] = end_fdot;
	cmd_frame.data[2] = 0x01;	//ASEN==1

	down(&inno_spi_mutex);
	ret = inno_i2c_send_cmd_frame(&cmd_frame);
	up(&inno_spi_mutex);

out:
	return ret;
}

int inno_scan_frequency(u16 start_freq, u16 end_freq)// freq unit MHZ
{
	u8 start_fdot = inno_freq2dot(start_freq);
	u8 end_fdot = inno_freq2dot(end_freq);
	if(start_fdot && end_fdot)
		return inno_scan_frequency_dot(start_fdot, end_fdot);
	else
		return -INNO_PARAM_ERR;
}

static u8 inno_cp_type = SHORT_CP;
/*
 * Set Tuner Frequency
 */
int inno_set_frequency_dot(u8 freq_dot)
{
	struct inno_cmd_frame cmd_frame = {0};
	int ret = INNO_SUCCESS;
	int retry = 0;
	
	inno_msg("%s freq_dot=%d", __FUNCTION__, freq_dot);
	
	cmd_frame.code = CMD_SET_FREQUENCY;
	cmd_frame.data[0] = freq_dot;
	cmd_frame.data[1] |= inno_cp_type;

	down(&inno_spi_mutex);
	for(retry = 0; retry < 5; retry ++){
		ret = inno_i2c_send_cmd_frame(&cmd_frame);
		if(ret == INNO_SUCCESS)
			break;
		inno_msg("inno_set_frequency_dot, retry=%d",retry);
	}
	up(&inno_spi_mutex);
	
	
	return ret;
}

int inno_set_frequency(u16 freq_mhz)
{
	u8 freq_dot = inno_freq2dot(freq_mhz);
	inno_msg("%s freq=%dMhz", __FUNCTION__, freq_mhz);
	return inno_set_frequency_dot(freq_dot);
}

int inno_set_cp_type(u8 cp_type)
{
	struct inno_cmd_frame cmd_frame = {0};
	int ret = INNO_ERROR;

	inno_msg("set cp type : %s", cp_type==0?"40TS":"36TS");

	cmd_frame.code = CMD_SET_FREQUENCY;
	cmd_frame.data[0] = 0;
	cmd_frame.data[1] = cp_type & CP_TYPE_MASK;

	ret = inno_i2c_send_cmd_frame(&cmd_frame);
	if(ret == INNO_SUCCESS)
		inno_cp_type = cp_type & CP_TYPE_MASK;
	return ret;
}

/*
 * Get version number of if101 fireware
 */
int inno_get_fireware_version(struct inno_fw_ver* version)
{
	struct inno_cmd_frame cmd_frame = {0};
	struct inno_response_frame resp_frame = {0};
	int ret = INNO_SUCCESS;

	down(&inno_spi_mutex);

	cmd_frame.code = CMD_GET_FW_VER;
	ret = inno_i2c_send_cmd_frame(&cmd_frame);
	dump_cmd_frame(&cmd_frame);
	if(INNO_SUCCESS != ret){
		inno_err("%s send cmd frame failed!", __FUNCTION__);
		goto out;
	}
	
	ret = inno_i2c_read_response_frame(&resp_frame);
	if(INNO_SUCCESS != ret){
		inno_err("%s read response frame failed!", __FUNCTION__);
		goto out;
	}
	version->major = resp_frame.data[0];
	version->minor = resp_frame.data[2];
	dump_response_frame(&resp_frame);
out:
	up(&inno_spi_mutex);
	return ret;

}

/*
 * Update if101 fireware
 */
int inno_update_fireware(void)
{
	int ret = INNO_SUCCESS;

	return ret;
}

extern void inno_chip_reset(void);
/*
 * Download fireware bin from user
 *
 * Notice:
 *  - IF101 as SPI Master, generate SPI_CLK & SPI_CS.
 * 	- AP as SPI Slave.  
 *
 * |-----------------------------------------|
 * | B1 | B2 | B3 | B4 | B5 | B6 | N-BYTES...|
 * |-----------------------------------------|
 * | 0Bh|  addr-offset |Dumy|   24KByte      |
 * |-----------------------------------------|
 */
int inno_download_fireware(struct inno_fw_bin* fw_bin)
{
#if 0
	int ret = INNO_SUCCESS;
	struct inno_spi_driver* spi_driver = g_inno_dev.spi_driver;	
	int try_times = 10;
	int  handshake = 0;
	u8  recv_data, offset_addr[3];

	inno_msg("enter %s", __FUNCTION__);
	down(&inno_spi_mutex);
	spi_driver->uninit();
	spi_driver->init(INNO_SPI_SLAVE);
	
	//until receive instruction 0x0B
	while(!handshake && try_times){
		inno_msg("reset if101 chip, (%d)", try_times);
		inno_chip_reset();
	
		ret = spi_driver->sync_cs();
		if(ret){
			ret = -INNO_TIMEOUT;
			inno_msg("spi sync chip select failed");
			goto out;
		}
		
		spi_driver->read(&recv_data, 1);
		if(recv_data == 0x0B){
			//inno_msg("handshake successful");
			handshake = 1;
			break;
		}else
			inno_msg("recv %x", recv_data);
		
		if(--try_times <= 0){
			inno_msg("download fireware failed");
			ret = -INNO_TIMEOUT;
			goto out;
		}
	
	}
	
	if(handshake){
		//receive 24bits address offset (no use)
		ret = spi_driver->read(offset_addr, 3);	
		//inno_msg("receive offset address : 0x%x", (offset_addr[0]<<16 | offset_addr[1]<<8 | offset_addr[2]));
		spi_driver->read(&recv_data, 1);	//receive one byte dummy data
	
		//inno_msg("begin download fireware...");
		ret = spi_driver->write(fw_bin->data, sizeof(struct inno_fw_bin));	
		if(ret == -1){
			inno_msg("write fireware bin timeout");
			ret = -INNO_TIMEOUT;
			goto out;
		}
		inno_msg("fireware download completed");
		inno_delay_ms(600);	// wait for IF101 internal operate complete
	}

out:
	spi_driver->uninit();
	spi_driver->init(INNO_SPI_MASTER);
	up(&inno_spi_mutex);
	inno_msg("leave %s with return code:%x", __FUNCTION__, ret);

	return ret;
#else
	return -EINVAL;
#endif
}

static int inno_check_fw_download(unsigned char *buffer, int len, int *pass)
{
	int ret = 0;
	struct inno_cmd_frame 		cmd = {0};
	struct inno_response_frame	rsp = {0};
	unsigned char addr_high = 0;
	unsigned char addr_low = 0;
	unsigned char check_high = 0;
	unsigned char check_low = 0;
	unsigned short file_checksum = 0;
	unsigned short read_checksum = 0;
	int i = 0;
	int retry = 0;

	if(sizeof(unsigned short) != 2)		// make sure "unsigned short" is 16bit 
		return -EINVAL;

	/************ calculate check sum of .bin ****************/
	if(len % 2 == 0){
		for(i = 0; i< len / 2; i++)
			file_checksum ^= *((unsigned short *)buffer + i);		// little endian
	}
	else{
		for(i = 0; i< (len - 1) / 2; i++)
			file_checksum ^= *((unsigned short *)buffer + i);		// little endian
		file_checksum ^= *(buffer + len - 1) | (0xFF << 8);
		
	}


	/************ Get check sum from IF101 *******************/
	addr_high = (len & 0xFF00) >> 8;	// calculate FW end address (high 8 bit)
	addr_low = (len & 0xFF);			// calculate FW end address (low 8 bit)

	cmd.code = CMD_GET_FW_CHECK_SUM;		// Prepare I2C command
	cmd.data[0] = 0;
	cmd.data[1] = 0;
	cmd.data[2] = addr_low;
	cmd.data[3] = addr_high;

	down(&inno_spi_mutex);

	for(retry = 0; retry < 3; retry ++){
		ret = inno_i2c_send_cmd_frame(&cmd);
		if(INNO_SUCCESS != ret)
			continue;

		/* need extra delay before response */
		inno_delay_ms(50);

		ret = inno_i2c_read_response_frame(&rsp);
		if(INNO_SUCCESS != ret)
			continue; 
		else break;
		
	}
	if(ret != INNO_SUCCESS)
		goto out; 

	check_high = rsp.data[0];								// little endian
	check_low = rsp.data[1];							// little endian
	read_checksum = (check_high << 8) | check_low;	// little endian
	//printf("diag_printf = 0x%X\n", read_checksum);

	/************************** Compare check sum ******************************/
	if(file_checksum == read_checksum)
		*pass = 1;
	else 
		*pass = 0;

out:
	up(&inno_spi_mutex);
	return ret;
}

/*
 * IF202 fireware download
 */
#define INNO_DOWNLOAD_FW_DMA

#ifdef INNO_DOWNLOAD_FW_DMA
/*
static int inno_spi_write_one_byte(unsigned char value)
{
	int ret;
	ret = g_inno_dev.spi_driver->write(&value, 1);
	udelay(100);
	return ret;
}

static int inno_spi_read_one_byte(unsigned char *data)
{
	int ret;
	ret = g_inno_dev.spi_driver->read(data, 1);
	udelay(100);
	return ret;
}
*/
#else

static
int inno_spi_write_one_byte_nocs(u8 value)
{
	return g_inno_dev.spi_driver->write(&value, 1);
}

static
int inno_spi_read_one_byte_nocs(u8 *data)
{
	return g_inno_dev.spi_driver->read(data, 1);
}

static int inno_spi_drive_cs(int status)
{
	return g_inno_dev.spi_driver->drive_cs(status);
}

static int inno_spi_write_one_byte(unsigned char data)
{
	inno_spi_drive_cs(0);
	inno_spi_write_one_byte_nocs(data);
	inno_spi_drive_cs(1);

	udelay(100);
	return 0;
}

static int inno_spi_read_one_byte(unsigned char *data)
{
	inno_spi_drive_cs(0);
	inno_spi_read_one_byte_nocs(data);
	inno_spi_drive_cs(1);

	udelay(100);
	return 0;
}
#endif

/*
 * Function:	the main process to download FW
 *
 * Paramter:	fw_buffer<in>: buffer pointer to FW data
 *				fw_size<in>: FW size
 */
static int if202_download_fw(unsigned char *fw_buffer, int fw_size)
{
	int ret = 0;
	int count = 0;
//	int download_complete = 0;
//	unsigned char status = 0;
//	int try_times = 5;
#ifdef INNO_DOWNLOAD_FW_DMA
	struct page **fw_pages;	/* physical pages */
	int fw_page_num;	/* physical pages count */
	char *fw_spi_data, *tmp;
	struct page *page = NULL;
	int i, j;

	/* alloc buffer for SPI transfer */
	fw_page_num = PAGE_ALIGN(fw_size*4) / PAGE_SIZE;
	fw_pages = (struct page **)kzalloc(fw_page_num * sizeof(struct page *), GFP_KERNEL);
	if (!fw_pages) {
		inno_err("SPI fw download: alloc array failed.");
		ret = -ENOMEM;
		goto alloc_error;
	}

	for (i=0; i<fw_page_num; i++) {
		page = alloc_page(GFP_KERNEL);

		if (!page) {
			inno_err("SPI fw download: alloc page %d failed, no enough memory.", i);
			ret = -ENOMEM;
			goto alloc_pages_error;
		}
		fw_pages[i] = page;
	}

	fw_spi_data = vmap(fw_pages, fw_page_num, VM_MAP, PAGE_KERNEL);

	/* check if the memory map is OK. */
	if (!fw_spi_data) {
		inno_err("SPI fw download: vmap() failed");
		ret = -EFAULT;
		goto vmap_error;
	}

	tmp = fw_spi_data;
	for(count = 0; count < fw_size - 1; count++)
	{
		*tmp++ = WRITE_DATA_CMD;
		*tmp++ = fw_buffer[count];
		*tmp++ = WRITE_STATUS_CMD;
		*tmp++ = SEND_CONTINUE;
	}
	// the last byte fw_buffer[fw_size - 1]
	*tmp++ = WRITE_DATA_CMD;
	*tmp++ = fw_buffer[count];
	*tmp++ = WRITE_STATUS_CMD;
	*tmp++ = SEND_OVER;
#endif

	//do{
		//if(try_times--<=0)
			//return -ETIME;

		//1. Reset IF202
		inno_msg("Begin downloading firmware...");
		inno_chip_reset();

#ifdef INNO_DOWNLOAD_FW_DMA
		g_inno_dev.spi_driver->write(fw_spi_data, fw_size*4);
#else
		inno_spi_drive_cs(0);	// !!! CS->low

		//2. Download fw_buffer[0] ~ fw_buffer[fw_size - 2]
		for(count = 0; count < fw_size - 1; count++)
		{
			inno_spi_write_one_byte_nocs(WRITE_DATA_CMD);
			inno_spi_write_one_byte_nocs(fw_buffer[count]);
			inno_spi_write_one_byte_nocs(WRITE_STATUS_CMD);
			inno_spi_write_one_byte_nocs(SEND_CONTINUE);
			// !! FIXME: Maybe need to add some delay here which is longher than 3us
		}

		//3. download the last byte fw_buffer[fw_size - 1]
		inno_spi_write_one_byte_nocs(WRITE_DATA_CMD);
		inno_spi_write_one_byte_nocs(fw_buffer[count]);
		inno_spi_write_one_byte_nocs(WRITE_STATUS_CMD);
		inno_spi_write_one_byte_nocs(SEND_OVER);

		inno_spi_drive_cs(1);	// !!! CS->high
#endif
/*
		//4. check if downloading is completed or not
		mdelay(1);
		download_complete = 0;
		inno_spi_write_one_byte(READ_STATUS_CMD);
		inno_spi_read_one_byte(&status);
		if(status == DOWNLOAD_OVER){
			download_complete = 1;
		}
		else{
			download_complete = 0;
			inno_msg("download firmware download_complete=0. try again!\n");
			continue;
		}
*/
//		inno_msg("download firmware complete!\n");
	//}while(download_complete != 1);

#ifdef INNO_DOWNLOAD_FW_DMA
	vunmap(fw_spi_data);
	i = fw_page_num;

vmap_error:
alloc_pages_error:
	for (j=0; j<i; j++) {
		page = fw_pages[j];
		__free_page(page);
	}
	kfree(fw_pages);
alloc_error:
#endif

	return ret;
}

int inno_download_firmware(char* fw_bin, int fw_size)
{
	int ret;
	int download_success = 0;
	int i = 0;
	
	for(i = 0; i< 5; i++){
		inno_msg("inno_download_firmware. i=%d", i);
		
		down(&inno_spi_mutex);
		ret = if202_download_fw(fw_bin, fw_size);
		up(&inno_spi_mutex);
		if(ret){
			inno_err("innodev: error, firmware download failed %d!\n", ret);
			ret =  -EIO;
			continue;
		}
		inno_delay_ms(100);

		// check if downloading is successful or not
		download_success = 0;
		if(inno_check_fw_download(fw_bin, fw_size, &download_success)){
			inno_err("innodev: error, check firmware download failed %d!\n", ret);
			ret =  -EIO;
			continue;
		}
		if(download_success == 0){
			inno_err("innodev: error, firmware download check failed\n");
		}else{
			inno_msg("firmware download check ok!");
			break;
		}
	}


	return ret;
}

/**********************************************************
Reset CMD FORMAT:
	UAM_RESET_CMD | 0x00 | 0x00 | crc
***********************************************************/
int INNO_UAM_Reset(unsigned char *pATRValue, unsigned int *pATRLen)
{
	int ret = INNO_ERROR;
	unsigned char cmd[8] = {0};
	unsigned int retry = 0;
//	int i = 0;

	cmd[0] = UAM_RESET_CMD;     // command
	cmd[1] = 0x00;
	cmd[2] = 0x00;

	for(retry = 0; retry < 3; retry ++){
		// Send CMD					
		ret = INNO_Send_UAM_Cmd(cmd, 4);
		//diag_printf("\n INNO_Send_Cmd ret = %d\r\n", ret);
		if(ret != INNO_SUCCESS)
			continue;
		
		// Get RSP
		ret = INNO_Read_UAM_Rsp(pATRValue, pATRLen);
		//diag_printf("\n INNO_Send_SIM_Reset ret = %d, rsplen = %d\r\n", ret, rsplen);
		//OutputStr(pATRValue, *pATRLen);
	#if 0
		inno_dbg( "Reset ATR:" );
		for(i = 0 ; i< *pATRLen; i++)
			inno_dbg(" 0x%2x ",*(pATRValue+i));
	#endif
		
		if(ret != INNO_SUCCESS)
			continue;
		else 
			break;
	}

	return ret;
}

static int UpdateCardCapabilities(void)
{
		unsigned char *atrString = cardCapabilities_UAM.ATR.Buffer;
		unsigned long atrLength = (unsigned long) cardCapabilities_UAM.ATR.Length;
		unsigned char	Tck, TA[4]={0}, TB[4]={0}, TC[4]={0}, TD[4]={0}, Y;
		unsigned char TA2Present = 0;
		unsigned long i, numProtocols = 0, protocolTypes = 0;
		int status =0;
		
		if(atrLength < 2)
		{
			inno_err("ATR is too short (Min. length is 2)!/r/n");
			return 0;
		}
		
		if(atrString[0] != 0x3b && atrString[0] != 0x3f && atrString[0] != 0x03) 
		{
			inno_err("Initial character %02xh of ATR is invalid\n",atrString[0]);
		    return 0;
		}
		
		// Test for invers convention
		if (*atrString == 0x03) 
		{
			cardCapabilities_UAM.InversConvention = 1;
		    inno_err("Card uses Inverse Convention\r\n");
			return 0;
		    //SET_REG(SIM_CONFIG,GET_REG(SIM_CONFIG)|SIM_DATA_INVERTER);
		}
		
		atrString += 1;
		atrLength -= 1;
		
		// Calculate check char, but do not test now since if only T=0	is present the ATR doesn't contain a check char
	        for(i = 0, Tck = 0; i < atrLength; i++) 
	  	{
			Tck ^= atrString[i];
		}
		
		// Set default values as described in ISO 7816-3
		TA[0] = 0x11;		// TA1 codes FI in high-byte and Dl in low-byte;
		TB[0] = 0x25;		// TB1 codes II in bits b7/b6 and PI1 in b5-b1. b8 has to be 0
		//TC[0] = 0x0;		// TC1 codes N in bits b8-b1
	        TC[1] = 10;			// TC2 codes T=0 WI
		
		// Translate ATR string to TA to TD values (See ISO)
		cardCapabilities_UAM.HistoricalChars.Length = *atrString & 0x0f;
		Y = *atrString & 0xf0;
		
		atrString += 1;
		atrLength -= 1;
		
		for(i=0; i<4; i++)
		{
			if(Y & 0x10)
			{
				if(i==1)
				{
					TA2Present = 1; 
				}
				
				TA[i] = *atrString++;
				atrLength--;
			}
			
			if(Y & 0x20)
			{
				TB[i] = *atrString++;
				atrLength--;
			}
			
			if(Y & 0x40)
			{
				TC[i] = *atrString++;
				atrLength--;
			}
			
			if(Y & 0x80)
			{
				Y = *atrString & 0xf0;
				TD[i] = *atrString++ & 0x0f;
				atrLength--;
				
				// Check if the next parameters are for a new protocol.
				if (((1 << TD[i]) & protocolTypes) == 0) 
				{
					// Count the number of protocols that the card supports
					numProtocols++;
				}
				protocolTypes |= 1 << TD[i];
			}else
			{
				break;
			}
		}
		
		// Check if the card supports a protocol other than T=0
		if (protocolTypes & ~1) 
		{
			// The atr contains a checksum byte. Exclude that from the historical byte length check
			atrLength -=1;		
	
			// This card supports more than one protocol or a protocol other than T=0, so test if the checksum is correct
			if (Tck != 0) 
			{
				inno_err("ATR Checksum is invalid\r\n");
				status = 0;
				goto _leave_lab;
			}
		}
		
		if (atrLength < 0 || atrLength != cardCapabilities_UAM.HistoricalChars.Length) 
		{
			inno_err("ATR length is inconsistent\r\n");
			status = 0;
			goto _leave_lab;
		}
		
	_leave_lab:
		if(status != 0)
		{
			return status;
		}
		
		// store historical characters
		memcpy(cardCapabilities_UAM.HistoricalChars.Buffer, atrString, cardCapabilities_UAM.HistoricalChars.Length);
		
		// Now convert TA - TD values to global interface bytes
		cardCapabilities_UAM.FI = (TA[0] & 0xf0) >> 4;// Clock rate conversion
		cardCapabilities_UAM.DI = (TA[0] & 0x0f);	// bit rate adjustment
		cardCapabilities_UAM.II = (TB[0] & 0xc0) >> 6;	// Maximum programming current factor
		cardCapabilities_UAM.P = (TB[1] ? TB[1] : (TB[0] & 0x1f) * 10); // Programming voltage in 0.1 Volts
		cardCapabilities_UAM.N = TC[0]; // Extra guard time
	
		if ((TA2Present || (numProtocols <= 1)) && (cardCapabilities_UAM.FI == 1) && (cardCapabilities_UAM.DI == 1)) 
		{
			// If the card supports only one protocol (or T=0 as default)
		// and only standard paramters then PTS selection is not available
			OperationMode_UAM = OP_SPECIFIC;
		} 
		else 
		{
			OperationMode_UAM = OP_NEGOTIABLE;
		}
	
		// Now find protocol specific data
		if (TD[0] == 0) 
		{		
			cardCapabilities_UAM.Protocol.Supported |= SCARD_PROTOCOL_T0;
			cardCapabilities_UAM.T0.WI = TC[1];
			inno_msg("T=0 Values from ATR:\n  WI = %x\n",cardCapabilities_UAM.T0.WI);
	  }
	  
		if (protocolTypes & SCARD_PROTOCOL_T1) 
		{
			for (i = 0; TD[i] != 1 && i < 4; i++)
				;
		
			for (; TD[i] == 1 && i < 4; i++) 
				;
	
			if (i == 4) 
			{
				return 0;			
		    }
	
			cardCapabilities_UAM.Protocol.Supported |= SCARD_PROTOCOL_T1;
	
			cardCapabilities_UAM.T1.IFSC = (TA[i] ? TA[i] : 32);
			cardCapabilities_UAM.T1.CWI = ((TB[i] & 0x0f) ? (TB[i] & 0x0f) : T1_CWI_DEFAULT);
			cardCapabilities_UAM.T1.BWI = ((TB[i] & 0xf0) >> 4 ? (TB[i] & 0xf0) >> 4 : T1_BWI_DEFAULT);
			cardCapabilities_UAM.T1.EDC = (TC[i] & 0x01);
	
		}
		
		if(OperationMode_UAM == OP_SPECIFIC)
		{
			if(TA2Present)
			{
				// TA2 is present in the ATR, so use the protocol indicated in the ATR
		  		cardCapabilities_UAM.Protocol.Selected = 1 << TA[1];
			}
			else
			{
				// The card only supports one protocol So make that one protocol the current one to use
			  	cardCapabilities_UAM.Protocol.Selected = cardCapabilities_UAM.Protocol.Supported;
			}
		}else
		{
			inno_msg("Mode: Negotiable\n\n");
		}
		
		return 1;	
}


/**********************************************************
PPS CMD FORMAT:
	UAM_PPS_CMD | 0x04 | 0x00 | pps[0] | pps[1] | pps[2] | pps[3] | crc
***********************************************************/
static int INNO_UAM_PPS(unsigned char *ppsReqValue)
{
	int ret = INNO_ERROR;
	unsigned char cmd[8] = {0};
	unsigned char rsp[33] = {0};
	unsigned int retry = 0, rsplen = 0;

	cmd[0] = UAM_PPS_CMD;			// command
	cmd[1] = 0x00;
	cmd[2] = 0x04;
	cmd[3] = ppsReqValue[0];
	cmd[4] = ppsReqValue[1];
	cmd[5] = ppsReqValue[2];
	cmd[6] = ppsReqValue[3];	

	for(retry = 0; retry < 3; retry ++){
		// Send CMD
		ret = INNO_Send_UAM_Cmd(cmd, 8);
		//diag_printf("\n INNO_Send_Cmd ret = %d\r\n", ret);
		if(ret != INNO_SUCCESS)
			continue;
		
		// Get RSP
		ret = INNO_Read_UAM_Rsp(rsp, &rsplen);
		//OutputStr(rsp, rsplen);
		
		if(ret != INNO_SUCCESS)
			continue;
		else 
			break;
	}

	if(ret != INNO_SUCCESS)
		return ret;

	if(memcmp(rsp, ppsReqValue, 4) == 0)
		return INNO_SUCCESS;
	else
		return INNO_ERROR;
}



static int inno_init_uam(void)
{
	unsigned char ATRBuf[40] ={0};	
	unsigned char pps_request[] ={0xFF, 0x10, 0, 0};
	//SCARD_CARD_CAPABILITIES  cardCapabilities_UAM;
	unsigned int atrLen = 0;
	int ret=0;
	
	ret = INNO_UAM_Reset(ATRBuf, &atrLen);
// AAAAAAA 2009/11/02
#if 1
	if(ret != INNO_SUCCESS)
	{
    		inno_err("innodev: UAM reset fail\n");
	 		return INNO_ERROR;
    }
	else
	{
		if(atrLen <= 0)
		{
			inno_err("innodev: ATR Value Length error!!!\n");
			return INNO_ERROR;
		}
		
		 memcpy(cardCapabilities_UAM.ATR.Buffer,ATRBuf,atrLen);		 
		 cardCapabilities_UAM.ATR.Length = atrLen;
		 
		 ret = UpdateCardCapabilities();
		  if(ret)
		  {
		    	//diag_printf("[INNO_UAM_Init] - INNO_UAM_Request_PPS\r\n");
				//diag_printf("[INNO_UAM_Init] - NO PPS\r\n");
	
			if(cardCapabilities_UAM.FI == 0x1 && cardCapabilities_UAM.DI == 0x8)
			{
				pps_request[2]=0x18;
			}
			else
			{
				pps_request[2]=0x94;
			}
			
			pps_request[3] = pps_request[0] ^ pps_request[1] ^pps_request[2];
			
			if(INNO_UAM_PPS(pps_request) != INNO_SUCCESS)
		 	{
		 		//diag_printf("UAM PPS Request failed\r\n");
				return INNO_ERROR;
		 	}
			
			inno_msg("innodev: UAM PPS Request successfully");
		   }
	 }
#endif
	
	 return INNO_SUCCESS;
}

int inno_uam_transfer(struct inno_uam_parameter *uam)
{
	inno_return_value_t	ret = INNO_SUCCESS;
	unsigned char 	cmd[260] = {0}, rsp[256] = {0};
	unsigned int	cmdLen = 0, rsplen = 0;
	//int i = 0;
	unsigned char *pBufIn;
	unsigned int bufInLen;
	unsigned char *pBufOut; 
	unsigned int pBufOutLen;
	unsigned short sw;
	int i;
	//int cmd_time = 0;

	pBufIn = uam->pBufIn;
	bufInLen = uam->bufInLen;
	pBufOut= uam->pBufOut; 
	pBufOutLen = uam->pBufOutLen;
	sw = uam->sw;

	inno_msg("+++++ inno_uam_transfer +++++"); 
 	inno_dbg("inno_uam_transfer, bufInLen=%d", bufInLen);
	inno_dbg("inno_uam_transfer, pBufIn: ");
	for(i = 0; i< bufInLen; i++){
		inno_dbg("0x%X,",pBufIn[i]);
	}

	if(bufInLen == 4)
	{
		inno_err("inno_uam_transfer, bufInLen = 4. return!\n");
		return INNO_ERROR;
	}
	else if(bufInLen == 5)
	{
		//inno_msg("inno_uam_transfer, bufInLen == 5\n");
		cmd[0] = GET_UAM_CMD;
		cmd[1] = 0x00;
		cmd[2] = 0x05;    //data len
		memcpy(cmd+3, pBufIn, 5);
		cmdLen = 0x09;  //bufInLen + 4

		ret = INNO_Send_UAM_Cmd(cmd, cmdLen);	
		if(ret != INNO_SUCCESS)  
		{
			inno_err("[T0Transfers (case 2)] - send cmd fail!!!\r\n");
			return INNO_ERROR;
		}

		ret = INNO_Read_UAM_Rsp(rsp, &rsplen);
		if(ret != INNO_SUCCESS)
		{
			inno_err("[T0Transfers (case 2)] - response fail!!!\r\n");
			return INNO_ERROR;      
		}
		
		if(rsplen < 3)
		{
			inno_err("[T0Transfers (case 2)] - Get SW1, SW2 error!!!");
			inno_err("inno_uam_transfer, bufInLen=%d", bufInLen);
			inno_err("inno_uam_transfer, pBufIn: ");
			for(i = 0; i< bufInLen; i++)
				inno_err("0x%X,",pBufIn[i]);
			inno_err("inno_uam_transfer,rspLen=%d", rsplen);
			inno_err("inno_uam_transfer rsp:");
			for(i = 0; i< rsplen; i++)
				inno_err("0x%X,", rsp[i]);
			return INNO_ERROR;    
		}

//		sw = MAKEWORD(rsp[rsplen-1], rsp[rsplen-2]);
		sw = ((unsigned short)((unsigned char)(rsp[rsplen-1])) | (((unsigned short)(unsigned char)(rsp[rsplen-2]))<<8));
		uam->sw = sw;
                
		//diag_printf("[T0Transfers (case 2)] - *sw = 0x%x\r\n", *sw);

		pBufOutLen = rsplen - 3;
		uam->pBufOutLen = pBufOutLen;
		memcpy(pBufOut, rsp+1, pBufOutLen);
		
	}
	else if(bufInLen > 5 && bufInLen == (int)pBufIn[4] + 5)
	{

		//inno_msg("kwjing - inno_uam_transfer, bufInLen > 5\n");
		cmd[0] = SEND_UAM_CMD;
		if(bufInLen <= 256)    
		{
			cmd[1] = 0x00;
			cmd[2] = bufInLen;			
		}
		else                //check?
		{		
			cmd[1] = bufInLen& 0xff;  
			cmd[2] = (bufInLen& 0xff00) >> 8; 			  
		}	
		memcpy(cmd+3, pBufIn, bufInLen);
		
		cmdLen = bufInLen + 4; 

		ret = INNO_Send_UAM_Cmd(cmd, cmdLen);
		if(ret != INNO_SUCCESS)  
		{
			inno_err("[T0Transfers (case 3)] - send cmd fail!!!\r\n");
			return INNO_ERROR;
		}

		//inno_msg("kwjing - inno_uam_trans , INNO_Read_UAM_Rsp\n");
		ret = INNO_Read_UAM_Rsp(rsp, &rsplen);
		if(ret != INNO_SUCCESS)
		{
			inno_err("[T0Transfers (case 3)] - response fail!!!\r\n");
			return INNO_ERROR;      
		}
		
		if(rsplen < 3)
		{
			inno_err("[T0Transfers (case 3)] - Get SW1, SW2 error!!!");
			inno_err("inno_uam_transfer, bufInLen=%d", bufInLen);
			inno_err("inno_uam_transfer, pBufIn: ");
			for(i = 0; i< bufInLen; i++)
				inno_err("0x%X,",pBufIn[i]);
			inno_err("inno_uam_transfer,rspLen=%d", rsplen);
			inno_err("inno_uam_transfer rsp:");
			for(i = 0; i< rsplen; i++)
				inno_err("0x%X,", rsp[i]);
			
			return INNO_ERROR;    
		}

//		inno_msg("kwjing - inno_uam_trans , after INNO_Send_UAM_Cmd 00: rsplen: %d, rsplen-1: 0x%x, rsplen-2: 0x%x, rsplen0: 0x%x\n", 
//		rsplen, rsp[rsplen-1], rsp[rsplen-2], rsp[0]);
//		sw = MAKEWORD(rsp[rsplen-1], rsp[rsplen-2]);
		sw = ((unsigned short)((unsigned char)(rsp[rsplen-1])) | (((unsigned short)(unsigned char)(rsp[rsplen-2]))<<8));
		uam->sw = sw;
		//diag_printf("[T0Transfers (case 3)] - sw = 0x%x\r\n", *sw);

		//inno_msg("kwjing - inno_uam_trans , after INNO_Send_UAM_Cmd 01: pBufOutLenaaa: %d, sw: 0x%x\n", (rsplen - 3), sw);
		pBufOutLen = rsplen - 3;
		uam->pBufOutLen = pBufOutLen;
		memcpy(pBufOut, rsp+1, pBufOutLen);
		//inno_msg("kwjing - inno_uam_trans , after INNO_Send_UAM_Cmd 02\n");
		
	}
	else if(bufInLen > 5 && bufInLen > (int)pBufIn[4] + 5)
	{
		inno_err("inno_uam_transfer, bufInLen > 5 return!\n");
		return INNO_ERROR;
	}

 	inno_dbg("inno_uam_transfer, pBufOutLen=%d", pBufOutLen);
	inno_dbg("inno_uam_transfer, pBufOut: ");
	for(i = 0; i< pBufOutLen; i++){
		inno_dbg("0x%X,", pBufOut[i]);
	}

	inno_msg("inno_uam_transfer, sw=0x%X",sw); 
	inno_msg("----- inno_uam_transfer -----"); 

	return INNO_SUCCESS;	
}

static int INNO_Send_UAM_Cmd(unsigned char *cmd, int cmdLen)
{
	unsigned char Local_CRC = 0xA5;
	int ret = 0;
	int i = 0;
	unsigned char end = 0;

	//OutputStr(cmd, cmdLen);
	//printk(KERN_INFO "innodev: INNO_Send_UAM_Cmd\n");
/*
	printk("INNO_Send_UAM_Cmd.cmdLen=%d,cmd=\n", cmdLen);
	for(i = 0; i< cmdLen -1; i++){
		printk("0x%02x,", cmd[i]);
		if(i%10 == 9)
			printk("\n");
	}
	printk("0x%02x\n", cmd[i]);
*/
	down(&inno_spi_mutex);
	
	ret = inno_i2c_send_cmd_byte(cmd[0], CMD_CODE);			// send cmd code 1 byte
	if(ret != INNO_SUCCESS)
		goto out;

	Local_CRC ^= cmd[0];

	for(i = 1; i< (cmdLen-1); i++){
		ret = inno_i2c_send_cmd_byte(cmd[i], CMD_PAYLOAD);		// send cmd payload cmdlen bytes
		if(ret != INNO_SUCCESS)
			goto out;
		Local_CRC ^= cmd[i];
	}

	cmd[cmdLen-1] = Local_CRC;

	ret = inno_i2c_send_cmd_byte(Local_CRC, CMD_PAYLOAD);		// send crc 1 byte
	if(ret != INNO_SUCCESS)
		goto out;

	ret = inno_i2c_read_response_byte(&end);							// check FW response
	//diag_printf("\n ret = %d, end = 0x%x\r\n", ret, end);

	if(end == 0xF9)
		ret =  INNO_SUCCESS;
	else 
		ret =  INNO_ERROR;

out: 
	up(&inno_spi_mutex);
	return ret; 
}

static int INNO_Read_UAM_Rsp(unsigned char *rsp, unsigned int *rsplen)
{

	int err = 0;
	struct inno_lgx* lgx = g_inno_dev.lgxs[UAM];
	struct inno_buffer* buffer = &lgx->inno_buffer;
	DECLARE_WAITQUEUE(wait, current);

	inno_dbg("innodev: enter Read_UAM_Rsp");
		
	add_wait_queue(&lgx->uam_wait, &wait);
	
	while(1) {
		err = -ERESTARTSYS;	
		if(signal_pending(current))
			break;
		if(buffer->valid == 1)
		{
			down(&buffer->sem);

			memcpy(rsp, buffer->start, buffer->valid_size);
			*rsplen = buffer->valid_size;
			inno_dbg("innodev: Read_UAM_Rsp GotRSP!\n");

			buffer->valid = 0;
			up(&buffer->sem);
			break;
		}
		set_current_state(TASK_INTERRUPTIBLE);
		err = -EAGAIN;
		
		schedule();
		//msleep(1);

		inno_dbg("waiting..., buffer->valid:%d, buffer->valid_size: %d", buffer->valid, buffer->valid_size);
	}
	
	set_current_state(TASK_RUNNING);
	remove_wait_queue(&lgx->uam_wait, &wait);
	//inno_msg("leave Read_UAM_Rsp. valid:%d, valid_size:[%d]", buffer->valid, buffer->valid_size);
/*
	printk("INNO_Read_UAM_Rsp.rsplen=%d,rsp=\n", *rsplen);
	for(i = 0; i< *rsplen -1; i++){
		printk("0x%02x,", rsp[i]);
		if(i%10 == 9)
			printk("\n");
	}
	printk("0x%02x\n", rsp[i]);
*/
	return 0;
}

