#include <linux/init.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/interrupt.h>

#include <mach/pxa-regs.h>
#include <mach/gpio.h>
#include <mach/mfp.h>
#include <asm/irq.h>
#include <linux/delay.h>
#include <linux/platform_device.h>
#include <linux/firmware.h>
#include <mach/innodev.h>

struct device *g_innodev_platform_dev;

/*
 * Innodev platform functions
*/

/*
 * register irq handler
 * parmaters pass by inno_core
 * @handler		-	if101 irq handler function pointer
 * @irq_type	-	if101 irq type (falling edge detect or rising)
 */
int inno_irq_setup(irqreturn_t (*handler)(int, void *), int irq_type)
{
	unsigned long irqflags;
	struct innodev_platform_data *pdata;

	if(!g_innodev_platform_dev || !g_innodev_platform_dev->platform_data)
		return -EINVAL;
	pdata = g_innodev_platform_dev->platform_data;

	if(irq_type == 0){		//falling edge
		irqflags = IRQF_TRIGGER_FALLING;
	}else{ // RISING_EDGE
		irqflags = IRQF_TRIGGER_RISING;
	}

	if(request_irq(pdata->irq, handler, irqflags, "innodev", NULL))
	{
		printk(KERN_ERR "innodev: register %d interrupt handler failed.\r\n", pdata->irq);
		return -EIO;
	}
	
	return 0;
}

void inno_irq_release(void)
{
	struct innodev_platform_data *pdata;

	if(!g_innodev_platform_dev || !g_innodev_platform_dev->platform_data)
		return;
	pdata = g_innodev_platform_dev->platform_data;

	free_irq(pdata->irq, NULL);
}

EXPORT_SYMBOL(inno_irq_setup);
EXPORT_SYMBOL(inno_irq_release);

void inno_chip_power_on(void)
{
	struct innodev_platform_data *pdata;
	if(!g_innodev_platform_dev || !g_innodev_platform_dev->platform_data)
		return;
	pdata = g_innodev_platform_dev->platform_data;
	pdata->voltage_on();
}

void inno_chip_power_off(void)
{
	struct innodev_platform_data *pdata;
	if(!g_innodev_platform_dev || !g_innodev_platform_dev->platform_data)
		return;
	pdata = g_innodev_platform_dev->platform_data;
	pdata->voltage_off();
}

void inno_chip_reset(void)
{
	struct innodev_platform_data *pdata;
	if(!g_innodev_platform_dev || !g_innodev_platform_dev->platform_data)
		return;
	pdata = g_innodev_platform_dev->platform_data;
	pdata->reset();
}

EXPORT_SYMBOL(inno_chip_power_on);
EXPORT_SYMBOL(inno_chip_power_off);
EXPORT_SYMBOL(inno_chip_reset);

/*
 * IF202 fireware download
 */
extern int inno_download_firmware(char* fw_bin, int fw_size);

static int inno_request_firmware(char *fw_name)
{
	int ret;
	/* uses the default method to get the firmware */
	const struct firmware *fw_entry;

	if(fw_name==NULL) {
		printk(KERN_ERR "innodev: error, firmware name is NULL\n");
		return -EINVAL;
	}

	printk(KERN_INFO "innodev: if2xx requesting firmware (%s)\n", fw_name);

	if(request_firmware(&fw_entry, fw_name, g_innodev_platform_dev)!=0)
	{
		printk(KERN_ERR "innodev: firmware (%s) not available!\n", fw_name);
		return -EINVAL;
	}

	printk(KERN_INFO "innodev: if2xx loaded firmware %d bytes\n", fw_entry->size);
	ret = inno_download_firmware(fw_entry->data, fw_entry->size);

	release_firmware(fw_entry);

	/* finish setting up the device */
	return ret;
}

int inno_check_firmware(void)
{
	struct innodev_platform_data *pdata;
	if(!g_innodev_platform_dev || !g_innodev_platform_dev->platform_data)
		return -EINVAL;
	pdata = g_innodev_platform_dev->platform_data;

	if(pdata->fw_spi_download)
		return inno_request_firmware(pdata->fw_name);
	return 0;
}

EXPORT_SYMBOL(inno_check_firmware);

static int __init innodev_probe(struct platform_device *pdev)
{
	g_innodev_platform_dev = &pdev->dev;
	return 0;
}

static int innodev_remove(struct platform_device *pdev)
{
	g_innodev_platform_dev = NULL;
	return 0;
}

static struct platform_driver innodev_driver = {
	.driver         = {
		.name   = "inno-dev",
	},
	.probe          = innodev_probe,
	.remove         = innodev_remove,
};

static int __init innodev_module_init(void)
{
	return platform_driver_register(&innodev_driver);
}

static void __exit innodev_module_exit(void)
{
	platform_driver_unregister(&innodev_driver);
}

module_init(innodev_module_init);
module_exit(innodev_module_exit);
