/***********************************************
 * FileName:	inno_lowe_spi.c
 *
 * Description:  
 *		SPI driver for Innofidei IF101 chip.
 * Platform:
 * 		Marvell PXA310
 * Author:	
 *		Jimmy.li <lizhengming@innofidei.com>
 * Date:	
 *		Oct. 15, 2007
 * History:
 * 		Dec.13, 2007	-- Add DMA support
 * 		Jan.3, 2008		-- add spi slave mode operations
 * 		Jan.7, 2008		-- add inno_spi_sync_cs function
 * Copyright (c) 2007, Innofidei Inc
 ***********************************************/
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/types.h>
#include <linux/delay.h>
#include <linux/sched.h>
#include <linux/wait.h>
#include <linux/clk.h>
#include <linux/platform_device.h>
#include <mach/ssp.h>
#include <mach/regs-ssp.h>
#include <asm/io.h>
#include <mach/gpio.h>
#include <mach/hardware.h>
#include <mach/innodev.h>

#define INNO_SPI_DMA
//#define INNO_SPI_DEBUG

#ifdef INNO_SPI_DEBUG
#define dbg(fmt, arg...)	printk(KERN_DEBUG "[inno_spi]:\t"fmt"\n", ##arg)
#else
#define dbg(fmt, arg...)	
#endif
#define err(fmt, arg...)	printk(KERN_ERR "[inno_spi ERROR!!]:\t"fmt"\n", ##arg);

#define INNO_SPI_MASTER		0
#define INNO_SPI_SLAVE		1

#define SPI_MODE_0  0     //idle low, transmit falling edge, receive rising edge.
#define SPI_MODE_1	1     //idle high, transmit rising edge, receive falling edge
#define SPI_MODE_2  2     //idle low, transmit rising edge, receive falling edge
#define SPI_MODE_3  3     //idle high, transmit falling edge, receive rising edge.

struct innodev_platform_spi *g_platform_spi = NULL;
static int ssp_port = -1;
static unsigned int spi_cs_gpio = 0;
static int spi_cs_drive = 1;

static int spi_inited = 0;
static int spi_type	= INNO_SPI_MASTER;

static unsigned short master_clk_mode = SPI_MODE_0;
static unsigned short slave_clk_mode = SPI_MODE_3;

static struct ssp_dev	inno_spi_dev;

#ifdef INNO_SPI_DMA
#include <mach/dma.h>
#include <linux/dma-mapping.h>

#define DMA_BUFFER_SIZE		PAGE_SIZE

struct inno_spi_dma_info{
	int				txdma;
	int   			rxdma;
	void			*txdma_addr;
	void      		*rxdma_addr;
	dma_addr_t  	txdma_addr_phys;
	dma_addr_t  	rxdma_addr_phys;
};

struct inno_spi_dma_info inno_dma;
static DECLARE_WAIT_QUEUE_HEAD(dma_rx_wq);
static int dma_rxflag = 0;
static DECLARE_WAIT_QUEUE_HEAD(dma_tx_wq);
static int dma_txflag = 0;

/*
 * inno spi dma functions
 */
static void inno_spi_dma_tx_intr(int dma, void *devid)
{
	if (DCSR(inno_dma.txdma) & DCSR_ENDINTR){
		dbg("TX DMA DCSR_ENDINTR");
		DCSR(inno_dma.txdma) |= DCSR_ENDINTR;
	}
	dma_txflag = 1;
	wake_up_interruptible(&dma_tx_wq);
}

static void inno_spi_dma_rx_intr(int dma, void *devid)
{
	if (DCSR(inno_dma.rxdma) & DCSR_ENDINTR){
		dbg("RX DMA DCSR_ENDINTR");	
		DCSR(inno_dma.rxdma) |= DCSR_ENDINTR;
	}
	dma_rxflag = 1;
	wake_up_interruptible(&dma_rx_wq);
}

static void inno_spi_dma_init(void)
{
	struct ssp_device *ssp = inno_spi_dev.ssp;
	uint32_t sscr1;

	dbg("SPI DMA Init");

	sscr1 = __raw_readl(ssp->mmio_base + SSCR1);
	sscr1 |= (SSCR1_TSRE | SSCR1_RSRE); 		//enable DMA request
	sscr1 |= SSCR1_TxTresh(2)|SSCR1_RxTresh(8); //set DMA request trig threshold
	sscr1 |= SSCR1_TRAIL;
	__raw_writel(sscr1, ssp->mmio_base + SSCR1);
	dbg("SSCR1(%d) = %x", ssp_port, sscr1);

	inno_dma.rxdma = pxa_request_dma("Inno Rx SPI DMA", DMA_PRIO_MEDIUM, inno_spi_dma_rx_intr, NULL);
	inno_dma.txdma = pxa_request_dma("Inno TX SPI DMA", DMA_PRIO_MEDIUM, inno_spi_dma_tx_intr, NULL);
	dbg("SPI DMA Channel, rx: %d, tx: %d", inno_dma.rxdma, inno_dma.txdma);
    
	inno_dma.txdma_addr = NULL;
	inno_dma.rxdma_addr = NULL;
	inno_dma.txdma_addr = dma_alloc_coherent(NULL, DMA_BUFFER_SIZE, &inno_dma.txdma_addr_phys, GFP_KERNEL);
	inno_dma.rxdma_addr = dma_alloc_coherent(NULL, DMA_BUFFER_SIZE, &inno_dma.rxdma_addr_phys, GFP_KERNEL);
	if((NULL == inno_dma.txdma_addr) || (NULL == inno_dma.rxdma_addr))
		err("SPI DMA alloc memory error");
	memset(inno_dma.txdma_addr, 0, PAGE_ALIGN(DMA_BUFFER_SIZE));
	memset(inno_dma.rxdma_addr, 0, PAGE_ALIGN(DMA_BUFFER_SIZE));

	DRCMR(ssp->drcmr_tx) = inno_dma.txdma | DRCMR_MAPVLD;
	DRCMR(ssp->drcmr_rx) = inno_dma.rxdma | DRCMR_MAPVLD;
}

static void inno_spi_dma_uninit(void)
{
	if(inno_dma.txdma != -1){
		pxa_free_dma(inno_dma.txdma);
		inno_dma.txdma = -1;
	}

	if(inno_dma.rxdma != -1){
		pxa_free_dma(inno_dma.rxdma);
		inno_dma.rxdma = -1;
	}

	if(inno_dma.txdma_addr != NULL){
		dma_free_coherent(NULL, DMA_BUFFER_SIZE, inno_dma.txdma_addr, inno_dma.txdma_addr_phys);
		inno_dma.txdma_addr = NULL;
	}

	if(inno_dma.rxdma_addr != NULL){
		dma_free_coherent(NULL, DMA_BUFFER_SIZE, inno_dma.rxdma_addr, inno_dma.rxdma_addr_phys);
		inno_dma.rxdma_addr = NULL;
	}

}

/*
 * start launch DMA for transfer @size bytes
 */

static void inno_spi_rxdma_start(u32 size)
{
	struct ssp_device *ssp = inno_spi_dev.ssp;

	DCSR(inno_dma.rxdma)  = DCSR_NODESC;
	DSADR(inno_dma.rxdma) = ssp->phys_base + SSDR;
	DTADR(inno_dma.rxdma) = inno_dma.rxdma_addr_phys;
	DCMD(inno_dma.rxdma)  = DCMD_INCTRGADDR | DCMD_FLOWSRC | DCMD_ENDIRQEN | DCMD_WIDTH1 | DCMD_BURST8 | (size & DCMD_LENGTH);

	DCSR(inno_dma.rxdma) |= DCSR_RUN;
}

static void inno_spi_txdma_start(u32 size)
{
	struct ssp_device *ssp = inno_spi_dev.ssp;

	DCSR(inno_dma.txdma)  = DCSR_NODESC;
	DSADR(inno_dma.txdma) = inno_dma.txdma_addr_phys;
	DTADR(inno_dma.txdma) = ssp->phys_base + SSDR;
	DCMD(inno_dma.txdma)  =  DCMD_INCSRCADDR | DCMD_FLOWTRG | DCMD_ENDIRQEN | DCMD_WIDTH1 | DCMD_BURST8 | (size & DCMD_LENGTH);

	DCSR(inno_dma.txdma) |= DCSR_RUN;
}

static void inno_spi_dma_start(u32 size)
{
	struct ssp_device *ssp = inno_spi_dev.ssp;
	uint32_t sscr0;

	sscr0 = __raw_readl(ssp->mmio_base + SSCR0);
	sscr0 &= ~SSCR0_SSE;
	sscr0 |= SSCR0_SSE;
	__raw_writel(sscr0, ssp->mmio_base + SSCR0);
	
	inno_spi_rxdma_start(size);
	inno_spi_txdma_start(size);
}

#endif	//INNO_SPI_DMA

/*
 * configure SPI GPIO Setting
 */

static void inno_spi_cs_to_gpio(void)
{
	/* change SSP CS pin to GPIO or SSP function */
	if(g_platform_spi->spi_cs_to_gpio)
		g_platform_spi->spi_cs_to_gpio();
}

static void inno_spi_cs_to_ssp(void)
{
	/* change SSP CS pin to GPIO or SSP function */
	if(g_platform_spi->spi_cs_to_ssp)
		g_platform_spi->spi_cs_to_ssp();
}

static void inno_spi_pin_config(int spi_type)
{
#if 0
	if (gpio_request(spi_cs_gpio, "IF101_SPI_CS")) {
		printk(KERN_ERR "Request GPIO failed,"
				"gpio: %d \n", spi_cs_gpio);
		return;
	}
#endif
	switch(spi_type){
		case INNO_SPI_MASTER:
		{
			/* change SSP CS pin to GPIO output function */
			gpio_direction_output(spi_cs_gpio, 1);
			inno_spi_cs_to_gpio();
			spi_cs_drive = 1;
			break;
		}
		case INNO_SPI_SLAVE:
		{
			/* change SSP CS pin to SSP function */
			inno_spi_cs_to_ssp();
			break;
		}
	}
#if 0
	gpio_free(spi_cs_gpio);
#endif
}

/*******************************
 * SPI master operations
 * *****************************/

static void inno_spi_master_start_frame(void)
{
	gpio_set_value(spi_cs_gpio, 0);
}

static void inno_spi_master_end_frame(void)
{
	/* do not de-assert SPI CS if drive to low */
	if(spi_cs_drive==1)
		gpio_set_value(spi_cs_gpio, 1);
}

/*
 * inno_spi_master_write_bytes
 * 		@pdata	-- data to be write
 * 		@length	--	how many bytes to be write
 * return
 *		@null
 */
static int inno_spi_master_write_bytes(u8* pdata, u32 length)
{
	int count = 0;
	u32 uu;
#ifdef INNO_SPI_DMA
	u32 len_dma = length;
	u8* pbuf = pdata;
#endif

	inno_spi_master_start_frame();
	
#ifdef INNO_SPI_DMA	//USE DMA
	if(length > 256){
		//filled to multiple of 8bytes
		len_dma = (length/8)*8;
		while(len_dma > 0){
			count = min(len_dma, (u32)DMA_BUFFER_SIZE);
			memcpy(inno_dma.txdma_addr, pbuf, count);
			dma_txflag = 0;
			inno_spi_dma_start(count);
			wait_event_interruptible(dma_tx_wq, dma_txflag!=0);
			len_dma -= count;
			length -= count;
			pbuf += count;
		}
		if(length>len_dma) {
			/* send left bytes */
			for(count=0; count<(length-len_dma); count++){
				ssp_write_word(&inno_spi_dev, pbuf[count]);
				/* Read null data back from device to prevent SSP overflow */
				ssp_read_word(&inno_spi_dev, &uu);
			}
		}
		goto out;
	}
#endif
	for(count=0; count<length; count++){
		ssp_write_word(&inno_spi_dev, pdata[count]);
		/* Read null data back from device to prevent SSP overflow */
		ssp_read_word(&inno_spi_dev, &uu);
		//ssp_flush(&inno_spi_dev);
	}

out:	
	inno_spi_master_end_frame();
	return 0;
}


/* 
 * note: if101 chip need nCS keep low staus when read until datas read end.
 * if we use spi controller to control nCS, the nCS will pull high after read every byte, 
 * so we should manual control nCS pin
 */
static int inno_spi_master_read_bytes(u8* pdata, u32 len)
{
	u32 count = 0;	
#ifdef INNO_SPI_DMA
	u32 len_dma = len;
	u8* pbuf = pdata;
#endif
	u32 data;

	inno_spi_master_start_frame();

#ifdef INNO_SPI_DMA	//USE DMA
	if(len > 256){
		if(len%8!=0)	//filled to multiple of 8bytes
			len_dma = (len+7)/8*8;
		while(len_dma > 0){
			count = min(len_dma, (u32)DMA_BUFFER_SIZE);
			dma_rxflag = 0;
			inno_spi_dma_start(count);
			wait_event_interruptible(dma_rx_wq, dma_rxflag!=0);
			len_dma -= count;
			if(pbuf != NULL)
				memcpy(pbuf, inno_dma.rxdma_addr, min(count, len));
			else return -1;
			
			len -= min(count, len);
			pbuf += min(count,len);
		}
		goto out;
	}
#endif
	for(count=0; count<len; count++){
		ssp_write_word(&inno_spi_dev, 0);	// in order to generate clock to receive
		ssp_read_word(&inno_spi_dev, &data);
		*(pdata+count) = data & 0xff;
	}
out:	
	inno_spi_master_end_frame();

	return 0;
}

static int inno_spi_master_init(void)
{
	int ret;
	u32 spi_clk_mode;

	if(spi_inited && spi_type==INNO_SPI_MASTER)
		return 0;

	switch(master_clk_mode){
		case SPI_MODE_3:	//idle high, transmit falling edge, receive rising edge.
			spi_clk_mode = SSCR1_SPH | SSCR1_SPO;
			break;
		case SPI_MODE_1:	//idle high, transmit rising edge, receive falling edge
			spi_clk_mode = SSCR1_SPO;
	        break;
		case SPI_MODE_2:	//idle low,transmit rising edge, receive falling edge
		    spi_clk_mode = SSCR1_SPH;
			break;
		case SPI_MODE_0:	//idle low, transmit falling edge, receive rising edge.
		default:
			spi_clk_mode = 0;
			break;
	}
	
	inno_spi_pin_config(INNO_SPI_MASTER);

	memset(&inno_spi_dev, 0, sizeof(struct ssp_dev));
	ret = ssp_init(&inno_spi_dev, ssp_port, 0);

	if(ret){
		err("%s Unable Init SPI %d.\r\n", __FUNCTION__, ssp_port);
	}else{
		ssp_disable(&inno_spi_dev);
		ssp_config(&inno_spi_dev, (SSCR0_Motorola | (SSCR0_DSS & 0x07)), // SPI, 8bit data
					(spi_clk_mode),	\
					0,	\
					SSCR0_SerClkDiv(2));		//6.5MHz
	#ifdef INNO_SPI_DMA
		inno_spi_dma_init();
	#endif
		ssp_enable(&inno_spi_dev);
		ssp_flush(&inno_spi_dev);
	}

	return 0;
}

static int inno_spi_master_uninit(void)
{
#ifdef INNO_SPI_DMA
	inno_spi_dma_uninit();
#endif
	ssp_exit(&inno_spi_dev);

	return 0;
}

/*******************************
 * SPI slave operations
 * *****************************/
static int inno_spi_slave_init(void)
{
	int ret;
	struct ssp_device *ssp = inno_spi_dev.ssp;
	uint32_t sscr0, sscr1;

	u32 spi_clk_mode = 0;

	switch(slave_clk_mode){
		case SPI_MODE_3:	//idle high, transmit falling edge, receive rising edge.
			spi_clk_mode = SSCR1_SPH | SSCR1_SPO;
			break;
		case SPI_MODE_1:	//idle high, transmit rising edge, receive falling edge
			spi_clk_mode = SSCR1_SPO;
	        break;
		case SPI_MODE_2:	//idle low,transmit rising edge, receive falling edge
		    spi_clk_mode = SSCR1_SPH;
			break;
		case SPI_MODE_0:	//idle low, transmit falling edge, receive rising edge.
		default:
			spi_clk_mode = 0;
			break;
	}
	memset(&inno_spi_dev, 0, sizeof(struct ssp_dev));
	ret = ssp_init(&inno_spi_dev, ssp_port, 0);
	if(ret){
		err("%s Unable Init SPI %d.\r\n", __FUNCTION__, ssp_port);
		return ret;
	}

	clk_disable(ssp->clk);
	sscr0 = __raw_readl(ssp->mmio_base + SSCR0);
	sscr0 &= ~SSCR0_SSE;
	__raw_writel(sscr0, ssp->mmio_base + SSCR0);

	__raw_writel(0, ssp->mmio_base + SSCR0);
	__raw_writel(0, ssp->mmio_base + SSCR1);
	__raw_writel(spi_clk_mode, ssp->mmio_base + SSCR1);
	sscr1 = __raw_readl(ssp->mmio_base + SSCR1);
	sscr1 |= SSCR1_SCLKDIR | SSCR1_SFRMDIR | SSCR1_SCFR;
	__raw_writel(sscr1, ssp->mmio_base + SSCR1);
	sscr0 = __raw_readl(ssp->mmio_base + SSCR0);
	sscr0 |= (SSCR0_DSS & 0x07) |SSCR0_Motorola;
	__raw_writel(sscr0, ssp->mmio_base + SSCR0);
	
	inno_spi_pin_config(INNO_SPI_SLAVE);	
	
	clk_enable(ssp->clk);
	sscr0 = __raw_readl(ssp->mmio_base + SSCR0);
	sscr0 |= SSCR0_SSE;
	__raw_writel(sscr0, ssp->mmio_base + SSCR0);

	return 0;
}

static int inno_spi_slave_uninit(void)
{
	struct ssp_device *ssp = inno_spi_dev.ssp;
	uint32_t sscr0;

	sscr0 = __raw_readl(ssp->mmio_base + SSCR0);
	sscr0 &= ~SSCR0_SSE;
	__raw_writel(sscr0, ssp->mmio_base + SSCR0);

	ssp_exit(&inno_spi_dev);
	return 0;
}

static int inno_spi_slave_write_bytes(u8* pdata, u32 length)
{
	struct ssp_device *ssp = inno_spi_dev.ssp;

	int i=0;
	int timeout;
	
	for(i=0; i<length; i++){
		timeout = 3;	
		while ((__raw_readl(ssp->mmio_base + SSSR) & SSSR_TNF)==0 && timeout--){
			udelay(100);
		}
		if(timeout <= 0){
			err("%s timeout(%d)", __FUNCTION__, i);	
			return -1;
		}
		__raw_writel(pdata[i], ssp->mmio_base + SSDR);
	
	}

	return 0;
}

static int inno_spi_slave_read_bytes(u8* pdata, u32 length)
{
	struct ssp_device *ssp = inno_spi_dev.ssp;
	uint32_t sssr;

	int i=0;
	int timeout;
	
	for(i=0; i<length; i++){
		timeout = 1000;
		if((sssr = __raw_readl(ssp->mmio_base + SSSR)) & SSSR_ROR){
			err("%x recv FIFO overrun(%d)", sssr, i);
			sssr |= SSSR_ROR;
			__raw_writel(sssr, ssp->mmio_base + SSSR);
		}
		while (((__raw_readl(ssp->mmio_base + SSSR) & SSSR_RNE)==0) && timeout--){
			udelay(1);
		}
		if(timeout<=0){
			err("%s timeout (%d), SSSR_P=0x%x", __FUNCTION__, i, __raw_readl(ssp->mmio_base + SSSR));
			return -1;
		}

		*(pdata+i)= (u8)__raw_readl(ssp->mmio_base + SSDR);
	}

	return 0;
}

/****************************
 * Export functions
 ****************************/
int inno_spi_write_bytes(u8* pdata, u32 len)
{
	if(INNO_SPI_MASTER == spi_type)
		return inno_spi_master_write_bytes(pdata, len);
	else	// INNO_SPI_SLAVE
		return inno_spi_slave_write_bytes(pdata, len);
	
}

int inno_spi_read_bytes(u8* pdata, u32 len)
{
	if(INNO_SPI_MASTER == spi_type)
		return inno_spi_master_read_bytes(pdata, len);
	else
		return inno_spi_slave_read_bytes(pdata, len);
}

int inno_spi_init(u8 type)
{
	int ret;

	if(ssp_port<0) {
		err("invalid ssp port %d", ssp_port);
		return -EINVAL;
	}

	switch(type){
		case INNO_SPI_SLAVE:
			ret = inno_spi_slave_init();
			break;
		case INNO_SPI_MASTER:
		default:
			ret = inno_spi_master_init();
			break;
	}

	if(ret){
		err("inno_spi_init(%s) failed.\r\n", (INNO_SPI_SLAVE==type)?"slave":"master");
		spi_inited = 0;
	}

	spi_inited = 1;
	spi_type = type;
	return 0;
}

int inno_spi_uninit(void)
{
	int ret = 0;

	if(INNO_SPI_MASTER == spi_type){
		if(spi_inited)
			ret = inno_spi_master_uninit();
		
	}else{	//slave mode
		if(spi_inited)
			ret = inno_spi_slave_uninit();
	}
	spi_inited = 0;

	return ret;
}

/*
 * This function may hard to understand,  it only use in download fireware process to sync chip select
 *
 * make sure that, when call this function, we are in slave mode.
 */
int inno_spi_sync_cs(void)
{
	struct ssp_device *ssp = inno_spi_dev.ssp;
	uint32_t sscr0;

	int times = 100;
	int ret = 0;
	
	if(INNO_SPI_MASTER == spi_type)
		goto out;

	sscr0 = __raw_readl(ssp->mmio_base + SSCR0);
	sscr0 &= ~SSCR0_SSE;
	__raw_writel(sscr0, ssp->mmio_base + SSCR0);

	/* change SSP CS pin to GPIO input function */
	gpio_direction_input(spi_cs_gpio);
	inno_spi_cs_to_gpio();

	while(gpio_get_value(spi_cs_gpio) == 0){
		udelay(10);
		if(!--times){
			err("%s timeout", __FUNCTION__);	
			ret = -1;
			goto out;
		}
	}
	/* change SSP CS pin to SSP function */
	inno_spi_cs_to_ssp();

	sscr0 = __raw_readl(ssp->mmio_base + SSCR0);
	sscr0 |= SSCR0_SSE;
	__raw_writel(sscr0, ssp->mmio_base + SSCR0);
out:
	return ret;
}

int inno_spi_drive_cs(int drive)
{
	if(drive) {
		/* drive to high */
		spi_cs_drive = 1;
		inno_spi_master_end_frame();
	}
	else{
		/* drive to low */
		inno_spi_master_start_frame();
		spi_cs_drive = 0;
	}
	return 0;
}

/*************************************
 * SPI Module Init / Exit
 ************************************/
static int __init inno_spi_probe(struct platform_device *pdev)
{
	struct innodev_platform_spi *platform_spi;

	platform_spi = pdev->dev.platform_data;
	if(platform_spi==NULL)
	{
		err("no platform data, can not initialize spi");
		return -EINVAL;
	}
	g_platform_spi = platform_spi;
	ssp_port = platform_spi->ssp_port;
	spi_cs_gpio = platform_spi->spi_cs_gpio;

	spi_inited = 0;
	spi_type = INNO_SPI_MASTER;
	return 0;
}

static int inno_spi_remove(struct platform_device *pdev)
{
	if(spi_inited)
		inno_spi_uninit();

	ssp_port = -1;
	g_platform_spi = NULL;
	return 0;
}

static struct platform_driver innodev_spi_driver = {
	.driver         = {
		.name   = "inno-spi",
	},
	.probe          = inno_spi_probe,
	.remove         = inno_spi_remove,
};

static int __init inno_spi_module_init(void)
{
	return platform_driver_register(&innodev_spi_driver);
}

static void __exit inno_spi_module_deinit(void)
{
	platform_driver_unregister(&innodev_spi_driver);
}

module_init(inno_spi_module_init);
module_exit(inno_spi_module_deinit);

module_param(master_clk_mode, ushort, (S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP));
MODULE_PARM_DESC(master_clk_mode, "innofidei if101 master spi clock mode");

module_param(slave_clk_mode, ushort, (S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP));
MODULE_PARM_DESC(slave_clk_mode, "innofidei if101 slave spi clock mode");

EXPORT_SYMBOL(inno_spi_init);
EXPORT_SYMBOL(inno_spi_uninit);
EXPORT_SYMBOL(inno_spi_write_bytes);
EXPORT_SYMBOL(inno_spi_read_bytes);
EXPORT_SYMBOL(inno_spi_sync_cs);
EXPORT_SYMBOL(inno_spi_drive_cs);

MODULE_AUTHOR("jimmy.li <lizhengming@innofidei.com>");
MODULE_DESCRIPTION("Innofidei IF101 chip SPI driver");
MODULE_LICENSE("GPL");
