/*
   Copyright (C) 2005, Intel Corporation.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.


 *(C) Copyright 2006 Marvell International Ltd.
 * All Rights Reserved
 */

/*****************************************************************************
 * Abstract:
 *	Contains all hardware related functions for INS3510
 *
 * Notes:Only valid for processor code named Monahans.
 *****************************************************************************/

#include <linux/init.h>
#include <linux/module.h>
#include <linux/delay.h>
#include <linux/err.h>
#include <linux/clk.h>
#include <linux/i2c.h>
#include <linux/sysfs.h>
#include <linux/firmware.h>
#if defined(CONFIG_PXA3xx_DVFM)
#include <mach/dvfm.h>
#endif
#ifdef CONFIG_ANDROID_POWER
#include <linux/android_power.h>
#endif

#include "camera.h"

//#define INS3510_STANDBY		/* enter standby while not using */

/*****************************************************************************
 * Constants & Structures
 *****************************************************************************/
#define CAMERA_DEBUG
//#undef CAMERA_DEBUG

static int cam_debug = 4;
static int i2c_debug = 1;

#ifdef  CAMERA_DEBUG
#define CAM_dbg(level, fmt, arg...)    if(level<=cam_debug) printk(KERN_INFO fmt, ##arg)
#define I2C_dbg(level, fmt, arg...)    if(level<=i2c_debug) printk(KERN_INFO fmt, ##arg)
#else 
#define CAM_dbg(level, fmt, arg...)    do {} while (0)
#define I2C_dbg(level, fmt, arg...)    do {} while (0)
#endif


/* CICLK for QCI to sensor */
#define CICLK	2600		/* 26 Mhz */

/* Revision constants */
#define INS3510_CHIP_VER_BOOT	0x090102ff

/* Output Size & Format */
#define INS3510_FORMAT_YUV_422  1
#define INS3510_FORMAT_JPG	4
#define INS3510_FORMAT_RGB565	5

#define INS3510_MAX_WIDTH	2592
#define INS3510_MAX_HEIGHT	1944

struct ins3510fw_header {
	unsigned long magic;
	unsigned long title_off; 	/* description string */
	unsigned long title_size;
	unsigned long spi_off;		/* SPI flash loader data */
	unsigned long spi_size;
	unsigned long patch_off;	/* patch data */
	unsigned long patch_size;
	unsigned long reserved;
};

struct ins3510hw_data {
	u32 orig_ae_upper_et_a;
	u32 orig_ae_upper_et_b;
	u32 orig_ae_max_et_a;
	u32 orig_ae_max_et_b;

	int clock_on;
	int voltage_on;
	int power_on;
	int standby;
	int bootup;
	int mode_change;
	int af_wait;

	int flashled_flag;
	int flashled_fire;
};

/****************************************************************************
   Gloable variable
*****************************************************************************/
struct frame_sizes_t {
	u32	width;
	u32	height;
};

/* preview YUV422 size list, supported preview frame sizes "preview-size-values" on Android */
static const struct frame_sizes_t ins3510_yuv422_frame_sizes[] = {
	{640, 480},
	{432, 320},
	{320, 240},
	{160, 120},
	{128, 96},
	{352, 288},
	{176, 144}
};

/* capture JPEG size list, supported capture frame sizes "picture-size-values" on Android */
static const struct frame_sizes_t ins3510_jpeg_frame_sizes[] = {
	{2048, 1536},
	{1600, 1200},
	{1280, 960},
	{1024, 768},
	{800, 600},
	{640, 480},
	{320, 240}
};

/* frame interval list,  frame rate = 1 / frame interval,
   supported frame rates "preview-frame-rate-values" on Android */
static const struct v4l2_fract ins3510_yuv422_frame_intervals[] = {
	{3, 15},	/* 1/5 seconds,  5fps, the frame rate could be down to 5 */
	{1, 15},	/* 1/15 seconds, 15fps */
	{1, 20},	/* 1/20 seconds, 20fps */
	{1, 24},	/* 1/24 seconds, 24fps */
	{1, 25},	/* 1/25 seconds, 25fps */
	{1, 30}		/* 1/30 seconds, 30fps */
};

/* Unique ID allocation */
static struct i2c_client *g_client;

static struct clk *sensor_clk = NULL;
static u32 sensor_mclk = CICLK;
static struct ins3510hw_data ins3510hw_data;

static struct delayed_work flash_work;
static struct workqueue_struct *flash_wqueue;

#if defined(CONFIG_PXA3xx_DVFM)
static int dvfm_dev_idx;
#endif

#ifdef CONFIG_ANDROID_POWER
static android_suspend_lock_t ins3510_ops_suspend_lock = {
        .name = "ins3510_ops"
};
#endif /* CONFIG_ANDROID_POWER */


/*****************************************************************************
 * Function Prototype
 *****************************************************************************/

static int   ins3510_i2c_read_reg8( const u8 subAddress, u8 *preg );
static int   ins3510_i2c_write_reg8( const u8 subAddress, u8 reg );
static int   ins3510_i2c_read_reg16( const u8 subAddress, u16 *preg );
static int   ins3510_i2c_write_reg16( const u8 subAddress, u16 reg );
static int   ins3510_i2c_read_reg32( const u8 subAddress, u32 *preg );
static int   ins3510_i2c_write_reg32( const u8 subAddress, u32 reg );

/*****************************************************************************
 *  Initialize
 *****************************************************************************/

static void ins3510hw_data_init(void)
{
	memset(&ins3510hw_data, 0, sizeof(ins3510hw_data));

	ins3510hw_data.clock_on = 0;
	ins3510hw_data.voltage_on = 0;
	ins3510hw_data.power_on = 0;
	ins3510hw_data.standby = 0;
	ins3510hw_data.bootup = 0;
	ins3510hw_data.mode_change = -1;
	ins3510hw_data.af_wait = 0;
	ins3510hw_data.flashled_flag = 0;
	ins3510hw_data.flashled_fire = 0;
}

static void ins3510hw_data_reset(void)
{
	ins3510hw_data.mode_change = -1;
	ins3510hw_data.af_wait = 0;
	ins3510hw_data.flashled_flag = 0;
	ins3510hw_data.flashled_fire = 0;
}

/*****************************************************************************
 *  Power & Reset
 *****************************************************************************/

static int sensor_clock_on(void)
{
#ifdef CONFIG_PXA3xx_DVFM
        dvfm_disable_op_name("D2", dvfm_dev_idx);
        dvfm_disable_op_name("D0CS", dvfm_dev_idx);
        dvfm_disable_op_name("104M", dvfm_dev_idx);
        dvfm_disable_op_name("156M", dvfm_dev_idx);
        dvfm_disable_op_name("208M", dvfm_dev_idx);
#endif
	sensor_clk = clk_get(NULL, "CAMCLK");
	if (IS_ERR(sensor_clk)) {
		printk(KERN_ERR "ins3510: sensor failed to get camera clock\n");
		sensor_clk = NULL;
		return -EINVAL;
	}
	clk_enable(sensor_clk);
	ci_set_clock(1, 1, sensor_mclk);

	ins3510hw_data.clock_on = 1;
	return 0;
}

static void sensor_clock_off(void)
{
	if ((sensor_clk==NULL) || IS_ERR(sensor_clk)) {
		printk(KERN_ERR "ins3510: sensor_clock_off() invalid handler\n");
		return;
	}
	ci_set_clock(0, 0, sensor_mclk);
	clk_disable(sensor_clk);
	clk_put(sensor_clk);

#ifdef CONFIG_PXA3xx_DVFM
        dvfm_enable_op_name("D2", dvfm_dev_idx);
        dvfm_enable_op_name("D0CS", dvfm_dev_idx);
        dvfm_enable_op_name("104M", dvfm_dev_idx);
        dvfm_enable_op_name("156M", dvfm_dev_idx);
        dvfm_enable_op_name("208M", dvfm_dev_idx);
#endif
	ins3510hw_data.clock_on = 0;
}

static void sensor_reset(void)
{
	struct sensor_platform_data *pdata;
	pdata = g_client->dev.platform_data;

	if(pdata && pdata->reset)
		pdata->reset(pdata->id);

	ins3510hw_data.bootup = 0;
}

static void sensor_power_on(void)
{
	struct sensor_platform_data *pdata;
	pdata = g_client->dev.platform_data;

	if(pdata && pdata->power_on)
		pdata->power_on(pdata->id);

	ins3510hw_data.power_on = 1;
	ins3510hw_data.bootup = 0;
}

static void sensor_power_off(void)
{
	struct sensor_platform_data *pdata;
	pdata = g_client->dev.platform_data;

	if(pdata && pdata->power_off)
		pdata->power_off(pdata->id);

	ins3510hw_data.power_on = 0;
	ins3510hw_data.bootup = 0;
}

static void sensor_voltage_on(void)
{
	struct sensor_platform_data *pdata;
	pdata = g_client->dev.platform_data;

	if(pdata && pdata->voltage_on)
		pdata->voltage_on(pdata->id);

	ins3510hw_data.voltage_on = 1;
	ins3510hw_data.power_on = 1;
	ins3510hw_data.bootup = 0;
}

static void sensor_voltage_off(void)
{
	struct sensor_platform_data *pdata;
	pdata = g_client->dev.platform_data;

	if(pdata && pdata->voltage_off)
		pdata->voltage_off(pdata->id);

	ins3510hw_data.voltage_on = 0;
	ins3510hw_data.power_on = 0;
	ins3510hw_data.bootup = 0;
}

static void sensor_standby(void)
{
	struct sensor_platform_data *pdata;
	pdata = g_client->dev.platform_data;

	if(pdata && pdata->standby)
		pdata->standby(pdata->id);

	ins3510hw_data.standby = 1;
}

static void sensor_wakeup(void)
{
	struct sensor_platform_data *pdata;
	pdata = g_client->dev.platform_data;

	if(pdata && pdata->wakeup)
		pdata->wakeup(pdata->id);

	ins3510hw_data.standby = 0;
}

static void cam_flash_fire(void)
{
	struct sensor_platform_data *pdata;
	pdata = g_client->dev.platform_data;

	if(pdata->flash_ctrl)
		pdata->flash_ctrl(1);
}

static void cam_flash_off(void)
{
	struct sensor_platform_data *pdata;
	pdata = g_client->dev.platform_data;

	if(pdata->flash_ctrl)
		pdata->flash_ctrl(0);
}

static void ins3510hw_print_all_regs(void)
{
	u8 regval;
	int i, j, n;
	int i2c_debug_back;
	char dump_str[80];

	i2c_debug_back = i2c_debug;
	i2c_debug = 4;		/* disable i2c detail log */
	printk(KERN_INFO "     0  1  2  3  4  5  6  7  8  9  a  b  c  d  e  f\n");
	for(i=0; i<0x10; i++)
	{
		n = 0;
		n += sprintf(dump_str+n, "%02x:", i<<4);
		for(j=0; j<0x10; j++)
		{
			ins3510_i2c_read_reg8(i*16+j, &regval);
			n += sprintf(dump_str+n, " %02x", regval);
		}
		printk(KERN_INFO "%s\n", dump_str);
	}
	i2c_debug = i2c_debug_back;
}

static int ins3510hw_dump_all_regs(char *dump_buf, int size)
{
	u8 regval;
	int i, j, n;

	n = 0;
	n += snprintf(dump_buf, size-n, "     0  1  2  3  4  5  6  7  8  9  a  b  c  d  e  f\n");
	if( n>=size )
		return n;
	for(i=0; i<0x10; i++)
	{
		n += snprintf(dump_buf+n, size-n, "%02x:", i<<4);
		if( n>=size )
			return n;
		for(j=0; j<0x10; j++)
		{
			ins3510_i2c_read_reg8(i*16+j, &regval);
			n += snprintf(dump_buf+n, size-n, " %02x", regval);
			if( n>=size )
				return n;
		}
		n += snprintf(dump_buf+n, size-n, "\n");
		if( n>=size )
			return n;
	}
	return n;
}

static int ins3510hw_bootup(void)
{
	int ret;
	u8 reg8;
	u16 reg16;
	u32 reg32, reg32s;
	int loop;

	/* patch bootup */
	CAM_dbg(4, "INS3510: bootup.\n");

	ret = ins3510_i2c_write_reg16(0xe0, 0x820c);
	if( ret<0 )
		return ret;
	loop = 500;
	while(loop-->0)
	{
		ret = ins3510_i2c_read_reg8(0xf8, &reg8);
		if( ret<0 )
			return ret;
		if( (reg8 & 0x20)!=0 )
			break;
		msleep(10);
	}
	if(loop<=0)
	{
		printk(KERN_ERR "ins3510: power on timeout to wait 0x820c8\n");
		ins3510hw_print_all_regs();
		return -ETIME;
	}

	/* IO slew rate */
	ret = ins3510_i2c_write_reg16(0xe0, 0x8201);
	if( ret<0 )
		return ret;
	ret = ins3510_i2c_read_reg8(0xf7, &reg8);
	if( ret<0 )
		return ret;
	//reg8 |= 0xf0;		/* TODO: */
	ret = ins3510_i2c_write_reg8(0xf7, reg8);		/* 0x82017 */
	if( ret<0 )
		return ret;

	/* clock setting */
	ret = ins3510_i2c_write_reg16(0xe0, 0x4009);
	if( ret<0 )
		return ret;
	ret = ins3510_i2c_write_reg32(0xf0, 0x0786f792);	/* 0x40090 */
	if( ret<0 )
		return ret;
	ret = ins3510_i2c_write_reg32(0xf4, 0x00000007);	/* 0x40094 */
	if( ret<0 )
		return ret;
	ret = ins3510_i2c_write_reg16(0xe0, 0x400f);
	if( ret<0 )
		return ret;
	ret = ins3510_i2c_write_reg32(0xf0, 0x11000601);	/* 0x400f0 */
	if( ret<0 )
		return ret;
	ret = ins3510_i2c_read_reg32(0xf4, &reg32);
	if( ret<0 )
		return ret;
	reg32 = (reg32 & 0x000000ff) | 0x00070100;
	ret = ins3510_i2c_write_reg32(0xf4, reg32);		/* 0x400f4 */
	if( ret<0 )
		return ret;

	/* read back PLL settings to debug bootup issue */
	reg32s = reg32;
	ins3510_i2c_write_reg16(0xe0, 0x4009);
	ins3510_i2c_read_reg32(0xf0, &reg32);
	if( reg32!=0x0786f792 )
		CAM_dbg(4, "INS3510: readback 0x40090 did not match 0x0786f792:0x%08x.\n", reg32);
	ins3510_i2c_read_reg32(0xf4, &reg32);
	if( reg32!=0x00000007 )
		CAM_dbg(4, "INS3510: readback 0x40094 did not match 0x00000007:0x%08x.\n", reg32);
	ins3510_i2c_write_reg16(0xe0, 0x400f);
	ins3510_i2c_read_reg32(0xf0, &reg32);
	if( reg32!=0x11000601 )
		CAM_dbg(4, "INS3510: readback 0x400f0 did not match 0x11000601:0x%08x.\n", reg32);
	ins3510_i2c_read_reg32(0xf4, &reg32);
	if( reg32!=reg32s )
		CAM_dbg(4, "INS3510: readback 0x400f0 did not match 0x%08x:0x%08x.\n", reg32s, reg32);
	/* end of debug */

	ret=ins3510_i2c_read_reg16(0xaa, &reg16);
	if( ret<0 )
		return ret;
	ret = ins3510_i2c_write_reg16(0xaa, 0x0014);
	if( ret<0 )
		return ret;
	msleep(1);
	ret = ins3510_i2c_write_reg16(0xe0, 0x820c);
	if( ret<0 )
		return ret;
	ret = ins3510_i2c_write_reg16(0xf8, 0x0000);
	if( ret<0 )
		return ret;
	ret = ins3510_i2c_write_reg16(0xfa, 0x0000);
	if( ret<0 )
		return ret;

	msleep(10);
	loop = 200;
	while(loop-->0)
	{
		ret = ins3510_i2c_read_reg8(0xf8, &reg8);
		if( ret<0 )
			return ret;
		if( (reg8 & 0x20)!=0 )
			break;
		msleep(10);
	}
	if(loop<=0)
	{
		printk(KERN_ERR "ins3510: bootup timeout to wait 0x820c8\n");
		ins3510hw_print_all_regs();
		return -ETIME;
	}

	loop = 500;
	while(loop-->0)
	{
		ret = ins3510_i2c_read_reg8(0x84, &reg8);
		if( ret<0 )
			return ret;
		if( (reg8 & 0x80)==0 )
			break;
		msleep(10);
	}
	if(loop<=0)
	{
		printk(KERN_ERR "ins3510: bootup timeout to wait 0x84\n");
		ins3510hw_print_all_regs();
		return -ETIME;
	}

	ins3510hw_data.bootup = 1;
	return 0;
}

static int ins3510hw_get_original(void)
{
	ins3510_i2c_read_reg32(0x28, &ins3510hw_data.orig_ae_upper_et_a);
	ins3510_i2c_read_reg32(0x68, &ins3510hw_data.orig_ae_upper_et_b);
	ins3510_i2c_read_reg32(0x2c, &ins3510hw_data.orig_ae_max_et_a);
	ins3510_i2c_read_reg32(0x6c, &ins3510hw_data.orig_ae_max_et_b);
	return 0;
}

static int ins3510hw_init_registers(void)
{
#ifdef CONFIG_MACH_CHERRY
	ins3510_i2c_write_reg16(0xe0, 0x42b9);	/* set gain */
	ins3510_i2c_write_reg32(0xf8, 0x00050000);
	ins3510_i2c_write_reg32(0x68, 0x101d0);	/*66ms*/
#endif
	return 0;
}

static int ins3510hw_reset(void)
{
	int ret = 0;

	if( cam_debug > 4)
		ins3510hw_print_all_regs();

	if(ins3510hw_data.bootup==0) {
		ret = ins3510hw_bootup();
		ins3510hw_init_registers();

		ins3510hw_get_original();
	}

	if( cam_debug > 4)
		ins3510hw_print_all_regs();

	ins3510hw_data.mode_change = -1;
	ins3510hw_data.af_wait = 0;

	return ret;
}

static void ins3510hw_outputpin_hiz(void)
{
	u8 reg8;

	if( ins3510_i2c_write_reg16(0xe0, 0x9800)<0 )
		return;
	ins3510_i2c_read_reg8(0xf3, &reg8);
	reg8 &= 0xef;
printk("ins3510: output hiz [0x98003]=%02x\n", reg8);
	ins3510_i2c_write_reg8(0xf3, reg8);

	msleep(1);
}

static void ins3510hw_outputpin_on(void)
{
	u8 reg8;

	if( ins3510_i2c_write_reg16(0xe0, 0x9800)<0 )
		return;
	ins3510_i2c_read_reg8(0xf3, &reg8);
	reg8 |= 0x10;
	ins3510_i2c_write_reg8(0xf3, reg8);
printk("ins3510: output on [0x98003]=%02x\n", reg8);
}

static int ins3510hw_power_down( u8 powerMode )
{
	/* INS3510 PWRDWN, 0 = NORMAL, 1=POWER DOWN */
	if( powerMode == CAMERA_POWER_OFF )
	{
		ins3510hw_outputpin_hiz();
#ifdef INS3510_STANDBY
		sensor_standby();
#else
		sensor_power_off();
#endif
	}
	else
	{
#ifdef INS3510_STANDBY
		sensor_wakeup();
#else
		sensor_power_on();
#endif
		ins3510hw_outputpin_on();
		return ins3510hw_reset();
	}
	return 0;
}

/*****************************************************************************
 * patch flash upgrage
 *****************************************************************************/

static int ins3510_firmware_load_spi(const u8 *firmware, size_t size)
{
	int ret;
	u32 addr;
	const u8 *data;

	ret = ins3510_i2c_write_reg16(0xe0, 0x8200);
	if( ret<0 )
		return ret;
	ret = ins3510_i2c_write_reg8(0xfb, 0x02);
	if( ret<0 )
		return ret;
	ret = ins3510_i2c_write_reg8(0xff, 0x01);
	if( ret<0 )
		return ret;

	data = firmware;
	addr = 0;
	ret = ins3510_i2c_write_reg16(0xe0, 0);
	if( ret<0 )
		return ret;
	while(addr<size)
	{
		if( (addr&0x0f)==0 )
		{
			ret = ins3510_i2c_write_reg16(0xe0, (addr>>4)&0xffff);
			if( ret<0 )
				return ret;
		}
		ret = ins3510_i2c_write_reg8(0xf0+(addr&0x0f), data[addr]);
		if( ret<0 )
			return ret;
		addr++;
	}

	ret = ins3510_i2c_write_reg16(0xe0, 0x0000);
	if( ret<0 )
		return ret;
	ret = ins3510_i2c_write_reg8(0xf4, 0x00);
	if( ret<0 )
		return ret;

	ret = ins3510_i2c_write_reg16(0xe0, 0x8200);
	if( ret<0 )
		return ret;
	ret = ins3510_i2c_write_reg8(0xfb, 0x00);
	if( ret<0 )
		return ret;

	return 0;
}

static int ins3510_firmware_load_patch(const u8 *firmware, size_t size)
{
	int ret;
	int loop, count;
	u32 addr, loadaddr, off;
	const u8 *data;
	u8 reg8;

	ret = ins3510_i2c_write_reg16(0xe0, 0x0000);
	if( ret<0 )
		return ret;

	loop = 10000;
	do {
		ret = ins3510_i2c_read_reg8(0xf4, &reg8);
		if( ret<0 )
			return ret;
		if( (reg8 & 0x04) == 0x04 )
			break;
		loop--;
		if( loop==0 )
		{
			printk(KERN_ERR "ins3510 load patch: timeout to wait SPI erased.\n");
			return -ETIME;
		}
		msleep(10);
	} while(1);

	data = firmware;
	loadaddr = (data[4]<<24) | (data[5]<<16) | (data[6]<<8) | data[7];
	off = 0;
	while( off<(size&0xfffffffc) )
	{
		ret = ins3510_i2c_write_reg8(0xf0, data[0]);
		if( ret<0 )
			return ret;
		ret = ins3510_i2c_write_reg8(0xf1, data[1]);
		if( ret<0 )
			return ret;
		ret = ins3510_i2c_write_reg8(0xf2, data[2]);
		if( ret<0 )
			return ret;
		ret = ins3510_i2c_write_reg8(0xf3, data[3]);
		if( ret<0 )
			return ret;
		ret = ins3510_i2c_write_reg8(0xf4, 0x01);
		if( ret<0 )
			return ret;
		loop = 1000;
		do {
			ret = ins3510_i2c_read_reg8(0xf4, &reg8);
			if( ret<0 )
				return ret;
			if( (reg8 & 0x01) == 0 )
				break;
			loop--;
			if( loop==0 )
			{
				printk(KERN_ERR "ins3510 load patch: timeout to wait data writing.\n");
				return -ETIME;
			}
			msleep(10);
		} while(1);
		data += 4;
		off += 4;
	}

	ret = ins3510_i2c_write_reg8(0xf4, 0x03);
	if( ret<0 )
		return ret;
	loop = 1000;
	do {
		ret = ins3510_i2c_read_reg8(0xf4, &reg8);
		if( ret<0 )
			return ret;
		if( (reg8 & 0x02) == 0 )
			break;
		loop--;
		if( loop==0 )
		{
			printk(KERN_ERR "ins3510 load patch: timeout to wait flash ended.\n");
			return -ETIME;
			break;
		}
		msleep(10);
	} while(1);

	ret = ins3510_i2c_write_reg16(0xe0, 0x8200);
	if( ret<0 )
		return ret;
	ret = ins3510_i2c_write_reg8(0xff, 0x00);
	if( ret<0 )
		return ret;
	ret = ins3510_i2c_write_reg8(0xfb, 0x02);
	if( ret<0 )
		return ret;
	ret = ins3510_i2c_write_reg8(0xfb, 0x00);
	if( ret<0 )
		return ret;

	ret = ins3510hw_bootup();
	if( ret<0 )
		return ret;

	data = firmware;
	addr = loadaddr;
	off = 0;
	ret = ins3510_i2c_write_reg16(0xe0, (addr>>4)&0xffff);
	if( ret<0 )
		return ret;
	count = 0;
	while(off<size)
	{
		if( (addr&0x0f)==0 )
		{
			ret = ins3510_i2c_write_reg16(0xe0, (addr>>4)&0xffff);
			if( ret<0 )
				return ret;
		}
		ret = ins3510_i2c_read_reg8(0xf0+(addr&0x0f), &reg8);
		if( ret<0 )
			return ret;
		if( reg8 != data[off] )
		{
			printk(KERN_ERR "ins3510 load patch: verify failed, [off:%08x]=%02x:%02x\n", off, data[off], reg8);
			count++;
			if( count>256 )
				return -EINVAL;
		}
		off++;
		addr++;
	}

	return 0;
}

static int ins3510_firmware_upgrade(void)
{
	int ret;
	/* uses the default method to get the firmware */
	const struct firmware *fw_entry_spi;
	const struct firmware *fw_entry_patch;

	printk(KERN_INFO "ins3510: requesting firmware SPI flash loader\n");
	if(request_firmware(&fw_entry_spi, "ins3510_spi_loader.fw", &g_client->dev)!=0)
	{
		printk(KERN_ERR "ins3510: Firmware ins3510_spi_loader.fw not available\n");
		return -EINVAL;
	}
	printk(KERN_INFO "ins3510: requesting firmware flash patch\n");
	if(request_firmware(&fw_entry_patch, "ins3510_flash_patch.fw", &g_client->dev)!=0)
	{
		printk(KERN_ERR "ins3510: Firmware ins3510_flash_patch.fw not available\n");
		release_firmware(fw_entry_spi);
		return -EINVAL;
	}

#ifdef CONFIG_ANDROID_POWER
	android_lock_suspend(&ins3510_ops_suspend_lock);
#endif

	sensor_clock_on();
	sensor_power_on();

	/* SPI loader */
	printk(KERN_INFO "ins3510 firmware load: SPI Flash loader %d bytes\n", fw_entry_spi->size);
	if( (ret=ins3510_firmware_load_spi(fw_entry_spi->data, fw_entry_spi->size))>=0 )
	{
		/* patch */
		printk(KERN_INFO "ins3510 firmware load: patch %d bytes\n", fw_entry_patch->size);
		if( (ret=ins3510_firmware_load_patch(fw_entry_patch->data, fw_entry_patch->size))>=0 )
			printk(KERN_INFO "ins3510: firmware upgrade ok!\n");
		else
			printk(KERN_ERR "ins3510: error, firmware upgrade failed (%d)!\n", ret);
	}
	else
		printk(KERN_ERR "ins3510: error, firmware upgrade failed (%d)!\n", ret);

	sensor_power_off();
	sensor_clock_off();

#ifdef CONFIG_ANDROID_POWER
	android_unlock_suspend(&ins3510_ops_suspend_lock);
#endif

	release_firmware(fw_entry_spi);
	release_firmware(fw_entry_patch);
	printk(KERN_INFO "ins3510: release firmware data\n");

	/* finish setting up the device */
	return 0;
}

/*****************************************************************************
 *  Register Settings Function
 *****************************************************************************/

static int ins3510hw_version_revision(u32 * pRevision)
{
	return ins3510_i2c_read_reg32( 0x80, pRevision );
}

static int ins3510hw_set_go_bit(void)
{
	int ret;
	int loop;
	u8 reg8;

	ret = ins3510_i2c_read_reg8(0x84, &reg8);
	if( ret<0 )
		return ret;

	/* set go bit */
  	ret = ins3510_i2c_write_reg8(0x84, reg8 | 0x80);
	if( ret<0 )
		return ret;

	loop = 5000;	/* 5s timeout */
	do {
		ret = ins3510_i2c_read_reg8(0x84, &reg8);
		if( ret<0 )
			return ret;
		msleep(2);
		if( --loop < 0 )
		{
			printk(KERN_ERR "error: ins3510: set go bit time out!\n");
			ins3510hw_print_all_regs();
			return -ETIME;
		}
	} while(reg8 & 0x80);
	return 0;
}

/* test pattern: 0-3 */
static int ins3510hw_set_test_pattern(int pattern)
{
	int ret;
	u8 reg8;
	ret = ins3510_i2c_read_reg8(0x84, &reg8);
	if( ret<0 )
		return ret;
	reg8 &= 0xcf;
	reg8 |= (pattern&0x03)<<4;	/* bit5:4 */
	return ins3510_i2c_write_reg8(0x84, reg8);
}

/* frame rate: 0-30, 0-auto for still capture */
static int ins3510hw_set_framerate(int fps)
{
	int ret;
	int ae_upper_et, ae_max_et;	/* us */

	if(fps==0)
	{
		ae_upper_et = ins3510hw_data.orig_ae_upper_et_b;
		ae_max_et = ins3510hw_data.orig_ae_max_et_b;
	}
	else if((fps>0) && (fps<=30))
	{
		ae_upper_et = ins3510hw_data.orig_ae_upper_et_b;
		ae_max_et = 1000000 / fps;
		if( ae_upper_et > ae_max_et )
			ae_upper_et = ae_max_et;
	}
	else
		return -EINVAL;

	CAM_dbg(4, "ins3510: set frame rate=%d, ae_upper_et=%d us, ae_max_et=%d us\n", fps, ae_upper_et, ae_max_et);

	ret = ins3510_i2c_write_reg8(0x97, fps);
	if( ret<0 )
		return ret;
	ret = ins3510_i2c_write_reg32(0x68, ae_upper_et);
	if( ret<0 )
		return ret;
	ret = ins3510_i2c_write_reg32(0x6c, ae_max_et);
	if( ret<0 )
		return ret;
	return 0;
}

/* sharpness: s.7 format, 0x80(-1) to disable */
static int ins3510hw_set_sharpness(int sharpness)
{
	return ins3510_i2c_write_reg8(0xa1, sharpness);
}

/* focus position: 0x0-0xff (0 - infinity, 0xFF - closest) */
static int ins3510hw_set_af_pos(int af_pos)
{
	return ins3510_i2c_write_reg8(0xa4, af_pos);
}

/* AE offset: s5.10 format */
static int ins3510hw_set_ae_offset(int ae_offset)
{
	return ins3510_i2c_write_reg16(0xb2, ae_offset);
}

static int ins3510hw_set_exif_orien(int exif_orien)
{
	return ins3510_i2c_write_reg8(0xc4, exif_orien);
}

/* gamms: s3.12 */
static int ins3510hw_set_gamma(int context, int gamma)
{
	if( context==0 )
		return ins3510_i2c_write_reg16(0x06, gamma);
	else
		return ins3510_i2c_write_reg16(0x46, gamma);
}

/* brightness: s3.12 */
static int ins3510hw_set_brightness(int context, int brightness)
{
	if( context==0 )
		return ins3510_i2c_write_reg16(0x08, brightness);
	else
		return ins3510_i2c_write_reg16(0x48, brightness);
}

/* contrast: s3.12 */
static int ins3510hw_set_contrast(int context, int contrast)
{
	if( context==0 )
		return ins3510_i2c_write_reg16(0x0a, contrast);
	else
		return ins3510_i2c_write_reg16(0x4a, contrast);
}

/* saturation: s3.12 */
static int ins3510hw_set_saturation(int context, int saturation)
{
	if( context==0 )
		return ins3510_i2c_write_reg16(0x0c, saturation);
	else
		return ins3510_i2c_write_reg16(0x4c, saturation);
}

/* jpeg qulaity: u4, 0x0-0x0f */
static int ins3510hw_set_jpeg_quality(int jpeg_quality)
{
	int ret;
	jpeg_quality &= 0x0f;
	ret = ins3510_i2c_write_reg8(0x10, jpeg_quality);
	if( ret<0 )
		return ret;
	return ins3510_i2c_write_reg8(0x50, jpeg_quality);
}

/* specail mode: 0 - normal */
static int ins3510hw_set_special_mode(int special_mode)
{
	int ret;
	ret = ins3510_i2c_write_reg8(0x11, special_mode);
	if( ret<0 )
		return ret;
	return ins3510_i2c_write_reg8(0x51, special_mode);
}

/* white balance mode: 0 - auto */
static int ins3510hw_set_wb_mode(int wb_mode)
{
	int ret;
	u8 reg8;

	/* context A */
	ret = ins3510_i2c_read_reg8(0x17, &reg8);
	if( ret<0 )
		return ret;
	if(wb_mode==0)
		reg8 |= 0x02;
	else
		reg8 &= 0xfd;
	ret = ins3510_i2c_write_reg8(0x17, reg8);
	if( ret<0 )
		return ret;

	ret = ins3510_i2c_read_reg8(0x30, &reg8);
	if( ret<0 )
		return ret;
	reg8 = (reg8 & 0xf0) | wb_mode;		/* bit3:0 */
	ret = ins3510_i2c_write_reg8(0x30, reg8);
	if( ret<0 )
		return ret;

	/* context B */
	ret = ins3510_i2c_read_reg8(0x57, &reg8);
	if( ret<0 )
		return ret;
	if(wb_mode==0)
		reg8 |= 0x02;
	else
		reg8 &= 0xfd;
	ret = ins3510_i2c_write_reg8(0x57, reg8);
	if( ret<0 )
		return ret;

	ret = ins3510_i2c_read_reg8(0x70, &reg8);
	if( ret<0 )
		return ret;
	reg8 = (reg8 & 0xf0) | wb_mode;		/* bit3:0 */
	return ins3510_i2c_write_reg8(0x70, reg8);
}

/* digital zoom: u4.12 format, 0x1000 - 1.0X */
static int ins3510hw_set_digital_zoom(int zoom)
{
	/* Digital Zoom Step Factor, u4.12 format */
	//ins3510_i2c_write_reg16(0x8a, 0x19);	/* step 0.025 */
	return ins3510_i2c_write_reg16(0x88, zoom);
}

static int ins3510hw_set_default(void)
{
#ifdef CONFIG_MACH_CHERRY
	ins3510hw_set_sharpness( 0x35 );	/* preview mode */
	ins3510hw_set_jpeg_quality( 0x0c );
	ins3510hw_set_contrast( 1, 0x7f00 );
	ins3510hw_set_saturation( 1, 0x1600 );
	ins3510hw_set_brightness( 1, 0x20 );
#endif
	return 0;
}

static void camerahw_flash_fire(void)
{
	if(!ins3510hw_data.flashled_fire){
		CAM_dbg(5, "ins3510: flash led on\n");
		cam_flash_fire();
		ins3510hw_data.flashled_fire = 1;

		/* start 5s timeout to off flash */
		queue_delayed_work(flash_wqueue, &flash_work, msecs_to_jiffies(5000));
	}
	else
		printk(KERN_ERR "ins3510: error, flash led already was on\n");
}

static void camerahw_flash_off(void)
{
	if(ins3510hw_data.flashled_fire){
		cam_flash_off();
		ins3510hw_data.flashled_fire = 0;
		CAM_dbg(5, "ins3510: flash led off\n");
	}
}

static void camerahw_flash_work_off(void)
{
	cancel_delayed_work(&flash_work);
	if(ins3510hw_data.flashled_fire){
		camerahw_flash_off();
	}
}

static void camera_flash_work_handler(struct work_struct *work)
{
	printk(KERN_INFO "ins3510: flashing timeout to be off\n");
	camerahw_flash_off();
}


/*****************************************************************************
 *  Viewfinder, still
 *****************************************************************************/

static int ins3510hw_enable_af(void)
{
	int ret;
	u8 af_mode;

	ret = ins3510_i2c_read_reg8(0xa3, &af_mode);
	if( ret<0 )
		return ret;
	af_mode = (af_mode & 0x0f) | 0x01;
	ret = ins3510_i2c_write_reg8(0xa3, af_mode);
	if( ret<0 )
		return ret;

	ins3510hw_data.af_wait = 1;
	return 0;
}

static int ins3510hw_wait_af_complete(void)
{
	int ret;
	u8 af_mode, af_pos, af_obj;
	int loop;

	loop = 100;
	while(loop-->0)
	{
		ret = ins3510_i2c_read_reg8(0xa3, &af_mode);
		if( ret<0 )
			return ret;
		if( (af_mode & 0x0f)==0 )
			break;
		msleep(10);
	}

	ret = ins3510_i2c_read_reg8(0xa4, &af_pos);
	if( ret<0 )
		return ret;
	ret = ins3510_i2c_read_reg8(0xa5, &af_obj);
	if( ret<0 )
		return ret;

	CAM_dbg(4, "ins3510: single AF, mode=%02x, pos=%d, obj=%d:%d\n",
		af_mode, af_pos, (af_obj&0xf0)>>4, af_obj&0x0f);

	ins3510hw_data.af_wait = 0;

	if(loop==0)
	{
		printk(KERN_ERR "ins3510: single AF timeout\n");
		return -ETIME;
	}

	/* wait a moment in order to viewfinder can show last frame */
	msleep(100);

	return 0;
}

static int ins3510hw_single_af(void)
{
	int ret;

	/* enable single auto focus */
	ret = ins3510hw_enable_af();
	if( ret<0 )
		return ret;

	ret = ins3510hw_set_go_bit();
	if( ret<0 )
	{
		printk(KERN_ERR "ins3510hw_single_af() set go failed(%d)\n", ret);
		return ret;
	}

	if( ins3510hw_data.af_wait )
	{
		CAM_dbg(4, "ins3510: wait focus complete-----------\n");
		ins3510hw_wait_af_complete();	/* ignore return status */
	}
	return 0;
}

static int ins3510hw_af_idle(void)
{
	return ins3510hw_set_af_pos(0);		/* 0 - infinity */
}

static int ins3510hw_sensor_preview2snap(void)
{
	int ret;
	u8 reg8;

	CAM_dbg(4, "ins3510_sensor_preview2snap\n");

	/* INS3510_set_context(0); */
	ret = ins3510_i2c_read_reg8(0x84, &reg8);
	if( ret<0 )
		return ret;
	reg8 &= 0xfe;
	ret = ins3510_i2c_write_reg8(0x84, reg8);
	if( ret<0 )
		return ret;

	return 0;
}

static int ins3510hw_sensor_snap2preview(void)
{
	int ret;
	u8 reg8;

	CAM_dbg(4, "ins3510_sensor_snap2preview\n");

	/* INS3510_set_context(1); */
	ret = ins3510_i2c_read_reg8(0x84, &reg8);
	if( ret<0 )
		return ret;
	reg8 |= 0x01;
	ret = ins3510_i2c_write_reg8(0x84, reg8);
	if( ret<0 )
		return ret;

	return 0;
}

static int ins3510hw_output_on(void)
{
	int ret;

	ret = ins3510hw_set_go_bit();
	if( ret<0 )
	{
		printk(KERN_ERR "ins3510hw_output_on() set go failed(%d)\n", ret);
		return ret;
	}

	if( cam_debug > 4)
		ins3510hw_print_all_regs();
	CAM_dbg(4, "INS3510: output on.\n");

	return 0;
}

static int ins3510hw_output_off(void)
{
	CAM_dbg(4, "INS3510: output off.\n");

	if( cam_debug > 4)
		ins3510hw_print_all_regs();
	return 0;
}


/*****************************************************************************
 *  Format
 *****************************************************************************/
static int ins3510hw_set_format( u32 width, u32 height, u32 colorFormat, u32 mode)
{
	int ret;
	u8 reg_fmt0, reg_fmt1;

	switch(colorFormat)
	{
	case INS3510_FORMAT_YUV_422:
	{
		reg_fmt0 = 0x30;
		reg_fmt1 = 0x01;	/* normal spoof mode */
	}
		break;
	case INS3510_FORMAT_JPG:
	{
		reg_fmt0 = 0x02;	/* EXIT JPEG */
		reg_fmt1 = 0x00;	/* continuous mode */
	}
		break;
	case INS3510_FORMAT_RGB565:
	{
		reg_fmt0 = 0x40;
		reg_fmt1 = 0x01;	/* normal spoof mode */
	}
		break;
	default:
		return -EPERM;
		break;
	}

	if(mode == CAMERA_MODE_STILL)
	{
		/* context A */
		ret = ins3510_i2c_write_reg8(0x04, reg_fmt0);
		if( ret<0 )
			return ret;
		ret = ins3510_i2c_write_reg8(0x05, reg_fmt1);
		if( ret<0 )
			return ret;
		ret = ins3510_i2c_write_reg16(0x00, width);
		if( ret<0 )
			return ret;
		ret = ins3510_i2c_write_reg16(0x02, height);
		if( ret<0 )
			return ret;
	}
	else
	{
		/* context B */
		ret = ins3510_i2c_write_reg8(0x44, reg_fmt0);
		if( ret<0 )
			return ret;
		ret = ins3510_i2c_write_reg8(0x45, reg_fmt1);
		if( ret<0 )
			return ret;
		ret = ins3510_i2c_write_reg16(0x40, width);
		if( ret<0 )
			return ret;
		ret = ins3510_i2c_write_reg16(0x42, height);
		if( ret<0 )
			return ret;
	}

	CAM_dbg(7, "ins3510: previous mode=%d, new mode=%d\n", ins3510hw_data.mode_change, mode);
	if(ins3510hw_data.mode_change!=mode)
	{
		if(mode == CAMERA_MODE_STILL) {
			if(ins3510hw_data.flashled_flag==1)
				camerahw_flash_fire();
			ret = ins3510hw_sensor_preview2snap();
		}
		else
			ret = ins3510hw_sensor_snap2preview();
		if( ret<0 )
			return ret;
		ins3510hw_data.mode_change = mode;
	}

	return 0;
}

/*****************************************************************************
 * INS3510 Functions
 *****************************************************************************/
static int ins3510_init( p_camera_context_t camera_context )
{
	u32 revision;
	int status;

	CAM_dbg(2, "ins3510_init\n");

	/* provide informat about the capabilities of the sensor */
	camera_context->sensor_status.caps |= SENSOR_CAP_MANUAL_CONTRAST |
		SENSOR_CAP_MANUAL_WHITEBALANCE |
		SENSOR_CAP_MANUAL_EXPOSURE;

	/* Configure CI according to INS3510's hardware
	 * master parallel with 8 data pins
	 */
	ci_set_mode(CI_MODE_MP, CI_DATA_WIDTH8);

	/* enable pixel clock(sensor will provide pclock)
	 * and master clock = 26MHZ
	 */
	ci_set_clock(1, 1, sensor_mclk);
	ins3510hw_data.clock_on = 1;

	/* data sample on rising and h,vsync active high */
	ci_set_polarity(0, 0, 0);

	/* fifo control */
	ci_set_fifo(0, CI_FIFO_THL_32, 1, 1);

	/* set black level */
	ci_cgu_set_black_level(0);

	/* CGU Mux Select */
	ci_cgu_set_addr_mux_select(CI_CGU_MUX_0_TO_7);

	/* power on */
	status = ins3510hw_power_down(CAMERA_POWER_FULL);
	if ( status<0 )
		return status;

	revision = 0;
	status = ins3510hw_version_revision(&revision);
	if ( status )
	{
		printk(KERN_ERR "ins3510: read revision failed(%d).\n", status);
		return status;
	}
	printk(KERN_INFO "ins3510: revision(%08x).\n", revision);

	/* initialize default camera parameters */
	/* default preview width/height */
	camera_context->mode[CAMERA_MODE_VIDEO].width = 432;
	camera_context->mode[CAMERA_MODE_VIDEO].height = 320;
	/* default still capture width/height */
	camera_context->mode[CAMERA_MODE_STILL].width = 2048;
	camera_context->mode[CAMERA_MODE_STILL].height = 1536;
	camera_context->mode[CAMERA_MODE_STILL].input_format = V4L2_PIX_FMT_JPEG;
	camera_context->mode[CAMERA_MODE_STILL].output_format = V4L2_PIX_FMT_JPEG;

	ins3510hw_data_reset();

	ins3510hw_set_default();

	/* turn sensor output off */
	ins3510hw_output_off();

	return 0;
}

static int ins3510_deinit(  p_camera_context_t camera_context )
{
	/* power off the external module */

	CAM_dbg(2, "ins3510_deinit() \n");
	ins3510hw_power_down( CAMERA_POWER_OFF );

	ci_set_clock(0, 0, sensor_mclk);
	ins3510hw_data.clock_on = 0;

	return 0;
}

static int ins3510_sleep(  p_camera_context_t camera_context )
{
	return ins3510_deinit( camera_context );
}

static int ins3510_wake(  p_camera_context_t camera_context )
{
	return ins3510_init( camera_context );
}

static int ins3510_set_capture_format(  p_camera_context_t camera_context )
{
	CI_MP_TIMING timing;
	int status;
	u32 insFormat;

	switch(camera_context->capture_input_format) 
	{
		case V4L2_PIX_FMT_YUV422P:
			insFormat = INS3510_FORMAT_YUV_422;
			break;
			
		case V4L2_PIX_FMT_JPEG:
			insFormat = INS3510_FORMAT_JPG;
			break;
		
		case V4L2_PIX_FMT_RGB565X:
			insFormat = INS3510_FORMAT_RGB565;
			break;

		default:
			insFormat = INS3510_FORMAT_YUV_422;
			break;
	}

	CAM_dbg(4, "ins3510: capture_input_width=%d\n", camera_context->capture_input_width);
	CAM_dbg(4, "ins3510: capture_input_height=%d\n", camera_context->capture_input_height);
	
	CAM_dbg(4, "ins3510: capture_input_format=%d\n", camera_context->capture_input_format);
		
	status = ins3510hw_set_format( camera_context->capture_input_width,
					camera_context->capture_input_height,
					insFormat, camera_context->capture_mode);
	if( status<0 )
	{
		printk(KERN_ERR "ins3510: hw_set_format: failed(%d).\n", status);
		return status;
	}

	timing.BFW = 0x00;
	timing.BLW = 0x00;
	
	ci_configure_mp(camera_context->capture_input_width-1,
			camera_context->capture_input_height-1, &timing);

	CAM_dbg(4, "ins3510: ci_configure_mp witd =%x heigh =%x\n",camera_context->capture_input_width,camera_context->capture_input_height);
	return 0;
}

static int ins3510_start_capture(p_camera_context_t camera_context,
		unsigned int frames)
{
#ifdef CONFIG_MACH_CHERRY
	if(camera_context->capture_mode == CAMERA_MODE_STILL)
		ins3510hw_set_sharpness( 0x17 );
	else
	{
		ins3510hw_set_sharpness( 0x35 );
	}
#endif
	return ins3510hw_output_on();
}

static int ins3510_stop_capture(  p_camera_context_t camera_context )
{
	if(ins3510hw_data.flashled_fire && camera_context->capture_mode == CAMERA_MODE_STILL) {
		camerahw_flash_work_off();
	}
	return ins3510hw_output_off();
}

static int ins3510_set_power_mode(p_camera_context_t camera_context, u8 mode)
{
	(void) camera_context;
	ins3510hw_power_down( mode );
	return 0;
}

/*****************************************************************************
 * Register read/write
 *****************************************************************************/

static int ins3510_read_8bit(p_camera_context_t camera_context,
		u8 reg_addr,  u8 *reg_val)
{
	(void) camera_context;
	return ins3510_i2c_read_reg8(reg_addr, reg_val);
}

static int ins3510_write_8bit( p_camera_context_t camera_context,
		u8 reg_addr,  u8 reg_val)
{
	(void) camera_context;
	return ins3510_i2c_write_reg8(reg_addr, reg_val);
}

static int ins3510_read_16bit(p_camera_context_t camera_context,
		u16 reg_addr,  u16 *reg_val)
{
	(void) camera_context;
	return ins3510_i2c_read_reg16(reg_addr, reg_val);
}

static int ins3510_write_16bit( p_camera_context_t camera_context,
		u16 reg_addr,  u16 reg_val)
{
	(void) camera_context;
	return ins3510_i2c_write_reg16(reg_addr, reg_val);
}

/*****************************************************************************
 * Contrast
 *****************************************************************************/
static int ins3510_try_contrast( p_camera_context_t camera_context, u32 value)
{
	(void) camera_context;

	if( (value == SENSOR_CONTRAST_LOWEST) ||
	    (value == SENSOR_CONTRAST_LOW)   ||
	    (value == SENSOR_CONTRAST_MIDDLE) ||
	    (value == SENSOR_CONTRAST_HIGH) ||
	    (value == SENSOR_CONTRAST_HIGHEST) )
		return 0;
	else
		return -ENOSYS;
}

static int ins3510_set_contrast( p_camera_context_t camera_context, u8 mode, u32 value)
{
	int ret, regval;
	(void) camera_context;

	switch(value) {
		case SENSOR_CONTRAST_LOWEST:
			regval = 0xe000;	/* -2, s3.12 */
			break;
		case SENSOR_CONTRAST_LOW:
			regval = 0xf000;	/* -1, s3.12 */
			break;
		case SENSOR_CONTRAST_MIDDLE:
			regval = 0;		/* 0, s3.12 */
			break;
		case SENSOR_CONTRAST_HIGH:
			regval = 0x1000;	/* 1, s3.12 */
			break;
		case SENSOR_CONTRAST_HIGHEST:
			regval = 0x2000;	/* 2, s3.12 */
			break;
		default:
			printk(KERN_ERR "ins3510: set contrast invalid %d\n", value);
			return -EINVAL;
	}

	/* set hw */
	ins3510hw_set_contrast( 0, regval );
	ins3510hw_set_contrast( 1, regval );

	ret = ins3510hw_set_go_bit();
	if( ret<0 )
	{
		printk(KERN_ERR "ins3510_set_contrast() set go failed(%d)\n", ret);
		return ret;
	}

	return 0;
}

/*****************************************************************************
 * Exposure
 *****************************************************************************/

static int ins3510_set_ev_offset( p_camera_context_t camera_context, int value)
{
	int ret, regval;

	if (value==0)			/* 0 EV */
		regval = 0;
	else if (value>0) {
		if (value<=33)		/* 0.3 EV */
			regval = 0x0155;
		else if (value<=70)	/* 0.7 EV */
			regval = 0x02aa;
		else if (value<=100)	/* 1.0 EV */
			regval = 0x0400;
		else if (value<=133)	/* 1.3 EV */
			regval = 0x0555;
		else if (value<=170)	/* 1.7 EV */
			regval = 0x06aa;
		else 			/* 2.0 EV */
			regval = 0x0800;
	}
	else {
		if (value>=-33)		/* -0.3 EV */
			regval = 0xfeab;
		else if (value>=-70)	/* -0.7 EV */
			regval = 0xfd56;
		else if (value>=-100)	/* -1.0 EV */
			regval = 0xfc00;
		else if (value>=-133)	/* -1.3 EV */
			regval = 0xfaab;
		else if (value>=-170)	/* -1.7 EV */
			regval = 0xf956;
		else			/* -2.0 EV */
			regval = 0xf800;
	}

	/* set hw */
	ins3510hw_set_ae_offset( regval );

	ret = ins3510hw_set_go_bit();
	if( ret<0 )
	{
		printk(KERN_ERR "ins3510_set_ev_offset() set go failed(%d)\n", ret);
		return ret;
	}

	return 0;
}

/*****************************************************************************
 * Auto White Balance
 *****************************************************************************/
static int ins3510_try_white_balance( p_camera_context_t camera_context, u32 value)
{
	(void) camera_context;

	if( (value == SENSOR_WHITEBALANCE_AUTO) ||
	    (value == SENSOR_WHITEBALANCE_DAYLIGHT) ||
	    (value == SENSOR_WHITEBALANCE_INCANDESCENT) ||
	    (value == SENSOR_WHITEBALANCE_FLUORESCENT) ||
	    (value == SENSOR_WHITEBALANCE_CLOUDY) )
		return 0;
	else
		return -ENOSYS;
}

static int ins3510_set_white_balance( p_camera_context_t camera_context,
		u8 mode, u32 value)
{
	int ret, wb_mode;
	(void) camera_context;

	switch(value) {
		case SENSOR_WHITEBALANCE_AUTO:		/* Auto */
			wb_mode = 0;
			break;
		case SENSOR_WHITEBALANCE_INCANDESCENT:	/* Incandescent */
			wb_mode = 6;			/* Tungsten */
			break;
		case SENSOR_WHITEBALANCE_DAYLIGHT:	/* Sunny */
			wb_mode = 1;			/* D65 */
			break;
		case SENSOR_WHITEBALANCE_FLUORESCENT:	/* Fluorescent */
			wb_mode = 4;			/* cool fluorescent */
			break;
		case SENSOR_WHITEBALANCE_CLOUDY:	/* Cloudy */
			wb_mode = 2;			/* D75 */
			break;
		default:
			printk(KERN_ERR "ins3510: set white balance invalid %d\n", value);
			return -EINVAL;
	}

	ins3510hw_set_wb_mode( wb_mode );

	ret = ins3510hw_set_go_bit();
	if( ret<0 )
	{
		printk(KERN_ERR "ins3510_set_white_balance() set go failed(%d)\n", ret);
		return ret;
	}

	return 0;
}

static int ins3510_set_framerate(u32 numerator, u32 denominator)
{
	int fps = 0;
	if (0<denominator && 0<=numerator)
	{
		fps = numerator / denominator;
		if( fps>30 )
			fps = 30;
	}
	return ins3510hw_set_framerate(fps);
}

static int ins3510_set_autofocus( p_camera_context_t camera_context, u32 value)
{
	int ret = 0;
	(void) camera_context;

	if (value==SENSOR_AUTOFOCUS_IDLE) {
		ret = ins3510hw_af_idle();
	}
	else if (value==SENSOR_AUTOFOCUS_SINGLE) {
		ret = ins3510hw_single_af();
	}
	else
		return -EINVAL;

	return ret;
}

/**saturation**/
static int ins3510_try_saturation( p_camera_context_t camera_context, u32 value)
{
	(void) camera_context;

	if( (value == SENSOR_SATURATION_LOWEST) ||
	    (value == SENSOR_SATURATION_LOW)   ||
	    (value == SENSOR_SATURATION_STANDARD) ||
	    (value == SENSOR_SATURATION_HIGH) ||
	    (value == SENSOR_SATURATION_HIGHEST) )
		return 0;
	else
		return -ENOSYS;
}

static int ins3510_set_saturation( p_camera_context_t camera_context, u32 value)
{
	int ret, regval;

	switch(value) {
		case SENSOR_SATURATION_LOWEST:
			regval = 0x0400;	/* 0.25, s3.12 format */
			break;
		case SENSOR_SATURATION_LOW:
			regval = 0x0800;	/* 0.5, s3.12 format */
			break;
		case SENSOR_SATURATION_STANDARD:
			regval = 0x1000;	/* 1, s3.12 format */
			break;
		case SENSOR_SATURATION_HIGH:
			regval = 0x1800;	/* 1.8, s3.12 format */
			break;
		case SENSOR_SATURATION_HIGHEST:
			regval = 0x2000;	/* 2, s3.12 format */
			break;
		default:
			printk(KERN_ERR "ins3510: set saturation invalid %d\n", value);
			return -EINVAL;
	}
	/* set hw */
	ins3510hw_set_saturation( 0, regval );
	ins3510hw_set_saturation( 1, regval );

	ret = ins3510hw_set_go_bit();
	if( ret<0 )
	{
		printk(KERN_ERR "ins3510_set_saturation() set go failed(%d)\n", ret);
		return ret;
	}

	return 0;
}

/*effect*/
static int ins3510_try_effect( p_camera_context_t camera_context, u32 value)
{
	(void) camera_context;

	if( (value == SENSOR_EFFECT_NONE) ||
	    (value == SENSOR_EFFECT_MONO) ||
	    (value == SENSOR_EFFECT_NEGATIVE) ||
	    (value == SENSOR_EFFECT_SEPIA) ||
	    (value == SENSOR_EFFECT_POSTERIZE) )
		return 0;
	else
		return -ENOSYS;
}

static int ins3510_set_effect( p_camera_context_t camera_context, u32 value)
{
	int ret, sp_mode;

	switch(value) {
		case SENSOR_EFFECT_NONE:
			sp_mode = 0x0;			/* normal */
			break;
		case SENSOR_EFFECT_MONO:
			sp_mode = 0x02;			/* Grayscale */
			break;
		case SENSOR_EFFECT_NEGATIVE:
			sp_mode = 0x01;			/* Negative */
			break;
		case SENSOR_EFFECT_SEPIA:
			sp_mode = 0x04;			/* Sepia */
			break;
		case SENSOR_EFFECT_POSTERIZE:
			sp_mode = 0x07;			/* Posterize 8 */
			break;
		default:
			printk(KERN_ERR "ins3510: set special mode invalid %d\n", value);
			return -EINVAL;
	}

	ins3510hw_set_special_mode( sp_mode );

	ret = ins3510hw_set_go_bit();
	if( ret<0 )
	{
		printk(KERN_ERR "ins3510_set_specialmode() set go failed(%d)\n", ret);
		return ret;
	}

	return 0;
}

/*sharpness*/
static int ins3510_try_sharpness( p_camera_context_t camera_context, u32 value)
{
	(void) camera_context;

	if( (value == SENSOR_SHARPNESS_LOWEST) ||
	    (value == SENSOR_SHARPNESS_LOW)   ||
	    (value == SENSOR_SHARPNESS_STANDARD) ||
	    (value == SENSOR_SHARPNESS_HIGH) ||
	    (value == SENSOR_SHARPNESS_HIGHEST) )
		return 0;
	else
		return -ENOSYS;
}

static int ins3510_set_sharpness( p_camera_context_t camera_context, u32 value)
{
	int ret, regval;

	switch(value) {
		case SENSOR_SHARPNESS_LOWEST:
			regval = 0x0400;	/* 0.25, s3.12 format */
			break;
		case SENSOR_SHARPNESS_LOW:
			regval = 0x0800;	/* 0.5, s3.12 format */
			break;
		case SENSOR_SHARPNESS_STANDARD:
			regval = 0x1000;	/* 1, s3.12 format */
			break;
		case SENSOR_SHARPNESS_HIGH:
			regval = 0x1800;	/* 1.8, s3.12 format */
			break;
		case SENSOR_SHARPNESS_HIGHEST:
			regval = 0x2000;	/* 2, s3.12 format */
			break;
		default:
			printk(KERN_ERR "ins3510: set sharpness invalid %d\n", value);
			return -EINVAL;
	}
	/* set hw */
	ins3510hw_set_sharpness( regval );

	ret = ins3510hw_set_go_bit();
	if( ret<0 )
	{
		printk(KERN_ERR "ins3510_set_sharpness() set go failed(%d)\n", ret);
		return ret;
	}

	return 0;
}

/**sensor jpeg quality**/
static int ins3510_set_jpeg_quality( p_camera_context_t camera_context, u32 value)
{
	int ret, regval;

	regval = (value*0x0f) / 100;
	if(regval<0)
		regval = 0;	/* lowest */
	if(regval>0x0f)
		regval = 0x0f;	/* best value */

	/* set hw */
	ins3510hw_set_jpeg_quality( regval );

	ret = ins3510hw_set_go_bit();
	if( ret<0 )
	{
		printk(KERN_ERR "ins3510_set_jpeg_quality() set go failed(%d)\n", ret);
		return ret;
	}

	return 0;
}

static int ins3510_set_digital_zoom( p_camera_context_t camera_context, u32 value)
{
	int ret, regval;
	if ((value<100) || (value>800))
		return -EINVAL;

	/* set hw */
	regval = value * 0x1000 / 100;
	CAM_dbg(4, "ins3510: set zoom %d, regval 0x%04x\n", value, regval);

	ins3510hw_set_digital_zoom( regval );

	ret = ins3510hw_set_go_bit();
	if( ret<0 )
	{
		printk(KERN_ERR "ins3510_set_digital_zoom() set go failed(%d)\n", ret);
		return ret;
	}
	return 0;
}

static int ins3510_enum_frame_sizes(p_camera_context_t camera_context, int mode, u32 format,
				   struct v4l2_frmsizeenum *frm_sizes)
{
	int index = frm_sizes->index;

	frm_sizes->type = V4L2_FRMSIZE_TYPE_DISCRETE;

	if (mode==CAMERA_MODE_VIDEO) {
		if (V4L2_PIX_FMT_YUV422P!=format) {
			printk(KERN_ERR "ins3510: only support YUV422 on video mode\n");
			return -EINVAL;
		}
		if (frm_sizes->index < ARRAY_SIZE(ins3510_yuv422_frame_sizes)) {
			frm_sizes->discrete.width = ins3510_yuv422_frame_sizes[index].width;
			frm_sizes->discrete.height = ins3510_yuv422_frame_sizes[index].height;
		}
		else
			return -EINVAL;
	}
	else {
		if (V4L2_PIX_FMT_JPEG!=format) {
			printk(KERN_ERR "ins3510: only support JPEG on still capture mode\n");
			return -EINVAL;
		}
		if (frm_sizes->index < ARRAY_SIZE(ins3510_jpeg_frame_sizes)) {
			frm_sizes->discrete.width = ins3510_jpeg_frame_sizes[index].width;
			frm_sizes->discrete.height = ins3510_jpeg_frame_sizes[index].height;
		}
		else
			return -EINVAL;
	}
	return 0;
}

static int ins3510_enum_frame_intervals(p_camera_context_t camera_context, u32 format,
					struct v4l2_frmivalenum *frm_ivals)
{
	int index = frm_ivals->index;

	frm_ivals->type = V4L2_FRMIVAL_TYPE_DISCRETE;

	if (V4L2_PIX_FMT_YUV422P!=format) {
		printk(KERN_ERR "ins3510: only YUV422 have frame interval\n");
		return -EINVAL;
	}
	if (frm_ivals->index < ARRAY_SIZE(ins3510_yuv422_frame_intervals)) {
		frm_ivals->discrete.numerator = ins3510_yuv422_frame_intervals[index].numerator;
		frm_ivals->discrete.denominator = ins3510_yuv422_frame_intervals[index].denominator;
	}
	else
		return -EINVAL;

	return 0;
}

static p_camera_function_t camera_functions;

static int ins3510_format_list[] = {

	V4L2_PIX_FMT_YUV422P,
	V4L2_PIX_FMT_RGB565X,
	V4L2_PIX_FMT_JPEG,
	-1	
};

static int ins3510_register(int id)
{

	/* allocte camera functions context */
	camera_functions = kzalloc(sizeof(camera_function_t), GFP_KERNEL);
	if (!camera_functions) {
		printk(KERN_ERR "ins3510: Can't allocate buffer for camera functions structure \n");
		return -ENOMEM;
	}

	camera_functions->width_max = INS3510_MAX_WIDTH;
	camera_functions->height_max = INS3510_MAX_HEIGHT;
	camera_functions->width_min = 16;
	camera_functions->height_min = 16;

	camera_functions->v_power = -1;		/* VAA 2.8V */
	camera_functions->v_io = -1;		/* VDD & VIO is controlled by GPIO */

	camera_functions->format_list = ins3510_format_list;

	camera_functions->init = ins3510_init;     	
	camera_functions->deinit = ins3510_deinit;     
	camera_functions->set_capture_format = ins3510_set_capture_format; 	
	camera_functions->start_capture =	ins3510_start_capture;	 		
	camera_functions->stop_capture =	ins3510_stop_capture; 			
	camera_functions->sleep = ins3510_sleep;	
	camera_functions->wakeup = ins3510_wake;	
	camera_functions->read_8bit = ins3510_read_8bit;
	camera_functions->write_8bit = ins3510_write_8bit;
	camera_functions->read_16bit = ins3510_read_16bit;
	camera_functions->write_16bit = ins3510_write_16bit;
	camera_functions->read_32bit = NULL;	
	camera_functions->write_32bit = NULL;	
	camera_functions->set_power_mode =	ins3510_set_power_mode;
	camera_functions->set_contrast =	ins3510_set_contrast;
	camera_functions->try_contrast =	ins3510_try_contrast;
	camera_functions->set_whitebalance =	ins3510_set_white_balance;
	camera_functions->try_whitebalance =	ins3510_try_white_balance;

	camera_functions->set_framerate = 	ins3510_set_framerate;

	camera_functions->set_autofocus = 	ins3510_set_autofocus;
	camera_functions->set_saturation = 	ins3510_set_saturation;
	camera_functions->try_saturation = 	ins3510_try_saturation;
	camera_functions->set_effect = 		ins3510_set_effect;
	camera_functions->try_effect = 		ins3510_try_effect;
	camera_functions->set_sharpness = 	ins3510_set_sharpness;
	camera_functions->try_sharpness = 	ins3510_try_sharpness;
	camera_functions->set_jpeg_quality_percent=ins3510_set_jpeg_quality;
	camera_functions->set_ev_offset = 	ins3510_set_ev_offset;
	camera_functions->set_digital_zoom = 	ins3510_set_digital_zoom;

	camera_functions->enum_frame_sizes	= ins3510_enum_frame_sizes;
	camera_functions->enum_frame_intervals	= ins3510_enum_frame_intervals;

	camera_functions->name = "INS3510";

	camera_functions->id = SENSOR_HIGH;

	/*register 3640 as high resolution */

	if(sensor_register(camera_functions, id) < 0){
		printk(KERN_ERR "ins3510: sensor_register failed !\n");
		kfree(camera_functions);
		camera_functions = NULL;
		return -1;
	}
	return 0;
}

static int ins3510_unregister(int id)
{
	sensor_unregister(id);
	
	if(!camera_functions){
		kfree(camera_functions);
		camera_functions = NULL;
	}
	return 0;
}

/*****************************************************************************
 *  sysfs
 *****************************************************************************/

static ssize_t ins3510_i2clog_show(struct device *dev, struct device_attribute *attr, char *buf)
{
	return sprintf(buf, "%d\n", i2c_debug);
}

static ssize_t ins3510_i2clog_store(struct device *dev, struct device_attribute *attr, const char *buf, size_t n)
{
	int val;
	sscanf(buf, "%d", &val);
	if( val<0 || val>255)
		printk(KERN_ERR "invalid input\n");
	else
		i2c_debug = val;
	return n;
}

static DEVICE_ATTR(i2clog, S_IRUGO|S_IWUSR, ins3510_i2clog_show, ins3510_i2clog_store);

static ssize_t ins3510_camlog_show(struct device *dev, struct device_attribute *attr, char *buf)
{
	return sprintf(buf, "%d\n", cam_debug);
}

static ssize_t ins3510_camlog_store(struct device *dev, struct device_attribute *attr, const char *buf, size_t n)
{
	int val;
	sscanf(buf, "%d", &val);
	if( val<0 || val>255)
		printk(KERN_ERR "invalid input\n");
	else
		cam_debug = val;
	return n;
}

static DEVICE_ATTR(camlog, S_IRUGO|S_IWUSR, ins3510_camlog_show, ins3510_camlog_store);

static ssize_t ins3510_mclock_show(struct device *dev, struct device_attribute *attr, char *buf)
{
	return sprintf(buf, "%d khz\n", sensor_mclk * 10);
}

static ssize_t ins3510_mclock_store(struct device *dev, struct device_attribute *attr, const char *buf, size_t n)
{
	int val;
	sscanf(buf, "%d", &val);
	if( val<0 || val>52000)
		printk(KERN_ERR "invalid input\n");
	else
		sensor_mclk = val / 10;
	return n;
}

static DEVICE_ATTR(mclock, S_IRUGO|S_IWUSR, ins3510_mclock_show, ins3510_mclock_store);

static ssize_t ins3510_clockon_show(struct device *dev, struct device_attribute *attr, char *buf)
{
	return sprintf(buf, "%d\n", ins3510hw_data.clock_on);
}

static ssize_t ins3510_clockon_store(struct device *dev, struct device_attribute *attr, const char *buf, size_t n)
{
	int val;
	sscanf(buf, "%d", &val);
	if( val<0 || val>1)
		printk(KERN_ERR "invalid input\n");
	else {
		if( val )
			sensor_clock_on();
		else
			sensor_clock_off();
	}
	return n;
}

static DEVICE_ATTR(clock_on, S_IRUGO|S_IWUSR, ins3510_clockon_show, ins3510_clockon_store);

static ssize_t ins3510_voltageon_show(struct device *dev, struct device_attribute *attr, char *buf)
{
	return sprintf(buf, "%d\n", ins3510hw_data.voltage_on);
}

static ssize_t ins3510_voltageon_store(struct device *dev, struct device_attribute *attr, const char *buf, size_t n)
{
	int val;
	sscanf(buf, "%d", &val);
	if( val<0 || val>1)
		printk(KERN_ERR "invalid input\n");
	else {
		if( val )
			sensor_voltage_on();
		else
			sensor_voltage_off();
	}
	return n;
}

static DEVICE_ATTR(voltage_on, S_IRUGO|S_IWUSR, ins3510_voltageon_show, ins3510_voltageon_store);

static ssize_t ins3510_poweron_show(struct device *dev, struct device_attribute *attr, char *buf)
{
	return sprintf(buf, "%d\n", ins3510hw_data.power_on);
}

static ssize_t ins3510_poweron_store(struct device *dev, struct device_attribute *attr, const char *buf, size_t n)
{
	int val;
	sscanf(buf, "%d", &val);
	if( val<0 || val>1)
		printk(KERN_ERR "invalid input\n");
	else {
		if( val )
			sensor_power_on();
		else
			sensor_power_off();
	}
	return n;
}

static DEVICE_ATTR(power_on, S_IRUGO|S_IWUSR, ins3510_poweron_show, ins3510_poweron_store);

static ssize_t ins3510_standby_show(struct device *dev, struct device_attribute *attr, char *buf)
{
	return sprintf(buf, "%d\n", ins3510hw_data.standby);
}

static ssize_t ins3510_standby_store(struct device *dev, struct device_attribute *attr, const char *buf, size_t n)
{
	int val;
	sscanf(buf, "%d", &val);
	if( val<0 || val>1)
		printk(KERN_ERR "invalid input\n");
	else {
		if( val )
			sensor_wakeup();
		else
			sensor_standby();
	}
	return n;
}

static DEVICE_ATTR(standby, S_IRUGO|S_IWUSR, ins3510_standby_show, ins3510_standby_store);

static ssize_t ins3510_bootup_store(struct device *dev, struct device_attribute *attr, const char *buf, size_t n)
{
	int val;
	sscanf(buf, "%d", &val);
	if( val==1 )
	{
		ins3510hw_bootup();
	}
	else
		printk(KERN_ERR "invalid input\n");
	return n;
}

static DEVICE_ATTR(bootup, S_IRUGO|S_IWUSR, NULL, ins3510_bootup_store);

static ssize_t ins3510_reset_store(struct device *dev, struct device_attribute *attr, const char *buf, size_t n)
{
	int val;
	sscanf(buf, "%d", &val);
	if( val!=1 )
		printk(KERN_ERR "invalid input\n");
	else {
		sensor_reset();
	}
	return n;
}

static DEVICE_ATTR(reset, S_IRUGO|S_IWUSR, NULL, ins3510_reset_store);

static ssize_t ins3510_dumpregs_show(struct device *dev, struct device_attribute *attr, char *buf)
{
	if( ins3510hw_data.power_on )
		return ins3510hw_dump_all_regs(buf, PAGE_SIZE);
	else
		return sprintf(buf, "not power on");
}

static DEVICE_ATTR(dumpregs, S_IRUGO|S_IWUSR, ins3510_dumpregs_show, NULL);

static ssize_t ins3510_i2ctest_store(struct device *dev, struct device_attribute *attr, const char *buf, size_t n)
{
	int val, i, ret = 0;
	u32 reg, ver;

	sscanf(buf, "%d", &val);
	if( val<0 )
		printk(KERN_ERR "invalid input\n");
	else if( ins3510hw_data.power_on )
	{
		ins3510_i2c_read_reg32(0x80, &ver);
		printk(KERN_ERR "ins3510: test register 0x80 chip id %08x\n", ver);
		for(i=0; i<val; i++)
		{
			reg = 0;
			ret = ins3510_i2c_read_reg32(0x80, &reg);
			if(ret<0)
			{
				printk(KERN_ERR "ins3510: reg test failed, count %d, return %d\n", i, ret);
				continue;
			}
			if( reg!=ver )
			{
				printk(KERN_ERR "ins3510: reg test failed, count %d, invalid chip id %08x\n", i, reg);
			}
		}
	}
	return n;
}

static DEVICE_ATTR(i2ctest, S_IRUGO|S_IWUSR, NULL, ins3510_i2ctest_store);

static ssize_t ins3510_i2crwtest_store(struct device *dev, struct device_attribute *attr, const char *buf, size_t n)
{
	int val, i, ret = 0;
	u8 reg8;
	u32 reg32, reg32s;
	int loop;

	sscanf(buf, "%d", &val);
	if( val<0 )
		printk(KERN_ERR "invalid input\n");
	else
	{
		for(i=0; i<val; i++)
		{
			sensor_clock_on();
			sensor_power_on();

			ret = ins3510_i2c_write_reg16(0xe0, 0x820c);
			if( ret<0 )
				return ret;
			loop = 500;
			while(loop-->0)
			{
				ret = ins3510_i2c_read_reg8(0xf8, &reg8);
				if( (reg8 & 0x20)!=0 )
					break;
				msleep(10);
			}
			if(loop<=0)
			{
				printk(KERN_ERR "ins3510: power on timeout to wait 0x820c8\n");
				ins3510hw_print_all_regs();
			}

			/* clock setting */
			ret = ins3510_i2c_write_reg16(0xe0, 0x4009);
			ret = ins3510_i2c_write_reg32(0xf0, 0x0786f792);	/* 0x40090 */
			ret = ins3510_i2c_write_reg32(0xf4, 0x00000007);	/* 0x40094 */
			ret = ins3510_i2c_write_reg16(0xe0, 0x400f);
			ret = ins3510_i2c_write_reg32(0xf0, 0x11000601);	/* 0x400f0 */
			ret = ins3510_i2c_read_reg32(0xf4, &reg32);
			reg32 = (reg32 & 0x000000ff) | 0x00070100;
			ret = ins3510_i2c_write_reg32(0xf4, reg32);		/* 0x400f4 */

			reg32s = reg32;
			ins3510_i2c_write_reg16(0xe0, 0x4009);
			ins3510_i2c_read_reg32(0xf0, &reg32);
			if( reg32!=0x0786f792 )
				CAM_dbg(4, "INS3510: readback 0x40090 did not match 0x0786f792:0x%08x.\n", reg32);
			ins3510_i2c_read_reg32(0xf4, &reg32);
			if( reg32!=0x00000007 )
				CAM_dbg(4, "INS3510: readback 0x40094 did not match 0x00000007:0x%08x.\n", reg32);
			ins3510_i2c_write_reg16(0xe0, 0x400f);
			ins3510_i2c_read_reg32(0xf0, &reg32);
			if( reg32!=0x11000601 )
				CAM_dbg(4, "INS3510: readback 0x400f0 did not match 0x11000601:0x%08x.\n", reg32);
			ins3510_i2c_read_reg32(0xf4, &reg32);
			if( reg32!=reg32s )
				CAM_dbg(4, "INS3510: readback 0x400f0 did not match 0x%08x:0x%08x.\n", reg32s, reg32);

			sensor_power_off();
			sensor_clock_off();
		}
	}
	return n;
}

static DEVICE_ATTR(i2crwtest, S_IRUGO|S_IWUSR, NULL, ins3510_i2crwtest_store);

static ssize_t ins3510_boottest_store(struct device *dev, struct device_attribute *attr, const char *buf, size_t n)
{
	int val, i, ret = 0;

	sscanf(buf, "%d", &val);
	if( val<0 )
		printk(KERN_ERR "invalid input\n");
	else
	{
		for(i=0; i<val; i++)
		{
			sensor_clock_on();
			sensor_power_on();
			if( (ret=ins3510hw_bootup())<0 )
			{
				printk(KERN_INFO "ins3510: bootup #%d failed %d\n", i, ret);
				break;
			}
			sensor_power_off();
			sensor_clock_off();
		}
	}
	return n;
}

static DEVICE_ATTR(boottest, S_IRUGO|S_IWUSR, NULL, ins3510_boottest_store);

static ssize_t ins3510_setgotest_store(struct device *dev, struct device_attribute *attr, const char *buf, size_t n)
{
	int val, i, ret = 0;
	u8 reg8;

	sscanf(buf, "%d", &val);
	if( val<0 )
		printk(KERN_ERR "invalid input\n");
	else
	{
		sensor_clock_on();
		sensor_power_on();
		if( (ret=ins3510hw_bootup())<0 )
			printk(KERN_INFO "ins3510: bootup failed %d\n", ret);
		for(i=0; i<val; i++)
		{
			/* INS3510_set_context(); */
			ret = ins3510_i2c_read_reg8(0x84, &reg8);
			if( ret<0 )
				return ret;
			reg8 = (reg8&0xfe)|(i&0x1);
			printk("ins3510: set context to %d\n", i&0x01);
			ret = ins3510_i2c_write_reg8(0x84, reg8);
			if( ret<0 )
				return ret;
			if( ins3510hw_set_go_bit()<0 )
			{
				printk(KERN_INFO "ins3510: set go #%d failed\n", i);
				break;
			}
		}
		sensor_power_off();
		sensor_clock_off();
	}
	return n;
}

static DEVICE_ATTR(setgotest, S_IRUGO|S_IWUSR, NULL, ins3510_setgotest_store);

static ssize_t ins3510_fwversion_show(struct device *dev, struct device_attribute *attr, char *buf)
{
	int ret;
	u32 ver;

	sensor_clock_on();
	sensor_power_on();
	if( (ret=ins3510hw_bootup())<0 )
		printk(KERN_INFO "ins3510: bootup failed %d\n", ret);
	ret = ins3510_i2c_read_reg32(0x80, &ver);
	sensor_power_off();
	sensor_clock_off();

	return sprintf(buf, "%08x\n", ver);
}

static DEVICE_ATTR(fwversion, S_IRUSR|S_IWUSR, ins3510_fwversion_show, NULL);

static int sysfs_fwupgrade = 0;

static ssize_t ins3510_fwupgrade_show(struct device *dev, struct device_attribute *attr, char *buf)
{
	return sprintf(buf, "%s\n", sysfs_fwupgrade ? "upgrading":"idle" );
}

static ssize_t ins3510_fwupgrade_store(struct device *dev, struct device_attribute *attr, const char *buf, size_t n)
{
	char *upgrade = "upgrade";
	if( n>=strlen(upgrade) && strncmp(upgrade, buf, strlen(upgrade))==0 )
	{
		printk(KERN_INFO "ins3510: start fireware upgrading... \n");
		sysfs_fwupgrade = 1;
		ins3510_firmware_upgrade();
		sysfs_fwupgrade = 0;
	}
	return n;
}

static DEVICE_ATTR(fwupgrade, S_IRUSR|S_IWUSR, ins3510_fwupgrade_show, ins3510_fwupgrade_store);

static ssize_t ins3510_testpattern_store(struct device *dev, struct device_attribute *attr, const char *buf, size_t n)
{
	int val;
	sscanf(buf, "%d", &val);
	if( val<0 || val>3)
        	printk(KERN_ERR "ins3510: invalid input %d\n", val);
	else if( ins3510hw_data.bootup )
	{
		ins3510hw_set_test_pattern(val);
		ins3510hw_set_go_bit();
	}
	else
        	printk(KERN_ERR "ins3510: not bootup\n");
	return n;
}

static DEVICE_ATTR(testpattern, S_IRUGO|S_IWUSR, NULL, ins3510_testpattern_store);

static ssize_t ins3510_fps_store(struct device *dev, struct device_attribute *attr, const char *buf, size_t n)
{
	int val;
	sscanf(buf, "%d", &val);
	if( val<0 || val>60)
        	printk(KERN_ERR "ins3510: invalid input %d\n", val);
	else if( ins3510hw_data.bootup )
	{
		ins3510hw_set_framerate(val);
		ins3510hw_set_go_bit();
	}
	else
        	printk(KERN_ERR "ins3510: not bootup\n");
	return n;
}

static DEVICE_ATTR(fps, S_IRUGO|S_IWUSR, NULL, ins3510_fps_store);

static ssize_t ins3510_af_pos_store(struct device *dev, struct device_attribute *attr, const char *buf, size_t n)
{
	int val;
	sscanf(buf, "%d", &val);
	if( val<0 || val>=0x100 )
        	printk(KERN_ERR "ins3510: invalid input 0x%x\n", val);
	else if( ins3510hw_data.bootup )
	{
		ins3510hw_set_af_pos(val);
		ins3510hw_set_go_bit();
	}
	else
        	printk(KERN_ERR "ins3510: not bootup\n");
	return n;
}

static DEVICE_ATTR(afpos, S_IRUGO|S_IWUSR, NULL, ins3510_af_pos_store);

static ssize_t ins3510_ae_offset_store(struct device *dev, struct device_attribute *attr, const char *buf, size_t n)
{
	int val;
	sscanf(buf, "%d", &val);
	if( val<0 || val>=0x10000 )
        	printk(KERN_ERR "ins3510: invalid input 0x%x\n", val);
	else if( ins3510hw_data.bootup )
	{
		ins3510hw_set_ae_offset(val);
		ins3510hw_set_go_bit();
	}
	else
        	printk(KERN_ERR "ins3510: not bootup\n");
	return n;
}

static DEVICE_ATTR(aeoff, S_IRUGO|S_IWUSR, NULL, ins3510_ae_offset_store);

static ssize_t ins3510_exif_orien_store(struct device *dev, struct device_attribute *attr, const char *buf, size_t n)
{
	int val;
	sscanf(buf, "%d", &val);
	if( val<0 || val>=0x100 )
        	printk(KERN_ERR "ins3510: invalid input 0x%x\n", val);
	else if( ins3510hw_data.bootup )
	{
		ins3510hw_set_exif_orien(val);
		ins3510hw_set_go_bit();
	}
	else
        	printk(KERN_ERR "ins3510: not bootup\n");
	return n;
}

static DEVICE_ATTR(exiforien, S_IRUGO|S_IWUSR, NULL, ins3510_exif_orien_store);

static ssize_t ins3510_sharpness_store(struct device *dev, struct device_attribute *attr, const char *buf, size_t n)
{
	int val;
	sscanf(buf, "%x", &val);
	if( val<0 || val>0xff)
        	printk(KERN_ERR "ins3510: invalid input 0x%x\n", val);
	else if( ins3510hw_data.bootup )
	{
		ins3510hw_set_sharpness(val);
		ins3510hw_set_go_bit();
	}
	else
        	printk(KERN_ERR "ins3510: not bootup\n");
	return n;
}

static DEVICE_ATTR(sharpness, S_IRUGO|S_IWUSR, NULL, ins3510_sharpness_store);

static ssize_t ins3510_a_gamma_store(struct device *dev, struct device_attribute *attr, const char *buf, size_t n)
{
	int val;
	sscanf(buf, "%x", &val);
	if( val<0 || val>0xffff)
        	printk(KERN_ERR "ins3510: invalid input 0x%x\n", val);
	else if( ins3510hw_data.bootup )
	{
		ins3510hw_set_gamma(0, val);
		ins3510hw_set_go_bit();
	}
	else
        	printk(KERN_ERR "ins3510: not bootup\n");
	return n;
}

static DEVICE_ATTR(a_gamma, S_IRUGO|S_IWUSR, NULL, ins3510_a_gamma_store);

static ssize_t ins3510_b_gamma_store(struct device *dev, struct device_attribute *attr, const char *buf, size_t n)
{
	int val;
	sscanf(buf, "%x", &val);
	if( val<0 || val>0xffff)
        	printk(KERN_ERR "ins3510: invalid input 0x%x\n", val);
	else if( ins3510hw_data.bootup )
	{
		ins3510hw_set_gamma(1, val);
		ins3510hw_set_go_bit();
	}
	else
        	printk(KERN_ERR "ins3510: not bootup\n");
	return n;
}

static DEVICE_ATTR(b_gamma, S_IRUGO|S_IWUSR, NULL, ins3510_b_gamma_store);

static ssize_t ins3510_a_brightness_store(struct device *dev, struct device_attribute *attr, const char *buf, size_t n)
{
	int val;
	sscanf(buf, "%x", &val);
	if( val<0 || val>0xffff)
        	printk(KERN_ERR "ins3510: invalid input 0x%x\n", val);
	else if( ins3510hw_data.bootup )
	{
		ins3510hw_set_brightness(0, val);
		ins3510hw_set_go_bit();
	}
	else
        	printk(KERN_ERR "ins3510: not bootup\n");
	return n;
}

static DEVICE_ATTR(a_brightness, S_IRUGO|S_IWUSR, NULL, ins3510_a_brightness_store);

static ssize_t ins3510_b_brightness_store(struct device *dev, struct device_attribute *attr, const char *buf, size_t n)
{
	int val;
	sscanf(buf, "%x", &val);
	if( val<0 || val>0xffff)
        	printk(KERN_ERR "ins3510: invalid input 0x%x\n", val);
	else if( ins3510hw_data.bootup )
	{
		ins3510hw_set_brightness(1, val);
		ins3510hw_set_go_bit();
	}
	else
        	printk(KERN_ERR "ins3510: not bootup\n");
	return n;
}

static DEVICE_ATTR(b_brightness, S_IRUGO|S_IWUSR, NULL, ins3510_b_brightness_store);

static ssize_t ins3510_a_contrast_store(struct device *dev, struct device_attribute *attr, const char *buf, size_t n)
{
	int val;
	sscanf(buf, "%x", &val);
	if( val<0 || val>0xffff)
        	printk(KERN_ERR "ins3510: invalid input 0x%x\n", val);
	else if( ins3510hw_data.bootup )
	{
		ins3510hw_set_contrast(0, val);
		ins3510hw_set_go_bit();
	}
	else
        	printk(KERN_ERR "ins3510: not bootup\n");
	return n;
}

static DEVICE_ATTR(a_contrast, S_IRUGO|S_IWUSR, NULL, ins3510_a_contrast_store);

static ssize_t ins3510_b_contrast_store(struct device *dev, struct device_attribute *attr, const char *buf, size_t n)
{
	int val;
	sscanf(buf, "%x", &val);
	if( val<0 || val>0xffff)
        	printk(KERN_ERR "ins3510: invalid input 0x%x\n", val);
	else if( ins3510hw_data.bootup )
	{
		ins3510hw_set_contrast(1, val);
		ins3510hw_set_go_bit();
	}
	else
        	printk(KERN_ERR "ins3510: not bootup\n");
	return n;
}

static DEVICE_ATTR(b_contrast, S_IRUGO|S_IWUSR, NULL, ins3510_b_contrast_store);

static ssize_t ins3510_a_saturation_store(struct device *dev, struct device_attribute *attr, const char *buf, size_t n)
{
	int val;
	sscanf(buf, "%x", &val);
	if( val<0 || val>0xffff)
        	printk(KERN_ERR "ins3510: invalid input 0x%x\n", val);
	else if( ins3510hw_data.bootup )
	{
		ins3510hw_set_saturation(0, val);
		ins3510hw_set_go_bit();
	}
	else
        	printk(KERN_ERR "ins3510: not bootup\n");
	return n;
}

static DEVICE_ATTR(a_saturation, S_IRUGO|S_IWUSR, NULL, ins3510_a_saturation_store);

static ssize_t ins3510_b_saturation_store(struct device *dev, struct device_attribute *attr, const char *buf, size_t n)
{
	int val;
	sscanf(buf, "%x", &val);
	if( val<0 || val>0xffff)
        	printk(KERN_ERR "ins3510: invalid input 0x%x\n", val);
	else if( ins3510hw_data.bootup )
	{
		ins3510hw_set_saturation(1, val);
		ins3510hw_set_go_bit();
	}
	else
        	printk(KERN_ERR "ins3510: not bootup\n");
	return n;
}

static DEVICE_ATTR(b_saturation, S_IRUGO|S_IWUSR, NULL, ins3510_b_saturation_store);

static ssize_t ins3510_jpeg_quality_store(struct device *dev, struct device_attribute *attr, const char *buf, size_t n)
{
	int val;
	sscanf(buf, "%x", &val);
	if( val<0 || val>0x0f )
        	printk(KERN_ERR "ins3510: invalid input 0x%x\n", val);
	else if( ins3510hw_data.bootup )
	{
		ins3510hw_set_jpeg_quality(val);
		ins3510hw_set_go_bit();
	}
	else
        	printk(KERN_ERR "ins3510: not bootup\n");
	return n;
}

static DEVICE_ATTR(jpegquality, S_IRUGO|S_IWUSR, NULL, ins3510_jpeg_quality_store);

static ssize_t ins3510_special_mode_store(struct device *dev, struct device_attribute *attr, const char *buf, size_t n)
{
	int val;
	sscanf(buf, "%x", &val);
	if( val<0 || val>0xff )
        	printk(KERN_ERR "ins3510: invalid input 0x%x\n", val);
	else if( ins3510hw_data.bootup )
	{
		ins3510hw_set_special_mode(val);
		ins3510hw_set_go_bit();
	}
	else
        	printk(KERN_ERR "ins3510: not bootup\n");
	return n;
}

static DEVICE_ATTR(specialmode, S_IRUGO|S_IWUSR, NULL, ins3510_special_mode_store);

static ssize_t ins3510_wb_mode_store(struct device *dev, struct device_attribute *attr, const char *buf, size_t n)
{
	int val;
	sscanf(buf, "%x", &val);
	if( val<0 || val>0xff )
        	printk(KERN_ERR "ins3510: invalid input 0x%x\n", val);
	else if( ins3510hw_data.bootup )
	{
		ins3510hw_set_wb_mode(val);
		ins3510hw_set_go_bit();
	}
	else
        	printk(KERN_ERR "ins3510: not bootup\n");
	return n;
}

static DEVICE_ATTR(wbmode, S_IRUGO|S_IWUSR, NULL, ins3510_wb_mode_store);

static ssize_t ins3510_af_store(struct device *dev, struct device_attribute *attr, const char *buf, size_t n)
{
	int val;
	sscanf(buf, "%x", &val);
	if( ins3510hw_data.bootup )
	{
		if( val ) {
			/* enable single auto focus */
			ins3510hw_single_af();
		}
		else
			ins3510hw_af_idle();
	}
	else
        	printk(KERN_ERR "ins3510: not bootup\n");
	return n;
}

static DEVICE_ATTR(af, S_IRUGO|S_IWUSR, NULL, ins3510_af_store);

static ssize_t ins3510_i2c8_store(struct device *dev, struct device_attribute *attr, const char *buf, size_t n)
{
	u32 addr, val;
	char cmd;
	int ret, autoOn;
	u8 reg8;

	sscanf(buf, "%c %d %x %x", &cmd, &autoOn, &addr, &val);

	if(autoOn)
	{
    	sensor_clock_on();
    	sensor_power_on();
    	if( (ret=ins3510hw_bootup())<0 )
    	{
    		printk(KERN_INFO "ins3510: bootup failed %d\n", ret);
    		sensor_power_off();
    	    sensor_clock_off();
    	    return n;
    	}
    }

	if ('r' == cmd)
	{
	    ins3510_i2c_read_reg8(addr, &reg8);
	    printk("ins3510: read i2c addr=%02x val=%02x\n", addr, reg8);
	}
	else
	{
		ins3510_i2c_write_reg8(addr, val);
		printk("ins3510: write i2c addr=%02x val=%02x\n", addr, val);
		ins3510_i2c_read_reg8(addr, &reg8);
	    printk("ins3510: read i2c addr=%02x val=%02x\n", addr, reg8);
	}

	if(autoOn)
	{
    	sensor_power_off();
    	sensor_clock_off();
    }

	return n;
}

static DEVICE_ATTR(i2c8rw, S_IRUGO|S_IWUSR, NULL, ins3510_i2c8_store);

static ssize_t ins3510_i2c16_store(struct device *dev, struct device_attribute *attr, const char *buf, size_t n)
{
	u32 addr, val;
	char cmd;
	int ret, autoOn;
	u16 reg16;

	sscanf(buf, "%c %d %x %x", &cmd, &autoOn, &addr, &val);

	if(autoOn)
	{
    	sensor_clock_on();
    	sensor_power_on();
    	if( (ret=ins3510hw_bootup())<0 )
    	{
    		printk(KERN_INFO "ins3510: bootup failed %d\n", ret);
    		sensor_power_off();
    	    sensor_clock_off();
    	    return n;
    	}
    }

	if ('r' == cmd)
	{
	    ins3510_i2c_read_reg16(addr, &reg16);
	    printk("ins3510: read i2c addr=%02x val=%04x\n", addr, reg16);

	}
	else
	{
		ins3510_i2c_write_reg16(addr, val);
		printk("ins3510: write i2c addr=%02x val=%04x\n", addr, val);
		ins3510_i2c_read_reg16(addr, &reg16);
	    printk("ins3510: read i2c addr=%02x val=%04x\n", addr, reg16);
	}

	if(autoOn)
	{
    	sensor_power_off();
    	sensor_clock_off();
    }

	return n;
}

static DEVICE_ATTR(i2c16rw, S_IRUGO|S_IWUSR, NULL, ins3510_i2c16_store);

static struct attribute *ins3510_sysfs_attr[] = {
	&dev_attr_fwversion.attr,
	&dev_attr_fwupgrade.attr,
	&dev_attr_testpattern.attr,
	&dev_attr_fps.attr,
	&dev_attr_afpos.attr,
	&dev_attr_aeoff.attr,
	&dev_attr_exiforien.attr,
	&dev_attr_sharpness.attr,
	&dev_attr_a_gamma.attr,
	&dev_attr_b_gamma.attr,
	&dev_attr_a_brightness.attr,
	&dev_attr_b_brightness.attr,
	&dev_attr_a_contrast.attr,
	&dev_attr_b_contrast.attr,
	&dev_attr_a_saturation.attr,
	&dev_attr_b_saturation.attr,
	&dev_attr_jpegquality.attr,
	&dev_attr_specialmode.attr,
	&dev_attr_wbmode.attr,
	&dev_attr_af.attr,
	NULL
};

static struct attribute *ins3510_sysfs_debug_attrs[] = {
	&dev_attr_i2clog.attr,
	&dev_attr_camlog.attr,
	&dev_attr_mclock.attr,
	&dev_attr_clock_on.attr,
	&dev_attr_voltage_on.attr,
	&dev_attr_power_on.attr,
	&dev_attr_standby.attr,
	&dev_attr_bootup.attr,
	&dev_attr_reset.attr,
	&dev_attr_dumpregs.attr,
	&dev_attr_i2ctest.attr,
	&dev_attr_i2crwtest.attr,
	&dev_attr_boottest.attr,
	&dev_attr_setgotest.attr,
	&dev_attr_i2c8rw.attr,
	&dev_attr_i2c16rw.attr,
	NULL
};

static struct attribute_group ins3510_sysfs_debug_attr_group = {
	.name	= "debug",
	.attrs	= ins3510_sysfs_debug_attrs,
};

static int ins3510_sysfs_add(struct kobject *kobj)
{
	int i, n, ret;
	n = ARRAY_SIZE(ins3510_sysfs_attr);
	for (i = 0; i < n; ++i) {
		if( ins3510_sysfs_attr[i] ) {
			ret = sysfs_create_file(kobj, ins3510_sysfs_attr[i]);
			if (ret)
				return ret;
		}
	}
	ret = sysfs_create_group(kobj, &ins3510_sysfs_debug_attr_group);
	if (ret)
		return ret;
	return 0;
}

static int ins3510_sysfs_rm(struct kobject *kobj)
{
	int i, n;
	sysfs_remove_group(kobj, &ins3510_sysfs_debug_attr_group);

	n = ARRAY_SIZE(ins3510_sysfs_attr);
	for (i = 0; i < n; i++) {
		if( ins3510_sysfs_attr[i] )
			sysfs_remove_file(kobj, ins3510_sysfs_attr[i]);
	}
	return 0;
}


/*****************************************************************************
 *                              INS3510 I2C Client Driver
 *****************************************************************************/

static int ins3510_i2c_read_reg8( const u8 subAddress, u8 *bufP )
{
	int ret;

	if( g_client == NULL )	/*	No global client pointer?	*/
		return -1;

	ret = i2c_smbus_read_byte_data(g_client, subAddress);
	if (ret >= 0) {
		*bufP = ret;
		I2C_dbg(7, "ins3510: read 8:[0x%02x]=0x%02x\n", subAddress, ret);
	}
	else
		printk(KERN_ERR "ins3510_i2c_read_reg8(0x%02x) failed, return %d\n", subAddress, ret);
	return ret;
}

static int ins3510_i2c_write_reg8( const u8 subAddress, u8 regval )
{
	int ret;

	if( g_client == NULL )	/*	No global client pointer?	*/
		return -1;

	I2C_dbg(7, "ins3510: write 8:[0x%02x] 0x%02x\n", subAddress, regval);
	ret = i2c_smbus_write_byte_data(g_client, subAddress, regval);
	if( ret<0 )
		printk(KERN_ERR "ins3510_i2c_write_reg8(0x%02x, 0x%02x) failed, return %d\n", subAddress, regval, ret);
	return ret;
}

static int ins3510_i2c_read_reg16( const u8 subAddress, u16 *bufP )
{
	int ret;

	if( g_client == NULL )	/*	No global client pointer?	*/
		return -1;

	ret = i2c_smbus_read_word_data(g_client, subAddress);
	if (ret >= 0) {
		/* INS3510 is big endian, convert */
		*bufP = ((ret>>8)&0x00ff) | ((ret&0x00ff)<<8);
		I2C_dbg(7, "ins3510: read 16:[0x%02x]=0x%04x\n", subAddress, *bufP);
	}
	else
		printk(KERN_ERR "ins3510_i2c_read_reg16(0x%02x) failed, return %d\n", subAddress, ret);
	return ret;
}

static int ins3510_i2c_write_reg16( const u8 subAddress, u16 regval )
{
	int ret;
	int i2cval;

	if( g_client == NULL )	/*	No global client pointer?	*/
		return -1;

	/* INS3510 is big endian, convert */
	i2cval = ((regval>>8)&0x00ff) | ((regval&0x00ff)<<8);
	I2C_dbg(7, "ins3510: write 16:[0x%02x] 0x%04x\n", subAddress, regval);

	ret = i2c_smbus_write_word_data(g_client, subAddress, i2cval);
	if( ret<0 )
		printk(KERN_ERR "ins3510_i2c_write_reg16(0x%02x, 0x%04x) failed, return %d\n", subAddress, regval, ret);
	return ret;
}

static int ins3510_i2c_read_reg32( const u8 subAddress, u32 *bufP )
{
	int ret1, ret2;

	if( g_client == NULL )	/*	No global client pointer?	*/
		return -1;

	ret1 = i2c_smbus_read_word_data(g_client, subAddress);
	if ( ret1< 0 )
	{
		printk(KERN_ERR "ins3510_i2c_read_reg32(0x%02x:0) failed, return %d\n", subAddress, ret1);
		return ret1;
	}

	ret2 = i2c_smbus_read_word_data(g_client, subAddress+2);
	if ( ret2< 0 )
	{
		printk(KERN_ERR "ins3510_i2c_read_reg32(0x%02x:2) failed, return %d\n", subAddress, ret2);
		return ret2;
	}

	/* INS3510 is big endian, convert */
	*bufP = ((ret1<<8)&0xff0000) | ((ret1&0x00ff)<<24)
	      |	((ret2>>8)&0x00ff) | ((ret2&0x00ff)<<8);

	I2C_dbg(7, "ins3510: read 32:[0x%02x]=0x%08x\n", subAddress, *bufP);
	return 0;
}

static int ins3510_i2c_write_reg32( const u8 subAddress, u32 regval )
{
	int ret;
	u8 data[5];

	if( g_client == NULL )	/*	No global client pointer?	*/
		return -1;

	I2C_dbg(7, "ins3510: write 32:[0x%02x] 0x%08x\n", subAddress, regval);
	data[0] = subAddress;
	data[1] = (regval>>24) & 0xff;
	data[2] = (regval>>16) & 0xff;
	data[3] = (regval>>8) & 0xff;
	data[4] = (regval) & 0xff;

	ret = i2c_master_send(g_client, data, 5);
	if( ret<0 )
		printk(KERN_ERR "ins3510_i2c_write_reg32(0x%02x, 0x%08x) failed, return %d\n", subAddress, regval, ret);
	return ret;
}

static int ins3510_i2c_ID_detect(void)
{
	u32 version;
	int ret;

	ret = ins3510_i2c_read_reg32(0x80, &version);
	if( ret<0 )
		return ret;

	printk(KERN_INFO "INS3510: version=%08x\n", version);

	if( version == INS3510_CHIP_VER_BOOT )
		return 0;
	else
		return -EINVAL;
}

static int ins3510_create_workqueue(void)
{
	INIT_DELAYED_WORK(&flash_work, camera_flash_work_handler);
	flash_wqueue = create_workqueue("ins3510_flash");
	if(!flash_wqueue){
		printk(KERN_ERR "ins3510: create_workqueue(flash) failed\n");
		return -ENOMEM;
	}
	return 0;
}

static void ins3510_destroy_workqueue(void)
{
	if(flash_wqueue){
		cancel_rearming_delayed_workqueue(flash_wqueue, &flash_work);
		destroy_workqueue(flash_wqueue);
		flash_wqueue = NULL;
	}
}


static int __init ins3510_i2c_probe(struct i2c_client *client, const struct i2c_device_id *_id)
{
	struct sensor_platform_data *pdata;
	int id = 0;

	pdata = client->dev.platform_data;
	id = pdata->id;

	if( sensor_clock_on()<0 )
		return -EIO;
	if(pdata->power_on)
		pdata->power_on(id);

	g_client = client;

	if (ins3510_i2c_ID_detect()<0){
		goto err;
	}

	if(ins3510_register(id)<0)
		goto err;

	printk(KERN_INFO "INS3510 detected.\n");

	//ins3510hw_bootup();
	//ins3510hw_outputpin_hiz();

#ifdef INS3510_STANDBY
	if(pdata->standby)
		pdata->standby(id);
#else
	if(pdata->power_off)
		pdata->power_off(id);
#endif
	sensor_clock_off();

	ins3510hw_data_init();
	/* voltage shall be already on when bootup */
	ins3510hw_data.voltage_on = 1;

	ins3510_sysfs_add(&client->dev.kobj);

	ins3510_create_workqueue();

	return 0;
err:
	printk(KERN_ERR "INS3510 detect failed.\n");

	if(pdata->power_off)
		pdata->power_off(id);
	sensor_clock_off();

	g_client = NULL;

	return -1;
}

static int ins3510_i2c_remove(struct i2c_client *client)
{
	struct sensor_platform_data *pdata;
	pdata = client->dev.platform_data;
	
	ins3510_unregister(pdata->id);

	ins3510_destroy_workqueue();

	ins3510_sysfs_rm(&client->dev.kobj);

	g_client = NULL;
	return 0;
}

static const struct i2c_device_id sensor_id[] = {
	{ "sensor_ins3510", 0 },
	{ }
};

static struct i2c_driver ins3510_i2c_driver = {
	.driver = {
		.name	= "sensor_ins3510",
	},
	.probe		= ins3510_i2c_probe,
	.remove		= ins3510_i2c_remove,
	.id_table	= sensor_id,
};

static int __init ins3510_module_init(void)
{
	int ret;

	if ( (ret = i2c_add_driver(&ins3510_i2c_driver)) ) {
		printk(KERN_ERR "ins3510: Driver registration failed,"
			"module not inserted.\n");
		return ret;
	}

#ifdef CONFIG_PXA3xx_DVFM
	dvfm_register("ins3510", &dvfm_dev_idx);
#endif
#ifdef CONFIG_ANDROID_POWER
	android_init_suspend_lock(&ins3510_ops_suspend_lock);
#endif
	return 0;
}

static void __exit ins3510_module_exit(void)
{
#ifdef CONFIG_ANDROID_POWER
	android_uninit_suspend_lock(&ins3510_ops_suspend_lock);
#endif
#ifdef CONFIG_PXA3xx_DVFM
	dvfm_unregister("ins3510", &dvfm_dev_idx);
#endif
	i2c_del_driver(&ins3510_i2c_driver);
}

MODULE_DESCRIPTION("INS3510 I2C Client driver");
MODULE_LICENSE("GPL");

module_init(ins3510_module_init);
module_exit(ins3510_module_exit);

