/*
   Copyright (C) 2005, Intel Corporation.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.


 *(C) Copyright 2006 Marvell International Ltd.
 * All Rights Reserved
 */

/*****************************************************************************
 * Abstract:
 *	Contains all hardware related functions for OV3640
 *
 * Notes:Only valid for processor code named Monahans.
 *****************************************************************************/
#include "camera.h"

#include <linux/init.h>
#include <linux/module.h>
#include <linux/delay.h>
#include <mach/mfp.h>
#include <mach/gpio.h>
#include <mach/hardware.h>
#include <asm/uaccess.h>
#include <linux/err.h>
#include <linux/proc_fs.h>

#include <linux/clk.h>
#include <linux/i2c.h>

/*****************************************************************************
 * Constants & Registers
 *****************************************************************************/
/* CICLK for QCI */
#define CICLK	2600
//#define CICLK	1300

/* Revision constants */
#define PID_VERH		0x36
#define PID_VERL		0x4C
#define PID_ADRRH		0x300a
#define PID_ADRRL		0x300b

#define AF_SETTING		0
/* Output Size & Format */
#define OV3640_SIZE_QVGA	1		//320x240
#define OV3640_SIZE_VGA		2		//640x480
#define OV3640_SIZE_SVAG	3		//800x600
#define OV3640_SIZE_D1		4		//864x600
#define OV3640_SIZE_XGA		5		//1024x768
#define OV3640_SIZE_SXGA	6		//1280x1024
#define OV3640_SIZE_UXGA	7		//1600X1200
#define OV3640_SIZE_QXGA	8		//2048x1536
#define OV3640_SIZE_QSXGA	9		//2592X1944
#define OV3640_SIZE_YGA		10		//480X360

#define OV3640_SIZE_SQCIF	100		//128x96
#define OV3640_SIZE_QCIF	101		//176x144
#define OV3640_SIZE_CIF		102		//352x288
#define OV3640_SIZE_JPEG_VGA	103		//640x480
#define OV3640_SIZE_256X192	200		//256X192 for 4X zoom
#define OV3640_SIZE_JPEG_QVGA	201		//320X240
#define OV3640_SIZE_JPEG_256X192 202		//256X192

#define OV3640_FORMAT_YUV_422	( 1 << 8)
#define OV3640_FORMAT_RAW10	( 2 << 8)
#define OV3640_FORMAT_RAW8	( 3 << 8)
#define OV3640_FORMAT_JPG	( 4 << 8)
#define OV3640_FORMAT_RGB565	( 5 << 8)

#define OV3640_QVGA_WIDTH	320
#define OV3640_QVGA_HEIGHT	240
#define OV3640_VGA_WIDTH	640
#define OV3640_VGA_HEIGHT	480
#define OV3640_YGA_WIDTH	480
#define OV3640_YGA_HEIGHT	360
#define OV3640_SQCIF_WIDTH	128
#define OV3640_SQCIF_HEIGHT	96
#define OV3640_QCIF_WIDTH	176
#define OV3640_QCIF_HEIGHT	144
#define OV3640_CIF_WIDTH	352
#define OV3640_CIF_HEIGHT	288
#define OV3640_XGA_WIDTH	1024
#define OV3640_XGA_HEIGHT	768
#define OV3640_UXGA_WIDTH	1600
#define OV3640_UXGA_HEIGHT	1200
#define OV3640_QXGA_WIDTH	2048
#define OV3640_QXGA_HEIGHT	1536

#define OV3640_REGEND		(0xFFFF)

/* Unique ID allocation */
static struct i2c_client *g_client;

#define I2C_DEBUG
//#undef I2C_DEBUG

#ifdef  I2C_DEBUG
#define ovI2C_dbg(fmt, arg...)    printk(KERN_INFO "%s(line %d): " fmt "\n",  \
                __FUNCTION__, __LINE__, ##arg)
#else 
#define ovI2C_dbg(fmt, arg...)        do {} while (0)
#endif

#define ARRAY_AND_SIZE(x)	(x), ARRAY_SIZE(x)

/****************************************************************************

   Gloable variable
   
*****************************************************************************/
struct frame_sizes_t {
	u32	width;
	u32	height;
};

/* preview YUV422 size list, supported preview frame sizes "preview-size-values" on Android */
static const struct frame_sizes_t ov3640_yuv422_frame_sizes[] = {
	{480, 360},
	{320, 240},
	{160, 120},
	{128, 96}
};

/* capture JPEG size list, supported capture frame sizes "picture-size-values" on Android */
static const struct frame_sizes_t ov3640_jpeg_frame_sizes[] = {
	{2048, 1536},
	{1600, 1200},
	{1024, 768},
	{640, 480},
	{320, 240}
};

/* frame interval list,  frame rate = 1 / frame interval,
   supported frame rates "preview-frame-rate-values" on Android */
static const struct v4l2_fract ov3640_yuv422_frame_intervals[] = {
	{2, 15},	/* 2/15 seconds, 7.5fps, the frame rate could be down to 7.5 due to night mode enabled */
	{1, 15},	/* 1/15 seconds, 15fps */
	{1, 25},	/* 1/25 seconds, 25fps */
	{1, 30}		/* 1/30 seconds, 30fps */
};

static int G_ov3640_mode_change = CAMERA_MODE_VIDEO;
static int g_ov3640_size = -1;
static int g_ov3640_format = -1;
#if 0
static u8  ov3640_gaincode = 0;
static int  ov3640_expcode = 0;
static u8  ov3640_recRgain =0,ov3640_recBgain=0,ov3640_recGgain=0;
#define QXGAMAXEXPLIN 1568
#define XGAMAXEXPLIN    779
const static u8 gSensorSlaveAddr = 0x3c;
#endif

#define FLASHLED_PROC_FILE	"driver/flashled"
#define OV3640_PROC_FILE	"driver/ov3640"

static struct proc_dir_entry *flashled_proc_file;
static struct proc_dir_entry *ov3640_proc_file;
static int flashled_flag= 0;
static int flashled_on = 0;
static int flashled_fire = 0;
static struct timeval flashled_fire_time;
static int af_flag= 0;
static int reg_format= 0, reg_size= 0;

static int ov3640_single_af_enabled = 0;
static int ov3640set_fps = 0;

static struct delayed_work flash_work;
static struct workqueue_struct *flash_wqueue;
static DECLARE_MUTEX(ov3640_flash_lock);

static DECLARE_MUTEX(ov3640_i2c_lock);

static inline int read_sensor_reg( const u16 subAddress, u8 *bufP );
static inline int write_sensor_reg( const u16 subAddress, u8 bufP );

static unsigned int atoi( char* str)
{
	unsigned int i= 0, radius= 10, val= 0;

	// hex
	if( ( str[ i] == '0') && ( str[ i+ 1] == 'x'))
	{
		radius= 16;
		i+= 2;
	}

	while( str[ i] != '\0')
	{
		val*= radius;

		if( ( str[ i] >= '0') && ( str[ i] <= '9'))
			val+= str[ i]- '0';
		else if( ( str[ i] >= 'A') && ( str[ i] <= 'F'))
			val+= 10+ str[ i]- 'A';
		else if( ( str[ i] >= 'a') && ( str[ i] <= 'f'))
			val+= 10+ str[ i]- 'a';
		else			// not a recognized char
		{
			val/= radius;
			break;
		}
		i++;
	}

	return val;
}

/*****************************************************************************

 *  Register Settings Function

 *****************************************************************************/

/* include AF firmware register tabs */
#include "ov3640_affw_dw9710.h"

typedef struct{
	u16	reg;
	u8	val;
	u8	mask;
} OV3640_REG_TABLE;

static
const OV3640_REG_TABLE OV3640_init[]=
{
//{ 0x3012, 0x80, 0x00},
{ 0x304d, 0x45, 0x00},
{ 0x3087, 0x16, 0x00},
{ 0x30aa, 0x45, 0x00},
{ 0x30b0, 0xff, 0x00},
{ 0x30b1, 0xff, 0x00},
{ 0x30b2, 0x11, 0x00},
{ 0x300e, 0x32, 0x00},
{ 0x300f, 0x21, 0x00},
{ 0x3010, 0x20, 0x00},
{ 0x3011, 0x00, 0x00},
{ 0x304c, 0x85, 0x00},
{ 0x3018, 0x38, 0x00},
{ 0x3019, 0x30, 0x00},
{ 0x301a, 0x61, 0x00},
{ 0x307d, 0x00, 0x00},
{ 0x3087, 0x02, 0x00},
{ 0x3082, 0x20, 0x00},
{ 0x3015, 0x22, 0x00},
{ 0x3014, 0x0c, 0x00},
{ 0x3013, 0xf7, 0x00},
{ 0x303c, 0x08, 0x00},
{ 0x303d, 0x18, 0x00},
{ 0x303e, 0x06, 0x00},
{ 0x303F, 0x0c, 0x00},

/* AEC Weight */
{ 0x3030, 0x62, 0x00},
{ 0x3031, 0x26, 0x00},
{ 0x3032, 0xe6, 0x00},
{ 0x3033, 0x6e, 0x00},
{ 0x3034, 0xea, 0x00},
{ 0x3035, 0xae, 0x00},
{ 0x3036, 0xa6, 0x00},
{ 0x3037, 0x6a, 0x00},
/* AEC Weight */

{ 0x3104, 0x02, 0x00},
{ 0x3105, 0xfd, 0x00},
{ 0x3106, 0x00, 0x00},
{ 0x3107, 0xff, 0x00},
{ 0x3300, 0x13, 0x00},
{ 0x3366, 0x15, 0x00},
{ 0x3301, 0xde, 0x00},
{ 0x3302, 0xef, 0x00},
{ 0x3316, 0xff, 0x00},
{ 0x3317, 0x00, 0x00},
{ 0x3308, 0xa5, 0x00},
{ 0x3312, 0x26, 0x00},
{ 0x3314, 0x42, 0x00},
{ 0x3313, 0x2b, 0x00},
{ 0x3315, 0x42, 0x00},
{ 0x3310, 0xd0, 0x00},
{ 0x3311, 0xbd, 0x00},
{ 0x330c, 0x18, 0x00},
{ 0x330d, 0x18, 0x00},
{ 0x330e, 0x56, 0x00},
{ 0x330f, 0x5c, 0x00},
{ 0x330b, 0x1c, 0x00},
{ 0x3306, 0x5c, 0x00},
{ 0x3307, 0x11, 0x00},
{ 0x336a, 0x52, 0x00},
{ 0x3370, 0x46, 0x00},
{ 0x3376, 0x38, 0x00},
{ 0x30b8, 0x20, 0x00},
{ 0x30b9, 0x17, 0x00},
{ 0x30ba, 0x04, 0x00},
{ 0x30bb, 0x08, 0x00},
{ 0x3100, 0x02, 0x00},
{ 0x3301, 0xde, 0x00},
{ 0x3304, 0xfc, 0x00},
{ 0x3400, 0x00, 0x00},
{ 0x3404, 0x02, 0x00},
{ 0x3012, 0x10, 0x00},
{ 0x3023, 0x06, 0x00},
{ 0x3026, 0x03, 0x00},
{ 0x3027, 0x04, 0x00},
{ 0x302a, 0x03, 0x00},
{ 0x302b, 0x10, 0x00},
{ 0x3075, 0x24, 0x00},
{ 0x300d, 0x01, 0x00},
{ 0x30d7, 0x90, 0x00},
{ 0x3069, 0x04, 0x00},
{ 0x303e, 0x00, 0x00},
{ 0x303f, 0xc0, 0x00},
{ 0x335f, 0x34, 0x00},
{ 0x3360, 0x0c, 0x00},
{ 0x3361, 0x04, 0x00},
{ 0x3362, 0x01, 0x00},
{ 0x3363, 0x48, 0x00},
{ 0x3364, 0xf4, 0x00},
{ 0x3403, 0x42, 0x00},
{ 0x3088, 0x01, 0x00},
{ 0x3089, 0x40, 0x00},
{ 0x308a, 0x00, 0x00},
{ 0x308b, 0xf0, 0x00},
{ 0x3600, 0xc4, 0x00},
{ 0x3012, 0x10, 0x00},
{ 0x3023, 0x06, 0x00},
{ 0x3026, 0x03, 0x00},
{ 0x3027, 0x04, 0x00},
{ 0x302a, 0x03, 0x00},
{ 0x302b, 0x10, 0x00},
{ 0x3075, 0x24, 0x00},
{ 0x300d, 0x01, 0x00},
{ 0x30d7, 0x90, 0x00},
{ 0x3069, 0x04, 0x00},
{ 0x303e, 0x00, 0x00},
{ 0x303f, 0xc0, 0x00},
{ 0x3302, 0xef, 0x00},
{ 0x335f, 0x34, 0x00},
{ 0x3360, 0x0c, 0x00},
{ 0x3361, 0x04, 0x00},
{ 0x3362, 0x11, 0x00},
{ 0x3363, 0xe8, 0x00},
{ 0x3364, 0x6c, 0x00},
{ 0x3403, 0x42, 0x00},
{ 0x3088, 0x01, 0x00},
{ 0x3089, 0xe0, 0x00},
{ 0x308a, 0x01, 0x00},
{ 0x308b, 0x68, 0x00},
{ 0x3020, 0x01, 0x00},
{ 0x3021, 0x1d, 0x00},
{ 0x3022, 0x00, 0x00},
{ 0x3023, 0x06, 0x00},
{ 0x3024, 0x08, 0x00},
{ 0x3025, 0x18, 0x00},
{ 0x3026, 0x03, 0x00},
{ 0x3027, 0x04, 0x00},
{ 0x304c, 0x85, 0x00},
{ 0x3100, 0x02, 0x00},
{ 0x3301, 0xde, 0x00},
{ 0x3304, 0xfc, 0x00},
{ 0x3400, 0x00, 0x00},
{ 0x3404, 0x02, 0x00},
{ 0x3600, 0x02, 0x00},
{ 0x3016, 0x81, 0x00},
{ 0x30d9, 0x05, 0x00},
{ 0x3317, 0x04, 0x00},
{ 0x3316, 0xf8, 0x00},
{ 0x3312, 0x39, 0x00},
{ 0x3314, 0x5D, 0x00},
{ 0x3313, 0x0F, 0x00},
{ 0x3315, 0x37, 0x00},
{ 0x3310, 0xDA, 0x00},
{ 0x3311, 0xDF, 0x00},
{ 0x330c, 0x0F, 0x00},
{ 0x330d, 0x0F, 0x00},
{ 0x330e, 0x6A, 0x00},
{ 0x330f, 0x47, 0x00},
{ 0x330b, 0x0F, 0x00},
{ 0x3306, 0x5C, 0x00},
{ 0x3307, 0x11, 0x00},
{ 0x3308, 0x25, 0x00},
{ 0x3340, 0x20, 0x00},
{ 0x3341, 0x64, 0x00},
{ 0x3342, 0x08, 0x00},
{ 0x3343, 0x30, 0x00},
{ 0x3344, 0x90, 0x00},
{ 0x3345, 0xc0, 0x00},
{ 0x3346, 0xa0, 0x00},
{ 0x3347, 0x98, 0x00},
{ 0x3348, 0x08, 0x00},
{ 0x3349, 0x98, 0x00},

/* @@ LensC FIH - 090703-01 */
{ 0x3300, 0x13, 0x00},
{ 0x3304, 0xfc, 0x00},
{ 0x3366, 0x15, 0x00},
{ 0x3367, 0x34, 0x00},
{ 0x3368, 0x00, 0x00},
{ 0x3369, 0x06, 0x00},
{ 0x336a, 0x52, 0x00},
{ 0x336b, 0x87, 0x00},
{ 0x336c, 0xc2, 0x00},
{ 0x336d, 0x34, 0x00},
{ 0x336e, 0x20, 0x00},
{ 0x336f, 0x06, 0x00},
{ 0x3370, 0x46, 0x00},
{ 0x3371, 0x87, 0x00},
{ 0x3372, 0xc2, 0x00},
{ 0x3373, 0x34, 0x00},
{ 0x3374, 0x00, 0x00},
{ 0x3375, 0x06, 0x00},
{ 0x3376, 0x46, 0x00},
{ 0x3377, 0x87, 0x00},
{ 0x3378, 0xc2, 0x00},

{ 0x331b, 0x03, 0x00},
{ 0x331c, 0x12, 0x00},
{ 0x331d, 0x2c, 0x00},
{ 0x331e, 0x55, 0x00},
{ 0x331f, 0x67, 0x00},
{ 0x3320, 0x78, 0x00},
{ 0x3321, 0x86, 0x00},
{ 0x3322, 0x93, 0x00},
{ 0x3323, 0x9e, 0x00},
{ 0x3324, 0xa8, 0x00},
{ 0x3325, 0xb5, 0x00},
{ 0x3326, 0xc1, 0x00},
{ 0x3327, 0xd2, 0x00},
{ 0x3328, 0xe1, 0x00},
{ 0x3329, 0xec, 0x00},
{ 0x332a, 0x1a, 0x00},
{ 0x303c, 0x02, 0x00},
{ 0x303d, 0x00, 0x00},
{ 0x303e, 0x01, 0x00},
{ 0x303F, 0x80, 0x00},
/* AEC Weight */
{ 0x3030, 0x62, 0x00},
{ 0x3031, 0x26, 0x00},
{ 0x3032, 0xe6, 0x00},
{ 0x3033, 0x6e, 0x00},
{ 0x3034, 0xea, 0x00},
{ 0x3035, 0xae, 0x00},
{ 0x3036, 0xa6, 0x00},
{ 0x3037, 0x6a, 0x00},
/* AEC Weight */
{ 0x3070, 0x00, 0x00},
{ 0x3071, 0xfd, 0x00},
{ 0x3072, 0x00, 0x00},
{ 0x3073, 0xd2, 0x00},
{ 0x301c, 0x02, 0x00},
{ 0x301d, 0x02, 0x00},
};

static
const OV3640_REG_TABLE OV3640_video_YUV422[]=
{
{ 0x3012, 0x10, 0x00},
{ 0x3023, 0x06, 0x00},
{ 0x3026, 0x03, 0x00},
{ 0x3027, 0x04, 0x00},
{ 0x302a, 0x03, 0x00},
{ 0x302b, 0x10, 0x00},
{ 0x3075, 0x24, 0x00},
{ 0x300d, 0x01, 0x00},
{ 0x30d7, 0x90, 0x00},
{ 0x3069, 0x04, 0x00},
{ 0x303e, 0x00, 0x00},
{ 0x303f, 0xc0, 0x00},
{ 0x3302, 0xef, 0x00},
{ 0x335f, 0x34, 0x00},
{ 0x3360, 0x0c, 0x00},
{ 0x3361, 0x04, 0x00},
{ 0x3362, 0x34, 0x00},
{ 0x3363, 0x08, 0x00},
{ 0x3364, 0x04, 0x00},
{ 0x3403, 0x42, 0x00},
{ 0x3088, 0x04, 0x00},
{ 0x3089, 0x00, 0x00},
{ 0x308a, 0x03, 0x00},
{ 0x308b, 0x00, 0x00},
{ 0x300e, 0x32, 0x00},
{ 0x300f, 0x21, 0x00},
{ 0x3010, 0x20, 0x00},
{ 0x3011, 0x00, 0x00},
{ 0x302e, 0x00, 0x00},
{ 0x302d, 0x00, 0x00},
{ 0x304c, 0x84, 0x00},
{ 0x3100, 0x02, 0x00},
{ 0x3301, 0xde, 0x00},
{ 0x3304, 0xfc, 0x00},
{ 0x3400, 0x00, 0x00},
{ 0x3404, 0x02, 0x00},
{ 0x3600, 0x02, 0x00},
{ 0x3366, 0x15, 0x00},
};

static
const OV3640_REG_TABLE OV3640_video_VGA[]=
{
{ 0x3302, 0xef, 0x00},
{ 0x335f, 0x34, 0x00},
{ 0x3360, 0x0c, 0x00},
{ 0x3361, 0x04, 0x00},
{ 0x3362, 0x12, 0x00},
{ 0x3363, 0x88, 0x00},
{ 0x3364, 0xe4, 0x00},
{ 0x3403, 0x42, 0x00},
{ 0x3088, 0x02, 0x00},
{ 0x3089, 0x80, 0x00},
{ 0x308a, 0x01, 0x00},
{ 0x308b, 0xe0, 0x00},
{ 0x304c, 0x83, 0x00},
};

static
const OV3640_REG_TABLE OV3640_video_QVGA[]=
{
{ 0x3302, 0xef, 0x00},
{ 0x335f, 0x34, 0x00},
{ 0x3360, 0x0c, 0x00},
{ 0x3361, 0x04, 0x00},
{ 0x3362, 0x01, 0x00},
{ 0x3363, 0x48, 0x00},
{ 0x3364, 0xf4, 0x00},
{ 0x3403, 0x42, 0x00},
{ 0x3088, 0x01, 0x00},
{ 0x3089, 0x40, 0x00},
{ 0x308a, 0x00, 0x00},
{ 0x308b, 0xf0, 0x00},
{ 0x304c, 0x85, 0x00},
};

static
const OV3640_REG_TABLE OV3640_video_YGA[]=
{
{ 0x3302, 0xef, 0x00},
{ 0x335f, 0x34, 0x00},
{ 0x3360, 0x0c, 0x00},
{ 0x3361, 0x04, 0x00},
{ 0x3362, 0x11, 0x00},
{ 0x3363, 0xe8, 0x00},
{ 0x3364, 0x6c, 0x00},
{ 0x3403, 0x42, 0x00},
{ 0x3088, 0x01, 0x00},
{ 0x3089, 0xe0, 0x00},
{ 0x308a, 0x01, 0x00},
{ 0x308b, 0x68, 0x00},
{ 0x304c, 0x84, 0x00},
};

static
const OV3640_REG_TABLE OV3640_video_YGA2[]=
{
{ 0x3302, 0xef, 0x00},
{ 0x335f, 0x34, 0x00},
{ 0x3360, 0x0c, 0x00},
{ 0x3361, 0x04, 0x00},
{ 0x3362, 0x11, 0x00},
{ 0x3363, 0xe8, 0x00},
{ 0x3364, 0x6c, 0x00},
{ 0x3403, 0x42, 0x00},
{ 0x3088, 0x01, 0x00},
{ 0x3089, 0xe0, 0x00},
{ 0x308a, 0x01, 0x00},
{ 0x308b, 0x68, 0x00},
{ 0x304c, 0x85, 0x00},
};

static
const OV3640_REG_TABLE OV3640_video_CIF[]=
{
{ 0x3012, 0x10, 0x00},
{ 0x3020, 0x03, 0x00},
{ 0x3021, 0x48, 0x00},
{ 0x3022, 0x00, 0x00},
{ 0x3023, 0x06, 0x00},
{ 0x3024, 0x03, 0x00},
{ 0x3025, 0xc2, 0x00},
{ 0x3026, 0x03, 0x00},
{ 0x3027, 0x04, 0x00},
{ 0x302a, 0x03, 0x00},
{ 0x302b, 0x10, 0x00},
{ 0x3075, 0x24, 0x00},
{ 0x300d, 0x01, 0x00},
{ 0x30d7, 0x90, 0x00},
{ 0x3069, 0x04, 0x00},
{ 0x3302, 0xef, 0x00},
{ 0x335f, 0x31, 0x00},
{ 0x3360, 0xe1, 0x00},
{ 0x3361, 0x04, 0x00},
{ 0x3362, 0x11, 0x00},
{ 0x3363, 0x68, 0x00},
{ 0x3364, 0x24, 0x00},
{ 0x3403, 0x42, 0x00},
{ 0x3088, 0x01, 0x00},
{ 0x3089, 0x60, 0x00},
{ 0x308a, 0x01, 0x00},
{ 0x308b, 0x20, 0x00},
};

static
const OV3640_REG_TABLE OV3640_video_QCIF[]=
{
{ 0x3012, 0x10, 0x00},
{ 0x3020, 0x03, 0x00},
{ 0x3021, 0x48, 0x00},
{ 0x3022, 0x00, 0x00},
{ 0x3023, 0x06, 0x00},
{ 0x3024, 0x03, 0x00},
{ 0x3025, 0xc2, 0x00},
{ 0x3026, 0x03, 0x00},
{ 0x3027, 0x04, 0x00},
{ 0x302a, 0x03, 0x00},
{ 0x302b, 0x10, 0x00},
{ 0x3075, 0x24, 0x00},
{ 0x300d, 0x01, 0x00},
{ 0x30d7, 0x90, 0x00},
{ 0x3069, 0x04, 0x00},
{ 0x3302, 0xef, 0x00},
{ 0x335f, 0x31, 0x00},
{ 0x3360, 0xe1, 0x00},
{ 0x3361, 0x04, 0x00},
{ 0x3362, 0x00, 0x00},
{ 0x3363, 0xb8, 0x00},
{ 0x3364, 0x94, 0x00},
{ 0x3403, 0x42, 0x00},
{ 0x3088, 0x00, 0x00},
{ 0x3089, 0xb0, 0x00},
{ 0x308a, 0x00, 0x00},
{ 0x308b, 0x90, 0x00},
};

static
const OV3640_REG_TABLE OV3640_video_SQCIF[]=
{
{ 0x3012, 0x10, 0x00},
{ 0x3020, 0x01, 0x00},
{ 0x3021, 0x1d, 0x00},
{ 0x3022, 0x00, 0x00},
{ 0x3023, 0x06, 0x00},
{ 0x3024, 0x08, 0x00},
{ 0x3025, 0x18, 0x00},
{ 0x3026, 0x03, 0x00},
{ 0x3027, 0x04, 0x00},
{ 0x302a, 0x03, 0x00},
{ 0x302b, 0x10, 0x00},
{ 0x3075, 0x24, 0x00},
{ 0x300d, 0x01, 0x00},
{ 0x30d7, 0x90, 0x00},
{ 0x3069, 0x04, 0x00},
{ 0x3302, 0xef, 0x00},
{ 0x335f, 0x34, 0x00},
{ 0x3360, 0x0c, 0x00},
{ 0x3361, 0x04, 0x00},
{ 0x3362, 0x00, 0x00},
{ 0x3363, 0x88, 0x00},
{ 0x3364, 0x64, 0x00},
{ 0x3403, 0x42, 0x00},
{ 0x3088, 0x00, 0x00},
{ 0x3089, 0x80, 0x00},
{ 0x308a, 0x00, 0x00},
{ 0x308b, 0x60, 0x00},
};

static const OV3640_REG_TABLE OV3640_video_256X192[]=
{
{ 0x3012, 0x10, 0x00},
{ 0x3020, 0x01, 0x00},
{ 0x3021, 0x1d, 0x00},
{ 0x3022, 0x00, 0x00},
{ 0x3023, 0x06, 0x00},
{ 0x3024, 0x08, 0x00},
{ 0x3025, 0x18, 0x00},
{ 0x3026, 0x03, 0x00},
{ 0x3027, 0x04, 0x00},
{ 0x302a, 0x03, 0x00},
{ 0x302b, 0x10, 0x00},
{ 0x3075, 0x24, 0x00},
{ 0x300d, 0x01, 0x00},
{ 0x30d7, 0x90, 0x00},
{ 0x3069, 0x04, 0x00},
{ 0x3302, 0xef, 0x00},
{ 0x335f, 0x34, 0x00},
{ 0x3360, 0x0c, 0x00},
{ 0x3361, 0x04, 0x00},
{ 0x3362, 0x01, 0x00},
{ 0x3363, 0x08, 0x00},
{ 0x3364, 0xC4, 0x00},
{ 0x3403, 0x42, 0x00},
{ 0x3088, 0x01, 0x00},
{ 0x3089, 0x00, 0x00},
{ 0x308a, 0x00, 0x00},
{ 0x308b, 0xC0, 0x00},
};

static
const OV3640_REG_TABLE OV3640_video_default_fps[]=
{
/* default, night mode enable */
{ 0x300e, 0x32, 0x00},
{ 0x3011, 0x00, 0x00},
{ 0x3010, 0x20, 0x00},
{ 0x3014, 0x0c, 0x00},
{ 0x302e, 0x00, 0x00},
{ 0x302d, 0x00, 0x00},
};

static
const OV3640_REG_TABLE OV3640_video_15fps[]=
{
/* 15fps */
{ 0x300e, 0x39, 0x00},
{ 0x3011, 0x00, 0x00},
{ 0x3010, 0x81, 0x00},
{ 0x3014, 0x04, 0x00},
{ 0x302e, 0x00, 0x00},
{ 0x302d, 0x00, 0x00},
};

static
const OV3640_REG_TABLE OV3640_video_30fps[]=
{
/* 30fps */
{ 0x300e, 0x32, 0x00},
{ 0x3011, 0x00, 0x00},
{ 0x3010, 0x81, 0x00},
{ 0x3014, 0x04, 0x00},
{ 0x302e, 0x00, 0x00},
{ 0x302d, 0x00, 0x00},
};
                
static
const OV3640_REG_TABLE OV3640_video_25fps[]=
{
/* 25fps */
{ 0x300e, 0x36, 0x00},
{ 0x3011, 0x00, 0x00},
{ 0x3010, 0x81, 0x00},
{ 0x3014, 0x04, 0x00},
{ 0x302e, 0x00, 0x00},
{ 0x302d, 0x00, 0x00},
};

static
const OV3640_REG_TABLE OV3640_capture_JPEG[]=
{
//{ 0x3013, 0xf0, 0x00},	/* do it in stop_preview() */
{ 0x3014, 0x04, 0x00},
{ 0x3366, 0x10, 0x00},
{ 0x300e, 0x32, 0x00},
{ 0x300f, 0x21, 0x00},
{ 0x3010, 0x20, 0x00},
//{ 0x3011, 0x01, 0x00},
{ 0x3011, 0x00, 0x00},
{ 0x304c, 0x84, 0x00},
{ 0x3366, 0x10, 0x00},
{ 0x3100, 0x32, 0x00},
{ 0x3304, 0xfc, 0x00},
{ 0x3400, 0x02, 0x00},
{ 0x3404, 0x22, 0x00},
{ 0x3500, 0x00, 0x00},
{ 0x3600, 0xc4, 0x00},
{ 0x3610, 0x00, 0x00},
{ 0x3611, 0x20, 0x00},
};

static
const OV3640_REG_TABLE OV3640_capture_QXGA[]=
{
{ 0x3012, 0x00, 0x00},
{ 0x3020, 0x01, 0x00},
{ 0x3021, 0x1d, 0x00},
{ 0x3022, 0x00, 0x00},
{ 0x3023, 0x0a, 0x00},
{ 0x3024, 0x08, 0x00},
{ 0x3025, 0x18, 0x00},
{ 0x3026, 0x06, 0x00},
{ 0x3027, 0x0c, 0x00},
{ 0x302a, 0x06, 0x00},
{ 0x302b, 0x20, 0x00},
{ 0x3075, 0x44, 0x00},
{ 0x300d, 0x00, 0x00},
{ 0x30d7, 0x10, 0x00},
{ 0x3069, 0x40, 0x00},
{ 0x303e, 0x01, 0x00},
{ 0x303f, 0x80, 0x00},
{ 0x3302, 0xee, 0x00},
{ 0x335f, 0x68, 0x00},
{ 0x3360, 0x18, 0x00},
{ 0x3361, 0x0c, 0x00},
{ 0x3362, 0x68, 0x00},
{ 0x3363, 0x08, 0x00},
{ 0x3364, 0x04, 0x00},
{ 0x3403, 0x42, 0x00},
{ 0x3088, 0x08, 0x00},
{ 0x3089, 0x00, 0x00},
{ 0x308a, 0x06, 0x00},
{ 0x308b, 0x00, 0x00},
};

static
const OV3640_REG_TABLE OV3640_capture_UXGA[]=
{
{ 0x3012, 0x00, 0x00},
{ 0x3020, 0x01, 0x00},
{ 0x3021, 0x1d, 0x00},
{ 0x3022, 0x00, 0x00},
{ 0x3023, 0x0a, 0x00},
{ 0x3024, 0x08, 0x00},
{ 0x3025, 0x18, 0x00},
{ 0x3026, 0x06, 0x00},
{ 0x3027, 0x0c, 0x00},
{ 0x302a, 0x06, 0x00},
{ 0x302b, 0x20, 0x00},
{ 0x3075, 0x44, 0x00},
{ 0x300d, 0x00, 0x00},
{ 0x30d7, 0x10, 0x00},
{ 0x3069, 0x44, 0x00},
{ 0x3302, 0xef, 0x00},
{ 0x335f, 0x68, 0x00},
{ 0x3360, 0x18, 0x00},
{ 0x3361, 0x0c, 0x00},
{ 0x3362, 0x46, 0x00},
{ 0x3363, 0x48, 0x00},
{ 0x3364, 0xb4, 0x00},
{ 0x3403, 0x42, 0x00},
{ 0x3088, 0x06, 0x00},
{ 0x3089, 0x40, 0x00},
{ 0x308a, 0x04, 0x00},
{ 0x308b, 0xb0, 0x00},
};

static
const OV3640_REG_TABLE OV3640_capture_XGA[]=
{
{ 0x3012, 0x00, 0x00},
{ 0x3020, 0x01, 0x00},
{ 0x3021, 0x1d, 0x00},
{ 0x3022, 0x00, 0x00},
{ 0x3023, 0x0a, 0x00},
{ 0x3024, 0x08, 0x00},
{ 0x3025, 0x18, 0x00},
{ 0x3026, 0x06, 0x00},
{ 0x3027, 0x0c, 0x00},
{ 0x302a, 0x06, 0x00},
{ 0x302b, 0x20, 0x00},
{ 0x3075, 0x44, 0x00},
{ 0x300d, 0x00, 0x00},
{ 0x30d7, 0x10, 0x00},
{ 0x3069, 0x44, 0x00},
{ 0x3302, 0xef, 0x00},
{ 0x335f, 0x68, 0x00},
{ 0x3360, 0x18, 0x00},
{ 0x3361, 0x0c, 0x00},
{ 0x3362, 0x34, 0x00},
{ 0x3363, 0x08, 0x00},
{ 0x3364, 0x04, 0x00},
{ 0x3403, 0x42, 0x00},
{ 0x3088, 0x04, 0x00},
{ 0x3089, 0x00, 0x00},
{ 0x308a, 0x03, 0x00},
{ 0x308b, 0x00, 0x00},
};

static
const OV3640_REG_TABLE OV3640_capture_VGA[]=
{
{ 0x3012, 0x00, 0x00},
{ 0x3020, 0x01, 0x00},
{ 0x3021, 0x1d, 0x00},
{ 0x3022, 0x00, 0x00},
{ 0x3023, 0x0a, 0x00},
{ 0x3024, 0x08, 0x00},
{ 0x3025, 0x18, 0x00},
{ 0x3026, 0x06, 0x00},
{ 0x3027, 0x0c, 0x00},
{ 0x302a, 0x06, 0x00},
{ 0x302b, 0x20, 0x00},
{ 0x3075, 0x44, 0x00},
{ 0x300d, 0x00, 0x00},
{ 0x30d7, 0x10, 0x00},
{ 0x3069, 0x44, 0x00},
{ 0x3302, 0xef, 0x00},
{ 0x335f, 0x68, 0x00},
{ 0x3360, 0x18, 0x00},
{ 0x3361, 0x0c, 0x00},
{ 0x3362, 0x12, 0x00},
{ 0x3363, 0x88, 0x00},
{ 0x3364, 0xe4, 0x00},
{ 0x3403, 0x42, 0x00},
{ 0x3088, 0x02, 0x00},
{ 0x3089, 0x80, 0x00},
{ 0x308a, 0x01, 0x00},
{ 0x308b, 0xe0, 0x00},
};

static const OV3640_REG_TABLE OV3640_capture_QVGA[]=
{
{ 0x3012, 0x00, 0x00},
{ 0x3020, 0x01, 0x00},
{ 0x3021, 0x1d, 0x00},
{ 0x3022, 0x00, 0x00},
{ 0x3023, 0x0a, 0x00},
{ 0x3024, 0x08, 0x00},
{ 0x3025, 0x18, 0x00},
{ 0x3026, 0x06, 0x00},
{ 0x3027, 0x0c, 0x00},
{ 0x302a, 0x06, 0x00},
{ 0x302b, 0x20, 0x00},
{ 0x3075, 0x44, 0x00},
{ 0x300d, 0x00, 0x00},
{ 0x30d7, 0x10, 0x00},
{ 0x3069, 0x44, 0x00},
{ 0x3302, 0xef, 0x00},
{ 0x335f, 0x68, 0x00},
{ 0x3360, 0x18, 0x00},
{ 0x3361, 0x0c, 0x00},
{ 0x3362, 0x01, 0x00},
{ 0x3363, 0x48, 0x00},
{ 0x3364, 0xf4, 0x00},
{ 0x3403, 0x42, 0x00},
{ 0x3088, 0x01, 0x00},
{ 0x3089, 0x40, 0x00},
{ 0x308a, 0x00, 0x00},
{ 0x308b, 0xf0, 0x00},
};

static const OV3640_REG_TABLE OV3640_capture_256X192[]=
{
{ 0x3012, 0x00, 0x00},
{ 0x3020, 0x01, 0x00},
{ 0x3021, 0x1d, 0x00},
{ 0x3022, 0x00, 0x00},
{ 0x3023, 0x0a, 0x00},
{ 0x3024, 0x08, 0x00},
{ 0x3025, 0x18, 0x00},
{ 0x3026, 0x06, 0x00},
{ 0x3027, 0x0c, 0x00},
{ 0x302a, 0x06, 0x00},
{ 0x302b, 0x20, 0x00},
{ 0x3075, 0x44, 0x00},
{ 0x300d, 0x00, 0x00},
{ 0x30d7, 0x10, 0x00},
{ 0x3069, 0x44, 0x00},
{ 0x3302, 0xef, 0x00},
{ 0x335f, 0x68, 0x00},
{ 0x3360, 0x18, 0x00},
{ 0x3361, 0x0c, 0x00},
{ 0x3362, 0x01, 0x00},
{ 0x3363, 0x08, 0x00},
{ 0x3364, 0xc4, 0x00},
{ 0x3403, 0x42, 0x00},
{ 0x3088, 0x01, 0x00},
{ 0x3089, 0x00, 0x00},
{ 0x308a, 0x00, 0x00},
{ 0x308b, 0xc0, 0x00},
};

////////////////////////////////////// AEC ////////////////////////////////////

static
const OV3640_REG_TABLE AEC_M1_7EV_tab[] =
{
	{ 0x3018, 0x10, 0x00},
	{ 0x3019, 0x08, 0x00},
	{ 0x301a, 0x21, 0x00},
};

static
const OV3640_REG_TABLE AEC_M1_3EV_tab[] =
{
	{ 0x3018, 0x18, 0x00},
	{ 0x3019, 0x10, 0x00},
	{ 0x301a, 0x31, 0x00},
};

static
const OV3640_REG_TABLE AEC_M1_0EV_tab[] =
{
	{ 0x3018, 0x20, 0x00},
	{ 0x3019, 0x18, 0x00},
	{ 0x301a, 0x41, 0x00},
};

static
const OV3640_REG_TABLE AEC_M0_7EV_tab[] =
{
	{ 0x3018, 0x28, 0x00},
	{ 0x3019, 0x20, 0x00},
	{ 0x301a, 0x51, 0x00},
};

static
const OV3640_REG_TABLE AEC_M0_3EV_tab[] =
{
	{ 0x3018, 0x30, 0x00},
	{ 0x3019, 0x28, 0x00},
	{ 0x301a, 0x61, 0x00},
};

static
const OV3640_REG_TABLE AEC_default_tab[] =
{
	{ 0x3047, 0x00, 0x80},
	{ 0x3018, 0x38, 0x00},
	{ 0x3019, 0x30, 0x00},
	{ 0x301a, 0x61, 0x00},
};

static
const OV3640_REG_TABLE AEC_0_3EV_tab[] =
{
	{ 0x3018, 0x40, 0x00},
	{ 0x3019, 0x38, 0x00},
	{ 0x301a, 0x71, 0x00},
};

static
const OV3640_REG_TABLE AEC_0_7EV_tab[] =
{
	{ 0x3018, 0x48, 0x00},
	{ 0x3019, 0x40, 0x00},
	{ 0x301a, 0x81, 0x00},
};

static
const OV3640_REG_TABLE AEC_1_0EV_tab[] =
{
	{ 0x3018, 0x50, 0x00},
	{ 0x3019, 0x48, 0x00},
	{ 0x301a, 0x91, 0x00},
};

static
const OV3640_REG_TABLE AEC_1_3EV_tab[] =
{
	{ 0x3018, 0x58, 0x00},
	{ 0x3019, 0x50, 0x00},
	{ 0x301a, 0x91, 0x00},
};

static
const OV3640_REG_TABLE AEC_1_7EV_tab[] =
{
	{ 0x3018, 0x60, 0x00},
	{ 0x3019, 0x58, 0x00},
	{ 0x301a, 0xa1, 0x00},
};

/////////////////////////////////// Contrast //////////////////////////////////

static
const OV3640_REG_TABLE contrast_4_tab[] =
{
	{ 0x3302, 0x80, 0x80},
	{ 0x3355, 0x04, 0x04},
	{ 0x335c, 0x30, 0x00},
	{ 0x335d, 0x30, 0x00},
	{ 0x3354, 0x08, 0x04},	////////// TODO: mask is smaller than value??
};

static
const OV3640_REG_TABLE contrast_3_tab[] =
{
	{ 0x3302, 0x80, 0x80},
	{ 0x3355, 0x04, 0x04},
	{ 0x335c, 0x2c, 0x00},
	{ 0x335d, 0x2c, 0x00},
	{ 0x3354, 0x00, 0x04},
};

static
const OV3640_REG_TABLE contrast_2_tab[] =
{
	{ 0x3302, 0x80, 0x80},
	{ 0x3355, 0x04, 0x04},
	{ 0x335c, 0x28, 0x00},
	{ 0x335d, 0x28, 0x00},
	{ 0x3354, 0x00, 0x04},
};

static
const OV3640_REG_TABLE contrast_1_tab[] =
{
	{ 0x3302, 0x80, 0x80},
	{ 0x3355, 0x04, 0x04},
	{ 0x335c, 0x24, 0x00},
	{ 0x335d, 0x24, 0x00},
	{ 0x3354, 0x00, 0x04},
};

static
const OV3640_REG_TABLE contrast_default_tab[] =
{
	{ 0x3302, 0x80, 0x80},
	{ 0x3355, 0x04, 0x04},
	{ 0x335c, 0x20, 0x00},
	{ 0x335d, 0x20, 0x00},
	{ 0x3354, 0x00, 0x04},
};

static
const OV3640_REG_TABLE contrast_M1_tab[] =
{
	{ 0x3302, 0x80, 0x80},
	{ 0x3355, 0x04, 0x04},
	{ 0x335c, 0x1c, 0x00},
	{ 0x335d, 0x1c, 0x00},
	{ 0x3354, 0x00, 0x04},
};

static
const OV3640_REG_TABLE contrast_M2_tab[] =
{
	{ 0x3302, 0x80, 0x80},
	{ 0x3355, 0x04, 0x04},
	{ 0x335c, 0x18, 0x00},
	{ 0x335d, 0x18, 0x00},
	{ 0x3354, 0x00, 0x04},
};

static
const OV3640_REG_TABLE contrast_M3_tab[] =
{
	{ 0x3302, 0x80, 0x80},
	{ 0x3355, 0x04, 0x04},
	{ 0x335c, 0x14, 0x00},
	{ 0x335d, 0x14, 0x00},
	{ 0x3354, 0x00, 0x04},
};

static
const OV3640_REG_TABLE contrast_M4_tab[] =
{
	{ 0x3302, 0x80, 0x80},
	{ 0x3355, 0x04, 0x04},
	{ 0x335c, 0x10, 0x00},
	{ 0x335d, 0x10, 0x00},
	{ 0x3354, 0x00, 0x04},
};

////////////////////////////////// Brightness /////////////////////////////////

static
const OV3640_REG_TABLE brightness_4_tab[] =
{
	{ 0x3302, 0x80, 0x80},
	{ 0x335e, 0x40, 0x00},
	{ 0x3355, 0x04, 0x04},
	{ 0x3354, 0x00, 0x08},
};

static
const OV3640_REG_TABLE brightness_3_tab[] =
{
	{ 0x3302, 0x80, 0x80},
	{ 0x335e, 0x30, 0x00},
	{ 0x3355, 0x04, 0x04},
	{ 0x3354, 0x00, 0x08},
};

static
const OV3640_REG_TABLE brightness_2_tab[] =
{
	{ 0x3302, 0x80, 0x80},
	{ 0x335e, 0x20, 0x00},
	{ 0x3355, 0x04, 0x04},
	{ 0x3354, 0x00, 0x08},
};

static
const OV3640_REG_TABLE brightness_1_tab[] =
{
	{ 0x3302, 0x80, 0x80},
	{ 0x335e, 0x10, 0x00},
	{ 0x3355, 0x04, 0x04},
	{ 0x3354, 0x00, 0x08},
};

static
const OV3640_REG_TABLE brightness_default_tab[] =
{
	{ 0x3302, 0x80, 0x80},
	{ 0x335e, 0x00, 0x00},
	{ 0x3355, 0x04, 0x04},
	{ 0x3354, 0x00, 0x08},
};

static
const OV3640_REG_TABLE brightness_M1_tab[] =
{
	{ 0x3302, 0x80, 0x80},
	{ 0x335e, 0x10, 0x00},
	{ 0x3355, 0x04, 0x04},
	{ 0x3354, 0x08, 0x08},
};

static
const OV3640_REG_TABLE brightness_M2_tab[] =
{
	{ 0x3302, 0x80, 0x80},
	{ 0x335e, 0x20, 0x00},
	{ 0x3355, 0x04, 0x04},
	{ 0x3354, 0x08, 0x08},
};

static
const OV3640_REG_TABLE brightness_M3_tab[] =
{
	{ 0x3302, 0x80, 0x80},
	{ 0x335e, 0x30, 0x00},
	{ 0x3355, 0x04, 0x04},
	{ 0x3354, 0x08, 0x08},
};

static
const OV3640_REG_TABLE brightness_M4_tab[] =
{
	{ 0x3302, 0x80, 0x80},
	{ 0x335e, 0x40, 0x00},
	{ 0x3355, 0x04, 0x04},
	{ 0x3354, 0x08, 0x08},
};

////////////////////////////////////// AWB ////////////////////////////////////

static
const OV3640_REG_TABLE wb_auto_tab[] =
{
	{ 0x332b, 0x00, 0x00},
};

static
const OV3640_REG_TABLE wb_manual_default_tab[] =
{
	{ 0x332b, 0x08, 0x00},
	{ 0x33a7, 0x40, 0x00},
	{ 0x33a8, 0x40, 0x00},
	{ 0x33a9, 0x40, 0x00},
};

static
const OV3640_REG_TABLE wb_manual_sunshine_tab[] =
{
	{ 0x332b, 0x08, 0x00},
	{ 0x33a7, 0x5a, 0x00},
	{ 0x33a8, 0x40, 0x00},
	{ 0x33a9, 0x48, 0x00},
};

static
const OV3640_REG_TABLE wb_manual_cloudy_tab[] =
{
	{ 0x332b, 0x08, 0x00},
	{ 0x33a7, 0x68, 0x00},
	{ 0x33a8, 0x40, 0x00},
	{ 0x33a9, 0x50, 0x00},
};

static
const OV3640_REG_TABLE wb_manual_fluorescence_tab[] =
{
	{ 0x332b, 0x08, 0x00},
	{ 0x33a7, 0x52, 0x00},
	{ 0x33a8, 0x40, 0x00},
	{ 0x33a9, 0x5a, 0x00},
};

static
const OV3640_REG_TABLE wb_manual_incandescence_tab[] =
{
	{ 0x332b, 0x08, 0x00},
	{ 0x33a7, 0x40, 0x00},
	{ 0x33a8, 0x40, 0x00},
	{ 0x33a9, 0x64, 0x00},
};

////////////////////////////////////// Gamma ////////////////////////////////////

static
const OV3640_REG_TABLE gamma_2_tab[] =
{
	{ 0x331b, 0x0e, 0x00},
	{ 0x331c, 0x1a, 0x00},
	{ 0x331d, 0x31, 0x00},
	{ 0x331e, 0x5a, 0x00},
	{ 0x331f, 0x69, 0x00},
	{ 0x3320, 0x75, 0x00},
	{ 0x3321, 0x7e, 0x00},
	{ 0x3322, 0x88, 0x00},
	{ 0x3323, 0x8f, 0x00},
	{ 0x3324, 0x96, 0x00},
	{ 0x3325, 0xa3, 0x00},
	{ 0x3326, 0xaf, 0x00},
	{ 0x3327, 0xc4, 0x00},
	{ 0x3328, 0xd7, 0x00},
	{ 0x3329, 0xe8, 0x00},
	{ 0x332a, 0x20, 0x00},
};

static
const OV3640_REG_TABLE gamma_1_tab[] =
{
	{ 0x331b, 0x0F, 0x00},
	{ 0x331c, 0x1E, 0x00},
	{ 0x331d, 0x3A, 0x00},
	{ 0x331e, 0x62, 0x00},
	{ 0x331f, 0x74, 0x00},
	{ 0x3320, 0x82, 0x00},
	{ 0x3321, 0x8E, 0x00},
	{ 0x3322, 0x98, 0x00},
	{ 0x3323, 0xA0, 0x00},
	{ 0x3324, 0xA7, 0x00},
	{ 0x3325, 0xB5, 0x00},
	{ 0x3326, 0xC0, 0x00},
	{ 0x3327, 0xD2, 0x00},
	{ 0x3328, 0xE1, 0x00},
	{ 0x3329, 0xED, 0x00},
	{ 0x332a, 0x19, 0x00},
};

static
const OV3640_REG_TABLE gamma_default_tab[] =
{
	{ 0x331b, 0x03, 0x00},
	{ 0x331c, 0x12, 0x00},
	{ 0x331d, 0x2c, 0x00},
	{ 0x331e, 0x55, 0x00},
	{ 0x331f, 0x67, 0x00},
	{ 0x3320, 0x78, 0x00},
	{ 0x3321, 0x86, 0x00},
	{ 0x3322, 0x93, 0x00},
	{ 0x3323, 0x9e, 0x00},
	{ 0x3324, 0xa8, 0x00},
	{ 0x3325, 0xb5, 0x00},
	{ 0x3326, 0xc1, 0x00},
	{ 0x3327, 0xd2, 0x00},
	{ 0x3328, 0xe1, 0x00},
	{ 0x3329, 0xec, 0x00},
	{ 0x332a, 0x1a, 0x00},
};

const static u16 ov3640_reg_flash_strobe_on[] =
{
	/* Enable strobe flash: */
	0x307a, 0x83,
	OV3640_REGEND, 0x00	    /* End of list delimiter */
};

const static u16 ov3640_reg_flash_strobe_off[] =
{
	/* Disable strobe flash: */
	0x307a, 0x00,
	OV3640_REGEND, 0x00	    /* End of list delimiter */
};

/*****************************************************************************
 *  Power & Reset
 *****************************************************************************/
static void sensor_power_on(void)
{
	struct sensor_platform_data *pdata;
	pdata = g_client->dev.platform_data;

	if(pdata->power_on)
		pdata->power_on(pdata->id);
}

static void sensor_power_off(void)
{
	struct sensor_platform_data *pdata;
	pdata = g_client->dev.platform_data;

	if(pdata->power_off)
		pdata->power_off(pdata->id);
}

static void sensor_hw_reset(void)
{
	struct sensor_platform_data *pdata;
	pdata = g_client->dev.platform_data;

	if(pdata->reset)
		pdata->reset(pdata->id);
}

static void cam_flash_fire(void)
{
	struct sensor_platform_data *pdata;
	pdata = g_client->dev.platform_data;

	if(pdata->flash_ctrl)
		pdata->flash_ctrl(1);
}

static void cam_flash_torch(void)
{
	struct sensor_platform_data *pdata;
	pdata = g_client->dev.platform_data;

	if(pdata->flash_ctrl)
		pdata->flash_ctrl(0);
}

static
void ov3640hw_power_down( u8 powerMode )
{
	/* OV3640 PWRDWN, 0 = NORMAL, 1=POWER DOWN */
	if( powerMode == CAMERA_POWER_OFF )
		sensor_power_off();
	else
		sensor_power_on();
	msleep(10);
}

static
void __attribute__ ((unused)) ov3640hw_reset(void)
{
	printk(KERN_ERR"[%s:%d]sensor reset\n", __FILE__, __LINE__);
	write_sensor_reg(0x3012, 0x80);	
	mdelay(5);
	return;
}

static void ov3640hw_enter_standby(void)
{
	write_sensor_reg(0x300E, 0x39);
	write_sensor_reg(0x300E, 0xb9);
	write_sensor_reg(0x308D, 0x14);
	write_sensor_reg(0x3086, 0x0F);
}

static void ov3640hw_resume_return(void)
{
	write_sensor_reg(0x3086, 0x08);
	write_sensor_reg(0x308D, 0x14);
	write_sensor_reg(0x300E, 0x32);
}

static
inline void ov3640_write_regs( const OV3640_REG_TABLE *tab, const unsigned len)
{
	u8 val= 0;
	unsigned i= 0;

	for( i= 0; i< len; i++)
	{
		if( tab[ i].mask != 0x00)
		{
			read_sensor_reg( tab[ i].reg, &val);
			val &= ~tab[ i].mask;
			val |= tab[ i].val;
		}
		else
			val= tab[ i].val;

		write_sensor_reg( tab[ i].reg, val);
	}
}

static
int ov3640_set_regs(const u16 *regP)
{
	u32	curReg = 0;
	u8 	regVal = 0;
	int	status = 0;
	/* The list is a register number followed by the value */
	while(regP[curReg << 1] != OV3640_REGEND)
	{
		regVal = regP[(curReg << 1) + 1];
		status = (write_sensor_reg( regP[curReg << 1], regVal) == 0) ?
			0 : -EIO;
		curReg++;
	}

	return status;
}

static
int init_AF( void)
{
	int init_ret= 0;
	int start, bytes;

	start = (AF_init_body_tab[0]<<8) | AF_init_body_tab[1];
	bytes = sizeof(AF_init_body_tab) - 2;
	printk(KERN_INFO "ov3640: download %s\n", AF_init_fw_name);
	printk(KERN_INFO "ov3640: load af firmware [%04x - %04x] %d bytes\n", start, start+bytes-1, bytes);

	ov3640_set_regs(AF_init_head_tab);

	init_ret = i2c_master_send( g_client, ARRAY_AND_SIZE( AF_init_body_tab));
	if( init_ret < 0)
	{
		printk(KERN_ERR"[%s:%d] @_@\n", __FILE__, __LINE__);
		return -EIO;
	}
//	else
//		printk(KERN_ERR"[%s:%d] 2737 - init_ret= %d\n", __FILE__, __LINE__, init_ret);
	
	ov3640_set_regs(AF_init_tail_tab);

	return init_ret;
}

static
void ov3640hw_update_setting(void)
{
	u8 value1=0, value2=0;
	int carry= 0;
	read_sensor_reg( 0x3002, &value1);
	read_sensor_reg( 0x3003, &value2);

	if( value2 > 127)
		carry= 1;
	
	value1*= 2;
	value2*= 2;

	value1+= carry;

	write_sensor_reg( 0x3002, value1);
	write_sensor_reg( 0x3003, value2);

	read_sensor_reg( 0x302d, &value1);
	read_sensor_reg( 0x302e, &value2);

	if( value2 > 127)
		carry= 1;
	else
		carry= 0;
	
	value1*= 2;
	value2*= 2;

	value1+= carry;

	write_sensor_reg( 0x302d, value1);
	write_sensor_reg( 0x302e, value2);
}

static
int ov3640_sensor_reg_init(int format)
{	
//	if( ( format == reg_format) || ( format == reg_size))
//		return 0;

	switch(format)
	{
	case OV3640_SIZE_QVGA:
printk(KERN_ERR"[%s:%d] OV3640_SIZE_QVGA\n", __FILE__, __LINE__);
		ov3640_write_regs( ARRAY_AND_SIZE( OV3640_video_QVGA));
		reg_size= format;
		break;
	case OV3640_SIZE_YGA:
printk(KERN_ERR"[%s:%d] OV3640_SIZE_YGA\n", __FILE__, __LINE__);
		ov3640_write_regs( ARRAY_AND_SIZE( OV3640_video_YGA));
		reg_size= format;
		break;
	case OV3640_SIZE_256X192:
printk(KERN_ERR"[%s:%d] OV3640_SIZE_256X192\n", __FILE__, __LINE__);
		ov3640_write_regs( ARRAY_AND_SIZE( OV3640_video_256X192));
		reg_size= format;
		break;
	case OV3640_SIZE_VGA:
printk(KERN_ERR"[%s:%d] OV3640_SIZE_VGA\n", __FILE__, __LINE__);
		ov3640_write_regs( ARRAY_AND_SIZE( OV3640_video_VGA));
		reg_size= format;
		break;
	case OV3640_SIZE_SQCIF:
printk(KERN_ERR"[%s:%d] OV3640_SIZE_SQCIF\n", __FILE__, __LINE__);
		ov3640_write_regs( ARRAY_AND_SIZE( OV3640_video_SQCIF));
		reg_size= format;
		break;
	case OV3640_SIZE_QCIF:
printk(KERN_ERR"[%s:%d] OV3640_SIZE_QCIF\n", __FILE__, __LINE__);
		ov3640_write_regs( ARRAY_AND_SIZE( OV3640_video_QCIF));
		reg_size= format;
		break;
	case OV3640_SIZE_CIF:
printk(KERN_ERR"[%s:%d] OV3640_SIZE_CIF\n", __FILE__, __LINE__);
		ov3640_write_regs( ARRAY_AND_SIZE( OV3640_video_CIF));
		reg_size= format;
		break;
	case OV3640_SIZE_QXGA:
printk(KERN_ERR"[%s:%d] OV3640_SIZE_QXGA\n", __FILE__, __LINE__);
		ov3640_write_regs( ARRAY_AND_SIZE( OV3640_capture_QXGA));
		reg_size= format;
		break;
	case OV3640_SIZE_UXGA:
printk(KERN_ERR"[%s:%d] OV3640_SIZE_UXGA\n", __FILE__, __LINE__);
		ov3640_write_regs( ARRAY_AND_SIZE( OV3640_capture_UXGA));
		reg_size= format;
		break;
	case OV3640_SIZE_XGA:
printk(KERN_ERR"[%s:%d] OV3640_SIZE_XGA\n", __FILE__, __LINE__);
		ov3640_write_regs( ARRAY_AND_SIZE( OV3640_capture_XGA));
		reg_size= format;
		break;
	case OV3640_SIZE_JPEG_VGA:
printk(KERN_ERR"[%s:%d] OV3640_SIZE_JPEG_VGA\n", __FILE__, __LINE__);
		ov3640_write_regs( ARRAY_AND_SIZE( OV3640_capture_VGA));
		reg_size= format;
		break;
	case OV3640_SIZE_JPEG_QVGA:
printk(KERN_ERR"[%s:%d] OV3640_SIZE_JPEG_QVGA\n", __FILE__, __LINE__);
		ov3640_write_regs( ARRAY_AND_SIZE( OV3640_capture_QVGA));
		reg_size= format;
		break;
	case OV3640_SIZE_JPEG_256X192:
printk(KERN_ERR"[%s:%d] OV3640_SIZE_JPEG_256X192\n", __FILE__, __LINE__);
		ov3640_write_regs( ARRAY_AND_SIZE( OV3640_capture_256X192));
		reg_size= format;
		break;

	case OV3640_FORMAT_JPG:
printk(KERN_ERR"[%s:%d] OV3640_FORMAT_JPG\n", __FILE__, __LINE__);
		ov3640_write_regs( ARRAY_AND_SIZE( OV3640_capture_JPEG));
		//ov3640hw_update_setting();
		reg_format= format;
		break;
	case OV3640_FORMAT_YUV_422:
printk(KERN_ERR"[%s:%d] OV3640_FORMAT_YUV_422\n", __FILE__, __LINE__);
		ov3640_write_regs( ARRAY_AND_SIZE( OV3640_video_YUV422));
		reg_format= format;
		break;

	case AF_SETTING:
		init_AF();
		break;
	default:
		printk("No such init table\n");
		return -1;
	}
	return 0;
}

static
int ov3640_sensorID_detect(void)
{
	u8 val_h= 0, val_l= 0;
	read_sensor_reg(PID_ADRRH, &val_h);
	read_sensor_reg(PID_ADRRL, &val_l);
	
	printk(KERN_ERR"[%s:%d] val_h= 0x%02X, val_l= 0x%02X\n", __FILE__, __LINE__, val_h, val_l);
	
	if( (val_h == PID_VERH ))
		return 0;
	else
		return -1;
}

static
int ov3640hw_read_sensor_reg( const u16 subAddress, u8 *bufP )
{
	return read_sensor_reg(subAddress, bufP);
}

static
int ov3640hw_write_sensor_reg( const u16 subAddress, u8 *bufP )
{
	return write_sensor_reg(subAddress, *bufP);
}

static void ov3640hw_flash_torch(void)
{
	down(&ov3640_flash_lock);
	if(!flashled_on){
		//printk(KERN_INFO "ov3640: flash led on\n");
		cam_flash_torch();
		flashled_fire = 0;
		ov3640_set_regs( ov3640_reg_flash_strobe_on );
		flashled_on = 1;

		/* start 5s timeout to off flash */
		queue_delayed_work(flash_wqueue, &flash_work, msecs_to_jiffies(5000));
	}
	else
		printk(KERN_ERR "ov3640: error, flash led already was on\n");
	up(&ov3640_flash_lock);
}

static void ov3640hw_flash_fire(void)
{
	struct timeval now;
        unsigned int ms;

	down(&ov3640_flash_lock);
	if(flashled_on){
		do_gettimeofday(&now);
		ms = 1000 * (now.tv_sec - flashled_fire_time.tv_sec);
		ms += (now.tv_usec - flashled_fire_time.tv_usec)/1000;
		if(ms<=5000) {
			printk(KERN_ERR "ov3640: warning, flash led off %d ms, must be off within 5s.\n", ms);
			up(&ov3640_flash_lock);
			return;
		}

		printk(KERN_INFO "ov3640: flash led fire\n");
		cam_flash_fire();
		flashled_fire = 1;
	}
	else
		printk(KERN_ERR "ov3640: fire error, flash led is not on\n");
	up(&ov3640_flash_lock);
}

static void ov3640hw_flash_off(void)
{
	down(&ov3640_flash_lock);
	if(flashled_on){
		cam_flash_torch();
		/* save flash fire off time */
		if(flashled_fire)
			do_gettimeofday(&flashled_fire_time);
		flashled_fire = 0;

		ov3640_set_regs( ov3640_reg_flash_strobe_off);
		flashled_on = 0;
		printk(KERN_INFO "ov3640: flash led off\n");
	}
	up(&ov3640_flash_lock);
}

static void ov3640hw_flash_work_off(void)
{
	cancel_delayed_work(&flash_work);
	if(flashled_on){
		ov3640hw_flash_off();
	}
}

static void ov3640_flash_work_handler(struct work_struct *work)
{
	printk(KERN_INFO "ov3640: flashing timeout to be off\n");
	ov3640hw_flash_off();
}

/*****************************************************************************
 *  Settings
 *****************************************************************************/
static
int ov3640hw_version_revision(u8 * pCmRevision, u8 *pSensorRevision)
{
	read_sensor_reg( PID_ADRRH, pCmRevision );
	read_sensor_reg( PID_ADRRL, pSensorRevision );
	return 0;
}


/*****************************************************************************
 *  Viewfinder, still
 *****************************************************************************/
static int ov3640hw_af_command(int cmd)
{
#if 0
	u8 reg = 0;
	int loop;

	/* Send auto focus firmware command to AF module */
	write_sensor_reg(0x3f00, cmd);

	loop = 100;
	do {
		read_sensor_reg(0x3f00, &reg);
		if(reg==0)
			return 0;
		msleep(10);
	} while(loop--);

	printk(KERN_ERR "ov3640: af command timeout, cmd=0x%02x, reg(0x%02x)\n", cmd, reg);
	return -ETIME;
#else
	/* Send auto focus firmware command to AF module */
	write_sensor_reg(0x3f00, cmd);
	return 0;
#endif
}

static int ov3640hw_af_finish(void)
{
	u8 state = 0;
	int loop;

	/* Send finish command for AF module */
	ov3640hw_af_command(0x08);

	/* Read out state register value state_current. 
	  If(state_current == STATE_INF)finish OK and focus on infinity */
	loop = 50;
	do {
		msleep(10);

		read_sensor_reg(0x3f07, &state);
		//printk(KERN_INFO "ov3640: af state(0x%02x)\n", state);

		if(state == 0x00)
		{
			ov3640_single_af_enabled = 0;
			printk(KERN_INFO "ov3640: af finished, ok\n");
			return 0;
		}
		else if(state == 0x02)
		{
			/* Send finish command again */
			printk(KERN_INFO "ov3640: af not finished, send finish command again! state(0x%02x)\n", state);
			ov3640hw_af_command(0x08);
		}
	} while(loop--);

	printk(KERN_ERR "ov3640: af finish timeout, state(0x%02x)\n", state);
	return -ETIME;
}

static int ov3640hw_single_af(void)
{
	u8 state = 0;
	int loop;
	int ret = 0;

	if(ov3640_single_af_enabled)
		ov3640hw_af_finish();

	read_sensor_reg(0x3f07, &state);

	if(state != 0x00)
	{
		printk(KERN_ERR "ov3640: af state(0x%02x) is not S_IDEL\n", state);
		ov3640hw_af_finish();
		return -EINVAL;
	}

	if(flashled_flag == 1){
		ov3640hw_flash_torch();
	}

	/* Send single focus command */
	ov3640hw_af_command(0x03);

	loop = 200;
	do {
		msleep(10);

		read_sensor_reg(0x3f07, &state);
		//printk(KERN_INFO "ov3640: af state(0x%02x)\n", state);

		if(state == 0x01)
			continue;
		else if(state == 0x02)
		{
			printk(KERN_INFO "ov3640: single af ok.\n");
			ov3640_single_af_enabled = 1;
			return 0;
		}
	} while(loop--);
	ret = -ETIME;
	printk(KERN_ERR "ov3640: single af timeout, af state(0x%02x)\n", state);

	if(flashled_on)
		ov3640hw_flash_work_off();

	ov3640hw_af_finish();
	return ret;
}

static int ov3640_sensor_stop_preview(void)
{
	u8 reg = 0;

	if(flashled_flag == 1){
		if(flashled_on) {
			ov3640hw_flash_fire();
			msleep(700);
		}
		else {
			ov3640hw_flash_torch();
			ov3640hw_flash_fire();
			msleep(1000);
		}
	}

	/* Stop AG/AE */
	read_sensor_reg(0x3013,&reg);
	reg &= 0xfa;
	write_sensor_reg(0x3013,reg);

	printk("ov3640_sensor_stop_preview\n");
	return 0;
}

static int ov3640_sensor_prepare_snap(void)
{
	ov3640hw_update_setting();

	printk("ov3640_sensor_prepare_snap\n");
	return 0;
}

static int ov3640_sensor_stop_snap(void)
{
	if(flashled_on) {
		ov3640hw_flash_work_off();
	}

	if(ov3640_single_af_enabled)
		ov3640hw_af_finish();

	printk("ov3640_sensor_stop_snap\n");
	return 0;
}

static int ov3640_sensor_back2preview(void)
{
	u8 reg = 0;

	/* Start AG/AE */
	read_sensor_reg(0x3013,&reg);
	reg |= 0x5;
	write_sensor_reg(0x3013,reg);

	printk("ov3640_sensor_back2preview\n");
	return 0;
}

/*****************************************************************************
 *  Format
 *****************************************************************************/

static int ov3640hw_set_framerate(int fps)
{
	if( fps<15 ) {
		ov3640_write_regs( ARRAY_AND_SIZE(OV3640_video_default_fps) );
		printk(KERN_INFO "ov3640: set video default fps.\n");
	}
	else if( fps==30 ) {
		ov3640_write_regs( ARRAY_AND_SIZE(OV3640_video_30fps) );
		printk(KERN_INFO "ov3640: set video 30fps.\n");
	}
	else if( fps==25 ) {
		ov3640_write_regs( ARRAY_AND_SIZE(OV3640_video_25fps) );
		printk(KERN_INFO "ov3640: set video 25fps.\n");
	}
	else if( fps==15 ) {
		ov3640_write_regs( ARRAY_AND_SIZE(OV3640_video_15fps) );
		printk(KERN_INFO "ov3640: set video 15fps.\n");
	}
	else {
		printk(KERN_INFO "ov3640: video invalid fps %d.\n", fps);
		return -EINVAL;
	}
	return 0;
}

static
int ov3640hw_set_format( u32 captureSizeFormat, u32 colorFormat, u32 mode)
{
	if(G_ov3640_mode_change!=mode)
	{
		if(mode == CAMERA_MODE_STILL)
			ov3640_sensor_stop_preview();
		else
			ov3640_sensor_stop_snap();
	}

#if 1
	ov3640_sensor_reg_init( colorFormat);
	ov3640_sensor_reg_init( captureSizeFormat);
#else
	ov3640_sensor_reg_init( captureSizeFormat);
	ov3640_sensor_reg_init( colorFormat);
#endif

	if( colorFormat == OV3640_FORMAT_YUV_422)
	{
		//printk("wait focus-----------\n");
		//ov3640_sensor_reg_init( AF_SETTING);//af function
		//printk("wait focus done-----------\n");
		ov3640hw_set_framerate(ov3640set_fps);
	}
/*
	{
	u8 value;
	int addr;
	addr= 0x3000;
	read_sensor_reg( addr, &value);
	printk( KERN_INFO"addr= 0x%04X, val= 0x%02X\n", addr, value);
	addr++;
	read_sensor_reg( addr, &value);
	printk( KERN_INFO"addr= 0x%04X, val= 0x%02X\n", addr, value);
	addr++;
	read_sensor_reg( addr, &value);
	printk( KERN_INFO"addr= 0x%04X, val= 0x%02X\n", addr, value);
	addr++;
	read_sensor_reg( addr, &value);
	printk( KERN_INFO"addr= 0x%04X, val= 0x%02X\n", addr, value);
	}
*/
	if(G_ov3640_mode_change!=mode)
	{
		if(mode == CAMERA_MODE_STILL)
			ov3640_sensor_prepare_snap();
		else
			ov3640_sensor_back2preview();
		G_ov3640_mode_change = mode;
	}

	return 0;
}

/*
int mode: sensor output mode, MODE_STILL for QXGA, MODE_VIDEO for XGA;
int iZoomLevel: zoom level, 100-1X, 110-1.1X, ... etc;
unsigned int iWidthOut: output width of the sensor;
unsigned int iHeightOut: output height of the sensor;
*/
static int ov3640hw_set_resolution_zoom(int mode, int iZoomLevel, 
				   unsigned int iWidthOut, unsigned int iHeightOut)
{
	unsigned int iWidthIn = 0;
	unsigned int iHeightIn = 0;
	unsigned int iCroppedX = 0;
	unsigned int iCroppedY = 0;
	unsigned int iTotalX = 0;
	unsigned int iTotalY = 0;
	unsigned int iOffsetX = 0;
	unsigned int iOffsetY = 0;

	unsigned char bTempValue = 0x00;

	if(mode == CAMERA_MODE_STILL)	//QXGA
	{
		iWidthIn = 2072;
		iHeightIn = 1548;
		iTotalX = 2642;
		iTotalY = 1568;

		iCroppedX = (unsigned int)(iWidthIn*100/iZoomLevel); 
		iCroppedY = (unsigned int)(iHeightIn*100/iZoomLevel);

		if(((iCroppedX-8) < iWidthOut) || ((iCroppedY-4) < iHeightOut))
		{
			iCroppedX = iWidthOut + 8;
			iCroppedY = iHeightOut + 4;
		}

		iOffsetX = (iTotalX - iCroppedX)/2;
		iOffsetY = (iTotalY - iCroppedY)/2;
	}
	else //XGA
	{
		iWidthIn = 1036;
		iHeightIn = 772;
		iTotalX = 1321;
		iTotalY = 784;

		iCroppedX = (unsigned int)(iWidthIn*100/iZoomLevel);
		iCroppedY = (unsigned int)(iHeightIn*100/iZoomLevel);

		if(((iCroppedX-8) < iWidthOut) || ((iCroppedY-4) < iHeightOut))
		{
			iCroppedX = iWidthOut + 8;
			iCroppedY = iHeightOut + 4;
		}

		iOffsetX = (iTotalX - iCroppedX);
		iOffsetY = (iTotalY - iCroppedY)/2;
	}

	//write_sensor_reg(0x3602, 0x32);	//test
	//delay(100);

	write_sensor_reg(0x3302, 0xef);

	bTempValue = (iCroppedY >> 4) & 0xf0; 
	bTempValue |= (iCroppedX >> 8);
	write_sensor_reg(0x335f, bTempValue);
	write_sensor_reg(0x3360, iCroppedX & 0xff);
	write_sensor_reg(0x3361, iCroppedY & 0xff);

	bTempValue = (iHeightOut >> 4) & 0xf0;
	bTempValue |= (iWidthOut >> 8);
	write_sensor_reg(0x3362, bTempValue);
	write_sensor_reg(0x3363, (iWidthOut & 0xff)+8);
	write_sensor_reg(0x3364, (iHeightOut & 0xff)+4);

	write_sensor_reg(0x3403, 0x42);

	write_sensor_reg(0x3088, (iWidthOut >> 8) & 0xff);
	write_sensor_reg(0x3089, iWidthOut & 0xff);
	write_sensor_reg(0x308a, (iHeightOut >> 8) & 0xff);
	write_sensor_reg(0x308b, iHeightOut & 0xff);

	write_sensor_reg(0x3020, (iOffsetX >> 8) & 0xff);
	write_sensor_reg(0x3021, iOffsetX & 0xff);
	write_sensor_reg(0x3022, (iOffsetY >> 8) & 0xff);
	write_sensor_reg(0x3023, iOffsetY & 0xff);

	if(mode != CAMERA_MODE_STILL)	//XGA
	{
		iCroppedX *= 2;
	}

	write_sensor_reg(0x3024, (iCroppedX >> 8) & 0xff);
	write_sensor_reg(0x3025, iCroppedX & 0xff);
	write_sensor_reg(0x3026, (iCroppedY >> 8) & 0xff);
	write_sensor_reg(0x3027, iCroppedY & 0xff);

	//write_sensor_reg(0x3602, 0x22);	//test

 	/* may need change clock on some size & zoom */
	//write_sensor_reg(0x304c, 0x84);
	return 0;
}

static
int ov3640hw_set_gamma( u32 value)
{
	switch( value)
	{
	case 0:
		ov3640_write_regs( ARRAY_AND_SIZE( gamma_default_tab));
		break;
	case 1:
		ov3640_write_regs( ARRAY_AND_SIZE( gamma_1_tab));
		break;
	case 2:
		ov3640_write_regs( ARRAY_AND_SIZE( gamma_2_tab));
		break;
	}
	
	return 0;
}

static
int ov3640hw_set_brightness( u32 value)
{
	switch( value)
	{
	case 0:
		ov3640_write_regs( ARRAY_AND_SIZE( brightness_default_tab));
		break;
	case 1:
		ov3640_write_regs( ARRAY_AND_SIZE( brightness_M4_tab));
		break;
	case 2:
		ov3640_write_regs( ARRAY_AND_SIZE( brightness_M3_tab));
		break;
	case 3:
		ov3640_write_regs( ARRAY_AND_SIZE( brightness_M2_tab));
		break;
	case 4:
		ov3640_write_regs( ARRAY_AND_SIZE( brightness_M1_tab));
		break;
	case 5:
		ov3640_write_regs( ARRAY_AND_SIZE( brightness_1_tab));
		break;
	case 6:
		ov3640_write_regs( ARRAY_AND_SIZE( brightness_2_tab));
		break;
	case 7:
		ov3640_write_regs( ARRAY_AND_SIZE( brightness_3_tab));
		break;
	case 8:
		ov3640_write_regs( ARRAY_AND_SIZE( brightness_4_tab));
		break;
	}
	
	return 0;
}

static
int ov3640hw_set_contrast( u32 value)
{
	switch( value)
	{
	case 0:
		ov3640_write_regs( ARRAY_AND_SIZE( contrast_default_tab));
		break;
	case 1:
		ov3640_write_regs( ARRAY_AND_SIZE( contrast_M4_tab));
		break;
	case 2:
		ov3640_write_regs( ARRAY_AND_SIZE( contrast_M3_tab));
		break;
	case 3:
		ov3640_write_regs( ARRAY_AND_SIZE( contrast_M2_tab));
		break;
	case 4:
		ov3640_write_regs( ARRAY_AND_SIZE( contrast_M1_tab));
		break;
	case 5:
		ov3640_write_regs( ARRAY_AND_SIZE( contrast_1_tab));
		break;
	case 6:
		ov3640_write_regs( ARRAY_AND_SIZE( contrast_2_tab));
		break;
	case 7:
		ov3640_write_regs( ARRAY_AND_SIZE( contrast_3_tab));
		break;
	case 8:
		ov3640_write_regs( ARRAY_AND_SIZE( contrast_4_tab));
		break;
	}

	return 0;
}

static
int ov3640hw_set_exposure( u32 value)
{
	switch( value)
	{
	case 0:				// 0 - auto
		ov3640_write_regs( ARRAY_AND_SIZE( AEC_default_tab));
		break;
	case 1:				// -1.7
		ov3640_write_regs( ARRAY_AND_SIZE( AEC_M1_7EV_tab));
		break;
	case 2:				// -1.3
		ov3640_write_regs( ARRAY_AND_SIZE( AEC_M1_3EV_tab));
		break;
	case 3:				// -1
		ov3640_write_regs( ARRAY_AND_SIZE( AEC_M1_0EV_tab));
		break;
	case 4:				// -0.7
		ov3640_write_regs( ARRAY_AND_SIZE( AEC_M0_7EV_tab));
		break;
	case 5:				// -0.3
		ov3640_write_regs( ARRAY_AND_SIZE( AEC_M0_3EV_tab));
		break;
	case 6:				// 0.3
		ov3640_write_regs( ARRAY_AND_SIZE( AEC_0_3EV_tab));
		break;
	case 7:				// 0.7
		ov3640_write_regs( ARRAY_AND_SIZE( AEC_0_7EV_tab));
		break;
	case 8:				// 1.0
		ov3640_write_regs( ARRAY_AND_SIZE( AEC_1_0EV_tab));
		break;
	case 9:				// 1.3
		ov3640_write_regs( ARRAY_AND_SIZE( AEC_1_3EV_tab));
		break;
	case 10:			// 1.7
		ov3640_write_regs( ARRAY_AND_SIZE( AEC_1_7EV_tab));
		break;
	}

	return 0;
}

/*****************************************************************************
 * Auto White Balance
 *****************************************************************************/
static
int ov3640hw_set_white_balance( u32 value)
{
	switch( value)
	{
	case 0:				// auto
		ov3640_write_regs( ARRAY_AND_SIZE( wb_auto_tab));
		break;
	case 1:				// default
		ov3640_write_regs( ARRAY_AND_SIZE( wb_manual_default_tab));
		break;
	case 2:				// sunshine
		ov3640_write_regs( ARRAY_AND_SIZE( wb_manual_sunshine_tab));
		break;
	case 3:				// cloudy
		ov3640_write_regs( ARRAY_AND_SIZE( wb_manual_cloudy_tab));
		break;
	case 4:				// Fluorescence
		ov3640_write_regs( ARRAY_AND_SIZE( wb_manual_fluorescence_tab));
		break;
	case 5:				// Incandescence
		ov3640_write_regs( ARRAY_AND_SIZE( wb_manual_incandescence_tab));
		break;
	}
	
	return 0;
}

static int ov3640hw_sensor_init( p_camera_context_t cam_ctx )
{
	G_ov3640_mode_change = CAMERA_MODE_VIDEO;
	g_ov3640_size = -1;
	g_ov3640_format = -1;

	flashled_flag = 0;
	flashled_on = 0;
	flashled_fire = 0;
	flashled_fire_time.tv_sec = 0;
	flashled_fire_time.tv_usec = 0;

	ov3640_single_af_enabled = 0;
	ov3640set_fps = 0;

	printk(KERN_INFO "ov3640: sensor init\n");
	ov3640_write_regs( ARRAY_AND_SIZE( OV3640_init));
	//printk(KERN_INFO "ov3640: video YUV422\n");
	ov3640_write_regs( ARRAY_AND_SIZE( OV3640_video_YUV422));
	//printk(KERN_INFO "ov3640: video 480x360\n");
	ov3640_write_regs( ARRAY_AND_SIZE( OV3640_video_YGA));

	ov3640_sensor_reg_init( AF_SETTING);//af function

	return 0;
}

/*****************************************************************************
 * OV3640 Functions
 *****************************************************************************/
static
int ov3640_init( p_camera_context_t camera_context )
{
	u8 cm_rev, cm_pid;
	u32 timeout;
	int status;

	printk(KERN_ERR"ov3640_init\n");

	af_flag= reg_format= reg_size= 0;

	/* provide informat about the capabilities of the sensor */
	camera_context->sensor_status.caps |= SENSOR_CAP_MANUAL_CONTRAST |
		SENSOR_CAP_MANUAL_WHITEBALANCE | /* SENSOR_CAP_MANUAL_BRIGHTNESS | */
		SENSOR_CAP_MANUAL_EXPOSURE | SENSOR_CAP_MANUAL_ZOOM;

	/* Configure CI according to OV3640's hardware
	 * master parallel with 8 data pins
	 */
	ci_set_mode(CI_MODE_MP, CI_DATA_WIDTH8);
	
	/* enable pixel clock(sensor will provide pclock)
	 * and master clock = 26MHZ
	 */
	/* software work around*/
	ci_set_clock(1, 1, CICLK);

	/* data sample on rising and h,vsync active high */
	ci_set_polarity(0, 0, 0);

	/* fifo control */
	ci_set_fifo(0, CI_FIFO_THL_32, 1, 1);

	/* set black level */
	ci_cgu_set_black_level(0);

	/* CGU Mux Select */
	ci_cgu_set_addr_mux_select(CI_CGU_MUX_0_TO_7);


	/* OV3640 Power on sequence
	 * Take out of Power down mode (GPIO_176), PWRDWN=1, NORMAL=0
	 * Assert Reset
	 * Delay
	 * Remove reset
	 * Delay
	 */
	ov3640hw_power_down(CAMERA_POWER_FULL);
	ov3640hw_resume_return();
	//ov3640hw_reset();

	/* read out version */
	timeout = 50;
	do
	{
		cm_pid = cm_rev = 0;
		status = ov3640hw_version_revision(&cm_pid, &cm_rev);

		/* Check to make sure we are working with an OV3640 */
		if( cm_pid != PID_VERH  )
		{
			printk("ov3640: pid(%x) is not correct, try again!\n", cm_pid);
			ov3640hw_power_down( CAMERA_POWER_OFF );
			ov3640hw_power_down( CAMERA_POWER_FULL );
			sensor_hw_reset();
			mdelay(1);
		}
		if (--timeout == 0) {
			return -EIO;
		}
	}
	while( cm_pid != PID_VERH );

	/* initialize default camera parameters */
	/* default preview width/height */
	camera_context->mode[CAMERA_MODE_VIDEO].width = OV3640_YGA_WIDTH;
	camera_context->mode[CAMERA_MODE_VIDEO].height = OV3640_YGA_HEIGHT;
	/* default preview framerate */
	camera_context->mode[CAMERA_MODE_VIDEO].timeperframe_numerator = 15;
	camera_context->mode[CAMERA_MODE_VIDEO].timeperframe_denominator = 2;
	/* default still capture width/height */
	camera_context->mode[CAMERA_MODE_STILL].width = OV3640_QXGA_WIDTH;
	camera_context->mode[CAMERA_MODE_STILL].height = OV3640_QXGA_HEIGHT;
	camera_context->mode[CAMERA_MODE_STILL].input_format = V4L2_PIX_FMT_JPEG;
	camera_context->mode[CAMERA_MODE_STILL].output_format = V4L2_PIX_FMT_JPEG;

	ov3640hw_sensor_init(camera_context);

	return 0;
}

static
int ov3640_deinit(p_camera_context_t camera_context)
{
	/* off flash if it is on */
	ov3640hw_flash_work_off();

	/* do af finish if the af is not finished */
	if(ov3640_single_af_enabled)
		ov3640hw_af_finish();

	/* power off the external module */
	ov3640hw_enter_standby();
	ov3640hw_power_down( CAMERA_POWER_OFF );

	return 0;
}

static
int ov3640_sleep(  p_camera_context_t camera_context )
{
	return ov3640_deinit( camera_context );
}

static
int ov3640_wake(  p_camera_context_t camera_context )
{
	return ov3640_init( camera_context );
}

static
int ov3640_set_capture_format(  p_camera_context_t camera_context )
{
	CI_MP_TIMING timing;
	int status;
	u32 ovSizeFormat, ovFormat;
	int capture_width, capture_height;
	int zoom, zoom_width, zoom_height;

	switch(camera_context->capture_input_format) 
	{
		case V4L2_PIX_FMT_YUV422P:
//			printk(KERN_ERR"[%s:%d] output format yuv422\n", __FILE__, __LINE__);
			ovFormat = OV3640_FORMAT_YUV_422;
			break;
			
		case V4L2_PIX_FMT_SRGGB10:
//			printk(KERN_ERR"[%s:%d] output format raw10\n", __FILE__, __LINE__);
			ovFormat = OV3640_FORMAT_RAW10;		
			break;
			
		case V4L2_PIX_FMT_SRGGB8:
//			printk(KERN_ERR"[%s:%d] output format raw8\n", __FILE__, __LINE__);
			ovFormat = OV3640_FORMAT_RAW8;		
			break;

			
		case V4L2_PIX_FMT_JPEG:
//			printk(KERN_ERR"[%s:%d] output format jpeg\n", __FILE__, __LINE__);
			ovFormat = OV3640_FORMAT_JPG;
			break;
		
		case V4L2_PIX_FMT_RGB565X:
//			printk(KERN_ERR"[%s:%d] output format rgb565\n", __FILE__, __LINE__);
			 ovFormat = OV3640_FORMAT_RGB565;
		
			 break;
		default:
//			printk(KERN_ERR"[%s:%d] output format default yuv422\n", __FILE__, __LINE__);
			ovFormat = OV3640_FORMAT_YUV_422;
			break;
	}

	capture_width	 = camera_context->capture_input_width;
	capture_height = camera_context->capture_input_height;

	zoom = camera_context->sensor_status.digital_zoom_value;
	if(zoom<100)
		zoom = 100;

	printk(KERN_INFO "ov3640: capture %d:%d zoom %d\n", capture_width, capture_height, zoom);

	if( ovFormat == OV3640_FORMAT_JPG )
	{	/* JPEG */
		zoom_width = OV3640_QXGA_WIDTH*100/zoom;
		zoom_height = OV3640_QXGA_HEIGHT*100/zoom;

		if(capture_width>zoom_width)
			capture_width = zoom_width;
		if(capture_height>zoom_height)
			capture_height = zoom_height;

		if( (capture_width >= OV3640_QXGA_WIDTH) && (capture_height >= OV3640_QXGA_HEIGHT) ) {
			ovSizeFormat = OV3640_SIZE_QXGA;
			capture_width = OV3640_QXGA_WIDTH;
			capture_height = OV3640_QXGA_HEIGHT;
		}
		else if( (capture_width >= OV3640_UXGA_WIDTH) && (capture_height >= OV3640_UXGA_HEIGHT) ) {
			ovSizeFormat = OV3640_SIZE_UXGA;
			capture_width = OV3640_UXGA_WIDTH;
			capture_height = OV3640_UXGA_HEIGHT;
		}
		else if( (capture_width >= OV3640_XGA_WIDTH) && (capture_height >= OV3640_XGA_HEIGHT) ) {
			ovSizeFormat = OV3640_SIZE_XGA;
			capture_width = OV3640_XGA_WIDTH;
			capture_height = OV3640_XGA_HEIGHT;
		}
		else if( (capture_width >= OV3640_VGA_WIDTH) && (capture_height >= OV3640_VGA_HEIGHT) ) {
			ovSizeFormat = OV3640_SIZE_JPEG_VGA;
			capture_width = OV3640_VGA_WIDTH;
			capture_height = OV3640_VGA_HEIGHT;
		}
		else if( (capture_width >= 320) && (capture_height>= 240) ) {
			ovSizeFormat = OV3640_SIZE_JPEG_QVGA;
			capture_width = 320;
			capture_height = 240;
		}
		else if( (capture_width >= 256) && (capture_height>= 192) ) {
			ovSizeFormat = OV3640_SIZE_JPEG_256X192;
			capture_width = 256;
			capture_height = 192;
		}
		else {
			ovSizeFormat = OV3640_SIZE_JPEG_VGA;
			capture_width = OV3640_VGA_WIDTH;
			capture_height = OV3640_VGA_HEIGHT;
			printk(KERN_ERR "ov3640: can not capture %d:%d at zoom %d\n", capture_width, capture_height, zoom);
		}

		printk(KERN_DEBUG "ov3640: jpeg capture %d:%d, zoom=%d, zoomed %d:%d\n",
			capture_width, capture_height, zoom, zoom_width, zoom_height);

		camera_context->capture_input_width = capture_width;
		camera_context->capture_input_height = capture_height;

	} else /* YUV422 */
	if( ( camera_context->capture_input_width <= OV3640_SQCIF_WIDTH) 
			&& (camera_context->capture_input_height <= OV3640_SQCIF_HEIGHT)) {
		ovSizeFormat = OV3640_SIZE_SQCIF;
	} else if( ( camera_context->capture_input_width <= OV3640_QCIF_WIDTH)
			&& (camera_context->capture_input_height <= OV3640_QCIF_HEIGHT)) {
		ovSizeFormat = OV3640_SIZE_QCIF;
	} else if( ( camera_context->capture_input_width <= 256)
			&& (camera_context->capture_input_height <= 192)) {
		ovSizeFormat = OV3640_SIZE_256X192;
	} else if( ( camera_context->capture_input_width <= OV3640_QVGA_WIDTH)
			&& (camera_context->capture_input_height <= OV3640_QVGA_HEIGHT)) {
		ovSizeFormat = OV3640_SIZE_QVGA;
	} else if( ( camera_context->capture_input_width <= OV3640_CIF_WIDTH)
			&& (camera_context->capture_input_height <= OV3640_CIF_HEIGHT)) {
		ovSizeFormat = OV3640_SIZE_CIF;
	} else if( ( camera_context->capture_input_width <= OV3640_YGA_WIDTH)
			&& (camera_context->capture_input_height <= OV3640_YGA_HEIGHT)) {
		ovSizeFormat = OV3640_SIZE_YGA;
	} else if( ( camera_context->capture_input_width <= OV3640_VGA_WIDTH)
			&& (camera_context->capture_input_height <= OV3640_VGA_HEIGHT)) {
		ovSizeFormat = OV3640_SIZE_VGA;
	} else {
		ovSizeFormat = OV3640_SIZE_VGA;
	}

	if( (ovSizeFormat!=g_ov3640_size)||(ovFormat!=g_ov3640_format)||(camera_context->capture_mode!=G_ov3640_mode_change) ) {
		status = ov3640hw_set_format( ovSizeFormat, ovFormat, camera_context->capture_mode);
		g_ov3640_size = ovSizeFormat;
		g_ov3640_format = ovFormat;
	}

	/* set digital zoom after size changed */
	ov3640hw_set_resolution_zoom(camera_context->capture_mode, zoom, capture_width, capture_height);

	if( ovFormat == OV3640_FORMAT_YUV_422)
		/* Update Focus Window, When Host update the sensor’s registers
		   about image resolution, this command must be sent. */
		ov3640hw_af_command(0x12);

	timing.BFW = 0x00;
	timing.BLW = 0x00;

	ci_configure_mp(camera_context->capture_input_width-1,
			camera_context->capture_input_height-1, &timing);
	return 0;
}

static
int ov3640_start_capture( p_camera_context_t camera_context,
		unsigned int frames)
{
	(void) camera_context;
	if(camera_context->capture_mode == CAMERA_MODE_STILL)
		/* wait about 2 frame before capture */
		camera_context->capture_drop_frames = 2;
	else
		camera_context->capture_drop_frames = 0;

	return 0;
}

static
int ov3640_stop_capture(  p_camera_context_t camera_context )
{
	(void) camera_context;
	if(flashled_on && camera_context->capture_mode == CAMERA_MODE_STILL) {
		ov3640hw_flash_work_off();
	}
	return 0;
}

static
int ov3640_set_power_mode(p_camera_context_t camera_context, u8 mode)
{
	(void) camera_context;

	ov3640hw_power_down( mode );
	return 0;
}

static
int ov3640_read_16bit( p_camera_context_t camera_context,
		u16 reg_addr,  u16 *reg_val)
{
	(void) camera_context;

	ov3640hw_read_sensor_reg( reg_addr, (u8 *)reg_val );
	return 0;
}

static
int ov3640_write_16bit( p_camera_context_t camera_context,
		u16 reg_addr,  u16 reg_val)
{
	u8 buffer;
	int status;

	(void) camera_context;

	buffer = (u8)reg_val;
	status = (ov3640hw_write_sensor_reg( reg_addr, &buffer ) == 0) ?
		0 : -EIO;

	return status;
}

static int ov3640_set_digital_zoom( p_camera_context_t camera_context, u32 value)
{
	if ((value<100))
		value = 100;
	else if ((value>800))
		value = 800;

	/* set digital zoom */
	return ov3640hw_set_resolution_zoom(camera_context->capture_mode, value,
					camera_context->capture_input_width,
					camera_context->capture_input_height);
}

/**get digital zoom max level*/
static int ov3640_get_digital_zoom_max( p_camera_context_t cam_ctx, u32 *value)
{
	int w, h, mw, mh;

	w = cam_ctx->mode[CAMERA_MODE_VIDEO].width;
	h = cam_ctx->mode[CAMERA_MODE_VIDEO].height;

	if( (w<=0) || (h<=0) )
		return -EINVAL;

	mw = 1024 * 100 / w;
	mh = 768 * 100 / h;
	if (mw<mh)
		*value = mw;
	else
		*value = mh;
	return 0;
}

/**get digital zoom step*/
static int ov3640_get_digital_zoom_step( p_camera_context_t camera_context, u32 *value)
{
	*value = 10;
	return 0;
}

static
int ov3640_set_gamma( p_camera_context_t camera_context, u32 value)
{
	(void) camera_context;

	if( value > 2)
		value= 2;

	return ov3640hw_set_gamma( value);
}

static
int __attribute__ ((unused)) ov3640_set_brightness( p_camera_context_t camera_context, u32 value)
{
	(void) camera_context;

	if( value > 8)
		value= 8;

	return ov3640hw_set_brightness( value);
}

static int ov3640_try_contrast( p_camera_context_t camera_context, u32 value)
{
	(void) camera_context;

	if( (value == SENSOR_CONTRAST_LOWEST) ||
	    (value == SENSOR_CONTRAST_LOW)   ||
	    (value == SENSOR_CONTRAST_MIDDLE) ||
	    (value == SENSOR_CONTRAST_HIGH) ||
	    (value == SENSOR_CONTRAST_HIGHEST) )
		return 0;
	else
		return -ENOSYS;
}

static
int ov3640_set_contrast( p_camera_context_t camera_context, u8 mode, u32 value)
{
	int ovindex;
	(void) camera_context;

	if( value == SENSOR_CONTRAST_LOWEST )
		ovindex = 1;	/* M4 */
	else if( value == SENSOR_CONTRAST_LOW )
		ovindex = 3;	/* M2 */
	else if( value == SENSOR_CONTRAST_MIDDLE )
		ovindex = 0;	/* default */
	else if( value == SENSOR_CONTRAST_HIGH )
		ovindex = 6;	/* 2 */
	else if( value == SENSOR_CONTRAST_HIGHEST )
		ovindex = 8;	/* 4 */
	else
		ovindex = 0;

	return ov3640hw_set_contrast( ovindex);
}

/**ev compesate**/
static int ov3640_set_ev_offset( p_camera_context_t camera_context, int value)
{
	int ovindex;
	ovindex = 0;

	if (value==0)			/* 0 EV */
		ovindex = 0;
	else if (value>0) {
		if (value<=33)		/* 0.3 EV */
			ovindex = 6;
		else if (value<=70)	/* 0.7 EV */
			ovindex = 7;
		else if (value<=100)	/* 1.0 EV */
			ovindex = 8;
		else if (value<=133)	/* 1.3 EV */
			ovindex = 9;
		else			/* 1.7 EV */
			ovindex = 10;
	}
	else {
		if (value>=-33)		/* -0.3 EV */
			ovindex = 5;
		else if (value>=-70)	/* -0.7 EV */
			ovindex = 4;
		else if (value>=-100)	/* -1.0 EV */
			ovindex = 3;
		else if (value>=-133)	/* -1.3 EV */
			ovindex = 2;
		else			/* -1.7 EV */
			ovindex = 1;
	}

	printk("ov3640_set_evoffset: %d,%d\n", value, ovindex);
	return ov3640hw_set_exposure( ovindex);
}

static int ov3640_try_white_balance( p_camera_context_t camera_context, u32 value)
{
	(void) camera_context;

	if( (value == SENSOR_WHITEBALANCE_AUTO) ||
	    (value == SENSOR_WHITEBALANCE_DAYLIGHT) ||
	    (value == SENSOR_WHITEBALANCE_INCANDESCENT) ||
	    (value == SENSOR_WHITEBALANCE_FLUORESCENT) ||
	    (value == SENSOR_WHITEBALANCE_CLOUDY) )
		return 0;
	else
		return -ENOSYS;
}

static
int ov3640_set_white_balance( p_camera_context_t camera_context,
		u8 mode, u32 value)
{
        int ovindex = 0;
	(void) camera_context;

	{
		if( value == SENSOR_WHITEBALANCE_AUTO )
			ovindex = 0;
		else if( value == SENSOR_WHITEBALANCE_DAYLIGHT )
			ovindex = 2;
		else if( value == SENSOR_WHITEBALANCE_INCANDESCENT )
			ovindex = 5;
		else if( value == SENSOR_WHITEBALANCE_FLUORESCENT )
			ovindex = 4;
		else if( value == SENSOR_WHITEBALANCE_CLOUDY )
			ovindex = 3;
		else
			ovindex = 0;
	}

	return ov3640hw_set_white_balance( ovindex);
}

static int ov3640_set_framerate(u32 numerator, u32 denominator)
{
	int fps;
	if (0<denominator && 0<=numerator) {
		fps = numerator / denominator;

		if( fps<=10 )
			ov3640set_fps = 0;
		else if( fps>25 )
			ov3640set_fps = 30;
		else if( fps>15 )
			ov3640set_fps = 25;
		else
			ov3640set_fps = 15;

		return ov3640hw_set_framerate(ov3640set_fps);
	}
	return 0;
}

static int ov3640_try_flashmode( p_camera_context_t camera_context, u32 value)
{
	(void) camera_context;

	if ( (value==CAMERA_FLASH_OFF) ||
	     (value==CAMERA_FLASH_ON) )
		return 0;
	else
		return -ENOSYS;
}

static int ov3640_set_flashmode( p_camera_context_t camera_context, u32 value)
{
	(void) camera_context;

	//printk("ov3640: set_flashmode(%d, %d)\n", mode, value);
	flashled_flag = value;
	return 0;
}

static int ov3640_set_autofocus( p_camera_context_t camera_context, u32 value)
{
	int ret = 0;
	(void) camera_context;

	if (value==SENSOR_AUTOFOCUS_IDLE) {
		if(ov3640_single_af_enabled)
			ret = ov3640hw_af_finish();
	}
	else if (value==SENSOR_AUTOFOCUS_SINGLE) {
		ret = ov3640hw_single_af();
	}
	else
		return -EINVAL;

	return ret;
}

static int ov3640_enum_frame_sizes(p_camera_context_t camera_context, int mode, u32 format,
				   struct v4l2_frmsizeenum *frm_sizes)
{
	int index = frm_sizes->index;

	frm_sizes->type = V4L2_FRMSIZE_TYPE_DISCRETE;

	if (mode==CAMERA_MODE_VIDEO) {
		if (V4L2_PIX_FMT_YUV422P!=format) {
			printk(KERN_ERR "ov3640: only support YUV422 on video mode\n");
			return -EINVAL;
		}
		if (frm_sizes->index < ARRAY_SIZE(ov3640_yuv422_frame_sizes)) {
			frm_sizes->discrete.width = ov3640_yuv422_frame_sizes[index].width;
			frm_sizes->discrete.height = ov3640_yuv422_frame_sizes[index].height;
		}
		else
			return -EINVAL;
	}
	else {
		if (V4L2_PIX_FMT_JPEG!=format) {
			printk(KERN_ERR "ov3640: only support JPEG on still capture mode\n");
			return -EINVAL;
		}
		if (frm_sizes->index < ARRAY_SIZE(ov3640_jpeg_frame_sizes)) {
			frm_sizes->discrete.width = ov3640_jpeg_frame_sizes[index].width;
			frm_sizes->discrete.height = ov3640_jpeg_frame_sizes[index].height;
		}
		else
			return -EINVAL;
	}
	return 0;
}

static int ov3640_enum_frame_intervals(p_camera_context_t camera_context, u32 format,
					struct v4l2_frmivalenum *frm_ivals)
{
	int index = frm_ivals->index;

	frm_ivals->type = V4L2_FRMIVAL_TYPE_DISCRETE;

	if (V4L2_PIX_FMT_YUV422P!=format) {
		printk(KERN_ERR "ov3640: only YUV422 have frame interval\n");
		return -EINVAL;
	}
	if (frm_ivals->index < ARRAY_SIZE(ov3640_yuv422_frame_intervals)) {
		frm_ivals->discrete.numerator = ov3640_yuv422_frame_intervals[index].numerator;
		frm_ivals->discrete.denominator = ov3640_yuv422_frame_intervals[index].denominator;
	}
	else
		return -EINVAL;

	return 0;
}

static char name[] = "Omnivision3640";

static p_camera_function_t camera_functions;

static int format_list[] = {

	V4L2_PIX_FMT_YUV422P,
	/* V4L2_PIX_FMT_SRGGB10, */
	
	/* V4L2_PIX_FMT_SRGGB8, */
	/* V4L2_PIX_FMT_RGB565X, */
	V4L2_PIX_FMT_JPEG, 
	
	-1	
};

static
int ov3640_register(int id)
{
	/* allocte camera functions context */
	camera_functions = kzalloc(sizeof(camera_function_t), GFP_KERNEL);
	if (!camera_functions) {
		printk("Can't allocate buffer for camera functions structure \n");
		return -ENOMEM;
	}

	camera_functions->width_max = 2048;
	camera_functions->height_max = 1536;
	camera_functions->width_min = 2;
	camera_functions->height_min = 2;

	camera_functions->v_power = 0; //2800;
	camera_functions->v_io = 0; //1800;
		

	camera_functions->format_list = format_list;

	camera_functions->init = ov3640_init;     	
	camera_functions->deinit = ov3640_deinit;     
	camera_functions->set_capture_format = ov3640_set_capture_format; 	
	camera_functions->start_capture = ov3640_start_capture;	 		
	camera_functions->stop_capture = ov3640_stop_capture; 			
	camera_functions->sleep = ov3640_sleep;	
	camera_functions->wakeup = ov3640_wake;	
	camera_functions->read_8bit = NULL;			
	camera_functions->write_8bit = NULL;			
	camera_functions->read_16bit = ov3640_read_16bit;		
	camera_functions->write_16bit = ov3640_write_16bit;		
	camera_functions->read_32bit = NULL;	
	camera_functions->write_32bit = NULL;	
	camera_functions->set_power_mode = ov3640_set_power_mode;
	camera_functions->try_contrast = ov3640_try_contrast;
	camera_functions->set_contrast = ov3640_set_contrast;
	camera_functions->try_whitebalance = ov3640_try_white_balance;
	camera_functions->set_whitebalance = ov3640_set_white_balance;
	camera_functions->set_ev_offset = ov3640_set_ev_offset;
	camera_functions->set_gamma= ov3640_set_gamma;

	camera_functions->set_framerate =	ov3640_set_framerate;
	camera_functions->set_autofocus =	ov3640_set_autofocus;
	camera_functions->try_flashmode =	ov3640_try_flashmode;
	camera_functions->set_flashmode =	ov3640_set_flashmode;
	camera_functions->set_digital_zoom = 	ov3640_set_digital_zoom;
	camera_functions->get_digital_zoom_max 	= ov3640_get_digital_zoom_max;
	camera_functions->get_digital_zoom_step	= ov3640_get_digital_zoom_step;

	camera_functions->enum_frame_sizes	= ov3640_enum_frame_sizes;
	camera_functions->enum_frame_intervals	= ov3640_enum_frame_intervals;

	camera_functions->name = name;
	camera_functions->id = id;

	/*register 3640 as high resolution */

	if(sensor_register(camera_functions, id) < 0){
		printk("sensor_register failed !\n");
		kfree(camera_functions);
		camera_functions = NULL;
		return -1;
	}
	return 0;
}

static
int ov3640_unregister(int id)
{
	sensor_unregister(id);
	
	if(!camera_functions){
		kfree(camera_functions);
		camera_functions = NULL;
	}
	return 0;
}

/*****************************************************************************
 *                              OV3640 I2C Client Driver
 *****************************************************************************/
static
inline int read_sensor_reg_16bit(u16 addr, u8 *pvalue)
{
	int ret;    
	u8 address[2];
	if( g_client == NULL )	/* Not initialize the client */	
	{
		ovI2C_dbg("not initialize the client\n");
		return -1;
	}
	address[0] = addr>>8;	
	address[1] = addr;
	ret = i2c_smbus_write_byte_data(g_client,address[0],address[1]);
	if(ret) {
		printk(KERN_ERR "ov3640: read reg [%04x], write address failed %d\n", addr, ret);
		return ret;
	}
	ret = i2c_smbus_read_byte(g_client);
	if(ret<0) {
		printk(KERN_ERR "ov3640: read reg [%04x], read byte failed %d\n", addr, ret);
		return ret;
	}
	*pvalue = ret;   

	//printk("ov3640: read reg [%04x]=%02x\n", addr, *pvalue);
	return 0;
}
static
inline int write_sensor_reg_16bit(u16 addr, u8 *value)
{
	int ret;
	u8 data[3];
	if( g_client == NULL )	/* Not initialize the client */
		return -1;			
	data[0] = addr>>8;
	data[1] = addr;	
	data[2]=  *value;
	
	//printk("ov3640: write reg [%04x]=%02x\n", addr, *value);
	ret = i2c_master_send(g_client, data, 3);
	if(ret<0) {
		printk(KERN_ERR "ov3640: write reg [%04x:%02x] failed %d\n", addr, *value, ret);
		return ret;
	}
	
	return 0;
}

static
int __attribute__ ((unused)) write_sensor_reg_16bit_m(u16 addr, u8 value, u8 mask)
{
	int ret = 0;
	u8 regval= 0;
	
	read_sensor_reg_16bit(addr, &regval);
	
	regval  &=  (~mask);
	value   &=  mask;
	
	regval|=value;
	
	ret = write_sensor_reg_16bit(addr,&regval);

	return ret;
}

static int write_sensor_reg(u16 addr, u8 value)
{
	int ret;
	down(&ov3640_i2c_lock);
	ret = write_sensor_reg_16bit(addr, &value);
	up(&ov3640_i2c_lock);
	return ret;
}

static int read_sensor_reg(u16 addr, u8 *pvalue)
{
	int ret;
	down(&ov3640_i2c_lock);
	ret = read_sensor_reg_16bit(addr,pvalue);
	up(&ov3640_i2c_lock);
	return ret;
}

static int ov3640_create_workqueue(void)
{
	INIT_DELAYED_WORK(&flash_work, ov3640_flash_work_handler);
	flash_wqueue = create_workqueue("ov3640_flash");
	if(!flash_wqueue){
		printk(KERN_ERR "ov3640: create_workqueue(flash) failed\n");
		return -ENOMEM;
	}
	return 0;
}

static void ov3640_destroy_workqueue(void)
{
	if(flash_wqueue){
		cancel_rearming_delayed_workqueue(flash_wqueue, &flash_work);
		destroy_workqueue(flash_wqueue);
		flash_wqueue = NULL;
	}
}

static int __devinit sensor_probe(struct i2c_client *client, const struct i2c_device_id *_id)
{
	struct sensor_platform_data *pdata;
	struct clk *clk = NULL;
	int id = 0;
	
	pdata = client->dev.platform_data;
	id = pdata->id;
	printk(KERN_ERR"[%s:%d] id= %d\n", __FILE__, __LINE__, id);

	clk = clk_get(NULL, "CAMCLK");
	if (IS_ERR(clk)) {
		printk("sensor failed to get camera clock\n");
		return -1;
	}
	clk_enable(clk);
	
	ci_set_clock(1, 1, CICLK);
	
	g_client = client;

	if( pdata->power_on && (pdata->power_on( id) != 0) )
		printk(KERN_ERR"[%s:%d] Camera power on failed!!!!!!!\n", __FILE__, __LINE__);

	if (ov3640_sensorID_detect()<0)
		goto err;

	if(ov3640_register(id)<0)
		goto err;

	printk("OV3640 detected.\n");
	
	ov3640hw_enter_standby();
	if( pdata->power_off )
		pdata->power_off(id);

	ci_set_clock(0, 0, CICLK);
	clk_disable(clk);
	clk_put(clk);

	if(ov3640_create_workqueue())
		return -1;

	return 0;
err:
	printk("OV3640 detect failed.\n");
	if( pdata->power_off )
		pdata->power_off(id);

	ci_set_clock(0, 0, CICLK);
	if(clk){
		clk_disable(clk);
		clk_put(clk);

	}
	return -1;
}

static int sensor_remove(struct i2c_client *client)
{
	struct sensor_platform_data *pdata;
	pdata = client->dev.platform_data;
	
	ov3640_unregister(pdata->id);
	g_client = NULL;

	ov3640_destroy_workqueue();
	return 0;
}

static const struct i2c_device_id sensor_id[] = {
	{ "ov3640_08p801a", 0 },
	{ }
};

static struct i2c_driver sensor_driver = {
	.driver = {
		.name	= "ov3640_08p801a",
	},
	.probe		= sensor_probe,
	.remove		= sensor_remove,
	.id_table	= sensor_id,
};

static ssize_t flashled_proc_read(struct file *filp,
		char *buffer, size_t length, loff_t *offset)
{
	printk(KERN_INFO "%d\n", flashled_flag);

	return 0;
}

static ssize_t flashled_proc_write(struct file *filp,
		const char *buff, size_t len, loff_t *off)
{
	char msg[ 3];

	if (len > 3)
		len = 3;

	if (copy_from_user(msg, buff, len))
		return -EFAULT;

	switch( msg[ 0])
	{
	case '0':
		flashled_flag= 0;
		break;
	case '1':
		flashled_flag= 1;
		break;
	}
	
	return len;
}

static struct file_operations flashled_proc_fops = {
	.owner		= THIS_MODULE,
	.read		= flashled_proc_read,
	.write		= flashled_proc_write,
};

static int create_flashled_proc_file(void)
{
	flashled_proc_file = create_proc_entry( FLASHLED_PROC_FILE, 0644, NULL);
	if (!flashled_proc_file) {
		printk(KERN_INFO "Create proc file for PXA Backlight failed\n");
		return -ENOMEM;
	}

	printk( KERN_INFO "PXA Backlight proc OK\n" );
	flashled_proc_file->proc_fops = &flashled_proc_fops;
	return 0;
}

static ssize_t ov3640_proc_read(struct file *filp,
		char *buffer, size_t length, loff_t *offset)
{
	printk( KERN_INFO "ccc\n");

	return 0;
}

static ssize_t ov3640_proc_write(struct file *filp,
		const char *buff, size_t len, loff_t *off)
{
	char msg[ 4096], tmp[ 10];
	unsigned int addr= 0, i= 0;
	u8 value= 0;

	if( len > 4096)
		len = 4096;

	if( copy_from_user( msg, buff, len))
		return -EFAULT;

	if( strncmp( msg, "QVGA", 4) == 0)
	{
		printk("set QVGA\n");
//		ov3640_write_regs( ARRAY_AND_SIZE( QVGA_yuv_init_tab));
//		ov3640_write_regs( ARRAY_AND_SIZE( XGA_QVGA_tab));
		ov3640_write_regs( ARRAY_AND_SIZE( OV3640_video_QVGA));
		i= len;					// force to leave the function

	}
	else if( strncmp( msg, "YGA2", 4) == 0)
	{
		printk("set YGA2\n");
//		ov3640_write_regs( ARRAY_AND_SIZE( QVGA_yuv_init_tab));
//		ov3640_write_regs( ARRAY_AND_SIZE( XGA_YGA2_tab));
		ov3640_write_regs( ARRAY_AND_SIZE( OV3640_video_YGA2));
		i= len;					// force to leave the function
	}
	else if( strncmp( msg, "YGA", 3) == 0)
	{
		printk("set YGA\n");
//		ov3640_write_regs( ARRAY_AND_SIZE( QVGA_yuv_init_tab));
//		ov3640_write_regs( ARRAY_AND_SIZE( XGA_YGA_tab));
		ov3640_write_regs( ARRAY_AND_SIZE( OV3640_video_YGA));
		i= len;					// force to leave the function
	}
	else if( strncmp( msg, "VGA", 3) == 0)
	{
		printk("set VGA\n");
//		ov3640_write_regs( ARRAY_AND_SIZE( QVGA_yuv_init_tab));
//		ov3640_write_regs( ARRAY_AND_SIZE( XGA_VGA_tab));
		ov3640_write_regs( ARRAY_AND_SIZE( OV3640_video_VGA));
		i= len;					// force to leave the function
	}
	else if( strncmp( msg, "QXGA", 4) == 0)
	{
		printk("set QXGA\n");
//		ov3640_write_regs( ARRAY_AND_SIZE( QVGA_yuv_init_tab));
//		ov3640_write_regs( ARRAY_AND_SIZE( QXGA_tab));
		ov3640_write_regs( ARRAY_AND_SIZE( OV3640_capture_QXGA));
		i= len;					// force to leave the function
	}
	else if( strncmp( msg, "AF", 2) == 0)
	{
		printk("Auto Focus\n");
		ov3640_sensor_reg_init( AF_SETTING);//af function
		i= len;					// force to leave the function
	}

	while( i< len)
	{
		// get address
		while( i< len)
		{
			if( ( msg[ i] == '0') && ( msg[ i+ 1] == 'x'))
			{
				strncpy( tmp, msg+ i, 6);
				tmp[ 6]= '\0';
				i+= 6;
				addr= atoi( tmp);
				break;
			}
			i++;
		}
	
		// get value
		while( i<= len)
		{
			// get value, not set value
			if( ( msg[ i] == '\n') || ( i == len))
			{
				read_sensor_reg( addr, &value);
				printk( KERN_INFO"addr= 0x%04X, val= 0x%02X\n", addr, value);
				break;
			}
			else if( ( msg[ i] == '0') && ( msg[ i+ 1] == 'x'))
			{
				strncpy( tmp, msg+ i, 6);
				tmp[ 4]= '\0';
				i+= 4;
				value= atoi( tmp);
printk("Write to addr= 0x%04X, val= 0x%02X\n", addr, value);
				write_sensor_reg( addr, value);
				break;
			}
	
			i++;
		}
	}
	
	return len;
}

static struct file_operations ov3640_proc_fops = {
	.owner		= THIS_MODULE,
	.read		= ov3640_proc_read,
	.write		= ov3640_proc_write,
};

static int create_ov3640_proc_file(void)
{
	ov3640_proc_file = create_proc_entry( OV3640_PROC_FILE, 0644, NULL);
	if (!ov3640_proc_file) {
		printk(KERN_INFO "Create proc file for PXA OV3640 camera failed\n");
		return -ENOMEM;
	}

	printk( KERN_INFO "PXA Camera OV3640 proc OK\n" );
	ov3640_proc_file->proc_fops = &ov3640_proc_fops;
	return 0;
}

static int __init sensor_init(void)
{
	create_flashled_proc_file();
	create_ov3640_proc_file();

	//ov3640hw_reset();

	return i2c_add_driver(&sensor_driver);
}

static void __exit sensor_exit(void)
{
	i2c_del_driver(&sensor_driver);
}

module_init(sensor_init);
module_exit(sensor_exit);

MODULE_DESCRIPTION("OV3643 I2C Client driver");
MODULE_LICENSE("GPL");
