/****************************************************************************
  Copyright (C) 2005, Intel Corporation.

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

  (C) Copyright 2006 Marvell International Ltd.
  All Rights Reserved

 *****************************************************************************/

/*****************************************************************************
 * Abstract:
 *  Contains all hardware related functions for SIV120A
 *
 * Notes:Only valid for processor code named Monahans.
 *****************************************************************************/

#include <linux/init.h>
#include <linux/module.h>
#include <linux/delay.h>
#include <linux/err.h>
#include <linux/clk.h>
#include <linux/i2c.h>
#include <linux/sysfs.h>
#include <linux/firmware.h>
#include <linux/crc32.h>
#include <linux/proc_fs.h>
#include <asm/uaccess.h>
#include <linux/cdev.h>

#if defined(CONFIG_PXA3xx_DVFM)
#include <mach/dvfm.h>
#endif
#ifdef CONFIG_ANDROID_POWER
#include <linux/android_power.h>
#endif
#include <mach/mfp.h>

#include "camera.h"
#include "siv120.h"

static int siv120_i2c_read_reg(const u8 blkAddr,const u8 regAddr, u8 *regVal);
static int siv120_i2c_write_reg(const u8 blkAddr,const u8 regAddr, u8 regVal);
static int siv120_register(int id, u8 chipVer);
static int siv120_unregister(int id, u8 chipVer);
static int siv120b_register(int id);
static int siv120hw_set_ae(u8 mode, u32 value);
static int siv120hw_set_wb(u8 mode, u32 value);
static int siv120_power_down(u8 powerMode);
static int siv120_i2c_read_reg(const u8 blkAddr,const u8 regAddr, u8 *regVal);
static int siv120hw_set_ae(u8 mode, u32 value);
static int siv120hw_set_wb(u8 mode, u32 value);
static int siv120hw_set_framerate(int fps);
static int siv120hw_regApply(const S_REG_SET *pRegTbl);
static int siv120_i2c_write_reg(const u8 blkAddr,const u8 regAddr, u8 regVal);
static int sensor_clock_on(void);
static void siv120_reg_init(void);
static void sensor_clock_off(void);


#if defined(CONFIG_PXA3xx_DVFM)
static int dvfm_dev_idx;
#endif

static int g_sivMajor = 105;
static int g_sivMinor = 0;
static int g_sivNrDev = 1;

static unsigned char g_chipId = 0;
static unsigned char g_chipVer = 0;

static struct sivCDev *g_psivDev=NULL;
static struct i2c_client *g_client = NULL;

static struct clk *sensor_clk = NULL;
static u32 sensor_mclk = CICLK;

static int bClockOn = 0;
static int bPowerOn = 0;

struct siv120_reg_config g_siv120RegCfg;
struct siv120hw_data 	siv120hw_data;
static p_camera_function_t camera_functions;

static CAMERA_SETTINGS cameraSettings = {
	SENSOR_AUTO_EXPOSURE,
	SENSOR_EXPOSURE_MIDDLE,
	SENSOR_AUTO_WHITEBALANCE,
	SENSOR_WHITEBALANCE_AUTO
};

static int siv120_format_list[] = {
	V4L2_PIX_FMT_YUV422P,
	V4L2_PIX_FMT_RGB565X,
	-1
};

/*siv120a chip initial configuration*/
//////////////////////////////////////////////////////////////////////
const static S_INIT_TBL siv120a_regs_init_26MHZ_60HZ[] = {
	/* Sensor Control Block */
	{0x00, 0x00}, /* select CTRL_BLK */
	{0x04, 0x02}, 
	{0x10, 0x01},
	{0x17, 0x80},
	{0x13, 0x1a},
	{0x42, 0x32},
	{0x43, 0x80},

	/* flicker& frame setting 30~7.5FPS on 24MHz AC60Hz */
	{0x20, 0x00}, /* upper HBNKT & VBNKT */
	{0x21, 0x36}, /* lower HBNKT */
	{0x23, 0x09}, /* lower VBNKT */

	/* AE Block */
	{0x00, 0x01}, /* select AE_BLK */
	{0x10, 0x00}, /* Disable AE */
	{0x11, 0x04}, /* MAX Shutstep */

	{0x34, 0x7f}, /* set Shutter step for Preview mode */
	{0x40, 0x80}, /* max gain for dark condition */

	/* AWB Block */
	{0x00, 0x02},
	{0x10, 0xd3}, /* Enable AWB */
	{0x11, 0x00},
	{0x15, 0xfe},
	{0x16, 0x80},
	{0x17, 0xea},
	{0x18, 0x80},
	{0x19, 0xb0},
	{0x1a, 0x68},
	{0x1b, 0xb0},
	{0x1c, 0x64},
	{0x1d, 0x90},
	{0x1e, 0x70},
	{0x20, 0xe8},
	{0x21, 0x20},
	{0x22, 0x9d},
	{0x23, 0x18},
	{0x25, 0x20},
	{0x26, 0x0f},
	{0x27, 0x0d},
	{0x28, 0x90},
	{0x29, 0xd0},
	{0x2a, 0x90},
	{0x30, 0x05},
	{0x46, 0x64},

	/* RGB to YCbCr (CSC) no used */
	{0x50, 0x33},
	{0x51, 0x20},
	{0x52, 0xe5},
	{0x53, 0xfb},
	{0x54, 0x13},
	{0x55, 0x26},
	{0x56, 0x07},
	{0x57, 0xf5},
	{0x58, 0xea},
	{0x59, 0x21},
	{0x63, 0xbd},
	{0x64, 0xc2},
	{0x65, 0xbd},
	{0x66, 0xc2},
	{0x67, 0xef},
	{0x68, 0xa0},
	{0x69, 0xef},
	{0x6a, 0xa0},

	/* IDP Block */
	{0x00, 0x03},
	{0x10, 0xEF},

	{0x11, 0x19},
	{0x14, 0xf5},

	/* shading _0415  */
	{0x2c, 0x44},
	{0x2d, 0x22},
	{0X2e, 0x11},
	{0x32, 0xA8},

	/* Gamma */
	{0x34, 0x00},
	{0x35, 0x08},
	{0x36, 0x11},
	{0x37, 0x25},
	{0x38, 0x45},
	{0x39, 0x5f},
	{0x3a, 0x74},
	{0x3b, 0x87},
	{0x3c, 0x97},
	{0x3d, 0xa5},
	{0x3e, 0xb2},
	{0x3f, 0xc9},
	{0x40, 0xdd},
	{0x41, 0xf0},
	{0x42, 0xf8},
	{0x43, 0xff},

	{0x44, 0xbb},
	{0x4c, 0xa1},
	{0x4d, 0x08},
	{0x4e, 0xff},
	{0x4f, 0x05},

	{0x50, 0x80},
	{0x51, 0x29},
	{0x52, 0xff},
	{0x53, 0x13},
	{0x54, 0x3a},
	{0x55, 0xcb},
	{0x56, 0xfb},
	{0x57, 0x13},
	{0x58, 0x26},
	{0x59, 0x07},
	{0x5a, 0xf3},
	{0x5b, 0xbd},
	{0x5c, 0x50},

	{0x70, 0x20},
	{0x71, 0x30},
	{0x73, 0x06},
	{0x74, 0x06},
	{0x79, 0x20},
	{0x7a, 0x30},
	{0x7b, 0x06},
	{0x7c, 0x06},
	{0x7d, 0x20},

	{0x86, 0x10},
	{0x87, 0x10},

	/* set window size to 320X240 */
	{0xa0, 0x10},
	{0xa2, 0x40},
	{0xa4, 0xf0},

	{0x00,0x01},
	{0x10,0x80}, /* AE On */

	{0x00,0x00},
	{0x13,0x16}, /* solve noise problem */
	{0x00,0x01},
	{0x10,0x80}, /* bit7:AE */
	{0x00,0x02},
	{0x10,0xd3}, /* bit6-7:AWB */
	{0x00,0x03},
	{0x10,0xeb}, /* bit0:LSC bit2:gamma corr bit3:BLC */
	{0x14,0xe5}, /* bit4:sharpness */
	{0x44,0xbb}, /* bit3:de-noise bit7:DPC */
	{0x27,0xfd}, /* shading gain */
	{0x28,0xcb},
	{0x29,0xa8},
	{0x2a,0x64},
	{0x2b,0x21},

	{SIV120_REG_END, 0x00}
};

/* Auto exposure */
const unsigned char siv120a_regs_AE[][2] = {
	{0x00, 0x01},
	{0x10, 0x80}, 	/* AE On */
	{0x11, 0x04},  	/* Max shuttle time */
	{SIV120_REG_END, 0x00}
};

/* AE disalbe, exopsure lowest */
const unsigned char siv120a_regs_ExposureLowest[][2] = {
	{0x00, 0x01},
	{0x10, 0x00}, 	/* AE Off */
	{0x30, 0x00}, 	/* Shutter time, upper 8 bits */
	{0x31, 0x7D},   /* Shutter time, lower 8 bits */
	{SIV120_REG_END, 0x00}
};

/* AE disalbe, exopsure low */
const unsigned char siv120a_regs_ExposureLow[][2] = {
	{0x00, 0x01},
	{0x10, 0x00}, 	/* AE Off */
	{0x30, 0x01}, 	/* Shutter time, upper 8 bits */
	{0x31, 0x7D},   /* Shutter time, lower 8 bits */
	{SIV120_REG_END, 0x00}
};

/* AE disalbe, exopsure middle */
const unsigned char siv120a_regs_ExposureMiddle[][2] = {
	{0x00, 0x01},
	{0x10, 0x00}, 	/* AE Off */
	{0x30, 0x02}, 	/* Shutter time, upper 8 bits */
	{0x31, 0x7D},   /* Shutter time, lower 8 bits */
	{SIV120_REG_END, 0x00}
};

/* AE disalbe, exopsure high */
const unsigned char siv120a_regs_ExposureHigh[][2] = {
	{0x00, 0x01},
	{0x10, 0x00}, 	/* AE Off */
	{0x30, 0x04}, 	/* Shutter time, upper 8 bits */
	{0x31, 0x7D},   /* Shutter time, lower 8 bits */
	{SIV120_REG_END, 0x00}
};

/* AE disalbe, exopsure highest */
const unsigned char siv120a_regs_ExposureHighest[][2] = {
	{0x00, 0x01},
	{0x10, 0x00}, 	/* AE Off */
	{0x30, 0x08}, 	/* Shutter time, upper 8 bits */
	{0x31, 0x7D},   /* Shutter time, lower 8 bits */
	{SIV120_REG_END, 0x00}
};

/* Auto white balance */
const unsigned char siv120a_regs_AutoWB[][2] = {
	{0x00, 0x02},
	{0x10, 0xC3}, 	/* AWB On */
	{0x60, 0xAA}, 	/* AWB R gain */
	{0x61, 0xBE},   /* AWB B gain */
	{SIV120_REG_END, 0x00}
};

/* white balance, incandescent */
const unsigned char siv120a_regs_WBIncandescent[][2] = {
	{0x00, 0x02},
	{0x10, 0x00}, 	/* AWB Off */
	{0x60, 0x98}, 	/* AWB R gain */
	{0x61, 0xC8},   /* AWB B gain */
	{SIV120_REG_END, 0x00}
};

/* white balance, sunny */
const unsigned char siv120a_regs_WBSunny[][2] = {
	{0x00, 0x02},
	{0x10, 0x00}, 	/* AWB Off */
	{0x60, 0xC2}, 	/* AWB R gain */
	{0x61, 0x9E},    /* AWB B gain */
	{SIV120_REG_END, 0x00}
};

/* white balance, fluorescent */
const unsigned char siv120a_regs_WBFluorescent[][2] = {
	{0x00, 0x02},
	{0x10, 0x00}, 	/* AWB Off */
	{0x60, 0xAA}, 	/* AWB R gain */
	{0x61, 0xBE},   /* AWB B gain */
	{SIV120_REG_END, 0x00}
};

/* white balance, cloudy */
const unsigned char siv120a_regs_WBCloudy[][2] = {
	{0x00, 0x02},
	{0x10, 0x00}, 	/* AWB Off */
	{0x60, 0xCA}, 	/* AWB R gain */
	{0x61, 0xCF},   /* AWB B gain */
	{SIV120_REG_END, 0x00}
};

/* output format, YUV422 */
const unsigned char siv120a_regs_fmtYUV422[][2] = {
	{0x00, 0x03},
	{0x12, 0x05},	/* Set output format */
	{SIV120_REG_END, 0x00}
};

/* output format, RGB565 */
const unsigned char siv120a_regs_fmtRGB565[][2] = {
	{0x00, 0x03},
	{0x12, 0x04}, 	/* Set output format */
	{SIV120_REG_END, 0x00}
};

/* output size, 640x480 */
const unsigned char siv120a_regs_size640_480[][2] = {
	{0x00, 0x03},   /* Width and Height */
	{0xA0, 0x24},
	{0xA2, 0x80},
	{0xA4, 0xE0},
	{SIV120_REG_END, 0x00}
};

/* output size, 352x288 */
const unsigned char siv120a_regs_size352_288[][2] = {
	{0x00, 0x03},   /* Width and Height */
	{0xA0, 0x14},
	{0xA2, 0x60},
	{0xA4, 0x20},
	{SIV120_REG_END, 0x00}
};

/* output size, 320x240 */
const unsigned char siv120a_regs_size320_240[][2] = {
	{0x00, 0x03},   /* Width and Height */
	{0xA0, 0x10},
	{0xA2, 0x40},
	{0xA4, 0xF0},
	{SIV120_REG_END, 0x00}
};

/* output size, 176x144 */
const unsigned char siv120a_regs_size176_144[][2] = {
	{0x00, 0x03},   /* Width and Height */
	{0xA0, 0x00},
	{0xA2, 0xB0},
	{0xA4, 0x90},
	{SIV120_REG_END, 0x00}
};

/* output size, 160x120 */
const unsigned char siv120a_regs_size160_120[][2] = {
	{0x00, 0x03},   /* Width and Height */
	{0xA0, 0x00},
	{0xA2, 0xA0},
	{0xA4, 0x78},
	{SIV120_REG_END, 0x00}
};

/* output size, 88x72 */
const unsigned char siv120a_regs_size88_72[][2] = {
	{0x00, 0x03},   /* Width and Height */
	{0xA0, 0x00},
	{0xA2, 0x58},
	{0xA4, 0x48},
	{SIV120_REG_END, 0x00}
};

/* still mode */
const unsigned char siv120a_regs_stillMode[][2] = {
	{0x00, 0x00},   /* Control block */
	{0x04, 0x10}, 	/* Set output format */
	{SIV120_REG_END, 0x00}
};

/* output pad enable */
const unsigned char siv120a_regs_opad[][2] = {
	{0x00, 0x00},   /* Control block */
	{0x03, 0x04}, 	/* Set output format */
	{SIV120_REG_END, 0x00}
};

/*siv120b chip initial configuration table*/
//////////////////////////////////////////////////////////////////////////
const unsigned char siv120b_regs_init_26MHZ_60HZ[][2] = {
	/* Sensor Control Block */
	{0x00, 0x00}, /* select CTRL_BLK */
	{0x04, 0x06}, /*{0x04, 0x02}, 1/2 MCLK CLK_DIV + Vertical filp image*/
	{0x13, 0x17}, //#anti_black sun
	{0x16, 0xcf}, //#0xb6

	{0x20, 0x00}, // #P_BNKT
	{0x21, 0x01}, // #P_HBNKT
	{0x22, 0x01}, // #P_ROWFIL
	{0x23, 0x01}, // #P_VBNKT

	{0x40, 0x00}, // #0x8b
	{0x41, 0x00}, // #0x96
	{0x42, 0x00}, // #0xda
	{0x43, 0x00}, // #0x83


	/* AE Block */
	{0x00, 0x01},
	{0x10, 0x80}, 	/* AE On */

	{0x00, 0x01}, /* select AE_BLK */
	{0x11, 0x1D}, // # 6fps at lowlux
	{0x12, 0x74}, // # D65 target 0x78
	{0x13, 0x74}, // # CWF target 0x78
	{0x14, 0x74}, // # A target   0x78
	{0x1E, 0x00}, // # ini gain	0x08
	{0x34, 0x7d},
	{0x40, 0x4c}, // # Max x7?	0x48

	{0x41, 0x20}, // #AG_TOP1	0x28
	{0x42, 0x20}, // #AG_TOP0	0x28
	{0x43, 0x00}, // #AG_MIN1	0x08
	{0x44, 0x00}, // #AG_MIN0	0x08
	{0x45, 0x00}, // #G50_dec	0x09
	{0x46, 0x0a}, // #G33_dec	0x17
	{0x47, 0x10}, // #G25_dec	0x1d
	{0x48, 0x13}, // #G20_dec	0x21
	{0x49, 0x15}, // #G12_dec	0x23
	{0x4a, 0x18}, // #G09_dec	0x24
	{0x4b, 0x1a}, // #G06_dec	0x26
	{0x4c, 0x1d}, // #G03_dec	0x27
	{0x4d, 0x20}, // #G100_inc	0x27
	{0x4e, 0x10}, // #G50_inc	0x1a
	{0x4f, 0x0a}, // #G33_inc	0x14
	{0x50, 0x08}, // #G25_inc	0x11
	{0x51, 0x06}, // #G20_inc	0x0f
	{0x52, 0x05}, // #G12_inc	0x0d
	{0x53, 0x04}, // #G09_inc	0x0c
	{0x54, 0x02}, // #G06_inc	0x0a
	{0x55, 0x01}, // #G03_inc	0x09

	/* AWB Block */
	{0x00, 0x02},
	{0x10, 0xd3}, /* Enable AWB */
	{0x11, 0xc0},
	{0x12, 0x80},
	{0x13, 0x80},
	{0x14, 0x80},
	{0x15, 0xfe}, //# R gain Top
	{0x16, 0x88}, //# R gain bottom
	{0x17, 0xea}, //# B gain Top
	{0x18, 0x70}, //# B gain bottom
	{0x19, 0x90}, //# Cr top value 0x90
	{0x1a, 0x70}, //# Cr bottom value 0x70
	{0x1b, 0x90}, //# Cb top value 0x90
	{0x1c, 0x70}, //# Cb bottom value 0x70
	{0x1d, 0x90}, //# 0x90
	{0x1e, 0x70}, //# 0x70
	{0x20, 0xe8}, //# AWB luminous top value
	{0x21, 0x30}, //# AWB luminous bottom value
	{0x22, 0xa4},
	{0x23, 0x20},
	{0x25, 0x20},
	{0x26, 0x0f},
	{0x27, 0x10}, //# ST for outdoor enable 0x10
	{0x28, 0x13}, //# ST for outdoor disable
	{0x29, 0xac}, //# AWB R gain at outdoor 0xb8
	{0x2a, 0x8c}, //# AWB B gain at outdoor 0x8c

	{0x30, 0x00},
	{0x31, 0x10},
	{0x32, 0x00},
	{0x33, 0x10},
	{0x34, 0x02},
	{0x35, 0x76},
	{0x36, 0x01},
	{0x37, 0xd6},
	{0x40, 0x01},
	{0x41, 0x04},
	{0x42, 0x08},
	{0x43, 0x10},
	{0x44, 0x12},
	{0x45, 0x35},
	{0x46, 0x64},
	{0x50, 0x33},
	{0x51, 0x20},
	{0x52, 0xe5},
	{0x53, 0xfb},
	{0x54, 0x13},
	{0x55, 0x26},
	{0x56, 0x07},
	{0x57, 0xf5},
	{0x58, 0xea},
	{0x59, 0x21},

	{0x63, 0x9c}, //# R D30 to D20
	{0x64, 0xca}, //# B D30 to D20
	{0x65, 0x9c}, //# R D20 to D30
	{0x66, 0xca}, //# B D20 to D30
	{0x67, 0xbb}, //# R D65 to D30
	{0x68, 0x99}, //# B D65 to D30
	{0x69, 0xbb}, //# R D30 to D65
	{0x6a, 0x99}, //# B D30 to D65

	/* IDP Block */
	{0x00, 0x03},
	{0x10, 0xef},
	{0x11, 0x19},
	{0x8c, 0x10},

	//Shading Register Setting
	{0x40, 0x00},
	{0x41, 0x21},
	{0x42, 0x32},
	{0x43, 0x43},
	{0x44, 0x77},
	{0x45, 0x77},
	{0x46, 0x11}, //# left R gain[7:4], right R gain[3:0] 0x22
	{0x47, 0x11}, //# top R gain[7:4], bottom R gain[3:0]
	{0x48, 0x11}, //# left Gr gain[7:4], right Gr gain[3:0]
	{0x49, 0x10}, //# top Gr gain[7:4], bottom Gr gain[3:0]
	{0x4a, 0x00}, //# left Gb gain[7:4], right Gb gain[3:0]
	{0x4b, 0x00}, //# top Gb gain[7:4], bottom Gb gain[3:0]
	{0x4c, 0x11}, //# left B gain[7:4], right B gain[3:0]
	{0x4d, 0x00}, //# top B gain[7:4], bottom B gain[3:0]
	{0x4e, 0x04}, //# X-axis center high[3:2], Y-axis center high[1:0]
	{0x4f, 0x50}, //# X-axis center low[7:0]
	{0x50, 0xf6}, //# Y-axis center low[7:0]
	{0x51, 0x80}, //# Shading Center Gain
	{0x52, 0x00}, //# Shading R Offset
	{0x53, 0x00}, //# Shading Gr Offset
	{0x54, 0x00}, //# Shading Gb Offset
	{0x55, 0x00}, //# Shading B Offset

	//Gamma
	{0x30, 0x0},// 	#0x0
	{0x31, 0x3},// 	#0x3
	{0x32, 0xA},// 	#0xb
	{0x33, 0x1E},//	#0x21
	{0x34, 0x42},//	#0x48
	{0x35, 0x5E},//	#0x63
	{0x36, 0x74},//	#0x78
	{0x37, 0x85},//	#0x89
	{0x38, 0x94},//	#0x98
	{0x39, 0xA2},//	#0xA6
	{0x3a, 0xAF},//	#0xB3
	{0x3b, 0xC6},//	#0xCA
	{0x3c, 0xDB},//	#0xDD
	{0x3d, 0xEF},//	#0xEF
	{0x3e, 0xF8},//	#0xF8
	{0x3f, 0xFF},//	#0xFF

	//# Color matrix (D65) - Daylight
	{0x71, 0x3e},//	#0x46
	{0x72, 0xbc},//	#0xb2
	{0x73, 0x06},//	#0x08
	{0x74, 0x17},//	#0x0d
	{0x75, 0x1e},//	#0x26
	{0x76, 0x0b},//	#0x0d
	{0x77, 0xf4},//	#0xf5
	{0x78, 0xcd},//	#0xbf
	{0x79, 0x3f},//	#0x4c

	//# Color matrix (D30) - CWF
	{0x7a, 0x38},//	#0x40
	{0x7b, 0xbc},//  #0xbe
	{0x7c, 0x0c},//  #0x02
	{0x7d, 0x14},//  #0x0d
	{0x7e, 0x1e},//  #0x26
	{0x7f, 0x0e},//  #0x0c
	{0x80, 0xf6},//  #0xf3
	{0x81, 0xcd},//  #0xce
	{0x82, 0x3d},//  #0x40

	//# Color matrix (D20) - A
	{0x83, 0x38},//	#0x3a
	{0x84, 0xc4},//  #0xcc
	{0x85, 0x04},//  #0xfa
	{0x86, 0x07},//  #0xfe
	{0x87, 0x25},//  #0x3a
	{0x88, 0x14},//  #0x08
	{0x89, 0xf0},//  #0xf9
	{0x8a, 0xc2},//  #0xc3
	{0x8b, 0x4f},//  #0x44

	//# DPCNR
	{0x17, 0xC8},//  # DPCNRCTRL
	{0x18, 0x08},//  # DPTHR
	{0x19, 0x48},//  # [7:6] G DP Number Thr @ Dark | [5:0] DPTHRMIN
	{0x1A, 0x48},//  # [7:6] G DP Number Thr @ Normal | [5:0] DPTHRMAX
	{0x1B, 0x24},//  # DPTHRSLP( [7:4] @ Normal | [3:0] @ Dark )
	{0x1C, 0x00},//  # NRTHR
	{0x1D, 0x48},//  # [7:6] C DP Number Thr @ Dark | [5:0] NRTHRMIN 0x4f
	{0x1E, 0x48},//  # [7:6] C DP Number Thr @ Normal | [5:0] NRTHRMAX 0x4f
	{0x1F, 0x2f},//  # NRTHRSLP( [7:4] @ Normal | [3:0] @ Dark )	0x58
	{0x20, 0x04},//  # IllumiInfo STRTNOR
	{0x21, 0x0f},//  # IllumiInfo STRTDRK

	{0x56, 0x10},//	# lowlux shading enable
	{0x57, 0x92},//	# lowlux shading
	{0x58, 0x00},//	# lowlux shading on

	{0x61, 0xb4},//	# INTCTRL outdoor
	{0x62, 0xb4},//	# INTCTRL indoor
	{0x63, 0xb4},//	# INTCTRL dark

	//# Edge gain
	{0x90, 0x18},//	# Green upper gain 0x18
	{0x91, 0x18},//	# Green down gain 0x18
	{0x92, 0x02},//	# upper coring value 0x04
	{0x96, 0x02},//	# down coring value 0x04

	{0x9f, 0x18},//	# Y upper gain 0x10
	{0xa0, 0x18},//	# Y down gain 0x10
	{0xa1, 0x04},//	# Y upper coring 0x02
	{0xa2, 0x04},//	# Y down coring 0x02

	{0xa9, 0x12},//	# Cr saturation 0x14
	{0xaa, 0x12},//	# Cb saturation 0x14
	{0xb9, 0x10},//	# 0x20 lowlux color
	{0xba, 0x20},//	# 0x10 lowlux color

	//#0xdd	0xc7
	{0xdd, 0x4f},//	# ENHCTRL5
	{0xde, 0xfa},//	# NOIZCTRL 0xba

	{0xe5, 0x15},
	{0xe6, 0x28},
	{0xe7, 0x04},

	/* set window size to 320X240 */
	{0xc0, 0x10},
	{0xc2, 0x40},
	{0xc4, 0xf0},

	{SIV120_REG_END, 0x00}
};

/* Auto exposure */
const unsigned char siv120b_regs_AE[][2] = {
	{0x00, 0x01},
	{0x10, 0x80}, 	/* AE On */
	{0x11, 0x1D}, 	/* Max shuttle time */
	{SIV120_REG_END, 0x00}
};


/* AE disalbe, exopsure lowest */
const unsigned char siv120b_regs_ExposureLowest[][2] = {
	{0x00, 0x01},
	{0x10, 0x00}, 	/* AE Off */
	{0x30, 0x00}, 	/* Shutter time, upper 8 bits */
	{0x31, 0x7D},   /* Shutter time, lower 8 bits */
	{SIV120_REG_END, 0x00}
};

/* AE disalbe, exopsure low */
const unsigned char siv120b_regs_ExposureLow[][2] = {
	{0x00, 0x01},
	{0x10, 0x00}, 	/* AE Off */
	{0x30, 0x01}, 	/* Shutter time, upper 8 bits */
	{0x31, 0x7D},   /* Shutter time, lower 8 bits */
	{SIV120_REG_END, 0x00}
};


/* AE disalbe, exopsure middle */
const unsigned char siv120b_regs_ExposureMiddle[][2] = {
	{0x00, 0x01},
	{0x10, 0x00}, 	/* AE Off */
	{0x30, 0x02}, 	/* Shutter time, upper 8 bits */
	{0x31, 0x7D},    /* Shutter time, lower 8 bits */
	{SIV120_REG_END, 0x00}
};

/* AE disalbe, exopsure high */
const unsigned char siv120b_regs_ExposureHigh[][2] = {
	{0x00, 0x01},
	{0x10, 0x00}, 	/* AE Off */
	{0x30, 0x04}, 	/* Shutter time, upper 8 bits */
	{0x31, 0x7D},   /* Shutter time, lower 8 bits */
	{SIV120_REG_END, 0x00}
};

/* AE disalbe, exopsure highest */
const unsigned char siv120b_regs_ExposureHighest[][2] = {
	{0x00, 0x01},
	{0x10, 0x00}, 	/* AE Off */
	{0x30, 0x08}, 	/* Shutter time, upper 8 bits */
	{0x31, 0x7D},   /* Shutter time, lower 8 bits */
	{SIV120_REG_END, 0x00}
};

/* Auto white balance */
const unsigned char siv120b_regs_AutoWB[][2] = {
	{0x00, 0x02},
	{0x10, 0xD3}, 	/* AWB On */
	{0x60, 0xAA}, 	/* AWB R gain */
	{0x61, 0xBE},   /* AWB B gain */
	{SIV120_REG_END, 0x00}
};

/* white balance, incandescent */
const unsigned char siv120b_regs_WBIncandescent[][2] = {
	{0x00, 0x02},
	{0x10, 0x00}, 	/* AWB Off */
	{0x60, 0x75}, 	/* AWB R gain */
	{0x61, 0xD3},   /* AWB B gain */
	{SIV120_REG_END, 0x00}
};

/* white balance, sunny */
const unsigned char siv120b_regs_WBSunny[][2] = {
	{0x00, 0x02},
	{0x10, 0x00}, 	/* AWB Off */
	{0x60, 0xA1}, 	/* AWB R gain */
	{0x61, 0x8B},   /* AWB B gain */
	{SIV120_REG_END, 0x00}
};

/* white balance, fluorescent */
const unsigned char siv120b_regs_WBFluorescent[][2] = {
	{0x00, 0x02},
	{0x10, 0x00}, 	/* AWB Off */
	{0x60, 0x9D}, 	/* AWB R gain */
	{0x61, 0xCD},   /* AWB B gain */
	{SIV120_REG_END, 0x00}
};

/* white balance, cloudy */
const unsigned char siv120b_regs_WBCloudy[][2] = {
	{0x00, 0x02},
	{0x10, 0x00}, 	/* AWB Off */
	{0x60, 0xCA}, 	/* AWB R gain */
	{0x61, 0xCF},   /* AWB B gain */
	{SIV120_REG_END, 0x00}
};

/* output format, YUV422 */
const unsigned char siv120b_regs_fmtYUV422[][2] = {
	{0x00, 0x03},
	{0x12, 0x05},	/* Set output format */
	{SIV120_REG_END, 0x00}
};

/* output format, RGB565 */
const unsigned char siv120b_regs_fmtRGB565[][2] = {
	{0x00, 0x03},
	{0x12, 0x03}, 	/* Set output format */
	{SIV120_REG_END, 0x00}
};

/* output size, 640x480 */
const unsigned char siv120b_regs_size640_480[][2] = {
	{0x00, 0x03},   /* Width and Height */
	{0xC0, 0x24},
	{0xC2, 0x80},
	{0xC4, 0xE0},
	{SIV120_REG_END, 0x00}
};

/* output size, 352x288 */
const unsigned char siv120b_regs_size352_288[][2] = {
	{0x00, 0x03},   /* Width and Height */
	{0xC0, 0x14},
	{0xC2, 0x60},
	{0xC4, 0x20},
	{SIV120_REG_END, 0x00}
};

/* output size, 320x240 */
const unsigned char siv120b_regs_size320_240[][2] = {
	{0x00, 0x03},   /* Width and Height */
	{0xC0, 0x10},
	{0xC2, 0x40},
	{0xC4, 0xF0},
	{SIV120_REG_END, 0x00}
};

/* output size, 176x144 */
const unsigned char siv120b_regs_size176_144[][2] = {
	{0x00, 0x03},   /* Width and Height */
	{0xC0, 0x00},
	{0xC2, 0xB0},
	{0xC4, 0x90},
	{SIV120_REG_END, 0x00}
};

/* output size, 160x120 */
const unsigned char siv120b_regs_size160_120[][2] = {
	{0x00, 0x03},   /* Width and Height */
	{0xC0, 0x00},
	{0xC2, 0xA0},
	{0xC4, 0x79},
	{SIV120_REG_END, 0x00}
};

/* output size, 88x72 */
const unsigned char siv120b_regs_size88_72[][2] = {
	{0x00, 0x03},   /* Width and Height */
	{0xC0, 0x00},
	{0xC2, 0x58},
	{0xC4, 0x48},
	{SIV120_REG_END, 0x00}
};

/* output pad enable */
const unsigned char siv120b_regs_opad[][2] = {
	{0x00, 0x00},   /* Control block */
	{0x03, 0x04}, 	/* Set output format */
	{SIV120_REG_END, 0x00}
};

////////////////////////////////////////////////////////////////////////////////////////
static int siv120_i2c_read_reg(const u8 blkAddr,const u8 regAddr, u8 *regVal)
{
	int ret;
	struct sensor_platform_data *pdata;

	if ( g_client == NULL )  /* No global client pointer?   */
		return -1;
	pdata = g_client->dev.platform_data;

	/* select block */
	ret = i2c_smbus_write_byte_data(g_client, BLK_SEL, blkAddr);
	if (ret < 0) {
		printk(KERN_ERR "siv120_i2c_read_reg write BLK_SEL %d fail, return %d\n", blkAddr, ret);
		return ret;
	}

	ret = i2c_smbus_read_byte_data(g_client, regAddr);
	if (ret >= 0) {
		*regVal = ret;
	} else
		printk(KERN_ERR "siv120_i2c_read_reg(0x%2x, 0x%02x) failed, return %d\n", blkAddr, regAddr, ret);

	return ret;
}

static int siv120_i2c_write_reg(const u8 blkAddr,const u8 regAddr, u8 regVal)
{
	int ret;
	struct sensor_platform_data *pdata;

	pdata = g_client->dev.platform_data;

	printk(KERN_ERR "siv120_i2c_write_reg: regAddr=0x02%x, regVal=0x%02x\n", regAddr, regVal);

	if ( g_client == NULL )  /* No global client pointer?   */
		return -1;

	/* select block */
	ret = i2c_smbus_write_byte_data(g_client, BLK_SEL, blkAddr);
	if (ret < 0) {
		printk(KERN_ERR "siv120_i2c_write_reg write BLK_SEL %d fail, return %d\n", blkAddr, ret);
		return ret;
	}

	ret = i2c_smbus_write_byte_data(g_client, regAddr, regVal);
	if ( ret<0 )
		printk(KERN_ERR "siv120_i2c_write_reg(0x%02x, 0x%02x, 0x%02x) failed, return %d\n", blkAddr, regAddr, regVal, ret);
	return ret;
}

static int siv120_read_8bit(p_camera_context_t camera_context,
		u8 reg_addr,  u8 *reg_val)
{
	printk(KERN_ERR "siv120_read_8bit: reg_addr 0x%X\n", reg_addr);

	(void) camera_context;
	return siv120_i2c_read_reg(CTRL_BLK, reg_addr, reg_val);
}

static int siv120_write_8bit( p_camera_context_t camera_context,
		u8 reg_addr,  u8 reg_val)
{
	/* Currently, this function should not be called from Camera Interface */
	return 0;
}

static void siv120hw_data_init(void)
{
	siv120hw_data.wb_mode = 0;  /* auto white balance */
	siv120hw_data.ae_mode = 0;  /* auto exposure */
	siv120hw_data.fps = 30;     /* frame rate */
}

static int siv120hw_set_framerate(int fps)
{
	printk("fps: %d\n", fps);
	switch (fps) {
		case 30:
			siv120_i2c_write_reg(CTRL_BLK, 0x20, 0x00);     /* Horizontal & Vertical blank time */
			siv120_i2c_write_reg(CTRL_BLK, 0x21, 0x36);
			siv120_i2c_write_reg(CTRL_BLK, 0x23, 0x09);
			siv120_i2c_write_reg(AE_BLK, 0x11, 0x04);       /* Max. Shutter time */
			siv120_i2c_write_reg(AE_BLK, 0x34, 0x7f);       /* STST_P */
			break;
		case 24:
			siv120_i2c_write_reg(CTRL_BLK, 0x20, 0x00);     /* Horizontal & Vertical blank time */
			siv120_i2c_write_reg(CTRL_BLK, 0x21, 0x36);
			siv120_i2c_write_reg(CTRL_BLK, 0x23, 0x88);
			siv120_i2c_write_reg(AE_BLK, 0x11, 0x05);       /* Max. Shutter time */
			siv120_i2c_write_reg(AE_BLK, 0x34, 0x7f);       /* STST_P */
			break;
		case 20:
			siv120_i2c_write_reg(CTRL_BLK, 0x20, 0x01);     /* Horizontal & Vertical blank time */
			siv120_i2c_write_reg(CTRL_BLK, 0x21, 0x36);
			siv120_i2c_write_reg(CTRL_BLK, 0x23, 0x07);
			siv120_i2c_write_reg(AE_BLK, 0x11, 0x06);       /* Max. Shutter time */
			siv120_i2c_write_reg(AE_BLK, 0x34, 0x7f);       /* STST_P */
			break;
		case 15:
			siv120_i2c_write_reg(CTRL_BLK, 0x20, 0x02);     /* Horizontal & Vertical blank time */
			siv120_i2c_write_reg(CTRL_BLK, 0x21, 0x36);
			siv120_i2c_write_reg(CTRL_BLK, 0x23, 0x05);
			siv120_i2c_write_reg(AE_BLK, 0x11, 0x08);       /* Max. Shutter time */
			siv120_i2c_write_reg(AE_BLK, 0x34, 0x7f);       /* STST_P */
			break;
		case 12:
			siv120_i2c_write_reg(CTRL_BLK, 0x20, 0x03);     /* Horizontal & Vertical blank time */
			siv120_i2c_write_reg(CTRL_BLK, 0x21, 0x36);
			siv120_i2c_write_reg(CTRL_BLK, 0x23, 0x03);
			siv120_i2c_write_reg(AE_BLK, 0x11, 0x0a);       /* Max. Shutter time */
			siv120_i2c_write_reg(AE_BLK, 0x34, 0x7f);       /* STST_P */
			break;
		case 10:
			siv120_i2c_write_reg(CTRL_BLK, 0x20, 0x03);     /* Horizontal & Vertical blank time */
			siv120_i2c_write_reg(CTRL_BLK, 0x21, 0x6f);
			siv120_i2c_write_reg(CTRL_BLK, 0x23, 0xa1);
			siv120_i2c_write_reg(AE_BLK, 0x11, 0x0c);       /* Max. Shutter time */
			siv120_i2c_write_reg(AE_BLK, 0x34, 0x77);       /* STST_P */
			break;
		case 8:
			siv120_i2c_write_reg(CTRL_BLK, 0x20, 0x13);     /* Horizontal & Vertical blank time */
			siv120_i2c_write_reg(CTRL_BLK, 0x21, 0x1c);
			siv120_i2c_write_reg(CTRL_BLK, 0x23, 0xe9);
			siv120_i2c_write_reg(AE_BLK, 0x11, 0x0f);       /* Max. Shutter time */
			siv120_i2c_write_reg(AE_BLK, 0x34, 0x64);       /* STST_P */
			break;
		case 6:
			siv120_i2c_write_reg(CTRL_BLK, 0x20, 0x23);     /* Horizontal & Vertical blank time */
			siv120_i2c_write_reg(CTRL_BLK, 0x21, 0xad);
			siv120_i2c_write_reg(CTRL_BLK, 0x23, 0xc1);
			siv120_i2c_write_reg(AE_BLK, 0x11, 0x14);       /* Max. Shutter time */
			siv120_i2c_write_reg(AE_BLK, 0x34, 0x49);       /* STST_P */
			break;
		default:
			printk(KERN_ERR "siv120hw_set_framerate:Not support !!\n");
			return -1;
	}

	return 0;
}

static int siv120hw_regApply(const S_REG_SET *pRegTbl)
{
	int status;
	struct i2c_client *i2c_client;

	status = 0;
	i2c_client = g_client;

	if (!pRegTbl)
		return -1;

	while ( pRegTbl->addr != SIV120_REG_END) {
		status = (i2c_smbus_write_byte_data(i2c_client, pRegTbl->addr, pRegTbl->val ) == 0) ?
			0 : -EIO;

		if (status == -EIO)
			printk("siv120hw_regApply 0x%X fail\n", pRegTbl->addr);

		pRegTbl++;
	}

	return 0;
}

static int siv120hw_set_ae(u8 mode, u32 value)
{
	const S_REG_SET *pRegTbl=NULL;

	if (mode == SENSOR_AUTO_EXPOSURE)
		pRegTbl = g_siv120RegCfg.autoExposureTable;
	else if (mode == SENSOR_MANUAL_EXPOSURE) {
		switch (value) {
			case SENSOR_EXPOSURE_LOWEST:
				pRegTbl = g_siv120RegCfg.exposureLowestTable;
				break;
			case SENSOR_EXPOSURE_LOW:
				pRegTbl = g_siv120RegCfg.exposureLowTable;
				break;
			case SENSOR_EXPOSURE_MIDDLE:
				pRegTbl = g_siv120RegCfg.exposureMiddleTable;
				break;
			case SENSOR_EXPOSURE_HIGH:
				pRegTbl = g_siv120RegCfg.exposureHighTable;
				break;
			default:
			case SENSOR_EXPOSURE_HIGHEST:
				pRegTbl = g_siv120RegCfg.exposureHighestTable;
				break;
		}
	}

	if (siv120hw_regApply(pRegTbl) < 0)
		return -1;

	return 0;
}

static int siv120hw_set_wb(u8 mode, u32 value)
{
	const S_REG_SET *pRegTbl=NULL;

	if (mode == SENSOR_AUTO_WHITEBALANCE)
		pRegTbl = g_siv120RegCfg.autoWBTable;
	else if (mode == SENSOR_MANUAL_WHITEBALANCE) {
		switch (value) {
			case SENSOR_WHITEBALANCE_AUTO:
				pRegTbl = g_siv120RegCfg.autoWBTable;
				break;
			case SENSOR_WHITEBALANCE_INCANDESCENT:  /* Incandescent */
				pRegTbl = g_siv120RegCfg.wbIncandescentTable;
				break;
			case SENSOR_WHITEBALANCE_SUNNY:     /* Sunny */
				pRegTbl = g_siv120RegCfg.wbSunnyTable;
				break;
			case SENSOR_WHITEBALANCE_FLUORESCENT:   /* Fluorescent */
				pRegTbl = g_siv120RegCfg.wbFluorescentTable;
				break;
			case SENSOR_WHITEBALANCE_CLOUDY:  /* ext: Cloudy */
				pRegTbl = g_siv120RegCfg.wbCloudyTable;
				break;
			default:
				return -EINVAL;
				break;
		}
	}

	if (siv120hw_regApply(pRegTbl) < 0)
		return -1;

	return 0;
}

static int sensor_clock_on(void)
{
	printk(KERN_ERR "siv120: sensor_clock_on\n");

#ifdef CONFIG_PXA3xx_DVFM
	dvfm_disable_op_name("D2", dvfm_dev_idx);
	dvfm_disable_op_name("D0CS", dvfm_dev_idx);
	dvfm_disable_op_name("104M", dvfm_dev_idx);
	dvfm_disable_op_name("208M", dvfm_dev_idx);
#endif
	msleep(10);

	sensor_clk = clk_get(NULL, "CAMCLK");
	if (IS_ERR(sensor_clk)) {
		printk(KERN_ERR "siv120: sensor failed to get camera clock\n");
		sensor_clk = NULL;
		return -EINVAL;
	}

	ci_set_clock(1, 1, sensor_mclk);
	clk_enable(sensor_clk);
	bClockOn = 1;
	return 0;
}

static void sensor_clock_off(void)
{
	printk(KERN_ERR "siv120: sensor_clock_off\n");

	if ((sensor_clk==NULL) || IS_ERR(sensor_clk)) {
		printk(KERN_ERR "siv120: sensor_clock_off() invalid handler\n");
		return;
	}

	ci_set_clock(0, 0, sensor_mclk);
	clk_disable(sensor_clk);
	clk_put(sensor_clk);
#ifdef CONFIG_PXA3xx_DVFM
	dvfm_enable_op_name("D2", dvfm_dev_idx);
	dvfm_enable_op_name("D0CS", dvfm_dev_idx);
	dvfm_enable_op_name("104M", dvfm_dev_idx);
	dvfm_enable_op_name("208M", dvfm_dev_idx);
#endif
	msleep(10);
	bClockOn = 0;
}


////////////////////////////////////////////////////////////////////////////////////////
static void siv120_reg_init(void)
{
	const S_REG_SET *pRegTbl;

	pRegTbl = g_siv120RegCfg.initTable;

	siv120hw_regApply(pRegTbl);
}

static ssize_t siv120_ae_mode_show(struct device *dev, struct device_attribute *attr, char *buf)
{
	return sprintf(buf, "%04x\n", siv120hw_data.ae_mode);
}

static ssize_t siv120_ae_mode_store(struct device *dev, struct device_attribute *attr, const char *buf, size_t n)
{
	int val;

	sscanf(buf, "%x", &val);
	if (val<0 || val>0x0A)
		printk(KERN_ERR "invalid input\n");
	else {
		siv120hw_data.ae_mode = val;

		if (bPowerOn) {
			/* apply */
			if (val == 0)
				siv120hw_set_ae(SENSOR_AUTO_EXPOSURE, 0);
			else
				siv120hw_set_ae(SENSOR_MANUAL_EXPOSURE, val-1);
		}
	}

	return n;
}
static DEVICE_ATTR(aemode, S_IRUGO|S_IWUSR, siv120_ae_mode_show, siv120_ae_mode_store);

static ssize_t siv120_wb_mode_show(struct device *dev, struct device_attribute *attr, char *buf)
{
	return sprintf(buf, "%04x\n", siv120hw_data.wb_mode);
}

static ssize_t siv120_wb_mode_store(struct device *dev, struct device_attribute *attr, const char *buf, size_t n)
{
	int val;

	sscanf(buf, "%x", &val);
	if (val<0 || val>0x4)
		printk(KERN_ERR "invalid input\n");
	else {
		siv120hw_data.wb_mode = val;

		if (bPowerOn) {
			/* apply */
			if (val == 0)
				siv120hw_set_wb(SENSOR_AUTO_WHITEBALANCE, 0);
			else
				siv120hw_set_wb(SENSOR_MANUAL_WHITEBALANCE, val);
		}
	}

	return n;
}
static DEVICE_ATTR(wbmode, S_IRUGO|S_IWUSR, siv120_wb_mode_show, siv120_wb_mode_store);

static ssize_t siv120_sensorID_show(struct device *dev, struct device_attribute *attr, char *buf)
{
	u8 sensorId=0;
	int clock_on=1,power_on=1;

	if (bClockOn == 0) {
		sensor_clock_on();
		clock_on=0;
	}

	if (bPowerOn == 0) {
		siv120_power_down(CAMERA_POWER_FULL);
		power_on=0;
	}

	/* verify sensor id */
	siv120_i2c_read_reg(CTRL_BLK, 0x01, &sensorId);

	if (power_on == 0) {
		siv120_power_down(CAMERA_POWER_OFF);
	}

	if (clock_on ==0) {
		sensor_clock_off();
	}

	return sprintf(buf, "siv120: sensor_id=%02x\n", sensorId);
}
static DEVICE_ATTR(sensor_id, S_IRUSR|S_IWUSR, siv120_sensorID_show, NULL);

static ssize_t siv120_fps_show(struct device *dev, struct device_attribute *attr, char *buf)
{
	return sprintf(buf, "%d\n", siv120hw_data.fps);
}

static ssize_t siv120_fps_store(struct device *dev, struct device_attribute *attr, const char *buf, size_t n)
{
	int val;
	sscanf(buf, "%d", &val);

	if (val<0 || val>30) {
		printk(KERN_ERR "invalid input\n");
		return -1;
	}

	switch (val) {
		case 30:
		case 24:
		case 20:
		case 15:
		case 12:
		case 10:
		case 8:
		case 6:
			siv120hw_data.fps = val;
			break;
		default:
			printk(KERN_ERR "Not support !!\n");
			return -2;
	}

	if (bPowerOn) {
		/* apply */
		siv120hw_set_framerate(siv120hw_data.fps);
	}

	return n;
}
static DEVICE_ATTR(fps, S_IRUGO|S_IWUSR, siv120_fps_show, siv120_fps_store);

static struct attribute *siv120_sysfs_attr[] = {
	&dev_attr_aemode.attr,
	&dev_attr_wbmode.attr,
	&dev_attr_sensor_id.attr,
	&dev_attr_fps.attr,
	NULL
};


static int siv120_sysfs_add(struct kobject *kobj)
{
	int i, n, ret;
	n = ARRAY_SIZE(siv120_sysfs_attr);
	for (i = 0; i < n; ++i) {
		if ( siv120_sysfs_attr[i] ) {
			ret = sysfs_create_file(kobj, siv120_sysfs_attr[i]);
			if (ret)
				return ret;
		}
	}

	if (ret)
		return ret;
	return 0;
}

static int siv120_sysfs_rm(struct kobject *kobj)
{
	int i, n;

	n = ARRAY_SIZE(siv120_sysfs_attr);
	for (i = 0; i < n; i++) {
		if ( siv120_sysfs_attr[i] )
			sysfs_remove_file(kobj, siv120_sysfs_attr[i]);
	}
	return 0;
}

static int siv120_set_contrast( p_camera_context_t camera_context, u8 mode, u32 value)
{
	int ret = 0;
	(void) camera_context;

	printk(KERN_ERR "siv120_set_contrast not supported\n");
	if ( mode!=SENSOR_MANUAL_CONTRAST )
		return -EINVAL;
	switch (value) {
		case SENSOR_CONTRAST_LOWEST:

			break;
		case SENSOR_CONTRAST_LOW:

			break;
		case SENSOR_CONTRAST_MIDDLE:

			break;
		case SENSOR_CONTRAST_HIGH:

			break;
		case SENSOR_CONTRAST_HIGHEST:

			break;
		default:
			return -EINVAL;
	}

	/* set hw */
	if (ret < 0) {
		printk(KERN_ERR "siv120_set_contrast() set go failed(%d)\n", ret);
		return ret;
	}

	return 0;
}

static int siv120_set_exposure( p_camera_context_t camera_context, u8 mode, u32 value)
{
	int ret = 0;
	(void) camera_context;

	printk(KERN_ERR "siv120_set_exposure\n");

	/* set hw */
	ret = siv120hw_set_ae(mode, value);

	if ( ret<0 ) {
		printk(KERN_ERR "siv120_set_exposure() set failed(%d)\n", ret);
		return ret;
	}

	return 0;
}

static int siv120_set_white_balance( p_camera_context_t camera_context,
		u8 mode, u32 value)
{
	int ret = 0;
	(void) camera_context;

	printk(KERN_ERR "siv120_set_white_balance\n");

	ret = siv120hw_set_wb(mode, value);

	if ( ret<0 ) {
		printk(KERN_ERR "siv120_set_white_balance() set go failed(%d)\n", ret);
		return ret;
	}

	return 0;
}

static int siv120_set_capture_format(  p_camera_context_t camera_context )
{
	CI_MP_TIMING timing;
	u8  regVal, regVal4Vmode, regVal4CNTRB, regOrg;
	const S_REG_SET *pRegTbl;

	printk(KERN_ERR "siv120: capture_mode=%d\n", camera_context->capture_mode);
	printk(KERN_ERR "siv120: capture_input_format=%d\n", camera_context->capture_input_format);
	printk(KERN_ERR "siv120: capture_input_width=%d\n", camera_context->capture_input_width);
	printk(KERN_ERR "siv120: capture_input_height=%d\n", camera_context->capture_input_height);

	/* apply AE and AWB */
	if (siv120hw_data.ae_mode == 0)
		siv120hw_set_ae(SENSOR_AUTO_EXPOSURE, 0);
	else
		siv120hw_set_ae(SENSOR_MANUAL_EXPOSURE, siv120hw_data.ae_mode-1);

	if (siv120hw_data.wb_mode == 0)
		siv120hw_set_wb(SENSOR_AUTO_WHITEBALANCE, 0);
	else
		siv120hw_set_wb(SENSOR_MANUAL_WHITEBALANCE, siv120hw_data.wb_mode);

	siv120hw_set_framerate(siv120hw_data.fps);

	/* set input format */
	siv120_i2c_read_reg(IDP_BLK, 0x12, &regVal);

	regOrg = regVal;
	regVal &= ~0x7;

	switch (camera_context->capture_input_format) {
		case V4L2_PIX_FMT_YUV422P:
			pRegTbl = g_siv120RegCfg.outFmtYUV422Table;
			regVal |= pRegTbl[1].val;
			break;

		case V4L2_PIX_FMT_RGB565X:
			pRegTbl = g_siv120RegCfg.outFmtRGB565Table;
			regVal |= pRegTbl[1].val;
			break;

		default:
			printk(KERN_ERR "siv120_set_capture_format invalid format %d\n", camera_context->capture_input_format);
			return -EINVAL;
			break;
	}

	if (regOrg != regVal)
		siv120_i2c_write_reg(IDP_BLK, 0x12, regVal);

	/* set width & height */
	siv120_i2c_read_reg(CTRL_BLK, 0x5, &regVal4Vmode);

	regOrg = regVal4Vmode;
	regVal4Vmode &= ~0x7;
	switch (camera_context->capture_input_width) {
		case 640:
			if (camera_context->capture_input_height != 480) {
				printk(KERN_ERR "siv120_set_capture_format invalid height %d\n", camera_context->capture_input_height);
				return -EINVAL;
			}
			regVal4Vmode |= 6;
			/* set window size */
			pRegTbl = g_siv120RegCfg.outSize640_480_Table;
			break;
		case 352:
			if (camera_context->capture_input_height != 288) {
				printk(KERN_ERR "siv120_set_capture_format invalid height %d\n", camera_context->capture_input_height);
				return -EINVAL;
			}
			regVal4Vmode |= 2;
			/* set window size */
			pRegTbl = g_siv120RegCfg.outSize352_288_Table;
			break;
		case 320:
			if (camera_context->capture_input_height != 240) {
				printk(KERN_ERR "siv120_set_capture_format invalid height %d\n", camera_context->capture_input_height);
				return -EINVAL;
			}
			regVal4Vmode |= 5;
			/* set window size */
			pRegTbl = g_siv120RegCfg.outSize320_240_Table;
			break;
		case 176:
			if (camera_context->capture_input_height != 144) {
				printk(KERN_ERR "siv120_set_capture_format invalid height %d\n", camera_context->capture_input_height);
				return -EINVAL;
			}
			regVal4Vmode |= 1;
			/* set window size */
			pRegTbl = g_siv120RegCfg.outSize176_144_Table;
			break;
		case 160:
			if (camera_context->capture_input_height != 120) {
				printk(KERN_ERR "siv120_set_capture_format invalid height %d\n", camera_context->capture_input_height);
				return -EINVAL;
			}
			regVal4Vmode |= 4;
			/* set window size */
			pRegTbl = g_siv120RegCfg.outSize160_120_Table;
			break;
		case 88:
			if (camera_context->capture_input_height != 72) {
				printk(KERN_ERR "siv120_set_capture_format invalid height %d\n", camera_context->capture_input_height);
				return -EINVAL;
			}
			regVal4Vmode |= 0;
			/* set window size */
			pRegTbl = g_siv120RegCfg.outSize88_72_Table;
			break;
		default:
			pRegTbl = NULL;
			break;
	}

	if (siv120hw_regApply(pRegTbl) < 0)
		return -1;

	if (regOrg != regVal4Vmode)
		siv120_i2c_write_reg(CTRL_BLK, 0x5, regVal4Vmode);

	pRegTbl = g_siv120RegCfg.stillModeTable;
	if (pRegTbl) {
		/* set capture mode */
		siv120_i2c_read_reg(CTRL_BLK, 0x4, &regVal4CNTRB);
		regOrg = regVal4CNTRB;

		if (camera_context->capture_mode == CAMERA_MODE_STILL)
			regVal4CNTRB |= pRegTbl[1].val;
		else
			regVal4CNTRB &= ~pRegTbl[1].val;

		if (regOrg != regVal4CNTRB)
			siv120_i2c_write_reg(CTRL_BLK, 0x4, regVal4CNTRB);
	}

	timing.BFW = 1;
	timing.BLW = 0x00;

	ci_configure_mp(camera_context->capture_input_width-1,
			camera_context->capture_input_height-1, &timing);

	printk(KERN_ERR "siv120_set_capture_format: ci_configure_mp witd = %d heigh = %d\n",
			camera_context->capture_input_width,camera_context->capture_input_height);

	return 0;
}

static int siv120_start_capture(p_camera_context_t camera_context,
		unsigned int frames)
{
	int ret = 0;
	u8  regVal,regOrg;

	(void) camera_context;

	printk(KERN_ERR "siv120_start_capture frames %d\n", frames);

	ret = siv120_i2c_read_reg(CTRL_BLK, 0x3, &regVal);
	if (ret < 0) {
		printk(KERN_ERR "siv120_start_capture fail\n");
		return ret;
	}

	regOrg = regVal;
	regVal |= 0x4; /* set OPAD_EN to normal operation mode */
	if (regOrg != regVal)
		ret = siv120_i2c_write_reg(CTRL_BLK, 0x3, regVal);

	if (ret < 0) {
		printk(KERN_ERR "siv120_start_capture fail\n");
		return ret;
	}

	/* sleep 2.5 seconds for AE and AWB stable period */
	msleep(2500);

	return 0;
}

static int siv120_stop_capture(  p_camera_context_t camera_context )
{
	int ret = 0;
	u8  regVal,regOrg;

	printk(KERN_ERR "siv120_stop_capture\n"); /* Kevin Ko */

	(void) camera_context;

	ret = siv120_i2c_read_reg(CTRL_BLK, 0x3, &regVal);
	if (ret < 0) {
		printk(KERN_ERR "siv120_stop_capture fail\n");
		return ret;
	}

	regOrg = regVal;
	regVal &= ~0x4; /* set OPAD_EN to idle operation mode */

	if (regOrg != regVal)
		ret = siv120_i2c_write_reg(CTRL_BLK, 0x3, regVal);

	if (ret < 0) {
		printk(KERN_ERR "siv120_stop_capture fail\n");
		return ret;
	}

	return 0;
}

static int siv120_get_framerate(u32 format, u32 width, u32 height, u32 *numerator, u32 *denominator)
{
	int ret = 0;

	printk(KERN_ERR "siv120_get_framerate failed\n");

	*numerator = 30;
	*denominator = 1;

	return ret;
}

static int siv120_set_framerate(u32 numerator, u32 denominator)
{
	int fps;

	printk(KERN_ERR "siv120_set_framerate failed\n");

	if (0<denominator && 0<=numerator) {
		fps = numerator / denominator;
		if ( fps>30 )
			fps = 30;
		siv120hw_data.fps = fps;
	}

	return siv120hw_set_framerate(siv120hw_data.fps);
}

static int siv120_power_down(u8 powerMode)
{
	u8  regVal;
	u8  regVal4Ldo;
	struct sensor_platform_data *pdata;

	printk(KERN_ERR "siv120_power_down(%d)\n", powerMode);

	if (powerMode == CAMERA_POWER_OFF) {
		pdata = g_client->dev.platform_data;
		pdata->standby(pdata->id);
		msleep(10);
		pdata->power_off(pdata->id);
	} else {
		pdata = g_client->dev.platform_data;
		pdata->power_on(pdata->id);
		msleep(100);
		pdata->wakeup(pdata->id);

		/* copy from Foxcon platform wakeup() */
		if (i2c_smbus_write_byte_data(g_client, 0x0, 0x0) < 0)
			printk("siv120 wakeup:  i2c_smbus_write_byte_data fail\n");
		regVal = i2c_smbus_read_byte_data(g_client, 0x03);
		regVal &= ~0x2; /* PWDN set to normal mode */
		regVal |= 1; /* SNR_EN set to dynamic mode */
		if (i2c_smbus_write_byte_data(g_client, 0x03, regVal) < 0)
			printk("siv120 wakeup:  i2c_smbus_write_byte_data fail\n");
		/* end copy from Foxcon platform wakeup() */
		msleep(10);

		siv120_i2c_read_reg(CTRL_BLK, 0x17, &regVal4Ldo);

		/* Enable internal LDO */
		regVal4Ldo |= 0x80;
		siv120_i2c_write_reg(CTRL_BLK, 0x17, regVal4Ldo);

		msleep(10);
	}

	return 0;
}

static int siv120_deinit(  p_camera_context_t camera_context )
{
	/* power off the external module */
	printk(KERN_ERR "siv120_deinit\n"); /* Kevin */

	if (camera_functions)
		camera_functions->set_power_mode(NULL,CAMERA_POWER_OFF);
	else
		printk(KERN_ERR "camera_functions is NULL!!!\n");

	sensor_clock_off();

	return 0;
}

static int siv120_init( p_camera_context_t camera_context )
{
	int status;

	printk(KERN_ERR "siv120_init\n");

	/* provide informat about the capabilities of the sensor */
	camera_context->sensor_status.caps |= SENSOR_CAP_MANUAL_CONTRAST |
		SENSOR_CAP_MANUAL_WHITEBALANCE |
		SENSOR_CAP_MANUAL_EXPOSURE;

	/* Configure CI according to SIV120A's hardware
	 * master parallel with 8 data pins
	 */
	ci_set_mode(CI_MODE_MP, CI_DATA_WIDTH8);

	/* enable pixel clock(sensor will provide pclock)
	 * and master clock = 26MHZ
	 */
	sensor_clock_on();

	/* Default setting : Output data synchronizing PCLK edge polarity control on Positive edge
	   active data output @ Low VSYNC & High HSYNC  */
	ci_set_polarity(0, 0, 1);

	/* fifo control */
	ci_set_fifo(0, CI_FIFO_THL_32, 1, 1);

	/* set black level */
	ci_cgu_set_black_level(0);

	/* CGU Mux Select */
	ci_cgu_set_addr_mux_select(CI_CGU_MUX_0_TO_7);

	/* power on */
	status = siv120_power_down(CAMERA_POWER_FULL);
	if ( status<0 ) {
		printk(KERN_ERR "siv120_init: power on failed\n");
		return status;
	}

	siv120_reg_init();

	return 0;
}


static int siv120_sleep(  p_camera_context_t camera_context )
{
	printk(KERN_ERR "siv120_sleep\n");
	return siv120_deinit( camera_context );
}

static int siv120_wake(  p_camera_context_t camera_context )
{
	printk(KERN_ERR "siv120_wake\n");
	return siv120_init( camera_context );
}


static int siv120_set_power_mode(p_camera_context_t camera_context, u8 mode)
{
	(void) camera_context;
	siv120_power_down( mode );

	return 0;
}

static int siv120b_register(int id)
{
	printk(KERN_ERR "siv120b_register\n");

	g_siv120RegCfg.initTable            = (S_REG_SET *)siv120b_regs_init_26MHZ_60HZ;
	g_siv120RegCfg.autoExposureTable    = (S_REG_SET *)siv120b_regs_AE;
	g_siv120RegCfg.exposureLowestTable	= (S_REG_SET *)siv120b_regs_ExposureLowest;
	g_siv120RegCfg.exposureLowTable 	= (S_REG_SET *)siv120b_regs_ExposureLow;
	g_siv120RegCfg.exposureMiddleTable  = (S_REG_SET *)siv120b_regs_ExposureMiddle;
	g_siv120RegCfg.exposureHighTable    = (S_REG_SET *)siv120b_regs_ExposureHigh;
	g_siv120RegCfg.exposureHighestTable = (S_REG_SET *)siv120b_regs_ExposureHighest;
	g_siv120RegCfg.autoWBTable          = (S_REG_SET *)siv120b_regs_AutoWB;
	g_siv120RegCfg.wbIncandescentTable  = (S_REG_SET *)siv120b_regs_WBIncandescent;
	g_siv120RegCfg.wbSunnyTable         = (S_REG_SET *)siv120b_regs_WBSunny;
	g_siv120RegCfg.wbFluorescentTable   = (S_REG_SET *)siv120b_regs_WBFluorescent;
	g_siv120RegCfg.wbCloudyTable        = (S_REG_SET *)siv120b_regs_WBCloudy;
	g_siv120RegCfg.outFmtYUV422Table    = (S_REG_SET *)siv120b_regs_fmtYUV422;
	g_siv120RegCfg.outFmtRGB565Table    = (S_REG_SET *)siv120b_regs_fmtRGB565;
	g_siv120RegCfg.outSize640_480_Table = (S_REG_SET *)siv120b_regs_size640_480;
	g_siv120RegCfg.outSize352_288_Table = (S_REG_SET *)siv120b_regs_size352_288;
	g_siv120RegCfg.outSize320_240_Table = (S_REG_SET *)siv120b_regs_size320_240;
	g_siv120RegCfg.outSize176_144_Table = (S_REG_SET *)siv120b_regs_size176_144;
	g_siv120RegCfg.outSize160_120_Table = (S_REG_SET *)siv120b_regs_size160_120;
	g_siv120RegCfg.outSize88_72_Table   = (S_REG_SET *)siv120b_regs_size88_72;
	g_siv120RegCfg.stillModeTable       = NULL;  /* still mode can't be set in siv120b due to it affects the AE function */
	g_siv120RegCfg.opadTable            = (S_REG_SET *)siv120b_regs_opad;

	/* allocte camera functions context */
	camera_functions = kzalloc(sizeof(camera_function_t), GFP_KERNEL);
	if (!camera_functions) {
		printk(KERN_ERR "siv120: Can't allocate buffer for camera functions structure \n");
		return -ENOMEM;
	}

	camera_functions->width_max 	= 640;
	camera_functions->height_max 	= 480;
	camera_functions->width_min 	= 88;
	camera_functions->height_min 	= 72;
	camera_functions->v_power = 	-1;     /* VAA 2.8V */
	camera_functions->v_io 		= -1;   /* VDD & VIO is controlled by GPIO */
	camera_functions->set_capture_format 	= siv120_set_capture_format;
	camera_functions->start_capture 	= siv120_start_capture;
	camera_functions->stop_capture 		= siv120_stop_capture;
	camera_functions->set_contrast 		= siv120_set_contrast;
	camera_functions->set_whitebalance 	= siv120_set_white_balance;
	camera_functions->set_exposure 		= siv120_set_exposure;
	camera_functions->get_framerate 	= siv120_get_framerate;
	camera_functions->name 			= "SIV120";
	camera_functions->init 			= siv120_init;
	camera_functions->set_power_mode 	= siv120_set_power_mode;
	camera_functions->format_list 		= siv120_format_list;
	camera_functions->deinit 		= siv120_deinit;
	camera_functions->sleep 		= siv120_sleep;
	camera_functions->wakeup 		= siv120_wake;
	camera_functions->read_8bit 		= siv120_read_8bit;
	camera_functions->write_8bit 		= siv120_write_8bit;
	camera_functions->read_16bit 		= NULL;
	camera_functions->write_16bit 		= NULL;
	camera_functions->read_32bit 		= NULL;
	camera_functions->write_32bit 		= NULL;
	camera_functions->set_zoom 		= NULL;
	camera_functions->set_framerate 	= NULL;
	camera_functions->set_autofocus 	= NULL;
	camera_functions->id = SENSOR_LOW;

	/*register siv120 as low resolution */
	if (sensor_register(camera_functions, id) < 0) {
		printk(KERN_ERR "siv120: sensor_register failed !\n");
		kfree(camera_functions);
		camera_functions = NULL;
		return -1;
	}
	return 0;

}

static int siv120b_unregister(int id)
{
	printk(KERN_ERR "siv120_unregister\n");

	sensor_unregister(id);

	if (!camera_functions) {
		kfree(camera_functions);
		camera_functions = NULL;
	}

	memset(&g_siv120RegCfg, 0, sizeof(g_siv120RegCfg));

	return 0;
}

////////////////////////////////////////////////////////////////////////////////////////
static int siv120a_set_power_mode(p_camera_context_t camera_context, u8 mode)
{
	(void) camera_context;
	siv120_power_down( mode );
	return 0;
}

static void siv120a_reg_init(void)
{
	int status = 0;
	const S_INIT_TBL *pInitTbl = siv120a_regs_init_26MHZ_60HZ;

	/* The list is a register number followed by the value */
	while ( pInitTbl->addr != SIV120A_REG_END) {
		status = (i2c_smbus_write_byte_data(g_client, pInitTbl->addr, pInitTbl->val ) == 0) ?
			0 : -EIO;
		if (status == -EIO)
			printk("siv120a_reg_init 0x%X fail\n", pInitTbl->addr);
		pInitTbl++;
	}
}

static int siv120a_init( p_camera_context_t camera_context )
{
	int status;
	u8  chipId = 0, chipVer = 0;

	printk(KERN_ERR "siv120a_init\n");

	/* provide informat about the capabilities of the sensor */
	camera_context->sensor_status.caps |= SENSOR_CAP_MANUAL_CONTRAST |
		SENSOR_CAP_MANUAL_WHITEBALANCE |
		SENSOR_CAP_MANUAL_EXPOSURE;

	/* Configure CI according to SIV120A's hardware
	 * master parallel with 8 data pins
	 */
	ci_set_mode(CI_MODE_MP, CI_DATA_WIDTH8);

	/* enable pixel clock(sensor will provide pclock)
	 * and master clock = 26MHZ
	 */
	ci_set_clock(1, 1, sensor_mclk);
	msleep(10);

	/* Default setting : Output data synchronizing PCLK edge polarity control on Positive edge
	   active data output @ Low VSYNC & High HSYNC  */
	ci_set_polarity(0, 0, 1);

	/* fifo control */
	ci_set_fifo(0, CI_FIFO_THL_32, 1, 1);

	/* set black level */
	ci_cgu_set_black_level(0);

	/* CGU Mux Select */
	ci_cgu_set_addr_mux_select(CI_CGU_MUX_0_TO_7);

	/* power on */
	status = siv120_power_down(CAMERA_POWER_FULL);
	if ( status<0 )
		return status;
	msleep(10);

	siv120_i2c_read_reg(CTRL_BLK, 0x01, &chipId);
	if (chipId != SIV120A_CHIP_ID) {
		printk(KERN_ERR "SIV120A detect failed chip ID is %d.\n", chipId);
		return -EIO;
	}

	siv120_i2c_read_reg(CTRL_BLK, 0x02, &chipVer);
	printk(KERN_ERR "SIV120A detect chip ID is 0x%x, chip ver is 0x%x\n", chipId, chipVer);

	siv120a_reg_init();

	return 0;
}

static int siv120a_deinit(  p_camera_context_t camera_context )
{
	/* power off the external module */
	printk(KERN_ERR "siv120a_deinit\n");

	siv120_power_down( CAMERA_POWER_OFF );
	ci_set_clock(0, 0, sensor_mclk);

	return 0;
}

static int siv120a_sleep(  p_camera_context_t camera_context )
{
	printk(KERN_ERR "siv120a_sleep\n");
	return siv120a_deinit( camera_context );
}

static int siv120a_wake(  p_camera_context_t camera_context )
{
	printk(KERN_ERR "siv120a_wake\n");
	return siv120a_init( camera_context );
}

static int siv120a_register(int id)
{
	g_siv120RegCfg.initTable            	= (S_REG_SET *)siv120a_regs_init_26MHZ_60HZ;
	g_siv120RegCfg.autoExposureTable    	= (S_REG_SET *)siv120a_regs_AE;
	g_siv120RegCfg.exposureLowestTable	= (S_REG_SET *)siv120a_regs_ExposureLowest;
	g_siv120RegCfg.exposureLowTable  	= (S_REG_SET *)siv120a_regs_ExposureLow;
	g_siv120RegCfg.exposureMiddleTable  	= (S_REG_SET *)siv120a_regs_ExposureMiddle;
	g_siv120RegCfg.exposureHighTable    	= (S_REG_SET *)siv120a_regs_ExposureHigh;
	g_siv120RegCfg.exposureHighestTable 	= (S_REG_SET *)siv120a_regs_ExposureHighest;
	g_siv120RegCfg.autoWBTable          	= (S_REG_SET *)siv120a_regs_AutoWB;
	g_siv120RegCfg.wbIncandescentTable  	= (S_REG_SET *)siv120a_regs_WBIncandescent;
	g_siv120RegCfg.wbSunnyTable         	= (S_REG_SET *)siv120a_regs_WBSunny;
	g_siv120RegCfg.wbFluorescentTable   	= (S_REG_SET *)siv120a_regs_WBFluorescent;
	g_siv120RegCfg.wbCloudyTable        	= (S_REG_SET *)siv120a_regs_WBCloudy;
	g_siv120RegCfg.outFmtYUV422Table    	= (S_REG_SET *)siv120a_regs_fmtYUV422;
	g_siv120RegCfg.outFmtRGB565Table    	= (S_REG_SET *)siv120a_regs_fmtRGB565;
	g_siv120RegCfg.outSize640_480_Table 	= (S_REG_SET *)siv120a_regs_size640_480;
	g_siv120RegCfg.outSize352_288_Table 	= (S_REG_SET *)siv120a_regs_size352_288;
	g_siv120RegCfg.outSize320_240_Table 	= (S_REG_SET *)siv120a_regs_size320_240;
	g_siv120RegCfg.outSize176_144_Table 	= (S_REG_SET *)siv120a_regs_size176_144;
	g_siv120RegCfg.outSize160_120_Table 	= (S_REG_SET *)siv120a_regs_size160_120;
	g_siv120RegCfg.outSize88_72_Table   	= (S_REG_SET *)siv120a_regs_size88_72;
	g_siv120RegCfg.stillModeTable       	= (S_REG_SET *)siv120a_regs_stillMode;
	g_siv120RegCfg.opadTable            	= (S_REG_SET *)siv120a_regs_opad;

	siv120hw_data.fps = 15;     /* The preview frame has many horizontal stripes
				       if fps is set 30 in SIV120a sensor */

	/* allocte camera functions context */
	camera_functions = kzalloc(sizeof(camera_function_t), GFP_KERNEL);
	if (!camera_functions) {
		printk(KERN_ERR "siv120: Can't allocate buffer for camera functions structure \n");
		return -ENOMEM;
	}

	camera_functions->width_max 	= 640;
	camera_functions->height_max 	= 480;
	camera_functions->width_min 	= 88;
	camera_functions->height_min 	= 72;

	camera_functions->v_power 	= -1;     /* VAA 2.8V */
	camera_functions->v_io 		= -1;     /* VDD & VIO is controlled by GPIO */

	camera_functions->set_capture_format 	= siv120_set_capture_format;
	camera_functions->start_capture 	= siv120_start_capture;
	camera_functions->stop_capture 		= siv120_stop_capture;
	camera_functions->set_contrast 		= siv120_set_contrast;
	camera_functions->set_whitebalance 	= siv120_set_white_balance;
	camera_functions->set_exposure 		= siv120_set_exposure;

	camera_functions->get_framerate 	= siv120_get_framerate;
	camera_functions->set_framerate 	= siv120_set_framerate;

	camera_functions->name = "SIV120";
	
	//////////////////////////////////////////////////////////
	//No Problem	
	camera_functions->read_8bit 		= siv120_read_8bit;
	camera_functions->write_8bit 		= siv120_write_8bit;

	camera_functions->format_list 		= siv120_format_list;

	/*Don't change these, maybe two chips difference*/
	camera_functions->init 			= siv120a_init;
	camera_functions->set_power_mode 	= siv120a_set_power_mode;
	camera_functions->deinit 		= siv120a_deinit;
	camera_functions->sleep 		= siv120a_sleep;
	camera_functions->wakeup 		= siv120a_wake;

	camera_functions->read_16bit 	= NULL;
	camera_functions->write_16bit 	= NULL;
	camera_functions->read_32bit 	= NULL;
	camera_functions->write_32bit 	= NULL;
	camera_functions->set_zoom 	= NULL;
	camera_functions->set_framerate = NULL;
	camera_functions->set_autofocus = NULL;
	camera_functions->id 		= SENSOR_LOW;

	/*register siv120 as low resolution */
	if (sensor_register(camera_functions, id) < 0) {
		printk(KERN_ERR "siv120: sensor_register failed !\n");
		kfree(camera_functions);
		camera_functions = NULL;
		return -1;
	}

	return 0;
}

static int siv120a_unregister(int id)
{
	printk(KERN_ERR "siv120a_unregister\n");

	sensor_unregister(id);
	if (!camera_functions) {
		kfree(camera_functions);
		camera_functions = NULL;
	}

	return 0;
}

static int sivOpen(struct inode *inode, struct file *file)
{
	return 0;
}

static int sivRelease(struct inode *inode, struct file *file)
{
	return 0;
}

static int sivIoctl(struct inode *inode, struct file *filp,
		unsigned int cmd, unsigned long arg)
{
	unsigned short *pval, val16;
	unsigned char reg, val;
	int ret = 0;
	int preClockOn = bClockOn;
	u8  mode;
	u32 value;

	printk(KERN_EMERG "[sivIoctl] ioctl cmd %x \n", cmd);

	if ((cmd == FXCI_CAMERA_SIV_I2C_READ) || (cmd == FXCI_CAMEAR_SIV_I2C_WRITE)) {
		if (bClockOn == 0) {
			sensor_clock_on();
			siv120_power_down(CAMERA_POWER_FULL);
		}

		msleep(10);
	}
	switch (cmd) {
		case FXCI_CAMERA_EXPOSURE_SET:
			mode = (arg & 0xFF000000) >> 24;
			value = arg & 0x00FFFFFF;
			printk(KERN_EMERG "[sivIoctl] FXCI_CAMERA_EXPOSURE_SET mode: %d value %d\n", mode, value);
			if ((mode != SENSOR_AUTO_EXPOSURE) && (mode != SENSOR_MANUAL_EXPOSURE)) {
				ret = -1;
				goto end;
			}
			if (value > SENSOR_EXPOSURE_HIGHEST) {
				ret = -1;
				goto end;
			}
			cameraSettings.exposureMode = mode;
			cameraSettings.exposureValue = value;
			break;
		case FXCI_CAMERA_WHITE_BALANCE_SET:
			mode = (arg & 0xFF000000) >> 24;
			value = arg & 0x00FFFFFF;
			printk(KERN_EMERG "[sivIoctl] FXCI_CAMERA_EXPOSURE_SET mode: %d value %d\n", mode, value);
			if ((mode != SENSOR_AUTO_WHITEBALANCE) && (mode != SENSOR_MANUAL_WHITEBALANCE)) {
				ret = -1;
				goto end;
			}
			if (value > SENSOR_WHITEBALANCE_CLOUDY) {
				ret = -1;
				goto end;
			}
			cameraSettings.whiteBalanceMode = mode;
			cameraSettings.whiteBalanceValue = value;
			break;
		case FXCI_CAMERA_SIV_I2C_READ:
			pval = (unsigned short *)arg;
			reg = (unsigned char)(((*pval) & 0xFF00) >> 8);
			ret = i2c_smbus_read_byte_data(g_client, reg);
			if (ret < 0) {
				ret = -1;
				goto end;
			}
			val = (unsigned char)ret;
			printk(KERN_EMERG "[sivIoctl] read reg = %x, val=%x \n", reg, val);
			val16 = reg;
			val16 = (val16 << 8);
			val16 |= val;
			copy_to_user((void *)arg, &val16, sizeof(unsigned short));
			break;
		case FXCI_CAMEAR_SIV_I2C_WRITE:
			pval = (unsigned short *)arg;
			reg = (unsigned char)(((*pval) & 0xFF00) >> 8);
			val = (unsigned char)((*pval) & 0x00FF);
			printk(KERN_EMERG "[sivIoctl] write reg = %x, val=%x \n", reg, val);
			if (i2c_smbus_write_byte_data(g_client, reg, val) < 0) {
				ret = -1;
				goto end;
			}
			break;
		default:
			ret = -ENOTTY;
			goto end;
	}

end:
	if ((cmd == FXCI_CAMERA_SIV_I2C_READ) || (cmd == FXCI_CAMEAR_SIV_I2C_WRITE)) {
		if (preClockOn == 0) {
			siv120_power_down(CAMERA_POWER_OFF);
			sensor_clock_off();
		}
	}

	return ret;
}

static struct file_operations g_sivFops = {
	.owner =    THIS_MODULE,
	.ioctl =    sivIoctl,
	.open =     sivOpen,
	.release =  sivRelease,
};

//////////////////////////////////////////////////////////////////////////////////////////
static int siv120_register(int id, u8 chipVer)
{
	printk(KERN_ERR "siv120 register chip Version: 0x%02x\n",chipVer);
	switch (chipVer) {
		case SIV120A_CHIP_VER:
			siv120a_register(id);
			break;
		case SIV120B_CHIP_VER:
			siv120b_register(id);
			break;
		default:
			printk(KERN_ERR "siv120_register chip version 0x%02x is not valid!\n", chipVer);
			return -1;
	}

	return 0;
}

static int siv120_unregister(int id, u8 chipVer)
{
	printk(KERN_ERR "siv120 register chip Version: 0x%02x\n",chipVer);
	switch (chipVer) {
		case SIV120A_CHIP_VER:
			siv120a_unregister(id);
			break;
		case SIV120B_CHIP_VER:
			siv120b_unregister(id);
			break;
			/* other siv120 chipVer we didn't support will return error */
		default:
			printk(KERN_ERR "siv120_register chip version 0x%02x is not valid!\n", chipVer);
			return -1;
	}

	return 0;
}
//////////////////////////////////////////////////////////////////////////////////////////
static int __init siv120_i2c_probe(struct i2c_client *client)
{
	struct sensor_platform_data *pdata;
	int id = 0;

	int result = -1;
	dev_t devno = 0;
	int err = 0;

	printk(KERN_ERR "siv120_i2c_probe\n");

	pdata = client->dev.platform_data;
	id = pdata->id;

	g_client = client;

	if (sensor_clock_on() < 0) {
		g_client = NULL;
		return -EIO;
	}
	msleep(10);

	/* power on */
	siv120_power_down(CAMERA_POWER_FULL);
	msleep(5);

	/* chip id & init table */
	siv120_i2c_read_reg(CTRL_BLK, 0x01, &g_chipId);
	if (g_chipId != SIV120_CHIP_ID) {
		printk(KERN_ERR "siv120 detect failed chip id is 0x%02x\n", g_chipId);
		return -EIO;
	}
	printk(KERN_ERR "siv120 detect chip id is 0x%02x\n", g_chipId);
	msleep(50);

	siv120_i2c_read_reg(CTRL_BLK, 0x02, &g_chipVer);
	printk(KERN_ERR "siv120 detect chip id is 0x%02x, chip version is 0x%02x\n", g_chipId, g_chipVer);
	msleep(50);

	if (siv120_register(id, g_chipVer) < 0)
		goto err;

	/* power off */
	siv120_power_down(CAMERA_POWER_OFF);
	msleep(5);

	sensor_clock_off();

	/* prepare device id */
	devno = MKDEV(g_sivMajor, g_sivMinor);
	result = register_chrdev_region(devno, g_sivNrDev, "fxmfg_siv120");

	if (result < 0) {
		printk(KERN_WARNING "[siv120_i2c_probe]: can't get major %d\n", g_sivMajor);
		return result;
	}
	printk(KERN_EMERG "[siv120_i2c_probe] major %d minor %d\n", g_sivMajor, g_sivMinor);

	g_psivDev = kmalloc(sizeof(struct sivCDev), GFP_KERNEL);
	if (!g_psivDev) {
		result = -ENOMEM;
		goto fail_malloc;
	}
	memset(g_psivDev, 0, sizeof(struct sivCDev));

	/* setup cdev */
	cdev_init(&g_psivDev->cdev, &g_sivFops);
	g_psivDev->cdev.owner = THIS_MODULE;
	g_psivDev->cdev.ops = &g_sivFops;
	err = cdev_add(&g_psivDev->cdev, devno, 1);
	printk(KERN_EMERG "[siv120_i2c_probe] cdev_add result %d \n", err);

	/* Fail gracefully if need be */
	if (err) {
		printk(KERN_EMERG "[siv120_i2c_probe] cdev_add Error %d", err);
		result = -ENOMEM;
		goto fail_setup;
	}

	siv120_sysfs_add(&client->dev.kobj);

	return 0;
err:
	printk(KERN_ERR "SIV120A detect failed.\n");
	sensor_clock_off();
	g_client = NULL;

	return -1;
fail_setup:
	kfree(g_psivDev);
fail_malloc:
	unregister_chrdev_region(devno, g_sivNrDev);
	return result;
}

static int siv120_i2c_remove(struct i2c_client *client)
{
	struct sensor_platform_data *pdata;
	pdata = client->dev.platform_data;

	printk(KERN_ERR "siv120_i2c_remove\n");

	siv120_unregister(pdata->id, g_chipVer);
	siv120_sysfs_rm(&client->dev.kobj);

	g_client = NULL;
	return 0;
}

static struct i2c_driver siv120_i2c_driver = {
	.driver = {
		.name  = "sensor_siv120",
	},
	.probe  = siv120_i2c_probe,
	.remove = siv120_i2c_remove,
};

static int __init siv120_module_init(void)
{
	int ret;

	printk(KERN_ERR "siv120_module_init\n");

	siv120hw_data_init();

	if (ci_get_p_add(0) == 0) {
		printk(KERN_ERR "siv120_module_init fail (should init camera frist)\n");
		return -EPERM;
	}

	if ((ret = i2c_add_driver(&siv120_i2c_driver)) < 0) {
		printk(KERN_ERR "siv120: Driver registration failed,"
				"module not inserted.\n");
		return ret;
	}
#ifdef CONFIG_PXA3xx_DVFM
	dvfm_register("siv120", &dvfm_dev_idx);
#endif
	return 0;
}

static void __exit siv120_module_exit(void)
{
	printk(KERN_ERR "siv120_module_exit\n");
#ifdef CONFIG_PXA3xx_DVFM
	dvfm_unregister("siv120", &dvfm_dev_idx);
#endif

	if (ci_get_p_add(0) == 0)
		return;

	i2c_del_driver(&siv120_i2c_driver);
}

MODULE_DESCRIPTION("SIV120A I2C Client driver");
MODULE_LICENSE("GPL");
module_init(siv120_module_init);
module_exit(siv120_module_exit);
