/*
 *  linux/drivers/media/video/pxa950_camera.c - PXA9XX MMCI driver
 *
 *  Based on linux/drivers/media/video/cafe_ccic.c
 *
 *  Copyright:	(C) Copyright 2008 Marvell International Ltd.
 *              Mingwei Wang <mwwang@marvell.com>
 *
 * A driver for the CMOS camera controller in the Marvell 88ALP01 "cafe"
 * multifunction chip.  Currently works with the Omnivision OV7670
 * sensor.
 *
 * The data sheet for this device can be found at:
 *    http://www.marvell.com/products/pcconn/88ALP01.jsp
 *
 * Copyright 2006 One Laptop Per Child Association, Inc.
 * Copyright 2006-7 Jonathan Corbet <corbet@lwn.net>
 *
 * Written by Jonathan Corbet, corbet@lwn.net.
 *
 * This file may be distributed under the terms of the GNU General
 * Public License, version 2.
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/fs.h>
#include <linux/i2c.h>
#include <linux/interrupt.h>
#include <linux/spinlock.h>
#include <linux/videodev2.h>
#include <media/v4l2-ioctl.h>
#include <media/v4l2-common.h>
#include <media/v4l2-chip-ident.h>
#include <linux/device.h>
#include <linux/wait.h>
#include <linux/list.h>
#include <linux/dma-mapping.h>
#include <linux/delay.h>
#include <linux/debugfs.h>
#include <linux/jiffies.h>

#include <linux/vmalloc.h>
#include <linux/platform_device.h>

#include <linux/uaccess.h>
#include <linux/io.h>

#include <linux/clk.h>
#include <mach/pxa-regs.h>
#include <mach/dma.h>
#include <mach/camera.h>
#include <asm/mach-types.h>

#define CCIC_VERSION 0x000001
/*
 * Parameters.
 */
MODULE_AUTHOR("Jonathan Corbet <corbet@lwn.net>");
MODULE_DESCRIPTION("Marvell 88ALP01 CMOS Camera Controller driver");
MODULE_LICENSE("GPL");
MODULE_SUPPORTED_DEVICE("Video");

/*
 * Registers
 */

/*
 * CSI register base: 0x50020000
 */
#define REG_CSSCR		0x0000
#define REG_CSGCR		0x000C
#define REG_CSxCR0		0x0010
#define REG_CSxSR		0x0014
#define REG_CSxINEN		0x0018
#define REG_CSxINST		0x001C
#define REG_CSxTIM0		0x0020
#define REG_CSxTIM1		0x0024
#define REG_CSxGENDAT	0x0028
#define REG_CSxPHYCAL	0x002C

/* CSxCR0 */
#define CSxCR0_CSIEN	0x0001

/* CAM_CSI_CONT_CSxCR0 bits */
#define CAM_CSI_CONT_CSI_EN            (0x1u<<0)  /* CSI controller Enable/Disable*/
#define CAM_CSI_CONT_NOL(n)            ((n)<<2)   /* Number of Active Lanes (2 bits)*/
#define CAM_CSI_CONT_VC0_CFG(n)        ((n)<<5)   /* Chan 0 Addr Conf (2 bits)*/
#define CAM_CSI_CONT_VC1_CFG(n)        ((n)<<7)   /* Chan 1 Addr Conf (2 bits)*/
#define CAM_CSI_CONT_VC2_CFG(n)        ((n)<<9)   /* Chan 2 Addr Conf (2 bits)*/
#define CAM_CSI_CONT_VC3_CFG(n)        ((n)<<11)  /* Chan 3 Addr Conf (2 bits)*/


/*CAM_CSI_CONT_CSxTIM0 bits*/
#define CAM_CSI_CONT_CLTERMEN(n)       ((n)<<0)   /* Time to wait before enabling clock HS termination (8 bits)*/
#define CAM_CSI_CONT_CLSETTLE(n)       ((n)<<8)   /* Time to wait before HS clock is valid (8 bits)*/
#define CAM_CSI_CONT_CLMISS(n)         ((n)<<16)  /* Time to detect that the clock has stopped toggling (8 bits)*/
#define CAM_CSI_CONT_HSTERMEN(n)       ((n)<<24)  /* Time to wait before enabling data HS termination (8 bits)*/

/* CAM_CSI_CONT_CSxTIM1 bits */
#define CAM_CSI_CONT_HS_Rx_TO(n)       ((n)<<0)   /* Time to wait before declaring an error on a packet reception. This counter counts on the escape mode clock which is 52 MHz (16 bits)*/
#define CAM_CSI_CONT_HSTSETTLE(n)      ((n)<<16)  /* Timeout at RX end to neglect transition effects (8 bits)*/

/* CAM_CSI_CONT_CSxPHYCAL bits */
#define CAM_CSI_CONT_MIPI_BG_VREF_EN   (0x1u<<0)  /* See DPHY specifications for details*/
#define CAM_CSI_CONT_MIPI_RCOMP_CLKSEL (0x1u<<1)  /* Used to control the source of the PHY calibration clock (External/Internal).*/
#define CAM_CSI_CONT_MIPI_RCOMP_LOAD   (0x1u<<2)  /* Used to enable loading the MIPI_REN bypass value*/
#define CAM_CSI_CONT_MIPI_REN_BYPASS(n) ((n)<<3)  /* An 8-bit value to bypass the internal MIPI_REN value.*/
#define CAM_CSI_CONT_MIPI_RCOMP_CLK_EXT (0x1u<<31) /* Used tp provide a SW generated extternal clock to the PHY calibration block.*/

/* CAM_CSI_MCLK_CSGCR bits */
#define CAM_CSI_CLK_DIV(n)             ((n)<<0)  /* Generated Clock Divisor. GCLK = 1/(DIV+2). 12 bits*/
#define CAM_CSI_CLK_GCLK_EN            (0x1u<<16)/* Generated Clock Enable*/

/* CAM_CSI_CONT_CSxINEN bits */
#define CAM_CSI_CONT_CSI_EN_INT        (0x1u<<0)  /* ENABLE Interrupt Enable*/
#define CAM_CSI_CONT_SoF_INT_EN        (0x1u<<1)  /* Start of Frame Interrupt Enable*/
#define CAM_CSI_CONT_EoF_INT_EN        (0x1u<<2)  /* End of Frame Interrupt Enable*/
#define CAM_CSI_CONT_OFLOW_INT_EN      (0x1u<<3)  /* Overflow Interrupt Enable*/
#define CAM_CSI_CONT_PHY_ERR_EN        (0x1u<<4)  /* D-PHY-related Error Interrupt Enable*/
#define CAM_CSI_CONT_GEN_PACK_INT_EN   (0x1u<<5)  /* Generic Packet Interrupt Enable*/
#define CAM_CSI_CONT_TIMEOUT_EN        (0x1u<<8)  /* Timeout Interrupt Enable*/
#define CAM_CSI_CONT_PROT_EN           (0x1u<<12) /* Protection Interupt Enable*/

/*
 * SCI register
 * SCI0 base: 0x50000000
 */
#define REG_SCICR0	0x0000
#define REG_SCICR1	0x0004
#define REG_SCISR	0x0008
#define REG_SCIMASK	0x000C
#define REG_SCIFIFO	0x00F8
#define REG_SCIFIFOSR	0x00FC
#define REG_SCIDADDR0	0x0200
#define REG_SCISADDR0	0x0204
#define REG_SCITADDR0	0x0208
#define REG_SCIDCMD0	0x020C
#define REG_SCIDADDR1	0x0210
#define REG_SCISADDR1	0x0214
#define REG_SCITADDR1	0x0218
#define REG_SCIDCMD1	0x021C
#define REG_SCIDADDR2	0x0220
#define REG_SCISADDR2	0x0224
#define REG_SCITADDR2	0x0228
#define REG_SCIDCMD2	0x022C
#define REG_SCIDBR0	0x0300
#define REG_SCIDCSR0	0x0304
#define REG_SCIDBR1	0x0310
#define REG_SCIDCSR1	0x0314
#define REG_SCIDBR2	0x0320
#define REG_SCIDCSR2	0x0324

#define	IRQ_EOFX	0x00000001
#define	IRQ_EOF		0x00000002
#define	IRQ_DIS		0x00000004
#define	IRQ_OFO		0x00000008
#define	IRQ_DBS		0x00000200
#define	IRQ_SOFX	0x00000400
#define	IRQ_SOF		0x00000800
#define	IRQ_FRAME 	(IRQ_EOF|IRQ_EOFX|IRQ_SOF|IRQ_SOFX)


/* FMT_YUV420 & FMT_RGB666 only for input format without output format */
#define   FMT_RAW8	  0x0000
#define   FMT_RAW10	  0x0002
#define   FMT_YUV422	  0x0003
#define   FMT_YUV420	  0x0004
#define   FMT_RGB565	  0x0005
#define   FMT_RGB666	  0x0006
#define   FMT_RGB888	  0x0007
#define   FMT_JPEG	  0x0008

#define SCICR1_FMT_OUT(n)       ((n) <<12)	/* Output Pixel Format (4 bits) */
#define SCICR1_FMT_IN(n)        ((n) <<28)	/* Input Pixel Format (4 bits) */

/* REG_SCIDCSR0 */
#define SCIDCSR_DMA_RUN        0x80000000  	/* DMA Channel Enable */

/* REG_SCICR0 */
#define SCICR0_CAP_EN          0x40000000 	/* Capture Enable */
#define SCICR0_CI_EN           0x80000000 	/* Camera Interface Enable */

/* REG_SCIFIFO */
#define SCIFIFO_Fx_EN        0x0010	/* FIFO 0 Enable */

#define SCIFIFO_F0_EN        0x0010	/* FIFO 0 Enable */
#define SCIFIFO_F1_EN        0x0020	/* FIFO 1 Enable */
#define SCIFIFO_F2_EN        0x0040	/* FIFO 2 Enable */
#define SCIFIFO_FX_EN(n)  ((n) <<4)		/* Input Pixel Format (4 bits) */

#define VGA_WIDTH       640
#define VGA_HEIGHT      480

#define SENSOR_MAX 2


/*
 * Internal DMA buffer management.  Since the controller cannot do S/G I/O,
 * we must have physically contiguous buffers to bring frames into.
 * These parameters control how many buffers we use, whether we
 * allocate them at load time (better chance of success, but nails down
 * memory) or when somebody tries to use the camera (riskier), and,
 * for load-time allocation, how big they should be.
 *
 * The controller can cycle through three buffers.  We could use
 * more by flipping pointers around, but it probably makes little
 * sense.
 */

#define DMA_POOL 0

/* csi only accept 422 and transfer to 420 */
#define CONVERT_420_422 1
#define CAM_BRIDGE

static int alloc_bufs_at_read = 0;
module_param(alloc_bufs_at_read, bool, 0444);
MODULE_PARM_DESC(alloc_bufs_at_read,
		"Non-zero value causes DMA buffers to be allocated when the "
		"video capture device is read, rather than at module load "
		"time.  This saves memory, but decreases the chances of "
		"successfully getting those buffers.");

static int n_dma_bufs = 2;	/*frame buffers number*/
module_param(n_dma_bufs, uint, 0644);
MODULE_PARM_DESC(n_dma_bufs,
		"The number of DMA buffers to allocate.  Can be either two "
		"(saves memory, makes timing tighter) or three.");

static int dma_buf_size = VGA_WIDTH * VGA_HEIGHT * 2;  /* Worst case */
module_param(dma_buf_size, uint, 0444);
MODULE_PARM_DESC(dma_buf_size,
		"The size of the allocated DMA buffers.  If actual operating "
		"parameters require larger buffers, an attempt to reallocate "
		"will be made.");

static int min_buffers = 1;
module_param(min_buffers, uint, 0644);
MODULE_PARM_DESC(min_buffers,
		"The minimum number of streaming I/O buffers we are willing "
		"to work with.");

static int max_buffers = 10;
module_param(max_buffers, uint, 0644);
MODULE_PARM_DESC(max_buffers,
		"The maximum number of streaming I/O buffers an application "
		"will be allowed to allocate.  These buffers are big and live "
		"in vmalloc space.");

static int flip = 0;
module_param(flip, bool, 0444);
MODULE_PARM_DESC(flip,
		"If set, the sensor will be instructed to flip the image "
		"vertically.");


enum ccic_state {
	S_NOTREADY,	/* Not yet initialized */
	S_IDLE,		/* Just hanging around */
	S_FLAKED,	/* Some sort of problem */
	S_SINGLEREAD,	/* In read() */
	S_SPECREAD,   	/* Speculative read (for future read()) */
	S_STREAMING	/* Streaming data */
};

/*
 * Tracking of streaming I/O buffers.
 */
struct ccic_sio_buffer {
	struct list_head list;
	struct v4l2_buffer v4lbuf;
	char *buffer;   /* Where it lives in kernel space */
	int mapcount;
	struct ccic_camera *cam;
};

/* descriptor needed for the PXA DMA engine */
struct pxa_cam_dma {
	dma_addr_t		sg_dma;
	struct pxa_dma_desc	*sg_cpu;
	size_t			sg_size;
	int                     sglen;
};

#define CHANNEL_NUM 3 		/*YUV*/
#define MAX_DMA_BUFS 3

struct pxa_buf_node {
	int order;	/* Internal buffer addresses */
	void *dma_bufs;	/* Internal buffer addresses */
	dma_addr_t dma_handles; /* Buffer bus addresses */

	struct pxa_cam_dma dma_desc[CHANNEL_NUM];
};

struct ccic_csi
{
	unsigned char __iomem *regs;
	int irq;

	spinlock_t dev_lock;  /* Access to device */
};

struct ccic_csi *g_csi;

/*
 * A description of one of our devices.
 * Locking: controlled by s_mutex.  Certain fields, however, require
 * 	    the dev_lock spinlock; they are marked as such by comments.
 *	    dev_lock is also required for access to device registers.
 */
struct ccic_camera
{
	enum ccic_state state;
	unsigned long flags;   		/* Buffer status, mainly (dev_lock) */
	int users;			/* How many open FDs */
	struct file *owner;		/* Who has data access (v4l2) */

	/*
	 * Subsystem structures.
	 */
	int irq;
	struct platform_device *pdev;
	struct video_device v4ldev;
	struct i2c_adapter i2c_adapter;
	struct i2c_client *sensor;
	struct i2c_client *sensors[SENSOR_MAX];

	unsigned char __iomem *regs;
	struct list_head dev_list;	/* link to other devices */

	/* DMA buffers */
	unsigned int nbufs;		/* How many are alloc'd */
	int next_buf;			/* Next to consume (dev_lock) */
	unsigned int dma_buf_size;  	/* allocated size */

	unsigned int specframes;	/* Unconsumed spec frames (dev_lock) */
	unsigned int sequence;		/* Frame sequence number */
	unsigned int buf_seq[MAX_DMA_BUFS]; /* Sequence for individual buffers */

	/* Streaming buffers */
	unsigned int n_sbufs;		/* How many we have */
	struct ccic_sio_buffer *sb_bufs; /* The array of housekeeping structs */
	struct list_head sb_avail;	/* Available for data (we own) (dev_lock) */
	struct list_head sb_full;	/* With data (user space owns) (dev_lock) */
	struct tasklet_struct s_tasklet;

	/* Current operating parameters */
	u32 sensor_type;		/* Currently ov7670 only */
	struct v4l2_pix_format pix_format;

	/* Locks */
	struct mutex s_mutex; /* Access to this structure */
	spinlock_t dev_lock;  /* Access to device */

	/* Misc */
	wait_queue_head_t iowait;	/* Waiting on frame data */


	struct clk *clk;

	struct pxa_buf_node buf_node[MAX_DMA_BUFS];
	int channels;
	unsigned int channel_size[CHANNEL_NUM];

	/* platform data */
	struct cam_platform_data *platform_ops;

};

/* Len should 8 bytes align, bit[2:0] should be 0 */
#define SINGLE_DESC_TRANS_MAX   (1 << 24)


/*
 * Status flags.  Always manipulated with bit operations.
 */
#define CF_BUF0_VALID	 0	/* Buffers valid - first three */
#define CF_BUF1_VALID	 1
#define CF_BUF2_VALID	 2
#define CF_DMA_ACTIVE	 3	/* A frame is incoming */
#define CF_CONFIG_NEEDED 4	/* Must configure hardware */

static int detected_high = 0;
static int detected_low = 0;
static int sensor_selected = 2;

/*
 * Start over with DMA buffers - dev_lock needed.
 */
static void ccic_reset_buffers(struct ccic_camera *cam)
{
	int i;

	cam->next_buf = -1;
	for (i = 0; i < cam->nbufs; i++)
		clear_bit(i, &cam->flags);
	cam->specframes = 0;
}

static inline int ccic_needs_config(struct ccic_camera *cam)
{
	return test_bit(CF_CONFIG_NEEDED, &cam->flags);
}

static void ccic_set_config_needed(struct ccic_camera *cam, int needed)
{
	if (needed)
		set_bit(CF_CONFIG_NEEDED, &cam->flags);
	else
		clear_bit(CF_CONFIG_NEEDED, &cam->flags);
}




/*
 * Debugging and related.
 */
#define cam_err(cam, fmt, arg...) \
	dev_err(&(cam)->pdev->dev, fmt, ##arg);
#define cam_warn(cam, fmt, arg...) \
	dev_warn(&(cam)->pdev->dev, fmt, ##arg);
#define cam_dbg(cam, fmt, arg...) \
	dev_dbg(&(cam)->pdev->dev, fmt, ##arg);


/* ---------------------------------------------------------------------*/
/*
 * We keep a simple list of known devices to search at open time.
 */
static LIST_HEAD(ccic_dev_list);
static DEFINE_MUTEX(ccic_dev_list_lock);
static int __ccic_cam_cmd(struct ccic_camera *cam, int cmd, void *arg);
static int __ccic_cam_reset(struct ccic_camera *cam);

static void ccic_add_dev(struct ccic_camera *cam)
{
	mutex_lock(&ccic_dev_list_lock);
	list_add_tail(&cam->dev_list, &ccic_dev_list);
	mutex_unlock(&ccic_dev_list_lock);
}

static void ccic_remove_dev(struct ccic_camera *cam)
{
	mutex_lock(&ccic_dev_list_lock);
	list_del(&cam->dev_list);
	mutex_unlock(&ccic_dev_list_lock);
}

static struct ccic_camera *ccic_find_dev(int minor)
{
	struct ccic_camera *cam;

	mutex_lock(&ccic_dev_list_lock);
	list_for_each_entry(cam, &ccic_dev_list, dev_list) {
		if (cam->v4ldev.minor == minor)
			goto done;
	}
	cam = NULL;
  done:
	mutex_unlock(&ccic_dev_list_lock);
	return cam;
}

static struct ccic_camera *ccic_find_by_pdev(struct platform_device *pdev)
{
	struct ccic_camera *cam;

	mutex_lock(&ccic_dev_list_lock);
	list_for_each_entry(cam, &ccic_dev_list, dev_list) {
		if (cam->pdev == pdev)
			goto done;
	}
	cam = NULL;
  done:
	mutex_unlock(&ccic_dev_list_lock);
	return cam;
}


/* ------------------------------------------------------------------------ */
/*
 * Device register I/O
 */
static inline void ccic_reg_write(struct ccic_camera *cam, unsigned int reg,
		unsigned int val)
{
	__raw_writel(val, cam->regs + reg);
}

static inline unsigned int ccic_reg_read(struct ccic_camera *cam,
		unsigned int reg)
{
	return __raw_readl(cam->regs + reg);
}


static inline void ccic_reg_write_mask(struct ccic_camera *cam, unsigned int reg,
		unsigned int val, unsigned int mask)
{
	unsigned int v = ccic_reg_read(cam, reg);

	v = (v & ~mask) | (val & mask);
	ccic_reg_write(cam, reg, v);
}

static inline void ccic_reg_clear_bit(struct ccic_camera *cam,
		unsigned int reg, unsigned int val)
{
	ccic_reg_write_mask(cam, reg, 0, val);
}

static inline void ccic_reg_set_bit(struct ccic_camera *cam,
		unsigned int reg, unsigned int val)
{
	ccic_reg_write_mask(cam, reg, val, val);
}

/*provided for sensor calling to enable clock at begining of probe*/
void ccic_set_clock(unsigned int reg, unsigned int val)
{
	struct ccic_camera *cam;
	cam = ccic_find_dev(0);
	ccic_reg_write(cam, reg, val);
}

unsigned int ccic_get_clock(unsigned int reg)
{
        struct ccic_camera *cam;
        cam = ccic_find_dev(0);
	return ccic_reg_read(cam, reg);
}
/*
	1. APMU CCIC clock reset control
	2. CCIC power on
	3. enable MCLK
*/

/* Somebody is on the bus */
static int ccic_cam_init(struct ccic_camera *cam);
static void ccic_ctlr_stop_dma(struct ccic_camera *cam);

static void csi_start(struct ccic_csi *csi);
static void csi_stop(struct ccic_csi *csi);
static void dump_csi_registers(struct ccic_csi *csi);
static void dump_registers(struct ccic_camera *cam);



int ccic_sensor_attach(struct i2c_client *client)
{
	struct ccic_camera *cam;
	int ret;
	struct v4l2_dbg_chip_ident chip = { V4L2_CHIP_MATCH_I2C_ADDR, 0, 0 };;

	cam = ccic_find_dev(0);		/*TODO - only support one camera controller*/
	if (cam == NULL) {
		printk(KERN_ERR "didn't find camera device!\n");
		return -ENODEV;
	}
	/*
	 * Don't talk to chips we don't recognize.
	 */

	mutex_lock(&cam->s_mutex);
	cam->sensor = client;
	ret = __ccic_cam_reset(cam);
	if (ret)
		goto out;
	chip.match.type = V4L2_CHIP_MATCH_I2C_ADDR;
	chip.match.addr = client->addr;
        ret = __ccic_cam_cmd(cam, VIDIOC_DBG_G_CHIP_IDENT, &chip);
	if (ret)
		goto out;
	cam->sensor_type = chip.ident;

	if ((cam->sensor_type == V4L2_IDENT_OV7690)
		|| (cam->sensor_type == V4L2_IDENT_OV7670)) {
		cam->sensors[SENSOR_LOW] = client;
		detected_low = 1;
	} else if (cam->sensor_type == V4L2_IDENT_OV3640) {
		cam->sensors[SENSOR_HIGH] = client;
		detected_high = 1;
	} else {
                cam_err(cam, "Unsupported sensor type %d\n", cam->sensor_type);
                ret = -EINVAL;
                goto out;
	}
	/* Get/set parameters? */
        ret = 0;
        cam->state = S_IDLE;
out:
	cam->sensor = NULL;
        mutex_unlock(&cam->s_mutex);
        return ret;

}

static int ccic_sensor_detach(struct i2c_client *client)
{
	struct ccic_camera *cam;

	cam = ccic_find_dev(0);
	if (cam == NULL)
		return -ENODEV;

	if (cam->sensor == client) {
		ccic_ctlr_stop_dma(cam);
		cam_err(cam, "lost the sensor!\n");
		cam->sensor = NULL;  /* Bummer, no camera */
		cam->state = S_NOTREADY;
	}
	return 0;
}

/* ------------------------------------------------------------------- */
/*
 * Deal with the controller.
 */
static int ccic_init_dma_desc(struct ccic_camera *cam)
{
	int i = 0;
	int index;
	unsigned int len = 0, len_tmp = 0;

	pxa_dma_desc 	*dma_desc_tmp;
	unsigned long 	dma_desc_p_tmp;
	unsigned long srcphyaddr, dstphyaddr;
	struct pxa_cam_dma *pxa_dma;

	for (index = 0; index < cam->nbufs; index++) {

		srcphyaddr = 0;	/* TBD */
		dstphyaddr = cam->buf_node[index].dma_handles;

		for (i = 0; i < cam->channels; i++) {
			pxa_dma = &cam->buf_node[index].dma_desc[i];
			len = cam->channel_size[i];
			pxa_dma->sglen = (len + SINGLE_DESC_TRANS_MAX - 1) / SINGLE_DESC_TRANS_MAX;
			pxa_dma->sg_size = (pxa_dma->sglen) * sizeof(struct pxa_dma_desc);
			pxa_dma->sg_cpu = dma_alloc_coherent(&(cam->pdev->dev), pxa_dma->sg_size,
					     &pxa_dma->sg_dma, GFP_KERNEL);

			if (!pxa_dma->sg_cpu)
				goto err;

			dma_desc_tmp = pxa_dma->sg_cpu;
			dma_desc_p_tmp = pxa_dma->sg_dma;

			while (len) {
				len_tmp = len > SINGLE_DESC_TRANS_MAX ?
					SINGLE_DESC_TRANS_MAX : len;

				dma_desc_tmp->ddadr = dma_desc_p_tmp + sizeof(pxa_dma_desc);
				dma_desc_tmp->dsadr = srcphyaddr; /* TBD */
				dma_desc_tmp->dtadr = dstphyaddr;
				dma_desc_tmp->dcmd = len_tmp;

				len -= len_tmp;
				dma_desc_tmp++;
				dma_desc_p_tmp += sizeof(pxa_dma_desc);
				dstphyaddr += len_tmp;

			}
		}
	}

	return 0;

err:
	for (index = 0; index<cam->nbufs; index++) {
		for (i = 0; i < cam->channels; i++) {
			pxa_dma = &cam->buf_node[index].dma_desc[i];
			if (pxa_dma->sg_cpu)
				dma_free_coherent(&cam->pdev->dev, pxa_dma->sg_size,
					    pxa_dma->sg_cpu, pxa_dma->sg_dma);
		}
	}

	return -ENOMEM;
}


static void ccic_link_dma_desc(struct ccic_camera *cam)
{
	int i = 0;
	int index = 0;
	struct pxa_cam_dma *pxa_dma, *pxa_dma_ex;	

	for (index = 0; index<cam->nbufs; index++) {	
		for (i = 0; i < cam->channels; i++) {
			pxa_dma = &cam->buf_node[index].dma_desc[i];
			
			if (0 == index) {
				/* fist, load to register */
				ccic_reg_write(cam, REG_SCIDADDR0 + i*0x10, pxa_dma->sg_dma);
			} else {
				/* link to ex descriptor */
				pxa_dma_ex = &cam->buf_node[index-1].dma_desc[i];
				pxa_dma_ex->sg_cpu[pxa_dma_ex->sglen-1].ddadr = (u32)pxa_dma->sg_dma;
			}
			
			if ((cam->nbufs-1) == index) {
				/* loop back */
				pxa_dma->sg_cpu[pxa_dma->sglen-1].ddadr =
					(u32)cam->buf_node[0].dma_desc[i].sg_dma;
			}
		}
	}
}

static void ccic_ctlr_dma(struct ccic_camera *cam)
{
	ccic_init_dma_desc(cam);
	ccic_link_dma_desc(cam);
}

static void ccic_ctlr_image(struct ccic_camera *cam)
{
	struct v4l2_pix_format *fmt = &cam->pix_format;
	unsigned int size = fmt->width*fmt->height;

	switch (fmt->pixelformat) {
		
	case V4L2_PIX_FMT_RGB565:
		cam->channels = 1;
		cam->channel_size[0] = size*2;
		ccic_reg_write(cam, REG_SCICR1, SCICR1_FMT_IN(FMT_RGB565) | SCICR1_FMT_OUT(FMT_RGB565));
	    break;

	case V4L2_PIX_FMT_JPEG:
		cam->channels = 1;
		cam->channel_size[0] = size;
		ccic_reg_write(cam, REG_SCICR1, SCICR1_FMT_IN(FMT_JPEG) | SCICR1_FMT_OUT(FMT_JPEG));
	    break;
	
	case V4L2_PIX_FMT_YUV422P:

		cam->channels = 3;
		cam->channel_size[0] = size;
		cam->channel_size[1] = cam->channel_size[2] = size/2;
		ccic_reg_write(cam, REG_SCICR1, SCICR1_FMT_IN(FMT_YUV422) | SCICR1_FMT_OUT(FMT_YUV422));
		break;
	
	case V4L2_PIX_FMT_YUV420:
		/* only accept YUV422 as input */
		cam->channels = 3;
		cam->channel_size[0] = size;
		cam->channel_size[1] = cam->channel_size[2] = size/4;
#if CONVERT_420_422
		ccic_reg_write(cam, REG_SCICR1, SCICR1_FMT_IN(FMT_YUV422) | SCICR1_FMT_OUT(FMT_YUV420));
#else
		ccic_reg_write(cam, REG_SCICR1, SCICR1_FMT_IN(FMT_YUV420) | SCICR1_FMT_OUT(FMT_YUV420));
#endif
		break;

	default:
		break;
	}

}


/*
 * Configure the controller for operation; caller holds the
 * device mutex.
 */
static int ccic_ctlr_configure(struct ccic_camera *cam)
{
	unsigned long flags;

	spin_lock_irqsave(&cam->dev_lock, flags);
	ccic_ctlr_image(cam);
	ccic_ctlr_dma(cam);
	ccic_set_config_needed(cam, 0);
	spin_unlock_irqrestore(&cam->dev_lock, flags);
	return 0;
}

static void ccic_sci_irq_enable(struct ccic_camera *cam)
{
	/*
	 * Clear any pending interrupts, since we do not
	 * expect to have I/O active prior to enabling.
	 */

	ccic_reg_write(cam, REG_SCISR, ccic_reg_read(cam, REG_SCISR));
	ccic_reg_clear_bit(cam, REG_SCIMASK, IRQ_FRAME);
}

static void ccic_sci_irq_disable(struct ccic_camera *cam)
{
	ccic_reg_set_bit(cam, REG_SCIMASK, IRQ_FRAME);
}

/*
 * Make the controller start grabbing images.  Everything must
 * be set up before doing this.
 */
static void ccic_sci_start(struct ccic_camera *cam)
{
	int i = 0;
	unsigned int val = 0;

	/* start_fifo */
	for (i = 0; i < cam->channels; i++) {
		val = SCIFIFO_F0_EN << i;
		ccic_reg_set_bit(cam, REG_SCIFIFO, val);
	}

	/* start_dma */
	for (i = 0; i < cam->channels; i++)
		ccic_reg_set_bit(cam, REG_SCIDCSR0 + i*0x10, SCIDCSR_DMA_RUN);

	/* start sci */	
	ccic_reg_set_bit(cam, REG_SCICR0, SCICR0_CAP_EN | SCICR0_CI_EN);
}

static void ccic_sci_stop(struct ccic_camera *cam)
{
	int i = 0;
	unsigned int val = 0;

	/* stop_fifo */
	for (i = 0; i < cam->channels; i++) {
		val = SCIFIFO_F0_EN << i;
		ccic_reg_clear_bit(cam, REG_SCIFIFO, val);	
	}	

	/* stop_dma */
	for (i = 0; i < cam->channels; i++)
		ccic_reg_clear_bit(cam, REG_SCIDCSR0 + i*0x10, SCIDCSR_DMA_RUN);

	/* stop sci */	
	ccic_reg_clear_bit(cam, REG_SCICR0, SCICR0_CAP_EN | SCICR0_CI_EN);
}

void ccic_ctlr_init(struct ccic_camera *cam)
{
	unsigned long flags;

	spin_lock_irqsave(&cam->dev_lock, flags);

	/*
	 * Turn off the enable bit.  It sure should be off anyway,
	 * but it's good to be sure.
	 */
	ccic_reg_clear_bit(cam, REG_SCICR0, SCICR0_CI_EN);
	/*
	 * Mask all interrupts.
	 */
	ccic_reg_write(cam, REG_SCIMASK, ~0);
	/*
	 * Clock the sensor appropriately.  Controller clock should
	 * be 48MHz, sensor "typical" value is half that.
	 */
	spin_unlock_irqrestore(&cam->dev_lock, flags);
}
/*
 * Stop the controller, and don't return until we're really sure that no
 * further DMA is going on.
 */
static void ccic_ctlr_stop_dma(struct ccic_camera *cam)
{
	unsigned long flags;

	/*
	 * Theory: stop the camera controller (whether it is operating
	 * or not).  Delay briefly just in case we race with the SOF
	 * interrupt, then wait until no DMA is active.
	 */

	csi_stop(g_csi);
	
	spin_lock_irqsave(&cam->dev_lock, flags);
	ccic_sci_stop(cam);
	spin_unlock_irqrestore(&cam->dev_lock, flags);
	
	wait_event_timeout(cam->iowait,
			!test_bit(CF_DMA_ACTIVE, &cam->flags), HZ);
	if (test_bit(CF_DMA_ACTIVE, &cam->flags))
		cam_err(cam, "Timeout waiting for DMA to end\n");
		/* This would be bad news - what now? */
	spin_lock_irqsave(&cam->dev_lock, flags);

	cam->state = S_IDLE;
	ccic_sci_irq_disable(cam);
	spin_unlock_irqrestore(&cam->dev_lock, flags);
}

/* -------------------------------------------------------------------- */
/*
 * Communications with the sensor.
 */

static int __ccic_cam_cmd(struct ccic_camera *cam, int cmd, void *arg)
{
	struct i2c_client *sc = cam->sensor;
	int ret;

	if (sc == NULL || sc->driver == NULL || sc->driver->command == NULL)
		return -EINVAL;
	ret = sc->driver->command(sc, cmd, arg);
	if (ret == -EPERM) /* Unsupported command */
		return 0;
	return ret;
}

static int __ccic_cam_reset(struct ccic_camera *cam)
{
	int zero = 0;
	return __ccic_cam_cmd(cam, VIDIOC_INT_RESET, &zero);
}

/*
 * We have found the sensor on the i2c.  Let's try to have a
 * conversation.
 */
static int ccic_cam_init(struct ccic_camera *cam)
{
	struct v4l2_dbg_chip_ident chip = { V4L2_CHIP_MATCH_I2C_ADDR, 0, 0 };;
	int ret;

	mutex_lock(&cam->s_mutex);
	if (cam->state != S_NOTREADY)
		cam_warn(cam, "Cam init with device in funky state %d",
				cam->state);
	ret = __ccic_cam_reset(cam);
	if (ret)
		goto out;
	chip.match.type = V4L2_CHIP_MATCH_I2C_ADDR;
	ret = __ccic_cam_cmd(cam, VIDIOC_DBG_G_CHIP_IDENT, &chip);
	if (ret)
		goto out;
	cam->sensor_type = chip.ident;
	if ((cam->sensor_type != V4L2_IDENT_OV7690)
		&&(cam->sensor_type != V4L2_IDENT_OV3640)) {
		cam_err(cam, "Unsupported sensor type %d", cam->sensor->addr);
		ret = -EINVAL;
		goto out;
	}
	/* Get/set parameters? */
	ret = 0;
	cam->state = S_IDLE;
out:
	mutex_unlock(&cam->s_mutex);
	return ret;
}

/*
 * Configure the sensor to match the parameters we have.  Caller should
 * hold s_mutex
 */
static int ccic_cam_set_flip(struct ccic_camera *cam)
{
	struct v4l2_control ctrl;

	memset(&ctrl, 0, sizeof(ctrl));
	ctrl.id = V4L2_CID_VFLIP;
	ctrl.value = flip;
	return __ccic_cam_cmd(cam, VIDIOC_S_CTRL, &ctrl);
}


static int ccic_cam_configure(struct ccic_camera *cam)
{
	struct v4l2_format fmt;
	int ret, zero = 0;

	if (cam->state != S_IDLE)
		return -EINVAL;
	fmt.fmt.pix = cam->pix_format;

#if CONVERT_420_422
	if (V4L2_PIX_FMT_YUV420 == fmt.fmt.pix.pixelformat)
		fmt.fmt.pix.pixelformat = V4L2_PIX_FMT_YUV422P;
#endif
	ret = __ccic_cam_cmd(cam, VIDIOC_INT_INIT, &zero);
	if (ret == 0)
		ret = __ccic_cam_cmd(cam, VIDIOC_S_FMT, &fmt);
	
	return ret;
}

/* -------------------------------------------------------------------- */
/*
 * DMA buffer management.  These functions need s_mutex held.
 */

/* FIXME: this is inefficient as hell, since dma_alloc_coherent just
 * does a get_free_pages() call, and we waste a good chunk of an orderN
 * allocation.  Should try to allocate the whole set in one chunk.
 */
static int ccic_alloc_dma_bufs(struct ccic_camera *cam, int loadtime)
{
	int i;

	ccic_set_config_needed(cam, 1);
	if (loadtime)
		cam->dma_buf_size = dma_buf_size;
	else
		cam->dma_buf_size = cam->pix_format.sizeimage;
	if (n_dma_bufs > 3)
		n_dma_bufs = 3;

	cam->nbufs = 0;
	for (i = 0; i < n_dma_bufs; i++) {
#if DMA_POOL
		cam->buf_node[i].dma_bufs = dma_alloc_coherent(&cam->pdev->dev,
				cam->dma_buf_size, &cam->buf_node[i].dma_handles,
				GFP_KERNEL);
#else
		cam->buf_node[i].order = get_order(cam->dma_buf_size);
		cam->buf_node[i].dma_bufs = (void *)__get_free_pages(GFP_KERNEL, cam->buf_node[i].order);
		cam->buf_node[i].dma_handles = __pa(cam->buf_node[i].dma_bufs);
#endif
		if (cam->buf_node[i].dma_bufs == NULL) {
			cam_warn(cam, "Failed to allocate DMA buffer\n");
			break;
		}

		/* For debug, remove eventually */
		memset(cam->buf_node[i].dma_bufs, 0xcc, cam->dma_buf_size);
		(cam->nbufs)++;
	}
	
	if (cam->nbufs < n_dma_bufs) {
		for (i = 0; i < cam->nbufs; i++) {
#if DMA_POOL
			dma_free_coherent(&cam->pdev->dev, cam->dma_buf_size,
					cam->buf_node[i].dma_bufs, cam->buf_node[i].dma_handles);
#else
			free_pages((unsigned long)cam->buf_node[i].dma_bufs, cam->buf_node[i].order);
#endif

		}
		cam->nbufs = 0;

		cam_err(cam, "Insufficient DMA buffers, cannot operate\n");
		return -ENOMEM;

	}

	return 0;
}

static void ccic_free_dma_bufs(struct ccic_camera *cam)
{
	int i;
	for (i = 0; i < cam->nbufs; i++) {
#if DMA_POOL
		dma_free_coherent(&cam->pdev->dev, cam->dma_buf_size,
				cam->buf_node[i].dma_bufs, cam->buf_node[i].dma_handles);
#else
		free_pages((unsigned long)cam->buf_node[i].dma_bufs, cam->buf_node[i].order);
#endif
		cam->buf_node[i].dma_bufs = NULL;
	}
	cam->nbufs = 0;
}





/* ----------------------------------------------------------------------- */
/*
 * Here starts the V4L2 interface code.
 */

/*
 * Read an image from the device.
 */
static ssize_t ccic_deliver_buffer(struct ccic_camera *cam,
		char __user *buffer, size_t len, loff_t *pos)
{
	int bufno;
	unsigned long flags;

	spin_lock_irqsave(&cam->dev_lock, flags);
	if (cam->next_buf < 0) {
		cam_err(cam, "deliver_buffer: No next buffer\n");
		spin_unlock_irqrestore(&cam->dev_lock, flags);
		return -EIO;
	}
	bufno = cam->next_buf;
	clear_bit(bufno, &cam->flags);
	if (++(cam->next_buf) >= cam->nbufs)
		cam->next_buf = 0;
	if (!test_bit(cam->next_buf, &cam->flags))
		cam->next_buf = -1;
	cam->specframes = 0;
	spin_unlock_irqrestore(&cam->dev_lock, flags);

	if (len > cam->pix_format.sizeimage)
		len = cam->pix_format.sizeimage;
#if !DMA_POOL
	dma_sync_single_for_device(&(cam->pdev->dev), cam->buf_node[bufno].dma_handles, len, DMA_FROM_DEVICE);
#endif		
	if (copy_to_user(buffer, cam->buf_node[bufno].dma_bufs, len))
		return -EFAULT;
	(*pos) += len;
	return len;
}

/*
 * Get everything ready, and start grabbing frames.
 */
static int ccic_read_setup(struct ccic_camera *cam, enum ccic_state state)
{
	int ret;
	unsigned long flags;

	/*
	 * Configuration.  If we still don't have DMA buffers,
	 * make one last, desperate attempt.
	 */
	if (cam->nbufs == 0)
		if (ccic_alloc_dma_bufs(cam, 0))
			return -ENOMEM;

	if (ccic_needs_config(cam)) {
		ccic_cam_configure(cam);
		ret = ccic_ctlr_configure(cam);
		if (ret)
			return ret;
	}

	/*
	 * Turn it loose.
	 */
	spin_lock_irqsave(&cam->dev_lock, flags);
	ccic_reset_buffers(cam);
	ccic_sci_irq_enable(cam);
	cam->state = state;
	
	csi_start(g_csi);	
	ccic_sci_start(cam);
	__ccic_cam_cmd(cam, VIDIOC_STREAMON, NULL);
	spin_unlock_irqrestore(&cam->dev_lock, flags);

/*
	dump_registers(cam);
	dump_csi_registers(g_csi);
*/
	return 0;
}


static ssize_t ccic_v4l_read(struct file *filp,
		char __user *buffer, size_t len, loff_t *pos)
{
	struct ccic_camera *cam = filp->private_data;
	int ret = 0;

	/*
	 * Perhaps we're in speculative read mode and already
	 * have data?
	 */
	mutex_lock(&cam->s_mutex);
	if (cam->state == S_SPECREAD) {
		if (cam->next_buf >= 0) {
			ret = ccic_deliver_buffer(cam, buffer, len, pos);
			if (ret != 0)
				goto out_unlock;
		}
	} else if (cam->state == S_FLAKED || cam->state == S_NOTREADY) {
		ret = -EIO;
		goto out_unlock;
	} else if (cam->state != S_IDLE) {
		ret = -EBUSY;
		goto out_unlock;
	}

	/*
	 * v4l2: multiple processes can open the device, but only
	 * one gets to grab data from it.
	 */
	if (cam->owner && cam->owner != filp) {
		ret = -EBUSY;
		goto out_unlock;
	}
	cam->owner = filp;

	/*
	 * Do setup if need be.
	 */
	if (cam->state != S_SPECREAD) {
		ret = ccic_read_setup(cam, S_SINGLEREAD);
		if (ret)
			goto out_unlock;
	}
	/*
	 * Wait for something to happen.  This should probably
	 * be interruptible (FIXME).
	 */
	wait_event_timeout(cam->iowait, cam->next_buf >= 0, HZ);
	if (cam->next_buf < 0) {
		cam_err(cam, "read() operation timed out\n");
		ccic_ctlr_stop_dma(cam);
		ret = -EIO;
		goto out_unlock;
	}
	/*
	 * Give them their data and we should be done.
	 */
	ret = ccic_deliver_buffer(cam, buffer, len, pos);

out_unlock:
	mutex_unlock(&cam->s_mutex);
	return ret;
}

/*
 * Streaming I/O support.
 */

static int ccic_vidioc_streamon(struct file *filp, void *priv,
		enum v4l2_buf_type type)
{
	struct ccic_camera *cam = filp->private_data;
	int ret = -EINVAL;

	if (type != V4L2_BUF_TYPE_VIDEO_CAPTURE)
		goto out;
	mutex_lock(&cam->s_mutex);
	if (cam->state != S_IDLE || cam->n_sbufs == 0)
		goto out_unlock;

	cam->sequence = 0;
	ret = ccic_read_setup(cam, S_STREAMING);
out_unlock:
	mutex_unlock(&cam->s_mutex);
out:
	return ret;
}

static unsigned int csi_reg_read(struct ccic_csi *csi, unsigned int reg);

static int ccic_vidioc_streamoff(struct file *filp, void *priv,
		enum v4l2_buf_type type)
{
	struct ccic_camera *cam = filp->private_data;
	int ret = -EINVAL;

	if (type != V4L2_BUF_TYPE_VIDEO_CAPTURE)
		goto out;
	mutex_lock(&cam->s_mutex);
	if (cam->state != S_STREAMING)
		goto out_unlock;

	__ccic_cam_cmd(cam, VIDIOC_STREAMOFF, NULL);
	ccic_ctlr_stop_dma(cam);
	
	ret = 0;

out_unlock:
	mutex_unlock(&cam->s_mutex);
out:
	return ret;
}



static int ccic_setup_siobuf(struct ccic_camera *cam, int index)
{
	struct ccic_sio_buffer *buf = cam->sb_bufs + index;

	INIT_LIST_HEAD(&buf->list);
	buf->v4lbuf.length = PAGE_ALIGN(cam->pix_format.sizeimage);
	buf->buffer = vmalloc_user(buf->v4lbuf.length);
	if (buf->buffer == NULL)
		return -ENOMEM;
	buf->mapcount = 0;
	buf->cam = cam;

	buf->v4lbuf.index = index;
	buf->v4lbuf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
	buf->v4lbuf.field = V4L2_FIELD_NONE;
	buf->v4lbuf.memory = V4L2_MEMORY_MMAP;
	/*
	 * Offset: must be 32-bit even on a 64-bit system.  videobuf-dma-sg
	 * just uses the length times the index, but the spec warns
	 * against doing just that - vma merging problems.  So we
	 * leave a gap between each pair of buffers.
	 */
	buf->v4lbuf.m.offset = 2*index*buf->v4lbuf.length;
	return 0;
}

static int ccic_free_sio_buffers(struct ccic_camera *cam)
{
	int i;

	/*
	 * If any buffers are mapped, we cannot free them at all.
	 */
	for (i = 0; i < cam->n_sbufs; i++)
		if (cam->sb_bufs[i].mapcount > 0)
			return -EBUSY;
	/*
	 * OK, let's do it.
	 */
	for (i = 0; i < cam->n_sbufs; i++)
		vfree(cam->sb_bufs[i].buffer);
	cam->n_sbufs = 0;
	kfree(cam->sb_bufs);
	cam->sb_bufs = NULL;
	INIT_LIST_HEAD(&cam->sb_avail);
	INIT_LIST_HEAD(&cam->sb_full);
	return 0;
}

static int ccic_vidioc_reqbufs(struct file *filp, void *priv,
		struct v4l2_requestbuffers *req)
{
	struct ccic_camera *cam = filp->private_data;
	int ret = 0;  /* Silence warning */

	/*
	 * Make sure it's something we can do.  User pointers could be
	 * implemented without great pain, but that's not been done yet.
	 */
	if (req->type != V4L2_BUF_TYPE_VIDEO_CAPTURE)
		return -EINVAL;
	if (req->memory != V4L2_MEMORY_MMAP)
		return -EINVAL;
	/*
	 * If they ask for zero buffers, they really want us to stop streaming
	 * (if it's happening) and free everything.  Should we check owner?
	 */
	mutex_lock(&cam->s_mutex);
	if (req->count == 0) {
		if (cam->state == S_STREAMING)
			ccic_ctlr_stop_dma(cam);
		ret = ccic_free_sio_buffers (cam);
		goto out;
	}
	/*
	 * Device needs to be idle and working.  We *could* try to do the
	 * right thing in S_SPECREAD by shutting things down, but it
	 * probably doesn't matter.
	 */
	if (cam->state != S_IDLE || (cam->owner && cam->owner != filp)) {
		ret = -EBUSY;
		goto out;
	}
	cam->owner = filp;

	if (req->count < min_buffers)
		req->count = min_buffers;
	else if (req->count > max_buffers)
		req->count = max_buffers;
	if (cam->n_sbufs > 0) {
		ret = ccic_free_sio_buffers(cam);
		if (ret)
			goto out;
	}

	cam->sb_bufs = kzalloc(req->count*sizeof(struct ccic_sio_buffer),
			GFP_KERNEL);
	if (cam->sb_bufs == NULL) {
		ret = -ENOMEM;
		goto out;
	}
	for (cam->n_sbufs = 0; cam->n_sbufs < req->count; (cam->n_sbufs++)) {
		ret = ccic_setup_siobuf(cam, cam->n_sbufs);
		if (ret)
			break;
	}

	if (cam->n_sbufs == 0)  /* no luck at all - ret already set */
		kfree(cam->sb_bufs);
	req->count = cam->n_sbufs;  /* In case of partial success */

out:
	mutex_unlock(&cam->s_mutex);
	return ret;
}


static int ccic_vidioc_querybuf(struct file *filp, void *priv,
		struct v4l2_buffer *buf)
{
	struct ccic_camera *cam = filp->private_data;
	int ret = -EINVAL;

	mutex_lock(&cam->s_mutex);
	if (buf->type != V4L2_BUF_TYPE_VIDEO_CAPTURE)
		goto out;
	if (buf->index < 0 || buf->index >= cam->n_sbufs)
		goto out;
	*buf = cam->sb_bufs[buf->index].v4lbuf;
	ret = 0;
out:
	mutex_unlock(&cam->s_mutex);
	return ret;
}

static int ccic_vidioc_qbuf(struct file *filp, void *priv,
		struct v4l2_buffer *buf)
{
	struct ccic_camera *cam = filp->private_data;
	struct ccic_sio_buffer *sbuf;
	int ret = -EINVAL;
	unsigned long flags;

	mutex_lock(&cam->s_mutex);
	if (buf->type != V4L2_BUF_TYPE_VIDEO_CAPTURE)
		goto out;
	if (buf->index < 0 || buf->index >= cam->n_sbufs)
		goto out;
	sbuf = cam->sb_bufs + buf->index;
	if (sbuf->v4lbuf.flags & V4L2_BUF_FLAG_QUEUED) {
		ret = 0; /* Already queued?? */
		goto out;
	}
	if (sbuf->v4lbuf.flags & V4L2_BUF_FLAG_DONE) {
		/* Spec doesn't say anything, seems appropriate tho */
		ret = -EBUSY;
		goto out;
	}
	sbuf->v4lbuf.flags |= V4L2_BUF_FLAG_QUEUED;
	spin_lock_irqsave(&cam->dev_lock, flags);
	list_add(&sbuf->list, &cam->sb_avail);
	spin_unlock_irqrestore(&cam->dev_lock, flags);
	ret = 0;
out:
	mutex_unlock(&cam->s_mutex);
	return ret;
}

static int ccic_vidioc_dqbuf(struct file *filp, void *priv,
		struct v4l2_buffer *buf)
{
	struct ccic_camera *cam = filp->private_data;
	struct ccic_sio_buffer *sbuf;
	int ret = -EINVAL;
	unsigned long flags;

	mutex_lock(&cam->s_mutex);
	if (buf->type != V4L2_BUF_TYPE_VIDEO_CAPTURE)
		goto out_unlock;
	if (cam->state != S_STREAMING)
		goto out_unlock;
	if (list_empty(&cam->sb_full) && filp->f_flags & O_NONBLOCK) {
		ret = -EAGAIN;
		goto out_unlock;
	}

	while (list_empty(&cam->sb_full) && cam->state == S_STREAMING) {
		mutex_unlock(&cam->s_mutex);
		if (wait_event_interruptible(cam->iowait,
						!list_empty(&cam->sb_full))) {
			ret = -ERESTARTSYS;
			goto out;
		}
		mutex_lock(&cam->s_mutex);
	}

	if (cam->state != S_STREAMING)
		ret = -EINTR;
	else {
		spin_lock_irqsave(&cam->dev_lock, flags);
		/* Should probably recheck !list_empty() here */
		sbuf = list_entry(cam->sb_full.next,
				struct ccic_sio_buffer, list);
		list_del_init(&sbuf->list);
		spin_unlock_irqrestore(&cam->dev_lock, flags);
		sbuf->v4lbuf.flags &= ~V4L2_BUF_FLAG_DONE;
		*buf = sbuf->v4lbuf;
		ret = 0;
	}

out_unlock:
	mutex_unlock(&cam->s_mutex);
out:
	return ret;
}

static void ccic_v4l_vm_open(struct vm_area_struct *vma)
{
	struct ccic_sio_buffer *sbuf = vma->vm_private_data;
	/*
	 * Locking: done under mmap_sem, so we don't need to
	 * go back to the camera lock here.
	 */
	sbuf->mapcount++;
}


static void ccic_v4l_vm_close(struct vm_area_struct *vma)
{
	struct ccic_sio_buffer *sbuf = vma->vm_private_data;

	mutex_lock(&sbuf->cam->s_mutex);
	sbuf->mapcount--;
	/* Docs say we should stop I/O too... */
	if (sbuf->mapcount == 0)
		sbuf->v4lbuf.flags &= ~V4L2_BUF_FLAG_MAPPED;
	mutex_unlock(&sbuf->cam->s_mutex);
}

static struct vm_operations_struct ccic_v4l_vm_ops = {
	.open = ccic_v4l_vm_open,
	.close = ccic_v4l_vm_close
};


static int ccic_v4l_mmap(struct file *filp, struct vm_area_struct *vma)
{
	struct ccic_camera *cam = filp->private_data;
	unsigned long offset = vma->vm_pgoff << PAGE_SHIFT;
	int ret = -EINVAL;
	int i;
	struct ccic_sio_buffer *sbuf = NULL;

	/*
	 * Find the buffer they are looking for.
	 */
	mutex_lock(&cam->s_mutex);
	for (i = 0; i < cam->n_sbufs; i++)
		if (cam->sb_bufs[i].v4lbuf.m.offset == offset) {
			sbuf = cam->sb_bufs + i;
			break;
		}
	if (sbuf == NULL)
		goto out;

	ret = remap_vmalloc_range(vma, sbuf->buffer, 0);
	if (ret)
		goto out;
	vma->vm_flags |= VM_DONTEXPAND;
	vma->vm_private_data = sbuf;
	vma->vm_ops = &ccic_v4l_vm_ops;
	sbuf->v4lbuf.flags |= V4L2_BUF_FLAG_MAPPED;
	ccic_v4l_vm_open(vma);
	ret = 0;
out:
	mutex_unlock(&cam->s_mutex);
	return ret;
}



static int ccic_v4l_open(struct file *filp)
{
	struct ccic_camera *cam;

	cam = ccic_find_dev(video_devdata(filp)->minor);
	if (cam == NULL)
		return -ENODEV;
	filp->private_data = cam;

	mutex_lock(&cam->s_mutex);
	if (cam->users == 0) {
		ccic_set_config_needed(cam, 1);
	/* FIXME make sure this is complete */
	}
	(cam->users)++;
	mutex_unlock(&cam->s_mutex);
	return 0;
}


static int ccic_v4l_release(struct file *filp)
{
	struct ccic_camera *cam = filp->private_data;
	struct sensor_platform_data *pdata;
	pdata = cam->sensor->dev.platform_data;

	mutex_lock(&cam->s_mutex);
	(cam->users)--;
	if (filp == cam->owner) {
		ccic_ctlr_stop_dma(cam);
		ccic_free_sio_buffers(cam);
		cam->owner = NULL;
	}
	if (cam->users == 0) {
		pdata->power_set(sensor_selected, SENSOR_CLOSE);
		if (alloc_bufs_at_read)
			ccic_free_dma_bufs(cam);
	}
	mutex_unlock(&cam->s_mutex);
	return 0;
}

static unsigned int ccic_v4l_poll(struct file *filp,
		struct poll_table_struct *pt)
{
	struct ccic_camera *cam = filp->private_data;

	poll_wait(filp, &cam->iowait, pt);
	if (cam->next_buf >= 0)
		return POLLIN | POLLRDNORM;
	return 0;
}



static int ccic_vidioc_queryctrl(struct file *filp, void *priv,
		struct v4l2_queryctrl *qc)
{
	struct ccic_camera *cam = filp->private_data;
	int ret;

	mutex_lock(&cam->s_mutex);
	ret = __ccic_cam_cmd(cam, VIDIOC_QUERYCTRL, qc);
	mutex_unlock(&cam->s_mutex);
	return ret;
}


static int ccic_vidioc_g_ctrl(struct file *filp, void *priv,
		struct v4l2_control *ctrl)
{
	struct ccic_camera *cam = filp->private_data;
	int ret;

	mutex_lock(&cam->s_mutex);
	ret = __ccic_cam_cmd(cam, VIDIOC_G_CTRL, ctrl);
	mutex_unlock(&cam->s_mutex);
	return ret;
}


static int ccic_vidioc_s_ctrl(struct file *filp, void *priv,
		struct v4l2_control *ctrl)
{
	struct ccic_camera *cam = filp->private_data;
	int ret;

	mutex_lock(&cam->s_mutex);
	ret = __ccic_cam_cmd(cam, VIDIOC_S_CTRL, ctrl);
	mutex_unlock(&cam->s_mutex);
	return ret;
}

static int ccic_vidioc_querycap(struct file *file, void *priv,
		struct v4l2_capability *cap)
{
	strcpy(cap->driver, "pxa950_camera");
	strcpy(cap->card, "pxa950_camera");
	cap->version = CCIC_VERSION;
	cap->capabilities = V4L2_CAP_VIDEO_CAPTURE |
		V4L2_CAP_READWRITE | V4L2_CAP_STREAMING;
	return 0;
}


/*
 * The default format we use until somebody says otherwise.
 */
static struct v4l2_pix_format ccic_def_pix_format = {
	.width		= VGA_WIDTH,
	.height		= VGA_HEIGHT,
	.pixelformat	= V4L2_PIX_FMT_YUYV,
	.field		= V4L2_FIELD_NONE,
	.bytesperline	= VGA_WIDTH*2,
	.sizeimage	= VGA_WIDTH*VGA_HEIGHT*2,
};

static int ccic_vidioc_enum_fmt_cap(struct file *filp,
		void *priv, struct v4l2_fmtdesc *fmt)
{
	struct ccic_camera *cam = priv;
	int ret;

	if (fmt->type != V4L2_BUF_TYPE_VIDEO_CAPTURE)
		return -EINVAL;
	mutex_lock(&cam->s_mutex);
	ret = __ccic_cam_cmd(cam, VIDIOC_ENUM_FMT, fmt);
	mutex_unlock(&cam->s_mutex);
	return ret;
}


static int ccic_vidioc_try_fmt_cap(struct file *filp, void *priv,
		struct v4l2_format *fmt)
{
	struct ccic_camera *cam = priv;
	int ret;
	mutex_lock(&cam->s_mutex);
	ret = __ccic_cam_cmd(cam, VIDIOC_TRY_FMT, fmt);
	mutex_unlock(&cam->s_mutex);
	return ret;
}

static int ccic_vidioc_s_fmt_cap(struct file *filp, void *priv,
		struct v4l2_format *fmt)
{
	struct ccic_camera *cam = priv;
	int ret;
	/*
	 * Can't do anything if the device is not idle
	 * Also can't if there are streaming buffers in place.
	 */
	if (cam->state != S_IDLE)
		return -EBUSY;
	/*
	 * See if the formatting works in principle.
	 */
	ret = ccic_vidioc_try_fmt_cap(filp, priv, fmt);
	if (ret)
		return ret;
	/*
	 * Now we start to change things for real, so let's do it
	 * under lock.
	 */
	mutex_lock(&cam->s_mutex);
	cam->pix_format = fmt->fmt.pix;
	/*
	 * Make sure we have appropriate DMA buffers.
	 */
	ret = -ENOMEM;

	if (cam->nbufs > 0 && cam->dma_buf_size < cam->pix_format.sizeimage)
		ccic_free_dma_bufs(cam);
	if (cam->nbufs == 0) {
		if (ccic_alloc_dma_bufs(cam, 0))
			goto out;
	}

	ccic_sci_stop(cam);

	/*
	 * It looks like this might work, so let's program the sensor.
	 */
	ret = ccic_cam_configure(cam);
	if (!ret)
		ret = ccic_ctlr_configure(cam);
out:
	mutex_unlock(&cam->s_mutex);
	return ret;
}

/*
 * Return our stored notion of how the camera is/should be configured.
 * The V4l2 spec wants us to be smarter, and actually get this from
 * the camera (and not mess with it at open time).  Someday.
 */
static int ccic_vidioc_g_fmt_cap(struct file *filp, void *priv,
		struct v4l2_format *f)
{
	struct ccic_camera *cam = priv;

	f->fmt.pix = cam->pix_format;
	return 0;
}

/*
 * We only have one input - the sensor - so minimize the nonsense here.
 */
static int ccic_vidioc_enum_input(struct file *filp, void *priv,
		struct v4l2_input *input)
{
	if (input->index != 0 && input->index != 1)
		return -EINVAL;

	input->type = V4L2_INPUT_TYPE_CAMERA;
	input->std = V4L2_STD_ALL; /* Not sure what should go here */
	strcpy(input->name, "Camera");
	return 0;
}

static int ccic_vidioc_g_input(struct file *filp, void *priv, unsigned int *i)
{
	*i = sensor_selected;
	return 0;
}

static int ccic_vidioc_s_input(struct file *filp, void *priv, unsigned int i)
{
	struct ccic_camera *cam = filp->private_data;
	int ret;

	struct sensor_platform_data *pdata;

	 if (((i == SENSOR_LOW) && (detected_low == 1))
		 || ((i == SENSOR_HIGH) && (detected_high == 1))) {
		 cam->sensor = cam->sensors[i];
		 sensor_selected = i;
	} else {
		printk(KERN_ERR "requested sensor %d is NOT attached!\n", i);
		return -1;
	}

	pdata = cam->sensor->dev.platform_data;

	/* enable clock */
	pdata->power_set(sensor_selected, SENSOR_OPEN);
	__ccic_cam_reset(cam);

	mutex_lock(&cam->s_mutex);
	ret = __ccic_cam_cmd(cam, VIDIOC_S_INPUT, &i);
	mutex_unlock(&cam->s_mutex);

#ifdef CAM_BRIDGE
	/* small resolution use high resolution as bridge in hardware, though it is not modulabe */

	if (SENSOR_LOW == sensor_selected) {
		struct i2c_client *sc = cam->sensors[SENSOR_HIGH];

		if (sc == NULL || sc->driver == NULL
				|| sc->driver->command == NULL) {
			printk(KERN_ERR "no briged could be used\n");
			return -EINVAL;
		}

		ret = sc->driver->command(sc, VIDIOC_SET_BRIDGE, NULL);
		if (ret == -EPERM) {
			printk(KERN_ERR "command no briged could be used\n");
			return -EINVAL;
		}

	}
#endif

	return ret;
}

/* from vivi.c */
static int ccic_vidioc_s_std(struct file *filp, void *priv, v4l2_std_id *a)
{
	return 0;
}

/*
 * G/S_PARM.  Most of this is done by the sensor, but we are
 * the level which controls the number of read buffers.
 */
static int ccic_vidioc_g_parm(struct file *filp, void *priv,
		struct v4l2_streamparm *parms)
{
	struct ccic_camera *cam = priv;
	int ret;

	mutex_lock(&cam->s_mutex);
	ret = __ccic_cam_cmd(cam, VIDIOC_G_PARM, parms);
	mutex_unlock(&cam->s_mutex);
	parms->parm.capture.readbuffers = n_dma_bufs;
	return ret;
}

static int ccic_vidioc_s_parm(struct file *filp, void *priv,
		struct v4l2_streamparm *parms)
{
	struct ccic_camera *cam = priv;
	int ret;
	mutex_lock(&cam->s_mutex);
	ret = __ccic_cam_cmd(cam, VIDIOC_S_PARM, parms);
	mutex_unlock(&cam->s_mutex);
	parms->parm.capture.readbuffers = n_dma_bufs;
	return ret;
}


static void ccic_v4l_dev_release(struct video_device *vd)
{
	struct ccic_camera *cam = container_of(vd, struct ccic_camera, v4ldev);

	kfree(cam);
}


/*
 * This template device holds all of those v4l2 methods; we
 * clone it for specific real devices.
 */

static const struct v4l2_file_operations ccic_v4l_fops = {
	.owner = THIS_MODULE,
	.open = ccic_v4l_open,
	.release = ccic_v4l_release,
	.read = ccic_v4l_read,
	.poll = ccic_v4l_poll,
	.mmap = ccic_v4l_mmap,
	.ioctl = video_ioctl2,
};
/* upgrade changes from .25 to .28 */
struct v4l2_ioctl_ops ccic_ioctl_ops = {
	.vidioc_querycap        = ccic_vidioc_querycap,
	.vidioc_enum_fmt_vid_cap    = ccic_vidioc_enum_fmt_cap,
	.vidioc_try_fmt_vid_cap     = ccic_vidioc_try_fmt_cap,
	.vidioc_s_fmt_vid_cap       = ccic_vidioc_s_fmt_cap,
	.vidioc_g_fmt_vid_cap       = ccic_vidioc_g_fmt_cap,
	.vidioc_enum_input      = ccic_vidioc_enum_input,
	.vidioc_g_input         = ccic_vidioc_g_input,
	.vidioc_s_input         = ccic_vidioc_s_input,
	.vidioc_s_std           = ccic_vidioc_s_std,
	.vidioc_reqbufs         = ccic_vidioc_reqbufs,
	.vidioc_querybuf        = ccic_vidioc_querybuf,
	.vidioc_qbuf            = ccic_vidioc_qbuf,
	.vidioc_dqbuf           = ccic_vidioc_dqbuf,
	.vidioc_streamon        = ccic_vidioc_streamon,
	.vidioc_streamoff       = ccic_vidioc_streamoff,
	.vidioc_queryctrl       = ccic_vidioc_queryctrl,
	.vidioc_g_ctrl          = ccic_vidioc_g_ctrl,
	.vidioc_s_ctrl          = ccic_vidioc_s_ctrl,
	.vidioc_g_parm          = ccic_vidioc_g_parm,
	.vidioc_s_parm          = ccic_vidioc_s_parm,
};

static struct video_device ccic_v4l_template = {
	.name = "pxa950-camera",
	.vfl_type = VID_TYPE_CAPTURE,
	.minor = -1, 				/* Get one dynamically */
	.tvnorms = V4L2_STD_NTSC_M,
	.current_norm = V4L2_STD_NTSC_M,	/* make mplayer happy */
	.fops = &ccic_v4l_fops,
	.release = ccic_v4l_dev_release,
	.ioctl_ops	= &ccic_ioctl_ops,
};

static void dump_registers(struct ccic_camera *cam)
{
	printk(KERN_ERR "SCICR0 0x%x\n", ccic_reg_read(cam, REG_SCICR0));
	printk(KERN_ERR "SCICR1 0x%x\n", ccic_reg_read(cam, REG_SCICR1));
	printk(KERN_ERR "SCISR 0x%x\n", ccic_reg_read(cam, REG_SCISR));
	printk(KERN_ERR "SCIMASK 0x%x\n", ccic_reg_read(cam, REG_SCIMASK));
	printk(KERN_ERR "SCIFIFO 0x%x\n", ccic_reg_read(cam, REG_SCIFIFO));
	printk(KERN_ERR "SCIFIFOSR 0x%x\n", ccic_reg_read(cam, REG_SCIFIFOSR));

	printk(KERN_ERR "SCIDADDR0 0x%x\n", ccic_reg_read(cam, REG_SCIDADDR0));
	printk(KERN_ERR "SCISADDR0 0x%x\n", ccic_reg_read(cam, REG_SCISADDR0));
	printk(KERN_ERR "SCITADDR0 0x%x\n", ccic_reg_read(cam, REG_SCITADDR0));
	printk(KERN_ERR "SCIDCMD0 0x%x\n", ccic_reg_read(cam, REG_SCIDCMD0));

	printk(KERN_ERR "SCIDADDR1 0x%x\n", ccic_reg_read(cam, REG_SCIDADDR1));
	printk(KERN_ERR "SCISADDR1 0x%x\n", ccic_reg_read(cam, REG_SCISADDR1));
	printk(KERN_ERR "SCITADDR1 0x%x\n", ccic_reg_read(cam, REG_SCITADDR1));
	printk(KERN_ERR "SCIDCMD1 0x%x\n", ccic_reg_read(cam, REG_SCIDCMD1));

	printk(KERN_ERR "SCIDADDR2 0x%x\n", ccic_reg_read(cam, REG_SCIDADDR2));
	printk(KERN_ERR "SCISADDR2 0x%x\n", ccic_reg_read(cam, REG_SCISADDR2));
	printk(KERN_ERR "SCITADDR2 0x%x\n", ccic_reg_read(cam, REG_SCITADDR2));
	printk(KERN_ERR "SCIDCMD2 0x%x\n", ccic_reg_read(cam, REG_SCIDCMD2));

	printk(KERN_ERR "SCIDBR0 0x%x\n", ccic_reg_read(cam, REG_SCIDBR0));
	printk(KERN_ERR "SCIDCSR0 0x%x\n", ccic_reg_read(cam, REG_SCIDCSR0));

	printk(KERN_ERR "SCIDBR1 0x%x\n", ccic_reg_read(cam, REG_SCIDBR1));
	printk(KERN_ERR "SCIDCSR1 0x%x\n", ccic_reg_read(cam, REG_SCIDCSR1));

	printk(KERN_ERR "SCIDBR2 0x%x\n", ccic_reg_read(cam, REG_SCIDBR2));
	printk(KERN_ERR "SCIDCSR2 0x%x\n", ccic_reg_read(cam, REG_SCIDCSR2));
}

/* ---------------------------------------------------------------------- */
/*
 * Interrupt handler stuff
 */

static void ccic_frame_tasklet(unsigned long data)
{
	struct ccic_camera *cam = (struct ccic_camera *) data;
	int i;
	unsigned long flags;
	struct ccic_sio_buffer *sbuf;
	spin_lock_irqsave(&cam->dev_lock, flags);
	for (i = 0; i < cam->nbufs; i++) {
		int bufno = cam->next_buf;
		if (bufno < 0) {  /* "will never happen" */
			cam_err(cam, "No valid bufs in tasklet!\n");
			break;
		}
		if (++(cam->next_buf) >= cam->nbufs)
			cam->next_buf = 0;
		if (!test_bit(bufno, &cam->flags))
			continue;
		if (list_empty(&cam->sb_avail)) {
			/* drop current frame because of buffer unavailable */
			clear_bit(bufno, &cam->flags);
			break;  /* Leave it valid, hope for better later */
		}
		clear_bit(bufno, &cam->flags);
#if !DMA_POOL
		dma_sync_single_for_device(&(cam->pdev->dev),
			cam->buf_node[bufno].dma_handles,
			cam->pix_format.sizeimage,
			DMA_FROM_DEVICE);
#endif
		if ((cam->pix_format.pixelformat == V4L2_PIX_FMT_JPEG)
			&& ((((char *)cam->buf_node[bufno].dma_bufs)[0] != 0xff)
			|| (((char *)cam->buf_node[bufno].dma_bufs)[1] != 0xd8))) {

			spin_unlock_irqrestore(&cam->dev_lock, flags);

			/* drop current JPEG frame because of wrong header*/
			printk(KERN_ERR "%s: JPEG ERROR	dropped this frame\n", __func__);
			return;
		}

		sbuf = list_entry(cam->sb_avail.next, struct ccic_sio_buffer, list);
		/*
		 * Drop the lock during the big copy.  This *should* be safe...
		 */
		spin_unlock_irqrestore(&cam->dev_lock, flags);
		memcpy(sbuf->buffer, cam->buf_node[bufno].dma_bufs,
				cam->pix_format.sizeimage);
		sbuf->v4lbuf.m.offset = 0;
		sbuf->v4lbuf.bytesused = cam->pix_format.sizeimage;
		sbuf->v4lbuf.sequence = cam->buf_seq[bufno];
		sbuf->v4lbuf.flags &= ~V4L2_BUF_FLAG_QUEUED;
		sbuf->v4lbuf.flags |= V4L2_BUF_FLAG_DONE;
		spin_lock_irqsave(&cam->dev_lock, flags);
		list_move_tail(&sbuf->list, &cam->sb_full);
		break;		/* just exit once current frame done */
	}
	if (!list_empty(&cam->sb_full))
		wake_up(&cam->iowait);
	spin_unlock_irqrestore(&cam->dev_lock, flags);
}



static void ccic_frame_complete(struct ccic_camera *cam, int frame)
{
	/*
	 * Basic frame housekeeping.
	 */
	if (test_bit(frame, &cam->flags) && printk_ratelimit())
		cam_err(cam, "Frame overrun on %d, frames lost\n", frame);
	set_bit(frame, &cam->flags);
	clear_bit(CF_DMA_ACTIVE, &cam->flags);

	if (cam->next_buf < 0)
		cam->next_buf = frame;
	cam->buf_seq[frame] = ++(cam->sequence);

	switch (cam->state) {
		/*
		 * If in single read mode, try going speculative.
		 */
		case S_SINGLEREAD:
			cam->state = S_SPECREAD;
			cam->specframes = 0;
			wake_up(&cam->iowait);
			break;

			/*
			 * If we are already doing speculative reads,
			 * and nobody is reading them, just stop.
			 */
		case S_SPECREAD:
			if (++(cam->specframes) >= cam->nbufs) {
				ccic_sci_stop(cam);
				ccic_sci_irq_disable(cam);
				cam->state = S_IDLE;
			}
			wake_up(&cam->iowait);
			break;
			/*
			 * For the streaming case, we defer the real work to the
			 * camera tasklet.
			 *
			 * FIXME: if the app is not consuming the buffers,
			 * we should eventually put things on hold and
			 * vrestart in idioc_dqbuf().
			 */
		case S_STREAMING:
			tasklet_schedule(&cam->s_tasklet);
			break;

		default:
			cam_err(cam, "Frame intin non-operational state\n");
			break;
	}
}

static void ccic_frame_irq(struct ccic_camera *cam, unsigned int irqs)
{
	unsigned int frame = 0;

	/*
	 * Handle any frame completions.
	 */
	if (irqs & IRQ_EOFX)
		ccic_frame_complete(cam, frame);
	/*
	 * If a frame starts, note that we have DMA active.  This
	 * code assumes that we won't get multiple frame interrupts
	 * at once; may want to rethink that.
	 */
	if (irqs & IRQ_SOFX)
		set_bit(CF_DMA_ACTIVE, &cam->flags);
}

static irqreturn_t ccic_irq(int irq, void *data)
{
	struct ccic_camera *cam = data;
	unsigned int irqs = 0;

	spin_lock(&cam->dev_lock);
	irqs = ccic_reg_read(cam, REG_SCISR);
	ccic_reg_write(cam, REG_SCISR, irqs);		/*clear irqs here*/

	if (irqs & IRQ_FRAME)
		ccic_frame_irq(cam, irqs);
	spin_unlock(&cam->dev_lock);
	return IRQ_HANDLED;
}


/* -------------------------------------------------------------------------- */
static int pxa950_camera_probe(struct platform_device *pdev)
{
	struct resource *res;
	int ret = -ENOMEM;
	struct ccic_camera *cam;

	/*
	 * Start putting together one of our big camera structures.
	 */

	cam = kzalloc(sizeof(struct ccic_camera), GFP_KERNEL);
	if (cam == NULL)
		goto out;
	memset(cam, 0, sizeof(struct ccic_camera));

	mutex_init(&cam->s_mutex);
	mutex_lock(&cam->s_mutex);
	spin_lock_init(&cam->dev_lock);
	cam->state = S_NOTREADY;
	ccic_set_config_needed(cam, 1);
	init_waitqueue_head(&cam->iowait);
	cam->pdev = pdev;

	cam->pix_format = ccic_def_pix_format;
	INIT_LIST_HEAD(&cam->dev_list);
	INIT_LIST_HEAD(&cam->sb_avail);
	INIT_LIST_HEAD(&cam->sb_full);
	tasklet_init(&cam->s_tasklet, ccic_frame_tasklet, (unsigned long) cam);

	cam->platform_ops = pdev->dev.platform_data;

	if (cam->platform_ops == NULL) {
		printk(KERN_ERR "camera no platform data defined\n");
		return -ENODEV;
	}

	if (cam->platform_ops->power_set)
		cam->platform_ops->power_set(SENSOR_OPEN);

	cam->irq = platform_get_irq(pdev, 0);
	if (cam->irq < 0)
		return -ENXIO;

	res = platform_get_resource(pdev, IORESOURCE_MEM, 0);
	if (res == NULL) {
		printk(KERN_ERR "no IO memory resource defined\n");
		return -ENODEV;
	}

	ret = -EIO;
	cam->regs = ioremap(res->start, SZ_4K);
	if (!cam->regs) {
		printk(KERN_ERR "Unable to ioremap pxa950-camera regs\n");
		goto out_free;
	}
	ret = request_irq(cam->irq, ccic_irq,
		IRQF_SHARED, "pxa950-camera", cam);
	if (ret)
		goto out_iounmap;
	/*
	 * Initialize the controller and leave it powered up.  It will
	 * stay that way until the sensor driver shows up.
	 */

	ccic_ctlr_init(cam);
	/*
	 * Set up I2C/SMBUS communications.  We have to drop the mutex here
	 * because the sensor could attach in this call chain, leading to
	 * unsightly deadlocks.
	 */
	mutex_unlock(&cam->s_mutex);  /* attach can deadlock */
	/*
	 * Get the v4l2 setup done.
	 */
	mutex_lock(&cam->s_mutex);
	cam->v4ldev = ccic_v4l_template;
	cam->v4ldev.debug = 0;
	cam->v4ldev.dev = pdev->dev;
	ret = video_register_device(&cam->v4ldev, VFL_TYPE_GRABBER, -1);
	if (ret)
		goto out_freeirq;
	/*
	 * If so requested, try to get our DMA buffers now.
	 */
	if (!alloc_bufs_at_read) {
		if (ccic_alloc_dma_bufs(cam, 1))
			cam_warn(cam, "Unable to alloc DMA buffers at load"
					" will try again later.");
	}
	mutex_unlock(&cam->s_mutex);
	ccic_add_dev(cam);
	return 0;

out_freeirq:
	free_irq(cam->irq, cam);
out_iounmap:
	iounmap(cam->regs);
out_free:
	kfree(cam);
out:
	return ret;
}


/*
 * Shut down an initialized device
 */
static void ccic_shutdown(struct ccic_camera *cam)
{
/* FIXME: Make sure we take care of everything here */
	if (cam->n_sbufs > 0)
		/* What if they are still mapped?  Shouldn't be, but... */
		ccic_free_sio_buffers(cam);
	ccic_remove_dev(cam);
	ccic_ctlr_stop_dma(cam);
	ccic_free_dma_bufs(cam);
	free_irq(cam->irq, cam);
	iounmap(cam->regs);
	video_unregister_device(&cam->v4ldev);
	/* kfree(cam); done in v4l_release () */
}


static int pxa950_camera_remove(struct platform_device *pdev)
{
	struct ccic_camera *cam = ccic_find_by_pdev(pdev);

	if (cam == NULL) {
		printk(KERN_WARNING "remove on unknown pdev %p\n", pdev);
		return -EIO;
	}
	mutex_lock(&cam->s_mutex);
	if (cam->users > 0)
		cam_warn(cam, "Removing a device with users!\n");
	ccic_shutdown(cam);

	return 0;
	/* No unlock - it no longer exists */
}


static struct platform_driver pxa950_camera_driver = {
	.driver = {
		.name = "pxa950-camera"
	},
	.probe 		= pxa950_camera_probe,
	.remove 	= pxa950_camera_remove,

};

/* -------------------------------------------------------------------------- */

/* CSI */

/* ------------------------------------------------------------------------ */
/*
 * Device register I/O
 */
static inline void csi_reg_write(struct ccic_csi *csi, unsigned int reg,
		unsigned int val)
{
	__raw_writel(val, csi->regs + reg);
}

static inline unsigned int csi_reg_read(struct ccic_csi *csi,
		unsigned int reg)
{
	return __raw_readl(csi->regs + reg);
}

static void csi_config(struct ccic_csi *csi)
{
	unsigned int val = 0;
	unsigned int tmp = 0;

	/* divide clock by 3. 200Mhz. */
	val = CAM_CSI_CLK_DIV(0x1) |
		   CAM_CSI_CLK_GCLK_EN;
	csi_reg_write(csi, REG_CSGCR, val);

	val = 0x01;
	csi_reg_write(csi, REG_CSSCR, val);

	val = CAM_CSI_CONT_CLTERMEN(0x00) |
		  CAM_CSI_CONT_CLSETTLE(0x0C) |
		  CAM_CSI_CONT_CLMISS(0x00)   |
		  CAM_CSI_CONT_HSTERMEN(0x04);
	csi_reg_write(csi, REG_CSxTIM0, val);

	/* set time out before declarating an error to maximum */
	val =  CAM_CSI_CONT_HS_Rx_TO(0xffff) |
		  CAM_CSI_CONT_HSTSETTLE(0x0e);
	csi_reg_write(csi, REG_CSxTIM1, val);
/*
	val = CAM_CSI_CONT_NOL(0x00) |
		  CAM_CSI_CONT_VC0_CFG(0x0) |
		  CAM_CSI_CONT_VC1_CFG(0x1) |
		  CAM_CSI_CONT_VC2_CFG(0x2) |
		  CAM_CSI_CONT_VC3_CFG(0x3);
*/

	val = CAM_CSI_CONT_NOL(0x01) | 		/* 2 lanes */
		  CAM_CSI_CONT_VC0_CFG(0x0) |
		  CAM_CSI_CONT_VC1_CFG(0x1);
	csi_reg_write(csi, REG_CSxCR0, val);

	val = CAM_CSI_CONT_MIPI_REN_BYPASS(0x0) |
		CAM_CSI_CONT_MIPI_BG_VREF_EN;
	csi_reg_write(csi, REG_CSxPHYCAL, val);

/*
	// enabling all interrupts
	val = CAM_CSI_CONT_CSI_EN_INT    | // CSI ENABLE Interrupt Enable
	CAM_CSI_CONT_SoF_INT_EN      | // CSI Start of Frame Interrupt Enable
	CAM_CSI_CONT_EoF_INT_EN      | // CSI End of Frame Interrupt Enable
	CAM_CSI_CONT_OFLOW_INT_EN    | // CSI Overflow Interrupt Enable
	CAM_CSI_CONT_PHY_ERR_EN      | // D-PHY-related Error Interrupt Enable
	CAM_CSI_CONT_GEN_PACK_INT_EN | // CSI Generic Packet Interrupt Enable
	CAM_CSI_CONT_TIMEOUT_EN      | // Timeout Interrupt Enable
	CAM_CSI_CONT_PROT_EN;
	csi_reg_write(csi, REG_CSxINEN, val);
*/

}

static void dump_csi_registers(struct ccic_csi *csi)
{
	printk(KERN_ERR "CSSCR  0x%x\n", csi_reg_read(csi, REG_CSSCR));
	printk(KERN_ERR "CSGCR  0x%x\n", csi_reg_read(csi, REG_CSGCR));
	printk(KERN_ERR "CSxCR0 0x%x\n", csi_reg_read(csi, REG_CSxCR0));
	printk(KERN_ERR "CSxSR  0x%x\n", csi_reg_read(csi, REG_CSxSR));
	printk(KERN_ERR "CSxINEN 0x%x\n", csi_reg_read(csi, REG_CSxINEN));
	printk(KERN_ERR "CSxINST 0x%x\n", csi_reg_read(csi, REG_CSxINST));

	printk(KERN_ERR "CSxTIM0 0x%x\n", csi_reg_read(csi, REG_CSxTIM0));
	printk(KERN_ERR "CSxTIM1 0x%x\n", csi_reg_read(csi, REG_CSxTIM1));
	printk(KERN_ERR "CSxGENDAT 0x%x\n", csi_reg_read(csi, REG_CSxGENDAT));
	printk(KERN_ERR "CSxPHYCAL 0x%x\n", csi_reg_read(csi, REG_CSxPHYCAL));

}

static void csi_start(struct ccic_csi *csi)
{
	unsigned int val = 0;

	val = csi_reg_read(csi, REG_CSxCR0);
	val |= CSxCR0_CSIEN;
	csi_reg_write(csi, REG_CSxCR0, val);
}

static void csi_stop(struct ccic_csi *csi)
{
	unsigned int val = csi_reg_read(csi, REG_CSxCR0);

	val &= ~CSxCR0_CSIEN;
	csi_reg_write(csi, REG_CSxCR0, val);
}


static irqreturn_t csi_irq(int irq, void *data)
{
	struct ccic_csi *csi = data;
	unsigned int irqs;

	spin_lock(&csi->dev_lock);
	irqs = csi_reg_read(csi, REG_CSxINST);
	csi_reg_write(csi, REG_CSxINST, irqs);
	spin_unlock(&csi->dev_lock);
	return IRQ_HANDLED;
}


static int pxa950_csi_probe(struct platform_device *pdev)
{
	struct resource *res;
	int ret = -ENOMEM;
	struct ccic_csi *csi;

	csi = kzalloc(sizeof(struct ccic_csi), GFP_KERNEL);
	if (csi == NULL)
		goto out;

	g_csi = csi;

	csi->irq = platform_get_irq(pdev, 0);
	if (csi->irq < 0) {
		ret = -ENXIO;
		goto out;
	}

	res = platform_get_resource(pdev, IORESOURCE_MEM, 0);
	if (res == NULL) {
		printk(KERN_ERR "no IO memory resource defined\n");
		ret = -ENODEV;
		goto out;
	}

	ret = -EIO;
	csi->regs = ioremap(res->start, SZ_4K);
	if (!csi->regs) {
		printk(KERN_ERR "Unable to ioremap pxa950-camera regs\n");
		goto out;
	}

	if (csi->irq > 0) {
		ret = request_irq(csi->irq, csi_irq,
			IRQF_SHARED, "pxa950-csi", csi);
		if (ret)
			goto out;
	}

	spin_lock_init(&csi->dev_lock);

	csi_config(csi);

	return 0;

 out:

	if (csi) {
		if (csi->irq > 0)
			free_irq(csi->irq, csi);

		if (csi->regs)
			iounmap(csi->regs);

		kfree(csi);
	}

	return ret;
}

static int pxa950_csi_remove(struct platform_device *pdev)
{
	struct ccic_csi *csi = g_csi;

	if (csi) {
		if (csi->irq > 0)
			free_irq(csi->irq, csi);

		if (csi->regs)
			iounmap(csi->regs);

		kfree(csi);
	}

	return 0;
}


static struct platform_driver pxa950_csi_driver = {
	.driver = {
		.name = "pxa950-camera-csi"
	},
	.probe 		= pxa950_csi_probe,
	.remove 	= pxa950_csi_remove,

};

static int __devinit pxa950_camera_init(void)
{
	int ret = 0;

	ret = platform_driver_register(&pxa950_camera_driver);
	ret |= platform_driver_register(&pxa950_csi_driver);

	return ret;
}

static void __exit pxa950_camera_exit(void)
{
	platform_driver_unregister(&pxa950_camera_driver);
	platform_driver_unregister(&pxa950_csi_driver);
}

module_init(pxa950_camera_init);
module_exit(pxa950_camera_exit);

