/*
 *  drivers/misc/sirf_gsc3lt.c
 *
 *  Supports for SiRF AGPS chip - GSC3LT
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2 as
 *  published by the Free Software Foundation.
 *
 * (C) Copyright 2008 Borqs Ltd.
 */

#include <linux/init.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/device.h>
#include <linux/sysfs.h>
#include <linux/proc_fs.h>
#include <linux/string.h>
#include <linux/delay.h>
#include <linux/platform_device.h>
#include <mach/gpio.h>
#include <mach/gps.h>
#include <mach/mfp.h>

#ifdef CONFIG_HAS_WAKELOCK
#include <linux/wakelock.h>
#endif

#include "sirf.h"

static struct sirf *sirf_data;

/* running mode */
static int run_mode = STOP_MODE;
/* power status */
static int pm_status = POWER_OFF;

#ifdef CONFIG_HAS_WAKELOCK
static struct wake_lock pm_ops_suspend_lock;
#endif /* CONFIG_HAS_WAKELOCK */

/* power on/off */
static void set_power(int on_off)
{
	if (!sirf_data->pdata->set_power)
	{
		printk(KERN_DEBUG "no gps set_power\n");
		return;
	}

	if (on_off == OP_POWER_ON || on_off == OP_POWER_OFF) {
		sirf_data->pdata->set_power(on_off == OP_POWER_ON);
		printk(KERN_DEBUG "<SiRF> Chip Power O%s\n",(on_off == OP_POWER_ON)?"n":"ff");
		if( on_off == OP_POWER_ON )
			msleep(200);
		pm_status = (on_off == OP_POWER_ON)?FULL_POWER:POWER_OFF;
		if (on_off == OP_POWER_OFF)
			run_mode = STOP_MODE;
	}
}

/* run mode switch
 *
 * 1 1 - flash download mode
 * 0 0 - flash run mode
 * 1 0 - ROM run mode
 */
static void set_mode(int mode)
{
	switch (mode) {
		case FLASH_DOWN_MODE:
		case OP_FLASH_DOWN:
			/* flash download mode */
			gpio_set_value(sirf_data->pdata->flash_pin1, 1);
			gpio_set_value(sirf_data->pdata->flash_pin2, 1);
			printk(KERN_DEBUG "<SiRF> Flash Download Mode\n");
			break;
		case FLASH_RUN_MODE:
		case OP_FLASH_RUN:
			/* flash run mode */
			gpio_set_value(sirf_data->pdata->flash_pin1, 0);
			gpio_set_value(sirf_data->pdata->flash_pin2, 0);
			printk(KERN_DEBUG "<SiRF> Flash Running Mode\n");
			break;
		case ROM_RUN_MODE:
		case OP_ROM_RUN:
			/* ROM run mode */
			gpio_set_value(sirf_data->pdata->flash_pin1, 1);
			gpio_set_value(sirf_data->pdata->flash_pin2, 0);
			printk(KERN_DEBUG "<SiRF> ROM Running Mode\n");
			break;
		default:
			printk(KERN_ERR "<SiRF> Unknown operation %d\n", mode);
	}
}

/* chip status query */
static void get_status(void)
{
//	GPIO_DIRECTION_INPUT(STATUS_QUERY_PIN); 
	printk(KERN_DEBUG "<SiRF> Chip Status: 0x%08x\n",
				gpio_get_value(sirf_data->pdata->status_pin));
}

/* power manage */
static void set_status(void)
{
	/* up pulse trigger */
	gpio_set_value(sirf_data->pdata->pm_status_pin, 0);
	msleep(100);
	gpio_set_value(sirf_data->pdata->pm_status_pin, 1);
	msleep(200);
	gpio_set_value(sirf_data->pdata->pm_status_pin, 0);
	msleep(100);

	if (pm_status == FULL_POWER)
		pm_status = HIBERNATION;
	else if (pm_status == HIBERNATION)
		pm_status = FULL_POWER;
	printk(KERN_DEBUG "<SiRF> Chip %s\n", pm_status ? "Full Power" : "Hibernation");
}

/* chip reset */
static void set_reset(void)
{
	gpio_set_value(sirf_data->pdata->reset_pin, 1);
	gpio_set_value(sirf_data->pdata->reset_pin, 0);
	msleep(200);
	gpio_set_value(sirf_data->pdata->reset_pin, 1);
	printk(KERN_DEBUG "<SiRF> Chip Reset\n");
}

static void set_runmode(int mode)
{
	set_power(OP_POWER_ON);
	set_mode(mode);
	set_reset();
	run_mode = mode;
	pm_status = FULL_POWER;
	printk(KERN_DEBUG "<SiRF> Set to Mode %d\n", run_mode);
}

static ssize_t show_runmode(struct kobject *kobj,
			    struct bin_attribute *attr, 
			    char *buf, loff_t off, size_t count)
{
	if (off >= SIRF_MODE_SIZE)
		return 0;
	
	if (off + count > SIRF_MODE_SIZE)
		count = SIRF_MODE_SIZE - off;

#ifdef CONFIG_MACH_JADES
    printk(KERN_DEBUG "<SiRF> GSC3LT on Jades\n");
#endif
	printk(KERN_DEBUG "<SiRF> Mode %d, Status %d\n", run_mode, pm_status);
	count = sprintf(buf, "%d:%d\n", run_mode, pm_status);

	return count;
}

#ifdef CONFIG_HAS_WAKELOCK
static void gsc3lt_pm_lock(void)
{
    wake_lock(&pm_ops_suspend_lock);
    printk(KERN_INFO "<SiRF> Working, prevent system sleep...\n");
}

static void gsc3lt_pm_unlock(void)
{
    wake_unlock(&pm_ops_suspend_lock);
    printk(KERN_INFO "<SiRF> Finish, system can sleep...\n");
}
#endif

static ssize_t change_runmode(struct kobject *kobj,
			      struct bin_attribute *attr,
			      char *buf, loff_t off, size_t count)
{
	char *tail;
	int op;

	op = simple_strtol(buf, &tail, 10);

	switch(op) {
		case FLASH_DOWN_MODE:
		case FLASH_RUN_MODE:
		case ROM_RUN_MODE:
			if (op != run_mode)
				set_runmode(op);
			break;
		case OP_POWER_ON:
		case OP_POWER_OFF:
			set_power(op);
			break;
		case OP_FLASH_DOWN:
		case OP_FLASH_RUN:
		case OP_ROM_RUN:
			set_mode(op);
			break;
		case OP_PM:
			set_status();
			break;
		case OP_STATUS_QUERY:
			get_status();
			break;
		case OP_RESET:
			set_reset();
			break;
#ifdef CONFIG_HAS_WAKELOCK
        case OP_PM_LOCK:
            gsc3lt_pm_lock();
            break;
        case OP_PM_UNLOCK:
            gsc3lt_pm_unlock();
            break;
#endif
		default:
			printk(KERN_ERR "<SiRF> Unknown operation %d\n", op);
	}
	return count;
}

static struct bin_attribute sirf_mode_attr = {
	.attr = {
		.name = "sirf",
		.owner = THIS_MODULE,
		.mode = S_IRUGO | S_IWUGO,
	},
	.size = SIRF_MODE_SIZE,
	.read = show_runmode,
	.write = change_runmode,
};

static int __init sirf_gsd3lt_probe(struct platform_device *pdev)
{
	int ret = 0;

	sirf_data = kzalloc(sizeof(struct sirf), GFP_KERNEL);
	if (!sirf_data)
	      return -ENOMEM;

	sirf_data->pdata = pdev->dev.platform_data;

	ret = sysfs_create_bin_file(&(module_kset->kobj), &sirf_mode_attr);
	if (ret) {
		printk(KERN_ERR "<SiRF> Failed to create sys file\n");
		return -ENOMEM;
	}

	printk(KERN_INFO "<SiRF> SiRF GSC3LT driver probe...\n");
	return 0;
}

static int __devexit sirf_gsd3lt_remove(struct platform_device *pdev)
{
	sysfs_remove_bin_file(&(module_kset->kobj), &sirf_mode_attr);
	printk(KERN_INFO "<SiRF> SiRF GSC3LT driver exit...\n");
	return 0;
}

static struct platform_driver sirf_gsd3lt_driver = {
	.driver		= {
		.name	= "sirf",
	},
	.probe		= sirf_gsd3lt_probe,
	.remove 	= sirf_gsd3lt_remove,
};

static int __init sirf_init(void)
{
#ifdef CONFIG_HAS_WAKELOCK
	wake_lock_init(&pm_ops_suspend_lock, WAKE_LOCK_SUSPEND, "sirf_gsc3lt_pm_ops");
	printk(KERN_INFO "<SiRF> Register to PM\n");
#endif
	return platform_driver_register(&sirf_gsd3lt_driver);
}

static void __exit sirf_exit(void)
{
#ifdef CONFIG_HAS_WAKELOCK
	wake_lock_destroy(&pm_ops_suspend_lock);
#endif
	platform_driver_unregister(&sirf_gsd3lt_driver);
	printk(KERN_INFO "<SiRF> SiRF GSD3lt driver exit...\n");
}

module_init(sirf_init);
module_exit(sirf_exit);

MODULE_AUTHOR("Borqs Ltd.");
MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("SiRF AGPS Chip Driver (GSC3LT)");
