/*
 *  drivers/misc/sirf_gsd3tw.c
 *
 *  Supports for SiRF AGPS chip - GSD3tw
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2 as
 *  published by the Free Software Foundation.
 *
 * (C) Copyright 2008 Borqs Ltd.
 */

#include <linux/init.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/device.h>
#include <linux/sysfs.h>
#include <linux/proc_fs.h>
#include <linux/string.h>
#include <linux/delay.h>
#include <linux/platform_device.h>
#include <linux/switch_usbuart.h>
#include <mach/gpio.h>
#include <mach/gps.h>

#ifdef CONFIG_HAS_WAKELOCK
#include <linux/wakelock.h>
#endif

#include "sirf.h"

static struct sirf *sirf_data;

/* running mode */
static int run_mode = STOP_MODE;
/* power status */
static int pm_status = FULL_POWER;

#ifdef CONFIG_HAS_WAKELOCK
static struct wake_lock pm_ops_suspend_lock;
#endif /* CONFIG_HAS_WAKELOCK */

/* power on/off */
static void set_power(int on_off)
{
	if (!sirf_data->pdata->set_power)
	{
		printk(KERN_DEBUG "no gps set_power\n");
		return;
	}

	if (on_off == OP_POWER_ON || on_off == OP_POWER_OFF) {
		sirf_data->pdata->set_power(on_off == OP_POWER_ON);
		printk(KERN_DEBUG "<SiRF> Chip Power O%s\n",(on_off == OP_POWER_ON)?"n":"ff");
		msleep(1);
		pm_status = (on_off == OP_POWER_ON)?FULL_POWER:POWER_OFF;
		if (on_off == OP_POWER_OFF)
			run_mode = STOP_MODE;
	}
}

/* chip on */
static void set_onoff(void)
{
	if (sirf_data->pdata->on_off_pin == -1)
	{
		printk(KERN_DEBUG "no gps on_off pin\n");
		return;
	}

	/* up pulse trigger */
    mdelay(1);
	gpio_set_value(sirf_data->pdata->on_off_pin, 0);
	mdelay(1);
	gpio_set_value(sirf_data->pdata->on_off_pin, 1);
	//mdelay(1);
    udelay(200);
	gpio_set_value(sirf_data->pdata->on_off_pin, 0);
	printk(KERN_DEBUG "<SiRF> Chip Run\n");
}

/* chip reset */
static void set_reset(void)
{
	/* low pulse trigger */
	gpio_set_value(sirf_data->pdata->reset_pin, 1);
	mdelay(1);
	gpio_set_value(sirf_data->pdata->reset_pin, 0);
	//mdelay(1);
    msleep(200);
	gpio_set_value(sirf_data->pdata->reset_pin, 1);
	printk(KERN_DEBUG "<SiRF> Chip Reset\n");
}

/* chip stop */
static void set_stop(void)
{
	if (sirf_data->pdata->on_off_pin != -1)
	      gpio_set_value(sirf_data->pdata->on_off_pin, 0);
	gpio_set_value(sirf_data->pdata->reset_pin, 1);
}

static ssize_t show_runmode(struct kobject *kobj,
			struct bin_attribute *attr, 
			char *buf, loff_t off, size_t count)
{
	if (off >= SIRF_MODE_SIZE)
	      return 0;

	if (off + count > SIRF_MODE_SIZE)
	      count = SIRF_MODE_SIZE - off;

	printk(KERN_DEBUG "<SiRF> Mode %d, Status %d\n", run_mode, pm_status);
	count = sprintf(buf, "%d:%d\n", run_mode, pm_status);

	return count;
}

#ifdef CONFIG_HAS_WAKELOCK
static void gsd3tw_pm_lock(void)
{
	wake_lock(&pm_ops_suspend_lock);
	printk(KERN_INFO "<SiRF> Working, prevent system sleep...\n");
}

static void gsd3tw_pm_unlock(void)
{
	wake_unlock(&pm_ops_suspend_lock);
	printk(KERN_INFO "<SiRF> Finish, system can sleep...\n");
}
#endif

extern void enable_oscc_tout_s0(void);
extern void disable_oscc_tout_s0(void);

static ssize_t change_runmode(struct kobject *kobj,
			struct bin_attribute *attr,
			char *buf, loff_t off, size_t count)
{
	char *tail;
	int op;

	op = simple_strtol(buf, &tail, 10);

	switch(op) {
		case OP_POWER_ON:
		case OP_POWER_OFF:
			set_power(op);
			break;
		case RUN_MODE:
			usbuart_set_state(USBUART_AGPS, 1);
			set_reset();
			set_onoff();
			run_mode = RUN_MODE;
			printk(KERN_INFO "<SiRF> Chip running...\n");
			break;
		case STOP_MODE:
			set_stop();
			run_mode = STOP_MODE;
			printk(KERN_INFO "<SiRF> Chip stopped...\n");
			usbuart_set_state(USBUART_AGPS, 0);
			break;
		case OP_ONOFF:
			set_onoff();
			break;
		case OP_RESET:
			set_reset();
			break;
		case OP_CLK_ENABLE:
			enable_oscc_tout_s0();
			printk(KERN_INFO "<SiRF> Enable 32K clock\n");
			break;
		case OP_CLK_DISABLE:
			disable_oscc_tout_s0();
			printk(KERN_INFO "<SiRF> Disable 32K clock\n");
			break;
#ifdef CONFIG_HAS_WAKELOCK
		case OP_PM_LOCK:
			gsd3tw_pm_lock();
			break;
		case OP_PM_UNLOCK:
			gsd3tw_pm_unlock();
			break;
#endif
		default:
			printk(KERN_ERR "<SiRF> Unknown operation %d\n", op);
	}
	return count;
}

static struct bin_attribute sirf_mode_attr = {
	.attr = {
		.name = "sirf",
		.owner = THIS_MODULE,
		.mode = S_IRUGO | S_IWUGO,
	},
	.size = SIRF_MODE_SIZE,
	.read = show_runmode,
	.write = change_runmode,
};

static int __init sirf_gsd3tw_probe(struct platform_device *pdev)
{
	int ret = 0;

	sirf_data = kzalloc(sizeof(struct sirf), GFP_KERNEL);
	if (!sirf_data)
	      return -ENOMEM;

	sirf_data->pdata = pdev->dev.platform_data;


	ret = sysfs_create_bin_file(&(module_kset->kobj), &sirf_mode_attr);
	if (ret) {
		printk(KERN_ERR "<SiRF> Failed to create sys file\n");
		return -ENOMEM;
	}

	printk(KERN_INFO "<SiRF> SiRF GSD3tw driver probe...\n");
	return 0;
}

static int __devexit sirf_gsd3tw_remove(struct platform_device *pdev)
{
	sysfs_remove_bin_file(&(module_kset->kobj), &sirf_mode_attr);
	printk(KERN_INFO "<SiRF> SiRF GSD3tw driver remove...\n");
	return 0;
}

static struct platform_driver sirf_gsd3tw_driver = {
	.driver		= {
		.name	= "sirf",
	},
	.probe		= sirf_gsd3tw_probe,
	.remove 	= sirf_gsd3tw_remove,
};

static int __init sirf_init(void)
{
#ifdef CONFIG_HAS_WAKELOCK
	wake_lock_init(&pm_ops_suspend_lock, WAKE_LOCK_SUSPEND, "sirf_gsd3tw_pm_ops");
	printk(KERN_INFO "<SiRF> Register to PM\n");
#endif
	return platform_driver_register(&sirf_gsd3tw_driver);
}

static void __exit sirf_exit(void)
{
#ifdef CONFIG_HAS_WAKELOCK
	wake_lock_destroy(&pm_ops_suspend_lock);
#endif
	platform_driver_unregister(&sirf_gsd3tw_driver);
	printk(KERN_INFO "<SiRF> SiRF GSD3tw driver exit...\n");
}

module_init(sirf_init);
module_exit(sirf_exit);

MODULE_AUTHOR("Borqs Ltd.");
MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("SiRF AGPS Chip Driver (GSD3tw)");
