/*
 *  linux/drivers/mmc/host/pxa9xx_mci.c - PXA9XX MMCI driver
 *
 *  Based on linux/drivers/mmc/host/pxamci.c
 *
 *  Copyright:	(C) Copyright 2008 Marvell International Ltd.
 *              Mingwei Wang <mwwang@marvell.com>
 *
 *  linux/drivers/mmc/host/pxa.c - PXA MMCI driver
 *
 *  Copyright (C) 2003 Russell King, All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 *  This hardware is really sick:
 *   - No way to clear interrupts.
 *   - Have to turn off the clock whenever we touch the device.
 *   - Doesn't tell you how many data blocks were transferred.
 *  Yuck!
 *
 *	1 and 3 byte data transfers not supported
 *	max block length up to 1023
 */
#include <linux/module.h>
#include <linux/init.h>
#include <linux/ioport.h>
#include <linux/platform_device.h>
#include <linux/delay.h>
#include <linux/interrupt.h>
#include <linux/dma-mapping.h>
#include <linux/clk.h>
#include <linux/err.h>
#include <linux/mmc/host.h>
#include <linux/mmc/card.h>
#include <linux/timer.h>

#include <asm/dma.h>
#include <asm/io.h>
#include <asm/sizes.h>

#include <mach/pxa-regs.h>
#include <mach/mmc.h>
#include <mach/hardware.h>

#include "pxa9xx_mci.h"

#ifdef CONFIG_PXA3xx_DVFM
#include <mach/dvfm.h>
#include <mach/pxa3xx_dvfm.h>
#endif

#define DRIVER_NAME	"pxa9xx-mci"

#define NR_SG	32
#define CLKRT_OFF	(~0)
#define PXA_TIMER_TIMEOUT	500
#define ADMA_DESC_TBL_SIZE	PAGE_SIZE

struct adma_desc_table {
	u16 entry_attr;
	u16 entry_length;
	u32 entry_address;
};

struct pxa9xx_mci_host {
	struct mmc_host		*mmc;
	spinlock_t		lock;
	struct resource		*res;
	void __iomem		*base;
	struct clk		*clk;
	struct clk		*island_clk;
	struct clk		*bus_clk;
	int			clk_enabled;
	unsigned long		clkrate;
	int			irq;
	unsigned int		clkrt;
	unsigned int		dma_bufsz;
	unsigned int		segment;
	unsigned int		i_stat_sig;
	unsigned int		power_mode;
	struct pxamci_platform_data *pdata;

	struct mmc_request	*mrq;
	struct mmc_command	*cmd;
	struct mmc_data		*data;

	struct adma_desc_table	*sg_cpu;
	dma_addr_t		sg_dma;
	unsigned int		dma_len;
	unsigned int		dma_dir;

	struct timer_list	timer;
};

#if defined(CONFIG_PXA3xx_DVFM)
static struct dvfm_lock dvfm_lock = {
	.lock		= SPIN_LOCK_UNLOCKED,
	.count		= 0,
	.dev_idx	= -1,
};

static void set_dvfm_constraint(void)
{
	spin_lock_irqsave(&dvfm_lock.lock, dvfm_lock.flags);
	if (dvfm_lock.count++ == 0) {
		/* Disable Low power mode */
		dvfm_disable_op_name("D1", dvfm_lock.dev_idx);
		dvfm_disable_op_name("D2", dvfm_lock.dev_idx);
		if (cpu_is_pxa935() || cpu_is_pxa950())
			dvfm_disable_op_name("CG", dvfm_lock.dev_idx);
	} else
		dvfm_lock.count--;
	spin_unlock_irqrestore(&dvfm_lock.lock, dvfm_lock.flags);
}

static void unset_dvfm_constraint(void)
{
	spin_lock_irqsave(&dvfm_lock.lock, dvfm_lock.flags);
	if (dvfm_lock.count == 0) {
		spin_unlock_irqrestore(&dvfm_lock.lock, dvfm_lock.flags);
		return;
	}
	if (--dvfm_lock.count == 0) {
		/* Enable Low power mode */
		dvfm_enable_op_name("D1", dvfm_lock.dev_idx);
		dvfm_enable_op_name("D2", dvfm_lock.dev_idx);
		if (cpu_is_pxa935() || cpu_is_pxa950())
			dvfm_enable_op_name("CG", dvfm_lock.dev_idx);
	} else
		dvfm_lock.count++;
	spin_unlock_irqrestore(&dvfm_lock.lock, dvfm_lock.flags);
}
#else
static void set_dvfm_constraint(void) {}
static void unset_dvfm_constraint(void) {}
#endif

static void set_mmc_cken(struct pxa9xx_mci_host *host, int enable)
{
	if (enable) {
		pr_debug("%s: ON\n", __FUNCTION__);
		if (!host->clk_enabled) {
			host->clk_enabled = 1;
			clk_enable(host->bus_clk);
			clk_enable(host->island_clk);
			clk_enable(host->clk);
		} else
			del_timer_sync(&host->timer);
	} else {
		pr_debug("%s: OFF\n", __FUNCTION__);
		mod_timer(&host->timer, jiffies + PXA_TIMER_TIMEOUT);
	}
}

static void __attribute__((unused))
dump_registers(struct pxa9xx_mci_host *host)
{
	unsigned int val;
	int offset;

	for (offset = 0; offset < 0x44; offset += 4) {
		val = readl(host->base + offset);
		pr_debug("%08x: %08x\n", (unsigned int)host->base + offset, val);
	}
	val = readl(host->base + 0x48);
	pr_debug("%08x: %08x\n", (unsigned int)host->base + 0x48, val);
	val = readl(host->base + 0xFC);
	pr_debug("%08x: %08x\n", (unsigned int)host->base + 0xFC, val);
}

static void pxa9xx_mci_stop_clock(struct pxa9xx_mci_host *host)
{
	unsigned int val;

	val = readl(host->base + MM4_CNTL2);
	writel(val & ~(MM4CLKEN | INTCLKEN), host->base + MM4_CNTL2);
}

static void pxa9xx_mci_start_clock(struct pxa9xx_mci_host *host)
{
	unsigned int val;
	unsigned long timeout = 10000;

	val = readl(host->base + MM4_CNTL2);
	writel(val | INTCLKEN, host->base + MM4_CNTL2);
	do {
		val = readl(host->base + MM4_CNTL2);
		if (val & INTCLKSTB)
			break;
		udelay(1);
	} while (timeout--);
	if (!(val & INTCLKSTB))
		dev_err(mmc_dev(host->mmc), "unable to start clock\n");

	val &= ~(0xff << SDFREQ_OFFSET);
	val |= host->clkrt << SDFREQ_OFFSET;
	writel(val, host->base + MM4_CNTL2);

	val = readl(host->base + MM4_CNTL2);
	writel(val | MM4CLKEN, host->base + MM4_CNTL2);
	do {
		val = readl(host->base + MM4_CNTL2);
		if (val & MM4CLKEN)
			break;
		udelay(1);
	} while (timeout--);

	if (!(val & MM4CLKEN))
		dev_err(mmc_dev(host->mmc), "unable to start clock\n");
}

static void pxa9xx_mci_enable_irq(struct pxa9xx_mci_host *host, unsigned int mask)
{
	unsigned long flags;

	spin_lock_irqsave(&host->lock, flags);
	writel(mask & ~MM4_I_STAT_RESERVED, host->base + MM4_I_STAT);
	host->i_stat_sig |= mask & ~MM4_I_STAT_RESERVED;
	writel(host->i_stat_sig, host->base + MM4_I_STAT_EN);
	writel(host->i_stat_sig, host->base + MM4_I_SIG_EN);
	spin_unlock_irqrestore(&host->lock, flags);
}

static void pxa9xx_mci_disable_irq(struct pxa9xx_mci_host *host, unsigned int mask)
{
	unsigned long flags;

	spin_lock_irqsave(&host->lock, flags);
	writel(mask & ~MM4_I_STAT_RESERVED, host->base + MM4_I_STAT);
	host->i_stat_sig &= ~mask;
	writel(host->i_stat_sig, host->base + MM4_I_STAT_EN);
	writel(host->i_stat_sig, host->base + MM4_I_SIG_EN);
	spin_unlock_irqrestore(&host->lock, flags);
}

static int pxa9xx_mci_wait_reset(struct pxa9xx_mci_host *host)
{
	u32 timeout = 1000;
	u16 val;

	do {
		val = readl(host->base + MM4_CNTL2);
		if(!(val & (MSWRST | DATSWRST | CMDSWRST)))
			break;
		udelay(1);
	} while (timeout--);
	if (timeout)
		return 0;

	printk(KERN_ERR "Fatal: Wait RESET timeout.\n");

	return 1;
}

static void __attribute__((unused))
pxa9xx_mci_setup_data(struct pxa9xx_mci_host *host, struct mmc_data *data)
{
	unsigned int nob = data->blocks;

	host->segment = 0;

	host->data = data;

	if (data->flags & MMC_DATA_STREAM)
		nob = 0xffff;

	if (data->flags & MMC_DATA_READ) {
		host->dma_dir = DMA_FROM_DEVICE;
	} else {
		host->dma_dir = DMA_TO_DEVICE;
	}

	host->dma_len = dma_map_sg(mmc_dev(host->mmc), data->sg, data->sg_len,
				   host->dma_dir);
	writel(sg_dma_address(&data->sg[host->segment]), host->base +
		MM4_SYSADDR);

	pr_debug("PXA9XX_MCI: nob = %d, blksz = %d\n", nob, data->blksz);
	writel(nob << BLK_CNT_OFFSET | host->dma_bufsz << DMA_BUFSZ_OFFSET |
		data->blksz, host->base + MM4_BLK_CNTL);
}

static void pxa9xx_mci_setup_adma_data(struct pxa9xx_mci_host *host, struct mmc_data *data)
{
	unsigned int nob = data->blocks;
	unsigned int val;
	int i;

	host->data = data;

	if (data->flags & MMC_DATA_STREAM)
		nob = 0xffff;

	if (data->flags & MMC_DATA_READ) {
		host->dma_dir = DMA_FROM_DEVICE;
	} else {
		host->dma_dir = DMA_TO_DEVICE;
	}

	host->dma_len = dma_map_sg(mmc_dev(host->mmc), data->sg, data->sg_len,
				   host->dma_dir);

	for (i = 0; i < host->dma_len; i++) {
		host->sg_cpu[i].entry_address = sg_dma_address(&data->sg[i]);
		host->sg_cpu[i].entry_length = sg_dma_len(&data->sg[i]);
		host->sg_cpu[i].entry_attr = 0x0021;
	}
	/*The Last Link, Set End bit*/
	host->sg_cpu[host->dma_len - 1].entry_attr |= 0x2;

	/*Set ADMA System Address Reg*/
	writel(host->sg_dma, host->base + MMC4_AMDA_ADDR1);

	pr_debug("PXA9XX_MCI: nob = %d, blksz = %d\n", nob, data->blksz);
	writel(nob << BLK_CNT_OFFSET | host->dma_bufsz << DMA_BUFSZ_OFFSET |
		data->blksz, host->base + MM4_BLK_CNTL);

	val = readl(host->base + MM4_CNTL1);
	writel(val | 0x2 << DMA_SEL_OFFSET, host->base + MM4_CNTL1);
}

static void pxa9xx_mci_start_cmd(struct pxa9xx_mci_host *host, struct mmc_command *cmd)
{
	unsigned int rsp = 0;
	unsigned int xfrmd = 0;

	host->cmd = cmd;

	/*Set Response Type*/
	switch (mmc_resp_type(cmd)) {
	case MMC_RSP_NONE:
		break;

	case MMC_RSP_R1: /* r1, r5, r6, r7 */
		rsp = CMD_RSP_48BIT;
		xfrmd |= CRCCHKEN | IDXCHKEN;
		break;

	case MMC_RSP_R2: /* r2 */
		rsp = CMD_RSP_136BIT;
		xfrmd |= CRCCHKEN;
		break;

	case MMC_RSP_R3: /* r3, r4*/
		rsp = CMD_RSP_48BIT;
		break;

	case MMC_RSP_R1B: /* r1b */
		rsp = CMD_RSP_48BITB;
		xfrmd |= CRCCHKEN | IDXCHKEN;
		break;

	default:
		break;
	}

	/*Set Transfer mode regarding to data flag*/
	if (cmd->data)
	{
		/*Set Data Present*/
		xfrmd |= DPSEL | BLKCNTEN | DMA_EN | MS_BLKSEL;

		if (cmd->data->flags & MMC_DATA_READ)
			xfrmd |= DXFRDIR;
		else
			xfrmd &= ~DXFRDIR;
	}

	//pxa9xx_mci_enable_irq(host, ~CDINT);

	writel(cmd->arg, host->base + MM4_ARG);
	xfrmd |= cmd->opcode << CMD_IDX_OFFSET | rsp << RES_TYPE_OFFSET;

	if (!host->mmc->card || !mmc_card_sdio(host->mmc->card)) {
		u16 val = readl(host->base + MM4_STATE) >> 16;
		u16 mask = 0x1ff0;
		if ((val & mask) != mask)
			dev_dbg(mmc_dev(host->mmc), "WARN: CMD/DATA pins are not all high, PRE_STAT=0x%04x\n",
				val);
	}
	writel(xfrmd, host->base + MM4_CMD_XFRMD);

	pxa9xx_mci_start_clock(host);
}

static void pxa9xx_mci_finish_request(struct pxa9xx_mci_host *host,
	struct mmc_request *mrq)
{
	unsigned int val, set = 0;

	if (host->cmd->error)
		set = CMDSWRST;
	if (host->mrq->data && host->mrq->data->error)
		set |= DATSWRST;
	if (set) {
		val = readl(host->base + MM4_CNTL2);
		writel(val | set, host->base + MM4_CNTL2);
	}

	host->mrq = NULL;
	host->cmd = NULL;
	host->data = NULL;
	mmc_request_done(host->mmc, mrq);
	set_mmc_cken(host, 0);
	unset_dvfm_constraint();
}

static int pxa9xx_mci_cmd_done(struct pxa9xx_mci_host *host)
{
	struct mmc_command *cmd = host->cmd;

	if (mmc_resp_type(cmd) == MMC_RSP_R2) {
		cmd->resp[0] = readl(host->base + MM4_RESP3) << 8 |
				readl(host->base + MM4_RESP2) >> 24;
		cmd->resp[1] = readl(host->base + MM4_RESP2) << 8 |
				readl(host->base + MM4_RESP1) >> 24;
		cmd->resp[2] = readl(host->base + MM4_RESP1) << 8 |
				readl(host->base + MM4_RESP0) >> 24;
		cmd->resp[3] = readl(host->base + MM4_RESP0) << 8;
	} else {
		cmd->resp[0] = readl(host->base + MM4_RESP0);
		cmd->resp[1] = readl(host->base + MM4_RESP1);
		cmd->resp[2] = readl(host->base + MM4_RESP2);
		cmd->resp[3] = readl(host->base + MM4_RESP3);
	}

	//pxa9xx_mci_disable_irq(host, CMDCOMP);

	if ((!host->data && !(mmc_resp_type(cmd) & MMC_RSP_BUSY)) // for R1b, R5b response type,
								  // needs to wait XFRCOMP interrupt
			|| cmd->error || host->mrq->cmd != cmd) {
		pxa9xx_mci_finish_request(host, host->mrq);
	}

	return 1;
}

static int pxa9xx_mci_data_done(struct pxa9xx_mci_host *host)
{
	struct mmc_data *data = host->data;

	if (!data) {
		if (host->cmd && (mmc_resp_type(host->cmd) & MMC_RSP_BUSY)) {
			if (host->mrq->stop) {
				pxa9xx_mci_start_cmd(host, host->mrq->stop);
			} else
				pxa9xx_mci_finish_request(host, host->mrq);
			return 1;
		} else
			return 1;//BUG_ON(1);
	}

	dma_unmap_sg(mmc_dev(host->mmc), data->sg, host->dma_len,
		     host->dma_dir);

	/*
	 * There appears to be a hardware design bug here.  There seems to
	 * be no way to find out how much data was transferred to the card.
	 * This means that if there was an error on any block, we mark all
	 * data blocks as being in error.
	 */
	if (!data->error)
		data->bytes_xfered = data->blocks * data->blksz;
	else
		data->bytes_xfered = 0;

	//pxa9xx_mci_disable_irq(host, XFRCOMP);

	if (host->mrq->stop) {
		pxa9xx_mci_start_cmd(host, host->mrq->stop);
	} else {
		pxa9xx_mci_finish_request(host, host->mrq);
	}

	return 1;
}

static inline int query_irq(u16 *irqs, u16 mask)
{
	u16 re = *irqs & mask;
	*irqs &= ~mask;
	return re;
}

static int err_int_handler(struct pxa9xx_mci_host *host,  u16 err_stat)
{
	u16 err_stat_bk = err_stat;

	BUG_ON(err_stat & CEATA); /* CE-ATA mode only, not support yet*/
	if (query_irq(&err_stat, AHBT)) {
		printk(KERN_ERR "AXI Bus Response Error, non-reconverable!\n");
		if (host->data) {
			host->data->error = -ENOTRECOVERABLE;
			pxa9xx_mci_data_done(host);
		} else {
			host->cmd->error = -ENOTRECOVERABLE;
			pxa9xx_mci_cmd_done(host);
		}
		goto out;
	}
	if (query_irq(&err_stat, ADMA)) {
		printk(KERN_ERR "ADMA Error(status: 0x%x)!\n",
			readl(host->base + MMC4_AMDA_ERSTS));
		host->mrq->data->error = -EIO;
		pxa9xx_mci_data_done(host);
		goto out;
	}
	BUG_ON(err_stat & AC12); /* Auto CMD12 not used yet*/
	if (err_stat & ILMT) {
		printk(KERN_WARNING "The card consumes power over contoller's current limit, stop it.\n");
		if (host->cmd) {
			host->cmd->error = -1;
			pxa9xx_mci_finish_request(host, host->mrq);
		}
		goto out;
	}
	if (query_irq(&err_stat, CMD_ERR_MASK)) {
		pr_debug("CMD Line Error(status: 0x%04x)!\n", err_stat_bk);
		if (err_stat_bk & CTO)
			host->cmd->error = -ETIMEDOUT;
		else
			host->cmd->error = -EIO;
		pxa9xx_mci_cmd_done(host);
		goto out;
	}
	if (query_irq(&err_stat, DATA_ERR_MASK)) {
		if ((err_stat_bk & DTO) &&
				(mmc_resp_type(host->cmd) & MMC_RSP_BUSY)) {
			host->cmd->error = -ETIMEDOUT;
			pxa9xx_mci_finish_request(host, host->mrq);
			goto out;
		}

		pr_debug("DATA Line Error(status: 0x%04x)!\n", err_stat_bk);
		if (err_stat_bk & DTO)
			host->mrq->data->error = -ETIMEDOUT;
		else
			host->mrq->data->error = -EIO;
		pxa9xx_mci_data_done(host);
		goto out;
	}

out:
	BUG_ON(err_stat);
	return 1;
}

static irqreturn_t pxa9xx_mci_irq(int irq, void *devid)
{
	struct pxa9xx_mci_host *host = devid;
	u16 nor_stat, err_stat;
	u32 val;

	val = readl(host->base + MM4_I_STAT) &
			readl(host->base + MM4_I_STAT_EN);
	nor_stat = val & 0xffff;
	err_stat = (val >> 16) & 0xffff;
	if (err_stat)
		nor_stat |= ERRINT;
	if ((nor_stat & XFRCOMP) && (err_stat && DTO)) {
		err_stat &= ~DTO;
		if (!err_stat)
			nor_stat &= ~ERRINT;
	}
	if (nor_stat == 0) {
		return IRQ_NONE;
	}
	writel(val, host->base + MM4_I_STAT);
	pr_debug("PXA9XX_MCI: stat %08x\n", val);

	if (query_irq(&nor_stat, CDINT)) {
		mmc_signal_sdio_irq(host->mmc);
	}

	if (!nor_stat)
		goto out;

	if (!host->cmd) {
		nor_stat = 0;
		goto out;
	}

	if (!host->mrq && (nor_stat == XFRCOMP)) {
			return IRQ_RETVAL(1);
	}

	if (query_irq(&nor_stat, ERRINT)) {
		BUG_ON(!err_stat);

		if (nor_stat) {
			printk(KERN_WARNING "combined irq detected, stat=0x%x.\n",
				val);
			//BUG_ON(1);
			nor_stat = 0;
		}
		err_int_handler(host, err_stat);
		goto out;
	}

	BUG_ON(err_stat);

	if (query_irq(&nor_stat, CMDCOMP)) {
		pxa9xx_mci_cmd_done(host);
	}
	if (query_irq(&nor_stat, BUFRDRDY
			| BUFWRRDY | BGEVNT)) {
		BUG_ON(1);
	}
	if (query_irq(&nor_stat, DMAINT)) {
#ifdef CONFIG_MMC_ADMA
		BUG_ON(1);
#else
		if (++host->segment < host->dma_len)
			writel(sg_dma_address(&host->data->sg[host->segment]),
				host->base + MM4_SYSADDR);
		pr_debug("PXA9XX_MCI: len = %d, seg = %d\n",
				host->dma_len, host->segment);
#endif
	}
	if (query_irq(&nor_stat, XFRCOMP)) {
		pxa9xx_mci_data_done(host);
	}

out:
	BUG_ON(nor_stat);

	return IRQ_RETVAL(1);
}


static void pxa9xx_mci_request(struct mmc_host *mmc, struct mmc_request *mrq)
{
	struct pxa9xx_mci_host *host = mmc_priv(mmc);
	unsigned int val;
	unsigned int timeout = 10000;

	WARN_ON(host->mrq != NULL);

	/* release this constraint when request is finished */
	set_dvfm_constraint();
	set_mmc_cken(host, 1);

	if (pxa9xx_mci_wait_reset(host)) {
		mrq->cmd->error = -EBUSY;
		pxa9xx_mci_finish_request(host, mrq);
		return;
	}

	do {
		val = readl(host->base + MM4_STATE);
		if (!(val & DCMDINHBT || val & CCMDINHBT))
			break;
		udelay(1);
	} while (timeout--);
	if (val & DCMDINHBT) {
		pr_debug("DCMDINHBT of MM4_STATE: %08x\n", val);
		val = readl(host->base + MM4_CNTL2);
		writel(val | DATSWRST, host->base + MM4_CNTL2);
	}
	if (val & CCMDINHBT) {
		pr_debug("CCMDINHBT of MM4_STATE: %08x\n", val);
		val = readl(host->base + MM4_CNTL2);
		writel(val | CMDSWRST, host->base + MM4_CNTL2);
	}

	host->mrq = mrq;

	if (mrq->data)
#define CONFIG_MMC_ADMA
#ifdef CONFIG_MMC_ADMA
		pxa9xx_mci_setup_adma_data(host, mrq->data);
#else
		pxa9xx_mci_setup_data(host, mrq->data);
#endif

	pxa9xx_mci_start_cmd(host, mrq->cmd);
}

static int pxa9xx_mci_get_ro(struct mmc_host *mmc)
{
	struct pxa9xx_mci_host *host = mmc_priv(mmc);

	if (host->pdata && host->pdata->get_ro)
		return host->pdata->get_ro(mmc_dev(mmc));
	/* Host doesn't support read only detection so assume writeable */
	return 0;
}

static void pxa9xx_mci_set_ios(struct mmc_host *mmc, struct mmc_ios *ios)
{
	unsigned int val;
	struct pxa9xx_mci_host *host = mmc_priv(mmc);

	set_dvfm_constraint();
	set_mmc_cken(host, 1);
	if (pxa9xx_mci_wait_reset(host))
		return;

	if (ios->clock) {
		unsigned long rate = host->clkrate;
		unsigned int clk = rate / ios->clock;
		unsigned int shift;

		if (ios->clock == 48000000) {
			host->clkrt = 0x00;
			val = readl(host->base + MM4_CNTL1);
			writel(val | HISPEED, host->base + MM4_CNTL1);
		} else {
			if (!clk)
				clk = 1;

			shift = ffs(clk);
			if (rate / clk > ios->clock)
				shift++;
			host->clkrt = 1 << (shift - 2);
		}
		pxa9xx_mci_stop_clock(host);
		pxa9xx_mci_start_clock(host);
	} else {
		pxa9xx_mci_stop_clock(host);
		if (host->clkrt != CLKRT_OFF) {
			host->clkrt = CLKRT_OFF;
			set_mmc_cken(host, 0);
		}
	}

	if (host->power_mode != ios->power_mode) {
		host->power_mode = ios->power_mode;

		if (host->pdata && host->pdata->setpower)
			host->pdata->setpower(mmc_dev(mmc), ios->vdd);

		if (ios->power_mode == MMC_POWER_ON) {
			val = readl(host->base + MM4_CNTL1);
			//val &= ~(0x7 << VLTGSEL_OFFSET);
			//val |= 0x6 << VLTGSEL_OFFSET;
			writel(val | BUSPWR, host->base + MM4_CNTL1);
		}
	}

	if (ios->bus_width == MMC_BUS_WIDTH_8) {
		BUG_ON(!(mmc->caps & MMC_CAP_8_BIT_DATA));
		val = readl(host->base + MM4_CNTL1);
		writel(val | _8BITMD, host->base + MM4_CNTL1);
	} else if (ios->bus_width == MMC_BUS_WIDTH_4) {
		val = readl(host->base + MM4_CNTL1);
		writel((val & ~_8BITMD) | _4BITMD, host->base + MM4_CNTL1);
	} else {
		val = readl(host->base + MM4_CNTL1);
		writel(val & _8BITMD & ~_4BITMD, host->base + MM4_CNTL1);
	}

	unset_dvfm_constraint();
	set_mmc_cken(host, 0);
	pr_debug("PXA9XX_MCI: clkrt = %08x\n", host->clkrt);
}

static void pxa9xx_mci_enable_sdio_irq(struct mmc_host *host, int enable)
{
	struct pxa9xx_mci_host *pxa_host = mmc_priv(host);

	if (enable)
		pxa9xx_mci_enable_irq(pxa_host, CDINT);
	else
		pxa9xx_mci_disable_irq(pxa_host, CDINT);
}

static const struct mmc_host_ops pxa9xx_mci_ops = {
	.request		= pxa9xx_mci_request,
	.get_ro			= pxa9xx_mci_get_ro,
	.set_ios		= pxa9xx_mci_set_ios,
	.enable_sdio_irq	= pxa9xx_mci_enable_sdio_irq,
};

static irqreturn_t pxa9xx_mci_detect_irq(int irq, void *devid)
{
	struct pxa9xx_mci_host *host = mmc_priv(devid);

	mmc_detect_change(devid, host->pdata->detect_delay);
	return IRQ_HANDLED;
}

static void mmc_timer_handler(unsigned long data)
{
	struct mmc_host *mmc = (struct mmc_host *)data;
	struct pxa9xx_mci_host *host = mmc_priv(mmc);

	if (!mmc->card) {
		if (host->clk_enabled) {
			host->clk_enabled = 0;
			clk_disable(host->clk);
			clk_disable(host->island_clk);
			clk_disable(host->bus_clk);
		}
	}
	if (mmc->card && !mmc_card_sdio(mmc->card)) {
		pxa9xx_mci_stop_clock(host);
		if (host->clk_enabled) {
			host->clk_enabled = 0;
			clk_disable(host->clk);
			clk_disable(host->island_clk);
			clk_disable(host->bus_clk);
		}
	}
}

static int pxa9xx_mci_probe(struct platform_device *pdev)
{
	struct mmc_host *mmc;
	struct pxa9xx_mci_host *host = NULL;
	struct resource *r;
	int ret, irq, pages;
	unsigned int val;

	r = platform_get_resource(pdev, IORESOURCE_MEM, 0);
	irq = platform_get_irq(pdev, 0);
	if (!r || irq < 0)
		return -ENXIO;

	r = request_mem_region(r->start, SZ_256, DRIVER_NAME);
	if (!r)
		return -EBUSY;

	mmc = mmc_alloc_host(sizeof(struct pxa9xx_mci_host), &pdev->dev);
	if (!mmc) {
		ret = -ENOMEM;
		goto out_clk;
	}

	mmc->ops = &pxa9xx_mci_ops;

	/*
	 * We can do SG-DMA, but we don't because we never know how much
	 * data we successfully wrote to the card.
	 */
	mmc->max_phys_segs = ADMA_DESC_TBL_SIZE / sizeof(struct adma_desc_table);
	mmc->max_hw_segs = ADMA_DESC_TBL_SIZE / sizeof(struct adma_desc_table);

	/*
	 * Our hardware DMA can handle a maximum of one page per SG entry.
	 */
	mmc->max_seg_size = (1 << 16);

	mmc->max_blk_size = 2048;

	/*
	 * Block count register is 16 bits.
	 */
	mmc->max_blk_count = (1 << 16) - 1;

	mmc->max_req_size = mmc->max_blk_size * mmc->max_blk_count;

	host = mmc_priv(mmc);
	host->mmc = mmc;
	host->pdata = pdev->dev.platform_data;
	host->clkrt = CLKRT_OFF;

	if (cpu_is_pxa935() || cpu_is_pxa950() || cpu_is_pxa968()) {
		host->bus_clk = clk_get(&pdev->dev, "AXICLK");
		if (IS_ERR(host->bus_clk)) {
			ret = PTR_ERR(host->bus_clk);
			goto out_clk;
		}
		host->island_clk = clk_get(&pdev->dev, "IMUCLK");
		if (IS_ERR(host->island_clk)) {
			ret = PTR_ERR(host->island_clk);
			clk_put(host->bus_clk);
			goto out_clk;
		}
	}
	host->clk = clk_get(&pdev->dev, "PXA935_MMCCLK");
	if (IS_ERR(host->clk)) {
		ret = PTR_ERR(host->clk);
		clk_put(host->island_clk);
		clk_put(host->bus_clk);
		goto out_clk;
	}

	host->clkrate = 48000000;

	/*
	 * Calculate minimum clock rate, rounding up.
	 */
	mmc->f_min = (host->clkrate + 255) / 256;
	mmc->f_max = host->clkrate;

	mmc->ocr_avail = host->pdata ?
			 host->pdata->ocr_mask :
			 MMC_VDD_32_33|MMC_VDD_33_34;

	mmc->caps = MMC_CAP_4_BIT_DATA | MMC_CAP_8_BIT_DATA | MMC_CAP_SDIO_IRQ |
			MMC_CAP_MMC_HIGHSPEED | MMC_CAP_SD_HIGHSPEED;

	host->sg_cpu = dma_alloc_coherent(&pdev->dev, ADMA_DESC_TBL_SIZE, &host->sg_dma, GFP_KERNEL);
	if (!host->sg_cpu) {
		ret = -ENOMEM;
		goto out;
	}

	spin_lock_init(&host->lock);
	host->res = r;
	host->irq = irq;
	host->i_stat_sig = 0;

	host->base = ioremap(r->start, SZ_256);
	if (!host->base) {
		ret = -ENOMEM;
		goto out;
	}

	/*
	 * Calculate dma_bufsz
	 */
	pages = mmc->max_seg_size/4096;
	host->dma_bufsz = ffs(pages);
	if (mmc->max_seg_size / pages == 4096)
		host->dma_bufsz--;

	/*
	 * Ensure that the host controller is shut down, and setup
	 * with our defaults.
	 */
	init_timer(&host->timer);
	host->timer.function = mmc_timer_handler;
	host->timer.data = (unsigned long)mmc;
	set_mmc_cken(host, 1);
	set_dvfm_constraint();

	val = readl(host->base + MM4_CNTL2);
	writel(0xe0000 | val, host->base + MM4_CNTL2);

	pxa9xx_mci_stop_clock(host);
	pxa9xx_mci_enable_irq(host, ~CDINT);
	writel(0xffffffff, host->base + MM4_I_STAT);

	ret = request_irq(host->irq, pxa9xx_mci_irq, 0, DRIVER_NAME, host);
	if (ret)
		goto out;

	platform_set_drvdata(pdev, mmc);

	if (host->pdata && host->pdata->init)
		host->pdata->init(&pdev->dev, pxa9xx_mci_detect_irq, mmc);

	mmc_add_host(mmc);
	unset_dvfm_constraint();
	set_mmc_cken(host, 0);
	return 0;

out:
	clk_put(host->clk);
	clk_put(host->island_clk);
	clk_put(host->bus_clk);
out_clk:
	if (host) {
		if (host->base)
			iounmap(host->base);
		if (host->sg_cpu)
			dma_free_coherent(&pdev->dev, ADMA_DESC_TBL_SIZE, host->sg_cpu, host->sg_dma);
	}
	if (mmc)
		mmc_free_host(mmc);
	release_mem_region(r->start, SZ_256);
	unset_dvfm_constraint();
	return ret;
}

static int pxa9xx_mci_remove(struct platform_device *pdev)
{
	struct mmc_host *mmc = platform_get_drvdata(pdev);
	struct pxa9xx_mci_host *host;

	platform_set_drvdata(pdev, NULL);

	if (mmc) {
		set_dvfm_constraint();
		host = mmc_priv(mmc);

		if (host->pdata && host->pdata->exit)
			host->pdata->exit(&pdev->dev, mmc);

		del_timer_sync(&host->timer);
		mmc_remove_host(mmc);

		pxa9xx_mci_stop_clock(host);
		pxa9xx_mci_disable_irq(host, 0xffffffff);

		free_irq(host->irq, host);
		iounmap(host->base);
		dma_free_coherent(&pdev->dev, ADMA_DESC_TBL_SIZE, host->sg_cpu, host->sg_dma);

		clk_put(host->clk);
		clk_put(host->island_clk);
		clk_put(host->bus_clk);

		release_mem_region(host->res->start, SZ_256);

		mmc_free_host(mmc);
		unset_dvfm_constraint();
	}
	return 0;
}

#ifdef CONFIG_PM
static int pxa9xx_mci_suspend(struct platform_device *dev, pm_message_t state)
{
	struct mmc_host *mmc = platform_get_drvdata(dev);
	int ret = 0;

	if (mmc)
		ret = mmc_suspend_host(mmc, state);

	return ret;
}

static int pxa9xx_mci_resume(struct platform_device *dev)
{
	struct mmc_host *mmc = platform_get_drvdata(dev);
	int ret = 0;

	if (mmc)
		ret = mmc_resume_host(mmc);

	return ret;
}
#else
#define pxa9xx_mci_suspend	NULL
#define pxa9xx_mci_resume	NULL
#endif

static struct platform_driver pxa9xx_mci_driver = {
	.probe		= pxa9xx_mci_probe,
	.remove		= pxa9xx_mci_remove,
	.suspend	= pxa9xx_mci_suspend,
	.resume		= pxa9xx_mci_resume,
	.driver		= {
		.name	= DRIVER_NAME,
	},
};

static int __init pxa9xx_mci_init(void)
{
#ifdef CONFIG_PXA3xx_DVFM
	dvfm_register("PXA9XX_MMC", &(dvfm_lock.dev_idx));
#endif
	return platform_driver_register(&pxa9xx_mci_driver);
}

static void __exit pxa9xx_mci_exit(void)
{
	platform_driver_unregister(&pxa9xx_mci_driver);
#ifdef CONFIG_PXA3xx_DVFM
	dvfm_unregister("PXA9XX_MMC", &(dvfm_lock.dev_idx));
#endif
}

module_init(pxa9xx_mci_init);
module_exit(pxa9xx_mci_exit);

MODULE_DESCRIPTION("PXA9XX Multimedia Card Interface Driver");
MODULE_LICENSE("GPL");
