/*
 * drivers/mtd/copydata.c
 *
 * Copyright (C) 2008~2010 Borqs Corporation.
 *
 * Author:  Emichael Li <emichael.li@borqs.com>
 *
 * This software program is licensed subject to the GNU General Public License
 * (GPL).Version 2,June 1991, available at http://www.fsf.org/copyleft/gpl.html
 *
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/sched.h>
#include <linux/delay.h>
#include <asm/io.h>
#include <linux/mtd/mtd.h>
#include <asm/memory.h>
//#include <asm/arch/ddr.h>

/*
 * Following defination should be updated with Flash Layout and CP image size!
 */

/*
 * Actually, in general release, the OBM should read one header in Arbel 
 * binary (from 0xBF600000), which contain the final Arbel beginning 
 * adders (it is 0xBFA00000 from the binary). The OBM will update the 
 * start address in NTIM to let OBM load the arbel to 0xBFA00000.
 * This is the reason that we use 0xBFA00000 otherwise OBM use 0xBF600000. 
 */

#define ARBEL_RAM_OFFSET 	(0xBFA00000)
#define ARBEL_SIZE		(0x004E0000)
#define ARBEL_ROM_OFFSET	(0x00000000)	/* the offset in mtd[2] */

#define GB_RAM_OFFSET		(0xBFF00000)
#define GB_SIZE			(0x00100000)
#define GB_ROM_OFFSET		(0x00700000)	/* the offset in mtd[2] */

#define COPYDATA_PROGRESS	"progressing"	
#define COPYDATA_FINISHED	"ready"	
#define COPYDATA_TIME_THRESHOLD (5000)

unsigned char *g_copydata_state = COPYDATA_PROGRESS;

static wait_queue_head_t cpimage_wait_q;

unsigned char* get_cp_state(void)
{
	return g_copydata_state;
}

int check_cp_image(void)
{
	wait_event_interruptible_timeout(cpimage_wait_q, 
			(!strcmp(g_copydata_state, COPYDATA_FINISHED)), 
			msecs_to_jiffies(COPYDATA_TIME_THRESHOLD));

	BUG_ON(strcmp(g_copydata_state, COPYDATA_FINISHED));
	return 0;
}

int copy_cp_image(void)
{
	struct mtd_info *mtd = NULL;
	unsigned int length = 0;
	unsigned int va_arbel;
	unsigned int va_gb;

	g_copydata_state = COPYDATA_PROGRESS;

	printk(KERN_WARNING "begin to copy cp image ...\n");

	va_arbel = (unsigned int)ioremap_nocache(ARBEL_RAM_OFFSET, ARBEL_SIZE);
	va_gb = (unsigned int)ioremap_nocache(GB_RAM_OFFSET, GB_SIZE);

	BUG_ON((!va_arbel) || (!va_gb));

	mtd = get_mtd_device_nm("cpImage");
	if (mtd == NULL) {
		printk(KERN_WARNING "can't find cp image partition\n");
		return -ENODEV;
	}
	/*
	 * The last 0x40 bytes in DDR is used to store some information. 
	 * We shouldn't overwrite them.
	 */
	mtd->read(mtd, GB_ROM_OFFSET, GB_SIZE - 0x40, &length,
			(unsigned char *)va_gb);
	BUG_ON(length != GB_SIZE - 0x40);

	mtd->read(mtd, ARBEL_ROM_OFFSET, ARBEL_SIZE, &length,
			(unsigned char *)va_arbel);
	BUG_ON(length != ARBEL_SIZE);

	__iounmap((void *)va_arbel);
	__iounmap((void *)va_gb);

	printk(KERN_WARNING "... copy cp image finished\n");

	g_copydata_state = COPYDATA_FINISHED;

	wake_up_interruptible(&cpimage_wait_q);

	return 0;
}

static int copydata_thread(void *data)
{
	daemonize("copydata");
	strcpy(current->comm, "copydata");

	/* cp image ddr address unlock */
//	ddr_address_isolation(0);

	copy_cp_image();

	/* cp image ddr address lock */
//	ddr_address_isolation(0xBFB);

	complete_and_exit(NULL, 0);

	return 0;
}

/*****************************************************************************/
/*                                                                           */
/*                        COPYDATA Init & Fini                               */
/*                                                                           */
/*****************************************************************************/

static int __init copydata_init(void)
{
	init_waitqueue_head(&cpimage_wait_q);

	printk(KERN_WARNING "start copydata thread");
	kernel_thread(copydata_thread, NULL, 0);
	return 0;
}

static void __exit copydata_exit(void)
{

}

module_init(copydata_init);
module_exit(copydata_exit);

MODULE_DESCRIPTION("COPYDATA");
MODULE_LICENSE("GPL");
