/*
 *  linux/drivers/mtd/onenand/pxa3xx.c
 *
 *  Copyright (C) 2009 Marvell Corporation
 *  Haojian Zhuang <haojian.zhuang@marvell.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */
#include <linux/init.h>
#include <linux/device.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/mtd/mtd.h>
#include <linux/mtd/onenand.h>
#include <linux/mtd/partitions.h>
#include <linux/mtd/pxa3xx_bbm.h>
#include <linux/interrupt.h>
#include <linux/platform_device.h>
#include <linux/dma-mapping.h>
#include <linux/clk.h>

#include <asm/io.h>
#include <asm/dma.h>
#include <asm/mach/flash.h>
#include <mach/pxa-regs.h>
#include <mach/pxa3xx-regs.h>
#include <mach/hardware.h>
#include <mach/dma.h>
#include <asm/cacheflush.h>
#if defined(CONFIG_PXA3xx_DVFM)
#include <mach/dvfm.h>
#include <mach/pxa3xx_dvfm.h>
#endif
#include <linux/wakelock.h>

#define DRIVER_NAME	"pxa3xx-onenand"

#define SEQUENCE_TWO

#ifdef CONFIG_MTD_PARTITIONS
static const char *part_probes[] = { "cmdlinepart", NULL,  };
#endif

#define MAX_DMA_CHANNEL		4
#define DMA_BUF_SIZE		2400
#define ONENAND_RAM_SIZE	(1 << 11)

/* static memory controll address space */
#define SMC_START		0x4a000000
#define SMC_END			0x4a0000ff

#define MSC0_OFF		0x0008
#define MSC1_OFF		0x000C
#define MECR_OFF		0x0014
#define SXCNFG_OFF		0x001C
#define MCMEM0_OFF		0x0028
#define MCATT0_OFF		0x0030
#define MCIO0_OFF		0x0038
#define MEMCLKCFG_OFF		0x0068
#define CSADRCFG0_OFF		0x0080
#define CSADRCFG1_OFF		0x0084
#define CSADRCFG2_OFF		0x0088
#define CSADRCFG3_OFF		0x008C
#define CSADRCFG_P_OFF		0x0090
#define CSMSADRCFG_OFF		0x00A0

static struct wake_lock pxa_onenand_suspend_lock;

struct pxa3xx_onenand_info {
	struct platform_device	*pdev;
	struct mtd_info		mtd;
	struct mtd_partition	*parts;
	struct onenand_chip	onenand;

	unsigned long		phys_base;
	int			channel_cnt;
	int			dma_ch[MAX_DMA_CHANNEL];
	int			dma_ch_done[MAX_DMA_CHANNEL];
	void			*sram_addr;
	dma_addr_t		sram_addr_phys;
	char			*kbuf;
	struct clk		*smc_clk;
	struct pxa3xx_bbm	*bbm;

	struct completion	irq_done;
	struct completion	dma_done;
};

#if defined(CONFIG_PXA3xx_DVFM)
static struct timer_list nand_timer;

static struct dvfm_lock dvfm_lock = {
	.lock		= SPIN_LOCK_UNLOCKED,
	.dev_idx	= -1,
	.count		= 0,
};

static void set_dvfm_constraint(void)
{
	spin_lock_irqsave(&dvfm_lock.lock, dvfm_lock.flags);
	if (dvfm_lock.count++ == 0) {
		/* Disable Low power mode */
		dvfm_disable_op_name("D1", dvfm_lock.dev_idx);
		dvfm_disable_op_name("D2", dvfm_lock.dev_idx);
		if (cpu_is_pxa935())
			dvfm_disable_op_name("CG", dvfm_lock.dev_idx);
		wake_lock(&pxa_onenand_suspend_lock);
	}
	spin_unlock_irqrestore(&dvfm_lock.lock, dvfm_lock.flags);
}

static void unset_dvfm_constraint(void)
{
	spin_lock_irqsave(&dvfm_lock.lock, dvfm_lock.flags);
	if (dvfm_lock.count == 0) {
		spin_unlock_irqrestore(&dvfm_lock.lock, dvfm_lock.flags);
		return;
	}
	if (--dvfm_lock.count == 0) {
		/* Enable Low power mode */
		dvfm_enable_op_name("D1", dvfm_lock.dev_idx);
		dvfm_enable_op_name("D2", dvfm_lock.dev_idx);
		if (cpu_is_pxa935())
			dvfm_enable_op_name("CG", dvfm_lock.dev_idx);
		wake_unlock(&pxa_onenand_suspend_lock);
	}
	spin_unlock_irqrestore(&dvfm_lock.lock, dvfm_lock.flags);
}

static void nand_timer_handler(unsigned long data)
{
	unset_dvfm_constraint();
}

#else
static void set_dvfm_constraint() {}
static void unset_dvfm_constraint() {}
#endif


void pxa3xx_onenand_get_device(struct mtd_info *mtd)
{
	struct pxa3xx_onenand_info *info =
			container_of(mtd, struct pxa3xx_onenand_info, mtd);
	clk_enable(info->smc_clk);
	set_dvfm_constraint();
}

void pxa3xx_onenand_release_device(struct mtd_info *mtd)
{
	struct pxa3xx_onenand_info *info =
			container_of(mtd, struct pxa3xx_onenand_info, mtd);
	unset_dvfm_constraint();
	clk_disable(info->smc_clk);
}
EXPORT_SYMBOL(pxa3xx_onenand_get_device);
EXPORT_SYMBOL(pxa3xx_onenand_release_device);

static void dma_nodesc_handler(int irq, void *data)
{
	struct pxa3xx_onenand_info *info = (struct pxa3xx_onenand_info *)data;
	int i;

	DCSR(irq) = DCSR_STARTINTR|DCSR_ENDINTR|DCSR_BUSERR;

	for (i = 0; i < info->channel_cnt; i++) {
		if (info->dma_ch[i] == irq)
			info->dma_ch_done[i] = 1;
	}
	for (i = 0; i < info->channel_cnt; i++) {
		if (info->dma_ch_done[i] == 0) {
			return;
		}
	}
	complete(&info->dma_done);
	return;
}

static inline int pxa3xx_onenand_bufferram_offset(struct mtd_info *mtd, int area)
{
	struct onenand_chip *this = mtd->priv;

	if (ONENAND_CURRENT_BUFFERRAM(this)) {
		if (area == ONENAND_DATARAM)
			return mtd->writesize;
		if (area == ONENAND_SPARERAM)
			return mtd->oobsize;
	}
	return 0;
}

/**
 * pxa3xx_onenand_read_bufferram - [OneNAND Interface] Read the bufferram area
 * 			with Sync. Burst mode
 * @param mtd		MTD data structure
 * @param area		BufferRAM area
 * @param buffer	the databuffer to put/get data
 * @param offset	offset to read from or write to
 * @param count		number of bytes to read/write
 *
 * Read the BufferRAM area with Sync. Burst Mode
 */
static int pxa3xx_onenand_read_bufferram(struct mtd_info *mtd, int area,
		unsigned char *buffer, int offset, size_t count)
{
	struct pxa3xx_onenand_info *info =
			container_of(mtd, struct pxa3xx_onenand_info, mtd);
	struct onenand_chip *this = &info->onenand;
	int bram_offset, len, off, i;
	dma_addr_t dma_src, dma_dst;
	void *buf = (void *)buffer;
	int sum, orig_offset = offset;

	bram_offset = pxa3xx_onenand_bufferram_offset(mtd, area) + area + offset;
	if (offset & (ONENAND_RAM_SIZE - 1)) {
		/* make offset aligned */
		offset = ALIGN(offset - ONENAND_RAM_SIZE, ONENAND_RAM_SIZE);
	}
	len = ALIGN((count + orig_offset - offset) / MAX_DMA_CHANNEL - 31, 32);
	dma_dst = info->sram_addr_phys + offset;
	dma_src = info->phys_base + area + offset + pxa3xx_onenand_bufferram_offset(mtd, area);

	this->mmcontrol(mtd, ONENAND_SYS_CFG1_SYNC_READ);
	for (i = 0; i < MAX_DMA_CHANNEL; i++)
		info->dma_ch_done[i] = 0;
	INIT_COMPLETION(info->dma_done);

	if (MAX_DMA_CHANNEL > 1) {
		/* use multiple channel */
		sum = count + orig_offset - offset;
		len = ALIGN(sum / MAX_DMA_CHANNEL - 31, 32);
		for (i = 0, off = 0; i < MAX_DMA_CHANNEL; i++, off += len) {
			/* update remaing count */
			if (i == (MAX_DMA_CHANNEL - 1))
				len = sum;
			sum = sum - len;
			DCSR(info->dma_ch[i]) = DCSR_NODESC;
			DSADR(info->dma_ch[i]) = dma_src + off;
			DTADR(info->dma_ch[i]) = dma_dst + off;
			DCMD(info->dma_ch[i]) = DCMD_INCSRCADDR | DCMD_INCTRGADDR
				| DCMD_BURST32 | DCMD_WIDTH2
				| DCMD_ENDIRQEN | len;
			DCSR(info->dma_ch[i]) |= DCSR_RUN;
		}
		info->channel_cnt = MAX_DMA_CHANNEL;
	} else {
		/* use single channel */
		DCSR(info->dma_ch[0]) = DCSR_NODESC;
		DSADR(info->dma_ch[0]) = dma_src;
		DTADR(info->dma_ch[0]) = dma_dst;
		DCMD(info->dma_ch[0]) = DCMD_INCSRCADDR | DCMD_INCTRGADDR
					| DCMD_BURST32 | DCMD_WIDTH2
					| DCMD_ENDIRQEN
					| (count + orig_offset - offset);
		DCSR(info->dma_ch[0]) |= DCSR_RUN;
		info->channel_cnt = 1;
	}
	wait_for_completion(&info->dma_done);
	memcpy(buf, info->sram_addr + orig_offset - offset, count);

	this->mmcontrol(mtd, 0);
	return 0;
}

/**
 * pxa3xx_onenand_block_address - [DEFAULT] Get block address
 * @param this		onenand chip data structure
 * @param block		the block
 * @return		translated block address if DDP, otherwise same
 *
 * Setup Start Address 1 Register (F100h)
 */
static int pxa3xx_onenand_block_address(struct onenand_chip *this, int block)
{
	/* Device Flash Core select, NAND Flash Block Address */
	if (block & this->density_mask)
		return ONENAND_DDP_CHIP1 | (block ^ this->density_mask);

	return block;
}

/**
 * pxa3xx_onenand_bufferram_address - [DEFAULT] Get bufferram address
 * @param this		onenand chip data structure
 * @param block		the block
 * @return		set DBS value if DDP, otherwise 0
 *
 * Setup Start Address 2 Register (F101h) for DDP
 */
static int pxa3xx_onenand_bufferram_address(struct onenand_chip *this, int block)
{
	/* Device BufferRAM Select */
	if (block & this->density_mask)
		return ONENAND_DDP_CHIP1;

	return ONENAND_DDP_CHIP0;
}

/**
 * pxa3xx_onenand_page_address - [DEFAULT] Get page address
 * @param page		the page address
 * @param sector	the sector address
 * @return		combined page and sector address
 *
 * Setup Start Address 8 Register (F107h)
 */
static int pxa3xx_onenand_page_address(int page, int sector)
{
	/* Flash Page Address, Flash Sector Address */
	int fpa, fsa;

	fpa = page & ONENAND_FPA_MASK;
	fsa = sector & ONENAND_FSA_MASK;

	return ((fpa << ONENAND_FPA_SHIFT) | fsa);
}

/**
 * pxa3xx_onenand_buffer_address - [DEFAULT] Get buffer address
 * @param dataram1	DataRAM index
 * @param sectors	the sector address
 * @param count		the number of sectors
 * @return		the start buffer value
 *
 * Setup Start Buffer Register (F200h)
 */
static int pxa3xx_onenand_buffer_address(int dataram1, int sectors, int count)
{
	int bsa, bsc;

	/* BufferRAM Sector Address */
	bsa = sectors & ONENAND_BSA_MASK;

	if (dataram1)
		bsa |= ONENAND_BSA_DATARAM1;	/* DataRAM1 */
	else
		bsa |= ONENAND_BSA_DATARAM0;	/* DataRAM0 */

	/* BufferRAM Sector Count */
	bsc = count & ONENAND_BSC_MASK;

	return ((bsa << ONENAND_BSA_SHIFT) | bsc);
}

/**
 * pxa3xx_onenand_command - [DEFAULT] Send command to OneNAND device
 * @param mtd		MTD device structure
 * @param cmd		the command to be sent
 * @param addr		offset to read from or write to
 * @param len		number of bytes to read or write
 *
 * Send command to OneNAND device. This function is used for middle/large page
 * devices (1KB/2KB Bytes per page)
 */
static int pxa3xx_onenand_command(struct mtd_info *mtd, int cmd,
				loff_t addr, size_t len)
{
	struct pxa3xx_onenand_info *info =
			container_of(mtd, struct pxa3xx_onenand_info, mtd);
	struct onenand_chip *this = &info->onenand;
	int value, block, page;
	int reloc_block;

	if (info->bbm && info->bbm->table_init && info->bbm->search) {
		block = (int)(addr >> info->bbm->erase_shift);
		reloc_block = info->bbm->search(mtd, info->bbm, block);
		addr = ((reloc_block << this->erase_shift) |
				(addr & ((1 << this->erase_shift) - 1)));
	}

	/* Address translation */
	switch (cmd) {
	case ONENAND_CMD_UNLOCK:
	case ONENAND_CMD_LOCK:
	case ONENAND_CMD_LOCK_TIGHT:
	case ONENAND_CMD_UNLOCK_ALL:
		block = -1;
		page = -1;
		break;

	case ONENAND_CMD_ERASE:
	case ONENAND_CMD_BUFFERRAM:
	case ONENAND_CMD_OTP_ACCESS:
		block = (int) (addr >> this->erase_shift);
		page = -1;
		break;

	default:
		block = (int) (addr >> this->erase_shift);
		page = (int) (addr >> this->page_shift);

		if (ONENAND_IS_2PLANE(this)) {
			/* Make the even block number */
			block &= ~1;
			/* Is it the odd plane? */
			if (addr & this->writesize)
				block++;
			page >>= 1;
		}
		page &= this->page_mask;
		break;
	}

	/* NOTE: The setting order of the registers is very important! */
	if (cmd == ONENAND_CMD_BUFFERRAM) {
		/* Select DataRAM for DDP */
		value = pxa3xx_onenand_bufferram_address(this, block);
		this->write_word(value, this->base + ONENAND_REG_START_ADDRESS2);

		if (ONENAND_IS_2PLANE(this))
			/* It is always BufferRAM0 */
			ONENAND_SET_BUFFERRAM0(this);
		else
			/* Switch to the next data buffer */
			ONENAND_SET_NEXT_BUFFERRAM(this);

		return 0;
	}

	if (block != -1) {
		/* Write 'DFS, FBA' of Flash */
		value = pxa3xx_onenand_block_address(this, block);
		this->write_word(value, this->base + ONENAND_REG_START_ADDRESS1);

		/* Select DataRAM for DDP */
		value = pxa3xx_onenand_bufferram_address(this, block);
		this->write_word(value, this->base + ONENAND_REG_START_ADDRESS2);
	}

	if (page != -1) {
		/* Now we use page size operation */
		int sectors = 4, count = 4;
		int dataram;

		switch (cmd) {
		case ONENAND_CMD_READ:
		case ONENAND_CMD_READOOB:
			dataram = ONENAND_SET_NEXT_BUFFERRAM(this);
			break;

		default:
			if (ONENAND_IS_2PLANE(this) && cmd == ONENAND_CMD_PROG)
				cmd = ONENAND_CMD_2X_PROG;
			dataram = ONENAND_CURRENT_BUFFERRAM(this);
			break;
		}

		/* Write 'FPA, FSA' of Flash */
		value = pxa3xx_onenand_page_address(page, sectors);
		this->write_word(value, this->base + ONENAND_REG_START_ADDRESS8);

		/* Write 'BSA, BSC' of DataRAM */
		value = pxa3xx_onenand_buffer_address(dataram, sectors, count);
		this->write_word(value, this->base + ONENAND_REG_START_BUFFER);
	}

	/* Interrupt clear */
	this->write_word(ONENAND_INT_CLEAR, this->base + ONENAND_REG_INTERRUPT);

	/* Write command */
	this->write_word(cmd, this->base + ONENAND_REG_COMMAND);

	return 0;
}

int pxa3xx_onenand_reinit_bbt(struct mtd_info *mtd)
{
	struct pxa3xx_onenand_info *info =
		container_of(mtd, struct pxa3xx_onenand_info, mtd);

	struct pxa3xx_bbm *bbm = info->bbm;

	bbm->uninit(mtd, bbm);
	free_pxa3xx_bbm(bbm);
	bbm = alloc_pxa3xx_bbm();
	if (!bbm)
	{
		printk("ERROR:alloc_pxa3xx_bbm.\n");
		free_pxa3xx_bbm(bbm);
		goto out;
	}
	bbm->table_init = 0;

	bbm->flash_type = FLASH_ONENAND;
	bbm->page_shift = info->onenand.page_shift;
	printk("info->onenand.page_shift = 0x%x!!!!!!!!!!!!!!!\n.", info->onenand.page_shift);
	bbm->erase_shift = info->onenand.erase_shift;

	info->bbm->init(&info->mtd, info->bbm);

out:
	return onenand_default_bbt(mtd);
}

/*
 *  * is_full_reloc_table  scan relocation table 
 *   * @param pxa3xx_bbm          bbm structure
 *    * check relocation table is full or no
 *    */
int is_full_reloc_table(struct pxa3xx_bbm *bbm)
{
    int retval = 1;

    if (bbm->max_reloc_entry > bbm->table->total + 1) {
        retval = 0;
    }

    return retval;
}

static int pxa3xx_onenand_bbt(struct mtd_info *mtd)
{
    int ret;
	struct pxa3xx_onenand_info *info =
			container_of(mtd, struct pxa3xx_onenand_info, mtd);

	info->bbm = alloc_pxa3xx_bbm();
	if (!info->bbm)
		goto out;

	if (info->bbm->init) {
		info->bbm->flash_type = FLASH_ONENAND;
		info->bbm->page_shift = info->onenand.page_shift;
		info->bbm->erase_shift = info->onenand.erase_shift;

		if (info->bbm->init(&info->mtd, info->bbm))
			goto out;
	}

    /*
     * relaction table is full, create a memory bbt                         
     * relocation table is not full, don`t read flash, 
     * create a empty memory bbt.
     */
    if (is_full_reloc_table(info->bbm)) {
        ret = onenand_default_bbt(mtd);
    } else {
        ret = onenand_empty_bbt(mtd);
    }

    return ret;

out:
	return onenand_default_bbt(mtd);
}

static int __devinit pxa3xx_onenand_probe(struct platform_device *dev)
{
	struct pxa3xx_onenand_info *info;
	struct platform_device *pdev = dev;
	struct flash_platform_data *pdata = pdev->dev.platform_data;
	struct resource *res = pdev->resource;
	struct onenand_chip *this = NULL;
	unsigned long size = res->end - res->start + 1;
	int err, i, ret;
	char name[80];

	info = kzalloc(sizeof(struct pxa3xx_onenand_info), GFP_KERNEL);
	if (!info)
		return -ENOMEM;

	this = &info->onenand;
	init_completion(&info->irq_done);
	init_completion(&info->dma_done);

	if (!request_mem_region(res->start, size, pdev->name)) {
		err = -EBUSY;
		goto out_free_info;
	}

	info->phys_base = res->start;
	info->onenand.base = ioremap(res->start, size);
	if (!info->onenand.base) {
		err = -ENOMEM;
		goto out_release_mem_region;
	}

	size = ONENAND_REG_MANUFACTURER_ID - ONENAND_DATARAM;
	info->onenand.data = ioremap_cached(res->start + ONENAND_DATARAM, size);
	if (!info->onenand.data) {
		err = -ENOMEM;
		goto out_iounmap;
	}

	if (pdata->mmcontrol) {
		/*
		 * If synchronous mode is to be used, mmcontrol() should be
		 * defined in platform driver.
		 */
		info->onenand.mmcontrol = pdata->mmcontrol;
		info->onenand.read_bufferram = pxa3xx_onenand_read_bufferram;
	}
	info->onenand.command = pxa3xx_onenand_command;
	info->onenand.scan_bbt = pxa3xx_onenand_bbt;
	info->onenand.irq = platform_get_irq(pdev, 0);

	info->mtd.name = pdev->dev.bus_id;
	info->mtd.priv = &info->onenand;
	info->mtd.owner = THIS_MODULE;

	/* allocate DMA buffer */
	info->kbuf = kmalloc(DMA_BUF_SIZE, GFP_KERNEL);
	if (!info->kbuf) {
		ret = -ENOMEM;
		goto out_dmamem;
	}

	info->smc_clk = clk_get(NULL, "SMCCLK");
	clk_enable(info->smc_clk);

	for (i = 0; i < MAX_DMA_CHANNEL; i++)
		info->dma_ch[i] = -1;	/* not allocated */

	for (i = 0; i < MAX_DMA_CHANNEL; i++) {
		sprintf(name, "ONENAND_%d", i);
		info->dma_ch[i] = pxa_request_dma(name, DMA_PRIO_HIGH,
						dma_nodesc_handler,
						(void *)info);
		if (info->dma_ch[i] < 0) {
			ret = -ENODEV;
			goto out_freedma;
		}
	}

	info->sram_addr = dma_alloc_coherent(NULL, 4096,
				&info->sram_addr_phys, GFP_KERNEL);
	if (onenand_scan(&info->mtd, 1)) {
		err = -ENXIO;
		goto out_freedma;
	}

	/* remount mmcontrol in order to enable private read_bufferram */
	if (pdata->mmcontrol)
		info->onenand.read_bufferram = pxa3xx_onenand_read_bufferram;

#ifdef CONFIG_MTD_PARTITIONS
	err = parse_mtd_partitions(&info->mtd, part_probes, &info->parts, 0);
	if (err > 0)
		add_mtd_partitions(&info->mtd, info->parts, err);
	else if (err <= 0 && pdata->parts)
		add_mtd_partitions(&info->mtd, pdata->parts, pdata->nr_parts);
	else
#endif
		err = add_mtd_device(&info->mtd);

	dev_set_drvdata(&pdev->dev, info);

	return 0;

out_freedma:
	for (i = 0; i < MAX_DMA_CHANNEL; i++) {
		if (info->dma_ch[i] >= 0)
			pxa_free_dma(info->dma_ch[i]);
	}
	kfree(info->kbuf);
out_dmamem:
	iounmap(info->onenand.data);
out_iounmap:
	iounmap(info->onenand.base);
out_release_mem_region:
	release_mem_region(res->start, size);
out_free_info:
	kfree(info);

	return err;
}

static int __devexit pxa3xx_onenand_remove(struct platform_device *dev)
{
	struct platform_device *pdev = dev;
	struct pxa3xx_onenand_info *info = dev_get_drvdata(&pdev->dev);
	struct resource *res = pdev->resource;
	unsigned long size = res->end - res->start + 1;
	int i;

	dev_set_drvdata(&pdev->dev, NULL);

	if (info) {
		for (i = 0; i < MAX_DMA_CHANNEL; i++) {
			if (info->dma_ch[i] >= 0)
				pxa_free_dma(info->dma_ch[i]);
		}

		if (info->parts)
			del_mtd_partitions(&info->mtd);
		else
			del_mtd_device(&info->mtd);

		onenand_release(&info->mtd);
		if (info->bbm && info->bbm->uninit) {
			info->bbm->uninit(&info->mtd, info->bbm);
			free_pxa3xx_bbm(info->bbm);
		}
		release_mem_region(res->start, size);
		iounmap(info->onenand.base);
		iounmap(info->onenand.data);
		kfree(info);
	}

	return 0;
}

static struct platform_driver pxa3xx_onenand_driver = {
	.driver = {
		.name	= DRIVER_NAME,
		.owner	= THIS_MODULE,
	},
	.probe		= pxa3xx_onenand_probe,
	.remove		= pxa3xx_onenand_remove,
#ifdef CONFIG_PM
	.suspend	= NULL,
	.resume		= NULL,
#endif
};

static int __init pxa3xx_onenand_init(void)
{
#if defined(CONFIG_PXA3xx_DVFM)
	wake_lock_init(&pxa_onenand_suspend_lock, WAKE_LOCK_SUSPEND, "pxa3xx-onenand");
    dvfm_register("OneNAND", &dvfm_lock.dev_idx);
	init_timer(&nand_timer);
	nand_timer.function = nand_timer_handler;
	nand_timer.data = 0;

	mod_timer(&nand_timer, (jiffies + 60 * HZ));
	set_dvfm_constraint();
#endif
	return platform_driver_register(&pxa3xx_onenand_driver);
}

static void __exit pxa3xx_onenand_exit(void)
{
#if defined(CONFIG_PXA3xx_DVFM)
	wake_lock_destroy(&pxa_onenand_suspend_lock);
    dvfm_unregister("OneNAND", &dvfm_lock.dev_idx);
#endif
	platform_driver_unregister(&pxa3xx_onenand_driver);
}
EXPORT_SYMBOL(pxa3xx_onenand_reinit_bbt);

module_init(pxa3xx_onenand_init);
module_exit(pxa3xx_onenand_exit);

MODULE_LICENSE("GPL");
MODULE_ALIAS(DRIVER_NAME);
MODULE_AUTHOR("Haojian Zhuang <haojian.zhuang@samsung.com>");
MODULE_DESCRIPTION("Glue layer for OneNAND flash on PXA3xx");
