/*
 * linux/drivers/mtd/pxa3xx_bbm.c
 *
 * Bad Block Table support for PXA3xx.
 * Copyright (C) 2007 Marvell International Ltd.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 */

#include <linux/slab.h>
#include <linux/mtd/mtd.h>
#include <linux/mtd/onenand.h>
#include <linux/mtd/compatmac.h>
#include <linux/mtd/pxa3xx_bbm.h>

#include <mach/hardware.h>

#if defined(CONFIG_MTD_CREATE_BBT)
#include <asm/bootinfo.h>
#include <linux/delay.h>

#define ONENAND_512

#ifdef ONENAND_256
#define RESERVED_BLOCK_NUMBER 40
#define NONRESERVED_BLOCK_NUMBER 0x800
#define RESERVED_BLOCK_START 2047
#define RESERVED_BLOCK_END  2008
#elif defined(ONENAND_512)
#define RESERVED_BLOCK_NUMBER 80
#define NONRESERVED_BLOCK_NUMBER 0x1000
#define RESERVED_BLOCK_START 4095
#define RESERVED_BLOCK_END  4016
#endif

extern int pxa3xx_onenand_reinit_bbt(struct mtd_info *mtd);
static int pxa3xx_mark_reloc_tb(struct mtd_info *mtd, struct pxa3xx_bbm *bbm, unsigned int block);


#define PART(x)  ((struct mtd_part *)(x))
/* Our partition node structure */
struct mtd_part {
	struct mtd_info mtd;
	struct mtd_info *master;
	u_int32_t offset;
	int index;
	struct list_head list;
	int registered;
};
#endif

/* MOBM V2 is used on MhnP B0/B1/B2, MhnPL B1 and MhnL A0
 * MOBM V3 is used on MhnLV A0
 */
enum {
	MHN_OBM_NULL,
	MHN_OBM_V2,
	MHN_OBM_V3,
	MHN_OBM_INVAL,
};

#define NAND_RELOC_HEADER	0x524e

#define	YFW_DBG

#ifdef	YFW_DBG
#define yfw_dbg(format, arg...) printk(KERN_ERR __FILE__ ": " format "\n" , ## arg)

void dump_reloc_table(struct pxa3xx_bbm *bbm)
{
	int i;

	for (i = 0; i < bbm->table->total; i++) {
		if(65535 == bbm->reloc[i].to)
			continue;
		printk("block: %d is relocated to block: %d\n",
				bbm->reloc[i].from, bbm->reloc[i].to);
	}
}
#else
#define	yfw_dbg(format, arg...) do {} while (0)
void dump_reloc_table(struct pxa3xx_bbm *bbm) {}
#endif

static int max_bbt_slots = 40;

void dump_buf(unsigned char *buf, int len)
{
	int i;

	for (i = 0; i < len; i++) {
		printk(" %02x", buf[i]);
		if (0 == ((i + 1) % 16))
			printk("\n");
	}
	printk("\n");
}

/*
 * The relocation table management is different between MOBM V2 and V3.
 *
 * MOBM V2 is applied on chips taped out before MhnLV A0.
 * MOBM V3 is applied on chips taped out after MhnLV A0. It's also applied
 * on MhnLV A0.
 */
static int calc_obm_ver(void)
{
	unsigned int	cpuid;

	cpuid = read_cpuid(CPUID_ID);

	/* TavorP65e series */
	if (cpu_is_pxa935() || cpu_is_pxa950() || cpu_is_pxa968()) {
		return MHN_OBM_V3;
	}

	/* It's not xscale chip. */
	if ((cpuid & 0xFFFF0000) != 0x69050000)
		return MHN_OBM_INVAL;
	/* It's MhnP Ax */
	if ((cpuid & 0x0000FFF0) == 0x00006420)
		return MHN_OBM_V2;
	/* It's MhnP Bx */
	if ((cpuid & 0x0000FFF0) == 0x00006820) {
		if ((cpuid & 0x0F) <= 6)
			return MHN_OBM_V2;
		else
			return MHN_OBM_V3;
	}
	/* It's MhnL Ax */
	if ((cpuid & 0x0000FFF0) == 0x00006880) {
		if ((cpuid & 0x0F) == 0)
			return MHN_OBM_V2;
		else
			return MHN_OBM_V3;
	}
	/* It's MhnLV Ax */
	if ((cpuid & 0x0000FFF0) == 0x00006890)
		return MHN_OBM_V3;

	/* It's Tavor P A0 */
	if ((cpuid & 0x0000FFFF) == 0x00006834)
		return MHN_OBM_V2;

	/* It's Tavor P B0*/
	if ((cpuid & 0x0000FFFF) == 0x00006835)
		return MHN_OBM_V3;

	/* It's Tavor P B1*/
	if ((cpuid & 0x0000FFFF) == 0x00006837)
		return MHN_OBM_V3;

	/* It's Tavor P B2*/
	if ((cpuid & 0x0000FFFF) == 0x00006838)
		return MHN_OBM_V3;

	return MHN_OBM_INVAL;
}

/* add the relocation entry into the relocation table
 * It's valid on MOBM V3.
 * If the relocated block is bad, an new entry will be added into the
 * bottom of the relocation table.
 */
static int update_reloc_tb(struct mtd_info *mtd, struct pxa3xx_bbm *bbm, int block)
{
	struct reloc_table *table = bbm->table;
	struct reloc_item *item = bbm->reloc;
	struct erase_info instr;
	int obm, reloc_block, ret, entry_num = -1;
	int i;

	obm = calc_obm_ver();
	if (obm == MHN_OBM_V3) {
		if (bbm->table_init == 0) {
			printk(KERN_ERR "Error: the initial relocation \
					table can't be read\n");
			memset(table, 0, sizeof(struct reloc_table));
			table->header = NAND_RELOC_HEADER;
			bbm->table_init = 1;
		}

		if (table->total > bbm->max_reloc_entry) {
			printk("Relocation table exceed max num,");
			printk("cannot relocate block 0x%x\n", block);
			return -ENOSPC;
		}
		reloc_block = (mtd->size  >> bbm->erase_shift) - 1;
		
		//identify whether the block has been relocated
		for(i = table->total - 1; i >= 0; i --) {
			if(block == item[i].from)
				entry_num = i;
		}

		//find the available block with the largest number in reservered area
		for(i = table->total - 1; i >= 0; i --) {
			if (item[i].to != 65535) {
				if (reloc_block >= item[i].to)
					reloc_block = item[i].to - 1;
			} else {
				if (reloc_block >= item[i].from)
					reloc_block = item[i].from - 1;
			}
		}

		if (reloc_block < ((mtd->size >> bbm->erase_shift) - bbm->max_reloc_entry))
			return -ENOSPC;

		/* Make sure that reloc_block is pointing to a valid block */
		for (; ; reloc_block--) {
			for (i = table->total-1; i >= 0; i--) {
				if (reloc_block == item[i].from) {
					if (item[i].to != 65535)
						printk(KERN_ERR "Res block marked invalid \
								in reloc table\n");
					reloc_block--;
					i = table->total-1;
					if (reloc_block <
							((mtd->size >> bbm->erase_shift) -
							 bbm->max_reloc_entry))
						return -ENOSPC;
				}
			}
			/* The relocate table is full */
			if (reloc_block <
					((mtd->size >> bbm->erase_shift) - bbm->max_reloc_entry))
				return -ENOSPC;
			
			memset(&instr, 0, sizeof(struct erase_info));
			instr.mtd = mtd;
			instr.addr = reloc_block << bbm->erase_shift;
			instr.len = (1 << bbm->erase_shift);

			ret = mtd->erase(mtd, &instr);
			if (!ret)
				break;
			else {
				printk("status:0x%x found at erasing reloc block %d\n",
					ret, reloc_block);
				/* skip it if the reloc_block is also a 
				 * bad block
				 */
				item[table->total].from = reloc_block;
				item[table->total].to = 65535;
				table->total++;
			}
		}
		/* Create the relocated block information in the table */
		//when the block is relocated before, blob should modify the original entry to new
		//relocated block and the old relocated block point to 65535. If not the situation, 
		//create a new entry
		if (entry_num != -1) {
			item[table->total].from = item[entry_num].to;
			item[table->total].to = 65535;
			table->total++;
			item[entry_num].to = reloc_block;
		} else {
			item[table->total].from = block;
			item[table->total].to = reloc_block;
			table->total++;
		}
	} else {
		return -ENOSPC;
	}
	
	return 0;
}

/* Write the relocation table back to device, if there's room. */
static int sync_reloc_tb(struct mtd_info *mtd, struct pxa3xx_bbm *bbm, int *idx)
{
	int obm, start_page, len, retlen;
	uint8_t *tmp;


	if (*idx >= max_bbt_slots) {
		printk(KERN_ERR "Can't write relocation table to device \
				any more.\n");
		return -1;
	}

	if (*idx < 0) {
		printk(KERN_ERR "Wrong Slot is specified.\n");
		return -1;
	}

	len = 4;
	len += bbm->table->total << 2;
	obm = calc_obm_ver();
	if (obm == MHN_OBM_V3) {
		/* write to device */
		start_page = 1 << (bbm->erase_shift - bbm->page_shift);
		start_page = start_page - 1 - *idx;

		printk(KERN_DEBUG "DUMP relocation table before write. \
				page:0x%x\n", start_page);

		tmp = (uint8_t *)bbm->data_buf;
		mtd->write(mtd, start_page << bbm->page_shift,
				1 << bbm->page_shift, &retlen, tmp);
		/* write to idx */
		(*idx)++;
	}
	return 0;
}

#if defined(CONFIG_MTD_CREATE_BBT)
typedef struct
{
	int badblock_number;
	int reserved_area_badblock_number;
	int *badblock_identifier;
	int badblock_identifier_len;
	int *reserved_area_bad_block;
	int reserved_area_bad_block_len;
	int block_num;
	int block_size;
}bad_block_info, *bad_block_info_p;

static int check_reserved_area_badblock(struct mtd_info *mtd,
					struct pxa3xx_bbm *bbm,
					bad_block_info_p bb_info)
{
	int i,j;
	int reserved_area_badblock_number = 0;
	int badblock_number, *badblock_identifier, *badblock_in_reserved_area;

	if ((bb_info->badblock_identifier_len <= 0) || (bb_info->reserved_area_bad_block_len <= 0))
	{
		printk("badblock_indentifier_len or reserved_area_bad_block_len is wrong!\n");
		return -1;
	}

	badblock_identifier = bb_info->badblock_identifier;
	badblock_in_reserved_area = bb_info->reserved_area_bad_block;
	badblock_number = bb_info->badblock_number;

	for (i = 0, j = 0; i < badblock_number; i++)
	{
		if (badblock_identifier[i] >= 4016)
		{
			badblock_in_reserved_area[j++] = badblock_identifier[i];
			badblock_identifier[i] = -1;
			reserved_area_badblock_number++;
		}
	}

	bb_info->reserved_area_badblock_number = reserved_area_badblock_number;
	return 0;
}

static int scan_whole_flash_bad_block(struct mtd_info *mtd, struct pxa3xx_bbm *bbm, bad_block_info_p bb_info)
{
	int i, badblock_number, block_num, block_size, *badblock_identifier;
	struct onenand_chip *master_chip;
	struct mtd_part *part = PART(mtd);
	struct mtd_info *master_mtd = part->master;

	printk("scan_whole_flash_bad_block start!\n");
	if (bb_info->badblock_identifier_len <= 0)
	{
		printk("bb->info->badblock_identifier_len is wrong!\n");
		return -1;
	}

	master_chip = (struct onenand_chip *)(master_mtd->priv);
	badblock_number = 0;
	block_num = bb_info->block_num;
	badblock_identifier = bb_info->badblock_identifier;
	block_size =  bb_info->block_size;

	printk("block_size = 0x%x!!!!!!.\n", block_size);
	printk("block_num = 0x%x!!!!!!!.\n", block_num);
	for (i = 0; i < block_num; i++)
	{
		printk("now block is %d, block_size = 0x%x.\n", i, block_size * i);
		if (master_mtd->block_isbad(master_mtd,  (i * block_size)))
		{
			badblock_identifier[badblock_number++] = i;
		}
	}

	bb_info->badblock_number = badblock_number;

	return 0;
}


static int reset_relocation_talbe(struct mtd_info *mtd, struct pxa3xx_bbm *bbm)
{
	int retlen, ret = 0;
	u_char *bbt_identifer;
	struct onenand_chip *master_chip;
	struct mtd_part *part = PART(mtd);
	struct mtd_info *master_mtd = part->master;

#if 0
	master_chip->reset_bbt(master_mtd);
	return;
#endif

	bbt_identifer = kmalloc(0x800, GFP_KERNEL);
	memset(bbt_identifer, 0xFF, 0x800);
	master_chip = (struct onenand_chip *)(master_mtd->priv);

	printk("WRITE SPACE BBT BEFORE!!!\n");
	bbt_identifer[0] = 0x4E;
	bbt_identifer[1] = 0x52;
	bbt_identifer[2] = 0x00;
	bbt_identifer[3] = 0x00;

	mtd->write(mtd, 0x1f800, 0x800, &retlen, bbt_identifer);
	mtd->sync(mtd);
	//msleep(1000);
	printk("RESET BBT START!!\n");
	pxa3xx_onenand_reinit_bbt(master_mtd);
	printk("RESET BBT END!!\n");
	kfree(bbt_identifer);

	return ret;
}

static int initialize_relocation_table(struct mtd_info *mtd, struct pxa3xx_bbm *bbm, bad_block_info_p bb_info)
{
	u_char *bbt_identifer;
	int i, j, k, *badblock_identifier, *badblock_in_reserved_area, ret, retlen = 0;
	short to_block_num;

	if (bb_info->badblock_number == 0)
	{
		reset_relocation_talbe(mtd, bbm);
		printk("No bad block!!!!\n");
		return 0;
	}

	if ((bb_info->badblock_identifier_len <= 0) || (bb_info->reserved_area_bad_block_len <= 0))
	{
		printk("bad_block_info_p->badblock_indentifier_len or reserved_area_bad_block_len is wrong!\n");
		return -1;
	}

	to_block_num = RESERVED_BLOCK_START;
	bbt_identifer = kmalloc(0x800, GFP_KERNEL);
	memset(bbt_identifer, 0xFF, 0x800);
	badblock_identifier = bb_info->badblock_identifier;
	badblock_in_reserved_area = bb_info->reserved_area_bad_block;


	bbt_identifer[0] = 0x4E;
	bbt_identifer[1] = 0x52;
	bbt_identifer[2] = (short)(bb_info->badblock_number) & 0x00FF;
	bbt_identifer[3] = (short)(bb_info->badblock_number) >> 8;


	for (i = 0, j = 4; i < bb_info->badblock_number; i++)
	{
		if (badblock_identifier[i] == -1)
		      continue;

		for (k = 0; k < RESERVED_BLOCK_NUMBER; k++)
		{
			if (to_block_num == badblock_in_reserved_area[k])
			{
				to_block_num--;
				k = -1;
			}
		}
		if (to_block_num <= RESERVED_BLOCK_END)
		{
			printk("reserve area if FULL!!!\n");
			return -3;
		}

		printk("RELOCATION FROM %x, TO %x.\n", badblock_identifier[i], to_block_num);
		printk("BAD_BLOCK_IDENTIFIER LOW 2BYTE = %x, HIGH 2BYTE = %x.\n", (badblock_identifier[i] & 0x000000FF), (badblock_identifier[i] & 0x0000FF00) >> 8);
		printk("TO_BLOCK_NUM LOW 2BYTE = %x, HIGH 2BYTE = %x.\n", (to_block_num & 0x000000FF), (to_block_num & 0x0000FF00) >> 8);

		bbt_identifer[j++] = (badblock_identifier[i] & 0x000000FF);
		bbt_identifer[j++] = (badblock_identifier[i] & 0x0000FF00) >> 8;
		bbt_identifer[j++] = (to_block_num & 0x00FF);
		bbt_identifer[j++] = (to_block_num-- & 0xFF00) >> 8;
	}

	for (i = 0; i < 40; i++)
	{
		if (badblock_in_reserved_area[i] == 0)
		      break;

		bbt_identifer[j++] = (badblock_in_reserved_area[i] & 0x000000FF);
		bbt_identifer[j++] = (badblock_in_reserved_area[i] & 0x0000FF00) >> 8;
		bbt_identifer[j++] = 0xFF;
		bbt_identifer[j++] = 0xFF;
	}

	printk("before mtd->write!\n");
	printk("j = %d\n", j);
	//ret = mtd->write(mtd, 0x1f800, j - 1, &retlen, bbt_identifer);
	ret = mtd->write(mtd, 0x1f800, 0x800, &retlen, bbt_identifer);
	printk("after mtd->write! ret = %d\n", ret);
	kfree(bbt_identifer);
	mtd->sync(mtd);
	//msleep(1000);

	return ret;
}


static int erase_block_zero(struct mtd_info *mtd, struct pxa3xx_bbm *bbm)
{
	int ret;
	struct erase_info *erase;

	erase = kmalloc(sizeof(struct erase_info), GFP_KERNEL);
	erase->addr = 0x00000000;
	erase->len =  0x00020000;
	erase->mtd =  mtd;
	erase->callback =  NULL;
	ret = mtd->erase(mtd, erase);
	kfree(erase);
	printk("erase ret = %d.", ret);

	return ret;
}
#endif



static int pxa3xx_scan_reloc_tb(struct mtd_info *mtd, struct pxa3xx_bbm *bbm)
{
	struct reloc_table *table = bbm->table;
	int page, maxslot, obm, valid = 0;
	int retlen, ret;

	obm = calc_obm_ver();

	if (obm == MHN_OBM_V2) {
		/* On MOBM V2, the relocation table resides in the last page
		 * of the first block.
		 */
		page = (1 << (bbm->erase_shift - bbm->page_shift)) - 1;
		memset(bbm->data_buf, 0, mtd->writesize + mtd->oobsize);
		ret = mtd->read(mtd, page << bbm->page_shift, mtd->writesize,
				&retlen, bbm->data_buf);

		if (ret == 0) {
			if (table->header == NAND_RELOC_HEADER)
				valid = 1;
		}
	} else if (obm == MHN_OBM_V3) {
		/* On MOBM V3, there're several relocation tables in the first
		 * block.
		 * When new bad blocks are found, a new relocation table will
		 * be generated and written back to the first block. But the
		 * original relocation table won't be erased. Even if the new
		 * relocation table is written wrong, system can still find an
		 * old one.
		 * One page contains one slot.
		 */
		maxslot = 1 << (bbm->erase_shift - bbm->page_shift);
		page = maxslot - max_bbt_slots;
		for (; page < maxslot; page++) {
			memset(bbm->data_buf, 0,
					mtd->writesize + mtd->oobsize);
			ret = mtd->read(mtd, (page << bbm->page_shift),
					mtd->writesize, &retlen, bbm->data_buf);

			if (ret == 0) {
				if (table->header != NAND_RELOC_HEADER) {
					continue;
				} else {
					bbm->current_slot = maxslot - page - 1;	
					valid = 1;
					break;
				}
			}
		}
	} else {
		printk(KERN_ERR "The version of MOBM isn't supported\n");
	}

	if (valid) {
		printk("relocation table at page:%d\n", page);
        bbm->table_init = 1;
        dump_reloc_table(bbm);
    } else {
#if defined(CONFIG_MTD_CREATE_BBT)
        if (get_boot_mode() == BOOT_MODE_SDL_FLASHMODE)
        {
            struct mtd_part *part = PART(mtd);
            struct mtd_info *master_mtd = part->master;
            int i, badblock_number = 0;
            int block_num = (uint32_t)(master_mtd->size) / master_mtd->erasesize;
            int page_num_per_block = master_mtd->erasesize / master_mtd->writesize;
            int page_size = master_mtd->writesize;
            int block_size = master_mtd->erasesize;
            int flash_whole_size = master_mtd->size;

            struct onenand_chip *master_chip;

            bad_block_info_p bb_info;
            bb_info = kmalloc(sizeof(bad_block_info), GFP_KERNEL);
            memset(bb_info, 0x00, sizeof(bad_block_info));


            bb_info->badblock_identifier = kmalloc(NONRESERVED_BLOCK_NUMBER, GFP_KERNEL);
            bb_info->badblock_identifier_len = NONRESERVED_BLOCK_NUMBER;
            memset(bb_info->badblock_identifier, 0xFF, NONRESERVED_BLOCK_NUMBER);


            bb_info->reserved_area_bad_block = kmalloc(RESERVED_BLOCK_NUMBER * 4, GFP_KERNEL);
            bb_info->reserved_area_bad_block_len = RESERVED_BLOCK_NUMBER;
            memset(bb_info->reserved_area_bad_block, 0x00, RESERVED_BLOCK_NUMBER * 4);

            master_chip = (struct onenand_chip *)(master_mtd->priv);

            reset_relocation_talbe(mtd, bbm);

            printk("nand flash size = 0x%x, block size = 0x%x, page size = 0x%x, block_num = 0x%x,\
                    page_num_per_block = 0x%x.\n", flash_whole_size, block_size, page_size,\
		    block_num, page_num_per_block);


            bb_info->block_num = block_num;
            bb_info->block_size = block_size;
            if (scan_whole_flash_bad_block(mtd, bbm, bb_info) == -1)
                printk("ERROR:scan_whole_flash_bad_block error!\n");


            printk("!!!!!badblock_number = %d.\n", badblock_number);
            for (i = 0; i < bb_info->badblock_number; i++)
                printk("BAD BLOCK = %d.\n", bb_info->badblock_identifier[i]);

            if (check_reserved_area_badblock(mtd, bbm, bb_info) == -1)
                printk("ERROR:check_reserved_area_badblock error!\n");

            erase_block_zero(mtd, bbm);
            for (i = 0; i < RESERVED_BLOCK_NUMBER; i++)
                printk("bad_block_in_reserved_area = %d.\n", bb_info->reserved_area_bad_block[i]);

            printk("!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!\n");
            mtd->sync(mtd);
            //msleep(1000);

            if (initialize_relocation_table(mtd, bbm, bb_info) != 0)
            {
                printk("ERROR:initialize_relocation_table is wrong!\n");
                return -EINVAL;
            }


            memset((unsigned char *)bbm->table, 0,
                    sizeof(struct reloc_table));
            bbm->table_init = 0;

            printk("kkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkk!!!\n");
            //while(1);

            //reinit the new relocation table structer which kernel used.
            printk("MASTER SCAN BBT, UPDATE TRUE BBT BEFORE!!!\n");
            //master_chip->reinit_bbt(master_mtd);
            pxa3xx_onenand_reinit_bbt(master_mtd);
            printk("MASTER SCAN BBT, UPDATE TRUE BBT END!!!\n");

            kfree(bb_info->badblock_identifier);
            kfree(bb_info->reserved_area_bad_block);
            kfree(bb_info);
        }
        else
        {
            /* There should be a valid relocation table slot at least. */
            printk(KERN_ERR "NO VALID relocation table can be \
                    recognized\n");
            printk(KERN_ERR "CAUTION: It may cause unpredicated error\n");
            printk(KERN_ERR "Please re-initialize the NAND flash.\n");
            memset((unsigned char *)bbm->table, 0,
                    sizeof(struct reloc_table));
            bbm->table_init = 0;
            return -EINVAL;
        }
#else
	/* There should be a valid relocation table slot at least. */
		printk(KERN_ERR "NO VALID relocation table can be \
				recognized\n");
		printk(KERN_ERR "CAUTION: It may cause unpredicated error\n");
		printk(KERN_ERR "Please re-initialize the NAND flash.\n");
		memset((unsigned char *)bbm->table, 0,
				sizeof(struct reloc_table));
		bbm->table_init = 0;
		return -EINVAL;
#endif
    }
    return 0;
}

static int pxa3xx_init_reloc_tb(struct mtd_info *mtd, struct pxa3xx_bbm *bbm)
{
	int size = mtd->writesize + mtd->oobsize;
	int page_per_block = 1 << (bbm->erase_shift - bbm->page_shift);

	bbm->max_reloc_entry = (int)(((int)mtd->size / (1 << bbm->erase_shift)) * 2) / 100;
	bbm->table_init = 0;

	if (page_per_block == 64){
		max_bbt_slots = 40;
		bbm->max_reloc_entry = max_bbt_slots;
	} else {
		max_bbt_slots = 8;
		bbm->max_reloc_entry = max_bbt_slots;
	}

	bbm->data_buf = kzalloc(size, GFP_KERNEL);
	if (!bbm->data_buf) {
		return -ENOMEM;
	}
	bbm->table = (struct reloc_table *)bbm->data_buf;
	memset(bbm->table, 0x0, sizeof(struct reloc_table));

	bbm->reloc = (struct reloc_item *)((uint8_t *)bbm->data_buf +
			sizeof (struct reloc_item));
	memset(bbm->reloc, 0x0,
			sizeof(struct reloc_item) * bbm->max_reloc_entry);

	return pxa3xx_scan_reloc_tb(mtd, bbm);
}

static int pxa3xx_uninit_reloc_tb(struct mtd_info *mtd, struct pxa3xx_bbm *bbm)
{
	kfree(bbm->data_buf);
	return 0;
}

/* Find the relocated block of the bad one.
 * If it's a good block, return 0. Otherwise, return a relocated one.
 * idx points to the next relocation entry
 * If the relocated block is bad, an new entry will be added into the
 * bottom of the relocation table.
 */
static int pxa3xx_search_reloc_tb(struct mtd_info *mtd, struct pxa3xx_bbm *bbm, unsigned int block)
{
	struct reloc_table *table = bbm->table;
	struct reloc_item *item = bbm->reloc;
	int i, max, reloc_block;

	if ((block <= 0) ||
			(block > ((mtd->size >> bbm->erase_shift) - bbm->max_reloc_entry)) ||
			(bbm->table_init == 0) || (table->total == 0))
		return block;

	if (table->total > bbm->max_reloc_entry)
		table->total = bbm->max_reloc_entry;

	/* If can't find reloc tb entry for block, return block */
	reloc_block = block;
	max = table->total;
	for (i = max-1; i >= 0; i--) {
		if (block == item[i].from) {
			reloc_block = item[i].to;
			break;
		}
	}

	return reloc_block;
}

static int pxa3xx_mark_reloc_tb(struct mtd_info *mtd, struct pxa3xx_bbm *bbm, unsigned int block)
{
	int ret = 0;

	ret = update_reloc_tb(mtd, bbm, block);
	if (ret)
		return ret;

	ret = sync_reloc_tb(mtd, bbm, &(bbm->current_slot));
	return ret;
}

struct pxa3xx_bbm* alloc_pxa3xx_bbm(void)
{
	/* FIXME: We don't want to add module_init entry
	 * here to avoid dependency issue.
	 */
	struct pxa3xx_bbm *bbm;

	bbm = kzalloc(sizeof(struct pxa3xx_bbm), GFP_KERNEL);
	if (!bbm)
		return NULL;

	bbm->init = pxa3xx_init_reloc_tb;
	bbm->uninit = pxa3xx_uninit_reloc_tb;
	bbm->search = pxa3xx_search_reloc_tb;
	bbm->markbad = pxa3xx_mark_reloc_tb;

	return bbm;
}
EXPORT_SYMBOL(alloc_pxa3xx_bbm);

void free_pxa3xx_bbm(struct pxa3xx_bbm *bbm)
{
	if (bbm) {
		kfree(bbm);
		bbm = NULL;
	}
}
EXPORT_SYMBOL(free_pxa3xx_bbm);

