/* 
 * drivers/omux/core/omux_core.c 
 *
 * Open MUX's core layer.
 *
 * Copyright (C) 2009~2010 Borqs, Inc.
 * Author: Emichael Li <emichael.li@borqs.com>
 *
 * This software is licensed under the terms of the GNU General Public
 * License version 2, as published by the Free Software Foundation, and
 * may be copied, distributed, and modified under those terms.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

#include <linux/module.h>
#include <linux/moduleparam.h>
#include <asm/system.h>
#include <asm/uaccess.h>
#include <linux/bitops.h>
#include <linux/string.h>
#include <linux/mm.h>
#include <linux/interrupt.h>
#include <linux/in.h>
#include <linux/tty.h>
#include <linux/errno.h>
#include <linux/delay.h>
#include <linux/init.h>
#include <linux/sysdev.h>
#include <linux/omux.h>

#define OMUX_DEBUG_MASK 	g_omux_core_debug_mask
#define OMUX_DEBUG_LOG 		OMUX_CORE_DEBUG_LOG
#define OMUX_DEBUG_FUNC 	OMUX_CORE_DEBUG_FUNC
#define OMUX_DEBUG_DATA 	OMUX_CORE_DEBUG_DATA

#include "omux_debug.h"

static struct kobject *omux_kobj;
static int omux_kobj_refcount;

static struct list_head omux_drv_list;	/* many drivers can be registered */
static struct list_head omux_dev_list;	/* many devices can be registered */

enum {
	OMUX_CORE_DEBUG_FUNC = 1U << 0,
	OMUX_CORE_DEBUG_DATA = 1U << 1,
	OMUX_CORE_DEBUG_LOG = 1U << 2,
};

unsigned int g_omux_core_debug_mask = 0;//OMUX_CORE_DEBUG_FUNC | OMUX_CORE_DEBUG_DATA | OMUX_CORE_DEBUG_LOG;

unsigned int g_omux_core_dump_data = 16;

#define OMUX_DUMP_BUF_SIZE (8192)
static char dumpbuf[OMUX_DUMP_BUF_SIZE];
void omux_dump_data(const unsigned char *buf, int count)
{
	int i, cnt = count;
	int len = 0;

	BUG_ON(count < 0);

	if (count == 0)
		return;

	memset(dumpbuf, 0, OMUX_DUMP_BUF_SIZE);

	cnt = cnt < (OMUX_DUMP_BUF_SIZE / 4) ? cnt : (OMUX_DUMP_BUF_SIZE / 4);

	for (i = 0; i < cnt; i++) {
		len += snprintf(dumpbuf + len, 10, "%02x ", buf[i]);
		if (0 == ((i + 1) % 16))
			len += snprintf(dumpbuf + len, 20, "\n            ");
	}

	printk("     (%4d) %s\n", cnt, dumpbuf);
}

int omux_tx_buf(struct omux_driver *drv, unsigned char *buf, ssize_t len)
{
	int retval = 0;
	struct omux_device *dev;

	DENTER("drv=%p buf=%p len=%d\n", drv, buf, len);

	BUG_ON(!drv || !buf || (len <= 0));

	dev = drv->dev;

	if (!dev || !dev->tx_buf) {
		printk(KERN_ERR "[OMUX] %s's device is not ready\n", drv->name);
		retval = 0;
	} else {
		OMUX_BUG_ON(dev->drv != drv);
		retval = dev->tx_buf(buf, len);
	}

	DLEAVE("ret=%d\n", retval);
	return retval;
}

/*
 * receive buf from low layer, it's called in ipc layer.
 */
int omux_rx_buf_handler(struct omux_device *dev, unsigned char *buf, ssize_t len)
{
	int retval = 0;
	struct omux_driver *drv;

	DENTER("dev=%p buf=%p len=%d\n", dev, buf, len);

	BUG_ON(!dev || !buf || (len <= 0));

	drv = dev->drv;

	if (!drv && !drv->rx_buf_handler) {
		printk(KERN_ERR "[OMUX] rx_buf_handler not ready\n");
		retval = -EINVAL;
	} else {
                OMUX_BUG_ON(drv->dev != dev);
                retval = drv->rx_buf_handler(buf, len);
	}

	DLEAVE("ret=%d\n", retval);
	return retval;
}

/*****************************************************************************/
/*                                                                           */
/*                     OMUX Driver(protocol) Interface                       */
/*                                                                           */
/*****************************************************************************/
int omux_register_driver(struct omux_driver *drv)
{
	struct list_head *list;
	struct omux_driver *p;
	struct omux_device *q;
	int retval = -ENXIO;

	DENTER("drv=%p\n", drv);

	BUG_ON(!drv || drv->dev || !drv->name);

	list_for_each(list, &omux_drv_list) {
		p = list_entry(list, struct omux_driver, list);
		if (!strncmp(drv->name, p->name, strlen(p->name))) {
			retval = -EEXIST;	/* driver is in list */
			goto reg_drv_exist;
		}
	}

	list_add_tail(&(drv->list), &omux_drv_list);

	retval = 0;
	list_for_each(list, &omux_dev_list) {
		q = list_entry(list, struct omux_device, list);
		if (!strncmp(drv->name, q->name, strlen(q->name))) {
			OMUX_BUG_ON(q->drv);   /* the device is busy! */
			drv->dev = q;
			q->drv = drv;
			break;
		}
	}
reg_drv_exist:
	OMUX_BUG_ON(0 != retval);

	DLEAVE("ret=%d\n", retval);
	return retval;
}

int omux_unregister_driver(struct omux_driver *drv)
{
	struct list_head *list;
	struct omux_driver *p;
	int retval = -ENXIO;

	DENTER("drv=%p\n", drv);

	BUG_ON(!drv || !drv->name);

	list_for_each(list, &omux_drv_list) {
		p = list_entry(list, struct omux_driver, list);
		if (!strncmp(drv->name, p->name, strlen(p->name))) {
			retval = 0;
			break;
		}
	}

	if (0 != retval)
		goto unreg_drv_nodrv;

	if (drv->dev) {
		OMUX_BUG_ON(drv->dev->drv != drv);
		drv->dev->drv = NULL;
		drv->dev = NULL;
	}
	list_del(&drv->list);

unreg_drv_nodrv:
	OMUX_BUG_ON(0 != retval);

	DLEAVE("ret=%d\n", retval);
	return retval;
}

/*****************************************************************************/
/*                                                                           */
/*                     OMUX Device Interface                                 */
/*                                                                           */
/*****************************************************************************/
int omux_register_device(struct omux_device *dev)
{
	struct list_head *list;
	struct omux_driver *p;
	struct omux_device *q;
	int retval = -ENXIO;

	DENTER("dev=%p\n", dev);

	BUG_ON(!dev || dev->drv || !dev->name);

	list_for_each(list, &omux_dev_list) {
		q = list_entry(list, struct omux_device, list);
		if (!strncmp(dev->name, q->name, strlen(q->name))) {
			retval = -EEXIST;
			OMUX_BUG_ON(1);
			goto reg_dev_exist; 
		}
	}

	list_add_tail(&(dev->list), &omux_dev_list);

	retval = 0;
	list_for_each(list, &omux_drv_list) {
		p = list_entry(list, struct omux_driver, list);

		if (!strncmp(dev->name, p->name, strlen(p->name))) {
			OMUX_BUG_ON(p->dev);	/* this driver is busy! */
			dev->drv = p;
			p->dev = dev;
			break;
		}
	}
reg_dev_exist:
	OMUX_BUG_ON(0 != retval);

	DLEAVE("ret=%d\n", retval);
	return retval;
}

int omux_unregister_device(struct omux_device *dev)
{
	struct list_head *list;
	struct omux_device *q;
	int retval = -ENXIO;

	DENTER("dev=%p\n", dev);

	BUG_ON(!dev || !dev->name);

	list_for_each(list, &omux_dev_list) {
		q = list_entry(list, struct omux_device, list);
		if (!strncmp(dev->name, q->name, strlen(q->name))) {
			retval = 0;
			break;
		}
	}
	
	if (0 != retval)
		goto unreg_dev_nodev;

	if (dev->drv) {
		OMUX_BUG_ON(dev->drv->dev != dev);
		dev->drv->dev = NULL;
		dev->drv = NULL;
	}
	list_del(&dev->list);

unreg_dev_nodev:
	OMUX_BUG_ON(0 != retval);

	DLEAVE("ret=%d\n", retval);
	return retval;
}

/*****************************************************************************/
/*                                                                           */
/*                           Sysfs Interface                                 */
/*                                                                           */
/*****************************************************************************/
static ssize_t core_dbg_mask_show(struct kobject *kobj,
				  struct kobj_attribute *attr, char *buf)
{
	char *s = buf;

	s += sprintf(s, "debug_mask=0x%08x\n", g_omux_core_debug_mask);
	s += sprintf(s,
		     "Help: [DEBUG_LOG | DEBUG_DATA | DEBUG_FUNC]\n");

	return (s - buf);
}

static ssize_t core_dbg_mask_store(struct kobject *kobj,
				   struct kobj_attribute *attr, const char *buf,
				   size_t n)
{
	sscanf(buf, "0x%x", &g_omux_core_debug_mask);
	return n;
}

OMUX_ATTR(core_dbg_mask);

static ssize_t core_dump_data_show(struct kobject *kobj,
				   struct kobj_attribute *attr, char *buf)
{
	char *s = buf;

	s += sprintf(s, "debug_mask=0x%08x\n", g_omux_core_dump_data);
	s += sprintf(s,
		     "Help: [DEBUG_LOG | DEBUG_DATA_HEAD | DEBUG_DATA | DEBUG_FUNC]\n");

	return (s - buf);
}

static ssize_t core_dump_data_store(struct kobject *kobj,
				    struct kobj_attribute *attr,
				    const char *buf, size_t n)
{
	sscanf(buf, "0x%x", &g_omux_core_dump_data);
	return n;
}

OMUX_ATTR(core_dump_data);

static struct attribute *omux_core_attrs[] = {
	&core_dbg_mask_attr.attr,
	&core_dump_data_attr.attr,
	NULL,
};

static struct attribute_group attr_group = {
	.attrs = omux_core_attrs,
};

int omux_sysfs_create_group(struct attribute_group *g)
{
	int retval = 0;
	if (!omux_kobj) {
		omux_kobj = kobject_create_and_add("omux", NULL);
		if (omux_kobj == NULL) {
			printk(KERN_ERR "subsystem_register failed\n");
			goto err_kobj;
		}
	}

	retval = sysfs_create_group(omux_kobj, g);
	if (retval) {
		printk(KERN_ERR "subsystem_register attr failed\n");
		goto err_kgrp;
	}
	++omux_kobj_refcount;

err_kobj:
	if (0 == omux_kobj_refcount) {
		kobject_del(omux_kobj);
		omux_kobj = NULL;
	}
	return retval;
err_kgrp:
	goto err_kobj;
}

void omux_sysfs_remove_group(struct attribute_group *g)
{
	BUG_ON(!omux_kobj);

	sysfs_remove_group(omux_kobj, g);
	--omux_kobj_refcount;

	if (0 == omux_kobj_refcount) {
		kobject_del(omux_kobj);
		omux_kobj = NULL;
	}
}

/*****************************************************************************/
/*                                                                           */
/*                     OMUX Init & Fini                                      */
/*                                                                           */
/*****************************************************************************/
static int __init omux_init(void)
{
	int retval;

	DENTER("\n");

	INIT_LIST_HEAD(&omux_drv_list);
	INIT_LIST_HEAD(&omux_dev_list);

	retval = omux_sysfs_create_group(&attr_group);
	if (retval) {
		printk(KERN_ERR "[OMUX]: Cannot create sysfs interfaces\n");
		goto err_sysg;
	}

ret_init:
	DLEAVE("ret=%d\n", retval);
	return retval;
err_sysg:
	OMUX_BUG_ON(1);
	goto ret_init;
}

static void __exit omux_exit(void)
{
	DENTER("\n");

	omux_sysfs_remove_group(&attr_group);

	DLEAVE("\n");
}

module_init(omux_init);
module_exit(omux_exit);

EXPORT_SYMBOL_GPL(omux_register_device);
EXPORT_SYMBOL_GPL(omux_unregister_device);
EXPORT_SYMBOL_GPL(omux_register_driver);
EXPORT_SYMBOL_GPL(omux_unregister_driver);
EXPORT_SYMBOL_GPL(omux_sysfs_remove_group);
EXPORT_SYMBOL_GPL(omux_sysfs_create_group);
EXPORT_SYMBOL_GPL(omux_tx_buf);
EXPORT_SYMBOL_GPL(omux_rx_buf_handler);
EXPORT_SYMBOL_GPL(omux_dump_data);

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Emichael Li<emichael.li@borqs.com>");
MODULE_DESCRIPTION("OMUX Core");
