/* 
 * drivers/omux/ipc_ldisc.c 
 *
 * Open MUX's ldisc IPC for TTY devices.
 *
 * Copyright (C) 2009 Borqs, Inc.
 * Author: Emichael Li <emichael.li@borqs.com>
 *
 * This software is licensed under the terms of the GNU General Public
 * License version 2, as published by the Free Software Foundation, and
 * may be copied, distributed, and modified under those terms.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

#include <linux/module.h>
#include <linux/moduleparam.h>
#include <asm/system.h>
#include <asm/uaccess.h>
#include <linux/bitops.h>
#include <linux/string.h>
#include <linux/mm.h>
#include <linux/interrupt.h>
#include <linux/in.h>
#include <linux/tty.h>
#include <linux/errno.h>
#include <linux/delay.h>
#include <linux/init.h>
#include <linux/sysdev.h>
#include <linux/omux.h>

#define OMUX_DEBUG_MASK 	g_omux_ldisc_debug_mask
#define OMUX_DEBUG_LOG          OMUX_LDISC_DEBUG_LOG
#define OMUX_DEBUG_FUNC         OMUX_LDISC_DEBUG_FUNC
#define OMUX_DEBUG_DATA         OMUX_LDISC_DEBUG_DATA
#define OMUX_DEBUG_DATA_HEAD    OMUX_LDISC_DEBUG_DATA_HEAD

#include "../core/omux_debug.h"

static void omux_ldisc_tx_worker(struct work_struct *work);

/*****************************************************************************/
/*                                                                           */
/*                                Debug variables                            */
/*                                                                           */
/*****************************************************************************/
enum {
	OMUX_LDISC_DEBUG_FUNC = 1U << 0,
	OMUX_LDISC_DEBUG_DATA = 1U << 1,
	OMUX_LDISC_DEBUG_LOG  = 1U << 2,
};

unsigned long g_omux_ldisc_debug_mask =
    OMUX_LDISC_DEBUG_FUNC | OMUX_LDISC_DEBUG_DATA | OMUX_LDISC_DEBUG_LOG;

int g_omux_ldisc_dump_data = 16;
/*****************************************************************************/
/*                                                                           */
/*                             omux ldiscter (ldisc)                         */
/*                                                                           */
/*****************************************************************************/
#define N_OMUX (5)
#define OMUX_LDISC_MAGIC (0x0505)

struct omux_ldisc {
	int magic;
	struct tty_struct *tty;
	struct list_head tx_list; 
	struct workqueue_struct *tx_workqueue;
	struct work_struct tx_work;
};

static struct omux_ldisc omux_ldisc_info;	/* only can be used once */

#if 0
/*****************************************************************************/
/*                                                                           */
/*                             omux ldisc functions                          */
/*                                                                           */
/*****************************************************************************/
static int omux_ldisc_open(struct tty_struct *tty)
{
	struct omux_ldisc *oml;
	int retval = 0;
	DENTER("tty=%p\n", tty);

	BUG_ON(!tty);

	oml = tty->disc_data;

	BUG_ON(oml && oml->magic == OMUX_LDISC_MAGIC);

	omux_ldisc_info.tx_workqueue = create_singlethread_workqueue("ldisc_tx");
	BUG_ON(!omux_ldisc_info.tx_workqueue);

	INIT_WORK(&omux_ldisc_info.tx_work, omux_ldisc_tx_worker);

	INIT_LIST_HEAD(&(omux_ldisc_info.tx_list)),

	oml = &omux_ldisc_info;

	oml->tty = tty;
	tty->disc_data = oml;

	DLEAVE("ret=%d\n", retval);
	return retval;
}

static void omux_ldisc_close(struct tty_struct *tty)
{
	struct omux_ldisc *oml;

	DENTER("tty=%p\n", tty);
	BUG_ON(!tty);

	oml = tty->disc_data;
	BUG_ON(!oml || oml->magic != OMUX_LDISC_MAGIC);

	destroy_workqueue(omux_ldisc_info.tx_workqueue);

	tty->disc_data = NULL;
	oml->tty = NULL;

	DLEAVE("\n");
}

static int omux_ldisc_ioctl(struct tty_struct *tty, struct file *file,
			    unsigned int cmd, unsigned long arg)
{
	DENTER("tty=%p cmd=0x%x arg=0x%lx\n", tty, cmd, arg);
	DLEAVE("\n");
	return 0;
}

static void omux_ldisc_receive_buf(struct tty_struct *tty,
				   const unsigned char *buf, char *cflags,
				   int count)
{
	int retval;
	int i, tmpcnt = count;
	struct omux_rb *mrb;

	DENTER("tty=%p buf=%p cflag=%p count=%d\n", tty, buf, cflags, count);

	if (unlikely(g_omux_ldisc_debug_mask & OMUX_LDISC_DEBUG_DATA)) {
		if (tmpcnt > g_omux_ldisc_dump_data)
			tmpcnt = g_omux_ldisc_dump_data;
	 
		printk("LR<: (%4d) (%2d)\n", count, tmpcnt);
		omux_dump_data(buf, tmpcnt);
	}

	mrb = omux_alloc_mrb(count);
	if (!mrb) {
		printk(KERN_ERR "Cannot alloc mrb!\n");
		OMUX_BUG_ON(1);
		return;
	}

	for (i = 0; i < count; ++i) {
		if (cflags && *cflags++) {
			printk(KERN_ERR "Data error!\n");
			BUG();
		}
	}

	memcpy(mrb->data, buf, count);
	mrb->tail += count;

	retval = omux_rx_mrb(mrb);

	omux_free_mrb(mrb);	// EL: to be improved!

	DLEAVE("\n");
}

static void omux_ldisc_write_wakeup(struct tty_struct *tty)
{
	DENTER("tty=%p\n", tty);
	DLEAVE("\n");
}

static struct tty_ldisc omux_ldisc_ops = {
	.owner = THIS_MODULE,
	.magic = TTY_LDISC_MAGIC,
	.name = "omux",
	.open = omux_ldisc_open,
	.close = omux_ldisc_close,
	.ioctl = omux_ldisc_ioctl,
	.receive_buf = omux_ldisc_receive_buf,
	.write_wakeup = omux_ldisc_write_wakeup,
};

/*****************************************************************************/
/*                                                                           */
/*                                   omux driver                             */
/*                                                                           */
/*****************************************************************************/
static void omux_ldisc_tx_worker(struct work_struct *work)
{
	struct omux_rb *mrb;
	int count, actual, tmpcnt;
	int retval = 0;
	unsigned long flags;

	DENTER("work=%p\n", work);

	local_irq_save(flags);
	while (!list_empty(&(omux_ldisc_info.tx_list))) {
		mrb = list_first_entry(&(omux_ldisc_info.tx_list), 
				struct omux_rb, list);
	
		BUG_ON(!mrb);
		list_del_init(&(mrb->list));
		local_irq_restore(flags);

		count = mrb->tail - mrb->data;
		BUG_ON(count <= 0);

		if (unlikely(g_omux_ldisc_debug_mask & OMUX_LDISC_DEBUG_DATA)) {
			if (tmpcnt > g_omux_ldisc_dump_data)
				tmpcnt = g_omux_ldisc_dump_data;

			printk("LT>: (%4d) (%2d)\n", count, tmpcnt);
			omux_dump_data(mrb->data, count);
		}

		actual = omux_ldisc_info.tty->driver->write(omux_ldisc_info.tty,
				mrb->data, count);

		if (actual < count) {
			mrb->data += actual;
			printk("TTY Buffer is full %d\n", count - actual);

			local_irq_save(flags);
			list_add(&(mrb->list), &(omux_ldisc_info.tx_list));
			local_irq_restore(flags);
		} else {
			BUG_ON(actual > count);
			mrb->status = 0;

			if (mrb->complete)
				mrb->complete(mrb);

			omux_free_mrb(mrb);     // EL: to be improved!
		}
	}
	DLEAVE("ret=%d\n", retval);
}

ssize_t omux_ldisc_xmit(struct omux_rb *mrb)
{
	unsigned long flags;
	int count;
	DENTER("mrb=%p\n", mrb);

	BUG_ON(!mrb);

	count = mrb->tail - mrb->data;
	local_irq_save(flags);
	list_add_tail(&(mrb->list), &(omux_ldisc_info.tx_list));
	local_irq_restore(flags);

	queue_work(omux_ldisc_info.tx_workqueue, &(omux_ldisc_info.tx_work));	

	DLEAVE("ret=%d\n", count);
	return count;		// to be inproved!
}

ssize_t omux_ldisc_chars_in_xmit_buffer(void)
{
	int retval = 0;
	DENTER("\n");

	retval =
	    omux_ldisc_info.tty->driver->chars_in_buffer(omux_ldisc_info.tty);

	DLEAVE("ret=%d\n", retval);
	return retval;
}

static struct omux_device omux_ttydev = {
	.owner = THIS_MODULE,
	.omux_tx = omux_ldisc_xmit,
	.chars_in_tx_buffer = omux_ldisc_chars_in_xmit_buffer,
};

/*****************************************************************************/
/*                                                                           */
/*                  sysfs interfaces of omux ldisc                           */
/*                                                                           */
/*****************************************************************************/
static ssize_t ldisc_dbg_mask_show(struct kobject *kobj,
				   struct kobj_attribute *attr, char *buf)
{
	char *s = buf;
	s += sprintf(s, "debug_mask=0x%08x\n", g_omux_ldisc_debug_mask);
	s += sprintf(s,
		     "Help: [DEBUG_LOG | DEBUG_DATA_HEAD | DEBUG_DATA | DEBUG_FUNC]\n");

	return (s - buf);
}

static ssize_t ldisc_dbg_mask_store(struct kobject *kobj,
				    struct kobj_attribute *attr,
				    const char *buf, size_t n)
{
	sscanf(buf, "0x%x", &g_omux_ldisc_debug_mask);
	return n;
}

OMUX_ATTR(ldisc_dbg_mask);

static struct attribute *omux_ldisc_attrs[] = {
	&ldisc_dbg_mask_attr.attr,
	NULL,
};

static struct attribute_group attr_group = {
	.attrs = omux_ldisc_attrs,
};
#endif
/*****************************************************************************/
/*                                                                           */
/*                         omux ldisc init & fini                            */
/*                                                                           */
/*****************************************************************************/
static int __init omux_ldisc_init(void)
{
	int retval;
	DENTER("\n");
#if 0
	retval = omux_register_device(&omux_ttydev);
	if (retval != 0) {
		printk(KERN_ERR "Failed to unregister omuxdev(%d).\n", retval);
		goto err_odrv;
	}

	retval = tty_register_ldisc(N_OMUX, &omux_ldisc_ops);
	if (retval != 0) {
		printk(KERN_ERR "Failed to register omux ldisc(%d).\n", retval);
		goto err_ldisc;
	}

	retval = omux_sysfs_create_group(&attr_group);
	if (retval) {
		printk(KERN_ERR "omux_init: subsystem_register attr failed\n");
		goto err_sysg;
	}

#endif
      ret_init:
	DLEAVE("ret=%d\n", retval);
	return retval;
#if 0
      err_sysg:
	tty_unregister_ldisc(N_OMUX);
      err_ldisc:
	omux_unregister_device(&omux_ttydev);
      err_odrv:
	OMUX_BUG_ON(1);
	goto ret_init;
#endif
}

static void __exit omux_ldisc_exit(void)
{
	int retval;
	DENTER("\n");
#if 0
	omux_sysfs_remove_group(&attr_group);

	retval = tty_unregister_ldisc(N_OMUX);
	if (retval != 0)
		printk(KERN_ERR "Failed to uninit omux ldisc(%d).\n", retval);

	retval = omux_unregister_device(&omux_ttydev);
	if (retval != 0)
		printk(KERN_ERR "Failed to unregister omuxdev(%d).\n", retval);

	DLEAVE("ret=%d\n", retval);
#endif
}

module_init(omux_ldisc_init);
module_exit(omux_ldisc_exit);

MODULE_LICENSE("GPLv2");
MODULE_AUTHOR("Emichael Li<emichael.li@borqs.com>");
MODULE_DESCRIPTION("OMUX Ldisc");
