/* 
 * drivers/omux/ipc/ipc_skeleton.c 
 *
 * the skeleton driver of Open MUX's IPC modules (loopback)
 *
 * Copyright (C) 2010 Borqs, Inc.
 * Author: Emichael Li <emichael.li@borqs.com>
 *
 * This software is licensed under the terms of the GNU General Public
 * License version 2, as published by the Free Software Foundation, and
 * may be copied, distributed, and modified under those terms.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

#include <linux/module.h>
#include <linux/moduleparam.h>
#include <asm/system.h>
#include <asm/uaccess.h>
#include <linux/bitops.h>
#include <linux/string.h>
#include <linux/mm.h>
#include <linux/errno.h>
#include <linux/delay.h>
#include <linux/sysdev.h>
#include <linux/omux.h>

#define OMUX_DEBUG_MASK 	g_omux_skel_debug_mask
#define OMUX_DEBUG_LOG          OMUX_SKEL_DEBUG_LOG
#define OMUX_DEBUG_FUNC         OMUX_SKEL_DEBUG_FUNC
#define OMUX_DEBUG_DATA         OMUX_SKEL_DEBUG_DATA

#include "../core/omux_debug.h"

/*****************************************************************************/
/*                                                                           */
/*                                Debug variables                            */
/*                                                                           */
/*****************************************************************************/
enum {
	OMUX_SKEL_DEBUG_FUNC = 1U << 0,
	OMUX_SKEL_DEBUG_DATA = 1U << 1,
	OMUX_SKEL_DEBUG_LOG = 1U << 2,
};

unsigned long g_omux_skel_debug_mask =
    OMUX_SKEL_DEBUG_FUNC | OMUX_SKEL_DEBUG_DATA | OMUX_SKEL_DEBUG_LOG;

int g_omux_skel_dump_data = 16;

/*****************************************************************************/
/*                                                                           */
/*                                   omux device                             */
/*                                                                           */
/*****************************************************************************/
static int ipc_skel_rx_buf_handler(unsigned char *buf, ssize_t len);

ssize_t ipc_skel_tx_buf(unsigned char *buf, ssize_t len)
{
	int count, tmpcnt;

	tmpcnt = len;
	if (unlikely(g_omux_skel_debug_mask & OMUX_SKEL_DEBUG_DATA)) {
		if (tmpcnt > g_omux_skel_dump_data)
			tmpcnt = g_omux_skel_dump_data;

		printk("ST>: (%4d) (%2d)\n", len, tmpcnt);
		omux_dump_data(buf, tmpcnt);
	}

	count = ipc_skel_rx_buf_handler(buf, len);

	return count;
}

static struct omux_device omux_skel_dev = {
	.owner = THIS_MODULE,
	.name = "muxraw",
	.tx_buf = ipc_skel_tx_buf,
};

static int ipc_skel_rx_buf_handler(unsigned char *buf, ssize_t len)
{
	return omux_rx_buf_handler(&omux_skel_dev, buf, len);
}

/*****************************************************************************/
/*                                                                           */
/*                   sysfs interfaces of omux skel                           */
/*                                                                           */
/*****************************************************************************/
static ssize_t skel_dbg_mask_show(struct kobject *kobj,
				  struct kobj_attribute *attr, char *buf)
{
	char *s = buf;
	s += sprintf(s, "debug_mask=0x%08lx\n", g_omux_skel_debug_mask);
	s += sprintf(s, "Help: [DEBUG_LOG | DEBUG_DATA | DEBUG_FUNC]\n");

	return (s - buf);
}

static ssize_t skel_dbg_mask_store(struct kobject *kobj,
				   struct kobj_attribute *attr,
				   const char *buf, size_t n)
{
	sscanf(buf, "0x%lx", &g_omux_skel_debug_mask);
	return n;
}

OMUX_ATTR(skel_dbg_mask);

static struct attribute *omux_skel_attrs[] = {
	&skel_dbg_mask_attr.attr,
	NULL,
};

static struct attribute_group attr_group = {
	.attrs = omux_skel_attrs,
};

/*****************************************************************************/
/*                                                                           */
/*                          omux skel init & fini                            */
/*                                                                           */
/*****************************************************************************/
static int __init omux_ipc_skel_init(void)
{
	int retval;
	DENTER("\n");

	retval = omux_register_device(&omux_skel_dev);
	if (retval != 0) {
		printk(KERN_ERR "Failed to unregister omuxdev(%d).\n", retval);
		goto err_odrv;
	}

	retval = omux_sysfs_create_group(&attr_group);
	if (retval) {
		printk(KERN_ERR "omux_init: subsystem_register attr failed\n");
		goto err_sysg;
	}

ret_init:
	DLEAVE("ret=%d\n", retval);
	return retval;
err_sysg:
	omux_unregister_device(&omux_skel_dev);
err_odrv:
	OMUX_BUG_ON(1);
	goto ret_init;
}

static void __exit omux_ipc_skel_exit(void)
{
	int retval;
	DENTER("\n");

	omux_sysfs_remove_group(&attr_group);

	retval = omux_unregister_device(&omux_skel_dev);
	OMUX_BUG_ON(retval);

	DLEAVE("ret=%d\n", retval);
}

module_init(omux_ipc_skel_init);
module_exit(omux_ipc_skel_exit);

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Emichael Li<emichael.li@borqs.com>");
MODULE_DESCRIPTION("OMUX IPC SKEL");
