/* 
 * drivers/omux/mux/raw.c 
 *
 * mux raw interface.
 *
 * Copyright (C) 2009~2010 Borqs, Inc.
 * Author: Emichael Li <emichael.li@borqs.com>
 *
 * This software is licensed under the terms of the GNU General Public
 * License version 2, as published by the Free Software Foundation, and
 * may be copied, distributed, and modified under those terms.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

#include <linux/module.h>
#include <linux/platform_device.h>
#include <linux/interrupt.h>
#include <linux/tty.h>
#include <linux/tty_flip.h>
#include <linux/errno.h>
#include <linux/omux.h>

#define OMUX_DEBUG_MASK         g_omux_raw_debug_mask
#define OMUX_DEBUG_LOG          OMUX_RAW_DEBUG_LOG
#define OMUX_DEBUG_FUNC         OMUX_RAW_DEBUG_FUNC
#define OMUX_DEBUG_DATA         OMUX_RAW_DEBUG_DATA

#include "../core/omux_debug.h"

#define MUX_RAW_TTY_MAJOR 	(250)
#define MUX_RAW_TTY_MINOR_START (64)
#define MUX_RAW_TTY_MINOR_NRS 	(1)

/*****************************************************************************/
/*                                                                           */
/*                                Debug variables                            */
/*                                                                           */
/*****************************************************************************/
enum {
	OMUX_RAW_DEBUG_FUNC = 1U << 0,
	OMUX_RAW_DEBUG_DATA = 1U << 1,
	OMUX_RAW_DEBUG_LOG = 1U << 2,
};

unsigned long g_omux_raw_debug_mask =
    OMUX_RAW_DEBUG_FUNC | OMUX_RAW_DEBUG_DATA | OMUX_RAW_DEBUG_LOG;

int g_omux_raw_dump_data = 16;

/*****************************************************************************/
/*                                                                           */
/*                             ts0710 mux tty opssc)                         */
/*                                                                           */
/*****************************************************************************/
struct raw_tty {
	spinlock_t lock;
	int opencount;

	struct tty_struct *tty;
};

static struct raw_tty *raw_ttys;
static DEFINE_MUTEX(raw_tty_lock);

/*****************************************************************************/
/*                                                                           */
/*                             omux driver                                   */
/*                                                                           */
/*****************************************************************************/
int raw_rx_buf_handler(unsigned char *buf, ssize_t len)
{
	int retval;
	int i;
	int line;
	struct raw_tty *rtty;

	DENTER("buf=%p, len=%d\n", buf, len);
	BUG_ON(!buf || len <= 0);

	omux_dump_data(buf, len);

	for (i = 0; i < len; ++i) {
		printk("%c ", buf[i]);
	}
	printk("\n");

	line = 0;		// only for the first line
	BUG_ON(!raw_ttys);
	rtty = &raw_ttys[line];
	BUG_ON(!rtty);

	if (!rtty->tty) {
		printk("ttyRAW0 is not opened\n");
		return 0;
	}

	retval = tty_insert_flip_string(rtty->tty, buf, len);
	BUG_ON(retval < len);

	tty_flip_buffer_push(rtty->tty);

	return retval;
}

struct omux_driver omux_raw_driver = {
	.name = "muxraw",
	.rx_buf_handler = raw_rx_buf_handler,
};

/*****************************************************************************/
/*                                                                           */
/*                             mux raw tty driver                            */
/*                                                                           */
/*****************************************************************************/
static int raw_open(struct tty_struct *tty, struct file *filp)
{
	int retval;
	int line;
	struct raw_tty *rtty;

	DENTER("tty=%p filp=%p\n", tty, filp);
	BUG_ON(!tty);

	line = tty->index;
	BUG_ON((line < 0) || (line >= MUX_RAW_TTY_MINOR_NRS));

	BUG_ON(!raw_ttys);
	rtty = &raw_ttys[line];

	mutex_lock(&raw_tty_lock);
	if (rtty->tty == NULL || rtty->tty == tty) {
		if (rtty->opencount++ == 0) {
			rtty->tty = tty;
		}
		retval = 0;
	} else
		retval = -EBUSY;

	mutex_unlock(&raw_tty_lock);

	DLEAVE("ret = %d\n", retval);
	return retval;
}

static void raw_close(struct tty_struct *tty, struct file *filp)
{
	int line;
	struct raw_tty *rtty;

	DENTER("tty=%p filp=%p\n", tty, filp);
	BUG_ON(!tty);

	line = tty->index;
	BUG_ON((line < 0) || (line >= MUX_RAW_TTY_MINOR_NRS));

	BUG_ON(!raw_ttys);
	rtty = &raw_ttys[line];
	BUG_ON(!rtty);

	mutex_lock(&raw_tty_lock);

	if (rtty->tty == tty) {
		if (--rtty->opencount == 0) {
			rtty->tty = NULL;
		}
	}
	mutex_unlock(&raw_tty_lock);

	DLEAVE("\n");
}

static int raw_write(struct tty_struct *tty, const unsigned char *buf,
		     int count)
{
	int line;
	struct raw_tty *rtty;
	int retcnt;

	DENTER("tty=%p buf=%p count=%d\n", tty, buf, count);
	BUG_ON(!tty || !buf || count <= 0);

	line = tty->index;
	BUG_ON((line < 0) || (line >= MUX_RAW_TTY_MINOR_NRS));

	BUG_ON(!raw_ttys);
	rtty = &raw_ttys[line];
	BUG_ON(!rtty);

	retcnt = omux_tx_buf(&omux_raw_driver, (unsigned char*)buf, count);

	// send data here
	DLEAVE("retcnt=%d\n", retcnt);
	return retcnt;
}

static int raw_write_room(struct tty_struct *tty)
{
	int retval;
	int line;
	struct raw_tty *rtty;
	unsigned long flags;

	BUG_ON(!tty);

	retval = 0;

	line = tty->index;
	BUG_ON((line < 0) || (line >= MUX_RAW_TTY_MINOR_NRS));

	BUG_ON(!raw_ttys);
	rtty = &raw_ttys[line];
	BUG_ON(!rtty);

	spin_lock_irqsave(&rtty->lock, flags);
	BUG_ON(1);
	spin_unlock_irqrestore(&rtty->lock, flags);

	return retval;
}

static void raw_flush_buffer(struct tty_struct *tty)
{
	int line;

	BUG_ON(!tty);

	line = tty->index;
	BUG_ON((line < 0) || (line >= MUX_RAW_TTY_MINOR_NRS));

	tty_wakeup(tty);
}

static int raw_chars_in_buffer(struct tty_struct *tty)
{
	int retval;
	int line;
	struct raw_tty *rtty;
	unsigned long flags;

	BUG_ON(!tty);

	retval = 0x10000;

	line = tty->index;
	BUG_ON((line < 0) || (line >= MUX_RAW_TTY_MINOR_NRS));

	BUG_ON(!raw_ttys);
	rtty = &raw_ttys[line];
	BUG_ON(!rtty);

	spin_lock_irqsave(&rtty->lock, flags);

	spin_unlock_irqrestore(&rtty->lock, flags);

	return retval;
}

static const struct tty_operations raw_tty_ops = {
	.open = raw_open,
	.close = raw_close,
	.write = raw_write,
	.write_room = raw_write_room,
	.flush_buffer = raw_flush_buffer,
	.chars_in_buffer = raw_chars_in_buffer,
};

/*****************************************************************************/
/*                                                                           */
/*                             ts0710 init & fini                            */
/*                                                                           */
/*****************************************************************************/
static int __init mux_raw_init(void)
{
	struct tty_driver *tty;
	int ret;

	ret = 0;
	raw_ttys =
	    kzalloc(sizeof(*raw_ttys) * MUX_RAW_TTY_MINOR_NRS, GFP_KERNEL);
	if (raw_ttys == NULL) {
		ret = -ENOMEM;
		goto err_alloc_ttys;
	}

	tty = alloc_tty_driver(MUX_RAW_TTY_MINOR_NRS);
	if (tty == NULL) {
		ret = -ENOMEM;
		goto err_alloc_tty_driver;
	}

	tty->magic = TTY_DRIVER_MAGIC;
	tty->driver_name = "ttyRAW";
	tty->name = "ttyRAW";
	tty->major = MUX_RAW_TTY_MAJOR;
	tty->minor_start = MUX_RAW_TTY_MINOR_START;
	tty->num = MUX_RAW_TTY_MINOR_NRS;
	tty->type = TTY_DRIVER_TYPE_SERIAL;
	tty->subtype = SERIAL_TYPE_NORMAL;
	tty->init_termios = tty_std_termios;
	tty->init_termios.c_iflag = 0;
	tty->init_termios.c_oflag = 0;
	tty->init_termios.c_cflag = B38400 | CS8 | CREAD;
	tty->init_termios.c_lflag = 0;
	tty->flags = TTY_DRIVER_RESET_TERMIOS | TTY_DRIVER_REAL_RAW;
	tty->other = NULL;
	tty->owner = THIS_MODULE;

	tty_set_operations(tty, &raw_tty_ops);

	if (tty_register_driver(tty)) {
		printk(KERN_ERR "tty register error\n");
		goto err_reg_tty_driver;
	}

	if (omux_register_driver(&omux_raw_driver)) {
		printk(KERN_ERR "omux register error\n");
		goto err_reg_omux_driver;
	}
	return ret;
err_reg_omux_driver:

err_reg_tty_driver:

err_alloc_tty_driver:
	kfree(raw_ttys);
err_alloc_ttys:
	BUG();
	return ret;
}

static void __exit mux_raw_exit(void)
{
}

module_init(mux_raw_init);
module_exit(mux_raw_exit);

MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("OMUX Raw TTY Driver");
