/*
 * drivers/power/bq2407x_charger.c
 *
 * Copyright @ 2009 Borqs Ltd.
 * 
 * Author: Ru Yi<yi.ru@borqs.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 */

#include <linux/module.h>
#include <linux/platform_device.h>
#include <linux/interrupt.h>
#include <linux/power_supply.h>
#include <linux/timer.h>
#include <linux/jiffies.h>
#include <linux/delay.h>
#include <linux/wakelock.h>
#include <linux/bq2407x_charger.h>

#include <mach/gpio.h>
#include <mach/mfp.h>
#include <mach/pxa-regs.h>

#define	POWER_SUPPLY_BATTERY_NAME	"battery"

struct bq2407x_charger_info {
	struct device *dev;
	struct bq2407x_charger_pdata *pdata;

	/* charger status */
	struct power_supply *charger_psy;

	int charger_status;
	int charger_old_status;
	int charger_type;
	int charger_old_type;
	int charge_current;
	int charge_old_current;
	int charge_enable; 
	int charge_old_enable; 

	/* battery info */
	struct power_supply *battery_psy;
	int battery_voltage;
	int battery_capacity;
	int battery_temperature;
	int battery_status;

#ifdef CONFIG_USB_CAN_SLEEP_TEST
	int charging_sleep;
#endif
	struct delayed_work charge_work;
	struct delayed_work monitor_work;
	struct workqueue_struct *charge_wqueue;
	struct wake_lock wake_lock;
};

struct bq2407x_charger_info *charger_info = NULL;

/* Update Battery Information */
static void update_battery_info(void)
{
	union power_supply_propval ret = {0,};
	
	if (!charger_info)
		return;

	if (!charger_info->battery_psy) {
		charger_info->battery_psy = power_supply_get_by_name(
					charger_info->pdata->supplied_to[0]);
		if (!charger_info->battery_psy) {
			TDBG("no battery found\n"); 
			return;
		}
	}

	/* update voltage */
	charger_info->battery_psy->get_property(charger_info->battery_psy, 
			POWER_SUPPLY_PROP_BATT_VOL, &ret);
	charger_info->battery_voltage = ret.intval;

	/* update capacity */
	charger_info->battery_psy->get_property(charger_info->battery_psy, 
			POWER_SUPPLY_PROP_CAPACITY, &ret);
	charger_info->battery_capacity = ret.intval;

	/* update temperature */
	charger_info->battery_psy->get_property(charger_info->battery_psy, 
			POWER_SUPPLY_PROP_BATT_TEMP, &ret);
	charger_info->battery_temperature = ret.intval;

	/* update status */
	charger_info->battery_psy->get_property(charger_info->battery_psy, 
			POWER_SUPPLY_PROP_STATUS, &ret);
	charger_info->battery_status = ret.intval;
}

/* Set charge ic charging current */
static void bq2407x_set_current(int current_type)
{
	if (!charger_info)
		return;

	/* check current current_type */
	if (CHARGING_SUSPEND > current_type 
	    || CHARGING_CURRENT_FULL < current_type) {
		TDBG("set current failed %d\n", current_type); 
		return;
	}
	
	if (charger_info->charge_current != current_type) {
		charger_info->charge_current = current_type;
		TDBG(KERN_DEBUG "set current type %d\n", charger_info->charge_current); 
	}
		
	return;
}

void usb_callback_set_current(unsigned int icurrent)
{
	TDBG(KERN_DEBUG "usb set current %u\n", icurrent);
	if(icurrent > 1500)
		icurrent = 1500;
	if(icurrent < 40)
		icurrent = 40;

	if(icurrent < 100)
		bq2407x_set_current(CHARGING_SUSPEND);
	else if(icurrent < 500)
		bq2407x_set_current(CHARGING_CURRENT_100);
	else if(icurrent < 1000)
		bq2407x_set_current(CHARGING_CURRENT_500);
	else
		bq2407x_set_current(CHARGING_CURRENT_FULL);
	return;
}
EXPORT_SYMBOL(usb_callback_set_current);

static int bq2407x_charger_get_property(struct power_supply *psy,
				  enum power_supply_property psp,
				  union power_supply_propval *val)
{
	int power_dev = 0;
	val->intval = 0;

	if (psy->type == POWER_SUPPLY_TYPE_MAINS)
		power_dev = BQ2407x_POWER_CHARGE_AC;
	else if (psy->type == POWER_SUPPLY_TYPE_USB)
		power_dev = BQ2407x_POWER_CHARGE_USB;

	switch (psp) {
	case POWER_SUPPLY_PROP_ONLINE:
		if (power_dev == charger_info->charger_type) 
			val->intval = 1;
		break;
	case POWER_SUPPLY_PROP_STATUS:
		if (power_dev == charger_info->charger_type)
			val->intval = charger_info->charger_status;
		else
			val->intval = POWER_SUPPLY_STATUS_NOT_CHARGING;
		break;
	case POWER_SUPPLY_PROP_CHARGE_ENABLE:
		if (power_dev == charger_info->charger_type 
		    && charger_info->charge_current != CHARGING_SUSPEND
		    && charger_info->charge_enable == BQ2407x_POWER_ENABLE)
			val->intval = 1;
		else
			val->intval = 0;
		break;
	case POWER_SUPPLY_PROP_CURRENT_NOW:
		if (power_dev == charger_info->charger_type) {
			switch (charger_info->charge_current) {
			case CHARGING_SUSPEND:
				val->intval = 0;
				break;
			case CHARGING_CURRENT_100:
				val->intval = 100;
				break;
			case CHARGING_CURRENT_500:
				val->intval = 500;
				break;
			case CHARGING_CURRENT_FULL:
				val->intval = 1000;
				break;
			default:
				val->intval = 0;
				break;
			}
		}
		else
			val->intval = 0;
		break;

	default:
		return -EINVAL;
	}
	return 0;
}

static int bq2407x_charger_put_property(struct power_supply *psy,
			    enum power_supply_property psp,
			    union power_supply_propval *val)
{
	switch (psp) {
	case POWER_SUPPLY_PROP_CHARGE_ENABLE:
		if (val->intval) {
			charger_info->charge_enable = BQ2407x_POWER_ENABLE;
		} else {
			charger_info->charge_enable = BQ2407x_POWER_DISABLE;
		}
		break;

	case POWER_SUPPLY_PROP_CURRENT_NOW:
		if(val->intval > 1500)
			val->intval = 1500;
		if(val->intval < 40)
			val->intval = 40;

		if(val->intval < 100)
			bq2407x_set_current(CHARGING_SUSPEND);
		else if(val->intval < 500)
			bq2407x_set_current(CHARGING_CURRENT_100);
		else if(val->intval < 1000)
			bq2407x_set_current(CHARGING_CURRENT_500);
		else
			bq2407x_set_current(CHARGING_CURRENT_FULL);
		break;
	default:
		return -EINVAL;
	}
	return 0;
}

static int bq2407x_charger_test_mode_get_property(struct power_supply *psy,
					    enum power_supply_property psp,
					    union power_supply_propval *val)
{
	val->intval = 0;
	switch (psp) {
	case POWER_SUPPLY_PROP_CHARGE_CURRENT:	
		val->intval = charger_info->charge_current;
		break;
#ifdef CONFIG_USB_CAN_SLEEP_TEST
	case POWER_SUPPLY_PROP_SLEEP_ON:
		val->intval = charger_info->charging_sleep;
		break;
#endif
	default:
		return -EINVAL;
	}
	return 0;
}

static int bq2407x_charger_test_mode_put_property(struct power_supply *psy,
					    enum power_supply_property psp,
					    union power_supply_propval *val)
{
	int ret = 0;
	switch (psp) {
	case POWER_SUPPLY_PROP_CHARGE_CURRENT:
		if (CHARGING_SUSPEND <= val->intval
			&& CHARGING_CURRENT_FULL >= val->intval)
			bq2407x_set_current(val->intval);
		else
			ret = -EINVAL;
		break;
#ifdef CONFIG_USB_CAN_SLEEP_TEST
        case POWER_SUPPLY_PROP_SLEEP_ON:
                if(val->intval)
                        charger_info->charging_sleep = 1;
                else
                        charger_info->charging_sleep = 0;
                TDBG("charging sleep enable %d\n",charger_info->charging_sleep);
		break;
#endif
	default:
		ret = -EINVAL;
	}
	return ret;
}

static enum power_supply_property bq2407x_charger_props[] = {
	POWER_SUPPLY_PROP_ONLINE,
	POWER_SUPPLY_PROP_STATUS,
	POWER_SUPPLY_PROP_CHARGE_ENABLE,
	POWER_SUPPLY_PROP_CURRENT_NOW,
};

static enum power_supply_property bq2407x_charger_test_mode_props[] = {
	POWER_SUPPLY_PROP_CHARGE_CURRENT,
#ifdef CONFIG_USB_CAN_SLEEP_TEST
	POWER_SUPPLY_PROP_SLEEP_ON,	
#endif	
};

static char *bq2407x_charger_supplied_to[] = {
	"battery",
};

static struct power_supply bq2407x_charger_supplies[] = {
	{
		.name = "ac",
		.type = POWER_SUPPLY_TYPE_MAINS,
		.supplied_to = bq2407x_charger_supplied_to,
		.num_supplicants = ARRAY_SIZE(bq2407x_charger_supplied_to),
		.properties = bq2407x_charger_props,
		.num_properties = ARRAY_SIZE(bq2407x_charger_props),
		.get_property = bq2407x_charger_get_property,
		.put_property = bq2407x_charger_put_property,
	},
	{
		.name = "usb",
		.type = POWER_SUPPLY_TYPE_USB,
		.supplied_to = bq2407x_charger_supplied_to,
		.num_supplicants = ARRAY_SIZE(bq2407x_charger_supplied_to),
		.properties = bq2407x_charger_props,
		.num_properties = ARRAY_SIZE(bq2407x_charger_props),
		.get_property = bq2407x_charger_get_property,
		.put_property = bq2407x_charger_put_property,
	},
	{
		.name = "test_mode",
		.type = POWER_SUPPLY_TYPE_CHARGER,
		.properties = bq2407x_charger_test_mode_props,
		.num_properties = ARRAY_SIZE(bq2407x_charger_test_mode_props),
		.get_property = bq2407x_charger_test_mode_get_property,
		.put_property = bq2407x_charger_test_mode_put_property,
	},
};

#define ENABLE_OVP_WORK_AROUND
static int count = 1;
static void update_charger(void)
{
	int status_change = 0;
	unsigned long flags;

	if (!charger_info)
		return;

	/* update charger type */	
	if (charger_info->charger_type != charger_info->charger_old_type) {
		TDBG("update charger (type: %d -> %d)\n",
			charger_info->charger_old_type,
			charger_info->charger_type);

		if (charger_info->charger_psy)
			power_supply_changed(charger_info->charger_psy);

		if (charger_info->charger_type == BQ2407x_POWER_CHARGE_AC)
			charger_info->charger_psy = 
					(void *)&bq2407x_charger_supplies[0];
		else if (charger_info->charger_type == BQ2407x_POWER_CHARGE_USB)
			charger_info->charger_psy = 
					(void *)&bq2407x_charger_supplies[1];
		else
			charger_info->charger_psy = NULL; 
	
		charger_info->charger_old_type = charger_info->charger_type;
		if (charger_info->charger_psy)
			power_supply_changed(charger_info->charger_psy);
	}

	/* Dummy battery detection */
	if (charger_info->battery_status == POWER_SUPPLY_STATUS_NOT_CHARGING) {
		//TDBG("Dummy battery do not charge!\n"); 
		charger_info->charge_enable = BQ2407x_POWER_DISABLE;
	}

	/* check charge enable change */
	if (charger_info->charge_enable != charger_info->charge_old_enable) {
		TDBG("update charger (control: %s)\n",
			(charger_info->charge_enable == BQ2407x_POWER_ENABLE
			  ? "enable" : "disable"));
		if (charger_info->pdata->set_charge)
			charger_info->pdata->set_charge(
						charger_info->charge_enable);

		charger_info->charge_old_enable = charger_info->charge_enable;
		status_change = 1;
	}

#ifdef ENABLE_OVP_WORK_AROUND
	/* workaround for reboot issue because of charger IC OVP */
	if (charger_info->charger_type == BQ2407x_POWER_CHARGE_AC) {
        if (charger_info->battery_capacity < 20  || charger_info->battery_capacity > 80) {
                if (charger_info->charge_current == CHARGING_CURRENT_FULL)
    			    bq2407x_set_current(CHARGING_CURRENT_500);
        } else if (charger_info->charge_current == CHARGING_CURRENT_500) 
                    bq2407x_set_current(CHARGING_CURRENT_FULL);
	}
#endif

    if (charger_info->charger_type == BQ2407x_POWER_CHARGE_AC ||
        charger_info->charger_type == BQ2407x_POWER_CHARGE_USB) {
        if (charger_info->battery_capacity == 99 && count == 1) {
            charger_info->pdata->set_current(CHARGING_SUSPEND);
            msleep(500);
            charger_info->pdata->set_current(CHARGING_CURRENT_500);
            count = 0;
        }

        if (charger_info->battery_capacity == 100 && !count)
            count = 1;
    }

	/* update current info */	
	if (charger_info->charge_current != charger_info->charge_old_current) {
		TDBG("update charger (current: %d -> %d)\n",
			charger_info->charge_old_current,
			charger_info->charge_current);

		/* set current to charger IC */
		if (charger_info->pdata->set_current) {	
			charger_info->pdata->set_current(
					charger_info->charge_current);
		}

		charger_info->charge_old_current = charger_info->charge_current;
		status_change = 1;
	}


	/* update charger status */	
	if (!charger_info->pdata->det_irq) {
		if (charger_info->pdata->is_charging) {
			if (charger_info->pdata->is_charging())
				charger_info->charger_status = 
					POWER_SUPPLY_STATUS_CHARGING;
			else
				charger_info->charger_status = 
					POWER_SUPPLY_STATUS_NOT_CHARGING;
		} else {
			if (charger_info->charge_current != CHARGING_SUSPEND
			&& charger_info->charge_enable == BQ2407x_POWER_ENABLE)
				charger_info->charger_status = 
					POWER_SUPPLY_STATUS_CHARGING;
			else
				charger_info->charger_status = 
					POWER_SUPPLY_STATUS_NOT_CHARGING;
		}
	}

	/* check charger status change */
	local_irq_save(flags);
	if (charger_info->charger_status != charger_info->charger_old_status) {
		TDBG("update charger (status:%sCharging)\n",
			charger_info->charger_status == 
			POWER_SUPPLY_STATUS_CHARGING ? " " : " Not ");

		charger_info->charger_old_status 
					= charger_info->charger_status;
		status_change = 1;
	}
	local_irq_restore(flags);

	/* update charger status */	
	if (status_change) {
		if (charger_info->battery_psy) {
			printk("%s: power supply changed\n", __func__);
			power_supply_changed(charger_info->battery_psy);
		}
	}

	return;
}

static void bq2407x_monitor_work(struct work_struct *work)
{
	if (!charger_info)
		return;

	update_battery_info();
	update_charger();

#ifdef CONFIG_USB_CAN_SLEEP_TEST 
	if(charger_info->charging_sleep){
		/* unlock suspend */
		TDBG("bq2407x_charger unlock suspend\n");
		wake_unlock(&charger_info->wake_lock);
		charger_info->charging_sleep = 0;
	}
#endif	
	queue_delayed_work(charger_info->charge_wqueue,
		&charger_info->monitor_work,
		msecs_to_jiffies(charger_info->pdata->wait_for_charger));
}

static irqreturn_t bq2407x_irq_handler(int irq, void *dev_id)
{
	struct bq2407x_charger_info *charger_info = dev_id;
	int wait;
	TDBG("bq2407x_charger_interrupt\n");
	wake_lock_timeout(&charger_info->wake_lock, 5 * HZ);
	if (charger_info && charger_info->pdata) {
		wait = charger_info->pdata->wait_for_status;
		if (charger_info->pdata->is_cable_on) {
		       if (!charger_info->pdata->is_cable_on()) {
			       wait = 0;
			       TDBG("%s: cable removed\n", __func__);
		       }
		} else
			TDBG("%s: is_cable_on is NULL\n", __func__);
		queue_delayed_work(charger_info->charge_wqueue,
				&charger_info->charge_work, msecs_to_jiffies(wait));
	} else {
		TDBG("%s: charger_info is NULL\n", __func__);
	}
	return IRQ_HANDLED;
}


#define FAST_CHARGE_SAFETIME_DETECT
/* 
 * According bq2407x spec. we enable CE once the fast charge timer expires.
 * 
 * if the battery current does not fall to ITERM before the fast charge timer 
 * expires, a fault is indicated. The CHG output flashes at approximately 2 Hz 
 * to indicate a fault condition. The fault condition is cleared by toggling 
 * CE or the input power, entering/exiting USB suspend mode, or an OVP event. 
 *
 */

#ifdef FAST_CHARGE_SAFETIME_DETECT
#define DETECT_IRQ_FREQ_TIMES	8
static int detect_irq_times_frequently = 0;
static unsigned long last_irq_jiffies = 0;

static void charger_safetime_expired_workaround(void)
{
	TDBG("charger_safetime_expired_workaround\n");
	if (charger_info->charge_enable == BQ2407x_POWER_ENABLE) {
		/* update_charger will reset the charge to enable */
		charger_info->pdata->set_charge(BQ2407x_POWER_DISABLE);
		charger_info->charge_old_enable = BQ2407x_POWER_DISABLE;
		/* force update_charger set current again */
		charger_info->charge_old_current = CHARGING_SUSPEND;
	}
}	
#endif

static irqreturn_t bq2407x_det_irq_handler(int irq, void *dev_id)
{
#ifdef FAST_CHARGE_SAFETIME_DETECT
/*
 * This section of code use to detect the CHG pin 2HZ wave when fast charge
 * timer expired. 
 * _______      ____      ____      ____      ____
 *        |____|    |____|    |____|    |____|    |____
 *         (HZ) (HZ) (HZ) (HZ) (HZ) (HZ) (HZ) (HZ)
 * 
 * the tolerated range of irq trigger time is 1/4 HZ. 
 * once we detect 8 times of CHG change, 4 entired 2HZ wave, we consider it
 * is the fast charge timer expired.
 *
 */
	unsigned long irq_jiffies = jiffies;

	if (time_in_range(irq_jiffies, last_irq_jiffies + HZ - (HZ >> 2), 
			  last_irq_jiffies + HZ + (HZ >> 2)))
		detect_irq_times_frequently++;
	else
		detect_irq_times_frequently = 0;

	last_irq_jiffies = irq_jiffies;

	if (unlikely(DETECT_IRQ_FREQ_TIMES < detect_irq_times_frequently)) {
		charger_safetime_expired_workaround();	
	}
#endif

	pr_debug("bq2407x_charger_det_interrupt\n");
	if (charger_info->pdata->is_charging) {
		if (charger_info->pdata->is_charging())
			charger_info->charger_status = 
					POWER_SUPPLY_STATUS_CHARGING;
		else
			charger_info->charger_status = 
					POWER_SUPPLY_STATUS_NOT_CHARGING;
	}

	return IRQ_HANDLED;
}

extern void stop_usb_device(void);
static void update_charger_type(void)
{
	switch (charger_info->charger_type) {	
	case BQ2407x_POWER_CHARGE_AC:
		TDBG("charger on (AC)\n");
		bq2407x_set_current(CHARGING_CURRENT_FULL);
		charger_info->charge_enable = BQ2407x_POWER_ENABLE;
		break;

	case BQ2407x_POWER_CHARGE_USB:
		TDBG("charger on (USB)\n");
		//bq2407x_set_current(CHARGING_SUSPEND);
		charger_info->charge_enable = BQ2407x_POWER_ENABLE;
		break;

	case BQ2407x_POWER_NO_CHARGER: 
		TDBG("charger remove\n");
		bq2407x_set_current(CHARGING_SUSPEND);
		stop_usb_device();
		charger_info->charge_enable = BQ2407x_POWER_DISABLE;
		break;

	default:
		break;
	}

	cancel_delayed_work_sync(&charger_info->monitor_work);
	queue_delayed_work(charger_info->charge_wqueue,
		&charger_info->monitor_work, 0);
}

static void bq2407x_charger_work(struct work_struct *work)
{
	if (!charger_info)
		return;

	if (charger_info->pdata->detect) {
		charger_info->charger_type = charger_info->pdata->detect(); 
	}

	if (charger_info->charger_type == BQ2407x_POWER_CHARGE_AC
		|| charger_info->charger_type == BQ2407x_POWER_CHARGE_USB) {	
		TDBG("bq2407x_charger lock for ac/usb\n");
		wake_lock(&charger_info->wake_lock);
	} else {
		TDBG("bq2407x_charger lock 5 seconds\n");
		wake_lock_timeout(&charger_info->wake_lock, 5 * HZ);
	}

	update_charger_type();
}

static int bq2407x_charger_probe(struct platform_device *pdev)
{
	int ret = 0, i;

	TDBG("bq2407x_charger_probe\n");

	charger_info = (struct bq2407x_charger_info *)kzalloc(
			sizeof(struct bq2407x_charger_info), GFP_KERNEL);
	if (!charger_info) {
		TWARN("kzalloc memory failed!\n");
		return -ENOMEM;
	}

	charger_info->dev = &pdev->dev;
	charger_info->pdata = pdev->dev.platform_data;

	wake_lock_init(&charger_info->wake_lock, WAKE_LOCK_SUSPEND,
			"bq2407x_charger");

	if (!charger_info->pdata->detect ||
	    !charger_info->pdata->is_charging ||
	    !charger_info->pdata->set_charge ||
	    !charger_info->pdata->set_current) {
		TWARN("Please implement the charger hardware functions!\n");
		goto pdata_failed;
	}

	charger_info->charger_status = POWER_SUPPLY_STATUS_NOT_CHARGING;
	charger_info->charger_old_status = POWER_SUPPLY_STATUS_NOT_CHARGING;
	charger_info->charger_type = BQ2407x_POWER_NO_CHARGER;
	charger_info->charger_old_type = BQ2407x_POWER_NO_CHARGER;
	charger_info->charge_current = CHARGING_SUSPEND;
	charger_info->charge_old_current = CHARGING_SUSPEND;
	charger_info->charge_enable = BQ2407x_POWER_DISABLE; 
	charger_info->charge_old_enable = BQ2407x_POWER_DISABLE; 
#ifdef CONFIG_USB_CAN_SLEEP_TEST 
	charger_info->charging_sleep = 0;
#endif

	if (!charger_info->pdata->wait_for_status)
		charger_info->pdata->wait_for_status = 1000;
	if (!charger_info->pdata->wait_for_charger)
		charger_info->pdata->wait_for_charger = 1500;

	if (charger_info->pdata->supplied_to) {
		bq2407x_charger_supplies[0].supplied_to = 
				charger_info->pdata->supplied_to;
		bq2407x_charger_supplies[1].supplied_to = 
				charger_info->pdata->supplied_to;
		bq2407x_charger_supplies[0].num_supplicants = 
				charger_info->pdata->num_supplicants;
		bq2407x_charger_supplies[1].num_supplicants = 
				charger_info->pdata->num_supplicants;
	}

	charger_info->battery_voltage = 0;
	charger_info->battery_capacity = 2;
	charger_info->battery_temperature = 0;
	charger_info->battery_status = POWER_SUPPLY_STATUS_NOT_CHARGING; 

	for (i = 0; i <= 2; i++) {
		ret = power_supply_register(&pdev->dev, 
				&bq2407x_charger_supplies[i]);
		if (ret) {
			TERR("failed to register %s power supply\n",
				bq2407x_charger_supplies[i].name);
			goto supply_failed;
		}
	}

	/* charger hardware init */
	if (charger_info->pdata->init)
		charger_info->pdata->init();

	/* charger set current suspend */
	if (charger_info->pdata->set_current)
		charger_info->pdata->set_current(charger_info->charge_current);

	/* charger set disable */
	if (charger_info->pdata->set_charge)
		charger_info->pdata->set_charge(charger_info->charge_enable);

	if (charger_info->pdata->irq) {
		ret = request_irq(charger_info->pdata->irq, 
				  bq2407x_irq_handler, 
				  IRQF_TRIGGER_RISING | IRQF_TRIGGER_FALLING,
				  "charger", charger_info);
	}

	if (charger_info->pdata->det_irq) {
		ret = request_irq(charger_info->pdata->det_irq, 
				  bq2407x_det_irq_handler, 
				  IRQF_TRIGGER_RISING | IRQF_TRIGGER_FALLING,
				  "charger_det", charger_info);
	}

	INIT_DELAYED_WORK(&charger_info->charge_work, bq2407x_charger_work);
	INIT_DELAYED_WORK(&charger_info->monitor_work, bq2407x_monitor_work);

	charger_info->charge_wqueue = create_singlethread_workqueue("charge");
	if (!charger_info->charge_wqueue) {
                TERR("failed to create work queue.\n");
		ret = -ESRCH;
		goto workqueue_failed;
	}
	queue_delayed_work(charger_info->charge_wqueue,
		&charger_info->monitor_work, msecs_to_jiffies(100));

	queue_delayed_work(charger_info->charge_wqueue,
		&charger_info->charge_work,
		msecs_to_jiffies(charger_info->pdata->wait_for_status));

	goto success;

workqueue_failed:
supply_failed:
	while (i--)
		power_supply_unregister(&bq2407x_charger_supplies[i]);
pdata_failed:
	wake_lock_destroy(&charger_info->wake_lock);
	kfree(charger_info);
	charger_info = NULL;
success:
	return ret;
}

static int bq2407x_charger_remove(struct platform_device *pdev)
{
	int i;
	cancel_rearming_delayed_workqueue(charger_info->charge_wqueue,
					  &charger_info->monitor_work);
	destroy_workqueue(charger_info->charge_wqueue);
	for (i = 2; i >= 0; i--)
		power_supply_unregister(&bq2407x_charger_supplies[i]);

	if (charger_info->pdata->exit)
		charger_info->pdata->exit();

	wake_lock_destroy(&charger_info->wake_lock);
	kfree(charger_info);
	charger_info = NULL;
	return 0;
}

#ifdef CONFIG_PM

static int bq2407x_charger_suspend(struct platform_device *pdev,
		  	pm_message_t state)
{
	return 0;
}

static int bq2407x_charger_resume(struct platform_device *pdev)
{
	return 0;
}

#else

#define bq2407x_charger_suspend NULL
#define bq2407x_charger_resume NULL

#endif /* CONFIG_PM */

static struct platform_driver bq2407x_charger_pdrv = {
	.driver = {
		.name = "charger",
	},
	.probe = bq2407x_charger_probe,
	.remove = bq2407x_charger_remove,
	.suspend = bq2407x_charger_suspend,
	.resume	= bq2407x_charger_resume,
};

static int __init bq2407x_charger_init(void)
{
	return platform_driver_register(&bq2407x_charger_pdrv);
}

static void __exit bq2407x_charger_exit(void)
{
	platform_driver_unregister(&bq2407x_charger_pdrv);
}

late_initcall(bq2407x_charger_init);
module_exit(bq2407x_charger_exit);

MODULE_LICENSE("GPL");
