/*
 * linux/drivers/power/bq27000_battery.c
 *
 * BQ27000 battery driver
 *
 * Copyright (C) 2009 Borqs Ltd.
 * Copyright (C) 2008 Texas Instruments, Inc.
 *
 * Author: Texas Instruments
 *
 * Modefied: River Wang <river.wang@borqs.com>
 *           Ru Yi <yi.ru@borqs.com>
 *
 * This package is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * THIS PACKAGE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 */
#include <linux/module.h>
#include <linux/param.h>
#include <linux/jiffies.h>
#include <linux/workqueue.h>
#include <linux/delay.h>
#include <linux/reboot.h>
#include <linux/platform_device.h>
#include <linux/power_supply.h>
#include <linux/wakelock.h>

#include <mach/pxa3xx_pmic.h>

#include "../w1/w1.h"

#define BQ_BATTERY_TEMP		0x06
#define BQ_BATTERY_TEMP_CK_OFFS	10926
#define BQ_BATTERY_TEMP_FACT_D	40
#define BQ_BATTERY_TEMP_FACT_M	10
#define BQ_BATTERY_VOLT		0x08
#define BQ_BATTERY_RSOC		0x0B /* Relative State-of-Charge */
#define BQ_BATTERY_CACT		0x10
#define BQ_BATTERY_CUR		0x14
#define BQ_BATTERY_CSOC		0x2C

#define	SENSE_RESISTOR_VALUE	20


#define HIGH_BYTE(A)		((A) << 8)

/* 2.9V will cause micco shutdown, we use 3.4v as micco vbat monitor voltage */
#define DEFAULT_VBAT_LOW_SHRESHOLD	3400
#define TEMPERATURE_COEFFICIENT		273
/* hot temp threshold temperature 273+43 */
#define DISCHARGER_HIGH_TEMPERATURE_THRESH	(TEMPERATURE_COEFFICIENT + 43)
/* low temp threshold temperature 273+2 */
#define DISCHARGER_LOW_TEMPERATURE_THRESH	(TEMPERATURE_COEFFICIENT + 2)
/* hot temp threshold temperature 273+60 */
#define POWEROFF_HIGH_TEMPERATURE_THRESH	(TEMPERATURE_COEFFICIENT + 60)
/* low temp threshold temperature 273-20 */
#define POWEROFF_LOW_TEMPERATURE_THRESH		(TEMPERATURE_COEFFICIENT - 20)

extern void w1_bq27000_write(struct device *dev, u8 buf, u8 reg);
extern int w1_bq27000_read(struct device *dev, u8 reg);
struct bq27000_device_info {
	struct device 		*dev;
	struct device		*w1_dev;

	unsigned long		update_time;
	int			status;
	int			health;
	int			voltage_uV;
	int			current_uA;
	int			temp_C;
	int			charge_csoc;
	int			charge_cact;
	int			last_csoc;
	int			bad_battery;

	struct power_supply	bat;
	struct delayed_work	monitor_work;
	struct workqueue_struct *battery_wq;
	struct wake_lock	wake_lock;
};

static enum power_supply_property bq27000_battery_props[] = {	
	POWER_SUPPLY_PROP_STATUS,
	POWER_SUPPLY_PROP_HEALTH,
	POWER_SUPPLY_PROP_PRESENT,
	POWER_SUPPLY_PROP_CAPACITY,
	POWER_SUPPLY_PROP_TECHNOLOGY,
	POWER_SUPPLY_PROP_VOLTAGE_NOW,
	POWER_SUPPLY_PROP_TEMP,
	POWER_SUPPLY_PROP_CURRENT_NOW,
};

static int bq27000_battery_probe(struct platform_device *dev);
static int bq27000_battery_remove(struct platform_device *dev);
#ifdef CONFIG_PM
static int bq27000_battery_suspend(struct platform_device *dev,
	pm_message_t state);
static int bq27000_battery_resume(struct platform_device *dev);
#endif

static struct platform_driver bq2700_battery_driver = {
	.probe = bq27000_battery_probe,
	.remove = bq27000_battery_remove,
#ifdef CONFIG_PM
	.suspend = bq27000_battery_suspend,
	.resume = bq27000_battery_resume,
#endif
	.driver = {
		.name = "bq27000-battery",
	},
};

static inline int read_bq_val(u8 reg, int *rt_value, int b_single,
	struct bq27000_device_info *di);

/*
 * Return the battery temperature in Celcius degrees
 * Or < 0 if something fails.
 */
static int bqbattery_temperature(struct bq27000_device_info *di)
{
	int ret, temp = 0;

	ret = read_bq_val(BQ_BATTERY_TEMP, &temp, 0, di);
	if (ret) {
		pr_err("BQ27000 battery driver:"
			"Error reading temperature from HDQ device\n");
		return ret;
	}

	return  ((temp * BQ_BATTERY_TEMP_FACT_M) - BQ_BATTERY_TEMP_CK_OFFS)
			/ BQ_BATTERY_TEMP_FACT_D;
}

/*
 * Return the battery Voltage in milivolts
 * Or < 0 if something fails.
 */
static int bqbattery_voltage(struct bq27000_device_info *di)
{
	int ret, volt = 0;

	ret = read_bq_val(BQ_BATTERY_VOLT, &volt, 0, di);
	if (ret) {
		pr_err("BQ27000 battery driver:"
			"Error reading battery voltage from HDQ device\n");
		return ret;
	}

	return volt;
}

/*
 * Return the battery average current
 * Or < 0 if something fails.
 */
static int bqbattery_current(struct bq27000_device_info *di)
{
	int ret, curr = 0;

	ret = read_bq_val(BQ_BATTERY_CUR, &curr, 0, di);
	if (ret) {
		pr_err("BQ27000 battery driver:"
			"Error reading battery current from HDQ device\n");
		return ret;
	}

	return curr*357/SENSE_RESISTOR_VALUE/100;
}

/*
 * Return the battery Relative State-of-Charge
 * Or < 0 if something fails.
 */
static int bqbattery_csoc(struct bq27000_device_info *di)
{
	int ret, csoc = 0;

	ret = read_bq_val(BQ_BATTERY_CSOC, &csoc, 1, di);
	if (ret) {
		pr_err("BQ27000 battery driver:"
			"Error reading battery Conpensated"
			"State-of-Charge from HDQ device\n");
		return ret;
	}
	return csoc;
}

/*
 * Return the battery Temperature Compensated capacity
 * Or < 0 if something fails.
 */
static int bqbattery_cact(struct bq27000_device_info *di)
{
	int ret, cact = 0;

	ret = read_bq_val(BQ_BATTERY_CACT, &cact, 0, di);
	if (ret) {
		pr_err("BQ27000 battery driver:"
			"Error reading battery Conpensated"
			"battery capacity from HDQ device\n");
		return ret;
	}
	return cact*357/SENSE_RESISTOR_VALUE/100;
}

/*
 * Read the content of battery reg through the HDQ interface.
 */
static inline int read_bq_val(u8 reg, int *rt_value, int b_single,
	struct bq27000_device_info *di)
{
	u8 high0, low0, high1,low1;;

	if (b_single) {
		low0 = w1_bq27000_read(di->w1_dev, reg);
		*rt_value = low0 & 0xff;
	}
	else {
		high0 = w1_bq27000_read(di->w1_dev, reg + 1);
		low0 = w1_bq27000_read(di->w1_dev, reg);
		high1 = w1_bq27000_read(di->w1_dev, reg + 1);
		if (high0 == high1 ) {
			*rt_value = low0 + HIGH_BYTE((int) high0);
		} else {
			low1 = w1_bq27000_read(di->w1_dev, reg);
			*rt_value = low1 + HIGH_BYTE((int) high1);
		}
	}

	return 0;
}

#if defined(CONFIG_SUPPORT_DUMMY_BATTERY)

#if defined(CONFIG_PXA3xx_MICCO)
#include <mach/micco.h>

static int micco_read_battery_voltage(void)
{
        /*Do ACD_IN0 measurement*/
        u8 val;
        int ret;
        micco_read(MICCO_VBAT_RES, &val);

        /*According to the formula from marvell*/
        ret =(2650+(val*2650)/255);
        pr_debug("%s: Battery voltage = %d mV\n", __func__, ret);
        return ret;
}
#else
static int micco_read_battery_voltage(void) { return 3800; }
#endif

static int check_dummy_battery(struct bq27000_device_info *di)
{
        u16 vbat_micco = 0;
	di->charge_csoc = bqbattery_csoc(di);

	vbat_micco = micco_read_battery_voltage();	
	if(di->charge_csoc == 0 && di->voltage_uV == 0 && vbat_micco > 3500)
        {
                printk("dummy battery capacity=%d vbat=%d vbat_micco=%d \n",
			di->charge_csoc,di->voltage_uV,vbat_micco);

		di->voltage_uV = 3800; 
		di->current_uA = 0; 
		di->temp_C = 20; 
		di->charge_csoc = 10;
		di->charge_cact = 0; 
                di->bad_battery = 1;
        } else {
                di->bad_battery = 0;
	}

	return di->bad_battery;
}
#endif

static void bq27000_battery_work(struct work_struct *work)
{
	struct bq27000_device_info *di = container_of(work,
			struct bq27000_device_info, monitor_work.work);
	int interval = HZ * 30;

	wake_lock(&di->wake_lock);
	/* read voltage first */
	di->voltage_uV = bqbattery_voltage(di);
#if defined(CONFIG_SUPPORT_DUMMY_BATTERY)
	if (unlikely(di->bad_battery < 0)) {
		if(check_dummy_battery(di)){
			wake_unlock(&di->wake_lock);
			return;
		}
	}
#endif
	if (di->voltage_uV == 0) {
		printk(KERN_ERR"%s: Battery already is bad or your battery"
			"had been taken off, force power down.\n", __func__);
		di->bad_battery = 1;
		di->charge_csoc = 0;
		di->temp_C = 0;
		power_supply_changed(&di->bat);
		wake_unlock(&di->wake_lock);
		return;
	}

	/* read current, temperature, capacity and capcity level */
	di->current_uA = bqbattery_current(di);
	di->temp_C = bqbattery_temperature(di);
	di->charge_csoc = bqbattery_csoc(di);
	di->charge_cact = bqbattery_cact(di);

	if (di->charge_csoc < 15)
		interval = HZ * 10;
	if (power_supply_am_i_supplied(&di->bat))
		interval = HZ * 5;
	
	if((di->voltage_uV <= DEFAULT_VBAT_LOW_SHRESHOLD)
			&& (di->charge_csoc < 6) && (di->charge_csoc > 1))
	{
		printk(KERN_INFO "%s: BQ27000 battery toolow (%duV %d)\n",
			__func__, di->voltage_uV, di->charge_csoc);
		di->last_csoc = di->charge_csoc = 0;
		power_supply_changed(&di->bat);
	}

	if (di->temp_C + TEMPERATURE_COEFFICIENT >= POWEROFF_HIGH_TEMPERATURE_THRESH) {
		di->health = POWER_SUPPLY_HEALTH_OVERHEAT;
		power_supply_changed(&di->bat);
	} else if (di->temp_C + TEMPERATURE_COEFFICIENT <= POWEROFF_LOW_TEMPERATURE_THRESH) {
		di->health = POWER_SUPPLY_HEALTH_COLD; 
		power_supply_changed(&di->bat);
	}

	if (di->charge_csoc != di->last_csoc) {
		di->last_csoc = di->charge_csoc;
		printk(KERN_DEBUG "%s: BQ27000 battery update (%duV %d)\n",
			__func__, di->voltage_uV, di->charge_csoc);
		power_supply_changed(&di->bat);
	}

	wake_unlock(&di->wake_lock);
	queue_delayed_work(di->battery_wq, &di->monitor_work, interval);
	return;
}

#define to_bq27000_device_info(x) container_of((x), \
				struct bq27000_device_info, bat);

static int bq27000_battery_get_property(struct power_supply *psy,
					enum power_supply_property psp,
					union power_supply_propval *val)
{
	struct bq27000_device_info *di = to_bq27000_device_info(psy);

	switch (psp) {
	case POWER_SUPPLY_PROP_STATUS:
#if defined(CONFIG_SUPPORT_DUMMY_BATTERY)
		if(unlikely(di->bad_battery == 1)){
			val->intval = POWER_SUPPLY_STATUS_NOT_CHARGING;
			break;
		}
#endif
		if (unlikely((di->temp_C + TEMPERATURE_COEFFICIENT) 
				> DISCHARGER_HIGH_TEMPERATURE_THRESH 
			|| (di->temp_C + TEMPERATURE_COEFFICIENT)
				< DISCHARGER_LOW_TEMPERATURE_THRESH)) {
			val->intval = POWER_SUPPLY_STATUS_NOT_CHARGING;
			break;
		}

		if(unlikely(di->charge_csoc == 100)) 
			val->intval = POWER_SUPPLY_STATUS_FULL;
		else if (power_supply_am_i_supplied(&di->bat))
			val->intval = POWER_SUPPLY_STATUS_CHARGING;
		else
			val->intval = POWER_SUPPLY_STATUS_DISCHARGING;
		break;
	case POWER_SUPPLY_PROP_HEALTH:
		val->intval = di->health;
		break;
	case POWER_SUPPLY_PROP_PRESENT:
		val->intval = di->voltage_uV > 0 ? 1 : 0;
		break;
	case POWER_SUPPLY_PROP_CAPACITY:
		val->intval = di->charge_csoc;
		break;
	case POWER_SUPPLY_PROP_CAPACITY_LEVEL:
		val->intval = di->charge_cact;
		break;
	case POWER_SUPPLY_PROP_TECHNOLOGY:
		val->intval = POWER_SUPPLY_TECHNOLOGY_LION;
		break;
	case POWER_SUPPLY_PROP_VOLTAGE_NOW:
		val->intval = (di->voltage_uV * 1000);
		break;
	case POWER_SUPPLY_PROP_TEMP:
		val->intval = di->temp_C;
		break;
	case POWER_SUPPLY_PROP_CURRENT_NOW:
		val->intval = (di->current_uA * 1000);
		break;

	default:
		return -EINVAL;
	}

	return 0;
}

static int bq27000_battery_put_property(struct power_supply *psy,
					enum power_supply_property psp,
					union power_supply_propval *val)
{
	switch (psp) {
	default:
		return -EINVAL;
	}

	return 0;
}

static int bq27000_battery_probe(struct  platform_device *pdev)
{
	struct bq27000_device_info *di;
	int retval = 0;

	di = kzalloc(sizeof(*di), GFP_KERNEL);
	if (!di) {
		pr_err("BQ27000 battery driver:"
				"Failed to allocate device info structure\n");
		return -ENOMEM;
	}

	platform_set_drvdata(pdev, di);

	di->dev = &pdev->dev;
	di->w1_dev = pdev->dev.parent;
	di->bat.name = "battery";
	di->bat.type = POWER_SUPPLY_TYPE_BATTERY;
	di->bat.properties = bq27000_battery_props;
	di->bat.num_properties = ARRAY_SIZE(bq27000_battery_props);
	di->bat.get_property = bq27000_battery_get_property;
	di->bat.put_property = bq27000_battery_put_property;
	di->bat.external_power_changed = NULL;
	di->voltage_uV = 3800; 
	di->current_uA = 0; 
	di->temp_C = 20; 
	di->charge_csoc = 2;
	di->charge_cact = 0; 
	di->last_csoc = 0;
	di->health = POWER_SUPPLY_HEALTH_GOOD;
	di->bad_battery = -1;
	wake_lock_init(&di->wake_lock, WAKE_LOCK_SUSPEND, "battery");

	retval = power_supply_register(&pdev->dev, &di->bat);
	if (retval) {
		pr_err("BQ27000 battery driver: Failed to register battery\n");
		goto batt_failed;
	}

	printk(KERN_INFO "%s: BQ27000 battery power_supply_register ok\n",__func__);

	di->battery_wq = create_singlethread_workqueue("battery");
	if (!di->battery_wq) {
		pr_err("failed to create work queue.\n");
		retval = -ESRCH;
		goto batt_wq_failed;
	}

	INIT_DELAYED_WORK(&di->monitor_work, bq27000_battery_work);
	queue_delayed_work(di->battery_wq, &di->monitor_work, 0);

	return 0;

batt_wq_failed:
	power_supply_unregister(&di->bat);
batt_failed:
	kfree(di);
	return retval;
}

static int bq27000_battery_remove(struct  platform_device *pdev)
{
	struct bq27000_device_info *di = platform_get_drvdata(pdev);

	flush_scheduled_work();
	power_supply_unregister(&di->bat);
	platform_set_drvdata(pdev, NULL);
	kfree(di);

	return 0;
}

#ifdef CONFIG_PM
static int bq27000_battery_suspend(struct platform_device *pdev,
	pm_message_t state)
{
	struct bq27000_device_info *di = platform_get_drvdata(pdev);

	cancel_delayed_work(&di->monitor_work);
	return 0;
}

static int bq27000_battery_resume(struct platform_device *pdev)
{
	struct bq27000_device_info *di = platform_get_drvdata(pdev);

	schedule_delayed_work(&di->monitor_work, 0);
	return 0;
}
#endif

static int __init bq27000_battery_init(void)
{

	return platform_driver_register(&bq2700_battery_driver);

}

static void __exit bq27000_battery_exit(void)
{
	platform_driver_unregister(&bq2700_battery_driver);
}

module_init(bq27000_battery_init);
module_exit(bq27000_battery_exit);

MODULE_AUTHOR("Texas Instruments");
MODULE_DESCRIPTION("BQ27000 battery moniter driver");
MODULE_LICENSE("GPL");
