/* 
 * driver/power/pda_battery.c
 *
 * Support for the pdaon battery driver.
 *
 * This software is licensed under the terms of the GNU General Public
 * License version 2, as published by the Free Software Foundation, and
 * may be copied, distributed, and modified under those terms.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

#include <linux/module.h>
#include <linux/param.h>
#include <linux/jiffies.h>
#include <linux/workqueue.h>
#include <linux/pm.h>
#include <linux/platform_device.h>
#include <linux/power_supply.h>
#include <linux/interrupt.h>
#include <linux/delay.h>
#include <linux/pda_battery.h>

#ifdef CONFIG_WAKELOCK
#include <linux/wakelock.h>
static struct wake_lock pda_battery_lock = {
//	.name = "pda_battery"
};
#endif

static int g_is_call = 0, g_call_signal = 0;
static char call_status[5];

struct pda_battery_info {
	struct device *dev;

	int technology;		/* */
	int health;		/* */
	int present;		/* */
	int soc;		/* in percents! */
	int soc_new;
	int temp_C;		/* C */
	int voltage_mV;		/* mV */
	int vol_toolow;

	int charging_status;	/* POWER_SUPPLY_STATUS_* */

	struct power_supply bat;
	struct workqueue_struct *monitor_wqueue;
	struct delayed_work monitor_work;
};

#define to_pda_battery_info(x) container_of((x), struct pda_battery_info, bat);

static int debug_mask = 1;
module_param_named(debug_mask, debug_mask, int, 0644);

static struct pda_battery_pdata *cb_pdata;

static int charging_is_off(int status)
{
	if (status != POWER_SUPPLY_STATUS_CHARGING && status != POWER_SUPPLY_STATUS_FULL)
		return 1;
	else
		return 0;
}

static int soc_smooth(int soc_new, int status)
{
#define SLOT_INTERVAL	(120)
	static int soc = -1;
	static struct timespec last_time = {.tv_sec = -100 * SLOT_INTERVAL};
	struct timespec time, interval;
	int w, slots;

	do_posix_clock_monotonic_gettime(&time);
	monotonic_to_bootbased(&time);
	interval = timespec_sub(time, last_time);
	slots = (interval.tv_sec + SLOT_INTERVAL) / SLOT_INTERVAL;

	soc_new *= 100;

	if (unlikely(debug_mask)) {
		printk("-- soc: %d soc_new: %d intval: %lu slots: %d status: %d\n", 
				soc, soc_new, interval.tv_sec, slots, status);
	}

	if (slots < 2)
		goto exit;

	/* in discharging mode, we can not increase the capacity */
	if ((charging_is_off(status)) && (soc_new >= soc) && (-1 != soc))
		goto exit;

	if ((soc - soc_new) > (500 * (slots - 1))) {
		soc_new = soc - 500 * (slots - 1);
	}

	if ((soc_new - soc) > 500 * (slots - 1)) {
		soc_new = soc + 500 * (slots - 1);
	}

	if (soc_new < 100 || -1 == soc)
		w = 0;
	else
		w = 4;

	soc = (soc * w + soc_new * 2) / (w + 2);
	if (soc > 9980)
		soc = 10000;

	last_time = time;

	if (unlikely(debug_mask)) {
		printk("++ soc: %d soc_new: %d w: %d intval: %lu slots: %d\n", 
				soc, soc_new, w, interval.tv_sec, slots);
	}
exit:
	return soc/100;
}

static int pda_battery_read_status(struct pda_battery_info *di)
{
	di->charging_status = power_supply_am_i_supplied_status(&di->bat);
	if (di->charging_status == 0)
		di->charging_status = POWER_SUPPLY_STATUS_DISCHARGING;

	if (!cb_pdata)
		return 0;

	di->health = POWER_SUPPLY_HEALTH_GOOD;
	di->present = 1;

	di->voltage_mV = 4000;
	if (cb_pdata->get_battery_volt)
		di->voltage_mV = cb_pdata->get_battery_volt();

	di->temp_C = 30;
	if (cb_pdata->get_battery_temp)
		di->temp_C = cb_pdata->get_battery_temp();

	di->soc = 80;

	if (cb_pdata->get_battery_soc) {
		di->soc_new = cb_pdata->get_battery_soc(di->voltage_mV, di->temp_C);
		di->soc = soc_smooth(di->soc_new, di->charging_status);
	}

	return 0;
}

static int get_battery_monitor_interval(struct pda_battery_info *di)
{
	int interval;
	if (di->soc <= 10)
		interval = 10 * HZ;
	else
		interval = 120 * HZ;
	return interval;
}

static int pda_battery_monitor_soc(struct pda_battery_info *di)
{
	if (di->soc > 10) {
		cb_pdata->set_monitor_soc(10);
	} else if (di->soc > 5) {
		cb_pdata->set_monitor_soc(5);
	} else {
		cb_pdata->set_monitor_soc(1);
	}
	return 0;
}

static void pda_battery_work(struct work_struct *work)
{
	int last_soc, last_status, monitor_interval;
	struct pda_battery_info *di;

	di = container_of(work, struct pda_battery_info, monitor_work.work);

	last_soc = di->soc;
	last_status = di->charging_status;

	pda_battery_read_status(di);

	if (last_soc != di->soc || last_status != di->charging_status)
		power_supply_changed(&di->bat);

	pda_battery_monitor_soc(di);
	monitor_interval = get_battery_monitor_interval(di);

	cancel_delayed_work(&di->monitor_work);
	queue_delayed_work(di->monitor_wqueue, &di->monitor_work, monitor_interval);
}


static void pda_battery_external_power_changed(struct power_supply *psy)
{
	struct pda_battery_info *di = to_pda_battery_info(psy);

//	cancel_delayed_work(&di->monitor_work);
//	queue_delayed_work(di->monitor_wqueue, &di->monitor_work, 0);
}
 
static int pda_battery_put_property(struct power_supply *psy,
				       enum power_supply_property psp,
				       union power_supply_propval *val)
{
	struct pda_battery_info *di = to_pda_battery_info(psy);

	switch (psp) {
	case POWER_SUPPLY_PROP_CALL_STATUS:
		sscanf(val->strval, "%d,%d", &g_is_call, &g_call_signal);
		if (((g_is_call != 0) && (g_is_call != 1)) || 
				(g_call_signal < 0) || (g_call_signal > 100))
			return -EINVAL;
		else
			return 0;
/*	case POWER_SUPPLY_PROP_BATT_VOL_TOO_LOW:
		di->vol_toolow = !!val->intval;
		break;
*/		
	default:
		break;
	}

	return 0;
}

static int pda_battery_get_property(struct power_supply *psy,
				     enum power_supply_property psp,
				     union power_supply_propval *val)
{
	struct pda_battery_info *di = to_pda_battery_info(psy);

	switch (psp) {
	case POWER_SUPPLY_PROP_STATUS:
		val->intval = di->charging_status;
		break;
	case POWER_SUPPLY_PROP_HEALTH:
		val->intval = di->health;
		break;
	case POWER_SUPPLY_PROP_PRESENT:
		val->intval = di->present;
		break;
	case POWER_SUPPLY_PROP_CAPACITY:
		val->intval = di->soc;
		break;
	case POWER_SUPPLY_PROP_TECHNOLOGY:
		val->intval = di->technology;
		break;
	case POWER_SUPPLY_PROP_BATT_VOL:
		val->intval = di->voltage_mV;
		break;
	case POWER_SUPPLY_PROP_BATT_TEMP:
		val->intval = di->temp_C;
		break;
/*	case POWER_SUPPLY_PROP_BATT_VOL_TOO_LOW:
		val->intval = di->vol_toolow;
		break;
*/
	case POWER_SUPPLY_PROP_CALL_STATUS:
		sprintf(call_status, "%d,%d", g_is_call, g_call_signal);
		val->strval = call_status;
		break;		
	default:
		return -EINVAL;
	}

	return 0;
}

static enum power_supply_property pda_battery_props[] = {
	POWER_SUPPLY_PROP_STATUS,
	POWER_SUPPLY_PROP_HEALTH,
	POWER_SUPPLY_PROP_PRESENT,
	POWER_SUPPLY_PROP_CAPACITY,
	POWER_SUPPLY_PROP_TECHNOLOGY,
	POWER_SUPPLY_PROP_BATT_VOL,
	POWER_SUPPLY_PROP_BATT_TEMP,
};

static int __init pda_battery_probe(struct platform_device *pdev)
{
	int retval = 0;
	struct pda_battery_info *di;

	di = kzalloc(sizeof(*di), GFP_KERNEL);
	if (!di) {
		retval = -ENOMEM;
		goto di_alloc_failed;
	}

	platform_set_drvdata(pdev, di);

	cb_pdata = pdev->dev.platform_data;
	di->dev = &pdev->dev;
	di->bat.name = cb_pdata->name;
	di->bat.type = POWER_SUPPLY_TYPE_BATTERY;
	di->bat.properties = pda_battery_props;
	di->bat.num_properties = ARRAY_SIZE(pda_battery_props);
	di->bat.get_property = pda_battery_get_property;
	di->bat.put_property = pda_battery_put_property;
	di->bat.external_power_changed = pda_battery_external_power_changed;
	di->charging_status = POWER_SUPPLY_STATUS_UNKNOWN;
	di->technology = POWER_SUPPLY_TECHNOLOGY_LION;
	memset(call_status, 0, sizeof(call_status));

	INIT_DELAYED_WORK(&di->monitor_work, pda_battery_work);
	di->monitor_wqueue = create_singlethread_workqueue(pdev->dev.bus_id);
	if (!di->monitor_wqueue) {
		printk("failed to create work queue.\n");
		retval = -ESRCH;
		goto workqueue_failed;
	}

//	queue_delayed_work(di->monitor_wqueue, &di->monitor_work, HZ);

	pda_battery_read_status(di);

	retval = power_supply_register(&pdev->dev, &di->bat);
	if (retval) {
		printk("failed to register battery\n");
		goto batt_failed;
	}


	goto success;

workqueue_failed:
	power_supply_unregister(&di->bat);
batt_failed:
	kfree(di);
di_alloc_failed:
success:
	return retval;
}

static int pda_battery_remove(struct platform_device *pdev)
{
	struct pda_battery_info *di = platform_get_drvdata(pdev);

	cancel_delayed_work_sync(&di->monitor_work);
	destroy_workqueue(di->monitor_wqueue);
	power_supply_unregister(&di->bat);
	cb_pdata = NULL;	

	return 0;
}

static struct platform_driver pda_battery_driver = {
	.driver = {
		.name = "pda_battery",
	},
	.probe = pda_battery_probe,
	.remove = pda_battery_remove,
};

static int __init pda_battery_init(void)
{
#ifdef CONFIG_WAKELOCK
	wake_lock_init(&pda_battery_lock, WAKE_LOCK_SUSPEND, "pda_battery");
#endif
	return platform_driver_register(&pda_battery_driver);
}

static void __exit pda_battery_exit(void)
{
	platform_driver_unregister(&pda_battery_driver);
#ifdef CONFIG_WAKELOCK
	wake_lock_destroy(&pda_battery_lock);
#endif
}

module_init(pda_battery_init);
module_exit(pda_battery_exit);

MODULE_LICENSE("GPL");
