/*
 * driver/sensors/alsps.c 
 *
 * Support CAPELLA CM36xx Short Distance Proximity Sensor with 
 * Ambient Light Sensor
 *
 * Copyright (C) 2008 Borqs, Ltd.
 * Author: Ma Yilei <yilei.ma@borqs.com>
 *         Ru Yi <yi.ru@borqs.com>
 *
 * This software is licensed under the terms of the GNU General Public
 * License version 2, as published by the Free Software Foundation, and
 * may be copied, distributed, and modified under those terms.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 */
#include <linux/types.h>
#include <linux/errno.h>
#include <linux/miscdevice.h>
#include <linux/fs.h>
#include <mach/pxa-regs.h>
#include <asm/uaccess.h>
#include <linux/seq_file.h>
#include <linux/proc_fs.h>
#include <linux/slab.h>
#include <mach/gpio.h>
#include <asm/signal.h>
#include <linux/interrupt.h>
#include <linux/device.h>
#include <linux/poll.h>
#include <linux/sysfs.h>
#include <linux/stat.h>
#include <linux/completion.h>
#include <linux/sensors.h>
#include <linux/time.h>
#include <linux/delay.h>
#include <linux/workqueue.h>
#include <linux/input.h>
#include <mach/alsps.h>
#include <linux/platform_device.h>
#include <linux/list.h>
#include <asm/atomic.h>
#include <mach/micco.h>
#include <mach/pxa3xx_pmic.h>
#include <linux/wakelock.h>
#include <linux/earlysuspend.h>

static struct wake_lock alsps_suspend_lock;

struct alsps_struct {
	int irq;
	int adc_event;
	int (*init)(void);
	int (*get_gpio_value)(void);
	void (*setpower)(int enable);
	int (*chip_verify) (int *, int);

	wait_queue_head_t proximity_workq_head;
	wait_queue_head_t ambient_workq_head;
	struct semaphore sem_ambient_interval;
	struct delayed_work ambient_workq;

	
	/* proximity sensor user count */
	int proximity_count;	
	/* ambient light sensor user count */
	int ambient_count;	
	/* ambient light interval */
	int ambient_interval;

	/* alsps sensor switch stat */
	int alsps_switch;	
	/* alsps sensor suspend stat. 1:0 */
	int alsps_suspend;	
};
static struct alsps_struct *alsps;

#define ALSPS_PROXIMITY_DATA_READY 	0
#define ALSPS_AMBIENT_DATA_READY 	1
#define ALSPS_PROXIMITY_FIRST_POLL	2
#define ALSPS_AMBIENT_FIRST_POLL	3

#define DEFAULT_AMBIENT_INTERVAL	500	
#define AMBITNT_ADC_THRESHOLD		2

struct alsps_stat {
	struct list_head list;
	unsigned long status;
};
static LIST_HEAD(proximity_process_list);
static LIST_HEAD(ambient_process_list);
static struct semaphore sem_ambient_list;
static struct semaphore sem_proximity_list;

static unsigned int previous_lumen = 0;
static unsigned int current_lumen = 0;

/**
 *now current lumen used for backlight adjusting, so this group of value ok!
 *back light  environment lum  ADC value
 *200 	      >12500           159
 *255         >400  and <12500 31~159
 *200         >200 and <400    5~31
 *150         >1 and <200      5
 *130 	      0                0
 */
static unsigned int ambient_get_lum(unsigned int adc)
{
	unsigned int lum;

	lum = 0;

	if ((adc > 0) && (adc <= 5)) {
		lum = 150;
	} else if ((adc > 5) && (adc <= 31)) {
		lum = 300;
	} else if ((adc > 31) && (adc < 159)) {
		lum = 8000;
	} else if (adc >= 159) {
		lum = 15000;
	}
	return lum;
}


/* 
 * Proximity sensor operations 
 */
static void proximity_handler(void)
{
	struct alsps_stat *ptr;

	if (alsps->proximity_count > 0) {
		list_for_each_entry(ptr, &proximity_process_list, list) {
			set_bit(ALSPS_PROXIMITY_DATA_READY, &ptr->status);
		}
		wake_up(&alsps->proximity_workq_head);
	}
}

static irqreturn_t proximity_isr(int irq, void * dev_id)
{
	printk(KERN_DEBUG "alsps: proximity trigger!\n");

	if (alsps->alsps_switch & PROXIMITY_ENABLE) {
		proximity_handler();
		wake_lock_timeout(&alsps_suspend_lock,HZ/2);
	} 
	return IRQ_HANDLED;
}

static ssize_t proximity_read(struct file *filep, char __user *buff, size_t size, loff_t *offset)
{
	int ret;
	int level;
	struct input_event  data;
	struct alsps_stat *ptr;

	ptr = (struct alsps_stat *)(filep->private_data);
	level = 0;

	ret = sizeof(struct input_event);
	if (alsps->alsps_switch & PROXIMITY_ENABLE) {
		level =	alsps->get_gpio_value();
		get_monotonic_time(&(data.time));
		data.type = EV_ABS;
		data.code = ABS_DISTANCE;
		data.value = (level > 0 ? 1 : 0);
		clear_bit(ALSPS_PROXIMITY_DATA_READY, &ptr->status);
		if (0 != copy_to_user(buff, (const void *)&data, 
				      sizeof(struct input_event))) {
			ret = -ENOMEM;
		}
		printk(KERN_DEBUG "alsps: copy %d bytes to user space %d\n", ret, level);
	}

	return ret;
}

static	unsigned int proximity_poll(struct file *filep, struct poll_table_struct *wait)
{
	unsigned int mask;
	struct alsps_stat *ptr;

	mask = 0;
	ptr = (struct alsps_stat *)filep->private_data;
	poll_wait(filep, &alsps->proximity_workq_head, wait);
	if (test_bit(ALSPS_PROXIMITY_FIRST_POLL, &ptr->status)) {
		mask |= (POLLIN | POLLRDNORM);
		clear_bit(ALSPS_PROXIMITY_FIRST_POLL, &ptr->status);
	}
	if (test_bit(ALSPS_PROXIMITY_DATA_READY, &ptr->status)) {
		mask |= (POLLIN | POLLRDNORM);
	}

	return mask;
}

static int proximity_open(struct inode * inode, struct file * filep)
{
	struct alsps_stat *ptr;

	ptr = kzalloc(sizeof(struct alsps_stat), GFP_KERNEL);
	if (ptr == NULL) {
		printk("alsps: proximity open kzalloc failed!\n");
		return -ENOMEM;
	}

	clear_bit(ALSPS_PROXIMITY_DATA_READY, &ptr->status);
	set_bit(ALSPS_PROXIMITY_FIRST_POLL, &ptr->status);
	filep->private_data = ptr;

	disable_irq(alsps->irq);
	down(&sem_proximity_list);
	list_add(&ptr->list, &proximity_process_list);
	++alsps->proximity_count;
	printk(KERN_DEBUG "proximity_count:%d\n", alsps->proximity_count);
	up(&sem_proximity_list);
	enable_irq(alsps->irq);

	return 0;
}

static int proximity_close(struct inode * inode, struct file *filep)
{
	struct alsps_stat *ptr;
	ptr = (struct alsps_stat *)filep->private_data;
	disable_irq(alsps->irq);
	down(&sem_proximity_list);
	list_del(&ptr->list);
	kfree(ptr);
	up(&sem_proximity_list);
	enable_irq(alsps->irq);

	if (--alsps->proximity_count <= 0) {
		alsps->proximity_count = 0;
		alsps->alsps_switch &= ~PROXIMITY_ENABLE;
		alsps->setpower(alsps->alsps_switch);	
	} 

	return 0;
}

/* 
 * Light sensor operations 
 */
static void ambient_delayed_worker(struct work_struct *work)
{
	if ((alsps->alsps_switch & AMBIENT_ENABLE) && !alsps->alsps_suspend) {
		//micco_auto_adc_enable(alsps->adc_event, 1);
		micco_enable_adc_irq(alsps->adc_event, 1);
	}
}

static void ambient_adc_interrupt(unsigned long event)
{
	struct alsps_stat *ptr;
	u8 adc = 0;

	if (alsps->alsps_switch & AMBIENT_ENABLE) {
		micco_enable_adc_irq(alsps->adc_event, 0);
		micco_auto_adc_read(alsps->adc_event, &adc);
		//micco_auto_adc_enable(alsps->adc_event, 0);

		current_lumen = ambient_get_lum(adc);

		printk(KERN_DEBUG "ambientd - adc: %d, lumen: %d\n", adc, current_lumen);
		if (current_lumen != previous_lumen) {
			printk(KERN_DEBUG "alsps: current lumen changeed!\n");
			down(&sem_ambient_list);
			list_for_each_entry(ptr, &ambient_process_list, list) {
				set_bit(ALSPS_AMBIENT_DATA_READY, &ptr->status);
			}
			up(&sem_ambient_list);
			wake_up(&alsps->ambient_workq_head);
			previous_lumen = current_lumen;
		}

		micco_auto_adc_set_trigger(alsps->adc_event, 
					   adc < (255 - AMBITNT_ADC_THRESHOLD) ? (adc + AMBITNT_ADC_THRESHOLD) : 255, 
					   adc > AMBITNT_ADC_THRESHOLD ? (adc - AMBITNT_ADC_THRESHOLD) : 0);
		schedule_delayed_work(&alsps->ambient_workq, 
				     msecs_to_jiffies(alsps->ambient_interval));
	}

	return;
}

static ssize_t ambient_read(struct file *filep, char __user *buff, size_t size, loff_t *offset)
{
	int ret;
	struct input_event  data;
	struct alsps_stat *ptr;

	ret = sizeof(struct input_event);

	down(&sem_ambient_list);
	ptr = (struct alsps_stat *)(filep->private_data);
	get_monotonic_time(&(data.time));
	data.type = EV_ABS;
	data.code = ABS_MISC;
	data.value = current_lumen;
	clear_bit(ALSPS_AMBIENT_DATA_READY, &ptr->status);
	if (0 != copy_to_user(buff, (const void *)&data, 
			      sizeof(struct input_event))) {
		ret = -ENOMEM;
	}
	up(&sem_ambient_list);

	return ret;
}

static	unsigned int ambient_poll(struct file *filep, struct poll_table_struct *wait)
{
	unsigned int mask;
	struct alsps_stat *ptr;

	mask = 0;
	ptr = (struct alsps_stat *)filep->private_data;

	poll_wait(filep, &alsps->ambient_workq_head, wait);
	if (test_bit(ALSPS_AMBIENT_FIRST_POLL, &ptr->status)) {
		mask |= (POLLIN | POLLRDNORM);
		clear_bit(ALSPS_AMBIENT_FIRST_POLL, &ptr->status);
	}
	if (test_bit(ALSPS_AMBIENT_DATA_READY, &ptr->status)) {
		mask |= (POLLIN | POLLRDNORM);
	}

	return mask;
}

static int ambient_open(struct inode * inode, struct file * filep)
{
	struct alsps_stat *ptr;

	ptr = kzalloc(sizeof(struct alsps_stat), GFP_KERNEL);
	if (ptr == NULL) {
		printk("alsps: ambient open kzalloc failed!\n");
		return -ENOMEM;
	}

	clear_bit(ALSPS_AMBIENT_DATA_READY, &ptr->status);
	set_bit(ALSPS_AMBIENT_FIRST_POLL, &ptr->status);
	filep->private_data = ptr;

	down(&sem_ambient_list);
	list_add(&ptr->list, &ambient_process_list);
	++alsps->ambient_count;
	printk(KERN_DEBUG "alsps->ambient_count = %d\n", alsps->ambient_count);	
	up(&sem_ambient_list);

	return 0;
}

static int ambient_close(struct inode * inode, struct file *filep)
{
	struct alsps_stat *ptr;

	ptr = (struct alsps_stat *)filep->private_data;
	down(&sem_ambient_list);
	list_del(&ptr->list);
	kfree(ptr);
	up(&sem_ambient_list);

	if (alsps->ambient_count > 0) {
		--alsps->ambient_count;
	}

	if (alsps->ambient_count == 0) {
		micco_enable_adc_irq(alsps->adc_event, 0);
		micco_auto_adc_poweroff(alsps->adc_event);

		alsps->alsps_switch &= ~AMBIENT_ENABLE;
		alsps->setpower(alsps->alsps_switch);	
	}

	return 0;
}

/*
 * sysfs operations
 */ 
static int alsps_get_property(struct sensors_dev *sdev,
			      enum sensors_property property,
			      union sensors_propval *val)
{
	int value = 0;

	switch (property) {
	case SENSORS_PROP_INTERVAL:
		if (sdev->id & SENSORS_LIGHT)
			val->intval = alsps->ambient_interval;	
		else
			val->intval = 0;
		break;

	case SENSORS_PROP_MAXRANGE:
		if (sdev->id & SENSORS_LIGHT)
			val->intval = 15000;	
		else
			val->intval = 1;
		break;

	case SENSORS_PROP_RESOLUTION:
		if (sdev->id & SENSORS_LIGHT)
			val->intval = 150;	
		else
			val->intval = 1;
		break;

	case SENSORS_PROP_VERSION:
		val->intval = 1;	
		break;

	case SENSORS_PROP_CURRENT:
		if (sdev->id & SENSORS_LIGHT)
			val->intval = 3;	
		else
			val->intval = 5;
		break;

	case SENSORS_PROP_SWITCH:
		if (sdev->id & SENSORS_PROXIMITY) {
			value = PROXIMITY_ENABLE;
		}	
		else if (sdev->id & SENSORS_LIGHT) {
			value = AMBIENT_ENABLE;	
		}
		else {
			printk("alsps: invailed sensor");
			return -EINVAL;
		}
		val->intval = (alsps->alsps_switch & value) ? 1 : 0;
		break;

	case SENSORS_PROP_VENDOR:
		sprintf(val->strval, "Capella");
		break;

	default:
		return -EINVAL;
	}
	return 0;
}

static int alsps_put_property(struct sensors_dev *sdev,
			      enum sensors_property property,
			      union sensors_propval *val)
{
	int value = 0;
	unsigned char adc;

	switch (property) {
	case SENSORS_PROP_SWITCH:

		if (sdev->id & SENSORS_PROXIMITY) {
			value = PROXIMITY_ENABLE;
		}	
		else if (sdev->id & SENSORS_LIGHT) {
			value = AMBIENT_ENABLE;	
		}
		else {
			printk("alsps: invailed sensor\n");
			return -EINVAL;
		}
		if (val->intval) {
			alsps->alsps_switch |= value;
			alsps->setpower(alsps->alsps_switch);	
			msleep(50);
			if (sdev->id & SENSORS_LIGHT) {
				/**
				 * cm3601 sampling timing is 300ms, freqeunctly query 
				 * the ADC_IN5 < 300ms is not suitable.
				 */
				msleep(300);

				/* init adc */
				micco_auto_adc_poweron(alsps->adc_event);
				micco_auto_adc_read(alsps->adc_event, &adc);
				micco_auto_adc_set_trigger(alsps->adc_event, 
					   adc < (255 - AMBITNT_ADC_THRESHOLD) ? (adc + AMBITNT_ADC_THRESHOLD) : 255, 
					   adc > AMBITNT_ADC_THRESHOLD ? (adc - AMBITNT_ADC_THRESHOLD) : 0);
				previous_lumen = current_lumen = ambient_get_lum(adc);

				/* Enable the ADC IRQ here */
				micco_auto_adc_enable(alsps->adc_event, 1);
				micco_enable_adc_irq(alsps->adc_event, 1);
			}

		} else {
			if ((alsps->alsps_switch & AMBIENT_ENABLE)
					&& (sdev->id & SENSORS_LIGHT)) {

				micco_enable_adc_irq(alsps->adc_event, 0);
				micco_auto_adc_enable(alsps->adc_event, 0);
			}

			alsps->alsps_switch &= ~value;
			alsps->setpower(alsps->alsps_switch);	
		}
		break;

	case SENSORS_PROP_INTERVAL:

		if ((sdev->id & SENSORS_LIGHT) && 
				(val->intval > DEFAULT_AMBIENT_INTERVAL)) {
			down(&alsps->sem_ambient_interval);
			alsps->ambient_interval = val->intval;
			up(&alsps->sem_ambient_interval);
		}
		break;

	default:
		return -EINVAL;
	}
	return 0;
}

static enum sensors_property alsps_properties[] = {
	SENSORS_PROP_INTERVAL,
	SENSORS_PROP_MAXRANGE,		/* read only */
	SENSORS_PROP_RESOLUTION,	/* read only */
	SENSORS_PROP_VERSION,		/* read only */ 
	SENSORS_PROP_CURRENT,		/* read only */
	SENSORS_PROP_SWITCH,
	SENSORS_PROP_VENDOR,		/* read only */
};

static struct file_operations proximity_fops = {
	.owner = THIS_MODULE,
	.open = proximity_open,
	.read = proximity_read,
	.poll = proximity_poll,
	.release = proximity_close,
};

static struct file_operations ambient_fops = {
	.owner = THIS_MODULE,
	.open = ambient_open,
	.read = ambient_read,
	.poll = ambient_poll,
	.release = ambient_close,
};

static struct sensors_dev proximity_dev = {
	.id = SENSORS_PROXIMITY,
	.name = "psensor",
	.num_properties = ARRAY_SIZE(alsps_properties),
	.properties = alsps_properties,
	.get_property = alsps_get_property,
	.put_property = alsps_put_property,
	.fops =	&proximity_fops,
};

static struct sensors_dev ambient_dev = {
	.id = SENSORS_LIGHT,
	.name = "lsensor",
	.num_properties = ARRAY_SIZE(alsps_properties),
	.properties = alsps_properties,
	.get_property = alsps_get_property,
	.put_property = alsps_put_property,
	.fops =	&ambient_fops,
};

#ifdef CONFIG_PROC_FS
static int show_stat(struct seq_file *f, void *v)
{
	u8 adc;
	int alsps_gpios[3];
	int ret;

	if (!alsps) {
		return -1;
	}

	if (alsps->alsps_switch) {
		seq_printf(f, "android sensors software interfaces\n");
		seq_printf(f, "proximity sensor status: %s\n", 
				(alsps->alsps_switch & PROXIMITY_ENABLE) ? 
				"on" : "off");
		seq_printf(f, "light sensor: %s\n",  
				(alsps->alsps_switch & AMBIENT_ENABLE) ? 
				"on" : "off");
		seq_printf(f, "proximity output: %s\n", 
				(alsps->get_gpio_value()) ? "high" : "low");
	}
	
	if (alsps->chip_verify) {
		ret = alsps->chip_verify(&alsps_gpios[0], 3);
		if (ret < 0) {
			return -1;
		}
		seq_printf(f, "alsps chip status\n");
		seq_printf(f, "proximity switch: pin value: %d, status: %s\n", 
				alsps_gpios[0],
				alsps_gpios[0] ? "off" : "on");
		seq_printf(f, "proximity output: pin value: %d, status: %s\n", 
				alsps_gpios[1],
				alsps_gpios[1] ? "high" : "low");
		seq_printf(f, "ambient switch  : pin value: %d, status: %s\n", 
				alsps_gpios[2],
				alsps_gpios[2] ? "off" : "on");
	
		adc = 0;
		ret = micco_auto_adc_read(alsps->adc_event, &adc);
		if (ret < 0) {
			seq_printf(f, "alsps: get adc value from micco failed!\n");
		} else {
			seq_printf(f, "alsps: light sensor adc output: %d\n", adc);
		}
	}

	return 0;
}

static int proc_open(struct inode *inode, struct file *file)
{
	return single_open(file, &show_stat, NULL);
}

static struct file_operations proc_ops = {
	.owner = THIS_MODULE,
	.open = proc_open,
	.read = seq_read,
	.llseek = seq_lseek,
	.release = single_release,
};
#endif

static void alsps_early_suspend(struct early_suspend *h)
{
	alsps->alsps_suspend = 1;

	if (alsps->alsps_switch & AMBIENT_ENABLE) {
		micco_enable_adc_irq(alsps->adc_event, 0);
		micco_auto_adc_enable(alsps->adc_event, 0);
	}
	printk(KERN_DEBUG "light sensor enter suspend ok\n");
}

static void alsps_late_resume(struct early_suspend *h)
{
	if (alsps->alsps_switch & AMBIENT_ENABLE) {
		previous_lumen = 0;
		micco_auto_adc_enable(alsps->adc_event, 1);
		micco_enable_adc_irq(alsps->adc_event, 1);
		ambient_adc_interrupt(alsps->adc_event);
	}
	alsps->alsps_suspend = 0;
	printk(KERN_DEBUG "light sensor resume ok\n");
}

static struct early_suspend alsps_early_suspend_desc = {
	.level = 1,
	.suspend = alsps_early_suspend,
	.resume = alsps_late_resume,
};

static int __devinit alsps_probe(struct platform_device *pdev)
{
	struct alsps_platform_data *pdata;
	int ret;

#ifdef CONFIG_PROC_FS
	struct proc_dir_entry *entry;
#endif
	pdata = pdev->dev.platform_data;
        if (!pdata) {
                dev_err(&pdev->dev,"platform data missing\n");
                return -ENODEV;
        }

	alsps = (struct alsps_struct *)kzalloc(sizeof(struct alsps_struct), 
			GFP_KERNEL);
	if (!alsps) {
		printk("alsps: kzalloc memory failed!\n");
		return -1;
	}

	init_MUTEX(&sem_ambient_list);
	init_MUTEX(&alsps->sem_ambient_interval);
	init_MUTEX(&sem_proximity_list);
	init_waitqueue_head(&alsps->proximity_workq_head);
	init_waitqueue_head(&alsps->ambient_workq_head);
	INIT_DELAYED_WORK(&alsps->ambient_workq, ambient_delayed_worker);

	alsps->proximity_count = 0;
	alsps->ambient_count = 0;
	alsps->ambient_interval = DEFAULT_AMBIENT_INTERVAL; 
	alsps->alsps_switch = 0;
	alsps->alsps_suspend = 0;

	alsps->irq = pdata->irq;
	alsps->adc_event = pdata->adc_event;
	alsps->init = pdata->init;
	alsps->chip_verify = pdata->chip_verify;

	alsps->get_gpio_value = pdata->get_gpio_value;
	if (!alsps->get_gpio_value) {
		printk("alsps: get_gpio_value no implement\n");
		goto failed;
	}

	alsps->setpower = pdata->setpower;
	if (!alsps->setpower) {
		printk("alsps: setpower no implement\n");
		goto failed;
	}

	if (alsps->init) {
		ret =alsps->init(); 
		if (ret) {
			printk("alsps hardware init failed\n");
			goto failed;
		}
	}

	/*set irq*/
	ret = request_irq(alsps->irq, proximity_isr,
			  IRQF_TRIGGER_RISING|IRQF_TRIGGER_FALLING, 
			  "alsps", NULL);
	if (ret) {
		printk("alsps: proximity request irq failed!\n");
		goto failed;
	}

	ret = pmic_callback_register(alsps->adc_event, ambient_adc_interrupt);
	if (ret < 0)
		goto pmic_failed;

	ret = sensors_dev_register(&proximity_dev);
	if (ret) {
		printk(KERN_ERR "alsps: proximity unable to register device.\n");
		goto failed_register_dev_proximity;
	}
	ret = sensors_dev_register(&ambient_dev);
	if (ret) {
		printk(KERN_ERR "alsps: ambient unable to register device.\n");
		goto failed_register_dev_ambient;
	}
	dev_set_drvdata(&pdev->dev, alsps);

	register_early_suspend(&alsps_early_suspend_desc);

#ifdef CONFIG_PROC_FS
	entry = create_proc_entry("driver/alsps",0,NULL);
	if (entry) {
		entry->proc_fops = &proc_ops;
	}
#endif

	return 0;

failed_register_dev_ambient:
	sensors_dev_unregister(&proximity_dev);
failed_register_dev_proximity:
	pmic_callback_unregister(alsps->adc_event, ambient_adc_interrupt);
pmic_failed:
	free_irq(alsps->irq, NULL);
failed:
	kfree(alsps);
	return -1;
}

static int __devexit alsps_remove(struct platform_device *pdev)
{
	if (alsps) {
		free_irq(alsps->irq, NULL);
		pmic_callback_unregister(alsps->adc_event, ambient_adc_interrupt);
	}

#ifdef CONFIG_PROC_FS
	remove_proc_entry("alsps", NULL);
#endif
	sensors_dev_unregister(&ambient_dev);
	sensors_dev_unregister(&proximity_dev);
	unregister_early_suspend(&alsps_early_suspend_desc);
	memset(alsps, 0, sizeof(struct alsps_struct));
	return 0;
}

static struct platform_driver alsps_device_driver = {
	.probe = alsps_probe,
	.remove = alsps_remove,
	.driver = {
		.name = "alsps",
	},
};

static int __init alsps_init(void)
{
	wake_lock_init(&alsps_suspend_lock, WAKE_LOCK_SUSPEND, "alsps");
	return platform_driver_register(&alsps_device_driver);
}

static void  __exit alsps_exit(void)
{
	platform_driver_unregister(&alsps_device_driver);
	wake_lock_destroy(&alsps_suspend_lock);
}

module_init(alsps_init);
module_exit(alsps_exit);
MODULE_AUTHOR("Ma Yilei <yilei.ma@borqs.com>");
MODULE_DESCRIPTION("Capella CM36xx short distance proximity and ambient light sensor");
MODULE_LICENSE("GPL");
