/*
 *  drivers/switch/switch_usbuart.c
 *
 * Copyright (C) 2009 BORQS, Inc.
 * Author: Emichael Li <emichael.li@borqs.com>
 *
 * This software is licensed under the terms of the GNU General Public
 * License version 2, as published by the Free Software Foundation, and
 * may be copied, distributed, and modified under those terms.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
*/

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/platform_device.h>
#include <linux/switch.h>
#include <linux/workqueue.h>
#include <linux/switch_usbuart.h>

struct usbuart_switch_data {
	struct switch_dev sdev;
	struct usbuart_switch_platform_data *pdata;

	enum usbuart_state state;

	struct work_struct work;
};

static struct usbuart_switch_data *udata = NULL;

static void usbuart_switch_work(struct work_struct *work)
{
	struct usbuart_switch_platform_data *pdata;

	if (udata == NULL)
		return;

	pdata = udata->pdata;
	BUG_ON(pdata == NULL);

	BUG_ON(pdata->get_state == NULL);
	udata->state = pdata->get_state();

	switch_set_state(&udata->sdev, udata->state);
}

int usbuart_set_state(enum usbuart_event event, int on)
{
	struct usbuart_switch_platform_data *pdata;

	if (udata == NULL)
		return 0;

	pdata = udata->pdata;
	BUG_ON(pdata == NULL);

	if (pdata->set_state)
		pdata->set_state(event, on);

	schedule_work(&udata->work);

	return 0;
}

EXPORT_SYMBOL_GPL(usbuart_set_state);

static ssize_t switch_usbuart_print_state(struct switch_dev *sdev, char *buf)
{
	if (udata == NULL)
		return 0;

	if (udata->state == USBUART_AP_UART)
		return sprintf(buf, "%s\n", "apuart");

	if (udata->state == USBUART_AP_USB)
		return sprintf(buf, "%s\n", "apusb");

	BUG();
	return 0;
}

static ssize_t switch_usbuart_switch_state(struct switch_dev *sdev, const char *buf, ssize_t n)
{
	if (udata == NULL)
		return n;
	
	if (!strncmp(buf, "apuart", strlen("apuart"))) {
		usbuart_set_state(USBUART_DBG, 0);
		return n;
	}

	if (!strncmp(buf, "apusb", strlen("apusb"))) {
		usbuart_set_state(USBUART_DBG, 1);
		return n;
	}

	return n;
}

static int usbuart_switch_probe(struct platform_device *pdev)
{
	struct usbuart_switch_platform_data *pdata = pdev->dev.platform_data;
	struct usbuart_switch_data *switch_data;
	int ret = 0;

	if (!pdata)
		return -EBUSY;

	switch_data = kzalloc(sizeof(struct usbuart_switch_data), GFP_KERNEL);
	if (!switch_data)
		return -ENOMEM;

	switch_data->pdata = pdata;
	switch_data->sdev.name = pdata->name;
	switch_data->sdev.print_state = switch_usbuart_print_state;
	switch_data->sdev.switch_state = switch_usbuart_switch_state;

	ret = switch_dev_register(&switch_data->sdev);
	if (ret < 0)
		goto err_switch_dev_register;

	INIT_WORK(&switch_data->work, usbuart_switch_work);

	/* Perform initial detection */
	udata = switch_data;
	usbuart_switch_work(&switch_data->work);
	return 0;

err_switch_dev_register:
	kfree(switch_data);
	return ret;
}

static int __devexit usbuart_switch_remove(struct platform_device *pdev)
{
	struct usbuart_switch_data *switch_data = platform_get_drvdata(pdev);

	cancel_work_sync(&switch_data->work);
	switch_dev_unregister(&switch_data->sdev);
	udata = NULL;
	kfree(switch_data);

	return 0;
}

static struct platform_driver usbuart_switch_driver = {
	.probe = usbuart_switch_probe,
	.remove = __devexit_p(usbuart_switch_remove),
	.driver = {
		   .name = "switch-usbuart",
		   .owner = THIS_MODULE,
		   },
};

static int __init usbuart_switch_init(void)
{
	return platform_driver_register(&usbuart_switch_driver);
}

static void __exit usbuart_switch_exit(void)
{
	platform_driver_unregister(&usbuart_switch_driver);
}

module_init(usbuart_switch_init);
module_exit(usbuart_switch_exit);

MODULE_AUTHOR("Emichael Li <emichael.li@borqs.com>");
MODULE_DESCRIPTION("USBUART Switch driver");
MODULE_LICENSE("GPL");
