/*
   * g_mtp.c -- USB gadget mtp driver
   *
   *
   * This code is based in part on the Gadget Zero driver, which
   * is Copyright (C) 2003 by David Brownell, all rights reserved.
   * This software is distributed under the terms of the GNU General
   * Public License ("GPL") as published by the Free Software Foundation,
   * either version 2 of that License or (at your option) any later version.
   *
   */
#include <linux/module.h>
#include <linux/slab.h>
#include <linux/errno.h>
#include <linux/list.h>
#include <linux/interrupt.h>
#include <linux/utsname.h>
#include <linux/device.h>
#include <linux/moduleparam.h>
#include <linux/fs.h>
#include <linux/proc_fs.h>
#include <linux/poll.h>
#include <linux/types.h>
#include <linux/ctype.h>
#include <linux/cdev.h>
#include <linux/workqueue.h>

#include <linux/irq.h>
#include <asm/system.h>

#include <linux/usb/ch9.h>
#include <linux/usb/gadget.h>

#include "gadget_chips.h"
#include "mtp.h"

#define DRIVER_DESC		"Gadget MTP"
#define DRIVER_NAME		"g_mtp"
#define DRIVER_VERSION		"13 Nov. 2008"

#define MTP_CLASS_IMAGE_INTERFACE                  0x06
#define MTP_SUBCLASS_STILL_IMAGE_CAPTURE_DEVICE    0x01
#define MTP_PROTOCOL_CODE                          0x01

#define MTP_EXT_STRING_ID			   0x3EE
#define MTP_VENDOR_CODE			           0x1C

#if 1
char *mtp_ext_desc ="2800000000010400010000000000000000014D545000000000003030000000000000000000000000";
char *mtp_ext_desc1="2800000000010400010000000000000000014D545000000000003300000000000000000000000000";
char *ext_str_desc="12034D005300460054003100300030001C00";
char *mtp_status ="240001200000000000000000000000000000000000000000000000000000000000000000";
char *mtp_status1 ="04000120";
char *serial_default="419464B90D2508";
#endif

static const char driver_desc [] = DRIVER_DESC;
static const char shortname[] = DRIVER_NAME;
static const char longname[] = DRIVER_DESC;


/*-------------------------------------------------------------------------*/

struct mtp_dev {
	spinlock_t		lock;		
	//spinlock_t              ctl_lock;
	/* lock buffer lists during read/write calls */
	spinlock_t		lock_mtp_io;
	struct usb_gadget	*gadget;
	struct usb_request	*req;		/* for control responses */
	u8			config;
	s8			interface;
	struct usb_ep		*in_ep, *out_ep, *intr_in;
	const struct usb_endpoint_descriptor
				*in, *out,*intr;
	 /* device request receive */
   	unsigned char *		req_recv_buf;
   	unsigned char *		req_recv_start;
   	unsigned char *		req_recv_end;
   	wait_queue_head_t 	req_recv_wait;
  	int 			req_ms_vendor_flag;

   	/* device request send */
  	int 			req_send_len;
  	wait_queue_head_t 	req_send_wait;
	struct list_head	rx_reqs;	/* List of free RX structs */
	int			rx_reqs_active;	/* num Active RX xfers */
	struct list_head	rx_buffers;	/* List of completed xfers */
	/* wait until there is data to be read. */
	wait_queue_head_t	rx_wait;
	struct list_head	tx_reqs;	/* List of free TX structs */
	int			tx_reqs_active; /* num Active TX xfers */
	/* Wait until there are write buffers available to use. */
	wait_queue_head_t	tx_wait;
	/* Wait until all write buffers have been sent. */
	wait_queue_head_t	tx_flush_wait;
	struct usb_request	*current_rx_req;
	size_t			current_rx_bytes;
	u8			*current_rx_buf;
	u8			reset_mtp;
	struct cdev		mtp_cdev;
	struct device		*pdev;
	u8			mtp_cdev_open;
	u8                      mtp_proc_open;
};
static	void mtp_dev_reset(struct mtp_dev *dev);
static  void rx_complete(struct usb_ep *ep, struct usb_request *req);
static struct mtp_dev *usb_mtp_gadget;

static dev_t g_mtp_devno;

static struct class *usb_gadget_class;
/*-------------------------------------------------------------------------*/

/* DO NOT REUSE THESE IDs with a protocol-incompatible driver!!  Ever!!
 * Instead:  allocate your own, using normal USB-IF procedures.
 */

/* Thanks to NetChip Technologies for donating this product ID.
 */
#define MTP_VENDOR_NUM		CONFIG_USB_G_MTP_VID
#define MTP_PRODUCT_NUM		CONFIG_USB_G_MTP_PID
//#define DRIVER_BCD_DEV		0x0101	// Linux-USB File-backed Storage Gadget/

/* Some systems will want different product identifers published in the
 * device descriptor, either numbers or strings or both.  These string
 * parameters are in UTF-8 (superset of ASCII's 7 bit characters).
 */

static ushort __initdata idVendor;
module_param(idVendor, ushort, S_IRUGO);
MODULE_PARM_DESC(idVendor, "USB Vendor ID");

static ushort __initdata idProduct;
module_param(idProduct, ushort, S_IRUGO);
MODULE_PARM_DESC(idProduct, "USB Product ID");

static ushort __initdata bcdDevice;
module_param(bcdDevice, ushort, S_IRUGO);
MODULE_PARM_DESC(bcdDevice, "USB Device version (BCD)");

static char *__initdata iManufacturer;
module_param(iManufacturer, charp, S_IRUGO);
MODULE_PARM_DESC(iManufacturer, "USB Manufacturer string");

static char *__initdata iProduct;
module_param(iProduct, charp, S_IRUGO);
MODULE_PARM_DESC(iProduct, "USB Product string");

static char *__initdata iSerialNum;
module_param(iSerialNum, charp, S_IRUGO);
MODULE_PARM_DESC(iSerialNum, "SerialNumber");


static unsigned bufnum = 16;
module_param(bufnum, uint, S_IRUGO|S_IWUSR);

static char *ext_desc = NULL;
module_param(ext_desc, charp, S_IRUGO);
MODULE_PARM_DESC(ext_desc, "MTP ext_desc");

static char *ext_string = NULL;
module_param(ext_string, charp, S_IRUGO);
MODULE_PARM_DESC(ext_string, "MTP ext_string");

//#define MTP_DEBUG

#ifdef MTP_DEBUG

#define MTP_DBG(args...) printk(KERN_INFO  args)	

#else

#define MTP_DBG(args...) printk(KERN_DEBUG args)

#endif

#define MTP_PRINT(args...) printk(KERN_INFO args)

#define MTP_ERR(args...) printk(KERN_ERR args)

#define LEN	bufnum



/*-------------------------------------------------------------------------*/

/*
 * Kbuild is not very cooperative with respect to linking separately
 * compiled library objects into one module.  So for now we won't use
 * separate compilation ... ensuring init/exit sections work to shrink
 * the runtime footprint, and giving us at least some parts of what
 * a "gcc --combine ... part1.c part2.c part3.c ... " build would.
 */
#include "usbstring.c"
#include "config.c"
#include "epautoconf.c"


/* USB DRIVER HOOKUP (to the hardware driver, below us), mostly
 * ep0 implementation:  descriptors, config management, setup().
 * also optional class-specific notification interrupt transfer.
 */

/*
 * mtp need ext string for windows identify
 */

#define STRING_MANUFACTURER		1
#define STRING_PRODUCT			2
#define STRING_SERIALNUM		3
#define STRING_CONFIG			4
#define STRING_INTERFACE		5
#define STRING_EXT_STRING       	0xee

/* holds our biggest descriptor */
#define USB_DESC_BUFSIZE		256
#define USB_BUFSIZE			1024*16

/* This device advertises one configuration. */
#define DEV_CONFIG_VALUE		1
#define	MTP_INTERFACE			0

#define HS_BULK_MAXPACKET_SIZE          512//HISPPE
#define FS_BULK_MAXPACKET_SIZE          64 //LOWSPPE

static struct usb_device_descriptor device_desc = {
	.bLength =		sizeof device_desc,
	.bDescriptorType =	USB_DT_DEVICE,
	.bcdUSB =		__constant_cpu_to_le16(0x0200),
	.bDeviceClass =		USB_CLASS_PER_INTERFACE,
	.bDeviceSubClass =	0,
	.bDeviceProtocol =	0,
	.idVendor =		__constant_cpu_to_le16(MTP_VENDOR_NUM),
	.idProduct =		__constant_cpu_to_le16(MTP_PRODUCT_NUM),
	.iManufacturer =	STRING_MANUFACTURER,
	.iProduct =		STRING_PRODUCT,
	.iSerialNumber =	STRING_SERIALNUM,
	.bNumConfigurations =	1
};

static struct usb_otg_descriptor otg_desc = {
	.bLength =		sizeof otg_desc,
	.bDescriptorType =	USB_DT_OTG,
	.bmAttributes =		USB_OTG_SRP
};

static struct usb_config_descriptor config_desc = {
	.bLength =		sizeof config_desc,
	.bDescriptorType =	USB_DT_CONFIG,

	/* compute wTotalLength on the fly */
	.bNumInterfaces =	1,
	.bConfigurationValue =	DEV_CONFIG_VALUE,
	.iConfiguration =	STRING_CONFIG,
	.bmAttributes =		USB_CONFIG_ATT_ONE,
	.bMaxPower =		500/2	/* Self-Powered */
};

static struct usb_interface_descriptor intf_desc = {
	.bLength =		sizeof intf_desc,
	.bDescriptorType =	USB_DT_INTERFACE,
	.bInterfaceNumber =	0,
	.bNumEndpoints =	3,
	.bInterfaceClass =	MTP_CLASS_IMAGE_INTERFACE,
	.bInterfaceSubClass =	MTP_SUBCLASS_STILL_IMAGE_CAPTURE_DEVICE,	/* mtp Sub-Class */
	.bInterfaceProtocol =	MTP_PROTOCOL_CODE,	/* Bi-Directional */
	.iInterface =		STRING_INTERFACE
};

static struct usb_endpoint_descriptor fs_ep_in_desc = {
	.bLength =		USB_DT_ENDPOINT_SIZE,
	.bDescriptorType =	USB_DT_ENDPOINT,
	.bEndpointAddress =	USB_DIR_IN,
	.wMaxPacketSize =       __constant_cpu_to_le16(FS_BULK_MAXPACKET_SIZE),
	.bmAttributes =		USB_ENDPOINT_XFER_BULK
};

static struct usb_endpoint_descriptor fs_ep_out_desc = {
	.bLength =		USB_DT_ENDPOINT_SIZE,
	.bDescriptorType =	USB_DT_ENDPOINT,
	.bEndpointAddress =	USB_DIR_OUT,
	.wMaxPacketSize =       __constant_cpu_to_le16(FS_BULK_MAXPACKET_SIZE),
	.bmAttributes =		USB_ENDPOINT_XFER_BULK
};
static struct usb_endpoint_descriptor fs_intr_in_desc = {
	.bLength =		USB_DT_ENDPOINT_SIZE,
	.bDescriptorType =	USB_DT_ENDPOINT,

	.bEndpointAddress =	USB_DIR_IN,
	.bmAttributes =		USB_ENDPOINT_XFER_INT,
	.wMaxPacketSize =	__constant_cpu_to_le16(16),
	.bInterval =		32,	// frames -> 32 ms
};
static const struct usb_descriptor_header *fs_mtp_function [] = {
	(struct usb_descriptor_header *) &otg_desc,
	(struct usb_descriptor_header *) &intf_desc,
	(struct usb_descriptor_header *) &fs_ep_in_desc,
	(struct usb_descriptor_header *) &fs_ep_out_desc,
	(struct usb_descriptor_header *) &fs_intr_in_desc,
	NULL
};

#ifdef	CONFIG_USB_GADGET_DUALSPEED

/*
 * usb 2.0 devices need to expose both high speed and full speed
 * descriptors, unless they only run at full speed.
 */

static struct usb_endpoint_descriptor hs_ep_in_desc = {
	.bLength =		USB_DT_ENDPOINT_SIZE,
	.bDescriptorType =	USB_DT_ENDPOINT,
	.bmAttributes =		USB_ENDPOINT_XFER_BULK,
	.wMaxPacketSize =	__constant_cpu_to_le16(HS_BULK_MAXPACKET_SIZE)
};

static struct usb_endpoint_descriptor hs_ep_out_desc = {
	.bLength =		USB_DT_ENDPOINT_SIZE,
	.bDescriptorType =	USB_DT_ENDPOINT,
	.bmAttributes =		USB_ENDPOINT_XFER_BULK,
	.wMaxPacketSize =	__constant_cpu_to_le16(HS_BULK_MAXPACKET_SIZE)
};
static struct usb_endpoint_descriptor hs_intr_in_desc = {
	.bLength =		USB_DT_ENDPOINT_SIZE,
	.bDescriptorType =	USB_DT_ENDPOINT,

	/* bEndpointAddress copied from fs_intr_in_desc during fsg_bind() */
	.bmAttributes =		USB_ENDPOINT_XFER_INT,
	.wMaxPacketSize =	__constant_cpu_to_le16(64),
	.bInterval =		0x0a,	// 2**(9-1) = 256 uframes -> 32 ms
};

static struct usb_qualifier_descriptor dev_qualifier = {
	.bLength =		sizeof dev_qualifier,
	.bDescriptorType =	USB_DT_DEVICE_QUALIFIER,
	.bcdUSB =		__constant_cpu_to_le16(0x0200),
	.bDeviceClass =		USB_CLASS_PER_INTERFACE,
	.bNumConfigurations =	1
};

static const struct usb_descriptor_header *hs_mtp_function [] = {
	(struct usb_descriptor_header *) &otg_desc,
	(struct usb_descriptor_header *) &intf_desc,
	(struct usb_descriptor_header *) &hs_ep_in_desc,
	(struct usb_descriptor_header *) &hs_ep_out_desc,
	(struct usb_descriptor_header *) &hs_intr_in_desc,
	NULL
};

/* maxpacket and other transfer characteristics vary by speed. */
#define ep_desc(g, hs, fs) (((g)->speed == USB_SPEED_HIGH)?(hs):(fs))

#else

/* if there's no high speed support, maxpacket doesn't change. */
#define ep_desc(g, hs, fs) (((void)(g)), (fs))

#endif	/* !CONFIG_USB_GADGET_DUALSPEED */

/*-------------------------------------------------------------------------*/

/* descriptors that are built on-demand */

static char				manufacturer[30],productor [50];
static char				product_desc [40] = DRIVER_DESC;
static char				serial[40];


/* static strings, in UTF-8 */
static struct usb_string		strings [] = {
/*	{ STRING_MANUFACTURER,	manufacturer, },
	{ STRING_PRODUCT,	product_desc, },
	{ STRING_SERIALNUM,	serial_num, },
	{  }	*/	/* end of list */
	{STRING_MANUFACTURER,	CONFIG_USB_DRIVER_MANUFACTURER},
	{STRING_PRODUCT,	productor},
	{STRING_SERIALNUM,	serial},
	{STRING_CONFIG,		"mtp"},
	{STRING_INTERFACE,	"MTP"},
	{STRING_EXT_STRING,	"1"},
	{}
};

/* data for BH of device_request */
struct mtp_dev_req_bh_private{
    struct work_struct    bh;
    unsigned char  *data;
    void *mtp_data;
    int data_len;
    bool pending_data;
};
/* for ep0 scheduling */
static unsigned char* pending_ep0_data = NULL;
static int pending_ep0_data_len = 0;
static bool ready_for_ep0_data = 0;
//typedef struct MTP_DEV_REQ_BH mtp_dev_req_bh_private;

static void *mtp_str2descriptor(u8 *buf, int buf_len, const char *str)
{
    int ext_len = strlen(str)/2;
    u8 val;
    int i;
    int j;

    if (ext_len > buf_len)
    {
        ext_len = buf_len;
    }
        
    MTP_DBG( "mtp_str2descriptor, ext_len=%d, str=%s", ext_len, str);

    for (i = 0; i < ext_len; i++)
    {
        val = 0;
        for (j = 2*i; j <= 2*i+1; j++)
        {
            val <<= 4;
            val += (str[j] >= 'A') ? str[j] - 'A' + 10 : str[j] - '0';
        }
        
        buf[i] = val;
        /* TRACE_MSG2(MTP, "mtp_str2descriptor, desc[%d]=0x%x", i, buf[i]); */
    }

    return buf;
}


static struct usb_gadget_strings	stringtab = {
	.language	= 0x0409,	/* en-us */
	.strings	= strings,
};

/*-------------------------------------------------------------------------*/

static struct usb_request *
mtp_req_alloc(struct usb_ep *ep, unsigned len, gfp_t gfp_flags)
{
	struct usb_request	*req;

	req = usb_ep_alloc_request(ep, gfp_flags);

	if (req != NULL) {
		req->length = len;
		req->buf = kmalloc(len, gfp_flags);
		if (req->buf == NULL) {
			usb_ep_free_request(ep, req);
			return NULL;
		}
	}

	return req;
}

static void
mtp_req_free(struct usb_ep *ep, struct usb_request *req)
{
	if (ep != NULL && req != NULL) {
		kfree(req->buf);
		usb_ep_free_request(ep, req);
	}
}

/*-------------------------------------------------------------------------*/

static void prepare_ep_queue(struct mtp_dev *mtp)
{
	struct usb_request              *req;
        struct mtp_dev *dev=mtp;
        int error;

     	while(likely(!list_empty(&dev->rx_reqs))){

                req = list_entry(dev->rx_reqs.next,
                                struct usb_request, list);
                MTP_DBG("req->buf12=%p\n",req->buf);

                list_del_init(&req->list);

               /* The USB Host sends us whatever amount of data it wants to
 		* so we always set the length field to the full USB_BUFSIZE.
 		* If the amount of data is more than the read() caller asked
 		* for it will be stored in the request buffer until it is
		* asked for by read().
 		*/
               req->length = USB_BUFSIZE;
               req->complete = rx_complete;

               error = usb_ep_queue(dev->out_ep, req, GFP_ATOMIC);
               if (error) {
                        MTP_ERR("rx submit --> %d\n", error);
                      	list_del_init(&req->list);
			list_add(&req->list, &dev->rx_reqs);
                        break;
                                    }
              else {
                        dev->rx_reqs_active++;
                        MTP_DBG("rx_reqs_active=%d in %s\n",dev->rx_reqs_active,__func__);  
			//list_add(&req->list, &dev->rx_reqs_active);
                   }
	}
}

static void rx_complete(struct usb_ep *ep, struct usb_request *req)
{
	struct mtp_dev	*dev = ep->driver_data;
	int			status = req->status;
	unsigned long		flags;

	spin_lock_irqsave(&dev->lock, flags);
	MTP_DBG("rx_complete and status=%d\n",status);
//    printk(KERN_INFO "rx_complete and status=%d\n",status);
	
	list_del_init(&req->list);

	if(dev->rx_reqs_active > 0)
	dev->rx_reqs_active --;
	
	MTP_DBG("mtp rx_complete and rx_reqs_active=%d\n",dev->rx_reqs_active);

	switch (status) {

	/* normal completion */
	case 0:
		list_add_tail(&req->list, &dev->rx_buffers);
		wake_up_interruptible(&dev->rx_wait);
		MTP_DBG("G_mtp : rx length %d\n", req->actual);
		break;

	/* software-driven interface shutdown */
	case -ECONNRESET:		/* unlink */
	case -ESHUTDOWN:		/* disconnect etc */
//		MTP_ERR("rx shutdown, code %d\n", status);
		list_add(&req->list, &dev->rx_reqs);
		break;

	/* for hardware automagic (such as pxa) */
	case -ECONNABORTED:		/* endpoint reset */
		MTP_DBG("rx %s reset\n", ep->name);
		list_add(&req->list, &dev->rx_reqs);
		break;

	/* data overrun */
	case -EOVERFLOW:
		/* FALLTHROUGH */

	default:
		MTP_DBG("rx status %d\n", status);
		list_add(&req->list, &dev->rx_reqs);
		break;
	}
	spin_unlock_irqrestore(&dev->lock, flags);
}

static void tx_complete(struct usb_ep *ep, struct usb_request *req)
{
	struct mtp_dev	*dev = ep->driver_data;
	MTP_DBG("tx err= %d in %s\n", req->status,__func__);

	if(dev->tx_reqs_active > 0)
	dev->tx_reqs_active --;
	
	MTP_DBG("mtp tx_complete and tx_reqs_active=%d\n",dev->tx_reqs_active);
	switch (req->status) {
	default:
		MTP_DBG("tx err %d\n", req->status);
		/* FALLTHROUGH */
	case -ECONNRESET:		/* unlink */
	case -ESHUTDOWN:		/* disconnect etc */
		break;
	case 0:
		break;
	}

	spin_lock(&dev->lock);
	/* Take the request struct off the active list and put it on the
	 * free list.
	 */
	list_del_init(&req->list);
	list_add(&req->list, &dev->tx_reqs);
	wake_up_interruptible(&dev->tx_wait);
	if (likely(dev->tx_reqs_active == 0))
		wake_up_interruptible(&dev->tx_flush_wait);
	spin_unlock(&dev->lock);

}

static void mtp_event_complete(struct usb_ep *ep, struct usb_request *req)
{
     MTP_DBG("event complete and status= %d in %s\n", req->status,__func__);
	
	switch (req->status) {
        default:
                MTP_DBG("mtp event err %d\n", req->status);
                /* FALLTHROUGH */
        case -ECONNRESET:               /* unlink */
        case -ESHUTDOWN:                /* disconnect etc */
                break;
        case 0:
                break;
        }
}
static void zlp_tx_complete(struct usb_ep *ep, struct usb_request *req)
{
        MTP_DBG("tx err= %d in %s\n", req->status,__func__);
	
	switch (req->status) {
        default:
                MTP_DBG("tx err %d\n", req->status);
                /* FALLTHROUGH */
        case -ECONNRESET:               /* unlink */
        case -ESHUTDOWN:                /* disconnect etc */
                break;
        case 0:
                break;
        }

        /* Take the request struct off the active list and put it on the
 *          * free list.
 *                   */
        //list_del_init(&req->list);
        //list_add(&req->list, &dev->tx_reqs);
        //wake_up_interruptible(&dev->tx_wait);
        //if (likely(list_empty(&dev->tx_reqs_active)))
          //      wake_up_interruptible(&dev->tx_flush_wait);

}
static void sent_zlp_from_ep0_complete(struct usb_ep *ep, struct usb_request *req)
{
        MTP_DBG("sent_zlp_from_ep0_complete and status=%d\n",req->status);
}

/*-------------------------------------------------------------------------*/

static int
mtp_open(struct inode *inode, struct file *fd)
{
	struct mtp_dev	*dev;
	unsigned long		flags;
	int			ret = -EBUSY;

        MTP_DBG("open mtp dev\n");   
	dev = container_of(inode->i_cdev, struct mtp_dev, mtp_cdev);

	spin_lock_irqsave(&dev->lock, flags);

	if (!dev->mtp_cdev_open) {
		dev->mtp_cdev_open = 1;
		fd->private_data = dev;
		ret = 0;
		/* Change the mtp status to show that it's on-line. */
	}

	spin_unlock_irqrestore(&dev->lock, flags);

	MTP_DBG("mtp_open returned %x\n", ret);
#if 0
	return 0;
#else
	return ret;
#endif
}

static int
mtp_close(struct inode *inode, struct file *fd)
{
	struct mtp_dev	*dev = fd->private_data;
	unsigned long		flags;

	spin_lock_irqsave(&dev->lock, flags);
	dev->mtp_cdev_open = 0;
	fd->private_data = NULL;
	/* Change mtp status to show that the mtp is off-line. */
	spin_unlock_irqrestore(&dev->lock, flags);

	MTP_DBG("mtp_close\n");

	return 0;
}

static ssize_t
mtp_read(struct file *fd, char __user *buf, size_t len, loff_t *ptr)
{
	struct mtp_dev		*dev = fd->private_data;
	unsigned long			flags;
	size_t				size;
	size_t				bytes_copied;
	struct usb_request		*req;
	/* This is a pointer to the current USB rx request. */
	struct usb_request		*current_rx_req;
	/* This is the number of bytes in the current rx buffer. */
	size_t				current_rx_bytes;
	/* This is a pointer to the current rx buffer. */
	u8				*current_rx_buf;

	if (len == 0)
		return -EINVAL;

	MTP_DBG("mtp_read trying to read %d bytes\n", (int)len);

	spin_lock(&dev->lock_mtp_io);
	spin_lock_irqsave(&dev->lock, flags);

	/* We will use this flag later to check if a mtp reset happened
 * 	 * after we turn interrupts back on.
 * 	 	 */
	dev->reset_mtp = 0;
	
	prepare_ep_queue(dev);
	bytes_copied = 0;
	current_rx_req = dev->current_rx_req;
	current_rx_bytes = dev->current_rx_bytes;
	current_rx_buf = dev->current_rx_buf;
	dev->current_rx_req = NULL;
	dev->current_rx_bytes = 0;
	dev->current_rx_buf = NULL;

	/* Check if there is any data in the read buffers. Please note that
  	 * current_rx_bytes is the number of bytes in the current rx buffer.
 	 * If it is zero then check if there are any other rx_buffers that
 	 * are on the completed list. We are only out of data if all rx
 	 *buffers are empty.
 	 */
	if ((current_rx_bytes == 0) &&
			(likely(list_empty(&dev->rx_buffers)))) {
		/* Turn interrupts back on before sleeping. */
		spin_unlock_irqrestore(&dev->lock, flags);

		/*
		 * If no data is available check if this is a NON-Blocking
 		 * call or not.
		 */
		if (fd->f_flags & (O_NONBLOCK|O_NDELAY)) {
			spin_unlock(&dev->lock_mtp_io);
			return -EAGAIN;
		}

		/* Sleep until data is available */
		wait_event_interruptible(dev->rx_wait,
				(likely(!list_empty(&dev->rx_buffers))));
		spin_lock_irqsave(&dev->lock, flags);
	}

	/* We have data to return then copy it to the caller's buffer.*/
	while ((current_rx_bytes || likely(!list_empty(&dev->rx_buffers)))
			&& len) {
		if (current_rx_bytes == 0) {
			req = container_of(dev->rx_buffers.next,
					struct usb_request, list);
			list_del_init(&req->list);

			if (req->actual && req->buf) {
				current_rx_req = req;
				current_rx_bytes = req->actual;
				current_rx_buf = req->buf;
			} else {
				list_add(&req->list, &dev->rx_reqs);
				continue;
			}
		}

		/* Don't leave irqs off while doing memory copies */
		spin_unlock_irqrestore(&dev->lock, flags);

		if (len > current_rx_bytes)
			size = current_rx_bytes;
		else
			size = len;

		size -= copy_to_user(buf, current_rx_buf, size);
		bytes_copied += size;
		len -= size;
		buf += size;

		spin_lock_irqsave(&dev->lock, flags);

		/* We've disconnected or reset free the req and buffer */
		if (dev->reset_mtp) {
			printk("reset_mtp=1 in read\n");
			//mtp_req_free(dev->out_ep, current_rx_req);
			spin_unlock_irqrestore(&dev->lock, flags);
			spin_unlock(&dev->lock_mtp_io);
			return -EAGAIN;
		}

		/* If we not returning all the data left in this RX request
  		 * buffer then adjust the amount of data left in the buffer.
 		 * Othewise if we are done with this RX request buffer then
  		 * requeue it to get any incoming data from the USB host.
  		 */
		if (size < current_rx_bytes) {
			current_rx_bytes -= size;
			current_rx_buf += size;
		} else {
			list_add(&current_rx_req->list, &dev->rx_reqs);
			current_rx_bytes = 0;
			current_rx_buf = NULL;
			current_rx_req = NULL;
		}
	}

	dev->current_rx_req = current_rx_req;
	dev->current_rx_bytes = current_rx_bytes;
	dev->current_rx_buf = current_rx_buf;

	spin_unlock_irqrestore(&dev->lock, flags);
	spin_unlock(&dev->lock_mtp_io);

	MTP_DBG("mtp_read returned %d bytes\n", (int)bytes_copied);

	if (bytes_copied)
		return bytes_copied;
	else
		return -EAGAIN;
}

static ssize_t
mtp_write(struct file *fd, const char __user *buf, size_t len, loff_t *ptr)
{
	struct mtp_dev	*dev = fd->private_data;
	unsigned long		flags;
	size_t			size;	/* Amount of data in a TX request. */
	size_t			bytes_copied = 0;
	struct usb_request	*req;
	int 			ret=0;
	MTP_DBG("mtp_write trying to send %d bytes\n", (int)len);
	//buf1[0]="0x20";
	if (len == 0)
		return -EINVAL;
	spin_lock(&dev->lock_mtp_io);
	spin_lock_irqsave(&dev->lock, flags);

	/* Check if a mtp reset happens while we have interrupts on */
	dev->reset_mtp = 0;

	/* Check if there is any available write buffers */
	if (likely(list_empty(&dev->tx_reqs))) {
		/* Turn interrupts back on before sleeping. */
		spin_unlock_irqrestore(&dev->lock, flags);

		/*
		 * If write buffers are available check if this is
		 * a NON-Blocking call or not.
		 */
	
		if (fd->f_flags & (O_NONBLOCK|O_NDELAY)) {
			spin_unlock(&dev->lock_mtp_io);
			return -EAGAIN;
		}

		/* Sleep until a write buffer is available */
		wait_event_interruptible(dev->tx_wait,
				(likely(!list_empty(&dev->tx_reqs))));
		spin_lock_irqsave(&dev->lock, flags);
	}

	while (likely(!list_empty(&dev->tx_reqs)) && len) {

		if (len > USB_BUFSIZE)
			size = USB_BUFSIZE;
		else
			size = len;
		req = container_of(dev->tx_reqs.next, struct usb_request,
				list);
		list_del_init(&req->list);

		MTP_DBG("req buf=%p in mtp write \n",req->buf);
		req->complete = tx_complete;
		req->length = size;

		/* Check if we need to send a zero length packet. */
		//if (len >= size)
			/* They will be more TX requests so no yet. */
		req->zero = 0;
	//	else 
			/* If the data amount is not a multple of the
			 * maxpacket size then send a zero length packet.
	//		 */
	//		req->zero = ((len % dev->in_ep->maxpacket) == 0);

		/* Don't leave irqs off while doing memory copies */
		spin_unlock_irqrestore(&dev->lock, flags);
	//	if(size==1)
	//	memcpy(buf, buf1, size);
		//else 
		//	
		if (copy_from_user(req->buf, buf, size)) {
			list_del_init(&req->list);
			list_add(&req->list, &dev->tx_reqs);
			spin_unlock(&dev->lock_mtp_io);
			MTP_ERR("copy err in %s\n",__func__);
			return bytes_copied;
		}
		bytes_copied += size;
		len -= size;
		buf += size;
		spin_lock_irqsave(&dev->lock, flags);

		/* We've disconnected or reset so free the req and buffer */
		if (dev->reset_mtp) {
			//mtp_req_free(dev->in_ep, req);
			printk("reset_mtp =1 and exit reset_mtp\n");
			spin_unlock_irqrestore(&dev->lock, flags);
			spin_unlock(&dev->lock_mtp_io);
			return -EAGAIN;
		}
		ret = usb_ep_queue(dev->in_ep, req, GFP_ATOMIC);
		if (ret) {
			
			
			list_del_init(&req->list);
			list_add(&req->list, &dev->tx_reqs);
			spin_unlock_irqrestore(&dev->lock, flags);
			spin_unlock(&dev->lock_mtp_io);
			MTP_ERR("queue err in %s and ret=%d\n",__func__,ret);
			return -EAGAIN;
		}

		dev->tx_reqs_active ++;
		MTP_DBG("mtp_write and tx_reqs_active=%d\n",dev->tx_reqs_active);

	}

	spin_unlock_irqrestore(&dev->lock, flags);
	spin_unlock(&dev->lock_mtp_io);

	MTP_DBG("mtp_write sent %d bytes\n", (int)bytes_copied);

	if (bytes_copied) {
		return bytes_copied;
	} else {
		return -EAGAIN;
	}
}

static int
mtp_fsync(struct file *fd, struct dentry *dentry, int datasync)
{
	struct mtp_dev	*dev = fd->private_data;
	unsigned long		flags;

	spin_lock_irqsave(&dev->lock, flags);
	spin_unlock_irqrestore(&dev->lock, flags);

	if (dev->tx_reqs_active > 0) {
                /* Sleep until all data has been sent */
                MTP_DBG("tx_reqs list is empty while sync\n");
                wait_event_interruptible(dev->tx_flush_wait,
		        	(likely(dev->tx_reqs_active == 0)));
	}

	return 0;
}

static unsigned int
mtp_poll(struct file *fd, poll_table *wait)
{
	struct mtp_dev	*dev = fd->private_data;
	unsigned long		flags;
	int			status = 0;

        MTP_DBG("enter to poll.....\n");
	poll_wait(fd, &dev->rx_wait, wait);
	poll_wait(fd, &dev->tx_wait, wait);

	spin_lock_irqsave(&dev->lock, flags);
	if (likely(!list_empty(&dev->tx_reqs)))
		status |= POLLOUT | POLLWRNORM;

	if (likely(!list_empty(&dev->rx_buffers)))
		status |= POLLIN | POLLRDNORM;

	spin_unlock_irqrestore(&dev->lock, flags);
        MTP_DBG("poll quit\n");
	return status;
}
static int
mtp_ioctl(struct inode *inode, struct file *fd, unsigned int code,
		unsigned long arg)
{
	struct mtp_dev	*dev = fd->private_data;
	unsigned long		flags;
	int			status=0,value = 0;
        struct usb_request              *req=dev->req;
	int len;
        int clen;
	struct mtp_event_data event;
	struct usb_gadget *gadget=dev->gadget;

	MTP_DBG("mtp_ioctl: cmd=0x%4.4x, arg=%lu\n", code, arg);

	/* handle ioctls */

	spin_lock_irqsave(&dev->lock, flags);

	switch (code)
    	{
       	case MTP_IOC_EVENT:
            	MTP_DBG("copy from user %d\n", min((int)_IOC_SIZE(code),MTP_EVENT_SIZE));
            	if(copy_from_user(event.data, (void *)arg, min((int)_IOC_SIZE(code), MTP_EVENT_SIZE))){
	    		status=-EINVAL;
	    		break;
	    	}
            	/* length is in little endian */
            memcpy(&len, event.data, sizeof(len));
            MTP_DBG("len=%d\n", len);
            clen = le32_to_cpu(len);
            MTP_DBG("cpu len=%d\n", clen);
                    /* send event through interrupt in */
            req->length = len;
            req->complete = mtp_event_complete;
            memcpy(req->buf, &event.data,clen);
            if(usb_ep_queue(dev->intr_in, req, GFP_ATOMIC))
                printk("err happen in mtp ioc event\n");
	        break;

       	case MTP_IOC_SEND_ZLP:
            	/* fsync before ZLP */
          	mtp_fsync(fd, NULL, 0);
        
	//	if(!list_empty(&dev->tx_reqs))
		{
	  //      	req = container_of(dev->tx_reqs.next, struct usb_request,                               
	//				 list);
          //     		list_del_init(&req->list);
	        	req->length = 0;
			//req->zero = 1;

			req->complete = zlp_tx_complete;

			if(usb_ep_queue(dev->in_ep, req, GFP_ATOMIC)) {
                		MTP_ERR("error MTP_IOC_SEND_ZLP\n");
			}

            //    	list_add(&req->list, &dev->tx_reqs_active);
		//mtp_write(fd, NULL,0x111, 0);
                //req->complete = tx_complete;
               // req->length = size;

	//    	req->length = 0;
            	//req->complete = tx_complete;
		//req->zero = 1;
       // 	usb_ep_queue(dev->in_ep, req, GFP_ATOMIC);
		
			MTP_DBG("MTP_IOC_SEND_ZLP...\n");
		}
	    	//mtp_fsync(fd, NULL, 0);
		break;

	case MTP_IOC_GET_EP_PK_SIZE:
            	/* get endpoint buffer size for bulk in */
#ifdef  CONFIG_USB_GADGET_DUALSPEED
	     	value = HS_BULK_MAXPACKET_SIZE;
#else		    
	        value = FS_BULK_MAXPACKET_SIZE;
#endif     
		if(copy_to_user((void *)arg,&value , sizeof(int)))
                	   status=-EINVAL;
       		break;

        case MTP_IOC_GET_VENDOR_FLAG:
            	/* get flag whether vender request has served */
        	value=1;
	    	if(copy_to_user((void *)arg, &value, sizeof(int)))
                	status=-EINVAL;
       		break;

        case MTP_IOC_CANCEL_IO:
            MTP_DBG("cancel IO \n");
            mtp_dev_reset(dev); 
			prepare_ep_queue(dev); 
            break;

        case MTP_IOC_DEVICE_RESET:
            	/*If we recieve a Reset IOCTL clear all pending URBs*/
	       	MTP_DBG("Flushing URBs due to Reset IOCTL\n");
	     	//       mtp_empty_cubuf(mtp);
			
           	//	mtp->recv_len = MTP_CANCEL_IO_MARKER;
           	mtp_dev_reset(dev); 
           	//usb_ep_fifo_flush(dev->out_ep);
	   	//usb_ep_fifo_flush(dev->in_ep);
           	//usb_ep_clear_halt(dev->in_ep);
	   	//usb_ep_clear_halt(dev->out_ep);
	//	prepare_ep_queue()
          // 	req->length = USB_BUFSIZE;
//
  //         	req->complete = rx_complete;
//
//	   	usb_ep_queue(dev->out_ep, req, GFP_ATOMIC);

            //wake_up_interruptible(&dev->rx_wait);
		prepare_ep_queue(dev);   
		req->length = 0;
		req->complete=sent_zlp_from_ep0_complete;
		if(usb_ep_queue(gadget->ep0,req,GFP_ATOMIC)){
                        
			MTP_ERR("ep_queue error in %s \n",__func__);
                        req->status = 0;
                        sent_zlp_from_ep0_complete(gadget->ep0, req);
                }
 
		break;
	}
	spin_unlock_irqrestore(&dev->lock, flags);

	return status;
}

/* used after endpoint configuration */
static struct file_operations mtp_io_operations = {
	.owner =	THIS_MODULE,
	.open =		mtp_open,
	.read =		mtp_read,
	.write =	mtp_write,
	.fsync =	mtp_fsync,
	.poll =		mtp_poll,
	.ioctl =	mtp_ioctl,
	.release =	mtp_close
};

#if 1

/*! mtp_dev_req_buf_avail - device request circular buffer free space available
 *
 * @param mtp     pointer to mtp private data
 * Return length of available buf in device request circular buffer
 */
 
  
 
int mtp_dev_req_buf_avail(struct mtp_dev *mtp)
{
    	int return_value = 0; 
    	unsigned char *local_buf_in_ptr; 
    	unsigned char *local_buf_out_ptr; 
	unsigned long flags;

    	if(!mtp)
    	return 0;
    	if((mtp->req_recv_end == NULL || mtp->req_recv_start == NULL))
    	return 0;
 
    	local_irq_save(flags);
    	local_buf_in_ptr = mtp->req_recv_end;
    	local_buf_out_ptr = mtp->req_recv_start;
    	local_irq_restore(flags);

    	/* Calculate how many bytes of free space available in the circular buffer.   
       	If the pointer is equal to the out pointer, no data is available.   
       	Therefore the maximum number of bytes of data the buffer can hold is  
       	local_buf_size-1.  The bytes immediately before the out pointer can not be used  
       	to hold data, because in that case there is no way to distinguish between an  
       	empty buffer or a full buffer  
    	*/ 
    	if ( local_buf_out_ptr > local_buf_in_ptr ){ 
        	return_value = local_buf_out_ptr - local_buf_in_ptr -1; 
    	} 
    	else{ 
	        return_value = MTP_EP0_RECV_MAX_SIZE -1 - (local_buf_in_ptr - local_buf_out_ptr); 
 	}

	return return_value;
}

/*! mtp_dev_req_get - copy out of circular buffer
 *
 * @param mtp      pointer to mtp private data
 * @param buf      pointer to data buffer
 * @param buf_len  len of buffer
 * Return length of data copied
 */
int mtp_dev_req_get(struct mtp_dev *mtp, unsigned char *buf, int buf_len)
{
    	int data_avail;
    	int bytes_to_copy;
    	unsigned char *local_buf_out_ptr; 
    	unsigned char *local_buf_start_addr; 
    	unsigned char *local_buf_end_boundary; 
    	int bytes_before_wraparound; 
    	int bytes_after_wraparound;
    	unsigned long flags;
    	//mtp = usb_mtp_gadget;
    	if(!mtp)
    	return 0;
    	if((mtp->req_recv_end == NULL || mtp->req_recv_start == NULL))
    	return 0;

    	data_avail = MTP_EP0_RECV_MAX_SIZE -1 - mtp_dev_req_buf_avail(mtp);
    	MTP_DBG("data_avail=%d and  req_recv_start=%p\n",data_avail,mtp->req_recv_start);
    	bytes_to_copy = min(data_avail, buf_len);

    	if (bytes_to_copy == 0){
        	MTP_DBG("no data copy\n");
        	return 0;
    	}

        spin_lock_irqsave(&mtp->lock, flags);
	local_buf_out_ptr = mtp->req_recv_start;
    	local_buf_start_addr = mtp->req_recv_buf;
	spin_unlock_irqrestore(&mtp->lock, flags); 
   	local_buf_end_boundary = local_buf_start_addr + MTP_EP0_RECV_MAX_SIZE; 
 
    	/* Decide how many bytes would fall before and after end of buffer boundary. */ 
    	if ( bytes_to_copy < ( local_buf_end_boundary - local_buf_out_ptr ) ){ 
        	bytes_before_wraparound = bytes_to_copy;
        	bytes_after_wraparound = 0; 
        	MTP_DBG("bytes_after_wraparound=0\n");
    	} 
    	else{ 
        	bytes_before_wraparound = local_buf_end_boundary - local_buf_out_ptr; 
        	bytes_after_wraparound = bytes_to_copy - bytes_before_wraparound;
		MTP_DBG("bytes_after_wraparound!=0\n");
 
    	} 
 
    	/* The number of bytes before the end of the circular buffer are 
       	now known - copy them from the circular buffer. */ 
#if 0
   	 value1=memcpy(buf, local_buf_out_ptr, bytes_before_wraparound); 
    	for(value1=0;value1<bytes_before_wraparound;value1++)
    	MTP_PRINT("buffer=%d.=%d.....\n",value1,local_buf_out_ptr[value1]);
    	MTP_PRINT("value1=%d.bytes_before_wraparound=%d.....\n",value1,bytes_before_wraparound);
    	//copy_from_user(buff1,buf,bytes_before_wraparound);
    	for(value1=0;value1<bytes_before_wraparound;value1++)
    	MTP_PRINT("get date=buff%d=%d\n",value1,buf[value1]);

#else
   	if(copy_to_user(buf,local_buf_out_ptr,bytes_before_wraparound)){
		MTP_ERR("error happen in %s t while copy_to_user.....\n",__func__);	
		return -EINVAL;
	}
    	//memcpy(buf,local_buf_out_ptr,bytes_before_wraparound);
#endif
    	/* If there are any remaining bytes to be copied, they must be copied at the start 
       	of the circular buffer. */ 
    	if( bytes_after_wraparound > 0){ 
        	memcpy(buf + bytes_before_wraparound, local_buf_start_addr, bytes_after_wraparound); 
        
       		/* Update local position of in ptr from start of buffer. */ 
        	local_buf_out_ptr = local_buf_start_addr + bytes_after_wraparound; 
    	} 
    	else{ /* Wraparound did not occur. Update local position of in ptr from present position. */ 
        	local_buf_out_ptr += bytes_before_wraparound; 
        
        	/* Check if the local_buf_ptr is pointing to the next bytes after 
           	the last byte of the circular buffer, it needs to point back to 
           	the beginning of teh circular buffer */ 
        	if (local_buf_out_ptr == local_buf_end_boundary){ 
            		local_buf_out_ptr = local_buf_start_addr; 
 		} 
    	}

    	/* Update the in pointer */ 
    	mtp->req_recv_start = local_buf_out_ptr; 

    	return bytes_to_copy;
}

/*! mtp_dev_req_put - copy into circular buffer
 *
 * @param mtp      pointer to mtp private data
 * @param data     pointer to data
 * @param data_len len of date
 * Return length of data copied
 */
int mtp_dev_req_put(struct mtp_dev *mtp, unsigned char *data, int data_len)
{
    	int buf_avail;
    	unsigned char *local_buf_in_ptr; 
    	unsigned char *local_buf_start_addr; 
    	unsigned char *local_buf_end_boundary; 
    	int bytes_before_wraparound; 
    	int bytes_after_wraparound; 
    	unsigned long flags;
    	//mtp = usb_mtp_gadget;
	
    	if(!mtp){
    	MTP_DBG("put data to null dev\n");
    	return 0;
    	}
    
    	if((mtp->req_recv_end == NULL || mtp->req_recv_start == NULL)){
    	MTP_DBG("buffer is null\n");
    	return 0;
    	}
    
   	buf_avail = mtp_dev_req_buf_avail(mtp);
    	if ( buf_avail < data_len ){
        	/* no point to copy incomplete req - this should never happen */
       		MTP_DBG("Not enough space to add dev req data, data_len=%d, buf_avail=%d\n", data_len, buf_avail);
        	return 0;
    	}
    	MTP_DBG(" req_recv_end=%p\n",mtp->req_recv_end);

    	local_irq_save(flags);
    	local_buf_in_ptr = mtp->req_recv_end;
    	local_buf_start_addr = mtp->req_recv_buf;
    	local_irq_restore(flags);
    
    	local_buf_end_boundary = local_buf_start_addr + MTP_EP0_RECV_MAX_SIZE; 
 
    	/* Decide how many bytes would fall before and after end of buffer boundary. */ 
    	if ( data_len < ( local_buf_end_boundary - local_buf_in_ptr ) ){ 
        	bytes_before_wraparound = data_len;
        	bytes_after_wraparound = 0; 
    	} 
    	else{ 
        	bytes_before_wraparound = local_buf_end_boundary - local_buf_in_ptr; 
        	bytes_after_wraparound = data_len - bytes_before_wraparound; 
    	} 
 
    	/* The number of bytes that can fit before the end of the circular buffer are 
      	 now known - copy them into the circular buffer. */ 
    	memcpy(local_buf_in_ptr, data, bytes_before_wraparound); 

    	/* If there are any remaining bytes to be copied, they must be copied at the start 
       		of the circular buffer. */ 
    	if( bytes_after_wraparound > 0){ 
       		memcpy(local_buf_start_addr, data + bytes_before_wraparound, bytes_after_wraparound); 
        
        	/* Update local position of in ptr from start of buffer. */ 
        	local_buf_in_ptr = local_buf_start_addr + bytes_after_wraparound; 
	} 
    	else{ /* Wraparound did not occur. Update local position of in ptr from present position. */ 
     
        	local_buf_in_ptr += bytes_before_wraparound; 
        
       		 /* Check if the local_buf_ptr is pointing to the next bytes after 
           	the last byte of the circular buffer, it needs to point back to 
           	the beginning of teh circular buffer */ 
		if (local_buf_in_ptr == local_buf_end_boundary){
       			local_buf_in_ptr = local_buf_start_addr; 
        	} 
	}

    	/* Update the in pointer */ 
    	mtp->req_recv_end = local_buf_in_ptr; 
	
   	return data_len;
}
/*!  
 * @brief mtp_device_request_bh - BH for handling device request
 * @param arg
 * @return void
 */ 
 

void mtp_device_request_bh (struct work_struct *work)
{
	static int i;
	struct mtp_dev *dev;
	int put_len=0;
	struct mtp_dev_req_bh_private *bh_priv = 
		container_of(work, struct mtp_dev_req_bh_private, bh);;
	
	dev = (struct mtp_dev *)bh_priv->mtp_data;
    
    	if(!dev){
    		MTP_DBG("mtp is no data,pls test it\n");
    		return;
   	 }
    	MTP_PRINT("mtp_device_request_bh: mtp=%p, data_len=%d, pending_data=%d\n", dev, bh_priv->data_len, bh_priv->pending_data);

    	put_len=mtp_dev_req_put(dev, (unsigned char*) bh_priv->data, bh_priv->data_len);
   	for(i=0;i<put_len;i++){
	
  		 MTP_ERR("put into buffer size=%d and data%d=%d\n",put_len,i,bh_priv->data[i]);
    	}
   	if (!bh_priv->pending_data){
        /* no data to receive for the req, pass it to the upper layer now.
         * otherwise the recv data will wake up upper layer.
         */
    
		wake_up_interruptible(&dev->req_recv_wait);
    	}
    	else
    	{
        	MTP_DBG("pending_ep0_data_len=%x\n", pending_ep0_data_len);
        	if (pending_ep0_data)
        	{
            	/* ep0 data arrived before the BH for dev_req */
            		mtp_dev_req_put(dev, pending_ep0_data, pending_ep0_data_len);
			kfree (pending_ep0_data);
  		  	pending_ep0_data = NULL;
            		pending_ep0_data_len = 0;
                 	wake_up_interruptible(&dev->req_recv_wait);
        	}
        	else
       		{
            	ready_for_ep0_data = 1;
       		}
    	}

   
    	if(bh_priv->data)
    	kfree (bh_priv->data);
    

    	if(bh_priv)
    	kfree (bh_priv);
   

   
}
static void sent_from_ep0_complete(struct usb_ep *ep, struct usb_request *req)
{
        struct mtp_dev  *dev = ep->driver_data;
	//int put_len=0;
	dev->req_send_len=0;
        MTP_DBG("ep0 in finished\n");
        wake_up_interruptible(&dev->req_send_wait);
}
static void get_extern_data_from_host(struct usb_ep *ep, struct usb_request *req)
{
	struct mtp_dev  *dev = ep->driver_data;
	unsigned long flags;
	int put_len = 0,i = 0;
	char * buf = req->buf;
	if (ready_for_ep0_data){
		ready_for_ep0_data = false;
   		/* copy data to buffer */
                local_irq_save(flags);
                put_len = mtp_dev_req_put(dev, (unsigned char*) req->buf, req->length);
		for(i = 0;i < put_len;i ++){

	                MTP_ERR("get_extern_data_from_host put into buffer size=%d and data%d=%d\n",put_len,i,buf[i]);
	        }

                local_irq_restore(flags);
                /* data ready */
                wake_up_interruptible(&dev->req_recv_wait);
         }
        else{
        /* data arrived before dev_req BH -- copy data to a tmp place */
             if ((pending_ep0_data = kmalloc (req->actual, GFP_KERNEL)) == NULL) {
	        MTP_ERR( "no space for pending_ep0_data\n");  
		return;
             }
        MTP_DBG( "ep0 out set up dinnished and req->actual=%d req->length=%d\n",req->actual,req->length);
        memcpy(pending_ep0_data, req->buf, req->actual);
        pending_ep0_data_len = req->actual;
        }
        MTP_DBG("ep0 out finished in %s\n",__func__);

//	put_len=mtp_dev_req_put(dev, (unsigned char*) req->buf, req->length);
}
static int mtp_device_request(struct mtp_dev *dev,const struct usb_ctrlrequest *ctrl)
{
	struct mtp_dev *mtp = dev; 
    	int count,ret=0;
    	int buf_avail;
    	int length;
    	u16 wLength = le16_to_cpu(ctrl->wLength);
    	struct usb_request *req = dev->req;
    	struct mtp_ctl_msg_header msg;
    	unsigned char* msg_data = NULL;
    	struct mtp_dev_req_bh_private *bh_priv = NULL;
    	int  has_data = 0;
  	struct usb_gadget *gadget = dev->gadget; 
    	/* alloc memory if it is the first use. */ 
#if 0
    	if (mtp->req_recv_buf == NULL){
	
		MTP_DBG("mtp->req_recv_buf is null,pls apply some memory for it \n");
    		if ((mtp->req_recv_buf = kmalloc (MTP_EP0_RECV_MAX_SIZE, GFP_KERNEL)) == NULL) {
        		return -ENOMEM;
   		}	
    		memset (mtp->req_recv_buf, 0, MTP_EP0_RECV_MAX_SIZE);
        	mtp->req_recv_start = mtp->req_recv_buf;
        	mtp->req_recv_end = mtp->req_recv_buf;
   	 }
#endif
    	/* prepare msg that needs to pass along to the upper layer */
    	count = MTP_CTL_MSG_HEADER_SIZE + sizeof(struct usb_ctrlrequest);
    	MTP_DBG("usb_ctrlrequest for usb class and cmd count =%d",count);
    	/* if it is from host to device, also need space for data, if any */
    	if ((ctrl->bRequestType & 0x80) == 0 && wLength > 0){
        	count += wLength;
        	has_data = 1;
	    }

    	buf_avail = mtp_dev_req_buf_avail(mtp);
    	MTP_PRINT("buf avail=%d, count=%d\n", buf_avail, count); 
    	if (count > buf_avail){
        	MTP_ERR("LEAVE ... device request overflow\n"); 
       		return -ENOMEM;
   	}
    	length = MTP_CTL_MSG_HEADER_SIZE + sizeof(struct usb_ctrlrequest);
    	if ((msg_data = kmalloc (length, GFP_KERNEL)) == NULL) {
        	return -ENOMEM;
    	}
    	memset (msg_data, 0, length);
    
    	bh_priv = kmalloc(sizeof(struct mtp_dev_req_bh_private), GFP_KERNEL);
    	if (bh_priv == NULL)
   	{
        	kfree (msg_data);
        	return -ENOMEM;
    	}


    	memset (bh_priv, 0, sizeof(struct mtp_dev_req_bh_private));
    	msg.msg_id = MTP_CTL_CLASS_REQ;
    	msg.msg_len =  count;
    	memcpy(msg_data, &msg, MTP_CTL_MSG_HEADER_SIZE);
    	memcpy(msg_data + MTP_CTL_MSG_HEADER_SIZE, (unsigned char*) ctrl, sizeof(struct usb_ctrlrequest));

    	MTP_DBG("dev req schedule work\n");
    	bh_priv->mtp_data=dev;
    	bh_priv->data = msg_data;
    	bh_priv->data_len = MTP_CTL_MSG_HEADER_SIZE + sizeof(struct usb_ctrlrequest);
   	bh_priv->pending_data = has_data;
    
  	  /* The host could send multiple dev reqs one after another. The dev req is handled in intr and recv data in BH.
     	* So to get the dev_req->data->dev_req in the right sequence, we need to process dev req in BH too.
    	 */
    	INIT_WORK (&bh_priv->bh, &mtp_device_request_bh);
    	schedule_work(&bh_priv->bh);
    	MTP_DBG("dev req work scheduled\n");

   	 /* if it is from host to device, also need to read data if wLength > 0 */
   	 if (has_data)
   	 {
       	 /* read data that follows the request */
		ret = wLength;
		req->length = ret;
		req->complete = get_extern_data_from_host;
        	ret = usb_ep_queue(gadget->ep0, req, GFP_ATOMIC);
		MTP_DBG("if has data,need continue to work,to be continue and ret=%d\n",ret);

    	}
	return ret;

}


#endif////////////waj debug
static int 
mtp_ctl_open(struct inode *inode, struct file *fd)
{
        int                     ret = -EBUSY;
	unsigned long           flags;
	struct mtp_dev *dev = PDE(inode)->data;
	spin_lock_irqsave(&dev->lock, flags);
        MTP_DBG("open mtp dev\n");
        if (!dev->mtp_proc_open) {
                dev->mtp_proc_open = 1;
		fd->private_data=dev;
                ret = 0;
                /* Change the mtp status to show that it's on-line. */
       }
        MTP_DBG("mtp_ctl_open returned %x\n", ret);
	spin_unlock_irqrestore(&dev->lock, flags);

#if 0
        return 0;
#else
        return ret;
#endif

	
}

static int
mtp_ctl_close(struct inode *inode, struct file *fd)
{
        int                     ret = -EBUSY;
	unsigned long           flags;
        struct mtp_dev  *dev = fd->private_data;
	spin_lock_irqsave(&dev->lock, flags);
        MTP_DBG("close mtp dev\n");

	fd->private_data = NULL;
	dev->mtp_proc_open = 0;
        ret = 0;
        /* Change the mtp status to show that it's on-line. */
        MTP_DBG("mtp_ctl_close returned %x\n", ret);
	spin_unlock_irqrestore(&dev->lock, flags);
        return ret;


}

static int  
mtp_ctl_read(struct file *fd, char __user*buf, size_t len, loff_t * pos)
{
	int bytes_read = 0;
        struct mtp_dev  *dev = fd->private_data;
	unsigned long           flags;
	MTP_DBG("mtp ctl read len=%d", len);
   	if(!dev){
   		MTP_ERR("dev is null.................\n");
    		return -EINVAL;
	}
	MTP_DBG("mtp_ctl_read is valid................\n");
	spin_lock_irqsave(&dev->lock, flags);
	if(dev->req_recv_end == dev->req_recv_start)
	{
		spin_unlock_irqrestore(&dev->lock, flags);
		if (fd->f_flags & (O_NONBLOCK|O_NDELAY)) {
                	MTP_DBG("mtp ctl no_block and return from read\n"); 
			return -EAGAIN;
               	}
		MTP_DBG("mtp ctl read SLEEPING\n");
 
		wait_event_interruptible(dev->req_recv_wait,(dev->req_recv_end!=dev->req_recv_start));
        }
	MTP_DBG("mtp ctl read awake\n");
    	bytes_read = mtp_dev_req_get(dev, buf, len);
#if 0 
       MTP_PRINT("set class 2.1\n");

                        mtp_str2descriptor(mtp_extended_config_descriptor, 4,mtp_status1);
                        memcpy(req->buf,mtp_extended_config_descriptor,4);
                        MTP_PRINT("\nset class 2\n");
	req->length = 4;
        MTP_PRINT("ctl write finished\n");

        req->complete=sent_from_ep0_complete;

			    ret=usb_ep_queue(gadget->ep0,req,GFP_ATOMIC);
       if (ret < 0) {
                        MTP_PRINT("ep_queue -->11111111111111 %d\n", ret);
                        req->status = 0;
                        sent_from_ep0_complete(gadget->ep0, req);
                }
	
 #endif	
	MTP_DBG("mtp ctl read return len=%d", bytes_read); 
	return bytes_read;
}

static int
mtp_ctl_write(struct file *fd, const char __user *buf, size_t len, loff_t * pos)  
{
    	//struct mtp_dev *dev = usb_mtp_gadget;
    	struct mtp_dev  *dev = fd->private_data;
    	//struct mtp_ctl_msg_header msg;
    	struct usb_gadget *gadget=dev->gadget;
    	struct usb_request *req = dev->req;
	//u8                      mtp_extended_config_descriptor[64];
	int ret=0;
	unsigned long           flags;
    	MTP_DBG("mtp_ctl_write len=%d\n", len);
#if 0
    	if (len > MTP_EP0_SEND_MAX_SIZE || len < MTP_CTL_MSG_HEADER_SIZE){
        	/* Error: write size too large or too small */
	        MTP_PRINT("mtp_ctl_write size invalid\n");
	        return -ENOMEM;
	}

	    /* msg info */
	if(copy_from_user(&msg, buf, MTP_CTL_MSG_HEADER_SIZE))
		return -EFAULT;
    	MTP_PRINT("ctl write msg len = %d, msg id = %d\n", msg.msg_len, msg.msg_id);
    	if (msg.msg_id != MTP_CTL_CLASS_REPLY){
        	/* unsupported */
        	MTP_PRINT("unsupported ctl msg id %d\n", msg.msg_id);
        	return -EINVAL;
    	}
#endif
	spin_lock_irqsave(&dev->lock, flags);	
	if(dev->req_send_len)
	{
		 spin_unlock_irqrestore(&dev->lock, flags);
		 if (fd->f_flags & (O_NONBLOCK|O_NDELAY)) {
			return -EAGAIN;
                }
 
	    	MTP_DBG("mtp->req_send_len = %d\n", dev->req_send_len);
        
		MTP_DBG("ctl write SLEEPING\n");
		wait_event_interruptible(dev->req_send_wait,!dev->req_send_len );
        }
	MTP_DBG("ctl write AWAKE\n");

    	/* sending the data */
    	//dev->req_send_len = len - MTP_CTL_MSG_HEADER_SIZE;
	spin_lock_irqsave(&dev->lock, flags);
	dev->req_send_len = len;
     	if (dev->req_send_len > MTP_EP0_SEND_MAX_SIZE){
        	MTP_ERR("len=%d oversize, truncate to %d\n", len, MTP_EP0_SEND_MAX_SIZE);
        	len = MTP_EP0_SEND_MAX_SIZE;
    	}
//	if(copy_from_user(req->buf, buf+MTP_CTL_MSG_HEADER_SIZE, len))
	req->length = len;
   	spin_unlock_irqrestore(&dev->lock, flags);
	if(copy_from_user(req->buf, buf, len))
		return -EFAULT;
	MTP_DBG("ctl write finished\n");
	

                        //p_str2descriptor(mtp_extended_config_descriptor, 4,mtp_status1);
                        //mcpy(req->buf,mtp_extended_config_descriptor,4);
	spin_lock_irqsave(&dev->lock, flags);
	req->complete=sent_from_ep0_complete;
    	ret=usb_ep_queue(gadget->ep0,req,GFP_ATOMIC);
     	if (ret < 0) {
                        MTP_ERR("ep_queue error %d in %s \n", ret,__func__);
                        req->status = 0;
                        sent_from_ep0_complete(gadget->ep0, req);
                }

	MTP_DBG("mtp_ctl_write finished and ret=%d",ret);
      
    	spin_unlock_irqrestore(&dev->lock, flags);
    	return len;
}

static unsigned int
mtp_ctl_poll(struct file *fd, poll_table *wait)
{
        struct mtp_dev  *dev = fd->private_data;
        unsigned long           flags;
        int                     status = 0;

        MTP_DBG("enter to ctl poll.....\n");
        poll_wait(fd, &dev->req_recv_wait, wait);
        poll_wait(fd, &dev->req_send_wait, wait);

        spin_lock_irqsave(&dev->lock, flags);
	//if(dev->req_recv_end != dev->req_recv_start)
	
	if (likely(!dev->req_send_len)){	
		MTP_DBG("req_send_len is valid.....\n");
                status |= POLLOUT | POLLWRNORM;
	}

        if(dev->req_recv_end != dev->req_recv_start){
		MTP_DBG("dev->req_recv_end!=dev->req_recv_start.....\n");
        
		status |= POLLIN | POLLRDNORM;
	}
        spin_unlock_irqrestore(&dev->lock, flags);
        MTP_DBG("poll ctl quit and status=%d\n",status);
        return status;
}

/* used after endpoint configuration */
static struct file_operations mtp_ctl_operations = {
	//owner =	THIS_MODULE,
	open:	mtp_ctl_open,
	release:  mtp_ctl_close,
	read: 	mtp_ctl_read,
	write:	mtp_ctl_write,
	poll:	mtp_ctl_poll,
};

/*-------------------------------------------------------------------------*/

static void mtp_proc_add_device(struct mtp_dev *dev)
{
        struct proc_dir_entry *mtp_proc = NULL;


       // mtp_proc=create_proc_entry("mtpctl",S_IRUSR|S_IWUSR, NULL);
	 mtp_proc=create_proc_entry("mtpctl",0777, NULL);
        if(mtp_proc){
                mtp_proc->proc_fops = &mtp_ctl_operations;
                mtp_proc->data = dev;
        }
        else
                MTP_DBG("%s: creating /proc/mtpctl failed\n", __FUNCTION__);

}
static int
set_mtp_interface(struct mtp_dev *dev)
{
	int	result = 0;
	dev->in = ep_desc(dev->gadget, &hs_ep_in_desc, &fs_ep_in_desc);
	dev->in_ep->driver_data = dev;

	dev->out = ep_desc(dev->gadget, &hs_ep_out_desc, &fs_ep_out_desc);
	dev->out_ep->driver_data = dev;

	dev->intr = ep_desc(dev->gadget, &hs_intr_in_desc,&fs_intr_in_desc);
	dev->intr_in->driver_data = dev;

	result = usb_ep_enable(dev->in_ep, dev->in);
	if (result != 0) {
		MTP_ERR("enable %s --> %d\n", dev->in_ep->name, result);
		goto done;
	}

	result = usb_ep_enable(dev->out_ep, dev->out);
	if (result != 0) {
		MTP_ERR("enable %s --> %d\n", dev->in_ep->name, result);
		goto done;
	}

	result = usb_ep_enable(dev->intr_in, dev->intr);
	if (result != 0) {
		MTP_ERR("enable %s --> %d\n", dev->intr_in->name, result);
		goto done;
	}

done:
	/*con error, disable any endpoints  */
	if (result != 0) {
		(void) usb_ep_disable(dev->in_ep);
		(void) usb_ep_disable(dev->out_ep);
		(void) usb_ep_disable(dev->intr_in);
		dev->in = NULL;
		dev->out = NULL;
		dev->intr = NULL;
		MTP_DBG("ep enable error\n");
	}
	prepare_ep_queue(dev);
	
	return result;
}

static void mtp_reset_interface(struct mtp_dev *dev)
{
	
	if (dev->interface < 0)
		return;

	MTP_DBG( "RESET %s\n", __FUNCTION__);

	if (dev->in){
		usb_ep_disable(dev->in_ep);
	}
	if (dev->out){
		usb_ep_disable(dev->out_ep);
	}
	if (dev->intr){
		usb_ep_disable(dev->intr_in);
	}
	dev->interface = -1;
}

/* change our operational config.  must agree with the code
 * that returns config descriptors, and altsetting code.
 */
static int
set_interface(struct mtp_dev *dev, unsigned number);
static int
mtp_set_config(struct mtp_dev *dev, unsigned number)
{
	int			result=0,rc=0;
	struct usb_gadget	*gadget = dev->gadget;
	MTP_DBG("mtp set config.%d..\n",number);
	if (gadget_is_sa1100(gadget) && dev->config) {
		/* tx fifo is full, but we can't clear it...*/
		MTP_ERR("can't change configurations\n");
		return -ESPIPE;
	}
	dev->config = 0; 
	//mtp_soft_reset(dev);
	switch (number) {
	case DEV_CONFIG_VALUE:
		mtp_reset_interface(dev);
		result = 0;
		break;
	default:
		result = -EINVAL;
		/* FALL THROUGH */
	case 0:
		break;
	}
	if (number == dev->config)
	{
		MTP_DBG("mtp set config. return 0..\n");
		return 0;
	}

	if ((rc = set_interface(dev, 0)) != 0)
			dev->config = 0;	// Reset on errors
	else {

		if (result) {
			usb_gadget_vbus_draw(dev->gadget,
			dev->gadget->is_otg ? 8 : 100);
		}
		else {

			char *speed;
			unsigned power;
		
			dev->config = number;
			power = 2 * config_desc.bMaxPower;
			usb_gadget_vbus_draw(dev->gadget, power);

			switch (gadget->speed) {
			case USB_SPEED_FULL:	speed = "full"; break;
#ifdef CONFIG_USB_GADGET_DUALSPEED
			case USB_SPEED_HIGH:	speed = "high"; break;
#endif
			default:		speed = "?"; break;
			}

			//prepare_ep_queue(dev);
			MTP_PRINT( "%s speed config #%d: %d mA, %s\n",
					speed, number, power, driver_desc);
		}
	}
	return result;
}

static int
populate_config_buf(enum usb_device_speed speed, u8 *buf, u8 type, unsigned index,
		int is_otg)
{
	int					len;
	const struct usb_descriptor_header	**function;
#ifdef CONFIG_USB_GADGET_DUALSPEED
	int					hs = (speed == USB_SPEED_HIGH);

	if (type == USB_DT_OTHER_SPEED_CONFIG)
		hs = !hs;

	if (hs) {
		function = hs_mtp_function;
	} else {
		function = fs_mtp_function;
	}
#else
	function = fs_mtp_function;
#endif

	if (index >= device_desc.bNumConfigurations)
		return -EINVAL;

	/* for now, don't advertise srp-only devices */
	if (!is_otg)
		function++;

	len = usb_gadget_config_buf(&config_desc, buf, USB_DESC_BUFSIZE,
			function);
	if (len < 0)
		return len;
	((struct usb_config_descriptor *) buf)->bDescriptorType = type;
	return len;
}

/* Change our operational Interface. */
static int
set_interface(struct mtp_dev *dev, unsigned number)
{
	int			result = 0;

	if (gadget_is_sa1100(dev->gadget) && dev->interface < 0) {
		/* tx fifo is full, but we can't clear it...*/
		MTP_ERR("can't change interfaces\n");
		return -ESPIPE;
	}
	
	 MTP_DBG("set inface %d\n",number);

	/* Free the current interface */
	switch (dev->interface) {
	case MTP_INTERFACE:
		MTP_DBG("set inface\n");
		mtp_reset_interface(dev);
		break;
	}

	switch (number) {
	case MTP_INTERFACE:
		result = set_mtp_interface(dev);
		if (result) {
			mtp_reset_interface(dev);
		} else {
			dev->interface = MTP_INTERFACE;
		}
		break;
	default:
		result = -EINVAL;
		/* FALL THROUGH */
	}

	if (!result)
		MTP_DBG("Using interface %x\n", number);

	return result;
}

static void mtp_setup_complete(struct usb_ep *ep, struct usb_request *req)
{
	if (req->status || req->actual != req->length){
	  	MTP_DBG("setup complete --> %d, %d/%d\n",
				req->status, req->actual, req->length);
	}
}

static void mtp_dev_reset(struct mtp_dev *dev)
{
	struct usb_request	*req;
	MTP_DBG( "Received mtp Reset Request\n");
#if 0
	if (usb_ep_disable(dev->in_ep))
		MTP_ERR("Failed to disable USB in_ep\n");
	if (usb_ep_disable(dev->out_ep))
		MTP_ERR( "Failed to disable USB out_ep\n");
	if (usb_ep_disable(dev->intr_in))
                MTP_ERR("Failed to disable USB intr_ep\n");
#endif
	usb_ep_fifo_flush(dev->in_ep);
	usb_ep_fifo_flush(dev->out_ep);
	//error = usb_ep_dequeue(dev->out_ep, req);
	if (dev->current_rx_req != NULL) {
		list_add(&dev->current_rx_req->list, &dev->rx_reqs);
		dev->current_rx_req = NULL;
	}
        dev->rx_reqs_active = 0;
        dev->tx_reqs_active = 0;
	dev->current_rx_bytes = 0;
	dev->current_rx_buf = NULL;
	dev->reset_mtp = 1;
        //dev->req_recv_start = dev->req_recv_buf;
        //dev->req_recv_end = dev->req_recv_buf;
	//memset (dev->req_recv_buf, 0, MTP_EP0_RECV_MAX_SIZE);
	//dev->req_send_len = 0;
#if 1
	while (likely(!(list_empty(&dev->rx_buffers)))) {
		req = container_of(dev->rx_buffers.next, struct usb_request,
				list);
		printk("rx_buffers is not full\n");
		list_del_init(&req->list);
		list_add(&req->list, &dev->rx_reqs);
	}
#endif
#if 0
	while (likely(!(list_empty(&dev->rx_reqs_active)))) {
		req = container_of(dev->rx_buffers.next, struct usb_request,
				list);
		list_del_init(&req->list);
		list_add(&req->list, &dev->rx_reqs);
	}

	while (likely(!(list_empty(&dev->tx_reqs_active)))) {
		req = container_of(dev->tx_reqs_active.next,
				struct usb_request, list);
		list_del_init(&req->list);
		list_add(&req->list, &dev->tx_reqs);
	}
#endif
#if 0 
	if (usb_ep_enable(dev->in_ep, dev->in))
		MTP_ERR("Failed to enable USB in_ep\n");
	if (usb_ep_enable(dev->out_ep, dev->out))
		MTP_ERR( "Failed to enable USB out_ep\n");
	if(usb_ep_enable(dev->intr_in, dev->intr))
		MTP_ERR("fail enable usb intr\n");
#endif
	MTP_DBG("reset mtp....\n");
	wake_up_interruptible(&dev->tx_wait);
	wake_up_interruptible(&dev->tx_flush_wait);
}

/*-------------------------------------------------------------------------*/

/*
 * The setup() callback implements all the ep0 functionality that's not
 * handled lower down.
 */
static int
mtp_setup(struct usb_gadget *gadget, const struct usb_ctrlrequest *ctrl)
{
	struct mtp_dev	*dev = get_gadget_data(gadget);
	struct usb_request	*req = dev->req;
	int			value = -EOPNOTSUPP;
	u16			wIndex = le16_to_cpu(ctrl->wIndex);
	u16			wValue = le16_to_cpu(ctrl->wValue);
	u16			wLength = le16_to_cpu(ctrl->wLength);
	int 			len;
	u8 			mtp_extended_config_descriptor[64];
	MTP_DBG( "mtp_setup ctrl req%02x.%02x v%04x i%04x l%d\n",
		ctrl->bRequestType, ctrl->bRequest, wValue, wIndex, wLength);

	req->complete = mtp_setup_complete;

	switch (ctrl->bRequestType&USB_TYPE_MASK) {

	case USB_TYPE_STANDARD:
		MTP_DBG("USB_TYPE_STANDARD..\n");
		
		switch (ctrl->bRequest) {

		case USB_REQ_GET_DESCRIPTOR:
			if (ctrl->bRequestType != USB_DIR_IN)
				break;
			switch (wValue >> 8) {

			case USB_DT_DEVICE:
				value = min(wLength, (u16) sizeof device_desc);
				memcpy(req->buf, &device_desc, value);
				break;
#ifdef CONFIG_USB_GADGET_DUALSPEED
			case USB_DT_DEVICE_QUALIFIER:
				if (!gadget->is_dualspeed)
					break;
				value = min(wLength,
						(u16) sizeof dev_qualifier);
				memcpy(req->buf, &dev_qualifier, value);
				break;

			case USB_DT_OTHER_SPEED_CONFIG:
				if (!gadget->is_dualspeed)
					break;
				/* FALLTHROUGH */
#endif /* CONFIG_USB_GADGET_DUALSPEED */
			case USB_DT_CONFIG:
				value = populate_config_buf(gadget->speed, req->buf,
						wValue >> 8,
						wValue & 0xff,
						gadget->is_otg);
				if (value >= 0)
					value = min(wLength, (u16) value);
				break;

			case USB_DT_STRING:
				value = usb_gadget_get_string(&stringtab,
						wValue & 0xff, req->buf);
				//MTP_PRINT("get string = %s\n",req->buf);
				//
				if(wValue == MTP_EXT_STRING_ID){
					if(ext_string)
						mtp_str2descriptor(mtp_extended_config_descriptor, MTP_EXT_STR_MAX_LEN, ext_string);		
					else
						mtp_str2descriptor(mtp_extended_config_descriptor, MTP_EXT_STR_MAX_LEN, ext_str_desc);
		                        len=mtp_extended_config_descriptor[0];
		                        value = min(wLength, (u16)len );
					memcpy(req->buf,mtp_extended_config_descriptor,value);
				}
				if (value >= 0)
					value = min(wLength, (u16) value);
				MTP_DBG("value=====%d\n",value);
				break;
			}
			break;

		case USB_REQ_SET_CONFIGURATION:
			MTP_DBG("set config...\n");
			if (ctrl->bRequestType != 0)
				break;
			value = mtp_set_config(dev, wValue);
			break;
		case USB_REQ_GET_CONFIGURATION:
			MTP_DBG("get config...\n");
			if (ctrl->bRequestType != USB_DIR_IN)
				break;
			*(u8 *)req->buf = dev->config;
			value = min(wLength, (u16) 1);
			break;

		case USB_REQ_SET_INTERFACE:
		//	if (ctrl->bRequestType != USB_RECIP_INTERFACE ||
		//
		///			!dev->config)
				//break;
			if (ctrl->bRequestType != (USB_DIR_OUT| USB_TYPE_STANDARD |
			USB_RECIP_INTERFACE))
			break;		
			value = set_interface(dev, MTP_INTERFACE);
			break;
		case USB_REQ_GET_INTERFACE:
			if (ctrl->bRequestType !=
					(USB_DIR_IN|USB_RECIP_INTERFACE)
					|| !dev->config)
				break;

			*(u8 *)req->buf = dev->interface;
			value = min(wLength, (u16) 1);
			break;

		default:
			goto unknown;
		}
		break;

	case USB_TYPE_CLASS:
		MTP_DBG("USB_TYPE_CLASS\n");
#if 0 
		switch (ctrl->bRequest) {
		case 0x67: /* Get Port Status */
			/* Only one mtp interface is supported. */
			MTP_PRINT("set class 2.1\n");

			if (wIndex != MTP_INTERFACE)
				break;
			mtp_str2descriptor(mtp_extended_config_descriptor, 4,mtp_status1);
			memcpy(req->buf,mtp_extended_config_descriptor,4);
			value = 4;//min(wLength, (u16) 4);
			MTP_PRINT("\nset class 2\n");
			break;

		case 0x66: /* Soft Reset */
			/* Only one mtp interface is supported. */
			MTP_PRINT("set class3 \n");
			if (wIndex != MTP_INTERFACE)
				break;
		
	
			//mtp_soft_reset(dev);
			value = 0;
			break;

		default:
			goto unknown;
		}
#else	
	
		value=mtp_device_request(dev,ctrl);
		//value=0;
        	MTP_DBG("quit from setup......wait......\n");
        	/* host either stalls (value < 0) or reports success */
        	return value;
#endif
        case USB_TYPE_VENDOR:
                MTP_DBG("USB_TYPE_VENDOR..\n");
                switch (ctrl->bRequest) {
                case MTP_VENDOR_CODE://USB_MOD_VENDOR_CODE
                        if(wIndex !=4){
	                        MTP_ERR("not ext device description\n");
        	                break;
                        }
			if(ext_desc){
				MTP_DBG("NO ext_desc input\n");
				mtp_str2descriptor(mtp_extended_config_descriptor, MTP_EXT_DESC_MAX_LEN, ext_desc);
			}
			else
				mtp_str2descriptor(mtp_extended_config_descriptor, MTP_EXT_DESC_MAX_LEN, mtp_ext_desc);
                        len=mtp_extended_config_descriptor[0];
			value = min(wLength, (u16)len );
			MTP_DBG("USB_TYPE_VENDOR len=%d.....\n",value);
			memcpy(req->buf, mtp_extended_config_descriptor,value );
			break;
                }
                break;
		
	default:
unknown:
		MTP_DBG("unknown ctrl req%02x.%02x v%04x i%04x l%d\n",
			ctrl->bRequestType, ctrl->bRequest,
			wValue, wIndex, wLength);
		break;
	}

	/* respond with data transfer before status phase? */
	if (value >= 0) {
		req->length = value;
		req->zero = value < wLength
				&& (value % gadget->ep0->maxpacket) == 0;
		value = usb_ep_queue(gadget->ep0, req, GFP_ATOMIC);
		if (value < 0) {
			MTP_DBG("ep_queue --> %d\n", value);
			req->status = 0;
			mtp_setup_complete(gadget->ep0, req);
		}
	}
	MTP_DBG("quit from setup\n");
        /* host either stalls (value < 0) or reports success */
        return value;

}

static void
mtp_disconnect(struct usb_gadget *gadget)
{
	struct mtp_dev	*dev = get_gadget_data(gadget);
	unsigned long		flags;

	MTP_DBG("%s\n", __FUNCTION__);

	spin_lock_irqsave(&dev->lock, flags);

	dev->config = 0;

	mtp_reset_interface(dev);

	spin_unlock_irqrestore(&dev->lock, flags);
}

static void
mtp_unbind(struct usb_gadget *gadget)
{
	struct mtp_dev	*dev = get_gadget_data(gadget);
	struct usb_request	*req;


	MTP_DBG("%s\n", __FUNCTION__);
	remove_proc_entry("mtpctl", NULL);
	/* Remove sysfs files */
	device_destroy(usb_gadget_class, g_mtp_devno);

	/* Remove Character Device */
	if (dev != NULL){

		cdev_del(&dev->mtp_cdev);


		/* we must already have been disconnected ... no i/o may be active */
		WARN_ON(dev->tx_reqs_active != 0);
		WARN_ON(dev->rx_reqs_active != 0);

		/* Free all memory for this driver. */
		while (!list_empty(&dev->tx_reqs)) {
			req = container_of(dev->tx_reqs.next, struct usb_request,
					list);
			list_del(&req->list);
			mtp_req_free(dev->in_ep, req);
		}

		if (dev->current_rx_req != NULL)
			mtp_req_free(dev->out_ep, dev->current_rx_req);

		while (!list_empty(&dev->rx_reqs)) {
			req = container_of(dev->rx_reqs.next,
					struct usb_request, list);
			list_del(&req->list);
			mtp_req_free(dev->out_ep, req);
		}

		while (!list_empty(&dev->rx_buffers)) {
			req = container_of(dev->rx_buffers.next,
					struct usb_request, list);
			list_del(&req->list);
			mtp_req_free(dev->out_ep, req);
		}

		if (dev->req) {
			mtp_req_free(gadget->ep0, dev->req);
			dev->req = NULL;
		}
		if (dev->req_recv_buf){ 
			kfree(dev->req_recv_buf);
			dev->req_recv_buf = NULL;
		}
		set_gadget_data(gadget, NULL);
	}
	if(usb_mtp_gadget) {
		kfree(usb_mtp_gadget);
		usb_mtp_gadget=NULL;
	}
	MTP_DBG("%s unbind finished\n",DRIVER_NAME);
}

static int __init
mtp_bind(struct usb_gadget *gadget)
{
	struct mtp_dev	*dev;
	struct usb_ep		*in_ep, *out_ep,*intr_in;
	int			status = -ENOMEM;
	int			gcnum;
	u32			i;
	struct usb_request	*req;
	usb_mtp_gadget = dev = kzalloc(sizeof(struct mtp_dev), GFP_KERNEL);
	
	if (dev == NULL)
	return -ENOMEM;

	/* Setup the sysfs files for the mtp gadget. */
	dev->pdev = device_create(usb_gadget_class, NULL, g_mtp_devno, NULL,
			"g_mtp");
	if (IS_ERR(dev->pdev)) {
		MTP_DBG("Failed to create device: g_mtp\n");
		goto fail;
	}

	/*
	 * Register a character device as an interface to a user mode
	 * program that handles the mtp specific functionality.
	 */
	cdev_init(&dev->mtp_cdev, &mtp_io_operations);
	dev->mtp_cdev.owner = THIS_MODULE;
	status = cdev_add(&dev->mtp_cdev, g_mtp_devno, 1);

	if (status) {
		MTP_DBG("Failed to open char device\n");
		goto fail;
	}

	mtp_proc_add_device(dev);

	if (gadget_is_sa1100(gadget)) {
		/* hardware can't write zero length packets. */
		MTP_DBG("SA1100 controller is unsupport by this driver\n");
		goto fail;
	}

	gcnum = usb_gadget_controller_number(gadget);
	if (gcnum >= 0) {
		device_desc.bcdDevice = cpu_to_le16(0x0200 + gcnum);
	} else {
		dev_warn(&gadget->dev, "controller '%s' not recognized\n",
			gadget->name);
		/* unrecognized, but safe unless bulk is REALLY quirky */
		device_desc.bcdDevice =
			__constant_cpu_to_le16(0xFFFF);
	}
	snprintf(productor, sizeof productor, "%s %s %s",
                    CONFIG_USB_DRIVER_MANUFACTURER,CONFIG_USB_DRIVER_PRODUCTOR,DRIVER_DESC);
//	device_desc.bcdDevice = cpu_to_le16(DRIVER_BCD_DEV);
	device_desc.idVendor =
		__constant_cpu_to_le16(MTP_VENDOR_NUM);
	device_desc.idProduct =
		__constant_cpu_to_le16(MTP_PRODUCT_NUM);
	//device_desc.bMaxPacketSize0 = gadget->ep0->maxpacket;
	/* support optional vendor/distro customization */
	if (idVendor) {
		if (!idProduct) {
			MTP_DBG("idVendor needs idProduct!\n");
			return -ENODEV;
		}
		device_desc.idVendor = cpu_to_le16(idVendor);
		device_desc.idProduct = cpu_to_le16(idProduct);
		if (bcdDevice)
			device_desc.bcdDevice = cpu_to_le16(bcdDevice);
	}

	if (iManufacturer)
		strlcpy(manufacturer, iManufacturer, sizeof manufacturer);

	if (iProduct)
		strlcpy(product_desc, iProduct, sizeof product_desc);

	if (iSerialNum)
		strlcpy(serial, iSerialNum, sizeof serial);
    else
        strlcpy(serial, serial_default, sizeof serial);
    
	usb_ep_autoconfig_reset(gadget);
	in_ep = usb_ep_autoconfig(gadget, &fs_ep_in_desc);
	if (!in_ep) {
autoconf_fail:
		MTP_DBG("can't autoconfigure on %s\n",
			gadget->name);
		return -ENODEV;
	}
	in_ep->driver_data = in_ep;	/* claim */

	out_ep = usb_ep_autoconfig(gadget, &fs_ep_out_desc);
	if (!out_ep)
		goto autoconf_fail;
	out_ep->driver_data = out_ep;	/* claim */


	intr_in = usb_ep_autoconfig(gadget, &fs_intr_in_desc);
	if (!intr_in)
		goto autoconf_fail;
	intr_in->driver_data = intr_in;	/* claim */
        device_desc.bMaxPacketSize0 = gadget->ep0->maxpacket;
#ifdef	CONFIG_USB_GADGET_DUALSPEED
	/* assumes ep0 uses the same value for both speeds ... */
	MTP_DBG("CONFIG_USB_GADGET_DUALSPEED is enable\n");
	dev_qualifier.bMaxPacketSize0 = device_desc.bMaxPacketSize0;

	/* and that all endpoints are dual-speed */
	hs_ep_in_desc.bEndpointAddress = fs_ep_in_desc.bEndpointAddress;
	hs_ep_out_desc.bEndpointAddress = fs_ep_out_desc.bEndpointAddress;
	hs_intr_in_desc.bEndpointAddress = fs_intr_in_desc.bEndpointAddress;
#endif	/* DUALSPEED */

	device_desc.bMaxPacketSize0 = gadget->ep0->maxpacket;
	usb_gadget_set_selfpowered(gadget);

	if (gadget->is_otg) {
		otg_desc.bmAttributes |= USB_OTG_HNP,
		config_desc.bmAttributes |= USB_CONFIG_ATT_WAKEUP;
		config_desc.bMaxPower = 4;
	}
	spin_lock_init(&dev->lock);
	spin_lock_init(&dev->lock_mtp_io);
	INIT_LIST_HEAD(&dev->tx_reqs);
	INIT_LIST_HEAD(&dev->rx_reqs);
	INIT_LIST_HEAD(&dev->rx_buffers);
	dev->rx_reqs_active = 0;
	dev->tx_reqs_active = 0;
	init_waitqueue_head(&dev->rx_wait);
	init_waitqueue_head(&dev->tx_wait);
	init_waitqueue_head(&dev->tx_flush_wait);
	init_waitqueue_head(&dev->req_recv_wait);
	init_waitqueue_head(&dev->req_send_wait);

	dev->req_recv_buf = NULL;
   	dev->req_recv_start = dev->req_recv_buf;
    	dev->req_recv_end = dev->req_recv_buf;
	dev->req_send_len = 0;
	

	dev->config = 0;
	dev->interface = -1;
	dev->mtp_cdev_open = 0;
	dev->mtp_proc_open = 0;
	dev->current_rx_req = NULL;
	dev->current_rx_bytes = 0;
	dev->current_rx_buf = NULL;

	dev->in_ep = in_ep;
	dev->out_ep = out_ep;
	dev->intr_in = intr_in;

	/* preallocate control message data and buffer */
	dev->req = mtp_req_alloc(gadget->ep0, USB_DESC_BUFSIZE,
			GFP_KERNEL);
	if (!dev->req) {
		status = -ENOMEM;
		goto fail;
	}

	for (i = 0; i < LEN; i++) {
		req = mtp_req_alloc(dev->in_ep, USB_BUFSIZE, GFP_KERNEL);
		if (!req) {
			while (!list_empty(&dev->tx_reqs)) {
				req = container_of(dev->tx_reqs.next,
						struct usb_request, list);
				list_del(&req->list);
				mtp_req_free(dev->in_ep, req);
			}
			return -ENOMEM;
		}
		list_add(&req->list, &dev->tx_reqs);
		MTP_DBG("req buf in=%p\n",req->buf);

	}

	for (i = 0; i < LEN; i++) {
		req = mtp_req_alloc(dev->out_ep, USB_BUFSIZE, GFP_KERNEL);
		if (!req) {
			while (!list_empty(&dev->rx_reqs)) {
				req = container_of(dev->rx_reqs.next,
						struct usb_request, list);
				list_del(&req->list);
				mtp_req_free(dev->out_ep, req);
			}
			return -ENOMEM;
		}
		list_add(&req->list, &dev->rx_reqs);
		if(!list_empty(&dev->rx_reqs))

		{
//                 req = container_of(dev->rx_reqs.next,
//	                                     struct usb_request, list);
		MTP_DBG("req buf1=%p\n",req->buf);

		MTP_DBG("(!list_empty(&dev->rx_reqs)).....\n");
		MTP_DBG("req buf2=%p\n",req->buf);
		}
		else
		 MTP_DBG("(null...............list_empty(&dev->rx_reqs)).....\n");
	}

#if 1
        if ((dev->req_recv_buf = kmalloc (MTP_EP0_RECV_MAX_SIZE, GFP_KERNEL)) == NULL) {
                MTP_PRINT("no buffer for it\n");
		goto fail;
        }
        memset (dev->req_recv_buf, 0, MTP_EP0_RECV_MAX_SIZE);
        dev->req_recv_start = dev->req_recv_buf;
        dev->req_recv_end = dev->req_recv_buf;

//	dev->req->complete = mtp_setup_complete;
#endif
	/* finish hookup to lower layer ... */
	dev->gadget = gadget;
	set_gadget_data(gadget, dev);
	gadget->ep0->driver_data = dev;
	//message->data=dev;
	MTP_DBG( "%s, version: " DRIVER_VERSION "\n", driver_desc);
	MTP_DBG( "using %s, OUT %s IN %s\n", gadget->name, out_ep->name,
			in_ep->name);

	return 0;

fail:
	MTP_DBG("error happen in bind\n");
	mtp_unbind(gadget);
	return status;
}

/*-------------------------------------------------------------------------*/

static struct usb_gadget_driver mtp_driver = {
//	.speed		= DEVSPEED,

#ifdef CONFIG_USB_GADGET_DUALSPEED
	.speed		= USB_SPEED_HIGH,
#else
	.speed		= USB_SPEED_FULL,
#endif
	.function	= (char *) longname,
	.bind		= mtp_bind,
	.unbind		= mtp_unbind,

	.setup		= mtp_setup,
	.disconnect	= mtp_disconnect,

	.driver		= {
		.name		= (char *) shortname,
		.owner		= THIS_MODULE,
	},
};

MODULE_DESCRIPTION(DRIVER_DESC);
MODULE_LICENSE("GPL");
static int __init
init(void)
{
	int status;
	
	usb_gadget_class = class_create(THIS_MODULE, "usb_mtp");
	if (IS_ERR(usb_gadget_class)) {
		status = PTR_ERR(usb_gadget_class);
		MTP_ERR("unable to create usb_gadget class %d\n", status);
		return status;
	}

	status = alloc_chrdev_region(&g_mtp_devno, 0, 1,
			"USB mtp gadget");
	if (status) {
		MTP_DBG("alloc_chrdev_region %d\n", status);
		class_destroy(usb_gadget_class);
		return status;
	}

	status = usb_gadget_register_driver(&mtp_driver);
	if (status) {
		class_destroy(usb_gadget_class);
		unregister_chrdev_region(g_mtp_devno, 1);
		MTP_DBG("usb_gadget_register_driver %x\n", status);
	}

	return status;
}
module_init(init);

static void __exit
cleanup(void)
{
	int status;
	spin_lock(&usb_mtp_gadget->lock_mtp_io);
	class_destroy(usb_gadget_class);
	unregister_chrdev_region(g_mtp_devno, 2);
	status = usb_gadget_unregister_driver(&mtp_driver);
	if (status)
		MTP_DBG("usb_gadget_unregister_driver %x\n", status);

	spin_unlock(&usb_mtp_gadget->lock_mtp_io);
}
module_exit(cleanup);
