/* drivers/video/backlight/levante_kp_bl.c
 *
 * Author:	yhanin
 * Created:	May 25, 2009
 * Copyright:	Marvell International Ltd. All Rights Reserved
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <linux/module.h>
#include <linux/init.h>
#include <linux/platform_device.h>
#include <linux/kernel.h>
#include <linux/spinlock.h>
#include <linux/backlight.h>
#include <linux/fb.h>

#include <mach/pxa-regs.h>
#include <mach/levante.h>
#include <mach/levante_hw.h>

#define KEYPAD_LED_MAX	100
#define	KEYPAD_LED_MIN	0

static int levante_kp_bl_suspended;
static int current_intensity = 0;

void kp_bl_power_on(unsigned char intensity)
{
	unsigned long flags;
	unsigned char val = 0x08;/*ISEt= 4ma ,boost is on automaticly if we enable one of the strings*/

	printk(LVT_DEBUG_MUST "------->kp_bl_power_on\n");

	local_irq_save(flags);
	if(intensity >= 0xFF)
	   val |= PORTOFINO_WLED_100DUTY;
	else
	   val &= ~PORTOFINO_WLED_100DUTY;

	portofino_write(PORTOFINO_WLED2B, val);
	local_irq_restore(flags);
}

void kp_bl_power_off(void)
{
	unsigned long flags;

	printk(LVT_DEBUG_MUST "------->kp_bl_power_off \n");

	local_irq_save(flags);
 	portofino_write(PORTOFINO_WLED2B, 0x00);
 	local_irq_restore(flags);
}

/* input: intensity in percentage 0% - 100% */
static int levante_kp_bl_send_intensity(struct backlight_device *bd)
{
	int user_intensity = bd->props.brightness;
	int plat_intensity = 0;

	if (bd->props.power != FB_BLANK_UNBLANK)
		user_intensity = 0;
	if (bd->props.fb_blank != FB_BLANK_UNBLANK)
		user_intensity = 0;
	if (levante_kp_bl_suspended)
		user_intensity = 0;

	if(user_intensity >= 100)
        plat_intensity = 0xFF;
     else
       plat_intensity = (user_intensity*256/100); // convert precentage to WLED_DUTY

	portofino_write(PORTOFINO_WLED2A, plat_intensity);

	printk(LVT_DEBUG_MUST \
	"------->levante_kp_bl_send_intensity = [%d]=[%d]\n", \
	user_intensity, plat_intensity);

	if (user_intensity && current_intensity == 0) {
		kp_bl_power_on(plat_intensity);
	} else if (user_intensity == 0 && current_intensity != 0) {
		kp_bl_power_off();
	}

	current_intensity = user_intensity;

	return 0;
}

#ifdef CONFIG_PM
static int levante_kp_bl_suspend(struct platform_device *pdev, pm_message_t state)
{
	struct backlight_device *bd = platform_get_drvdata(pdev);

	levante_kp_bl_suspended = 1;
	levante_kp_bl_send_intensity(bd);
	return 0;
}

static int levante_kp_bl_resume(struct platform_device *pdev)
{
	struct backlight_device *bd = platform_get_drvdata(pdev);

	levante_kp_bl_suspended = 0;
	levante_kp_bl_send_intensity(bd);
	return 0;
}
#else
#define levante_bl_suspend	NULL
#define levante_bl_resume	NULL
#endif

static int levante_bl_set_intensity(struct backlight_device *bd)
{
	levante_kp_bl_send_intensity(bd);
	return 0;
}

static int levante_bl_get_intensity(struct backlight_device *bd)
{
    printk(LVT_DEBUG_MUST "------->levante_kp_bl(levante_bl_get_intensity)= %d\n",current_intensity);
	return current_intensity;
}

static struct backlight_ops levante_bl_ops = {
	.get_brightness = levante_bl_get_intensity,
	.update_status  = levante_bl_set_intensity,
};

static int __init levante_kp_bl_probe(struct platform_device *pdev)
{
	struct backlight_device *bd;

	printk(LVT_DEBUG_MUST "------->levante_kp_bl_probe\n");

	bd = backlight_device_register("levante-kp-bl", &pdev->dev, NULL,
		    &levante_bl_ops);
	if (IS_ERR(bd))
		return PTR_ERR(bd);

	platform_set_drvdata(pdev, bd);

{
	unsigned char val= 0x00;
	/* turn on BOOST_EN to enable lcd backlight */
	/* allow WLED strings you need to allow reference voltage */
 	portofino_read(PORTOFINO_VSYS, &val);
	if(!(val & PORTOFINO_VSYS_ENABLE)){
		val |= PORTOFINO_VSYS_ENABLE;
		portofino_write(PORTOFINO_VSYS, val);
	}
	/* allow WLED strings you need to allow reference frequency*/
 	portofino_read(PORTOFINO_USB_MISC, &val);
	if(!(val & PRTOFINIO_USB_MISC_OSC_EN_BIT)){
		val |= PRTOFINIO_USB_MISC_OSC_EN_BIT;
		portofino_write(PORTOFINO_USB_MISC, val);
	}
}
	printk(LVT_DEBUG_MUST "------->levante_kp_bl_probe:PORTOFINO_VSYS+PORTOFINO_USB_MISC\n");

	bd->props.power = FB_BLANK_UNBLANK;
	bd->props.brightness = KEYPAD_LED_MAX;
	bd->props.max_brightness = KEYPAD_LED_MAX;
	levante_kp_bl_send_intensity(bd);

	return 0;
}

static int levante_kp_bl_remove(struct platform_device *pdev)
{
	struct backlight_device *bd = platform_get_drvdata(pdev);

	bd->props.brightness = 0;
	bd->props.power = 0;
	levante_kp_bl_send_intensity(bd);

	backlight_device_unregister(bd);

	return 0;
}

static struct platform_driver levante_bl_driver = {
	.probe		= levante_kp_bl_probe,
	.remove		= levante_kp_bl_remove,
#ifdef	CONFIG_PM
	.suspend	= levante_kp_bl_suspend,
	.resume		= levante_kp_bl_resume,
#endif
	.driver		= {
		.name	= "levante-kp-bl",
	},
};

static int __init levante_backlight_init(void)
{
	return platform_driver_register(&levante_bl_driver);
}

static void __exit levante_backlight_exit(void)
{
	platform_driver_unregister(&levante_bl_driver);
}

late_initcall(levante_backlight_init);
module_exit(levante_backlight_exit);
MODULE_LICENSE("GPL");
