/*
 *  linux/drivers/video/pxa950fb.c
 */

#include "pxa950fb.h"

#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/errno.h>
#include <linux/string.h>
#include <linux/interrupt.h>
#include <linux/slab.h>
#include <linux/fb.h>
#include <linux/delay.h>
#include <linux/init.h>
#include <linux/ioport.h>
#include <linux/cpufreq.h>
#include <linux/platform_device.h>
#include <linux/dma-mapping.h>
#include <linux/clk.h>
#include <linux/err.h>
#include <linux/uaccess.h>

#include <mach/io.h>
#include <mach/irqs.h>
#include <mach/hardware.h>
#include <mach/pxa3xx-regs.h>

#include <mach/gpio.h>
#include <mach/pxa950fb.h>
#include <mach/micco.h>

#ifdef CONFIG_PXA3xx_DVFM
#include <mach/dvfm.h>
#include <mach/pxa3xx_dvfm.h>
#endif

#define MAX_QUEUE_NUM   30
#define RESET_BUF       0x1
#define FREE_ENTRY      0x2

extern void pxa950fb_writel(unsigned int val, unsigned int *reg_base, unsigned int offset);

static int pxa950fb_set_par(struct fb_info *fi);
static int wait_for_vsync(struct pxa950fb_info *fbi);

static int pxa950fb_switch_buff(struct fb_info *fi);
static int addFreeBuf(u8 **bufList, u8 *freeBuf);
static void clearFreeBuf(u8 **bufList, int iFlag);
static void collectFreeBuf(u8 **filterList, u8 **freeList, int count);
static u8 *filterBufList[MAX_QUEUE_NUM];
static u8 *freeBufList[MAX_QUEUE_NUM];

static struct _sViewPortInfo gViewPortInfo = {
	.srcWidth = 480,	/* video source size */
	.srcHeight = 640,
	.zoomXSize = 480,	/* size after zooming */
	.zoomYSize = 640,
};

static struct _sViewPortOffset gViewPortOffset = {
	.xOffset = 0,	/* position on screen */
	.yOffset = 0
};

static volatile LCD_Controller_Configurations LCD_Conf;

u32 *ovly_dma_addr_cpu;
u32 ovly_dma_addr_dma;

pxa950fb_registers pxa950fb_ovly_regs;

#ifdef CONFIG_PXA3xx_DVFM
static int dvfm_dev_idx;
static void set_dvfm_constraint(void)
{
	/* Disable Lowpower mode */
	dvfm_disable_op_name("D0CS", dvfm_dev_idx);
	dvfm_disable_op_name("D1", dvfm_dev_idx);
	dvfm_disable_op_name("D2", dvfm_dev_idx);
	if (cpu_is_pxa935() || cpu_is_pxa950())
		dvfm_disable_op_name("CG", dvfm_dev_idx);
	/* Disable below frequency since they only support DSI clock up to 156Mhz */
	dvfm_disable_op_name("156M", dvfm_dev_idx);
	dvfm_disable_op_name("208M", dvfm_dev_idx);
	dvfm_disable_op_name("156M_HF", dvfm_dev_idx);
}

static void unset_dvfm_constraint(void)
{
	/* Enable Lowpower mode */
	dvfm_enable_op_name("D0CS", dvfm_dev_idx);
	dvfm_enable_op_name("D1", dvfm_dev_idx);
	dvfm_enable_op_name("D2", dvfm_dev_idx);
	if (cpu_is_pxa935() || cpu_is_pxa950())
		dvfm_enable_op_name("CG", dvfm_dev_idx);
	/* Enable frequency */
	dvfm_enable_op_name("156M", dvfm_dev_idx);
	dvfm_enable_op_name("208M", dvfm_dev_idx);
	dvfm_enable_op_name("156M_HF", dvfm_dev_idx);
}

#else
static void set_dvfm_constraint(void) {}
static void unset_dvfm_constraint(void) {}
#endif

static void set_fetch(struct pxa950fb_info *fbi, struct fb_var_screeninfo *var, int channel)
{
	/* set DMA descriptor register */
	DisplayControllerFrameDescriptor *dmadesc_cpu;
	uint dmadesc_dma;
	dmadesc_cpu = (DisplayControllerFrameDescriptor *)((unsigned int)ovly_dma_addr_cpu - 16*(channel+1));
	dmadesc_dma = ovly_dma_addr_dma - 16*(channel+1);
	dmadesc_cpu->LCD_FR_IDx = 0;
	dmadesc_cpu->LCD_CH_CMDx = (u32)(var->xres*var->yres*3/2);
	dmadesc_cpu->LCD_NXT_DESC_ADDRx = dmadesc_dma;

	if ((channel == 3) || (channel == 4))
		dmadesc_cpu->LCD_FR_ADDRx = ovly_dma_addr_dma;
	if ((channel == 5) && (fbi->pix_fmt > 3))
		dmadesc_cpu->LCD_FR_ADDRx = ovly_dma_addr_dma + var->xres*var->yres;
	if ((channel == 6) && (fbi->pix_fmt > 3)) {
		if (fbi->pix_fmt == PIX_FMTIN_YUV420)
			dmadesc_cpu->LCD_FR_ADDRx = ovly_dma_addr_dma + var->xres*var->yres*5/4;
		if (fbi->pix_fmt == PIX_FMTIN_YUV422)
			dmadesc_cpu->LCD_FR_ADDRx = ovly_dma_addr_dma + var->xres*var->yres*3/2;
		if (fbi->pix_fmt == PIX_FMTIN_YUV444)
			dmadesc_cpu->LCD_FR_ADDRx = ovly_dma_addr_dma + var->xres*var->yres*2;
	}
	pxa950fb_ovly_regs.descregs[channel].LCD_NXT_DESC_ADDRx = dmadesc_dma;
	/* set fetch registers */
	pxa950fb_ovly_regs.fetchregs[channel].LCD_FETCH_CTLx &= ~(LCD_FETCH_CTLx_SRC_FOR(0x7));
	/* set format */
	pxa950fb_ovly_regs.fetchregs[channel].LCD_FETCH_CTLx |= LCD_FETCH_CTLx_CHAN_EN|LCD_FETCH_CTLx_SRC_FOR(fbi->pix_fmt);
}

static void set_window(struct pxa950fb_info *fbi, struct fb_var_screeninfo *var)
{
	/* set input image resolution */
	int xzoom, yzoom;
	int xscale, yscale;
	int wgid;
	yzoom = fbi->surface.viewPortInfo.zoomYSize;
	xzoom = fbi->surface.viewPortInfo.zoomXSize;
	if (yzoom == 0)
		yzoom = var->yres;
	if (xzoom == 0)
		xzoom = var->xres;
	if (fbi->pix_fmt > 3)
		wgid = 4;
	else
		wgid = LCD_Conf.BP_ID - 0x10;
	pxa950fb_ovly_regs.wgregs[wgid].LCD_WINx_CTL &= ~(LCD_WINx_CTL_WIN_XRES(0x1ff)|LCD_WINx_CTL_WIN_YRES(0x1ff));
	pxa950fb_ovly_regs.wgregs[wgid].LCD_WINx_CTL = LCD_WINx_CTL_WIN_XRES(var->xres/4)|LCD_WINx_CTL_WIN_YRES(var->yres/4);
	if (xzoom < var->xres) {
		pxa950fb_ovly_regs.wgregs[wgid].LCD_WINx_CFG &= ~(1<<24);
		xscale = (100000 - xzoom*100000/(var->xres))/3125 - 1;
	} else {
		pxa950fb_ovly_regs.wgregs[wgid].LCD_WINx_CFG |= 1<<24;
		xscale = (xzoom*1000/(var->xres) - 1000)/125;
	}
	if (yzoom < var->yres) {
		pxa950fb_ovly_regs.wgregs[wgid].LCD_WINx_CFG &= ~(1<<18);
		yscale = (100000 - yzoom*100000/(var->yres))/3125 - 1;
	} else {
		pxa950fb_ovly_regs.wgregs[wgid].LCD_WINx_CFG |= 1<<18;
		yscale = (yzoom*1000/(var->yres) - 1000)/125;
	}
	pxa950fb_ovly_regs.wgregs[wgid].LCD_WINx_CFG &= ~(LCD_WINx_CFG_WIN_XSCALE(0x1f)|LCD_WINx_CFG_WIN_YSCALE(0x1f));
	pxa950fb_ovly_regs.wgregs[wgid].LCD_WINx_CFG |= LCD_WINx_CFG_WIN_XSCALE(xscale)|LCD_WINx_CFG_WIN_YSCALE(yscale);
}

static void set_mixer(struct pxa950fb_info *fbi, struct fb_var_screeninfo *var, u8 overlay)
{
	uint LCD_MIXERx_CTL0_Value;
	int index;
	int x, y;

	x = fbi->surface.viewPortInfo.zoomXSize ? fbi->surface.viewPortInfo.zoomXSize : var->xres;
	y = fbi->surface.viewPortInfo.zoomYSize ? fbi->surface.viewPortInfo.zoomYSize : var->yres;

	for (index = 0; index < 2; index++) {
		if ((index == 0) && (LCD_Conf.LCD_Mixer[0] == LCD_MIXER_DISABLE))
			continue;
		if ((index == 1) && (LCD_Conf.LCD_Mixer[1] == LCD_MIXER_DISABLE))
			continue;
		pxa950fb_ovly_regs.mixerregs[index].LCD_MIXERx_CTL0 =
			readl(fbi->reg_base + ((index == 0) ? LCD_MIXER0_CTL0 : LCD_MIXER1_CTL0));
		switch (overlay) {
			case LCD_Controller_Base_Plane:
//				pxa950fb_ovly_regs.mixerregs[index].LCD_MIXERx_CTL1 &= ~(LCD_MIXERx_CTL1_DISP_XRES(0x7ff) | LCD_MIXERx_CTL1_DISP_YRES(0x7ff));
//				pxa950fb_ovly_regs.mixerregs[index].LCD_MIXERx_CTL1 |= LCD_MIXERx_CTL1_DISP_XRES(var->xres) | LCD_MIXERx_CTL1_DISP_YRES(var->yres);
				/* base plane is always full screen */
				pxa950fb_ovly_regs.mixerregs[index].LCD_MIXERx_BP_CFG0 &= ~(LCD_MIXERx_CFG0_XPOS(0x7ff) | LCD_MIXERx_CFG0_YPOS(0x7ff));
//				pxa950fb_ovly_regs.mixerregs[index].LCD_MIXERx_BP_CFG0 |= LCD_MIXERx_CFG0_XPOS(LCD_Conf.LCD_Position_x) | LCD_MIXERx_CFG0_YPOS(LCD_Conf.LCD_Position_y);
				pxa950fb_ovly_regs.mixerregs[index].LCD_MIXERx_BP_CFG0 |= LCD_MIXERx_CFG0_XPOS(fbi->surface.viewPortOffset.xOffset) | LCD_MIXERx_CFG0_YPOS(fbi->surface.viewPortOffset.yOffset);
				pxa950fb_ovly_regs.mixerregs[index].LCD_MIXERx_BP_CFG1 &= ~(LCD_MIXERx_CFG1_XRES(0x7ff) | LCD_MIXERx_CFG1_YRES(0x7ff));
//				pxa950fb_ovly_regs.mixerregs[index].LCD_MIXERx_BP_CFG1 |= LCD_MIXERx_CFG1_XRES(var->xres) | LCD_MIXERx_CFG1_YRES(var->yres);
				pxa950fb_ovly_regs.mixerregs[index].LCD_MIXERx_BP_CFG1 |= LCD_MIXERx_CFG1_XRES(x) | LCD_MIXERx_CFG1_YRES(y);
				LCD_MIXERx_CTL0_Value = LCD_MIXERx_CTL0_BP_ID(LCD_Conf.BP_ID);
				pxa950fb_ovly_regs.mixerregs[index].LCD_MIXERx_CTL0 &= ~LCD_MIXERx_CTL0_BP_ID_MASK;
				pxa950fb_ovly_regs.mixerregs[index].LCD_MIXERx_CTL0 |= LCD_MIXERx_CTL0_Value | LCD_MIXERx_CTL0_MIX_EN;
				break;
			case LCD_Controller_Overlay1:

				pxa950fb_ovly_regs.mixerregs[index].LCD_MIXERx_OL1_CFG0 &= ~(LCD_MIXERx_CFG0_XPOS(0x7ff) | LCD_MIXERx_CFG0_YPOS(0x7ff));
				pxa950fb_ovly_regs.mixerregs[index].LCD_MIXERx_OL1_CFG0 |= LCD_MIXERx_CFG0_XPOS(fbi->surface.viewPortOffset.xOffset) | LCD_MIXERx_CFG0_YPOS(fbi->surface.viewPortOffset.yOffset);
				pxa950fb_ovly_regs.mixerregs[index].LCD_MIXERx_OL1_CFG1 &= ~(LCD_MIXERx_CFG1_XRES(0x7ff) | LCD_MIXERx_CFG1_YRES(0x7ff));
				pxa950fb_ovly_regs.mixerregs[index].LCD_MIXERx_OL1_CFG1 |= LCD_MIXERx_CFG1_XRES(x) | LCD_MIXERx_CFG1_YRES(y);
				LCD_MIXERx_CTL0_Value = LCD_MIXERx_CTL0_OL1_ID(LCD_Conf.OL1_ID);
				pxa950fb_ovly_regs.mixerregs[index].LCD_MIXERx_CTL0 &= ~LCD_MIXERx_CTL0_OL1_ID_MASK;
				/* 1. 	remove LCD_MIXERx_CTL0_DISP_UPDATE here so user must call
				   ioctl FB_IOCTL_SWITCH_VID_OVLY to turn on/off overlay,
				   otherwise overlay will display on once initialized.
				   2. 	removed LCD_MIXERx_CTL0_MIX_EN here since it is enabled in
				   base layer and should always on */
				pxa950fb_ovly_regs.mixerregs[index].LCD_MIXERx_CTL0 |= LCD_MIXERx_CTL0_Value;
				break;
			case LCD_Controller_Overlay2:
				pxa950fb_ovly_regs.mixerregs[index].LCD_MIXERx_OL2_CFG0 &= ~(LCD_MIXERx_CFG0_XPOS(0x7ff) | LCD_MIXERx_CFG0_YPOS(0x7ff));
				pxa950fb_ovly_regs.mixerregs[index].LCD_MIXERx_OL2_CFG0 |= LCD_MIXERx_CFG0_XPOS(fbi->surface.viewPortOffset.xOffset) | LCD_MIXERx_CFG0_YPOS(fbi->surface.viewPortOffset.yOffset);
				pxa950fb_ovly_regs.mixerregs[index].LCD_MIXERx_OL2_CFG1 &= ~(LCD_MIXERx_CFG1_XRES(0x7ff) | LCD_MIXERx_CFG1_YRES(0x7ff));
				pxa950fb_ovly_regs.mixerregs[index].LCD_MIXERx_OL2_CFG1 |= LCD_MIXERx_CFG1_XRES(x) | LCD_MIXERx_CFG1_YRES(y);
				LCD_MIXERx_CTL0_Value = LCD_MIXERx_CTL0_OL2_ID(LCD_Conf.OL2_ID);
				pxa950fb_ovly_regs.mixerregs[index].LCD_MIXERx_CTL0 &= ~LCD_MIXERx_CTL0_OL2_ID_MASK;
				pxa950fb_ovly_regs.mixerregs[index].LCD_MIXERx_CTL0 |= LCD_MIXERx_CTL0_Value;
				break;
			case LCD_Controller_Overlay3:
				pxa950fb_ovly_regs.mixerregs[index].LCD_MIXERx_OL3_CFG0 &= ~(LCD_MIXERx_CFG0_XPOS(0x7ff) | LCD_MIXERx_CFG0_YPOS(0x7ff));
				pxa950fb_ovly_regs.mixerregs[index].LCD_MIXERx_OL3_CFG0 |= LCD_MIXERx_CFG0_XPOS(LCD_Conf.LCD_Position_x) | LCD_MIXERx_CFG0_YPOS(LCD_Conf.LCD_Position_y);
				pxa950fb_ovly_regs.mixerregs[index].LCD_MIXERx_OL3_CFG1 &= ~(LCD_MIXERx_CFG1_XRES(0x7ff) | LCD_MIXERx_CFG1_YRES(0x7ff));
				pxa950fb_ovly_regs.mixerregs[index].LCD_MIXERx_OL3_CFG1 |= LCD_MIXERx_CFG1_XRES(LCD_Conf.LCD_Horizontal) | LCD_MIXERx_CFG1_YRES(LCD_Conf.LCD_Vertical);
				LCD_MIXERx_CTL0_Value = LCD_MIXERx_CTL0_OL3_ID(LCD_Conf.OL3_ID);
				pxa950fb_ovly_regs.mixerregs[index].LCD_MIXERx_CTL0 &= ~LCD_MIXERx_CTL0_OL3_ID_MASK;
				pxa950fb_ovly_regs.mixerregs[index].LCD_MIXERx_CTL0 |= LCD_MIXERx_CTL0_Value;
				break;

		}
	}
}

static void set_configuration(struct pxa950fb_info *fbi)
{
	struct pxa950fb_mach_info *mi = fbi->dev->platform_data;

	/* Modify Configuration structre */
	LCD_Conf.LCD_FRAME_BUFFER = (u32)ovly_dma_addr_dma;	/* TODO */

	LCD_Conf.LCD_Horizontal = 480;
	LCD_Conf.LCD_Vertical = 640;

	LCD_Conf.LCD_Position_x = 0;
	LCD_Conf.LCD_Position_y = 0;

	/*
	 * parallel setting don't depend on below LCD_Conf.LCD_XXW
	 * since mi->modes are copied to var.
	 * TODO: need to think about the using mode for DSI
	 */
	LCD_Conf.LCD_HSW = mi->modes->hsync_len;
	LCD_Conf.LCD_VSW = mi->modes->vsync_len;
	LCD_Conf.LCD_BLW = mi->modes->right_margin;
	LCD_Conf.LCD_ELW = mi->modes->left_margin;
	LCD_Conf.LCD_BFW = mi->modes->lower_margin;
	LCD_Conf.LCD_EFW = mi->modes->upper_margin;

	LCD_Conf.LCD_BPP = BYTES_PER_PIXEL;
	LCD_Conf.LCD_Display = mi->panel_type;
	LCD_Conf.LCD_OP_FOR = mi->pix_fmt_out;
	LCD_Conf.LCD_SRC_FOR = mi->pix_fmt_in;
	/*
	 * NOTE:
	 * currently set WGe4/WGe3 (overlay layer) as BP_ID in mixer
	 * to make sure overlay is behind of base (WGe0)
	 */
	LCD_Conf.LCD_Overlay = LCD_Controller_Base_Plane;
	LCD_Conf.BP_ID = LCD_Controller_WGe4;
	LCD_Conf.OL1_ID = LCD_Controller_WGe0;

	LCD_Conf.LCD_Mixer[0] = mi->mixer0;
	LCD_Conf.LCD_Mixer[1] = mi->mixer1;
	return;
}

static void pxa950fb_enable_controller(struct pxa950fb_info *fbi)
{
	if (fbi->pix_fmt == 7) {
		pxa950fb_writel(pxa950fb_ovly_regs.descregs[4].LCD_NXT_DESC_ADDRx,
				fbi->reg_base, LCD_NXT_DESC_ADDR4);
		pxa950fb_writel(pxa950fb_ovly_regs.fetchregs[4].LCD_FETCH_CTLx,
				fbi->reg_base, LCD_FETCH_CTL4);
		pxa950fb_writel(pxa950fb_ovly_regs.wgregs[4].LCD_WINx_CTL,
				fbi->reg_base, LCD_WIN4_CTL);
		pxa950fb_writel(pxa950fb_ovly_regs.wgregs[4].LCD_WINx_CFG,
				fbi->reg_base, LCD_WIN4_CFG);
	} else {

		if (fbi->pix_fmt > 3) {
			pxa950fb_writel(pxa950fb_ovly_regs.descregs[4].LCD_NXT_DESC_ADDRx,
					fbi->reg_base, LCD_NXT_DESC_ADDR4);
			pxa950fb_writel(pxa950fb_ovly_regs.descregs[5].LCD_NXT_DESC_ADDRx,
					fbi->reg_base, LCD_NXT_DESC_ADDR5);
			pxa950fb_writel(pxa950fb_ovly_regs.descregs[6].LCD_NXT_DESC_ADDRx,
					fbi->reg_base, LCD_NXT_DESC_ADDR6);

			pxa950fb_writel(pxa950fb_ovly_regs.fetchregs[4].LCD_FETCH_CTLx,
					fbi->reg_base, LCD_FETCH_CTL4);
			pxa950fb_writel(pxa950fb_ovly_regs.fetchregs[5].LCD_FETCH_CTLx,
					fbi->reg_base, LCD_FETCH_CTL5);
			pxa950fb_writel(pxa950fb_ovly_regs.fetchregs[6].LCD_FETCH_CTLx,
					fbi->reg_base, LCD_FETCH_CTL6);

			pxa950fb_writel(pxa950fb_ovly_regs.wgregs[4].LCD_WINx_CTL,
					fbi->reg_base, LCD_WIN4_CTL);
			pxa950fb_writel(pxa950fb_ovly_regs.wgregs[4].LCD_WINx_CFG,
					fbi->reg_base, LCD_WIN4_CFG);

		} else {
			pxa950fb_writel(pxa950fb_ovly_regs.descregs[3].LCD_NXT_DESC_ADDRx,
					fbi->reg_base, LCD_NXT_DESC_ADDR3);
			pxa950fb_writel(pxa950fb_ovly_regs.fetchregs[3].LCD_FETCH_CTLx,
					fbi->reg_base, LCD_FETCH_CTL3);
			pxa950fb_writel(pxa950fb_ovly_regs.wgregs[3].LCD_WINx_CTL,
					fbi->reg_base, LCD_WIN3_CTL);
			pxa950fb_writel(pxa950fb_ovly_regs.wgregs[3].LCD_WINx_CFG,
					fbi->reg_base, LCD_WIN3_CFG);
		}
	}
#if 0
	if (LCD_Conf.LCD_Mixer[0] != LCD_MIXER_DISABLE) {
		pxa950fb_writel(pxa950fb_ovly_regs.mixerregs[0].LCD_MIXERx_OL1_CFG0,
				fbi->reg_base, LCD_MIXER0_OL1_CFG0);
		pxa950fb_writel(pxa950fb_ovly_regs.mixerregs[0].LCD_MIXERx_OL1_CFG1,
				fbi->reg_base, LCD_MIXER0_OL1_CFG1);
		pxa950fb_writel(pxa950fb_ovly_regs.mixerregs[0].LCD_MIXERx_CTL0,
				fbi->reg_base, LCD_MIXER0_CTL0);
	}
	if (LCD_Conf.LCD_Mixer[1] != LCD_MIXER_DISABLE) {
		pxa950fb_writel(pxa950fb_ovly_regs.mixerregs[1].LCD_MIXERx_OL1_CFG0,
				fbi->reg_base, LCD_MIXER1_OL1_CFG0);
		pxa950fb_writel(pxa950fb_ovly_regs.mixerregs[1].LCD_MIXERx_OL1_CFG1,
				fbi->reg_base, LCD_MIXER1_OL1_CFG1);
		pxa950fb_writel(pxa950fb_ovly_regs.mixerregs[1].LCD_MIXERx_CTL0,
				fbi->reg_base, LCD_MIXER1_CTL0);
	}
#endif
	if (LCD_Conf.LCD_Mixer[0] != LCD_MIXER_DISABLE) {
		pxa950fb_writel(pxa950fb_ovly_regs.mixerregs[0].LCD_MIXERx_BP_CFG0,
				fbi->reg_base, LCD_MIXER0_BP_CFG0);
		pxa950fb_writel(pxa950fb_ovly_regs.mixerregs[0].LCD_MIXERx_BP_CFG1,
				fbi->reg_base, LCD_MIXER0_BP_CFG1);
		pxa950fb_writel(pxa950fb_ovly_regs.mixerregs[0].LCD_MIXERx_CTL0,
				fbi->reg_base, LCD_MIXER0_CTL0);
	}
	if (LCD_Conf.LCD_Mixer[1] != LCD_MIXER_DISABLE) {
		pxa950fb_writel(pxa950fb_ovly_regs.mixerregs[1].LCD_MIXERx_BP_CFG0,
				fbi->reg_base, LCD_MIXER1_BP_CFG0);
		pxa950fb_writel(pxa950fb_ovly_regs.mixerregs[1].LCD_MIXERx_BP_CFG1,
				fbi->reg_base, LCD_MIXER1_BP_CFG1);
		pxa950fb_writel(pxa950fb_ovly_regs.mixerregs[1].LCD_MIXERx_CTL0,
				fbi->reg_base, LCD_MIXER1_CTL0);
	}
}

#define MAX_HWC_SIZE		(64*64*2)
#define MAX_REFRESH_RATE	100	/* Hz */
#define MIN_REFRESH_RATE	47	/* Hz */
#define DEFAULT_REFRESH		60	/* Hz */

static void pxa950fb_do_tasklet(unsigned long data)
{
	struct fb_info *fi;

	fi = (struct fb_info *)data;
	pxa950fb_switch_buff(fi);
}

static int determine_best_pix_fmt(struct fb_var_screeninfo *var)
{
	/*
	 * Pseudocolor mode?
	 */
	if (var->bits_per_pixel == 8)
		return PIX_FMT_PSEUDOCOLOR;

	/*
	 * Check for YUV422PLANAR.
	 */
	if (var->bits_per_pixel == 16 && var->red.length == 8 &&
			var->green.length == 4 && var->blue.length == 4) {
		if (var->red.offset >= var->blue.offset)
			return PIX_FMTIN_YUV422;
		else
			printk("%s: YVU422PLANAR format is not supported!\n", __func__);
	}

	/*
	 * Check for YUV420PLANAR.
	 */
	if (var->bits_per_pixel == 12 && var->red.length == 8 &&
			var->green.length == 2 && var->blue.length == 2) {
		if (var->red.offset >= var->blue.offset)
			return PIX_FMTIN_YUV420;
		else
			printk("%s: YVU420PLANAR format is not supported!\n", __func__);
	}

	/*
	 * Check for YUV444PLANAR.
	 */
	if (var->bits_per_pixel == 24 && var->red.length == 8 &&
			var->green.length == 8 && var->blue.length == 8) {
		if (var->red.offset >= var->blue.offset)
			return PIX_FMTIN_YUV444;
		else
			printk("%s: YVU420PLANAR format is not supported!\n", __func__);
	}
	/*
	 * Check for YUV422PACK.
	 */
	if (var->bits_per_pixel == 16 && var->red.length == 16 &&
			var->green.length == 16 && var->blue.length == 16) {
		if ((var->red.offset >= var->blue.offset) && (var->red.offset == 8))
			return PIX_FMTIN_YUV422IL;
		else
			printk("%s: YVU422PACK or UYVY format is not supported!\n", __func__);
	}

	/*
	 * Check for 565/1555.
	 */
	if (var->bits_per_pixel == 16 && var->red.length <= 5 &&
			var->green.length <= 6 && var->blue.length <= 5) {
		if (var->transp.length == 0) {
			if (var->red.offset >= var->blue.offset)
				return PIX_FMTIN_RGB_16;
			else
				printk("%s: BGR565 format is not supported!\n", __func__);
		}
	}

	/*
	 * Check for 888/A888.
	 */
	if (var->bits_per_pixel <= 32 && var->red.length <= 8 &&
			var->green.length <= 8 && var->blue.length <= 8) {
		if (var->bits_per_pixel == 24 && var->transp.length == 0) {
			if (var->red.offset >= var->blue.offset)
				return PIX_FMTIN_RGB_24_PACK;
			else
				printk("%s: BGR888 packed format is not supported!\n", __func__);
		}

		if (var->bits_per_pixel == 32 && var->transp.length == 8) {
			if (var->red.offset >= var->blue.offset)
				return PIX_FMTIN_RGB_32;
			else
				printk("%s: BGRA888 format is not supported!\n", __func__);
		} else {
			if (var->red.offset >= var->blue.offset)
				return PIX_FMTIN_RGB_24;
			else
				printk("%s: BGR888 format is not supported!\n", __func__);;
		}

		/* fall through */
	}

	return -EINVAL;


}

static void set_pix_fmt(struct fb_var_screeninfo *var, int pix_fmt)
{
	switch (pix_fmt) {
		case PIX_FMTIN_RGB_16:
			var->bits_per_pixel = 16;
			var->red.offset = 11;    var->red.length = 5;
			var->green.offset = 5;   var->green.length = 6;
			var->blue.offset = 0;    var->blue.length = 5;
			var->transp.offset = 0;  var->transp.length = 0;
			break;
		case PIX_FMTIN_RGB_24:
			var->bits_per_pixel = 32;
			var->red.offset = 16;    var->red.length = 8;
			var->green.offset = 8;   var->green.length = 8;
			var->blue.offset = 0;    var->blue.length = 8;
			var->transp.offset = 0;  var->transp.length = 0;
			break;
		case PIX_FMTIN_RGB_32:
			var->bits_per_pixel = 32;
			var->red.offset = 16;    var->red.length = 8;
			var->green.offset = 8;   var->green.length = 8;
			var->blue.offset = 0;    var->blue.length = 8;
			var->transp.offset = 24; var->transp.length = 8;
			break;
		case PIX_FMTIN_RGB_24_PACK:
			var->bits_per_pixel = 24;
			var->red.offset = 16;    var->red.length = 8;
			var->green.offset = 8;   var->green.length = 8;
			var->blue.offset = 0;    var->blue.length = 8;
			var->transp.offset = 0;  var->transp.length = 0;
			break;
		case PIX_FMTIN_YUV420:
			var->bits_per_pixel = 12;
			var->red.offset = 4;     var->red.length = 8;
			var->green.offset = 2;   var->green.length = 2;
			var->blue.offset = 0;   var->blue.length = 2;
			var->transp.offset = 0;  var->transp.length = 0;
			break;
		case PIX_FMTIN_YUV422:
			var->bits_per_pixel = 16;
			var->red.offset = 8;     var->red.length = 8;
			var->green.offset = 4;   var->green.length = 4;
			var->blue.offset = 0;   var->blue.length = 4;
			var->transp.offset = 0;  var->transp.length = 0;
			break;
		case PIX_FMTIN_YUV444:
			var->bits_per_pixel = 16;
			var->red.offset = 16;     var->red.length = 8;
			var->green.offset = 8;   var->green.length = 8;
			var->blue.offset = 0;   var->blue.length = 8;
			var->transp.offset = 0;  var->transp.length = 0;
			break;
		case PIX_FMTIN_YUV422IL:
			var->bits_per_pixel = 16;
			var->red.offset = 8;     var->red.length = 16;
			var->green.offset = 4;   var->green.length = 16;
			var->blue.offset = 0;   var->blue.length = 16;
			var->transp.offset = 0;  var->transp.length = 0;
			break;
		case PIX_FMT_PSEUDOCOLOR:
			var->bits_per_pixel = 8;
			var->red.offset = 0;     var->red.length = 8;
			var->green.offset = 0;   var->green.length = 8;
			var->blue.offset = 0;    var->blue.length = 8;
			var->transp.offset = 0;  var->transp.length = 0;
			break;
	}
}


static int convert_pix_fmt(u32 vmode)
{
	/*      printk(KERN_INFO "vmode=%d\n", vmode); */
	switch (vmode) {
		case FB_VMODE_YUV422PACKED:
			return PIX_FMTIN_YUV422IL;
		case FB_VMODE_YUV422PLANAR:
			return PIX_FMTIN_YUV422;
		case FB_VMODE_YUV420PLANAR:
			return PIX_FMTIN_YUV420;
			/* TODO - add U/V and R/B SWAP format and YUV444 */
		case FB_VMODE_RGB565:
			return PIX_FMTIN_RGB_16;
		case FB_VMODE_RGB888PACK:
			return PIX_FMTIN_RGB_24_PACK;
		case FB_VMODE_RGBA888:
			return PIX_FMTIN_RGB_32;
		case FB_VMODE_RGB888UNPACK:
			return PIX_FMTIN_RGB_24;
		default:
			return -1;
	}
}

static u32 pxa168fb_ovly_set_colorkeyalpha(struct pxa950fb_info *fbi)
{
	unsigned int x = 0;
	struct _sColorKeyNAlpha *color_a = &fbi->ckey_alpha;

	dev_dbg(fbi->fb_info->dev, "Enter %s\n", __FUNCTION__);

	if (color_a->win_alpha_en)
		/* YCbCr formate can only select window specific alpha */
		x = LCD_CHx_ALPHA_W_ALPHA_EN | LCD_CHx_ALPHA_CH_ALPHA(color_a->default_alpha_val);
	else {
		/*
		 * color_a->alpha_method = 0: color match alpha method (only for RGB16 and RGB24)	
		 * color_a->alpha_method = 1: lower bit alpha method (only for RGB16 and RGB24)
		 * color_a->alpha_method = 2: source alpha method (only for RGB32)
		 */
		if ((LCD_CHx_ALPHA_CLR_KEY_EN(color_a->alpha_method)) == 0) {
			x = LCD_CHx_ALPHA_CH_ALPHA(color_a->default_alpha_val);
			if (LCD_Conf.BP_ID == LCD_Controller_WGe4)
				pxa950fb_writel(color_a->color_match, fbi->reg_base, LCD_CH4_CLR_MATCH);
			else if (LCD_Conf.BP_ID == LCD_Controller_WGe3)
				pxa950fb_writel(color_a->color_match, fbi->reg_base, LCD_CH3_CLR_MATCH);
		} else
			x = LCD_CHx_ALPHA_CLR_KEY_EN(color_a->alpha_method);
	}
	if (LCD_Conf.BP_ID == LCD_Controller_WGe4)
		pxa950fb_writel(x, fbi->reg_base, LCD_CH4_ALPHA);
	else if (LCD_Conf.BP_ID == LCD_Controller_WGe3)
		pxa950fb_writel(x, fbi->reg_base, LCD_CH3_ALPHA);
	return 0;
}

static int check_surface(struct fb_info *fi,
		FBVideoMode new_mode,
		struct _sViewPortInfo *new_info,
		struct _sViewPortOffset *new_offset,
		struct _sVideoBufferAddr *new_addr)
{
	struct pxa950fb_info *fbi = (struct pxa950fb_info *)fi->par;
	struct fb_var_screeninfo *var = &fi->var;
	int changed = 0;

	dev_dbg(fi->dev, "Enter %s\n", __FUNCTION__);

	/*
	 * check mode
	 */
	if (new_mode >= 0 && fbi->surface.videoMode != new_mode) {
		fbi->surface.videoMode = new_mode;
		fbi->pix_fmt = convert_pix_fmt(new_mode);
		set_pix_fmt(var, fbi->pix_fmt);
		changed = 1;
	}

	/*
	 * check view port settings.
	 */
	if (new_info &&
			(fbi->surface.viewPortInfo.srcWidth != new_info->srcWidth ||
			 fbi->surface.viewPortInfo.srcHeight != new_info->srcHeight ||
			 fbi->surface.viewPortInfo.zoomXSize != new_info->zoomXSize ||
			 fbi->surface.viewPortInfo.zoomYSize != new_info->zoomYSize)) {
		if (!(new_addr && new_addr->startAddr)) {
			if (((new_info->srcWidth * new_info->srcHeight * var->bits_per_pixel / 8) * 2) > DEFAULT_FB_SIZE) {
				printk("%s: requested memory buffer size %d exceed the max limit %d!\n", __func__,
						(new_info->srcWidth * new_info->srcHeight * var->bits_per_pixel / 4), DEFAULT_FB_SIZE);
				return changed;
			}
		}
		var->xres_virtual = new_info->srcWidth;
		var->yres_virtual = new_info->srcHeight * 2;
		var->xres = new_info->srcWidth;
		var->yres = new_info->srcHeight;
		fbi->surface.viewPortInfo = *new_info;
		set_pix_fmt(var, fbi->pix_fmt);
		changed = 1;
	}

	/*
	 * Check offset
	 */
	if (new_offset &&
			(fbi->surface.viewPortOffset.xOffset != new_offset->xOffset ||
			 fbi->surface.viewPortOffset.yOffset != new_offset->yOffset)) {
		fbi->surface.viewPortOffset.xOffset = new_offset->xOffset;
		fbi->surface.viewPortOffset.yOffset = new_offset->yOffset;
		changed = 1;
	}

	/*
	 * Check buffer address
	 */
	if (new_addr && new_addr->startAddr &&
			fbi->new_addr != (unsigned long)new_addr->startAddr) {
		fbi->new_addr = (unsigned long)new_addr->startAddr;
		changed = 1;
	}

	return changed;
}

static int pxa950fb_ioctl(struct fb_info *fi, unsigned int cmd,
		unsigned long arg)
{
	void __user *argp = (void __user *)arg;
	struct pxa950fb_info *fbi = (struct pxa950fb_info *)fi->par;
	u32 x;
	int vmode = 0;
	int gfx_on = 1;
	int vid_on = 1;

	switch (cmd) {
		case FB_IOCTL_WAIT_VSYNC:
			wait_for_vsync(fbi);
			break;
		case FB_IOCTL_GET_VIEWPORT_INFO:
			return copy_to_user(argp, &fbi->surface.viewPortInfo,
					sizeof(struct _sViewPortInfo)) ? -EFAULT : 0;
		case FB_IOCTL_SET_VIEWPORT_INFO:
			mutex_lock(&fbi->access_ok);
			if (copy_from_user(&gViewPortInfo, argp,
						sizeof(gViewPortInfo))) {
				mutex_unlock(&fbi->access_ok);
				return -EFAULT;
			}

			if (check_surface(fi, -1, &gViewPortInfo, 0, 0)) {
				pxa950fb_set_par(fi);
			}

			mutex_unlock(&fbi->access_ok);
			break;
		case FB_IOCTL_SET_VIDEO_MODE:
			/*
			 * Get data from user space.
			 */
			if (copy_from_user(&vmode, argp, sizeof(vmode)))
				return -EFAULT;

			if (check_surface(fi, vmode, 0, 0, 0)) {
				pxa950fb_set_par(fi);
			}
			break;
		case FB_IOCTL_GET_VIDEO_MODE:
			return copy_to_user(argp, &fbi->surface.videoMode,
					sizeof(u32)) ? -EFAULT : 0;
		case FB_IOCTL_FLIP_VID_BUFFER:
			{
				struct _sOvlySurface *surface = 0;
				u8 *start_addr, *input_data, *dst_addr;
				u32 length;
				surface = kmalloc(sizeof(struct _sOvlySurface),
						GFP_KERNEL);

				/* Get user-mode data. */
				if (copy_from_user(surface, argp,
							sizeof(struct _sOvlySurface))) {
					kfree(surface);
					return -EFAULT;
				}

				length = surface->videoBufferAddr.length;
				dst_addr = fbi->surface.videoBufferAddr.startAddr;
				start_addr = surface->videoBufferAddr.startAddr;
				input_data = surface->videoBufferAddr.inputData;

				/*
				 * Has DMA addr?
				 */
				if (start_addr &&
						(!input_data)) {
					if (0 != addFreeBuf(freeBufList, (u8 *)surface)) {
						printk(KERN_INFO "Error: addFreeBuf()\n");
						mutex_unlock(&fbi->access_ok);
						kfree(surface);
						return -EFAULT;
					}
				} else {
					if (check_surface(fi, surface->videoMode,
								&surface->viewPortInfo,
								&surface->viewPortOffset,
								&surface->videoBufferAddr)) {
						pxa950fb_set_par(fi);
					}

					/* copy buffer */
					if (input_data) {
						wait_for_vsync(fbi);
						/* if support hw DMA, replace this. */
						if (copy_from_user(fbi->fb_start,
									input_data, length))
							return -EFAULT;
						return 0;
					}

					/*
					 * if it has its own physical address,
					 * switch to this memory.
					 */
					if (start_addr) {
						if (fbi->mem_status)
							free_pages(
									(unsigned long)fbi->fb_start,
									get_order(fbi->fb_size));
						else
							dma_free_writecombine(fbi->dev,
									fbi->fb_size,
									fbi->fb_start,
									fbi->fb_start_dma);

						fbi->fb_start = __va(start_addr);
						fbi->fb_size = length;
						fbi->fb_start_dma =
							(dma_addr_t)__pa(fbi->fb_start);
						fbi->mem_status = 1;
						fi->fix.smem_start = fbi->fb_start_dma;
						fi->fix.smem_len = fbi->fb_size;
						fi->screen_base = fbi->fb_start;
						fi->screen_size = fbi->fb_size;
					}

					kfree(surface);
				}

				mutex_unlock(&fbi->access_ok);
				return 0;
			}
		case FB_IOCTL_GET_FREELIST:
			{
				mutex_lock(&fbi->access_ok);

				if (copy_to_user(argp, filterBufList,
							MAX_QUEUE_NUM*sizeof(u8 *))) {
					mutex_unlock(&fbi->access_ok);
					return -EFAULT;
				}

				clearFreeBuf(filterBufList, RESET_BUF);

				mutex_unlock(&fbi->access_ok);
				return 0;
			}
		case FB_IOCTL_GET_BUFF_ADDR:
			{
				return copy_to_user(argp, &fbi->surface.videoBufferAddr,
						sizeof(struct _sVideoBufferAddr)) ? -EFAULT : 0;
			}
		case FB_IOCTL_SET_VID_OFFSET:
			mutex_lock(&fbi->access_ok);
			if (copy_from_user(&gViewPortOffset, argp,
						sizeof(gViewPortOffset))) {
				mutex_unlock(&fbi->access_ok);
				return -EFAULT;
			}

			if (check_surface(fi, -1, 0, &gViewPortOffset, 0)) {
				pxa950fb_set_par(fi);
			}
			mutex_unlock(&fbi->access_ok);
			break;
		case FB_IOCTL_GET_VID_OFFSET:
			return copy_to_user(argp, &fbi->surface.viewPortOffset,
					sizeof(struct _sViewPortOffset)) ? -EFAULT : 0;
		case FB_IOCTL_SET_MEMORY_TOGGLE:
			break;
		case FB_IOCTL_SET_COLORKEYnALPHA:
			if (copy_from_user(&fbi->ckey_alpha, argp,
						sizeof(struct _sColorKeyNAlpha)))
				return -EFAULT;

			pxa168fb_ovly_set_colorkeyalpha(fbi);
			break;
		case FB_IOCTL_GET_COLORKEYnALPHA:
			if (copy_to_user(argp, &fbi->ckey_alpha,
						sizeof(struct _sColorKeyNAlpha)))
				return -EFAULT;
                	break;
		case FB_IOCTL_SWITCH_VID_OVLY:
			if (copy_from_user(&vid_on, argp, sizeof(int)))
				return -EFAULT;
			if (0 == vid_on) {
				if (LCD_Conf.LCD_Mixer[0] !=
							LCD_MIXER_DISABLE) {
					x = readl(fbi->reg_base
							+ LCD_MIXER0_CTL0);
					x |= LCD_MIXERx_CTL0_BP_ID_MASK|
						LCD_MIXERx_CTL0_DISP_UPDATE;
					pxa950fb_writel(x, fbi->reg_base,
							 LCD_MIXER0_CTL0);
				}
				if (LCD_Conf.LCD_Mixer[1] !=
							 LCD_MIXER_DISABLE) {
					x = readl(fbi->reg_base +
							 LCD_MIXER1_CTL0);
					x |= LCD_MIXERx_CTL0_BP_ID_MASK|
						LCD_MIXERx_CTL0_DISP_UPDATE;
					pxa950fb_writel(x, fbi->reg_base,
						 LCD_MIXER1_CTL0);
				}
			} else {
				if (LCD_Conf.LCD_Mixer[0] !=
							 LCD_MIXER_DISABLE) {
					x = readl(fbi->reg_base +
							 LCD_MIXER0_CTL0);
					x &= ~LCD_MIXERx_CTL0_BP_ID_MASK;
					x |= LCD_MIXERx_CTL0_BP_ID(LCD_Conf.BP_ID)|
						 LCD_MIXERx_CTL0_DISP_UPDATE;
					pxa950fb_writel(x, fbi->reg_base,
							 LCD_MIXER0_CTL0);
				}
				if (LCD_Conf.LCD_Mixer[1] !=
							 LCD_MIXER_DISABLE) {
					x = readl(fbi->reg_base +
							 LCD_MIXER1_CTL0);
					x &= ~LCD_MIXERx_CTL0_BP_ID_MASK;
					x |= LCD_MIXERx_CTL0_BP_ID(LCD_Conf.BP_ID)|
						 LCD_MIXERx_CTL0_DISP_UPDATE;
					pxa950fb_writel(x, fbi->reg_base,
							 LCD_MIXER1_CTL0);
				}
			}
			break;

		case FB_IOCTL_SWITCH_GRA_OVLY:
			if (copy_from_user(&gfx_on, argp, sizeof(int))) {
				return -EFAULT;
			}
			if (0 == gfx_on) {
				if (LCD_Conf.LCD_Mixer[0] !=
							 LCD_MIXER_DISABLE) {
					x = readl(fbi->reg_base +
							 LCD_MIXER0_CTL0);
					x |= LCD_MIXERx_CTL0_OL1_ID_MASK |
						 LCD_MIXERx_CTL0_DISP_UPDATE;
					pxa950fb_writel(x, fbi->reg_base,
							 LCD_MIXER0_CTL0);
				}
				if (LCD_Conf.LCD_Mixer[1] !=
							 LCD_MIXER_DISABLE) {
					x = readl(fbi->reg_base +
							 LCD_MIXER1_CTL0);
					x |= LCD_MIXERx_CTL0_OL1_ID_MASK |
						 LCD_MIXERx_CTL0_DISP_UPDATE;
					pxa950fb_writel(x, fbi->reg_base,
							 LCD_MIXER1_CTL0);
				}
			} else {
				if (LCD_Conf.LCD_Mixer[0] !=
							 LCD_MIXER_DISABLE) {
					x = readl(fbi->reg_base +
							 LCD_MIXER0_CTL0);
					x &= ~LCD_MIXERx_CTL0_OL1_ID_MASK;
					x |= LCD_MIXERx_CTL0_OL1_ID(LCD_Conf.OL1_ID)|
						LCD_MIXERx_CTL0_DISP_UPDATE;
					pxa950fb_writel(x, fbi->reg_base,
							 LCD_MIXER0_CTL0);
				}
				if (LCD_Conf.LCD_Mixer[1] !=
							 LCD_MIXER_DISABLE) {
					x = readl(fbi->reg_base +
							 LCD_MIXER1_CTL0);
					x &= ~LCD_MIXERx_CTL0_OL1_ID_MASK;
					x |= LCD_MIXERx_CTL0_OL1_ID(LCD_Conf.OL1_ID)|
						LCD_MIXERx_CTL0_DISP_UPDATE;
					pxa950fb_writel(x, fbi->reg_base,
							 LCD_MIXER1_CTL0);
				}
			}
			break;
		default:
			break;
	}

	return 0;
}

static void collectFreeBuf(u8 **filterList, u8 **freeList, int count)
{
	int i = 0, j = 0;
	struct _sOvlySurface *srf = 0;
	u8 *ptr;

	if ((count < 1) || !filterList || !freeList)
		return;

	for (i = 0, j = 0; i < count; i++) {

		ptr = freeList[i];

		/* Check freeList's content. */
		if (ptr) {
			for (; j < MAX_QUEUE_NUM; j++) {
				if (!(filterList[j])) {
					/* get surface's ptr. */
					srf = (struct _sOvlySurface *)ptr;

					/*
					 * save ptrs which need
					 * to be freed.
					 */
					filterList[j] =
						srf->videoBufferAddr.startAddr;

					break;
				}
			}

			if (j >= MAX_QUEUE_NUM)
				break;

			kfree(freeList[i]);
			freeList[i] = 0;
		} else {
			/* till content is null. */
			break;
		}
	}

	freeList[0] = freeList[count];
	freeList[count] = 0;
}

static int addFreeBuf(u8 **ppBufList, u8 *pFreeBuf)
{
	int i = 0 ;
	struct _sOvlySurface *srf0 = (struct _sOvlySurface *)(pFreeBuf);
	struct _sOvlySurface *srf1 = 0;

	/* Error checking */
	if (!srf0)
		return -1;

	for (; i < MAX_QUEUE_NUM; i++) {
		srf1 = (struct _sOvlySurface *)ppBufList[i];

		if (!srf1) {
			/* printk(KERN_INFO "Set pFreeBuf "
			   "into %d entry.\n", i);
			 */
			ppBufList[i] = pFreeBuf;
			return 0;
		}

		if (srf1->videoBufferAddr.startAddr ==
				srf0->videoBufferAddr.startAddr) {
			/* same address, free this request. */
			kfree(pFreeBuf);
			return 0;
		}
	}


	if (i >= MAX_QUEUE_NUM)
		printk(KERN_INFO "buffer overflow\n");

	return -3;
}

static void clearFreeBuf(u8 **ppBufList, int iFlag)
{
	int i = 0;

	/* Check null pointer. */
	if (!ppBufList)
		return;

	/* free */
	if (FREE_ENTRY & iFlag) {
		for (i = 0; i < MAX_QUEUE_NUM; i++) {
			if (ppBufList && ppBufList[i])
				kfree(ppBufList[i]);
		}
	}

	if (RESET_BUF & iFlag)
		memset(ppBufList, 0, MAX_QUEUE_NUM * sizeof(u8 *));
}

static int pxa950fb_open(struct fb_info *fi, int user)
{
	struct pxa950fb_mach_info *mi;
	struct pxa950fb_info *fbi = (struct pxa950fb_info *)fi->par;
	struct fb_var_screeninfo *var = &fi->var;

	dev_dbg(fi->dev, "Enter %s\n", __FUNCTION__);

	set_dvfm_constraint();

	mi = fbi->dev->platform_data;
	fbi->new_addr = 0;
	fi->fix.smem_start = fbi->fb_start_dma;
	fi->screen_base = fbi->fb_start;
	fbi->surface.videoMode = -1;
	fbi->surface.viewPortInfo.srcWidth = var->xres;
	fbi->surface.viewPortInfo.srcHeight = var->yres;
	fbi->pix_fmt = LCD_Conf.LCD_SRC_FOR;	/* get pixel format from default config */
	set_pix_fmt(var, fbi->pix_fmt);
	pxa950fb_set_par(fi);

	if (mutex_is_locked(&fbi->access_ok))
		mutex_unlock(&fbi->access_ok);

	/* clear buffer list. */
	mutex_lock(&fbi->access_ok);
	clearFreeBuf(filterBufList, RESET_BUF);
	clearFreeBuf(freeBufList, RESET_BUF|FREE_ENTRY);
	mutex_unlock(&fbi->access_ok);

	return 0;
}

static int pxa950fb_release(struct fb_info *fi, int user)
{
	struct pxa950fb_info *fbi = (struct pxa950fb_info *)fi->par;
	struct fb_var_screeninfo *var = &fi->var;

	dev_dbg(fi->dev, "Enter %s\n", __FUNCTION__);
	/* clear buffer list. */
	mutex_lock(&fbi->access_ok);
	clearFreeBuf(filterBufList, RESET_BUF);
	clearFreeBuf(freeBufList, RESET_BUF|FREE_ENTRY);
	mutex_unlock(&fbi->access_ok);

	/* Turn off compatibility mode */

	var->nonstd &= ~0xff000000;

	unset_dvfm_constraint();

	return 0;
}

static int pxa950fb_switch_buff(struct fb_info *fi)
{
	struct pxa950fb_info *fbi = (struct pxa950fb_info *)fi->par;
	int i = 0;
	struct _sOvlySurface *pOvlySurface = 0;

	/*
	 * Find the latest frame.
	 */
	for (i = (MAX_QUEUE_NUM-1); i >= 0; i--) {
		if (freeBufList[i]) {
			pOvlySurface = (struct _sOvlySurface *)freeBufList[i];
			break;
		}
	}

	/*
	 * Got new frame?
	 */
	if (pOvlySurface && (fbi->new_addr !=
				(unsigned long)pOvlySurface->videoBufferAddr.startAddr)) {
		/*
		 * Collect expired frame to list.
		 */
		collectFreeBuf(filterBufList, freeBufList, (i));

		/*
		 * Update new surface.
		 */
		if (check_surface(fi, pOvlySurface->videoMode,
					&pOvlySurface->viewPortInfo,
					&pOvlySurface->viewPortOffset,
					&pOvlySurface->videoBufferAddr))
			pxa950fb_set_par(fi);
	} else {
		/*printk(KERN_INFO "********Oops: pOvlySurface"
		  " is NULL!!!!\n\n");
		 */
		return -1;
	}

	return 0;
}

static void set_mode(struct pxa950fb_info *fbi, struct fb_var_screeninfo *var,
		struct fb_videomode *mode, int pix_fmt, int ystretch)
{
	set_pix_fmt(var, pix_fmt);

	var->xres = mode->xres;
	var->yres = mode->yres;
	var->xres_virtual = max(var->xres, var->xres_virtual);
	if (ystretch)
		var->yres_virtual = fbi->fb_size /
			(var->xres_virtual * (var->bits_per_pixel >> 3));
	else
		var->yres_virtual = max(var->yres, var->yres_virtual);
	var->grayscale = 0;
	var->accel_flags = FB_ACCEL_NONE;
	var->pixclock = mode->pixclock;
	var->left_margin = mode->left_margin;
	var->right_margin = mode->right_margin;
	var->upper_margin = mode->upper_margin;
	var->lower_margin = mode->lower_margin;
	var->hsync_len = mode->hsync_len;
	var->vsync_len = mode->vsync_len;
	var->sync = mode->sync;
	var->vmode = FB_VMODE_NONINTERLACED;
	var->rotate = FB_ROTATE_UR;
}

static int pxa950fb_check_var(struct fb_var_screeninfo *var,
		struct fb_info *info)
{
	struct pxa950fb_info *fbi = info->par;

	if (var->bits_per_pixel == 8)
		return -EINVAL;
	/*
	 * Basic geometry sanity checks.
	 */
	if (var->xoffset + var->xres > var->xres_virtual)
		return -EINVAL;
	if (var->yoffset + var->yres > var->yres_virtual)
		return -EINVAL;
	if (var->xres + var->right_margin +
			var->hsync_len + var->left_margin > 2048)
		return -EINVAL;
	if (var->yres + var->lower_margin +
			var->vsync_len + var->upper_margin > 2048)
		return -EINVAL;

	/*
	 * Check size of framebuffer.
	 */
	if (var->xres_virtual * var->yres_virtual *
			(var->bits_per_pixel >> 3) > fbi->fb_size)
		return -EINVAL;

	return 0;
}

static int wait_for_vsync(struct pxa950fb_info *fbi)
{
	if (fbi) {
		wait_event_interruptible(fbi->w_intr_wq,
				atomic_read(&fbi->w_intr));
		atomic_set(&fbi->w_intr, 0);
		return 0;
	}

	return 0;
}

static void set_graphics_start(struct fb_info *info, int xoffset, int yoffset)
{
	struct pxa950fb_info *fbi = info->par;
	DisplayControllerFrameDescriptor *dmadesc_cpu;
	int pixel_offset;
	struct fb_var_screeninfo *var = &info->var;

	pixel_offset = ((yoffset * var->xres_virtual) + xoffset) * var->bits_per_pixel / 8;
	if (fbi->pix_fmt == 7) {
		dmadesc_cpu = (DisplayControllerFrameDescriptor *)((unsigned int)ovly_dma_addr_cpu - 16*5);
		dmadesc_cpu->LCD_FR_ADDRx = ovly_dma_addr_dma + pixel_offset;
	} else {
		if (fbi->pix_fmt > 3) {
			dmadesc_cpu = (DisplayControllerFrameDescriptor *)((unsigned int)ovly_dma_addr_cpu - 16*5);
			dmadesc_cpu->LCD_FR_ADDRx = ovly_dma_addr_dma + pixel_offset;

			pixel_offset += (var->xres * var->yres);
			dmadesc_cpu = (DisplayControllerFrameDescriptor *)((unsigned int)ovly_dma_addr_cpu - 16*6);
			dmadesc_cpu->LCD_FR_ADDRx = ovly_dma_addr_dma + pixel_offset;

			if (fbi->pix_fmt == PIX_FMTIN_YUV420)
				pixel_offset += (var->xres * var->yres) / 4;
			if (fbi->pix_fmt == PIX_FMTIN_YUV422)
				pixel_offset += (var->xres * var->yres) / 2;
			if (fbi->pix_fmt == PIX_FMTIN_YUV444)
				pixel_offset += (var->xres * var->yres);
			dmadesc_cpu = (DisplayControllerFrameDescriptor *)((unsigned int)ovly_dma_addr_cpu - 16*7);
			dmadesc_cpu->LCD_FR_ADDRx = ovly_dma_addr_dma + pixel_offset;
		} else {
			dmadesc_cpu = (DisplayControllerFrameDescriptor *)((unsigned int)ovly_dma_addr_cpu - 16*4);
			dmadesc_cpu->LCD_FR_ADDRx = ovly_dma_addr_dma + pixel_offset;
		}
	}
	return;
}

static void dump_regs_ovly(struct pxa950fb_info *fbi)
{
	if (fbi->pix_fmt == 7) {
		printk("LCD_NXT_DESC_ADDR4 = %x\n", readl(fbi->reg_base + LCD_NXT_DESC_ADDR4));
		printk("LCD_FETCH_CTL4 = %x\n", readl(fbi->reg_base + LCD_FETCH_CTL4));
		printk("LCD_WIN4_CTL = %x\n", readl(fbi->reg_base + LCD_WIN4_CTL));
		printk("LCD_WINx_CFG = %x\n", readl(fbi->reg_base + LCD_WIN4_CFG));
		printk("LCD_CH4_ALPHA = %x\n", readl(fbi->reg_base + LCD_CH4_ALPHA));
	} else {
		if (fbi->pix_fmt > 3) {
			printk("LCD_NXT_DESC_ADDR4 = %x\n", readl(fbi->reg_base + LCD_NXT_DESC_ADDR4));
			printk("LCD_NXT_DESC_ADDR5 = %x\n", readl(fbi->reg_base + LCD_NXT_DESC_ADDR5));
			printk("LCD_NXT_DESC_ADDR6 = %x\n", readl(fbi->reg_base + LCD_NXT_DESC_ADDR6));
			printk("LCD_FETCH_CTL4 = %x\n", readl(fbi->reg_base + LCD_FETCH_CTL4));
			printk("LCD_FETCH_CTL5 = %x\n", readl(fbi->reg_base + LCD_FETCH_CTL5));
			printk("LCD_FETCH_CTL6 = %x\n", readl(fbi->reg_base + LCD_FETCH_CTL6));
			printk("LCD_WIN4_CTL = %x\n", readl(fbi->reg_base + LCD_WIN4_CTL));
			printk("LCD_WIN4_CFG = %x\n", readl(fbi->reg_base + LCD_WIN4_CFG));
			printk("LCD_CH4_ALPHA = %x\n", readl(fbi->reg_base + LCD_CH4_ALPHA));
		} else {
			printk("LCD_NXT_DESC_ADDR3 = %x\n", readl(fbi->reg_base + LCD_NXT_DESC_ADDR3));
			printk("LCD_FETCH_CTL3 = %x\n", readl(fbi->reg_base + LCD_FETCH_CTL3));
			printk("LCD_WIN3_CTL = %x\n", readl(fbi->reg_base + LCD_WIN3_CTL));
			printk("LCD_WIN3_CFG = %x\n", readl(fbi->reg_base + LCD_WIN3_CFG));
			printk("LCD_CH3_ALPHA = %x\n", readl(fbi->reg_base + LCD_CH3_ALPHA));
		}
	}
	printk("LCD_MIXER0_OL1_CFG0 = %x\n", readl(fbi->reg_base + LCD_MIXER0_OL1_CFG0));
	printk("LCD_MIXER0_OL1_CFG1 = %x\n", readl(fbi->reg_base + LCD_MIXER0_OL1_CFG1));
	printk("LCD_MIXER0_CTL0 = %x\n", readl(fbi->reg_base + LCD_MIXER0_CTL0));
}

static int pxa950fb_set_par(struct fb_info *info)
{
	struct pxa950fb_info *fbi = info->par;
	struct fb_var_screeninfo *var = &info->var;
	int pix_fmt;
	/*
	 * Determine which pixel format we're going to use.
	 */
	pix_fmt = determine_best_pix_fmt(&info->var);
	if (pix_fmt < 0)
		return pix_fmt;
	fbi->pix_fmt = pix_fmt;

	/*
	 * Set additional mode info.
	 */
	if (pix_fmt == PIX_FMT_PSEUDOCOLOR)
		info->fix.visual = FB_VISUAL_PSEUDOCOLOR;
	else
		info->fix.visual = FB_VISUAL_TRUECOLOR;
	info->fix.line_length = var->xres_virtual * var->bits_per_pixel / 8;
	/* info->fix.ypanstep = var->yres; */
	/* set fetch registers */
	if (fbi->pix_fmt == 7)
		set_fetch(fbi, var, 4);
	else {
		if (fbi->pix_fmt > 3) {
			set_fetch(fbi, var, 4);
			set_fetch(fbi, var, 5);
			set_fetch(fbi, var, 6);
		} else {
			set_fetch(fbi, var, 3);
		}
	}
	if (fbi->pix_fmt > 3)
		LCD_Conf.BP_ID = LCD_Controller_WGe4;
	else
		LCD_Conf.BP_ID = LCD_Controller_WGe3;
	/* set window registers */
	set_window(fbi, var);		/* must behind of OV1_ID setting */
	/* set mixer registers */
	set_mixer(fbi, var, LCD_Conf.LCD_Overlay);
	/* configure graphics dma */
	set_graphics_start(info, info->var.xoffset, info->var.yoffset);
	/* write registers setting and enable controller */
	pxa950fb_enable_controller(fbi);

	info->fix.smem_len = var->xres_virtual * var->yres_virtual * var->bits_per_pixel/8;
	info->screen_size = info->fix.smem_len;


	return 0;
}

static unsigned int chan_to_field(unsigned int chan, struct fb_bitfield *bf)
{
	return ((chan & 0xffff) >> (16 - bf->length)) << bf->offset;
}

static u32 to_rgb(u16 red, u16 green, u16 blue)
{
	red >>= 8;
	green >>= 8;
	blue >>= 8;

	return (red << 16) | (green << 8) | blue;
}

	static int
pxa950fb_setcolreg(unsigned int regno, unsigned int red, unsigned int green,
		unsigned int blue, unsigned int trans, struct fb_info *info)
{
	struct pxa950fb_info *fbi = info->par;
	u32 val;

	if (info->fix.visual == FB_VISUAL_TRUECOLOR && regno < 16) {
		val =  chan_to_field(red,   &info->var.red);
		val |= chan_to_field(green, &info->var.green);
		val |= chan_to_field(blue , &info->var.blue);
		fbi->pseudo_palette[regno] = val;
	}

	if (info->fix.visual == FB_VISUAL_PSEUDOCOLOR && regno < 256) {
		val = to_rgb(red, green, blue);
		/* TODO */
	}

	return 0;
}

static int pxa950fb_blank(int blank, struct fb_info *info)
{
	struct pxa950fb_info *fbi = info->par;

	fbi->is_blanked = (blank == FB_BLANK_UNBLANK) ? 0 : 1;

	return 0;	/* TODO */
}

static int pxa950fb_pan_display(struct fb_var_screeninfo *var,
		struct fb_info *info)
{
	set_graphics_start(info, var->xoffset, var->yoffset);

	return 0;
}

static int pxa950fb_fb_sync(struct fb_info *info)
{
	struct pxa950fb_info *fbi = (struct pxa950fb_info *)info->par;

	return wait_for_vsync(fbi);
}

static irqreturn_t pxa950fb_handle_irq(int irq, void *dev_id)
{
	struct pxa950fb_info *fbi = (struct pxa950fb_info *)dev_id;
	u32     isr;
	int 	index;

	for (index = 0; index < 2; index++) {
		if ((index == 0) && (LCD_Conf.LCD_Mixer[0] == LCD_MIXER_DISABLE))
			continue;
		if ((index == 1) && (LCD_Conf.LCD_Mixer[1] == LCD_MIXER_DISABLE))
			continue;
		isr = readl(fbi->reg_base + 
				1000 * ((index == 0) ? LCD_Conf.LCD_Mixer[0] : LCD_Conf.LCD_Mixer[1]) +
				LCD_DSI_DxINST0_OFFSET);
		writel(isr, fbi->reg_base +
				1000 * ((index == 0) ? LCD_Conf.LCD_Mixer[0] : LCD_Conf.LCD_Mixer[1]) +
				LCD_DSI_DxINST0_OFFSET);
		/* FIXME: currently only support one mixer */
		if (isr & LCD_CONVx_CTL_DISP_EOF_INT_EN) {
			/* wake up queue. */
			atomic_set(&fbi->w_intr, 1);
			wake_up(&fbi->w_intr_wq);
			tasklet_schedule(&fbi->tasklet);
		}
	}
	return IRQ_HANDLED;
}

static struct fb_ops pxa950fb_ops = {
	.owner		= THIS_MODULE,
	.fb_open        = pxa950fb_open,
	.fb_release     = pxa950fb_release,

	.fb_check_var	= pxa950fb_check_var,
	.fb_set_par	= pxa950fb_set_par,
	.fb_setcolreg	= pxa950fb_setcolreg,	/* TODO */
	.fb_blank	= pxa950fb_blank,
	.fb_pan_display	= pxa950fb_pan_display,
	.fb_fillrect	= cfb_fillrect,
	.fb_copyarea	= cfb_copyarea,
	.fb_imageblit	= cfb_imageblit,
	.fb_sync        = pxa950fb_fb_sync,
	.fb_ioctl       = pxa950fb_ioctl,
};

static int __init pxa950fb_probe(struct platform_device *pdev)
{
	struct pxa950fb_mach_info *mi;
	struct fb_info *info = 0;
	struct pxa950fb_info *fbi = 0;
	struct resource *res;
	int irq, ret;

	mi = pdev->dev.platform_data;
	if (mi == NULL) {
		dev_err(&pdev->dev, "no platform data defined\n");
		return -EINVAL;
	}

	res = platform_get_resource(pdev, IORESOURCE_MEM, 0);
	if (res == NULL) {
		dev_err(&pdev->dev, "no IO memory defined\n");
		return -ENOENT;
	}

	irq = platform_get_irq(pdev, 0);
	if (irq < 0) {
		dev_err(&pdev->dev, "no IRQ defined\n");
		return -ENOENT;
	}

	info = framebuffer_alloc(sizeof(struct pxa950fb_info), &pdev->dev);
	if (info == NULL)
		return -ENOMEM;

	/* Initialize private data */
	fbi = info->par;
	platform_set_drvdata(pdev, fbi);
	fbi->fb_info = info;
	fbi->dev = &pdev->dev;
	fbi->id = pdev->id;
	fbi->panel_rbswap = mi->panel_rbswap;
	fbi->is_blanked = 0;
	fbi->debug = 0;
	fbi->active = mi->active;

	init_waitqueue_head(&fbi->w_intr_wq);

	/*
	 * Initialise static fb parameters.
	 */
	info->flags = FBINFO_DEFAULT | FBINFO_PARTIAL_PAN_OK |
		FBINFO_HWACCEL_XPAN | FBINFO_HWACCEL_YPAN;
	info->node = -1;
	strcpy(info->fix.id, mi->id);
	info->fix.type = FB_TYPE_PACKED_PIXELS;
	info->fix.type_aux = 0;
	info->fix.xpanstep = 1;
	info->fix.ypanstep = 1;
	info->fix.ywrapstep = 0;
	info->fix.mmio_start = res->start;
	info->fix.mmio_len = res->end - res->start + 1;
	info->fix.accel = FB_ACCEL_NONE;
	info->fbops = &pxa950fb_ops;
	info->pseudo_palette = fbi->pseudo_palette;

	/*
	 * Map LCD controller registers.
	 */
	fbi->reg_base = ioremap_nocache(res->start, res->end - res->start);
	if (fbi->reg_base == NULL) {
		ret = -ENOMEM;
		goto failed;
	}
	/*
	 * Allocate framebuffer memory.
	 */
	fbi->fb_size = PAGE_ALIGN(DEFAULT_FB_SIZE);

	fbi->fb_start = dma_alloc_writecombine(fbi->dev, fbi->fb_size + PAGE_SIZE,
			&fbi->fb_start_dma,
			GFP_KERNEL);
	if (fbi->fb_start == NULL) {
		ret = -ENOMEM;
		goto failed;
	}
	memset(fbi->fb_start, 0x0, fbi->fb_size);
	fbi->fb_start = fbi->fb_start + PAGE_SIZE;
	fbi->fb_start_dma = fbi->fb_start_dma + PAGE_SIZE;

	ovly_dma_addr_dma = fbi->fb_start_dma;
	ovly_dma_addr_cpu = fbi->fb_start;
	info->fix.smem_start = fbi->fb_start_dma;
	info->fix.smem_len = fbi->fb_size;
	info->screen_base = fbi->fb_start;
	info->screen_size = fbi->fb_size;

	/* save machine info to LCD_Conf */
	set_configuration(fbi);

	/*
	 * Set video mode according to platform data.
	 */
	set_mode(fbi, &info->var, mi->modes, LCD_Conf.LCD_SRC_FOR, 1);

	/* read register value as default */
	pxa950fb_ovly_regs.fetchregs[4].LCD_FETCH_CTLx = readl(fbi->reg_base + LCD_FETCH_CTL4);
	pxa950fb_ovly_regs.fetchregs[5].LCD_FETCH_CTLx = readl(fbi->reg_base + LCD_FETCH_CTL5);
	pxa950fb_ovly_regs.fetchregs[6].LCD_FETCH_CTLx = readl(fbi->reg_base + LCD_FETCH_CTL6);

	pxa950fb_ovly_regs.wgregs[4].LCD_WINx_CTL = readl(fbi->reg_base + LCD_WIN4_CTL);

	pxa950fb_ovly_regs.mixerregs[0].LCD_MIXERx_CTL0 = readl(fbi->reg_base + LCD_MIXER0_CTL0);
	pxa950fb_ovly_regs.mixerregs[0].LCD_MIXERx_CTL1 = readl(fbi->reg_base + LCD_MIXER0_CTL1);
	pxa950fb_ovly_regs.mixerregs[0].LCD_MIXERx_CTL2 = readl(fbi->reg_base + LCD_MIXER0_CTL2);
	pxa950fb_ovly_regs.mixerregs[0].LCD_MIXERx_OL1_CFG0 = readl(fbi->reg_base + LCD_MIXER0_OL1_CFG0);
	pxa950fb_ovly_regs.mixerregs[0].LCD_MIXERx_OL1_CFG1 = readl(fbi->reg_base + LCD_MIXER0_OL1_CFG1);

	pxa950fb_ovly_regs.mixerregs[0].LCD_MIXERx_BP_CFG0 = readl(fbi->reg_base + LCD_MIXER0_BP_CFG0);
	pxa950fb_ovly_regs.mixerregs[0].LCD_MIXERx_BP_CFG1 = readl(fbi->reg_base + LCD_MIXER0_BP_CFG1);

	pxa950fb_ovly_regs.mixerregs[1].LCD_MIXERx_CTL0 = readl(fbi->reg_base + LCD_MIXER1_CTL0);
	pxa950fb_ovly_regs.mixerregs[1].LCD_MIXERx_CTL1 = readl(fbi->reg_base + LCD_MIXER1_CTL1);
	pxa950fb_ovly_regs.mixerregs[1].LCD_MIXERx_CTL2 = readl(fbi->reg_base + LCD_MIXER1_CTL2);
	pxa950fb_ovly_regs.mixerregs[1].LCD_MIXERx_OL1_CFG0 = readl(fbi->reg_base + LCD_MIXER1_OL1_CFG0);
	pxa950fb_ovly_regs.mixerregs[1].LCD_MIXERx_OL1_CFG1 = readl(fbi->reg_base + LCD_MIXER1_OL1_CFG1);
	pxa950fb_ovly_regs.mixerregs[1].LCD_MIXERx_BP_CFG0 = readl(fbi->reg_base + LCD_MIXER1_BP_CFG0);
	pxa950fb_ovly_regs.mixerregs[1].LCD_MIXERx_BP_CFG1 = readl(fbi->reg_base + LCD_MIXER1_BP_CFG1);

	pxa950fb_ovly_regs.convregs.LCD_CONVx_CTL = readl(fbi->reg_base + LCD_CONV0_CTL);

	/*
	 * Allocate color map.
	 */
	if (fb_alloc_cmap(&info->cmap, 256, 0) < 0) {
		ret = -ENOMEM;
		goto failed_free_clk;
	}

	/*
	 * Register irq handler.
	 */
	ret = request_irq(irq, pxa950fb_handle_irq, IRQF_SHARED, mi->id, fbi);
	if (ret < 0) {
		dev_err(&pdev->dev, "unable to request IRQ\n");
		ret = -ENXIO;
		goto failed_free_cmap;
	}

	fbi->tasklet.next = NULL;
	fbi->tasklet.state = 0;
	atomic_set(&fbi->tasklet.count, 0);
	fbi->tasklet.func = pxa950fb_do_tasklet;
	fbi->tasklet.data = (unsigned long)info;

	pxa950fb_writel(0x800000ff, fbi->reg_base, LCD_CH3_ALPHA);
	pxa950fb_writel(0x800000ff, fbi->reg_base, LCD_CH4_ALPHA);

	if (LCD_Conf.LCD_Mixer[0] == LCD_M2PARALELL_CONVERTER) {
		pxa950fb_ovly_regs.convregs.LCD_CONVx_CTL |= LCD_CONVx_CTL_DISP_EOF_INT_EN;
		pxa950fb_writel(pxa950fb_ovly_regs.convregs.LCD_CONVx_CTL, fbi->reg_base, LCD_CONV0_CTL);
	} else if (LCD_Conf.LCD_Mixer[0] == LCD_M2DSI0)
		writel(LCD_DSI_DxSCR0_DISP_EOF_INT_EN, fbi->reg_base +
				1000 * LCD_Conf.LCD_Mixer[0] +
				LCD_DSI_DxINEN_OFFSET);

	if (LCD_Conf.LCD_Mixer[1] != LCD_M2DSI1)
		writel(LCD_DSI_DxSCR0_DISP_EOF_INT_EN, fbi->reg_base +
				1000 * LCD_Conf.LCD_Mixer[1] +
				LCD_DSI_DxINEN_OFFSET);

	/*
	 * Register framebuffer.
	 */
	ret = register_framebuffer(info);
	if (ret < 0) {
		dev_err(&pdev->dev, "Failed to register pxa950-fb: %d\n", ret);
		ret = -ENXIO;
		goto failed_free_irq;
	}
	printk(KERN_INFO "pxa950fb_ovly: frame buffer device was loaded"
			" to /dev/fb%d <%s>.\n", info->node, info->fix.id);

	return 0;

failed_free_irq:
	free_irq(irq, fbi);
failed_free_cmap:
	fb_dealloc_cmap(&info->cmap);
failed_free_clk:
failed:
	pr_err("pxa950-fb: frame buffer device init failed\n");
	platform_set_drvdata(pdev, NULL);
	fb_dealloc_cmap(&info->cmap);

	if (fbi && fbi->reg_base) {
		iounmap(fbi->reg_base);
		kfree(fbi);
	}

	return ret;
}

static struct platform_driver pxa950fb_driver = {
	.driver		= {
		.name	= "pxa950fb-ovly",
		.owner	= THIS_MODULE,
	},
	.probe		= pxa950fb_probe,
};

static int __devinit pxa950fb_init(void)
{
#ifdef CONFIG_PXA3xx_DVFM
	dvfm_register("pxa950fb-ovly", &dvfm_dev_idx);
#endif
	return platform_driver_register(&pxa950fb_driver);
}
/* module_init(pxa950fb_init); */
late_initcall(pxa950fb_init);

MODULE_AUTHOR("Lennert Buytenhek <buytenh@marvell.com>");
MODULE_DESCRIPTION("Framebuffer driver for PXA950");
MODULE_LICENSE("GPL");
