/*
 * drivers/w1/masters/micco_w1.c
 *
 * Copyright @ 2009 Borqs Ltd.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include <linux/module.h>
#include <linux/types.h>
#include <linux/errno.h>
#include <linux/delay.h>
#include <linux/slab.h>
#include <mach/micco.h>
#include <mach/pxa3xx_pmic.h>

#include "../w1.h"
#include "../w1_int.h"

static int W1_ID;

struct micco_w1_data {
	u8 w1_reg_addr;
	int w1_addr_data;	/* 0-write addr, 1-write data */
	struct completion rx_done;
	int rx_timeout;
};

static struct micco_w1_data * w1_data = NULL;

static void micco_w1_interrupt(unsigned long event)
{
	if(event & (PMIC_EVENT_RX_COMP | PMIC_EVENT_TIMEOUT))
	{
		if( event & PMIC_EVENT_TIMEOUT )
			w1_data->rx_timeout = 1;
		complete(&w1_data->rx_done);
	}
}

/* One-wire uart write 
 * u8 data: data being transfered 
 * u8 addr: register address of the device	
 */
static void micco_w1_write_byte(void *data, u8 byte)
{
	struct micco_w1_data *wd = (struct micco_w1_data *)data;
	u8 cntl;
	cntl = 0x07;	/*write and reset pulse and send the command data */

	if (!wd->w1_addr_data) {
		wd->w1_addr_data = 1;
		wd->w1_reg_addr = byte;
		return;
	}
	/* Set target address */
	micco_write(MICCO_UART_ADDR, wd->w1_reg_addr);

	/* Fill transition register with data */
	micco_write(MICCO_UART_TX_DATA, byte);

	/* Send command */
	micco_write(MICCO_UART_CONTROL, cntl);

	wd->w1_addr_data = 0;
}

/* One-wire uart read 
 * u8 addr: register address of the device
 */
static u8 micco_w1_read_byte(void *data)
{
	struct micco_w1_data *wd = (struct micco_w1_data *)data;
	unsigned long timeout = 0;
	u8 cntl = 0x05;	/*read and go, and send the command data */
	u8 val = 0, tmp;
	int index = 0;

RETRY:
	INIT_COMPLETION(wd->rx_done);
	wd->rx_timeout=0;

	/* Set target address */
	micco_write(MICCO_UART_ADDR, wd->w1_reg_addr);

	/* Send control command */
	micco_write(MICCO_UART_CONTROL, cntl);

	/* if timeout == 0, means no response for 1 second
	 * rx_timeout == 1, means PMIC_TIMEOUT event happens
 	 */
	timeout = wait_for_completion_timeout(&(wd->rx_done), HZ>>1);
	val = pxa3xx_pmic_get_rx_value(); 

	if (timeout == 0 || wd->rx_timeout) {
		printk(KERN_WARNING "++++++++%s:%d timeout = %lu value = %d\n", __func__, index, timeout, val);
		if (3 > index++){
			micco_write(MICCO_UART_CONTROL, 0);
			msleep(50);
			micco_write(MICCO_UART_CONTROL, 0x4);
			msleep(50);
			micco_read(MICCO_UART_CONTROL, &tmp);
			if( tmp != 0 )
				printk(KERN_WARNING"%s:Fail to reset 1-wire\n",__func__);
			msleep(50);
			goto RETRY;
		}
	}

	wd->w1_addr_data = 0;
	return val;
}

static u8 micco_w1_reset(void *data)
{
	struct micco_w1_data *wd = (struct micco_w1_data *)data;
	wd->w1_addr_data = 0;
	return micco_write(MICCO_UART_CONTROL, 0x04);
}

static void micco_w1_search(void *data, struct w1_master *master_dev,
		u8 search_type, w1_slave_found_callback slave_found)
{
	u64 module_id, rn_le, cs, id;

	if (W1_ID)
		module_id = W1_ID;
	else
		module_id = 0x1;

	rn_le = cpu_to_le64(module_id);

	/* HDQ might not obely truly the 1-wire spec.
	 * So calculate CRC based on module parameter.
	 */
	cs = w1_calc_crc8((u8 *) & rn_le, 7);
	id = (cs << 56) | module_id;

	slave_found(master_dev, id);
	//slave_found(data, id);
}

static struct w1_bus_master micco_w1_bus_master = {
	.read_byte = micco_w1_read_byte,
	.write_byte = micco_w1_write_byte,
	.reset_bus = micco_w1_reset,
	.search = micco_w1_search,
};

static int __devinit micco_w1_init(void)
{
	u8 val;

	w1_data = kzalloc(sizeof(struct micco_w1_data), GFP_KERNEL);
	if (!w1_data) {
		pr_err("micco_w1_init:"
			"Failed to allocate device info structure\n");
		return -ENOMEM;
	}

	micco_w1_reset((void *)w1_data);

	/* enable micco uart interrupts for w1-battery IC */
	micco_read(MICCO_IRQ_MASK_D, &val);
	val &= ~(IRQ_MASK_D_UART_TX |
		 IRQ_MASK_D_UART_RX | IRQ_MASK_D_UART_TIMEOUT);
	micco_write(MICCO_IRQ_MASK_D, val);

	init_completion(&(w1_data->rx_done));

	pmic_callback_register(PMIC_EVENT_TX_COMP | PMIC_EVENT_RX_COMP |
			       PMIC_EVENT_TIMEOUT, micco_w1_interrupt);

	micco_w1_bus_master.data = (void *)w1_data;
	return w1_add_master_device(&micco_w1_bus_master);
}

static void __devexit micco_w1_fini(void)
{
	pmic_callback_unregister(PMIC_EVENT_TX_COMP | PMIC_EVENT_RX_COMP |
				 PMIC_EVENT_TIMEOUT, micco_w1_interrupt);

	w1_remove_master_device(&micco_w1_bus_master);
}

module_init(micco_w1_init);
module_exit(micco_w1_fini);

module_param(W1_ID, int, S_IRUSR);

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Borqs Ltd");
MODULE_DESCRIPTION("Micco HDQ/1-Wire master driver");
