/*
 * kernel/power/pm_check.c 
 *
 * Copyright (C) 2008 Borqs Ltd.
 * Author:  Emichael Li <emichael.li@borqs.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/errno.h>
#include <linux/pm.h>

#include "power.h"

#define PM_CHECK_RUNNING_INTERVAL 20000

static int pm_check_shown;

static spinlock_t g_pmc_list_lock = SPIN_LOCK_UNLOCKED;
static LIST_HEAD(g_pmc_list);
static struct timer_list g_pm_check_timer;

int pm_check_register(struct pm_check *pmc)
{
	unsigned long irqflags;

	spin_lock_irqsave(&g_pmc_list_lock, irqflags);
	list_add(&pmc->link, &g_pmc_list);
	spin_unlock_irqrestore(&g_pmc_list_lock, irqflags);
	return 0;
}

int pm_check_unregister(struct pm_check *pmc)
{
	unsigned long irqflags;

	spin_lock_irqsave(&g_pmc_list_lock, irqflags);
	list_del(&pmc->link);
	spin_unlock_irqrestore(&g_pmc_list_lock, irqflags);
	return 0;
}

int pm_check_standby(void)
{
	struct list_head *pos;
	struct pm_check *e;
	unsigned long irqflags;

	spin_lock_irqsave(&g_pmc_list_lock, irqflags);

	if (pm_check_shown) {
		list_for_each(pos, &g_pmc_list) {
			e = list_entry(pos, struct pm_check, link);
			if (e->standby) {
				if (!e->standby(e)) {
					e->standby_ok++;
				}
				e->standby_tm++;
			}
		}
	}

	spin_unlock_irqrestore(&g_pmc_list_lock, irqflags);
	return 0;
}

static void __attribute__ ((unused)) pm_check_running_timer_handler(unsigned long data)
{
	struct list_head *pos;
	struct pm_check *e;
	unsigned long irqflags;

	spin_lock_irqsave(&g_pmc_list_lock, irqflags);

	if (pm_check_shown) {
		list_for_each(pos, &g_pmc_list) {
			e = list_entry(pos, struct pm_check, link);
			if (e->running) {
				if (!e->running(e)) {
					e->running_ok++;
				}
				e->running_tm++;
			}
		}
	}

	spin_unlock_irqrestore(&g_pmc_list_lock, irqflags);
	mod_timer(&g_pm_check_timer, jiffies + msecs_to_jiffies(PM_CHECK_RUNNING_INTERVAL));
}

ssize_t pm_check_show(struct kobject *kobj, struct kobj_attribute *attr, char *buf)
{
	char *s;
	struct list_head *pos;
	struct pm_check *e;
	unsigned long irqflags;

	s = buf;

	spin_lock_irqsave(&g_pmc_list_lock, irqflags);

	s += sprintf(s, "%d\n", pm_check_shown);

	list_for_each(pos, &g_pmc_list) {
		e = list_entry(pos, struct pm_check, link);
		s += sprintf(s, "%12d %12d %12d %12d %12d %12d %s\n", 
				e->running_tm - e->running_ok,
				e->standby_tm - e->standby_ok,
				e->running_tm,
				e->running_ok,
				e->standby_tm,
				e->standby_ok,
				e->name);
	}

	spin_unlock_irqrestore(&g_pmc_list_lock, irqflags);

	return s - buf;
}

ssize_t pm_check_store(struct kobject *kobj, struct kobj_attribute *attr, const char *buf, size_t n)
{
	int val;
	struct list_head *pos;
	struct pm_check *e;
	unsigned long irqflags;

	if (sscanf(buf, "%d", &val) == 1) {
		spin_lock_irqsave(&g_pmc_list_lock, irqflags);
		pm_check_shown = !!val;

		if (pm_check_shown) {
			list_for_each(pos, &g_pmc_list) {
				e = list_entry(pos, struct pm_check, link);
				e->running_ok = 0;
				e->running_tm = 0;
				e->standby_ok = 0;
				e->standby_tm = 0;
			}
		}
		spin_unlock_irqrestore(&g_pmc_list_lock, irqflags);	
		return n;
	}
	return -EINVAL;
}
